# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os

from Library.Util.Config import Config

from qtcanvas import QCanvasRectangle
from qt import QBrush, Qt, QObject, SIGNAL, QFontDialog, QColorDialog, QFont
from qt import QColor
from qtui import QWidgetFactory

class CanvasComment(QCanvasRectangle):
    def __init__(self, canvas_view, canvas_comment_data, created_on_user_interaction):
        """Creates a canvas comment
        
        parameters:
           *canvas_view: of type QCanvasView
           *canvas_comment_data: of type CanvasCommentData
           *created_from_mouse_event: of type bool
        """
        
        self._settings_dialog = None
        self._border_colour = Qt.black
        self._canvas_view = canvas_view
        self._canvas = self._canvas_view.canvas()
        self._canvas_comment_data = canvas_comment_data

        xpos, ypos = self._canvas_comment_data.get_pos()
        QCanvasRectangle.__init__(self, int(xpos), int(ypos), 120, 40, self._canvas)

        if created_on_user_interaction:
            self._update_start_position()

        self.show()


    def _update_start_position(self):
        """This method should only be called when a new comment is created by
        adding it using the comment tool. The middle of the comment should be
        located where user pressed the mouse button"""
        self.redraw()
        new_x = self.x() - int(self.width()/2)
        new_y = self.y() - int(self.height()/2)
        self._canvas_comment_data.set_pos((int(new_x), int(new_y)))
        QCanvasRectangle.move(self, int(new_x), int(new_y))

    def get_canvas_comment_data(self):
        return self._canvas_comment_data

        
    def move(self, dx, dy):
        new_x = self.x() + dx
        new_y = self.y() + dy
        self._canvas_comment_data.set_pos((int(new_x), int(new_y)))
        QCanvasRectangle.move(self, int(new_x), int(new_y))

        
    def redraw(self):
        """Updates the position and contents of this object"""

        self.update()
        self._canvas.setAllChanged()
        self._canvas.update()


    def delete(self):
        self.setCanvas(None)
        self._canvas.update()
        self._canvas_view.remove_canvas_comment(self)        


    def drawShape(self, painter):
        """
        Draws this object. This method implements the interface of
        QCanvasRectangle. It'll be called whenever the canvasview is updated.
        """

        text = self._canvas_comment_data.get_text()
        
        x = y = w = h = 0
        painter.setFont(self._canvas_comment_data.get_font())
        painter.setPen(self._canvas_comment_data.get_font_colour())
        background_brush = QBrush(self._canvas_comment_data.get_background_colour())
        rect = painter.boundingRect(x, y, w, h, 0, text)
        
        padding = 10
        border = 2

        self.setSize(rect.width() + 2 * (padding + border),
                     rect.height() + 2 * (padding + border))

        painter.fillRect(self.x(), self.y(), self.width(), self.height(),
                         QBrush(self._border_colour))

        painter.fillRect(self.x() + border, self.y() + border,
                         self.width() - 2*border, self.height() - 2*border,
                         background_brush)

        painter.drawText(self.x() + padding + border,
                         self.y() + padding + border,
                         rect.width(), rect.height(), 0, text)


    def set_highlight(self, value=True):
        if value:
            self._border_colour = Qt.blue
        else:
            self._border_colour = Qt.black
        self.redraw()

    def comment_dialog(self):
        self._settings_dialog = CanvasCommentSettingsDialog(self._canvas_view,
                                                            self)
        self._settings_dialog.show()


class CanvasCommentSettingsDialog:
        def __init__(self, canvas_view, canvas_comment):
            """Creates a settings dialog"""

            self._canvas_comment = canvas_comment
            self._canvas_comment_data = canvas_comment.get_canvas_comment_data()
            self._canvas_view = canvas_view
            self._widget = QWidgetFactory.create(os.path.join(Config.uidir, 'comment_settings.ui'),
                                                 None, canvas_view)
            self._widget.reparent(canvas_view, Qt.WType_Dialog | Qt.WShowModal,
                                  self._widget.pos(), False)

            # button connections
            QObject.connect(self._widget.child('push_button_ok'),
                            SIGNAL('clicked()'), self._slot_ok_pressed)
            QObject.connect(self._widget.child('push_button_cancel'),
                            SIGNAL('clicked()'), self._slot_cancel_pressed)


            # action menu connections
            action_edit_font = self._widget.child('action_edit_font')
            action_edit_font_colour = self._widget.child('action_edit_font_colour')
            action_edit_background_colour = self._widget.child('action_edit_background_colour')

            QObject.connect(action_edit_font,
                            SIGNAL('activated()'), self._slot_edit_font_pressed)
            QObject.connect(action_edit_font_colour,
                            SIGNAL('activated()'), self._slot_edit_font_colour_pressed)
            QObject.connect(action_edit_background_colour,
                            SIGNAL('activated()'), self._slot_edit_background_colour_pressed)


            # other connections
            self._text_edit_comment_text = self._widget.child('text_edit_comment_text')
            self._text_label_text_preview = self._widget.child('text_label_text_preview')
            self._frame_preview = self._widget.child('frame_preview')

            self._new_font               = None
            self._new_font_colour        = None
            self._new_background_colour  = None

            self._text_edit_comment_text.setText(self._canvas_comment_data.get_text())
            self._text_label_text_preview.setFont(self._canvas_comment_data.get_font())
            self._text_label_text_preview.setPaletteForegroundColor(self._canvas_comment_data.get_font_colour())
            self._text_label_text_preview.setPaletteBackgroundColor(self._canvas_comment_data.get_background_colour())
            self._frame_preview.setPaletteBackgroundColor(self._canvas_comment_data.get_background_colour())


        def _slot_edit_font_pressed(self):
            previous_font = self._canvas_comment_data.get_font()
            results = QFontDialog.getFont(previous_font, self._widget)
            font = results[0]
            ok_pressed = results[1]

            if ok_pressed:
                self._new_font = font
                self._text_label_text_preview.setFont(font)                

                
        def _slot_edit_font_colour_pressed(self):
            previous_font_colour = self._canvas_comment_data.get_font_colour()
            colour = QColorDialog.getColor(previous_font_colour, self._widget)

            if colour.isValid():
                self._new_font_colour = colour
                self._text_label_text_preview.setPaletteForegroundColor(colour)

                
        def _slot_edit_background_colour_pressed(self):
            previous_background_colour = self._canvas_comment_data.get_background_colour()
            colour = QColorDialog.getColor(previous_background_colour, self._widget)

            if colour.isValid():
                self._new_background_colour = colour
                self._text_label_text_preview.setPaletteBackgroundColor(colour)
                self._frame_preview.setPaletteBackgroundColor(colour)


        def _slot_ok_pressed(self):
            self._canvas_comment_data.set_text(self._text_edit_comment_text.text())
            if self._new_font:
                self._canvas_comment_data.set_font(self._new_font)
            if self._new_font_colour:
                self._canvas_comment_data.set_font_colour(self._new_font_colour)
            if self._new_background_colour:
                self._canvas_comment_data.set_background_colour(self._new_background_colour)
            self._canvas_comment.redraw()
            self._widget.close()


        def _slot_cancel_pressed(self):
            self._widget.close()


        def show(self):
            self._widget.show()


class CanvasCommentData:
    
    def __init__(self, text="", xpos=0, ypos=0, font=None, font_colour=None,
                 background_colour=None):
        """Contains data for a CanvasComment.

        parameters:
           * text: string that should be displayed on the comment
           * xpos: integer describing the x-position of the comment
           * ypos: integer describing the y-position of the comment
           * font: QFont describing the font the text should have
           * font_colour:       QColor describing the colour of the text
           * background_colour: QColor describing the background colour of the comment
        """

        self._text = text
        self._xpos = xpos
        self._ypos = ypos           

        if font:
            self._font = font
        else:
            #creates the default font used by the application
            self._font = QFont()
            self._font.setUnderline(False)
            self._font.setStrikeOut(False)

        if font_colour:
            self._font_colour = font_colour
        else:
            self._font_colour = QColor(0, 0, 0)

        if background_colour:
            self._background_colour = background_colour
        else:
            self._background_colour = QColor(255, 255, 255)

    def from_xml_element(self, root):
        self.set_pos((int(root.get("xpos")),
                      int(root.get("ypos"))))
        self.set_text(root.find("text").text)
        font_element = root.find("font")
        font = QFont(font_element.find("family").text,
                     int(font_element.find("point_size").text),
                     int(font_element.find("weight").text))
        font.setItalic(font_element.find("italic").text == "True")
        font.setUnderline(font_element.find("under_line").text == "True")
        font.setStrikeOut(font_element.find("strike_out").text == "True")        
        self.set_font(font)

        color = font_element.find("colour")
        self.set_font_colour(QColor(int(color.find("red").text),
                                    int(color.find("green").text),
                                    int(color.find("blue").text)))

        bg_color = root.find("background_colour")
        self.set_background_colour(QColor(int(bg_color.find("red").text),
                                          int(bg_color.find("green").text),
                                          int(bg_color.find("blue").text)))

    def set_text(self, text):
        self._text = text


    def set_pos(self, pos):
        """Sets the position of the comment

        parameters:
           * pos: a tuple (xpos, ypos) describing the position of the comment 
        """
        
        self._xpos, self._ypos = pos


    def set_font(self, font):
        """Sets the font of the comment

        parameters:
           * font: should be of type QFont
        """
        
        self._font = font


    def set_font_colour(self, font_colour):
        """Sets the font colour of the comment

        parameters:
           * font_colour: should be of type QColor
        """
        
        self._font_colour = font_colour


    def set_background_colour(self, background_colour):
        """Sets the background colour of the comment

        parameters:
           * background_colour: should be of type QColor
        """

        self._background_colour = background_colour


    def get_text(self):
        return self._text


    def get_pos(self):
        """Returns a tuple (x, y) describing the position"""
        return (self._xpos, self._ypos)


    def get_font(self):
        """Returns a QFont object describing the font of the comment"""
        return self._font


    def get_font_colour(self):
        """Returns a QColor object describing the font colour of the comment"""
        return self._font_colour


    def get_background_colour(self):
        """Returns a QColor object describing the background of the comment"""
        return self._background_colour


    def to_xml(self):
        """
        Makes XML-code of this object
        Return:
           *list: Contains the XML representation of this object
        """
        
        xpos, ypos        = self.get_pos()
        xpos              = str(self._xpos)
        ypos              = str(self._ypos)
        text              = str(self._text)
        family            = str(self._font.family())
        point_size        = str(self._font.pointSize())
        weight            = str(self._font.weight())
        italic            = str(self._font.italic())
        under_line        = str(self._font.underline())
        strike_out        = str(self._font.strikeOut())
        font_colour_red   = str(self._font_colour.red())
        font_colour_green = str(self._font_colour.green())
        font_colour_blue  = str(self._font_colour.blue())
        bg_colour_red     = str(self._background_colour.red())
        bg_colour_green   = str(self._background_colour.green())
        bg_colour_blue    = str(self._background_colour.blue())

        return ["<canvascomment xpos=\"" + xpos + "\" ypos=\"" + ypos + "\">",
                "\t<text>" + text + "</text>",
                "\t<font>",
                "\t\t<family>" + family + "</family>",
                "\t\t<point_size>" + point_size + "</point_size>",
                "\t\t<weight>" + weight + "</weight>",
                "\t\t<italic>" + italic + "</italic>",
                "\t\t<under_line>" + under_line + "</under_line>",
                "\t\t<strike_out>" + strike_out + "</strike_out>",
                "\t\t<colour>",
                "\t\t\t<red>" + font_colour_red + "</red>",
                "\t\t\t<green>" + font_colour_green + "</green>",
                "\t\t\t<blue>" + font_colour_blue + "</blue>",
                "\t\t</colour>",
                "\t</font>",
                "\t<background_colour>",
                "\t\t<red>" + bg_colour_red + "</red>",
                "\t\t<green>" + bg_colour_green + "</green>",
                "\t\t<blue>" + bg_colour_blue + "</blue>",
                "\t</background_colour>",
                "</canvascomment>"]
