# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os
import sre

from qt import Qt, QScrollView, QColor, QSizePolicy, QRect, QPoint, QPixmap
from qt import QCursor, QMessageBox, QFileDialog, qRed, qGreen, qBlue, qGray
from qt import QObject, SIGNAL, QLayout
from qtui import QWidgetFactory
from qtcanvas import QCanvasView, QCanvas, QCanvasLine

from GUI.Canvas.CanvasLine import CanvasLine
from GUI.Canvas.CanvasLabel import CanvasLabel
from GUI.Canvas.CanvasNode import CanvasNode
from GUI.Canvas.canvascomment import CanvasComment
from GUI.Canvas.CanvasMarkLine import CanvasMarkLine
from GUI.Canvas.tools import PointingTool, LinkingTool, CommentTool
from GUI.guinode import GUINode
from GUI.links import LinkCreationError
from GUI.connection import Connection
from GUI.Nodes import guiplot3dnode
from GUI.ProgressDockWindow import ProgressDockWindow
from GUI.EventPopup import EventPopup
from Library.Util.Config import Config
from Library.Util.gui import EventReceiver
from GUI.Pixmaps.Icons import Icons
from GUI.Nodes.gaussdal_node import Gaussdal_gui_node
from GUI.Nodes.guiplot2dnode import GUIPlot2DNode
from GUI.Nodes.spreadsheetnode import SpreadSheetGUINode
from System.events import StartDiagramEvent
from GUI.Nodes.multipleoutputnode import FileWriterGUINode
from GUI.Nodes.guiuserquerynode import GUIUserQueryNode
from GUI.Nodes.guidissectionnode import GUIDissectionNode
from GUI.Nodes.inputnodegui import InputGUINode
from GUI.Nodes.guidevelplotnode import GUIDevelPlotNode

connect = QObject.connect
    
# Helper functions for MDView when setting positions to new nodes


def bounding_rect(gui_elements):
    """Return bounding rectangle of a set of gui elements. This function
    requires that each element in the given list has functions left(), right(),
    top(), bottom().

    Returns a QRect
    """
    if not gui_elements:
        return QRect(0, 0, 0, 0)

    first_rect = gui_elements[0].boundingRect()
    xmin = first_rect.left()
    ymin = first_rect.top()
    xmax = first_rect.right()
    ymax = first_rect.bottom()

    for gui_element in gui_elements[1:]:
        iter_rect = gui_element.boundingRect()
        xmin_ = iter_rect.left()
        ymin_ = iter_rect.top()
        xmax_ = iter_rect.right()
        ymax_ = iter_rect.bottom()
        
        xmin = min(xmin, xmin_)
        ymin = min(ymin, ymin_)
        xmax = max(xmax, xmax_)
        ymax = max(ymax, ymax_)

    return QRect(QPoint(xmin, ymin), QPoint(xmax, ymax))

def find_free_position(node, nodes, area_size):
    """Find free (x, y) position in area for placing node with node_size."""
    
    free_x = 0
    free_y = 0
    increment_x = 22 # horizontal space between nodes.
    increment_y = 30 # vertical spacing...
    
    while True:
        collided = False 
        for other_node in nodes:
            node_rect = (free_x, free_x + node.width(),
                         free_y, free_y + node.height())
            if is_overlapping(node_rect, other_node):
                collided = True
        if not collided:
            return (free_x, free_y)
        free_x += increment_x
        
        if area_size[0]-free_x < (node.width()+increment_x):
            
            free_y += increment_y
            free_x = 0
            if area_size[1]-free_y < (node.height()):
                free_x = 0
                free_y = 0
                return (free_x, free_y)
            
def is_overlapping(node_rect, other_node):
    """Return if rectangle overlaps GUINode.

    node_rect consists of (x1, x2, y1, y2).
    other_node is GUINode, supporting x, y, width and height methods.
    """
    return max(node_rect[0], other_node.x()) < min(node_rect[1], other_node.x() + other_node.width()) and \
        max(node_rect[2], other_node.y()) < min(node_rect[3], other_node.y() + other_node.height())

def find_canvas_size(nodes_size, visible_size, visible_scrollbars_size):
    """Calculate size of canvas to show scrollbars only when necessary.

    nodes_size is (max_x, max_y) of all nodes
    visible_size is (width, height) of view without scrollbars
    visible_scrollbars_size is (width, height) of view with scrollbars

    Return (width, height) tuple of best canvas size.
    """
    if nodes_size[0] <= visible_size[0] and nodes_size[1] <= visible_size[1]:
        return visible_size
    else:
        canvas_width = max(nodes_size[0], visible_scrollbars_size[0])
        canvas_height = max(nodes_size[1], visible_scrollbars_size[1])
        return (canvas_width, canvas_height)


class IgnorableEventPopup:
    """Displays warning event, unless user has chosen to ignore it."""
    def __init__(self, parent, handler, event):
        self._handler = handler
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          "event_popup.ui"),
                                             None, parent)
        
        self._widget.child('warning_text').setText(event.getDescription())

        preferred_height = self._widget.layout().heightForWidth(self._widget.width())

        self._widget.setMinimumWidth(self._widget.width())
        self._widget.setMinimumHeight(preferred_height)
        self._widget.layout().setResizeMode(QLayout.Minimum)

        self._widget.connect(self._widget.child('ok_button'),
                             SIGNAL('clicked()'), self._accepted)
        self._event_id = event.get_warning_id()

        # make sure we get notified if our warning type suddenly is
        # ignored by another popup
        handler.add_change_listener(self)

    def show(self):
        """Show this event if it's not ignored by the user."""
        if self._handler.is_ignored(self._event_id):
            return

        self._widget.show()
        self._widget.updateGeometry()

    def _accepted(self):
        if self._widget.child('checkbox').isChecked():
            self._handler.add_ignored_warning(self._event_id)

    def warnings_changed(self):
        # If another (or this) popup makes us ignored, we should go away.
        if self._handler.is_ignored(self._event_id):
            self._widget.hide()

class IgnorableWarningPopupHandler:
    """Notify warning popups about changes in which warnings to ignore.

    Also keep reference of popup-widgets that are
    active. Referenceless widgets tend to lose ability to send
    signals.
    """
    def __init__(self):
        self._listeners = []
        
    def is_ignored(self, warn_type):
        """Find out if user wants to ignore this type of warnings.

        warn_type is stringbased and should not include colon (:).
        """
        return warn_type in Config.ignored_warnings.split(":")

    def add_change_listener(self, listener):
        self._listeners.append(listener)

    def add_ignored_warning(self, warn_type):
        """Add warn_type to list of ignored_warnings and notify listeners."""
        ignored = filter(None, Config.ignored_warnings.split(":"))
        if warn_type in ignored:
            return
        
        ignored += (warn_type,)
        Config.ignored_warnings = ":".join(ignored)
        
        for listener in self._listeners:
            listener.warnings_changed()

    def remove_change_listener(self, listener):
        self._listeners.remove(listener)


class MDView(QCanvasView, EventReceiver):

    def __init__(self, workspace, application, module_diagram,
                 event_server, filename, caption):

        QCanvasView.__init__(self, workspace, "canvas view",
                             Qt.WDestructiveClose )
        EventReceiver.__init__(self)

        # canvas must be assigned to variable to keep from being deleted by GC
        self._canvas = QCanvas()
        self.setCanvas(self._canvas)
        
        self._workspace = workspace
        self._application = application        
        
        self._caption = caption
        
        self.event_server = event_server
        self._module_diagram = module_diagram

        # keeps xml-data in order to compare when user closes the diagram
        self._xml_data = ""

        self.canvas().setBackgroundColor(QColor(Config.bgR, Config.bgG,
                                                Config.bgB))
        
        self.setVScrollBarMode(QScrollView.Auto)
        self.setHScrollBarMode(QScrollView.Auto)

        event_server.addEventHandler(self.post_event)
        
        # List with CanvasComment objects from module
        # GUI.Canvas.canvascomment
        self._comments = [] 
        # List of the currently selected items in the Canvas
        self._selected_items = []
        # The currently selected link
        self._selected_link = None
        # list with GUI.guinode.GUINode objects, indexed by node
        # ModuleDiagramID
        self.nodes = {}
        self.links = []

        self.viewport().setMouseTracking(1)
        self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)

        self._progress_bar = ProgressDockWindow(self._application._widget)
        self._node_status_data_list = self._application.get_node_status_data_list()
        self.activate() #sets filename
        self._module_diagram.setFilename(filename)

        self.setIcon(QPixmap(Icons.s))
        
        # enable drag and drop
        self.setAcceptDrops(1)
        
        self._populate_diagram()
    
        self._resize_canvas_to_fit()

        self._pointing_tool = PointingTool(self)
        self._linking_tool  = LinkingTool(self)
        self._comment_tool  = CommentTool(self)
        self._current_tool  = None
        self._previous_tool = None
        self.set_pointing_tool_as_current_tool()

        # save current state of module diagram's XML
        self._update_diagram_state()

        # warning dialogs need reference kept to be fully functional
        self._popup_handler = IgnorableWarningPopupHandler()

    def _set_current_tool(self, tool):
        if self._current_tool:
            self._current_tool.reset()
        self._current_tool = tool
        self._current_tool.activate()

    def set_pointing_tool_as_current_tool(self):
        self._set_current_tool(self._pointing_tool)
        self._application.enable_pointer_tool()

    def set_linking_tool_as_current_tool(self):
        self._set_current_tool(self._linking_tool)
        self._application.enable_link_tool()

    def set_comment_tool_as_current_tool(self):
        self._set_current_tool(self._comment_tool)
        self._application.enable_comment_tool()

    def get_current_tool(self):
        return self._current_tool

    def set_previous_tool(self, tool):
        """Sets the previous tool used for this MDView"""
        self._previous_tool = tool

    def get_previous_tool(self):
        return self._previous_tool

    def _update_diagram_state(self):
        """Update the current state of the module diagram.

        Call this method when an update to the module diagram has
        happened, which you don't want to ask the user to save.
        """
        self._xml_data = self._module_diagram.toXML()
    
    def _populate_diagram(self):
        for node in self._module_diagram.get_nodes():
            new_node = self._create_gui_node(node)
            self.nodes[node.getId()] = new_node
            new_node.getCanvasNode().newZ(2*len(self.nodes)-1)

        # creates canvas comments
        for comment in self._module_diagram.get_canvas_comments():
            gui_comment = CanvasComment(self, comment, False)
            gui_comment.redraw()
            self._comments.append(gui_comment)

        # gets a list of (from outputport, to inputport)-tuples.
        node_pairs = self._module_diagram.getConnections()

        # fill this list with new node->node connections
        linkable = {}

        for output_node, input_node in node_pairs:
            
            # get nodes
            fromnode = output_node.getOwnerNode()
            tonode = input_node.getOwnerNode()
                
            # if nodes haven't been added before, add them!
            if not linkable.has_key((fromnode.getId(), tonode.getId())):

                # position is simply where from the node the
                # connection should be drawn
                topos = output_node.getPosition()
                linkable[fromnode.getId(), tonode.getId()] = (topos[0], topos[1])
        
        # add links between nodes with no port connections
        for snode_id, enode_id in self._module_diagram.get_connectionless_links():
            if not linkable.has_key((snode_id, enode_id)):
                topos = self.nodes[enode_id].getDiagramNode().getPosition()
                linkable[snode_id, enode_id] = (topos[0], topos[1])

        # create the actual objects!
        for keys in linkable:
            fromnode = self.nodes[keys[0]]
            tonode = self.nodes[keys[1]]        

            start_point = linkable[keys][0], linkable[keys][0]
            end_point = linkable[keys][1], linkable[keys][1]
            
            self.create_link(fromnode, tonode, start_point, end_point)

    def create_canvas_comment(self, point, created_on_user_interaction = False):
        comment_data = self._module_diagram.create_canvas_comment("new comment",
                                                                  point.x(),
                                                                  point.y())
        new_comment = CanvasComment(self, comment_data,
                                    created_on_user_interaction)
        self._comments.append(new_comment)
        return new_comment

    def remove_canvas_comment(self, comment):
        self._module_diagram.del_canvas_comment(comment.get_canvas_comment_data())
        self._comments.remove(comment)

    def get_canvas_comments(self):
        return self._comments
    
    def _create_gui_node(self, node):
        """Create GUINode related to specific Node."""
        node_class = node.getNodeRepName()
        if sre.match("Plot(Nodes)?\.Plot3DNode", node_class):
            return guiplot3dnode.GUIPlot3DNode(self, node, self.event_server)
        elif sre.match("Plot(Nodes)?\.Plot2DNode", node_class):
            return GUIPlot2DNode(self, node, self.event_server)
        elif sre.match("Plot(Nodes)?\.(Custom|Saved )plots", node_class):
            return GUIPlot2DNode(self, node, self.event_server)
        elif node_class.startswith("View.SpreadSheetNode"):
            return SpreadSheetGUINode(self, node)
        elif node_class.startswith("Tools.MoleculeEditorNode"):
            # imported here in case pymol doesnt exist, in which case it
            # raises an exception 
            from GUI.Nodes.MoleculeEditorNode import MoleculeEditorNode
            return MoleculeEditorNode(self, node)
        elif node_class.startswith("Tools.GaussdalNode"):
            return Gaussdal_gui_node(self, node)
        elif node_class.startswith("File.OutputNode"):
            return FileWriterGUINode(self, node)
        elif node_class.startswith("Input.UserQueryNode"):
            return GUIUserQueryNode(self, node)
        elif node_class.startswith('Tools.DissectionNode'):
            return GUIDissectionNode(self, node)
        elif sre.search("InputNode", node_class):
            return InputGUINode(self, node)
        elif node_class.endswith("DevelPlotNode"):
            return GUIDevelPlotNode(self, node)
        else:
            return GUINode(self, node)

    def create_node(self, node_class, x=None, y=None,):
        node = self._module_diagram.add_node_by_name(node_class, None)
        position_set = False

        if x is not None and y is not None:
            position_set = True
            node.setPosition(x, y)
            
        new_node = self._create_gui_node(node)
        
        if not position_set:
            new_position = find_free_position(new_node, self.nodes.values(),
                                              (self.canvas().width(),
                                               self.canvas().height()))
            new_node.setPosition(new_position[0], new_position[1])        
                        
        self.nodes[node.getId()] = new_node
        # Update node status dock window.
        self._node_status_data_list.update_list()
        
        new_node.getCanvasNode().newZ(2*len(self.nodes)-1)
        self.set_selected_items([new_node])
        self.dehighlight_none_selected_items()
        self.highlight_selected_items()
            
        self.setFocus()
        return new_node
        
    def remove_node(self, node):
        diagram_node = node.getDiagramNode()
        node_id = diagram_node.getId()
        del self.nodes[node_id]
        self._module_diagram.remove_node(diagram_node)
        # Update status list 
        self._node_status_data_list.update_list() 

    def create_link(self, start_node, end_node, start_point, end_point):
        """Create a link between two GUINode items.

        Creates a link at the exit/entry-point closest to given QPoint
        at each node.
        """
        diagram_start = start_node.getDiagramNode()
        diagram_end = end_node.getDiagramNode()
        
        for link in self.links: #first check if the link already exists
            if (diagram_start, diagram_end) == link.get_nodes():
                return

        start = start_node.get_closest_exit_line(start_point)
        end = end_node.get_closest_entry_line(end_point)

        if not start or not end:
            raise LinkCreationError(
                '%s and %s cannot be connected like that.\n' \
                % (diagram_start.getName(),
                   diagram_end.getName()))
        
        start_point = start_node.x() + start_node.width()/2., start_node.y() + start_node.height()/2.
        end_point = end_node.x() + end_node.width()/2., end_node.y() + end_node.height()/2.

        connection = Connection(self,
                                diagram_start,
                                diagram_end,
                                start_point, end_point,
                                end_node.width(), end_node.height())
        start_node.addStartLink(connection)
        end_node.addEndLink(connection)

        self.links.append(connection)
        self._module_diagram.add_node_link(diagram_start.getId(), 
                                           diagram_end.getId())
        
        return connection

    def get_links(self):
        return self.links

    def set_canvas_view_cursor(self, pixmap_data):
        cursor_pixmap = QPixmap(pixmap_data)
        cursor = QCursor(cursor_pixmap)
        self.setCursor(cursor)

    def _resize_canvas_to_fit(self):
        canvas_items = self.nodes.values()
        canvas_items.extend(self._comments)
        rect = bounding_rect(canvas_items)
        
        xmax = rect.right()
        ymax = rect.bottom()

        visible_size = (self.visibleWidth(), self.visibleHeight())

        visible_scrollbars_size = (self.visibleWidth() - self.verticalScrollBar().width(),
                                   self.visibleHeight() - self.horizontalScrollBar().height())
        
        canvas_width, canvas_height = find_canvas_size((xmax, ymax),
                                                       visible_size,
                                                       visible_scrollbars_size)
        self.canvas().resize(canvas_width, canvas_height)

    def resizeEvent(self, e):
        """Method overridden from QCanvasView"""
        save_needed_before_resize = self._check_save_needed()
        QCanvasView.resizeEvent(self, e)

        self._resize_canvas_to_fit()
        self._module_diagram.setVariable("canvasSize", (self.canvas().width(),
                                                        self.canvas().height()))

        # we don't want to ask the user if changes should be saved,
        # when the only change is resizing
        if not save_needed_before_resize:
            self._update_diagram_state()

    def _get_collided_gui_element(self, point):
        """Gets the gui element of interest with highest z-order in the
        collision list. If none found, the returned item is None.

        parameters:
           * point : The location where the mouse currently is

        """
        item_list = self.canvas().collisions(point)
        for item in item_list:
            if isinstance(item, CanvasMarkLine):
                continue
            if isinstance(item, CanvasNode) or isinstance(item, QCanvasLine): 
                return item.getParent()
            elif isinstance(item, CanvasComment):
                return item
            elif isinstance(item, CanvasLabel):
                return item
        return None
        
    def contentsMousePressEvent(self, e): # QMouseEvent e
        """Method overridden from QCanvasView"""
        point = self.inverseWorldMatrix().map(e.pos())
        gui_element = self._get_collided_gui_element(point)

        if e.button() == Qt.LeftButton: 
            try:
                self._current_tool.left_mouse_pressed(point, gui_element)
                self._progress_bar.setup()    
            except LinkCreationError, e:
                QMessageBox.warning(self, "Link creation error", str(e))

        elif e.button() == Qt.RightButton:
            self._current_tool.right_mouse_pressed(point, gui_element)

        self.setFocus() # need this since the dockwindows will grab the keyboard focus, do _not_ use grabKeyboard()
        self._canvas.update()
                
    def contentsMouseReleaseEvent(self, e):
        """Method overridden from QCanvasView"""
        point = self.inverseWorldMatrix().map(e.pos())
        gui_element = self._get_collided_gui_element(point)
        
        if e.button() == Qt.LeftButton:
            try:
                self._current_tool.left_mouse_released(point, gui_element)
                self._progress_bar.setup()
            except LinkCreationError, e:
                QMessageBox.warning(self, "Link creation error", str(e))
        elif e.button() == Qt.RightButton:
            self._progress_bar.setup()

        self._canvas.update()

    def contentsMouseMoveEvent(self, e):
        """Method overridden from QCanvasView"""
        point = self.inverseWorldMatrix().map(e.pos())
        self._current_tool.mouse_moved(point)
        if e.state() == Qt.LeftButton:
            self._canvas.update()

    def contentsMouseDoubleClickEvent(self, e):
        """Method overridden from QCanvasView"""
        point = self.inverseWorldMatrix().map(e.pos())
        gui_element = self._get_collided_gui_element(point)
        if e.button() == Qt.LeftButton:
            self._current_tool.left_mouse_double_clicked(point, gui_element)

    def keyPressEvent(self, e):        
        """Method overridden from QCanvasView"""
        self._current_tool.key_pressed(e.key())

    def bring_to_front(self, item):
        # A node has two z-values; the node icon and label have value
        # z and the label text z + 1.
        old_z = item.z()
        item.newZ(2 * len(self.nodes) - 1)
        for node in self.nodes.values():
            view_node = node.getCanvasNode()
            if view_node != item and view_node.z() > old_z:
                view_node.newZ(view_node.z() - 2)

    def update(self):
        """updates the background color. Method overridden from QCanvasView."""
        self.canvas().setBackgroundColor(QColor(Config.bgR, Config.bgG,
                                                Config.bgB))
        for line in self.links: # updates line-color
            line.update()
            
        self.canvas().update()

    def save_node_settings(self):
        """Save GUI settings for all nodes."""
        for node in self.nodes.values():
            node.saveGuiSettings()

    def get_nodes(self):
        return self.nodes

    def get_diagram_id(self):
        return self._module_diagram.getId()
    
    def handle_event(self, event):
        """Function that will be called by eventserver with all events sent 
            from the diagram. All events that is handled by this eventHandler
            will be in the main event thread through EventReceiver(superclass).
        """
        # need to check if this event is for the current diagram
        owner_id = event.get_module_diagram_id()
        class_ = event.get_event_class()
        type_ = event.getType()
        if owner_id != self._module_diagram.getId():
            return
        
        if class_ == "DiagramEvent":
            if type_ == "START":
                if isinstance(event, StartDiagramEvent):
                    num_nodes = event.get_num_nodes()
                else:
                    num_nodes = len(self.nodes)
                    
                self._progress_bar.setup(num_nodes)
                self._progress_bar.startProgress()
            if type_ == "FINISH":
                self._progress_bar.setFinished()
            if type_ == "HALT":
                self._progress_bar.setHalted()
            if type_ == "EXCEPTION":
                if not self._progress_bar.isCrashed():
                    self._progress_bar.setCrashed()

        elif class_ == "NodeEvent":
            # Update node status on all node events.
            self._node_status_data_list.update_list()
            if type_ == "FINISH":
                if not self._module_diagram.getIsInterrupted():
                    self._progress_bar.addProgress(1)
            elif type_ == "EXCEPTION":
                self._progress_bar.setCrashed()
                EventPopup(self._application._widget, event)
            elif type_ == "INFO":
                EventPopup(self._application._widget, event)
        elif class_ == "Warning":
            if type_ == "Ignorable":
                popup = IgnorableEventPopup(self._application._widget, self._popup_handler, event)
                popup.show()
            else:
                EventPopup(self._application._widget, event)

    def activate(self, filename = 0):
        if filename:
            self._module_diagram.setFilename(filename)
            self._caption = os.path.basename(filename)
            self.setCaption(self._caption)
        else:
            self.setCaption(self._caption)
            
        self._application.set_progress_bar(self._progress_bar)
        self._node_status_data_list.update_list()
        self._progress_bar.refreshBar()

    def closeEvent(self, e):
        """Method overridden from QCanvasView"""
        if self._check_save_needed():
            msgtext = "Save changes to diagram %s?" % str(self.caption())
            check = QMessageBox.question(None, "Diagram is modified", msgtext,
                                         "&Save", "&Discard", "&Cancel")
            if check == 0: # Save
                # ... and let the diagram close
                if not self.save_diagram():
                    e.ignore() # user cancelled saving the diagram
                    return 
            elif check == 2: # Cancel
                e.ignore()
                return 
            else: # Discard
                # we are discarding changes, let the diagram close
                pass

        self._module_diagram.on_close()
        self._workspace.remove_canvas(self._module_diagram.getId())            
        e.accept()

    def save_diagram(self, save_as=0):
        """Saves the current diagram.

        If save_as is True, the user will be asked to give a
        filename. If, for any reason, SciCraft is unable to open the
        diagram's filename for writing, the user will be notified that
        the changes could not be saved and False is returned.

        If the user cancels the 'Save as...' file dialog, False is
        returned. If everything goes just fine, True is returned.
        """
        self.save_node_settings()

        filename = self.get_filename()

        if not filename or save_as:
            filename = QFileDialog.getSaveFileName(Config.last_path_modulediagram,
                                                   "SciCraft diagrams (*.zmd)",
                                                   self._application._widget,
                                                   "Save File",
                                                   "Save File").latin1()
            
            if not filename: # jump out if no filename returned from the dialog
                return False
            
            filename = str(filename)
            Config.last_path_modulediagram = os.path.dirname(filename)
            
            if not filename.split(".")[-1] == "zmd":
                filename = filename + ".zmd"
        try:
            self._module_diagram.save_to_file(filename)
        except:
            msg_text = """SciCraft was unable to save your changes in \"%s\".
Do you have the proper write-permissions?""" % filename
            button_pressed = QMessageBox.critical(self,
                                                  "File not writable", msg_text,
                                                  QMessageBox.Abort,
                                                  QMessageBox.Retry)
            if button_pressed == QMessageBox.Retry:
                return self.save_diagram(save_as)
            else: # Abort...
                return False
        
        self.activate(filename)
        self._update_diagram_state()
        from GUI.Workspace import add_to_recent_files
        add_to_recent_files(filename)
        return True        

    def get_filename(self):
        return self._module_diagram.getFilename()

    def to_xml(self):
        """Create XML from module diagram."""
        return self._module_diagram.toXML()

    def to_eps(self, filename,colour_mode):
        """Creates an eps-image file of the scicraft diagram through pyx."""        
        try:
            from pyx import color, deco, canvas, unit, box, path, bitmap
        except ImportError:
            QMessageBox.information(self, "Message",
                                    "PyX must be installed for EPS export to work.")
            return
        
        #pyx canvas with unit "points"
        unit.set(defaultunit="pt")
        c = canvas.canvas()

        for link in self.links:
            mi = 0
            while mi < (len(link.line_points)-1):
                p1 = link.line_points[mi]
                p2 = link.line_points[mi+1]

                #if last point, add arrow-head
                if mi == (len(link.line_points)-2):
                    # removes whitespace at end of line before
                    # arrow-head - must see to believe :)
                    add_x2 = 0
                    if p1.x() >= p2.x():
                        add_x2 = -4
                    else:
                        add_x2 = 4

                    c.stroke(path.line(p1.x(), 0-p1.y(), p2.x()+add_x2,
                                       0-p2.y()),
                             [deco.earrow.normal])
                else:
                    #Just draw line
                    c.stroke(path.line(p1.x(), 0-p1.y(), p2.x(), 0-p2.y()))

                mi+=1

        font_height = 9.7 # how high is one character
        font_width = 5.5 # how wide is one character

        for comment in self._comments:
            comment_data = comment.get_canvas_comment_data()
            x_pos, y_pos = comment_data.get_pos()
            text = str(comment_data.get_text()).splitlines()

            # get the width of the comment box (the longest line)
            max_length = 0
            for line in text:
                max_length = max(max_length, len(line))

            # the height of the box (in lines)
            no_lines = len(text)

            text_x_offset = 1.0
            text_y_offset = font_height / 2.0

            # draw the comment box
            boxrect = box.rect(x_pos, 0 - y_pos - no_lines * font_height,
                               font_width * max_length + 2 * text_x_offset,
                               no_lines * font_height + 2 * text_y_offset)
            c.stroke(boxrect.path(bezierradius=0.0), [color.rgb.black])
            
            # get the background and foreground color
            bgcolour = comment_data.get_background_colour()
            fgcolour = comment_data.get_font_colour()
            pyx_bg_colour = None
            pyx_fg_colour = None
            if colour_mode:
                pyx_bg_colour = color.rgb(bgcolour.red() / 255.0,
                                          bgcolour.green() / 255.0,
                                          bgcolour.blue() / 255.0)
                pyx_fg_colour = color.rgb(fgcolour.red() / 255.0,
                                          fgcolour.green() / 255.0,
                                          fgcolour.blue() / 255.0)
            else:
                pyx_bg_colour = color.gray(qGray(bgcolour.rgb()) / 255.0)
                pyx_fg_colour = color.gray(qGray(fgcolour.rgb()) / 255.0)

            # fill the comment box
            c.fill(boxrect.path(bezierradius=0.0), [pyx_bg_colour])

            # draw the text inside the box
            for i, line in enumerate(text):
                c.text(x_pos + text_x_offset,
                       0 - y_pos - i * font_height - text_y_offset, line,
                       [pyx_fg_colour])

        for node in self.nodes.values():
            #pyx node
            node_im = QPixmap(node.image).convertToImage()
            bitstring = ""
            mode_string = ""
            if colour_mode:
                mode_string = "RGB"
                for y in range(node_im.height()):
                    for x in range(node_im.width()):
                        px = node_im.pixel(x, y)
                        bitstring += chr(qRed(px)) +\
                                     chr(qGreen(px)) +\
                                     chr(qBlue(px))
            else:
                mode_string = "L"
                for y in range(node_im.height()):
                    for x in range(node_im.width()):
                        bitstring += chr(qGray(node_im.pixel(x, y)))

            pyx_image = bitmap.image(node_im.width(), node_im.height(), mode_string,
                                     bitstring)
            pyx_bitmap = bitmap.bitmap(node.x(), 0 - node.y() - 40, pyx_image,
                                       width=node.width(), height=node.height())
            c.insert(pyx_bitmap)

            #pyx node-label
            new_name = node.getDiagramNode().getName().replace("Node","")
            labelrect = box.rect(node.x() + 1.2, 0 - node.y() - 48.3,
                               font_width * len(new_name), font_height)
            c.stroke(labelrect.path(bezierradius=0.0), [color.rgb.black])
            c.fill(labelrect.path(bezierradius=0.0), [color.rgb.white])

            #pyx label-text
            c.text(node.x()+2.2, 0-node.y()-46.3, new_name, [color.rgb.black])

        try:
            c.writeEPSfile(filename)
        except AttributeError:
            QMessageBox.information(self, "Message",
                                    "The diagram is empty, nothing to export.")
    
    def validate_diagram(self):
        """Validate module diagram.
        Return:
                a list consisting of:
                    - List of NodeInputPort which doesn't have their data set
                    - List of nodes which have not defined all parameters"""
        
        return self._module_diagram.validate()
        
    def run_diagram(self):
        """Start execution of module diagram."""
        self._module_diagram.start()
    
    def get_module_diagram(self):
        return self._module_diagram
    
    def dragEnterEvent(self, event):
        """Method overridden from QCanvasView"""
        # Drag-and-drops entering this QWidget are accepted"""
        
        if event.format().startswith("text/plain"):
            if str(event.encodedData("text/plain")).startswith("SciCraft node"):
                event.accept()
        
    def dropEvent(self, event):
        """Create a new node where the node icon is dropped.

        Method overridden from QCanvasView.
        """
        if event.format().startswith("text/plain"):
            if str(event.encodedData("text/plain")).startswith("SciCraft node"):
                dragObject = event.source().dragObject

                if not (dragObject.pixmapWidth or dragObject.pixmapHeight):
                    return
                
                dropNodeName = dragObject.nodeName
                dropPos = self.viewportToContents(event.pos())
                dropX = dropPos.x() - (dragObject.pixmapWidth / 2) - 2
                dropY = dropPos.y() - (dragObject.pixmapHeight / 2) - 2

                if dropX < 0:
                    dropX = 0
                elif dropX + dragObject.pixmapWidth > self.canvas().width():
                    dropX = self.canvas().width() - dragObject.pixmapWidth
               
                if dropY < 0:
                    dropY = 0
                elif dropY + dragObject.pixmapHeight > self.canvas().height():
                    dropY = self.canvas().height() - dragObject.pixmapHeight
               
                self.create_node(dropNodeName, dropX, dropY)
                dragObject.source().nodeItem = None
                dragObject.source().dragObject = None

    def _check_save_needed(self):
        """Returns True if the diagram is modified and needs to be saved."""
        return not(self._xml_data == self._module_diagram.toXML())

    def show_hidden_links(self):
        for link in self.links:
            link.show_link()
                
    def hide_hidden_links(self):
        for link in self.links:
            link.hidden_link()

    def set_selected_link(self, link):
        """ Sets the currently selected link.

        parameters:
           * link : the selected link of type Link
        """
        self._selected_link = link

    def get_selected_link(self):
        return self._selected_link


    def reset_selected_link(self):
        self._selected_link = None

    def set_selected_items(self, item_list):
        """ Sets the currently selected items.

        parameters:
           * item_list: a list of currently selected items.
        """
        self._selected_items = item_list

    def get_selected_items(self):
        return self._selected_items

    def reset_selected_items(self):
        self._selected_items = []

    def dehighlight_none_selected_items(self):
        # dehighlights all nodes and links which aren't selected
        for link in self.links:
            link.set_highlight(False)
        for node in self.nodes.values():
            if not node in self._selected_items:
                node.set_highlight(False)
        for comment in self._comments:
            if not comment in self._selected_items:
                comment.set_highlight(False)

    def highlight_selected_items(self):
        # highlights the node(s) currently selected
        if self._selected_link:
            self._selected_link.set_highlight()
        elif self._selected_items:
            for gui_element in self._selected_items:
                gui_element.set_highlight()

    def remove_line(self, line):
        if line in self.links:
            self.links.remove(line)
            self._module_diagram.remove_node_link(line.get_start_node().getId(),
                                                  line.get_end_node().getId())
    
