# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QWidget, QDialog, QColorDialog, QColor, SIGNAL, QFileDialog
from qt import QMessageBox
from qtui import QWidgetFactory

from Library.Util.Config import Config, write_config
from Library.Util.file import CacheDirectory, PathList, InvalidPathException
from GUI.viewers import HelpViewer
import os


class EditPrefs(QWidget):

    def __init__(self, parent, existing_cache):
        
        QWidget.__init__(self, parent)
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir, "edit_prefs.ui"), None, self)
        self._style_combo = self._widget.child('style_combo')
        self._background_colour = self._widget.child('background_button')
        self._line_colour = self._widget.child('line_colour_button')
        self._line_size = self._widget.child('line_size_spin')
        self._composite_check = self._widget.child('composite_check')

        function_browse_button = self._widget.child('browse_button')
        self._add_button = self._widget.child('add_button')
        self._add_line = self._widget.child('add_line')
        self._path_list = self._widget.child('path_list')
        remove_button = self._widget.child('remove_button')

        self._ok_button = self._widget.child('ok_button')
        cancel_button = self._widget.child('cancel_button')
        help_button = self._widget.child('help_button')
        self._cache_location = self._widget.child('cache_location')
        self._cache_enabled = self._widget.child('cache_enabled')
        self._current_cache_size = self._widget.child('current_size')
        self._cache_warn_limit = self._widget.child('cache_warn_limit')
        self._clear_cache = self._widget.child('clear_cache')
        cache_browse_button = self._widget.child('cache_browse_button')
        self._enable_keep_output = self._widget.child('enable_keep_output')
        

        self._parent = parent
        self.setCaption("SciCraft - Edit Preferences")

        # buttons
        
        if (Config.style == "QCDEStyle"):
            style = 0
        elif (Config.style == "QMotifStyle"):
            style = 1
        elif (Config.style == "QMotif plusStyle"):
            style = 2
        elif (Config.style == "QPlatinumStyle"):
            style = 3
        elif (Config.style == "QSGIStyle"):
            style = 4
        elif (Config.style == "QWindowsStyle"):
            style = 5
        elif (Config.style == "QWindowsXPStyle"):
            style = 6
        else:
            style = 0

        self._style_combo.setCurrentItem(style)
        self._background_colour.setPaletteBackgroundColor(QColor(Config.bgR,
                                                                 Config.bgG,
                                                                 Config.bgB))
        self._line_colour.setPaletteBackgroundColor(QColor(Config.lineR,
                                                     Config.lineG,
                                                     Config.lineB))
        self._line_size.setValue(Config.lineWidth)

        self._composite_check.setChecked(Config.use_composite)
        self._enable_keep_output.setChecked(Config.enable_keep_output)

        self.connect(self._clear_cache, SIGNAL("clicked()"), self._slot_clear_cache)
        self.connect(self._ok_button, SIGNAL("clicked()"), self.slot_ok) 
        self.connect(cancel_button, SIGNAL("clicked()"), self.slot_cancel)
        self.connect(help_button, SIGNAL("clicked()"), self.slot_help)
        self.connect(self._background_colour, SIGNAL("clicked()"), self.slot_background)
        self.connect(self._line_colour, SIGNAL("clicked()"), self.slot_line)
        self.connect(function_browse_button, SIGNAL("clicked()"), self._browse_function_path)
        self.connect(self._add_button, SIGNAL("clicked()"), self.slow_add_path)
        self.connect(remove_button, SIGNAL("clicked()"), self.slot_remove_from_path)
        self.connect(self._add_line, SIGNAL("textChanged(const QString &)"),
                     self.slot_change_default_button)
        self.connect(self._cache_location, SIGNAL("textChanged(const QString &)"),
                     self._update_cache_info)
        self.connect(cache_browse_button, SIGNAL("clicked()"), self._browse_cache_path)
        
        self._path_handler = PathList()
        self._draw_paths()

        self._cache = CacheDirectory(existing_cache.get_path())
        self._cache.set_enabled(existing_cache.get_enabled())
        self._cache_enabled.setChecked(existing_cache.get_enabled())
        self._cache_location.setText(existing_cache.get_path())
        self._cache_warn_limit.setValue(Config.cachewarnlimit)

        self._result = QDialog.Rejected

    def _slot_clear_cache(self):
        self._cache.clear_files()
        self._update_cache_info()

    def _update_cache_info(self):
        self._cache.set_path(str(self._cache_location.text()))
        try:
            used_space = self._cache.get_used_space()
        except:
            used_space = 0

        used_mb = used_space / (1024 * 1024)
        self._current_cache_size.setText("%s MB (%s bytes)" % (used_mb, used_space))

        if used_space > 0:
            self._clear_cache.setEnabled(True)
        else:
            self._clear_cache.setEnabled(False)
        
            

    def exec_loop(self):
        self._widget.exec_loop()
        return self._result

    def _draw_paths(self):
        self._path_list.clear()
        for item in self._path_handler.get_paths():
            if item:
                self._path_list.insertItem(item)

    
    def slot_remove_from_path(self):
        if str(self._path_list.currentText()):
            self._path_handler.remove_paths([str(self._path_list.currentText())])
            self._draw_paths() 

    
    def slow_add_path(self):
        try:
            if str(self._add_line.text()):
                self._path_handler.add_path(str(self._add_line.text()))
                self._draw_paths()
                self._add_line.setText("")
                self._ok_button.setDefault(True)
        except Exception,e:
            print e

    def _browse_cache_path(self):
        path = QFileDialog.getExistingDirectory(Config.homedir, self, 
            "getExistingDir", "Choose a directory")
        if path:
            self._cache_location.setText(path)
    
    def _browse_function_path(self):
        path = QFileDialog.getExistingDirectory(Config.homedir, self, 
            "getExistingDir", "Choose a directory", True, True)
        if path:
            self._path_handler.add_path(str(path))
            self._draw_paths()


    def slot_change_default_button(self):
        self._add_button.setDefault(True)


    def slot_ok(self):
        try:
            self._path_handler.write_paths()
        except InvalidPathException, e:
            invalid = self._path_handler.get_invalid()
            if len(invalid) > 1:
                fault_string = "\n".join(invalid)
                remove = QMessageBox.information(self, "Warning", \
                "The following paths do not exist:\n" + fault_string,\
                "OK","Remove from function path")
                if remove:
                    self._path_handler.remove_paths(invalid)
                    self._draw_paths()
                return
            else:
                remove = QMessageBox.information(self, "Warning", \
                    "The path " + invalid[0] + " does not exist.", "OK", \
                    "Remove from function path")
                if remove:
                    self._path_handler.remove_paths(invalid)
                    self._draw_paths()
                return

        Config.style = "Q" + str(self._style_combo.currentText()) + "Style"
        Config.lineWidth = self._line_size.value()
        bg_colour = self._background_colour.paletteBackgroundColor()
        line_colour = self._line_colour.paletteBackgroundColor()
      
        Config.bgR = bg_colour.red()
        Config.bgG = bg_colour.green()
        Config.bgB = bg_colour.blue()
        Config.lineR = line_colour.red()
        Config.lineG = line_colour.green()
        Config.lineB = line_colour.blue()
        Config.use_composite = self._composite_check.isChecked()
        Config.enable_keep_output = self._enable_keep_output.isChecked()
        
        Config.cacheenabled = self._cache_enabled.isChecked()
        Config.cachepath = str(self._cache_location.text())
        Config.cachewarnlimit = self._cache_warn_limit.value()
      
        write_config()
        self._result = QDialog.Accepted
        self._widget.close()


    def slot_cancel(self):
        self._widget.close()
        self._result = QDialog.Rejected
        
    def slot_background(self):

        bgColor = 0
        bgColor = QColorDialog.getColor(self._background_colour.paletteBackgroundColor(), self)        
        if bgColor.isValid():
            self._background_colour.setPaletteBackgroundColor(bgColor)

            
    def slot_line(self):
        
        lineColor = 0
        lineColor = QColorDialog.getColor(self._line_colour.paletteBackgroundColor(), self)
        if lineColor.isValid():
            self._line_colour.setPaletteBackgroundColor(lineColor)


    def slot_help(self):

        HelpViewer(self, os.path.join(Config.htmldocpath,
                                      "node_Changing_SciCraft_Settings.html"),
                   modal=True)
