# -*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from vtk import vtkMath
from qt import QCheckBox, QComboBox, QGridLayout, QLabel, QLineEdit, QMessageBox
from qt import QPushButton, QSizePolicy, SIGNAL, qApp
from GUI.Plots import PlotSetup
from GUI.Plots.Plots3D.MolPlot3D.FragmentDialog import FragmentDialog
import traceback
from Library.Util.math3d.Point3D import Point3D
from Library.Util.math3d.Vector3D import Vector3D
from Library.Util.Config import Config

#hack for redhat-installed pymol
import sys
import os.path
for path in sys.path:
    combopath = os.path.join(path, 'pymol/modules')
    if os.path.exists(combopath):
        if sys.path.count(combopath) == 0:
            sys.path.append(combopath)


class MoleculeEditorSetup(PlotSetup.PlotSetup):

    BNSTEXT = "Ball'n'Stick"
    SFTEXT = "Space Filling"
    LTEXT = "Line"
    
    def __init__(self, parent, node):
        """See GUI/Plots/PlotSetup """
        PlotSetup.PlotSetup.__init__(self, parent, None, 0)

        self.gui = parent.windowController.gui
        self.controller = parent.windowController
        self.node = node

        self.setName("MoleculeEditorSetup")
        
        MoleculeEditorSetupLayout = QGridLayout(self,1,1,11,6,"MoleculeEditorSetupLayout")

        self.name = QLineEdit(None,"name")

        # Source of data
        self.source = QComboBox(0,None,"Source")

        # Show index for each atom
        self.showIndexLabel = QLabel(self,"showIndexLabel")
        self.showIndexLabel.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed,0,0,self.showIndexLabel.sizePolicy().hasHeightForWidth()))
        self.showIndexLabel.setText(self.__tr("Show indices:"))

        self.showIndexBox = QCheckBox(self, "showIndexBox")
        self.connect(self.showIndexBox, SIGNAL("toggled(bool)"), self.slotIndexBoxToggled)

        # Show names of atoms
        self.showNameLabel = QLabel(self,"showNameLabel")
        self.showNameLabel.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed,0,0,self.showNameLabel.sizePolicy().hasHeightForWidth()))
        self.showNameLabel.setText(self.__tr("Show symbol/name:"))

        self.showNameBox = QCheckBox(self, "showNameBox")
        self.connect(self.showNameBox, SIGNAL("toggled(bool)"), self.slotNameBoxToggled)
        
        # Plot type
        self.plotTypeLabel = QLabel(self,"plotTypeLabel")
        self.plotTypeLabel.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed,0,0,self.plotTypeLabel.sizePolicy().hasHeightForWidth()))
        self.plotTypeLabel.setText(self.__tr("Plot Type:"))

        self.plotType = QComboBox(0,self,"plotType")
        self.plotType.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed,0,0,self.plotType.sizePolicy().hasHeightForWidth()))

        self.plotType.insertItem(self.BNSTEXT) # make sure ball'n'stick is default by inserting it first
        self.plotType.insertItem(self.SFTEXT)
        self.plotType.insertItem(self.LTEXT)
        self.connect(self.plotType, SIGNAL("activated(int)"), self.slotChangePlotType)

        self.applyButton = self.createButton("applyButton", "Apply setup", self.slotApply)
        self.deleteButton = self.createButton("deleteButton", "Delete atoms", self.slotDelete)
        self.invertButton = self.createButton("invertButton", "Invert selection", self.slotInvert)
        self.hideButton = self.createButton("hideButton", "Hide selection", self.slotHide)
        self.angleButton = self.createButton("angleButton", "Angle selection", self.slotAngle)
        self.extendButton = self.createButton("extendButton", "Extend to model", self.slotExtendSelectionToModel)
        self.extendAlongButton = self.createButton("extendAlongButton", "Extend along bonds", self.slotExtendSelectionAlongBonds)
        self.unhideButton = self.createButton("unhideButton", "Unhide all", self.slotUnhideAll)
        self.addHydrogenButton = self.createButton("addHydrogenButton", "Fill hydrogen", self.slotAddHydrogen)
        self.selectAllButton = self.createButton("selectAllButton", "Select all", self.slotSelectAll)
#        self.minimizeButton = self.createButton("minimizeButton", "Minimize energy", self.slotMinimize)
        self.singleAttachButton = self.createButton("singleAttachButton", "1-Point Fragments", self.slotSingleAttach)
        self.doubleAttachButton = self.createButton("doubleAttachButton", "2-Point Fragments", self.slotDoubleAttach)

        self.typeLabel = QLabel(self,"typeLabel")
        self.typeLabel.setSizePolicy(QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed,0,0,self.typeLabel.sizePolicy().hasHeightForWidth()))

        self.setCaption(self.__tr("Plot Config"))

        MoleculeEditorSetupLayout.addWidget(self.showNameLabel,2,0)
        MoleculeEditorSetupLayout.addWidget(self.showNameBox,2,1)
        MoleculeEditorSetupLayout.addWidget(self.showIndexBox,3,1)
        MoleculeEditorSetupLayout.addWidget(self.showIndexLabel,3,0)
        MoleculeEditorSetupLayout.addWidget(self.plotTypeLabel,1,0)
        MoleculeEditorSetupLayout.addWidget(self.plotType,1,1)
        MoleculeEditorSetupLayout.addWidget(self.applyButton,6,3)
        MoleculeEditorSetupLayout.addWidget(self.deleteButton,3,4)
        MoleculeEditorSetupLayout.addWidget(self.invertButton,1,3)
        MoleculeEditorSetupLayout.addWidget(self.hideButton,1,4)
        MoleculeEditorSetupLayout.addWidget(self.angleButton,4,4)
        MoleculeEditorSetupLayout.addWidget(self.extendButton,5,4)
        MoleculeEditorSetupLayout.addWidget(self.extendAlongButton,6,4)
        MoleculeEditorSetupLayout.addWidget(self.unhideButton,2,4)
        MoleculeEditorSetupLayout.addWidget(self.addHydrogenButton,5,3)
        MoleculeEditorSetupLayout.addWidget(self.selectAllButton,2,3)
        MoleculeEditorSetupLayout.addWidget(self.singleAttachButton,3,3)
        MoleculeEditorSetupLayout.addWidget(self.doubleAttachButton,4,3)
        MoleculeEditorSetupLayout.addMultiCellWidget(self.typeLabel,0,0,0,1)
#        MoleculeEditorSetupLayout.addWidget(self.minimizeButton,2,3)

    def createButton(self, name, text, callback):
        button = QPushButton(self,name)
        button.setText(self.__tr(text))
        self.connect(button, SIGNAL("clicked()"), callback)
        return button

    def updateInputPorts(self, ports):
        self.source.clear()
        for port in ports:
            self.source.insertItem(port)

    def slotApply(self):
        config = self.getConfig()
        
        filter = self.plot.getDataFilter()
        filter.clearFilters()
        filter.addFilter(config['source'])
        PlotSetup.PlotSetup.slotApply(self)

        
    def updateDataFields(self):
        pass
        
    def __tr(self,s,c = None):
        return qApp.translate("MoleculeEditorSetup",s,c)

    
    def getConfig(self):
        config = {}
        config['name'] = str(self.name.text())
        config['source'] = str(self.source.currentText())
        config['plotType'] = str(self.plotType.currentText())
        config['showIndex'] = str(self.showIndexBox.isChecked())
        config['showName'] = str(self.showNameBox.isChecked())
        return config

    def setConfig(self, config):
        self.name.setText(config["name"])

        for i in range(self.source.count()):
            self.source.setCurrentItem(i)
            if str(self.source.currentText()) == config["source"]:
                break

        try:
            for i in range(self.plotType.count()):
                self.plotType.setCurrentItem(i)
                self.slotChangePlotType(i)
                if str(self.plotType.currentText()) == config["plotType"]:
                    break
        except KeyError:
            pass

        try:
            if config['showIndex'] == "True":
                if not self.showIndexBox.isChecked():
                    self.showIndexBox.toggle()
            else:
                if self.showIndexBox.isChecked():
                    self.showIndexBox.toggle()
        except KeyError:
            pass
        
        try:
            if config['showName'] == "True":
                if not self.showNameBox.isChecked():
                    self.showNameBox.toggle()
            else:
                if self.showNameBox.isChecked():
                    self.showNameBox.toggle()
        except KeyError:
            pass

    def slotSelectAll(self):
        self.plot.selectAll()


    def addFillFreeBondsWithHydrogen(self, atoms, viewRight):
        """Add hydrogen atoms to all atoms in model with free bonds.

        atoms - List of atoms to add hydrogen
        viewRight - direction to use when direction is arbitrary
        """
        # imported here to enable instantiation of class without pymol
        from Library.Util import molecular
        
        # bondlengths in Angstrom
        hydrogenBondLength = {'Br':1.41,
                              'C':1.08,
                              'Cl':1.28,
                              'F':0.92,
                              'H':0.74,
                              'I':1.60,
                              'N':1.01,
                              'O':0.97,
                              'P':1.42,
                              'S':1.34,
                              'Si':1.48}                

        for atom in atoms:
            model = self.plot.getMoleculeForAtom(atom)
            thisIndex = model.index_atom(atom)
            attachedAtoms = map(lambda index: model.atom[index],
                                    molecular.getAttached(model, thisIndex))
            bonds = molecular.getRelatedBonds(model, thisIndex)
            usedValence = 0
            for b in bonds:
                usedValence += b.order
                
            try:
                numHydrogens = atom.get_implicit_valence()[0] - usedValence
                
                bondedAtoms = []
                bondDirections = []
                # to contain all results
                hydrogenDirections = []

                for b in bonds:
                    otherAtom = model.atom[b.index[b.index.index(thisIndex)-1]]
                    bondedAtoms.append(otherAtom)
                    direction = Vector3D(endPoint=otherAtom.coord,
                                         startPoint=atom.coord)
                    direction.normalize()
                    bondDirections.append(direction)
                    
                if numHydrogens == 1:
                    if len(bonds) == 0:
                        rotatedDirection = Vector3D(viewRight)
                    elif len(bonds) == 1:
                        if bonds[0].order < 3:
                            rotateAngle = 109.5
                        else:
                            rotateAngle = 180

                        rotatedDirection = Vector3D(bondDirections[0])
                        rotatedDirection.rotateAboutPerpendicular(rotateAngle*vtkMath.DegreesToRadians())
                            
                    elif len(bonds) == 2:
                        # when one of bonds is a double bond, place at
                        # center of greatest angle between existing
                        # bonds (which is also at the opposite end of
                        # the smallest angle
                        if (bonds[0].order > 1) or (bonds[1].order > 1):                          
                            rotatedDirection = Vector3D(bondDirections[0])
                            rotatedDirection.add(bondDirections[1])
                            rotatedDirection.negate()
                        else:
                            # otherwise place in a pyramid
                            perpendicular = bondDirections[0].cross(bondDirections[1])

                            centerDirection = Vector3D(bondDirections[0])
                            centerDirection.add(bondDirections[1])

                            rotationAxis = perpendicular.cross(centerDirection)

                            rotatedDirection = Vector3D(centerDirection)
                            rotatedDirection.rotateAboutAxis(rotationAxis, (180-(109.5/2))*vtkMath.DegreesToRadians())

                    elif len(bonds) == 3:
                        # take average coordinate of three other atoms, reflected through atom itself
                        # to get direction
                        rotatedDirection = Vector3D(bondDirections[0])
                        rotatedDirection.add(bondDirections[1])
                        rotatedDirection.add(bondDirections[2])
                        rotatedDirection.negate()

                    hydrogenDirections.append(rotatedDirection)

                elif numHydrogens == 2:
                    if len(bonds) == 0:
                        firstDirection = Vector3D(viewRight)
                        secondDirection = Vector3D(firstDirection)
                        secondDirection.rotateAboutPerpendicular(109.5*vtkMath.DegreesToRadians())
                    elif len(bonds) == 1:
                        # place at 109.5 from eachother in a pyramid
                        rotationAxis = Vector3D(bondDirections[0])
                        rotationAxis.rotateAboutPerpendicular(109.5*vtkMath.DegreesToRadians())
                        firstDirection = Vector3D(bondDirections[0])
                        firstDirection.rotateAboutAxis(rotationAxis, 120*vtkMath.DegreesToRadians())
                        secondDirection = Vector3D(bondDirections[0])
                        secondDirection.rotateAboutAxis(rotationAxis, -120*vtkMath.DegreesToRadians())
                    elif len(bonds) == 2:
                        #place at 109.5 from eachother with center at opposite of biggest angle
                        #between existing, and perpendicular to existing
                        perpendicular = bondDirections[0].cross(bondDirections[1])

                        centerDirection = Vector3D(bondDirections[0])
                        centerDirection.add(bondDirections[1])

                        rotationAxis = perpendicular.cross(centerDirection)

                        firstDirection = Vector3D(centerDirection)
                        firstDirection.rotateAboutAxis(rotationAxis, (180-(109.5/2))*vtkMath.DegreesToRadians())

                        secondDirection = Vector3D(centerDirection)
                        secondDirection.rotateAboutAxis(rotationAxis, (-180+(109.5/2))*vtkMath.DegreesToRadians())
                    hydrogenDirections.append(firstDirection)
                    hydrogenDirections.append(secondDirection)
                elif numHydrogens == 3:
                    if len(bonds) == 0:
                        # place at 109.5 from eachother ?
                        # should look better with 120 degrees
                        firstDirection = Vector3D(viewRight)
                        rotationAxis = Vector3D(firstDirection)
                        rotationAxis.rotateAboutPerpendicular(109.5*vtkMath.DegreesToRadians())
                        secondDirection = Vector3D(firstDirection)
                        secondDirection.rotateAboutAxis(rotationAxis, 120*vtkMath.DegreesToRadians())
                        thirdDirection = Vector3D(firstDirection)
                        thirdDirection.rotateAboutAxis(rotationAxis, -120*vtkMath.DegreesToRadians())
                    elif len(bonds) == 1:
                        # place at 109.5 from eachother, opposite existing
                        # rotate 120 degrees about axis of opposite bond to get
                        # correct angle
                        firstDirection = Vector3D(bondDirections[0])
                        firstDirection.rotateAboutPerpendicular(109.5*vtkMath.DegreesToRadians())
                        secondDirection = Vector3D(firstDirection)
                        secondDirection.rotateAboutAxis(bondDirections[0], 120*vtkMath.DegreesToRadians())
                        thirdDirection = Vector3D(firstDirection)
                        thirdDirection.rotateAboutAxis(bondDirections[0], -120*vtkMath.DegreesToRadians())
                        
                    hydrogenDirections.append(firstDirection)
                    hydrogenDirections.append(secondDirection)
                    hydrogenDirections.append(thirdDirection)
                elif numHydrogens == 4:
                    if len(bonds) == 0:
                        #place as tetraheder
                        fourthDirection = Vector3D(viewRight)
                        firstDirection = Vector3D(fourthDirection)
                        firstDirection.rotateAboutPerpendicular(109.5*vtkMath.DegreesToRadians())
                        secondDirection = Vector3D(firstDirection)
                        secondDirection.rotateAboutAxis(fourthDirection, 120*vtkMath.DegreesToRadians())
                        thirdDirection = Vector3D(firstDirection)
                        thirdDirection.rotateAboutAxis(fourthDirection, -120*vtkMath.DegreesToRadians())
                        hydrogenDirections.append(firstDirection)
                        hydrogenDirections.append(secondDirection)
                        hydrogenDirections.append(thirdDirection)
                        hydrogenDirections.append(fourthDirection)

                        
                for direction in hydrogenDirections:
                    hydrogenAtom = molecular.createAtom('H')
                    position = Point3D(atom.coord)
                    direction.normalize()
                    direction.multiply(hydrogenBondLength[atom.symbol])
                    position.add(direction)
                    hydrogenAtom.coord = position[0:3]
                    index = model.add_atom(hydrogenAtom)
                    bond = molecular.createBond(thisIndex, index, order=1)
                    model.add_bond(bond)
                    
            except Exception, e:
                traceback.print_exc()
                

    def slotAddHydrogen(self):

        sceneAtoms = []

        for actor in self.plot.picked:
            if hasattr(actor, "atom"):
                sceneAtoms.append(actor.atom)

        # get direction to right of camera, used for some bonds where "arbitrary" directions
        # are allowed
        cam = self.plot.renderer.GetActiveCamera()
        cam.OrthogonalizeViewUp()
        cam.ComputeViewPlaneNormal()
        viewUp = list(cam.GetViewUp())
        vtkMath.Normalize(viewUp)
        viewLook = cam.GetViewPlaneNormal()
        viewRight = [0.0,0.0,0.0]
        vtkMath.Cross(viewUp, viewLook, viewRight)
        vtkMath.Normalize(viewRight)

        # convert to Vector3D-objects
        viewRight = Vector3D(viewRight)

        self.addFillFreeBondsWithHydrogen(sceneAtoms, viewRight)

        self.updatePlotData()
        
        
    def slotHide(self):
        self.plot.hidePicked()

    def slotUnhideAll(self):
        self.plot.unhideAll()

    def slotInvert(self):
        self.plot.invertPicked()

    def slotExtendSelectionToModel(self):
        self.plot.extendSelectionToModel()

    def slotExtendSelectionAlongBonds(self):
        self.plot.extendSelectionAlongBonds()
        
    def slotDelete(self):
        # imported here to enable instantiation of class without pymol
        from Library.Util import molecular

        atoms = []
        for actor in self.plot.picked:
            if not hasattr(actor, "atom"):
                print "missing atom-attribute"
            else:
                atoms.append(actor.atom)

            self.node.applyCommand(lambda models : molecular.deleteAtoms(models,atoms))
        self.updatePlotData()

    def handleEvent(self,code,data):
        if code == "MDSTOPPED":
            if self.plot and self.plot.renderer and self.plot.renderWindow:
                self.plot.redrawPlot()

    def updatePlotData(self):
                
        if self.plot and self.plot.renderer and self.plot.renderWindow:
            self.plot.renderer.ResetCameraClippingRange()
            self.plot.redrawPlot()

    def slotMinimize(self):
        QMessageBox.information(self,
                            "Missing functionality",
                            "Not yet implemented.",
                            QMessageBox.Ok)
        return
    

    def slotSingleAttach(self):
        dialog = FragmentDialog(Config.singlefragmentdir, self.singlePointFragmentCallback, self)
        if dialog.count():
            dialog.exec_loop()
        else:
            QMessageBox.warning(self,
                                "No fragments available",
                                "No single attachment point fragments are available.",
                                QMessageBox.Ok)
            
    def slotDoubleAttach(self):
        if len(self.plot.picked) != 2:
            QMessageBox.warning(self,
                                "Needs two attachment points",
                                "To attach these fragments, you have to select exactly two attachment points.",
                                QMessageBox.Ok)
        else:
            dialog = FragmentDialog(Config.doublefragmentdir, self.doublePointFragmentCallback, self)
            if dialog.count():
                dialog.exec_loop()
            else:
                QMessageBox.warning(self,
                                    "No fragments available",
                                    "No fragments with dual attachment points are available.",
                                    QMessageBox.Ok)
                

    def singlePointFragmentCallback(self, iconViewItem):
        # imported here to be able to instantiate this class without pymol
        from chempy.pkl import PKL
        from Library.Util import molecular
        
        fragmentFile = iconViewItem.fragmentPath
        dialog = iconViewItem.dialog # evil hack to close dialog after doubleclicking items. must be a better way...
        dialog.accept()
        fragment = PKL().fromFile(fragmentFile)
        if self.plot.picked:
            targets = []
            for p in self.plot.picked:
                targets.append(p.atom)
            method = lambda models : molecular.connectToFragment(models,targets,fragment)
        else:
            method = lambda models : models.append(fragment.getModel())
        self.node.applyCommand(method)
        self.updatePlotData()

    def doublePointFragmentCallback(self, iconViewItem):
        fragmentFile = iconViewItem.fragmentPath
        dialog = iconViewItem.dialog # evil hack to close dialog after doubleclicking items. must be a better way...
        dialog.accept()
        fragment = PKL().fromFile(fragmentFile)
        
        if len(self.plot.picked) == 2:
            if fragment.getNumAttachmentPoints() == 2:
                targets = map(lambda p : p.atom, self.plot.picked)
                molA = self.plot.getMoleculeForAtom(targets[0])
                molB = self.plot.getMoleculeForAtom(targets[1])
                if molA != molB:
                    molA.merge(molB)
                replaceList = map(lambda atom : molA.index_atom(atom), targets)
                fragment.attachToMolecule(molA, replaceList)
                self.node.applyCommand(lambda models : models.remove(molB))
                self.updatePlotData()
            
    def rotatePickedAtoms(self):
        self.plot.originalPositions = {}

        for target in self.plot.picked:
            if hasattr(target, "atom"):
                atom = target.atom
                for k in (0,1,2):
                    center = target.GetCenter()
                    atom.coord[k] = center[k]
                # edits atom directly, won't work when middleware is more separated from gui
        self.updatePlotData()
    
    def movePickedAtoms(self, movedDistance):
        self.plot.originalPositions = {}

        for target in self.plot.picked:
            if hasattr(target, "atom"):
                atom = target.atom
                # edits atom directly, won't work when middleware is more separated from gui
                for k in (0,1,2):
                    atom.coord[k] += movedDistance[k]
        self.updatePlotData()
                        

    def slotAngle(self):
        self.plot.printPlaneAngleForSelectedFour()

    def slotChangePlotType(self, int):
        if self.plotType.currentText() == self.BNSTEXT:
            self.plot.setBallNStickPlot()
        elif self.plotType.currentText() == self.SFTEXT:
            self.plot.setSpaceFillingPlot()
        elif self.plotType.currentText() == self.LTEXT:
            self.plot.setLinePlot()

    def slotNameBoxToggled(self,on):
        self.plot.showName = self.showNameBox.isChecked()
        if self.plot.renderer and self.plot.renderWindow:
            self.plot.updateAnnotation()
            self.plot.renderWindow.update()
            self.plot.renderer.Render()

    def slotIndexBoxToggled(self,on):
        self.plot.showIndex = self.showIndexBox.isChecked()
        if self.plot.renderer and self.plot.renderWindow:
            self.plot.updateAnnotation()
            self.plot.renderWindow.update()
            self.plot.renderer.Render()


