# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from vtk import vtkBMPWriter, vtkJPEGWriter, vtkPNGWriter, vtkPNMWriter
from vtk import vtkPostScriptWriter, vtkTIFFWriter, vtkWindowToImageFilter
from qt import QAction, QCloseEvent, QDialog, QFileDialog, QFrame, QGridLayout
from qt import QIconSet, QKeySequence, QMainWindow, QMenuBar, QMessageBox
from qt import QPixmap, QPoint, QPopupMenu, QSplitter, QString, QToolBar
from qt import QWidget, Qt, SIGNAL, SLOT
import os, types

from GUI.Editors.ModeHandler import ModeHandler
from GUI.Editors.MoleculeEditorSetup import MoleculeEditorSetup
from GUI.Editors.MoleculeEditorWindowControl import MoleculeEditorWindowControl
from GUI.Pixmaps.Icons import Icons
from GUI.Plots import PlotRenderWindow, PlotConfig, Plot
from GUI.Plots.Plots3D.MolPlot3D.MolPlot3D import MolPlot3D
from GUI.Plots.interaction.CameraMode import CameraMode
from GUI.Plots.interaction.MoveMode import MoveMode
from GUI.Plots.interaction.PickMode import PickMode
from GUI.Plots.interaction.RotateMode import RotateMode
from GUI.Plots.interaction.SelectBoxMode import SelectBoxMode
from Library.Util.Config import Config

class MoleculeEditorWindow(QMainWindow):
    """This class will handle all the common elements of a molecule editor window GUI."""

    def __init__(self, node):
        QMainWindow.__init__(self,
                             None,
                             "MoleculeEditorWindow",
                             Qt.WType_TopLevel | Qt.WStyle_ContextHelp)
        
        self.setCaption(self.tr("Molecule Editor"))
        self.node = node
        
        # a timer for the duration of status messages.
        # Could be ignored, but holds a reasonable value for all
        # time dependent messages.
        self.statusTime = 3000

        # self.plotType = plotType
        # Storing the data to be plotted
        self.data = []

        self.renderWindow = None
        
        self.resize(500, 400)
        self.setCentralWidget(QWidget(self, "centralWidget"))
                
        self.controller = None
        
        self.plotList = []
        self.newRenderPosition = (0,0)

        self.plotLayout = QGridLayout(self.centralWidget(), 0, 0, 0, 0, "plotLayout")

        self.plot = None

        self.lineSplitter = QSplitter(self.centralWidget(), "lineSplitter")

        self.lineSplitter.setOrientation(QSplitter.Vertical)
        self.activeColSplitter = QSplitter(self.lineSplitter)
        self.activeColSplitter.setOrientation(QSplitter.Horizontal)

        self.plotLayout.addWidget(self.lineSplitter,0,0)
        
        self.statusBar = self.statusBar()

        # Add general and specific menus
        self.addWindowController()
        self.initToolbars()
        self.addModeSelectionItems()
        
        
        # self.show()
        # self.viewControllerAction.toggle()

        self.loadGuiSettings()
        
    def show(self):
        QMainWindow.show(self)

        if self.controller.isFinished():
            self.viewControllerAction.setOn(False)
        else:
            self.viewControllerAction.setOn(True)

    def setData(self, data):
        self.data = data

    def getData(self):
        return self.data


    def setActiveRenderer(self, renderer):
        self.renderWindow = renderer
            

    def addRenderWindows(self, rows = 1, columns = 1):
        """
        Params:
               - rows : int - Number of rows with plot renderers in this window
               - columns : int - Number of colums in each row

        Creates a matrix, with the specified dimension, with render windows."""

        if len(self.controller.getRenderWindows()) == 1:
            return

        amount = rows * columns

        # Crude hack to make the existing spiltters go away.
        for splitter in self.lineSplitter.children():
            if isinstance(splitter,QSplitter):
                point = QPoint()
                splitter.reparent(None, point)


        frame = QFrame(self, "frame")
        framelayout = QGridLayout(frame, 0, 0, 2, 0, "framelayout")
        render = PlotRenderWindow.PlotRenderWindow(self, self.controller, frame, "renderWindow")

        plot = MolPlot3D(lambda renderSetup : MoleculeEditorSetup(renderSetup, self.node))
        self.plot = plot
        self.renderWindow = render

        render.setupWidget.addPlot(plot)

        plot.renderWindow = render
        render.hidePlotList()
        plot.getSetupWidget(render.setupWidget).slotApply()

        self.controller.setRenderTab(render.getSetupWidget())
        
        framelayout.addWidget(render, 0, 0)

        self.renderWindow = render

        self.activeColSplitter = QSplitter(self.lineSplitter)
        self.activeColSplitter.setOrientation(QSplitter.Horizontal)
        self.activeColSplitter.show()

        point = QPoint()
        render.parent.reparent(self.activeColSplitter, point)
        render.parent.show()              
        
        # Move the controller to the bottom of the window.
        self.lineSplitter.moveToLast(self.controller)
       
        # Try not to grow the controller, it is very annoying
        self.lineSplitter.setResizeMode(self.controller, QSplitter.KeepSize)
        

    
    def addWindowController(self):
        # The controller needs a reference to this window
        self.controller = MoleculeEditorWindowControl(self, self.node, self.lineSplitter, "WindowControl")
        # Adding a default render window. This may be removed later for xml compability.
        if self.controller.plotSetup:
            self.controller.plotSetup.slotAddRenderWindows()


    def createAction(self, text, menuText, iconPixmap, toolTip, whatsThis, accelerator, callback, addTo):
        """Create and setup a new QAction for use as a button or toolbar element."""
        action = QAction(self)
        action.setText(self.tr(text))
        action.setMenuText(self.tr(menuText))
        if iconPixmap:
            action.setIconSet(QIconSet(QPixmap(iconPixmap)))
        if toolTip:
            action.setToolTip(self.tr(toolTip))
        if whatsThis:
            action.setWhatsThis(self.tr(whatsThis))
        if accelerator:
            if isinstance(accelerator, str):
                action.setAccel(QKeySequence(self.tr(accelerator)))
            elif isinstance(accelerator, QKeySequence):
                action.setAccel(accelerator)
        if isinstance(addTo, list):
            for target in addTo:
                action.addTo(target)
        elif addTo:
            action.addTo(addTo)
        if callback:
            if isinstance(callback, (list,tuple)):
                target = callback[0]
                slot = callback[1]
                self.connect(action, SIGNAL("activated()"), target, slot)
            else:
                self.connect(action, SIGNAL("activated()"), callback)
        return action

    def createToggleAction(self, text, menuText, iconPixmap, toolTip, whatsThis, accelerator, toggleCallback, addTo):
        action = self.createAction(text, menuText, iconPixmap, toolTip, whatsThis, accelerator, None, addTo)
        action.setToggleAction(True)
        if toggleCallback:
            if isinstance(toggleCallback, (list,tuple)):
                target = toggleCallback[0]
                slot = toggleCallback[1]
                self.connect(action, SIGNAL("toggled(bool)"), target, slot)
            else:
                self.connect(action, SIGNAL("toggled(bool)"), toggleCallback)
        return action

    def addModeSelectionItems(self):
        """
        Add icons and menuitems to select current editing mode."""
        self.modeHandler = ModeHandler(self)

        cameraModeAction = self.createToggleAction("Camera Mode", "Camera Mode", Icons.viewPic,
                                                   "Move mouse in editor to rotate view.",
                                                   "This tool allows the user to move around in the scene by moving the mouse combined with pushin the mousebuttons.",
                                                   None, self.cameraModeCallback, None)

        selectModeAction = self.createToggleAction("Select Box", "Select Box", Icons.selectBoxPic,
                                                   "Select all atoms within a boxed area of the screen.",
                                                   "This tool sets the current mode to Select Box. Dragging the mouse in the editor will then select all atoms within a selection rectangle.",
                                                   None, self.selectModeCallback, None)
        pickModeAction = self.createToggleAction("Pick Mode", "Pick Mode", Icons.pickPic,
                                                 "Pick atoms individually.", None,
                                                 None, self.pickModeCallback, None)

        moveModeAction = self.createToggleAction("Move Mode", "Move Atoms", Icons.movePic,
                                                 "Move all selected atoms interactively.",
                                                 "By clicking and moving the mouse on one of the selected atoms (or a single atom), you can move atoms interactively.",
                                                 None, self.moveModeCallback, None)

        rotateModeAction = self.createToggleAction("Rotate Mode", "Rotate Atoms", Icons.rotatePic,
                                                   "Rotate all selected atoms interactively.",
                                                   "By clicking and moving the mouse while having a series of atoms selected, the atoms will rotate about their joint center of mass.",
                                                   None, self.rotateModeCallback, None)
        self.modeHandler.addModeAction(cameraModeAction)
        self.modeHandler.addModeAction(selectModeAction)
        self.modeHandler.addModeAction(pickModeAction)
        self.modeHandler.addModeAction(moveModeAction)
        self.modeHandler.addModeAction(rotateModeAction)

    def pickModeCallback(self, active):
        if active:
            self.renderWindow.setActiveMode(PickMode(self.renderWindow))

    def cameraModeCallback(self, active):
        if active:
            self.renderWindow.setActiveMode(CameraMode(self.renderWindow))
        
    def selectModeCallback(self, active):
        if active:
            self.renderWindow.setActiveMode(SelectBoxMode(self.renderWindow, self.plot.selectWithinBox))

    def moveModeCallback(self, active):
        if active:
            renderSetupWidget = self.renderWindow.setupWidget
            plotSetup = self.plot.getSetupWidget(renderSetupWidget)
            self.renderWindow.setActiveMode(MoveMode(self.renderWindow,
                                                     self.plot.interactiveMove,
                                                     plotSetup.movePickedAtoms))

    def rotateModeCallback(self, active):
        if active:
            renderSetupWidget = self.renderWindow.setupWidget
            plotSetup = self.plot.getSetupWidget(renderSetupWidget)
            self.renderWindow.setActiveMode(RotateMode(self.renderWindow,
                                                       self.plot.interactiveRotate,
                                                       plotSetup.rotatePickedAtoms))


    def initToolbars(self):
        """
        Creates the default menu and toolbaritems for this window. This includes
        actions for saving and printing of the plot."""
        
        # Create a toolbar and a menubar for the window
        self.toolBar = QToolBar(self)
        self.menuBar = QMenuBar(self)
        self.plotToolBar = QToolBar(self)


        # Fill the menubar with popup menus
        self.fileMenu = QPopupMenu(self.menuBar)
        self.menuBar.insertItem(self.tr("&File"),self.fileMenu)

        # Create button and menuentry for saving data
        self.createAction("Save Plot Configuration ...",
                          "&Save Plot Configuration ...",
                          Icons.savePic,
                          "Save the configuration of the plot window to file...",
                          "Displays a file dialogue and lets you choose a filename to save the plot configuration to. When you restart SciCraft, a new plot will appear in the \"Custom\" menu, and you can use the setup you saved instantly.",
                          "CTRL+S",
                          self.saveConfig,
                          [self.fileMenu, self.toolBar])

        #Create button and menuentry for loading data
        self.createAction("Load Plot Configuration ...",
                          "&Load Plot Configuration ...",
                          Icons.openPic,
                          "Load a plot configuration from file",
                          "Displays a file dialogue and lets you choose a filename to load the plot configuration from.",
                          "CTRL+L",
                          self.loadConfig,
                          [self.fileMenu, self.toolBar])

        #Create button end menuentry for saving plot image
        self.createAction("Save Plot ...",
                          "S&ave Plot ...",
                          Icons.plotSavePicturePic,
                          "Save the plot(s) in the active render window as a picture",
                          "Displays a file dialogue and lets you choose a filename and an image type for the plot to be saved in.",
                          "CTRL+G",
                          self.savePicture,
                          [self.fileMenu, self.toolBar])
        #Create button and menuentry for fullscreen mode
        fsa = self.createToggleAction("Fullscreen Mode",
                                      "Fullscreen Mode",
                                      Icons.fullscreenPic,
                                      "Fullscreen. Toggle with ALT+Return",
                                      "Expands the plots to cover the whole screen (Removing most of the window decorations). Toggle this mode by clicking ALT+Return",
                                      "Alt+Return",
                                      self.slotToggleFullscreen,
                                      self.plotToolBar)
        fsa.setOn(False)

        self.viewControllerAction = self.createToggleAction("Window Controller",
                                                            "Window Controller",
                                                            Icons.plotControlPic,
                                                            "Show/hide window controller.",
                                                            "Shows and hides the plot window controller in the bottom of this window.",
                                                            None,
                                                            self.slotToggleController,
                                                            self.plotToolBar)

        self.viewControllerAction.setOn(True)
        
        # Exit window action should be separated
        self.fileMenu.insertSeparator()

        self.createAction("Close Window",
                          "Close",
                          None,
                          "Close the window",
                          "Close the window (obviously)",
                          "CTRL+Q",
                          (self, SLOT("close()")),
                          self.fileMenu)
                          

        self.helpMenu = QPopupMenu(self.menuBar)
        self.menuBar.insertItem("&Help", self.helpMenu)

        self.createAction("Help",
                          "Help",
                          None,
                          "Display help pages for plotting",
                          None,
                          "F1",
                          self.slotShowHelp,
                          self.helpMenu)


    def savePicture(self, defaultFileName = ""):
        """
        Params:
               - defaultFileName : String - The filename to be secified when the user
               is presentet to the file save dialog. This parameter can very well be left
               unspecified, it has some value when this method is called recurively (rare)

        This method is connected to the "save as picture" button in the plot
        window. It will open a file dialog and try to save the current plot to
        a file specified by the user. The user can choose from several different image formats,
        as well as "by extension". When the latter is chosen, the method will find the image format
        by looking at the extension (file suffix) given by the user. This method is fairly
        robust, and will try to recover from any mistake made by the user. It will a.o. check
        for file name validity and prompt the user before overwriting anything."""

        # The the last directory the user visited, usually the user would
        # want to save here
        lastdir = Config.homedir

        # Check if the last used dir still exists,
        # otherwise continue to use the home directory
        if os.path.isdir(PlotConfig.PlotConfig.lastsavedir):
            lastdir = PlotConfig.PlotConfig.lastsavedir
        
        # The valid image formats gif may be added later
        validSuffices = {".jpeg": "JPEG Image (*.jpg *.jpeg)",
                         ".jpg" : "JPEG Image (*.jpg *.jpeg)",
                         ".bmp" : "Bitmap Picture (*.bmp)",
                         ".png" : "Portable Network Graphics (*.png)",
                         ".pnm" : "Portable Anymap (*.pnm)",
                         ".ps"  : "Encapsulated Postscript (*.ps *.eps)",
                         ".eps" : "Encapsulated Postscript (*.ps *.eps)",
                         ".tif" : "Tag Image File Format (*.tif *.tiff)",
                         ".tiff": "Tag Image File Format (*.tif *.tiff)"}
                        

        # Remember the file selected
        selectedFile = defaultFileName
        
        # The file save dialog
        fileDialog = QFileDialog(lastdir,
                                 str(self.tr("By extension (*)")),
                                 self,
                                 "fileDialog",
                                 1)

        # Add the rest of the filters
                
        fileDialog.addFilter(validSuffices[".bmp"])
        fileDialog.addFilter(validSuffices[".jpg"])
        fileDialog.addFilter(validSuffices[".png"])
        fileDialog.addFilter(validSuffices[".pnm"])
        fileDialog.addFilter(validSuffices[".eps"])
        fileDialog.addFilter(validSuffices[".tif"])
        
        # Set the selected file to the previous seleced file
        fileDialog.setSelection(selectedFile)
        fileDialog.setMode(QFileDialog.AnyFile)
        
        # Remember the last file filter used
        fileDialog.setSelectedFilter(PlotConfig.PlotConfig.plotPictureSaveFilter)
        fileDialog.setCaption("Save plot as picture")
        
        # User has pushed the save button
        if fileDialog.exec_loop() == QDialog.Accepted:
            
            fileFilter = fileDialog.selectedFilter()
            PlotConfig.PlotConfig.plotPictureSaveFilter = fileFilter
            fullFileName = fileDialog.selectedFile()
                        
                
        # User has cancelled saving
        else:
            self.setStatusMessage(self.tr("Save to picture cancelled"), self.statusTime)
            return
        
        # Strip the name from the file path
        filePath, fileName = os.path.split(str(fullFileName))
        PlotConfig.PlotConfig.lastsavedir = filePath
        PlotConfig.writeConfig()

        # Strip away the suffix (if any). Will be used for testing
        prefix, suffix = os.path.splitext(fileName)

        # if user has entered an invalid suffix, stuff
        # it together with the file prefix and let the
        # file filter decide the format
        if not suffix in validSuffices:
            prefix = prefix + suffix
            suffix = ""

        # if the user has not enterd a valid suffix but want to save by an extension,
        # send message to user
        if str(fileFilter) == str(self.tr("By extension (*)")) and not suffix:

            # I'm bored
            message = str(self.tr(" is an invalid image file type."))
            if os.path.splitext(prefix)[1] == ".gif":
                message += str(self.tr(" Future GIF image support is possible and probable."))
            message += "\n"
                        
            retval = QMessageBox.warning(self,
                                    "SciCraft",
                                    fileName +
                                    message +
                                    str(self.tr("Do you want to abort the save or specify a file with a valid type?")),
                                    QMessageBox.Retry,
                                    QMessageBox.Abort,
                                    QMessageBox.NoButton)

            # User aborts the save
            if retval == QMessageBox.Abort:
                self.setStatusMessage(self.tr("Save to picture cancelled"), self.statusTime)
                return
            # User retries. Call this method again
            else:
                self.savePicture(prefix)
                return
                
        # User has entered a valid suffix and save by extension.
        # Find the right filter.
        elif str(fileFilter) == str(self.tr("By extension (*)")):

            # Not necessary to make i QString, but lets keep
            # the type consistent throughout the method
            fileFilter = QString(validSuffices[suffix])

        # Check for the chosen filter, create the appropriate writer
        # and add the correct file suffix to the file
        if str(fileFilter) == validSuffices[".bmp"]:
            writer = vtkBMPWriter()
            if not suffix == ".bmp":
                prefix = prefix + suffix
                suffix = ".bmp"

        elif str(fileFilter) == validSuffices[".jpg"]:
            writer = vtkJPEGWriter()
            if not (suffix == ".jpg" or suffix == ".jpeg"):
                prefix = prefix + suffix
                suffix = ".jpg"

        elif str(fileFilter) == validSuffices[".png"]:
            writer = vtkPNGWriter()
            if not suffix == ".png":
                prefix = prefix + suffix
                suffix = ".png"

        elif str(fileFilter) == validSuffices[".pnm"]:
            writer = vtkPNMWriter()
            if not suffix == ".pnm":
                prefix = prefix + suffix
                suffix = ".pnm"

        elif str(fileFilter) == validSuffices[".eps"]:
            writer = vtkPostScriptWriter()
            if not (suffix == ".eps" or suffix == ".ps"):
                prefix = prefix + suffix
                suffix = ".eps"

        elif str(fileFilter) == validSuffices[".tif"]:
            writer = vtkTIFFWriter()
            if not (suffix == ".tif" or suffix == ".tiff"):
                prefix = prefix + suffix
                suffix = ".tif"
        
        # This should not occur at all
        else:
            QMessageBox.error(fileDialog,
                              "SciCraft",
                              self.tr("Found no matching file filter. This is a BUG! Please file a bug report. The image will not be saved."),
                              QMessageBox.Ok,
                              QMessageBox.NoButton,
                              QMessageBox.NoButton)
            
            return

        
        # Check if file already exists. If it does, prompt user
        # for advice
        if os.path.isfile(os.path.join(filePath, prefix + suffix)):
            retval = QMessageBox.warning(fileDialog,
                                         "SciCraft",
                                         str(self.tr("Overwrite ")) + prefix + suffix + "?",
                                         QMessageBox.Yes,
                                         QMessageBox.No,
                                         QMessageBox.NoButton)


            # User won't overwrite, recall this method
            if retval == QMessageBox.No:
                self.savePicture(prefix + suffix)
                return

        # Ready to write the file
        # self.repaint()

        imageFilter = vtkWindowToImageFilter()

        # This line makes the plot renderer rerender the plot before
        # saving it to file. If this call is not made, a picture of the
        # save dialog itself is written to file.
        self.renderWindow._RenderWindow.Render()

        imageFilter.SetInput(self.renderWindow._RenderWindow)

        # Just make sure
        imageFilter.Update()
        
        writer.SetInput(imageFilter.GetOutput())
        writer.SetFileName(os.path.join(filePath, prefix + suffix))

        self.renderWindow.Render()

        # Things could go wrong when writing the file
        try:

            # VTK is not very good at throwing exceptions. To catch some of the
            # errors that could occur when trying to write a file, a test file
            # is first written, by python, in the place of the image file.
            # If this goes allright, vtk will overwrite the file with
            # the actual image file. This method is not bullet proof, and some
            # underlying assumptions must hold if the program should be able to
            # avoid a crash.
            testFile = open(os.path.join(filePath, prefix + suffix), "w")
            testFile.write("SciCraft: Should contain image data! Something is very wrong :(")
            testFile.close()

            # VTK
            writer.Write()
        # This is the exception that the python file handlig would throw
        except IOError, e:

            # Give the user a new set of choices
            retval = QMessageBox.critical(self,
                                         "SciCraft",
                                         "Could not save " + str(fullFileName) + ":" +
                                         e.strerror + "!\n"+
                                         "Do you want to try another location or abort the save?" ,
                                         QMessageBox.Retry,
                                         QMessageBox.Abort,
                                         QMessageBox.NoButton)
            
            # Retry the save, call this method again
            if retval == QMessageBox.Retry:
                # Recursive call
                self.savePicture(prefix + suffix)
                return

            # User gives up 
            else:
                self.setStatusMessage(self.tr("Save to picture cancelled"), self.statusTime)
                return

        self.setStatusMessage(self.tr("Plot saved to " + os.path.join(filePath, prefix + suffix)),
                              self.statusTime)

    def setStatusMessage(self, text, time = 0):
        """
        Params:
               - text : String - The message to be displayed
               - time : int - An optional time limit for the message to be shown.
               If 0 is specified (default) the message will be displayed until another
               message is displayed

        Displays a text message in the plot windows statusbar."""
        
        if time:
            self.statusBar.message(text, time)
        else:
            self.statusBar.message(text)
    
    def slotToggleController(self, on):
        if on:
            self.controller.show()
        else:
            self.controller.hide()

    def slotToggleFullscreen(self, on):
        if on:
            self.viewControllerAction.setOn(False)
            self.toolBar.hide()
            self.menuBar.hide()
            self.statusBar.hide()
            self.showFullScreen()
        else:
            #self.slotToggleController(0)
            self.toolBar.show()
            self.menuBar.show()
            self.statusBar.show()
            self.showNormal()

    def closeEvent(self, event):
        """
        Params:
               - event : QCloseEvent? - Event sent from Qt when user clicks the close
               button on the window frame.
        Redefines the closeEent method in the QMainWindow class. A plot window
        should never be closed as such, but just hidden. This method therefore
        only hides the window when it is instructed to close."""
        
        self.hide()
        
        # QMainWindow.closeEvent(self, event)
        
    def saveConfig(self):
        self.controller.saveConfig()

    def loadGuiSettings(self):
        """Get GUI settings as defined in module diagram file."""
        lambdaish = self.node.getCustomXMLasLambda()
        if lambdaish:
            lamb = lambdaish.plotconfig

            # Make sure render window and plot is created
            self.addRenderWindows()
        
            if lamb.__dict__.has_key("renderer"):
                config = lamb.renderer
                if not type(lamb.renderer) == types.ListType:
                    lamb.renderer = [lamb.renderer]
                for config in lamb.renderer:
                    renderSetupWidget = self.renderWindow.setupWidget

                    if hasattr(config,"plot"):
                        if not type(config.plot) == types.ListType:
                            config.plot = [config.plot]
                        # traverse the list of plots and create them
                        for plot in config.plot:
                            if plot.type == "MolPlot3D":

                                # make dictionary from the lambda expression
                                dict = {}
                                for key in plot.config.__dict__:
                                    dict[key] = getattr(plot.config,key)

                                # set the plots configuration
                                self.plot.getSetupWidget(renderSetupWidget).setConfig(dict)
                                self.plot.getSetupWidget(renderSetupWidget).slotApply()
                    
            self.setStatusMessage("Config loaded" , self.statusTime)
        else:
            # TODO - report error
            pass


    def loadConfig(self, filename=None):
        custompath = os.path.join(Config.userconfigdir, "Plots")
        if not os.path.isdir(custompath) :
            os.mkdir(custompath, 0755)
        custompath = os.path.join(custompath, "Custom")
        if not os.path.isdir(custompath) :
            os.mkdir(custompath, 0755)

        if not filename:
            filename = str(QFileDialog.getOpenFileName(custompath,
                                                       "SciCraft plot files (*.zpl)",
                                                       self,
                                                       "load file dialog",
                                                       "Choose a file"))
        
        if len(filename) != 0 and filename.rfind(".zpl") > 0:
            from xml.dom import minidom
            dom = minidom.parse(filename)
            lamb = xml2lambda([dom]).plotconfig

            if hasattr(lamb, "renderer"):
                if not type(lamb.renderer) == types.ListType:
                    lamb.renderer = [lamb.renderer]
                for config in lamb.renderer:
                    if config.__dict__.has_key("plot"):
                        renderSetupWidget = self.renderWindow.setupWidget
                        if not type(config.plot) == types.ListType:
                            config.plot = [config.plot]
                        # traverse the list of plots and create them
                        for plot in config.plot:
                            if plot.type == "MolPlot3D":

                                # make dictionary from the lambda expression
                                dict = {}
                                for key in plot.config.__dict__:
                                    dict[key] = getattr(plot.config,key)

                                # set the plots configuration
                                self.plot.getSetupWidget(renderSetupWidget).setConfig(dict)
                                self.plot.getSetupWidget(renderSetupWidget).slotApply()
            self.setStatusMessage("Config loaded from " + filename, self.statusTime)
        else:
            self.setStatusMessage("Load operation cancelled", self.statusTime)

    def saveNodeDiagram(self):
        self.controller.saveNodeDiagram()

    def slotShowHelp(self):
        pass

def xml2lambda(elementlist):

    def handleElement(element):
            
        
            if hasattr(element,"data"):
                # element is a textnode. Return text right away
                return element.data

            stringValue = ""
            returnvalue = lambda : None

            # check each child this element has
            for child in element.childNodes:

                # iterate this over this element                
                value = handleElement(child)

                # check if checked node was a textNode
                if child.nodeType == child.TEXT_NODE:
                    stringValue = value

                elif hasattr(returnvalue,str(child.localName)):
                    # the node already has checked an occurence of this childNode

                    # get the attribute
                    attr = getattr(returnvalue,str(child.localName))

                    # if not list, make it a list
                    if not isinstance(attr,list):
                        attr = [attr]

                    # add the value, and put it back on the returnvalue object
                    attr.append(value)
                    setattr(returnvalue,str(child.localName),attr)

                else:
                    # new node, just adding the variable to the dict
                    setattr(returnvalue,str(child.localName),value)

            if returnvalue.__dict__:
                # object had non-text childNodes. Return parsed input from these
                return returnvalue
            else:
                # object had no non-text childNodes, return text node if any, else return empty string
                return stringValue


    # if element exists, parse its first element
    if elementlist:
                
            # call recursive method.
            return handleElement(elementlist[0])

    # no object to play with, returing nothing :(
    return None     
 
