#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QDialog, QGridLayout, QListBox, QMainWindow, QMessageBox, QObject
from qt import QPushButton, QVBoxLayout, SIGNAL
from qtui import QWidgetFactory
from GUI.guinode import GUINode
from Library.Util.gui import size
from Library.Util.Config import Config
import os

class Gaussdal_gui_node(GUINode):
    def __init__(self, parent, node):
        self.node = node
        self.parent = parent
        self._dialog = None
        GUINode.__init__(self, parent, node)

    def nodeDialog(self):
        if not self._dialog:
            try:
                import vtk
                self._dialog = MoleculeMapper(self.parent, self.node)
            except ImportError:
                QMessageBox.warning(None, "Missing module VTK",
                                    "You need VTK with python bindings installed to use Gaussdal node.")
        elif not self._dialog.is_shown():
            self._dialog.show()        

    def procrustes(self):
        reference_molecule_id = self.molecule_mapper.left_molecule_view.molecules.current_index()
        self.node.procrustes(reference_molecule_id)

    def _update_molecules(self):
        self.node.update_molecules()
        self.molecule_mapper.left_molecule_view.render_widget.display_new_model()
        self.molecule_mapper.right_molecule_view.render_widget.display_new_model()

        # update global selection list box as well
        self.molecule_mapper.control_panel.set_global_list_box()

        # and molecule_view list boxes
        self.molecule_mapper.left_molecule_view.set_selection_widget()
        self.molecule_mapper.right_molecule_view.set_selection_widget()

class MoleculeMapper:
    """Main user interface"""
    def __init__(self, parent, node):
        self.node = node
        self.parent = parent
        self.page = -1
        self._page_index_in_stack = [0, 1, 1, 2]
        self.render_widget = None
        self._ref_render_widget = None
        self._test_render_widget = None
        self.ref_mol_index = 1

        self.window = QWidgetFactory.create(
            os.path.join(Config.uidir, 'molecule_mapper.ui'),
            parent,
            parent)
        
        
        # Back & Next buttons
        _next_button = self.window.child("qt_central_widget").child("next_button")
        _back_button = self.window.child("qt_central_widget").child("back_button")
        QObject.connect(_next_button,
                        SIGNAL("clicked()"),
                        self._next_page)
        QObject.connect(_back_button,
                        SIGNAL("clicked()"),
                        self._previous_page)

        # Next molecule button
        # Browsing through test molecules
        QObject.connect(self.window.child('next_molecule_button'),
                        SIGNAL('clicked()'),
                        self._next_test_molecule)

        # Layouts for the render widgets
        self._mol_sel_frame_layout = QGridLayout(
            self.window.child('ref_mol_select_frame'), 1, 1)
        self._ref_frame_layout = QGridLayout(
            self.window.child('ref_mol_map_frame'), 1, 1)
        self._test_frame_layout = QGridLayout(
            self.window.child('test_mol_map_frame'), 1, 1)

        self._next_page()

        # Hack
        self.window.event = self.event

        self.window.show()
        
    def event(self, e):
        """Hack"""

    def show(self):
        self.window.show()

    def is_shown(self):
        return self.window.isShown()

    def _next_page(self):
        self.page += 1
        if self.page == 0:
            self._settings_page()
        elif self.page == 1 and self.node.manual_mapping:
            self._molecule_select_page()
        elif self.page == 1 and (not self.node.manual_mapping or self.node.input_port_mapping):
            self.page -= 1
            self._done_button_action()
        elif self.page == 2:
            self._atom_select_page()
        elif self.page == 3:
            self._mapping_page()
        elif self.page == 4:
            self._finalize_mapping_wizard()
            self.page -= 1

    def _previous_page(self):
        self.page -= 1
        if self.page == 0:   # Settings page
            self.window.child("back_button").setEnabled(0)
        if self.page == 1:  # Molecule select page
            self.window.child('mol_list_box').setEnabled(1)
            self.window.child('select_state').setText("Choose Reference Molecule")
        self.window.child("widget_stack").raiseWidget(self._page_index_in_stack[self.page])

    def _set_default_mapping(self):
        self.window.child("next_button").setText("Done")
        self.node.set_default_mapping()

    def _set_manual_mapping(self):
        self.window.child("next_button").setText("Next")
        self.node.set_manual_mapping()

    def _set_port_mapping(self):
        self.window.child("next_button").setText("Done")
        self.node.set_port_mapping()

    def _done_button_action(self):
        self.window.close()

    def _finalize_mapping_wizard(self):
        answer = QMessageBox.question(self.window,
                             "Mapping wizard done",
                             "Do you want to run the procrustes routine?",
                             QMessageBox.Yes, QMessageBox.No)
        if answer == QMessageBox.Yes:
            self.node.procrustes(self.ref_mol_index)
            procrustes_view = Procrustes_view(self.window, self.node)
        elif answer == QMessageBox.No:
            self._done_button_action()

    def _settings_page(self):
        """Raises the settings page of the QWidgetStack"""
        #Signals
        QObject.connect(self.window.child("default_mapping"),
                        SIGNAL("clicked()"),
                        self._set_default_mapping)
        QObject.connect(self.window.child("manual_mapping"),
                        SIGNAL("clicked()"),
                        self._set_manual_mapping)
        QObject.connect(self.window.child("input_port"),
                        SIGNAL("clicked()"),
                        self._set_port_mapping)

        # Data column selection
        # I use a QListBox here because if the node reads data out of a GaussDal
        # database, the titles of the columns may be displayed in the QListBox.
        # LOCAL - per atom attributes
        _local_list_box = self.window.child("qt_central_widget").child("local_list_box")
        QObject.connect(_local_list_box,
                        SIGNAL("selectionChanged()"),
                        self._update_local_selection)
        _global_list_box = self.window.child("qt_central_widget").child("global_list_box")
        QObject.connect(_global_list_box,
                        SIGNAL("selectionChanged()"),
                        self._update_global_selection)
        if self.node.col_names:
            for name in self.node.col_names:
                _local_list_box.insertItem(str(name))
                _global_list_box.insertItem(str(name))
        for i in self.node.local_data_cols:
            _local_list_box.setSelected(i, 1)

        _molecule_id = self.window.child("qt_central_widget").child("combo_molecule_id")
        _atomic_number = self.window.child("qt_central_widget").child("combo_atomic_number")
        _x_coordinate = self.window.child("qt_central_widget").child("combo_x_coordinate")
        _y_coordinate = self.window.child("qt_central_widget").child("combo_y_coordinate")
        _z_coordinate = self.window.child("qt_central_widget").child("combo_z_coordinate")

        if self.node.is_executed:        
            _molecule_id.insertStrList(self.node.col_names)
            _atomic_number.insertStrList(self.node.col_names)
            _x_coordinate.insertStrList(self.node.col_names)
            _y_coordinate.insertStrList(self.node.col_names)
            _z_coordinate.insertStrList(self.node.col_names)

            # atom.moleculeid
            _molecule_id.setCurrentItem(self.node.col_mol_id)
            QObject.connect(_molecule_id,
                            SIGNAL("highlighted(int)"),
                            self._change_moleculeid_col)
            # atom.atomicnumber
            _atomic_number.setCurrentItem(self.node.col_atom_no)
            QObject.connect(_atomic_number,
                            SIGNAL("highlighted(int)"),
                            self._change_n_atomic_col)

            # coordinate.{x, y, z}
            _x_coordinate.setCurrentItem(self.node.col_x)
            _y_coordinate.setCurrentItem(self.node.col_y)
            _z_coordinate.setCurrentItem(self.node.col_z)
            QObject.connect(_x_coordinate,
                            SIGNAL("highlighted(int)"),
                            self._change_x_col)
            QObject.connect(_y_coordinate,
                            SIGNAL("highlighted(int)"),
                            self._change_y_col)
            QObject.connect(_z_coordinate,
                            SIGNAL("highlighted(int)"),
                            self._change_z_col)

    def _molecule_select_page(self):
        from GUI.Nodes.gaussdalnodevtk import HybridRenderWidget

        self.node.update_molecules()

        # Enable the Back button
        self.window.child('back_button').setEnabled(1)

        self.window.child('widget_stack').raiseWidget(1)
        _frame = self.window.child("ref_mol_select_frame")

        # Render widget
        if self.render_widget:
            self.render_widget.hide()
            del self.render_widget
            
        self.render_widget = HybridRenderWidget(_frame, self.node)
        self._mol_sel_frame_layout.addWidget(self.render_widget, 0, 0)
        self.render_widget.show()

        _mol_select_list_box = self.window.child("mol_list_box")
        _mol_select_list_box.clear()

        if self.node.is_executed:
            for molecule in self.node.molecule_list:
                _mol_select_list_box.insertItem(str(molecule.id)) 
            _mol_select_list_box.setCurrentItem(_mol_select_list_box.topItem())
        QObject.connect(_mol_select_list_box,
                     SIGNAL("highlighted(int)"),
                     self._select_reference_molecule)

    def _atom_select_page(self):
        self.render_widget.enter_selection_mode()
        self.window.child('mol_list_box').setDisabled(1)
        self.window.child('select_state').setText("Select Reference Atoms")

    def _mapping_page(self):
        # imported here to make vtk optional
        from GUI.Nodes.gaussdalnodevtk import Render_widget, InteractionRenderWidget

        self.window.child('widget_stack').raiseWidget(2)
        # Ref molecule
        if self._ref_render_widget:
            self._ref_render_widget.hide()
            del self._ref_render_widget
        _ref_frame = self.window.child('ref_mol_map_frame')
        self._ref_render_widget = Render_widget(_ref_frame, self.node)
        self._ref_frame_layout.addWidget(self._ref_render_widget, 0, 0)
        self._ref_render_widget.show()
        # Test molecule
        if self._test_render_widget:
            self._test_render_widget.hide()
            del self._test_render_widget
        _test_frame = self.window.child('test_mol_map_frame')
        self._test_render_widget = InteractionRenderWidget(_test_frame,
                                                           self.node,
                                                           self._ref_render_widget)

        #-1 because without reference molecule
        self._total_steps = len(self.node.molecule_list)-1
        #self._progress_bar = qt.QProgressBar(self._total_steps, self.window)
        #self._progress_bar = self.window.child("qt_central_widget").child("progress_bar")
                
        #QObject.connect(self.window.child('next_molecule_button'),
        #                SIGNAL('clicked()'),
        #                self._test_render_widget.redisplay_model)
        self._mol_select_counter = 0
        #self._progress_bar.reset()


        # Undo button
        QObject.connect(self.window.child('undo'),
                        SIGNAL('clicked()'),
                        self._test_render_widget.undo_last_selection)

        # Clear selection button
        QObject.connect(self.window.child("push_button_clear_selection"),
                        SIGNAL("clicked()"),
                        self._test_render_widget.clear_selection)

        self._test_frame_layout.addWidget(self._test_render_widget, 0, 0)
        self._test_render_widget.show()


    def _next_test_molecule(self):
        """Displays the next test molecule if the selecting of the current
        test molecule is finished.
        """
        reference_molecule = self.node.molecule_browser.get_reference_molecule()
        current_molecule = self.node.molecule_browser.current()
        if len(reference_molecule.mapping) == len(current_molecule.mapping):
            self._test_render_widget.redisplay_model()
            self._mol_select_counter += 1
            #self._progress_bar.setIndicator(str(self._mol_select_counter)+"\/"+str(self._total_steps),
            #                            self._mol_select_counter, self._total_steps)
            #self._progress_bar.setProgress(self._mol_select_counter, self._total_steps)
            
        else:
            error_string = "Inconsistent number of test atoms selected. Please \
select the same number of test atoms as\nthe number of selected reference atoms \
before proceeding to the next test molecule."
            QMessageBox("Mapping Error", error_string, QMessageBox.Critical, QMessageBox.Ok, QMessageBox.NoButton, QMessageBox.NoButton, self.window).show()

    def set_global_list_box(self):
        """GLOBAL - per molecule attributes"""
        # Clear list box
        self.panel.child('global_list_box').clear()

    def _update_global_selection(self):
        selected = []
        list_box = self.window.child("qt_central_widget").child("global_list_box")
        #list_box = self.panel.child('global_list_box')
        for i in range(list_box.count()):
            if list_box.isSelected(i):
                selected.append(i)
                #selected.append(list_box.index(i))
                #selected.append(list_box.text(i).toInt()[0])
        self.node.global_data_cols = selected

    def _update_local_selection(self):
        selected = []
        list_box = self.window.child("local_list_box")
        for i in range(list_box.count()):
            if list_box.isSelected(i):
                selected.append(i)
        self.node.local_data_cols = selected

    def _next_molecule_action(self):
        self._test_render_widget.redisplay_model()
        #self._progress_bar.setIndicator(str(self._mol_select_counter)+"\/"+str(self._total_steps),
        #                                self._mol_select_counter, self._total_steps)
        self._progress_bar.setProgress(self._mol_select_counter, self._total_steps)
        self._mol_select_counter += 1

    def _change_moleculeid_col(self, index):
        self.node.col_mol_id = index

    def _change_n_atomic_col(self, index):
        self.node.col_atom_no = index

    def _change_x_col(self, index):
        self.node.col_x = index

    def _change_y_col(self, index):
        self.node.col_y = index

    def _change_z_col(self, index):
        self.node.col_z = index

    def _select_reference_molecule(self, n):
        self.node.molecule_browser.set_reference_molecule_by_index(n)
        self.render_widget.redisplay_model(n)



class Molecule_view(QDialog):
    """A VTK render widget on the top and a QListBox
    on the bottom for selecting molecules"""
    def __init__(self, parent, node):
    #def __init__(self, parent, is_reference=1):
        QMainWindow.__init__(self, parent)
        #QVBox.__init__(self, parent.hor_layout)
        self.layout = QVBoxLayout(self)
        
        
        #self.show()
        
        self.parent = parent
        is_reference = 1
        self.is_reference = is_reference
        self.node = node
        self.molecules = self.node.get_molecules()
        
        # label on the top
        #top_label = QLabel(self._title(is_reference), self)
        #self.insertChild(top_label)
        #layout.setStretchFactor(top_label, 1)
        #layout.setSpacing(50)
        #top_label.show()
        # Render widget

        # imported here to make vtk optional
        from GUI.Nodes.gaussdalnodevtk import Render_widget
        init_mol_index = 1
        if is_reference:
            init_mol_index = 0
        self.render_widget = Render_widget(self, self.node)
        #self.render_widget.show()                

        # label in the middle
        #middle_label = QLabel("Molecule Selection by 'moleculeid'", self)
        #self.insertChild(middle_label)
        #layout.setStretchFactor(middle_label, 1)
        #layout.setSpacing(50)
        #middle_label.show()
        # list box on the bottom
        self._mol_select = QListBox(self)
                
        self.set_selection_widget()
        #self._mol_select.show()
        #layout.addWidget(top_label)
        self.layout.addWidget(self.render_widget)
        #layout.addWidget(middle_label)
        self.layout.addWidget(self._mol_select)
        #self.render_widget.show()
        self.resize(400, 400)
        self.show()
        
        #self.setCentralWidget(layout)
    #def add_renderer(self):
        #self.render_widget.add_renderer()

    def _title(self, is_reference):
        if is_reference:
            return "Reference Molecule"
        else:
            return "Test Molecule"

    def _display_model(self, n):
        self.render_widget.redisplay_model(n)

    def set_selection_widget(self):
        '''list box on the bottom'''
        self._mol_select.clear()
        for molecule in self.node.molecule_list:
            self._mol_select.insertItem(str(molecule.id)) 
        self._mol_select.setCurrentItem(self.molecules.current_index())
        self._mol_select.setMaximumHeight(100)
        
class Mapper_view(Molecule_view):
    """An interactive molecule mapper view"""
    def __init__(self, parent, is_reference=1):
        Molecule_view.__init__(self, parent, is_reference)
        self.connect(self._mol_select,
                     SIGNAL("highlighted(int)"),
                     self._display_model)

class Procrustes_view(Molecule_view):
    """A view of a selection of molecules all in a single renderer.
    
    Similar to Molecule_view, only that it has an 'Update' button on the bottom
    for procrustes-izing of the molecules
    """
    def __init__(self, parent, node):
        Molecule_view.__init__(self, parent, node)
        self._mol_select.setSelectionMode(QListBox.Multi)
        _update_button = QPushButton("Update", self)
        self.connect(_update_button,
                     SIGNAL("clicked()"),
                     self._update)
        self.layout.addWidget(_update_button)
        _update_button.show()

    def _title(self, *args):
        return "Procrustes View"

    def _update(self):
        self.render_widget.renderer.RemoveAllProps()
        for i in range(self._mol_select.count()):
            if self._mol_select.isSelected(i):
                for actor in self.molecules.specific_by_index(i).get_actors():
                    self.render_widget.renderer.AddActor(actor)
        self.render_widget.Render()

    def _add_model(self, n):
        for actor in self.molecules.specific_by_index(n).get_actors():
            self.render_widget.renderer.AddActor(actor)
        self.render_widget.Render()

    def _remove_model(self, n):
        for actor in self.molecules.specific_by_index(n).get_actors():
            self.render_widget.renderer.RemoveActor(actor)
        self.render_widget.Render()

