#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Plots.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
import vtk
from qt import QSizePolicy, QKeyEvent

class Render_widget(QVTKRenderWindowInteractor):
    """A VTK render widget"""
    def __init__(self, parent_widget, node):
        """Renders model onto a VTK render widget"""
        QVTKRenderWindowInteractor.__init__(self, parent_widget)

        # Size
        self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))

        self.parent = parent_widget
        self.molecules = node.molecule_browser
        self.current_molecule = None
        self.node = node
        self.highlight_index = 1
        self.cycle_atom = None
        self.init_mol_index = 1

        # Render molecule
        self.renderer = vtk.vtkRenderer()
        self.GetRenderWindow().AddRenderer(self.renderer)
        self.display_model()
        self._Iren.SetInteractorStyle(                                                                                                          
                vtk.vtkInteractorStyleTrackballCamera())

    def display_new_model(self):
        self.molecules = self.node.molecule_browser()
        self.redisplay_model()

    def display_model(self, n=None):
        """Replaces the current model in the render widget with the new one"""
        self.renderer.RemoveAllProps()
        if n == None:
            self.current_molecule = self.node.molecule_browser.get_reference_molecule()
        else:
            self.current_molecule = self.molecules.specific_by_index(n)
        actors = self.current_molecule.get_actors()
        for actor in actors:
            self.renderer.AddActor(actor)

    def redisplay_model(self, n=None):
        self.display_model(n)
        self.Render()

    def cycle(self, index=None):
        """Cycles through selected atoms and highlights them"""
        if index:
            self.highlight_index = index

        current_molecule = self.molecules.current()
        
        if self.highlight_index <= len(current_molecule.mapping):
            atomid = current_molecule.mapping[self.highlight_index]
        else:
            self.cycle(1)
            return
        
        if self.cycle_atom:
            # Negate color
            self.cycle_atom.textActor.GetProperty().SetColor(map(lambda a:1-a, self.cycle_atom.textActor.GetProperty().GetColor()))

        self.cycle_atom = current_molecule.atoms[atomid]

        # Negate color of actor
        self.cycle_atom.textActor.GetProperty().SetColor(map(lambda a:1-a, self.cycle_atom.textActor.GetProperty().GetColor()))

        self.highlight_index += 1
        self.Render()

    def mousePressEvent(self, event):
        """[hack] Send a 'P' key event when an atom is selected with a mouse button
        press event"""
        self.keyPressEvent(QKeyEvent(QKeyEvent.KeyRelease,
                                     QKeyEvent.Key_P,
                                     QKeyEvent.Key_P, 0))
        QVTKRenderWindowInteractor.mousePressEvent(self, event)

class HybridRenderWidget(Render_widget):
    """A stateful VTK render widget for displaying models and selecting atoms"""
    def __init__(self, parent_widget, node, is_reference=1):
        """Renders model onto a VTK render widget"""
        Render_widget.__init__(self, parent_widget, node)

    def enter_selection_mode(self):
        # Atom selection
        self.picker = self.GetPicker()
        self.picker.AddObserver("PickEvent", lambda a,b : self.atom_select())

    def display_model(self, n=None):
        """Replaces the current model in the render widget with the new one"""
        self.renderer.RemoveAllProps()
        if n == None:
            self.current_molecule = self.molecules.current()
        else:
            self.current_molecule = self.molecules.specific_by_index(n)
        actors = self.current_molecule.get_actors()
        for actor in actors:
            self.renderer.AddActor(actor)

    def atom_select(self):
        actor = self.picker.GetActor()

        # If the selected actor is an atom actor
        if hasattr(actor, "id"):

            # Find the atom of the selected id
            atomid = actor.id
            selected_atom = None
            for atom in self.current_molecule.atoms:
                if actor.id == atom.actor.id:
                    selected_atom = atom

            camera = self.renderer.GetActiveCamera()
            self.current_molecule.select_atom(atomid, camera)

            # show mapping label
            self.renderer.AddActor(selected_atom.textActor)

    def undo_last_selection(self):
        self.current_molecule.deselect_latest_selected_atom()
        self._update()

    def clear_selection(self):
        self.current_molecule.clear_atom_selection()
        self._update()

    def _update(self):
        self.renderer.RemoveAllProps()
        actors = self.current_molecule.get_actors()
        for actor in actors:
            self.renderer.AddActor(actor)
        self.Render()

class InteractionRenderWidget(HybridRenderWidget):
    """A VTK render widget for displaying models and selecting atoms"""
    def __init__(self, parent_widget, node, _ref_render_widget):
        HybridRenderWidget.__init__(self, parent_widget, node)
        self._ref_render_widget = _ref_render_widget
        self.enter_selection_mode()

    def display_model(self, n=None):
        """Replaces the current model in the render widget with the new one"""
        self.renderer.RemoveAllProps()
        if n == None:
            self.current_molecule = self.node.molecule_browser.get_next_test_molecule()
        else:
            self.current_molecule = self.molecules.specific_by_index(n)
        actors = self.current_molecule.get_actors()
        for actor in actors:
            self.renderer.AddActor(actor)

    def atom_select(self):
        actor = self.picker.GetActor()

        # If the selected actor is an atom actor
        if hasattr(actor, "id"):

            # Find the atom of the selected id
            atomid = actor.id
            selected_atom = None
            for atom in self.current_molecule.atoms:
                if actor.id == atom.actor.id:
                    selected_atom = atom

            self.current_molecule.select_atom(atomid, self.renderer.GetActiveCamera())

            # show mapping label
            self.renderer.AddActor(selected_atom.textActor)

            # When an atom is picked in the test molecule, the next atom is
            # highlighted in the reference molecule.
            #self._ref_render_widget.cycle()

    def display_next_molecule(self):
        actors = self.node.molecule_browser.next().get_actors()
        for actor in actors:
            self.renderer.AddActor(actor)
        
