# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.guinode import GUINode
from Library.Util.Config import Config
from Library.Nodes.Plot.develplotnode import UserPlot
from GUI.Plots.plot2d.imageplot import blow_up_region

from qtui import QWidgetFactory
from qt import QDockWindow, Qt, QObject, SIGNAL, QGrid, QPoint, QWidget
from qt import QDialog, PYSIGNAL, QSizePolicy, QColor, QPen, QMessageBox
from qt import QImage, QBrush, QRect, QToolTip, QSize
from qwt import QwtPlot, QwtPicker, QwtPlotPicker, QwtEventPattern
from qwt import QwtPlotZoomer, QwtDoubleRect, QwtAutoScale, QwtPlotCurve
from qwt import QwtCurve
from sets import Set

import thread

connect = QObject.connect

import os

class DockSettings(QDockWindow):
    """Show individual plot settings for all plot squares in tabs."""
    def __init__(self, parent, node):
        QDockWindow.__init__(self, QDockWindow.InDock, parent)
        self._node = node
        widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                         "userplotdocktabs.ui"),
                                            None, self)
        self._tabs = widget.child("tabs")
        self.setWidget(widget)
        self.setCloseMode(QDockWindow.Always)
        self.setResizeEnabled(True)
        self._pane_to_widget = {}
        self._widget_to_pane = {}
        node.add_active_pane_listener(self._active_tab_changed)
        connect(self._tabs, SIGNAL("currentChanged(QWidget*)"), self._tab_clicked)

    def _tab_clicked(self, widget):
        pane = self._node.get_active_pane()
        if pane not in self._pane_to_widget:
            return
        
        if widget != self._pane_to_widget[pane]:
            self._node.set_active_pane(self._widget_to_pane[widget])

    def _active_tab_changed(self):
        if self._tabs.currentPage() != self._pane_to_widget[self._node.get_active_pane()]:
            self._tabs.showPage(self._pane_to_widget[self._node.get_active_pane()])

    def update_layout(self):
        width, height = self._node.get_layout()
        self._pane_to_widget = {}
        self._widget_to_pane = {}
        self._tabs.hide()
        while self._tabs.currentPage():
            self._tabs.removePage(self._tabs.currentPage())

        index = 0
        for y in range(height):
            for x in range(width):
                settings_widget = SettingsWidget(self._tabs, self._node, x, y)
                widget = settings_widget.get_widget()
                self._tabs.addTab(widget, "%d, %d" % (x+1, y+1))
                self._pane_to_widget[self._node.get_pane_at(x, y)] = widget
                self._widget_to_pane[widget] = self._node.get_pane_at(x, y)
                index += 1

        self._tabs.show()

    def hideEvent(self, e):
        self.emit(PYSIGNAL("hidden()"), ())


class SettingsWidget:
    """Show settings for all plots in a single layout location."""
    def __init__(self, parent, node, x, y):
        self._controllers = []
        self._node = node
        self._pane = self._node.get_pane_at(x, y)
        self._pos = x, y
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          "userplotdocksettings.ui"),
                                             None, parent)
        self._widget.show_addable = lambda : self.show_addable_plots(x, y)
        self._list_widget = self._widget.child("added_plots")
        connect(self._widget.child("add_button"), SIGNAL("clicked()"),
                self._show_addable_plots)
        connect(self._widget.child("remove_button"), SIGNAL("clicked()"),
                self._remove_plot)
        connect(self._list_widget, SIGNAL("highlighted(int)"),
                self._update_settings)
        self._update_plot_list()
        self._parent = parent
        if self._plot_list:
            self._list_widget.setCurrentItem(0)

    def get_widget(self):
        return self._widget

    def _update_plot_list(self):
        self._plot_list = self._pane.get_plots()
        self._list_widget.clear()
        plot_names = [plot.get_type_name() for plot in self._plot_list]
        self._list_widget.insertStrList(plot_names)
        if self._plot_list:
            self._list_widget.setCurrentItem(self._plot_list.index(self._pane.get_active_plot()))
        else:
            box = self._widget.child("box")
            for child in box.children():
                if isinstance(child, QWidget):
                    child.reparent(None, 0, QPoint(0, 0))

    def _update_settings(self, current_plot_index):
        for controller in self._controllers:
            controller.disconnect()

        self._controllers = []
            
        box = self._widget.child("box")
        for child in box.children():
            if isinstance(child, QWidget):
                child.reparent(None, 0, QPoint(0, 0))

        plot = self._plot_list[current_plot_index]
        widget = plot.get_widget(box)
        for choice in plot.get_choices():
            if widget:
                widget.show()
                self._controllers.append(choice.create_controller_widget(box, widget.child(choice.get_id())))
            else:
                self._controllers.append(choice.create_controller_widget(box))
        self._pane.set_active_plot(plot)

    def _show_addable_plots(self):
        list_dialog = QWidgetFactory.create(os.path.join(Config.uidir,
                                                         "userplotaddlist.ui"),
                                            None, self.get_widget())
        plot_list = list_dialog.child("plot_list")
        if not self._node.get_available_plot_types():
            QMessageBox.warning(self._parent, "No plots available", "No available plot types could be found in " + Config.userplotdir + ".")
            return
            
        types, names = zip(*self._node.get_available_plot_types())
        plot_list.insertStrList(list(names))

        if list_dialog.exec_loop() == QDialog.Accepted:
            if plot_list.currentItem() != -1:
                self._node.add_plot_by_type_at(types[plot_list.currentItem()], *self._pos)
                self._update_plot_list()

    def _remove_plot(self):
        if self._list_widget.currentItem() != -1:
            self._node.remove_plot(self._plot_list[self._list_widget.currentItem()])
            self._update_plot_list()


class PointsTipper(QToolTip):
    def __init__(self, widget, plot):
        QToolTip.__init__(self, widget)
        self._plot = plot

    def maybeTip(self, point):
        """point is pixel coordinate in widget-coordinates."""
        xmap = self._plot.canvasMap(QwtPlot.xBottom)
        ymap = self._plot.canvasMap(QwtPlot.yLeft)
        key, dist, xval, yval, index = self._plot.closestCurve(point.x(), point.y())

        if key == 0:
            # no curves found
            return
        
        # we get size of the point, and add abit, to make sure
        # points are easy to hit with the marker, without making
        # them so big they make it hard to hit other points.
        size = self._plot.curve(key).symbol().size() + QSize(2, 2)
        x, y = xmap.transform(xval), ymap.transform(yval)
        point_area = QRect(QPoint(x-size.width()/2, y-size.height()/2), size)

        if point_area.contains(point):
            source = self._plot.get_tooltip_source(key)
            if source:
                text = source.get_tooltip_text(key, index)
                if text:
                    self.tip(QRect(QPoint(x-size.width()/2,
                                          y-size.height()/2), size), text)


class QwtWidget(QwtPlot):
    def __init__(self, parent, grid, node, pane, x, y):
        QwtPlot.__init__(self, parent)
        self._node = node
        self._grid = grid
        self._coord = x, y
        self._painters = []
        self._components = []
        self._images = {}
        self._has_image_plot = False
        self.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored, True)
        self.setMinimumWidth(1)
        self.setMinimumHeight(1)
        self.setCanvasLineWidth(0)
        self._is_dirty = False
        self._gui_thread = thread.get_ident()
        self._tooltips = {}
        self._tipper = PointsTipper(self.canvas(), self)
        pane.add_component_change_listener(self._update_painters)
        pane.add_active_plot_listener(self._update_labels)
        
        picker = QwtPlotPicker(self.canvas())
        picker.setRubberBandPen(QPen(Qt.red))
        picker.setCursorLabelMode(QwtPicker.ActiveOnly)
        picker.setSelectionFlags(QwtPicker.RectSelection | QwtPicker.DragSelection)
        picker.setRubberBand(QwtPicker.RectRubberBand)

        zoomer = QwtPlotZoomer(self.canvas())
        zoomer.setRubberBandPen(QPen(Qt.black))
        zoomer.setCursorLabelMode(QwtPicker.ActiveOnly)
        zoomer.setSelectionFlags(QwtPicker.RectSelection | QwtPicker.DragSelection)
        zoomer.setRubberBand(QwtPicker.RectRubberBand)

        # set patterns for zoom, zoom-base and pop-zoom-stack
        mouse_pattern = [QwtEventPattern.MousePattern(Qt.RightButton, Qt.NoButton),
                         QwtEventPattern.MousePattern(Qt.MidButton, Qt.NoButton),
                         QwtEventPattern.MousePattern(Qt.RightButton, Qt.ShiftButton)]

        zoomer.setMousePattern(mouse_pattern)

        

        self.connect(self, SIGNAL("plotMousePressed(const QMouseEvent &)"),
                     self._activate_pane)

        self._pane = pane
        self._picker = picker
        self._zoomer = zoomer

        connect(picker, SIGNAL("selected(const QwtDoubleRect &)"),
                self._update_selection)

        if self._pane == self._node.get_active_pane():
            self._selected = True

        self._update_painters()
        self._update_labels()

    def insert_tooltip_curve(self, curve, tooltip_source):
        """Add curve with source for curve tooltips.

        If the mousepointer hovers over a curve point,
        tooltip_source.get_tooltip_text(curve_key, index) will be
        called to find the tooltip text to display to the user.
        
        If tooltip_source.get_tooltip_text(...) returns None, no
        tooltip will be displayed for that curve position.
        """
        key = self.insertCurve(curve)
        self._tooltips[key] = tooltip_source
        return key

    def remove_tooltip_curve(self, key):
        del self._tooltips[key]
        self.removeCurve(key)

    def get_tooltip_source(self, curve_key):
        if curve_key in self._tooltips:
            return self._tooltips[curve_key]
        else:
            return None

    def _activate_pane(self, e):
        self._node.set_active_pane(self._pane)

    def _update_labels(self):
        plot = self._pane.get_active_plot()
        if plot:
            self.setAxisTitle(QwtPlot.yLeft, plot.get_y_label() or "")
            self.setAxisTitle(QwtPlot.xBottom, plot.get_x_label() or "")
            self.setTitle(plot.get_title() or "")
        else:
            self.setAxisTitle(QwtPlot.yLeft, "")
            self.setAxisTitle(QwtPlot.xBottom, "")
            self.setTitle("")

    def set_dirty(self):
        """Allow painters to mark this widget as needing repainting."""
        self._is_dirty = True

    def _redraw_plot(self):
        reverse_y = False
        floating_scale = False
        for plot in self._pane.get_plots():
            if plot.get_should_reverse_y():
                reverse_y = True
            if plot.get_use_floating_scale():
                floating_scale = True

        self.changeAxisOptions(QwtPlot.yLeft, QwtAutoScale.Inverted, reverse_y)
        self.changeAxisOptions(QwtPlot.yLeft, QwtAutoScale.Floating, floating_scale)
        self.changeAxisOptions(QwtPlot.xBottom, QwtAutoScale.Floating, floating_scale)
        
        self._update_labels()
        if self._zoomer.zoomRect() == self._zoomer.zoomBase():
            self.setAxisAutoScale(QwtPlot.yLeft)
            self.setAxisAutoScale(QwtPlot.xBottom)
            
        if thread.get_ident() == self._gui_thread:
            self.replot()
        else:
            self.updateAxes()
            self.canvas().invalidateCache()
            self.canvas().update()
        if self._zoomer.zoomRect() == self._zoomer.zoomBase():
            self._zoomer.setZoomBase()
        self._is_dirty = False
        
    def update_if_dirty(self):
        if self._is_dirty:
            self._redraw_plot()

    def disconnect(self):
        """Disconnect listener from PlotPane before destruction."""
        for painter in self._painters:
            painter.disconnect()
        self._pane.disconnect_component_change_listener(self._update_painters)
        self._pane.disconnect_active_plot_listener(self._update_labels)

    def _update_selection(self, rect):
        self.clear()
        self._images = {}
        self._pane.select_area(rect.x1(), rect.x2(), rect.y1(), rect.y2())
        self._grid.update_all_if_dirty()

    def _update_painters(self):
        for painter in self._painters:
            painter.disconnect()
        self._painters = []
        self._components = []
        self.clear()
        self._images = {} 
        for userplot in self._pane.get_plots():
            self._components.extend(userplot.get_cached_components())
        for component in self._components:
            self._painters.append(component.create_painter(self))

        if self._has_image_plot:
            self.enableGridX(False)
            self.enableGridY(False)
        else:
            self.enableGridX(True)
            self.enableGridY(True)
                        
        self._redraw_plot()

    def drawCanvasItems(self, painter, rect, maps, pfilter):
        # we want to draw the image first incase there are curves above it
        for image_item in self._images.values():
            image_item.draw(painter, rect, maps)

        QwtPlot.drawCanvasItems(self, painter, rect, maps, pfilter)

    def add_image(self, image, image_area, selection):
        image_item = QwtImageItem(self, image, image_area, selection)
        self._has_image_plot = True
        self._images[image] = image_item

    def remove_image(self, image):
        if image in self._images.keys():
            del self._images[image]
        if not self._images:
            self._has_image_plot = False

class PlotGrid:
    """Display a grid of all plots."""
    def __init__(self, parent, node):
        self._parent = parent
        self._grid = None
        self._node = node
        self._plots = {}
        self._pickers = []
        self._selectors = {}
        self._full = False
        self._last_full = False
        self._last_layout = None
        self._last_pane = None
        self.update_layout()
        node.add_active_pane_listener(self.update_layout)

    def toggle_fullplot(self, value):
        self._full = value
        self.update_layout()

    def _update_full(self):
        width, height = self._node.get_layout()
        for y in range(height):
            for x in range(width):
                if self._node.get_pane_at(x, y) == self._node.get_active_pane():
                    pane = self._node.get_pane_at(x, y)
                    if self._last_full and self._last_pane == pane:
                        return
                    
        self._last_pane = pane
        
        self._pickers = []
        if self._grid:
            self._grid.reparent(None, 0, QPoint(0, 0))

        # disconnect old plots
        for pos, plot in self._plots.items():
            plot.disconnect()
        self._plots = {}

        self._grid = QGrid(1, self._parent)

        plot = QwtWidget(self._grid, self, self._node, pane, x, y)
        self._plots[x, y] = plot

        self._grid.show()        

    def update_layout(self):
        if self._full:
            self._update_full()
            self._last_full = True
        else:
            self._update_multiple()
            self._last_full = False
        
    def _update_multiple(self):
        width, height = self._node.get_layout()
        if self._last_layout == (width, height) and not self._last_full:
            # we won't update if nothing changed
            return
        
        self._last_layout = width, height

        self._pickers = []
        if self._grid:
            self._grid.reparent(None, 0, QPoint(0, 0))

        # disconnect old plots
        for pos, plot in self._plots.items():
            plot.disconnect()
        self._plots = {}

        self._grid = QGrid(width, self._parent)

        for y in range(height):
            for x in range(width):
                plot = QwtWidget(self._grid, self, self._node, self._node.get_pane_at(x, y), x, y)
                self._plots[x, y] = plot
                
        self._grid.show()

    def update_all_if_dirty(self):
        for plot in self._plots.values():
            plot.update_if_dirty()

class GUIDevelPlotNode(GUINode):
    def __init__(self, parent, node):
        GUINode.__init__(self, parent, node)
        self._parent = parent

        self._window = QWidgetFactory.create(os.path.join(Config.uidir,
                                                    "userplotwindow.ui"),
                                       None, self._parent)
        self._dock = DockSettings(self._window, node)
        self._window.moveDockWindow(self._dock, Qt.DockBottom)
        self._grid = PlotGrid(self._window.child("box"), self.getDiagramNode())

        self._configuration_action = self._window.child("toggleConfigurationAction")
        self._full_plot_action = self._window.child("fullPlotAction")
        self._pane_combo = self._window.child("pane_combo")
        self._panes = []

        connect(self._configuration_action, SIGNAL("toggled(bool)"), self._dock.setShown)
        connect(self._full_plot_action, SIGNAL("toggled(bool)"), self._grid.toggle_fullplot)
        connect(self._window.child("showSettingsAction"), SIGNAL("activated()"),
                self.show_settings)
        connect(self._dock, PYSIGNAL("hidden()"), self._unset_dock_action)
        connect(self._pane_combo, SIGNAL("activated(int)"), self._pane_selected)
        node.add_active_pane_listener(self._pane_changed)

        # we initially hide settings dock
        self._configuration_action.setOn(False)
        self._shown_missing_warning = False

    def _pane_selected(self, index):
        """Set active pane when user selects in pane combo."""
        self.node.set_active_pane(self._panes[index])

    def _pane_changed(self):
        """When active pane changes, update plot list and pane combo."""
        pane = self.node.get_active_pane()
        self._pane_combo.setCurrentItem(self._panes.index(pane))

        combo = self._window.child("plot_combo")
        combo.clear()
        plots = pane.get_plots()
        plot_names = [plot.get_type_name() for plot in plots]
        combo.insertStrList(plot_names)

    def _unset_dock_action(self):
        self._configuration_action.setOn(False)

    def nodeDialog(self):
        self._window.show()
        self._window.raiseW()
        self.update_layout()
        missing_plots = list(Set(self.node.get_missing_plot_types()))
        missing_plots.sort()

        if not self._shown_missing_warning:
            for filename, message in self.node.get_failed_plot_imports().items():
                QMessageBox.warning(self._window, "Could not load " + filename,
                                    "Could not import plots from file: %s\n\n%s" % (filename, message))
                
            if missing_plots:
                QMessageBox.warning(self._window, "Missing plots", "The following plot types are used in this plot node, but could not be found:\n\n" + "\n".join(missing_plots) + "\n\nThese plots should be located in " + Config.userplotdir + ".")
            self._shown_missing_warning = True

    def show_settings(self):
        settings_window = QWidgetFactory.create(os.path.join(Config.uidir,
                                                             "userplotsettings.ui"),
                                                None, self._window)

        width, height = self.getDiagramNode().get_layout()
        width_box, height_box = map(settings_window.child, ["plot_width",
                                                            "plot_height"])
        width_box.setValue(width)
        height_box.setValue(height)
        
        if settings_window.exec_loop() == QDialog.Accepted:
            self.getDiagramNode().set_layout(width_box.value(), height_box.value())
            self.update_layout()

    def update_layout(self):
        self._dock.update_layout()
        self._grid.update_layout()
        self._pane_combo.clear()
        width, height = self.node.get_layout()
        self._panes = []
        for y in range(height):
            for x in range(width):
                self._pane_combo.insertItem("%d, %d" % (x+1, y+1))
                self._panes.append(self.node.get_pane_at(x, y))


def get_pixel_area(plot_area, (image_width, image_height), image_area):
    """Convert area from plot coordinates to pixel coordinates.

    Pixel coordinates are from (0, 0) to (width, height) for the
    same area covered by image area in plot coordinates.
    """
    x1 = (plot_area.x1() - image_area.x1()) * image_width / image_area.width()
    x2 = (plot_area.x2() - image_area.x1()) * image_width / image_area.width()
    y1 = (plot_area.y1() - image_area.y1()) * image_height / image_area.height()
    y2 = (plot_area.y2() - image_area.y1()) * image_height / image_area.height()
    return QwtDoubleRect(x1, x2, y1, y2)

def get_plot_area(pixel_area, (image_width, image_height), image_area):
    """Convert area from pixel coordinates to plot coordinates.
                
    Pixel coordinates are from (0, 0) to (width, height) for the
    same area covered by image area in plot coordinates.
    """
    x1 = image_area.x1() + pixel_area.x1() * image_area.width() / image_width
    x2 = image_area.x1() + pixel_area.x2() * image_area.width() / image_width
    y1 = image_area.y1() + pixel_area.y1() * image_area.height() / image_height
    y2 = image_area.y1() + pixel_area.y2() * image_area.height() / image_height
    return QwtDoubleRect(x1, x2, y1, y2)


class QwtImageItem:

    def __init__(self, qwt_plot, image, image_area, selection):
        self._image = image
        self._selection = selection
        self._image_area = image_area
        self._qwt_plot  = qwt_plot
        self._outline_curve = None
        self._draw_helper_curve()

    def _draw_helper_curve(self):
        curve = QwtPlotCurve(self._qwt_plot)
        
        x_points = self._image_area[0], self._image_area[1]
        y_points = self._image_area[2], self._image_area[3]

        curve.setData(x_points, y_points)
        curve.setStyle(QwtCurve.NoCurve)

        self._outline_curve = self._qwt_plot.insertCurve(curve)

    def draw(self, painter, rect, maps):
        x_map = maps[QwtPlot.xBottom]
        y_map = maps[QwtPlot.yLeft]

        # first we want to know how to handle our y-axis
        
        if y_map.invTransform(rect.top()) < y_map.invTransform(rect.bottom()):
            origin_topleft = True
        else:
            origin_topleft = False

        # we make all our relevant areas
        if origin_topleft:
            plot_area = QwtDoubleRect(x_map.invTransform(rect.left()),
                                      x_map.invTransform(rect.right()),
                                      y_map.invTransform(rect.top()),
                                      y_map.invTransform(rect.bottom()))
            image_area = QwtDoubleRect(*self._image_area)
        else:
            plot_area = QwtDoubleRect(x_map.invTransform(rect.left()),
                                      x_map.invTransform(rect.right()),
                                      y_map.invTransform(rect.bottom()),
                                      y_map.invTransform(rect.top()))
            x1, x2, y1, y2 = self._image_area
            image_area = QwtDoubleRect(x1, x2, y1, y2)

        # we find which part of the image to get in plot coordinates
        plot_intersection = plot_area.intersect(image_area)

        # we find how big it is in pixel coordinates

        pixel_area = get_pixel_area(plot_intersection, (self._image.width(),
                                                        self._image.height()),
                                    image_area)

        # we then find the resulting pixel area to use in the plot
        # (the width and height in pixels we should scale the image to
        new_size = (x_map.transform(plot_intersection.x2()) - x_map.transform(plot_intersection.x1()),
                    y_map.transform(plot_intersection.y2()) - y_map.transform(plot_intersection.y1()))

        for dim in new_size:
            if dim <= 0:
                # no use in drawing image or selection if image is too small
                return

        y1 = min(pixel_area.y1(), pixel_area.y2())
        y2 = max(pixel_area.y1(), pixel_area.y2())

        crop_rect = [val + 0.5 for val in (pixel_area.x1(), y1,
                                           pixel_area.x2(), y2)]

        sub_image = blow_up_region(self._image, crop_rect, new_size)

        if origin_topleft:
            painter.drawImage(x_map.transform(plot_intersection.x1()),
                              y_map.transform(plot_intersection.y1()),
                              sub_image)
        else:
            if y1 > y2:
                image = sub_image.mirror()
            else:
                image = sub_image
            
            painter.drawImage(x_map.transform(plot_intersection.x1()),
                              y_map.transform(plot_intersection.y1()),
                              image)

        # paint selection
        if not self._selection:
            return       

        pixels_by_row, pixels_by_column, rows, columns = self._selection

        boxel_areas = []
        
        for row in rows:
            boxel_areas.append(QwtDoubleRect(0, self._image.width(), row, row+1))

        for column in columns:
            boxel_areas.append(QwtDoubleRect(column, column+1, 0,
                                             self._image.height()))

        for pixel in pixels_by_row:
            row = pixel / self._image.width()
            column = pixel % self._image.width()
            boxel_areas.append(QwtDoubleRect(column, column+1, row, row+1))

        for pixel in pixels_by_column:
            row = pixel % self._image.height()
            column = pixel / self._image.height()
            boxel_areas.append(QwtDoubleRect(column, column+1, row, row+1))


        painter.setRasterOp(Qt.XorROP)
        brush = QBrush(QColor(255, 255, 255))

        for area in boxel_areas:
            covered_area = get_plot_area(area, (self._image.width(),
                                                self._image.height()),
                                         image_area)
            # no use in painting outside
            covered_area = covered_area.intersect(plot_area)

            if origin_topleft:
                x1 = x_map.transform(covered_area.x1())
                y1 = y_map.transform(covered_area.y1())
                x2 = x_map.transform(covered_area.x2()) - x_map.transform(covered_area.x1())
                y2 = y_map.transform(covered_area.y2()) - y_map.transform(covered_area.y1())
            else:
                x1 = x_map.transform(covered_area.x1())
                y1 = y_map.transform(abs(covered_area.y1()-self._image.height()))
                x2 = x_map.transform(covered_area.x2()) - x_map.transform(covered_area.x1())
                y2 = y_map.transform(abs(covered_area.y2() - self._image.height())) - \
                     y_map.transform(abs(covered_area.y1()-self._image.height()))
            
            painter.fillRect(x1, y1, x2, y2, brush)
            
        painter.setRasterOp(Qt.CopyROP)

    def __del__(self):
        if self._outline_curve: 
            self._qwt_plot.removeCurve(self._outline_curve)
