#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Nodes.guiplot2dnodesettings import WidgetSettingsDialog
from GUI.Pixmaps.Icons import Icons
from GUI.Plots.plot2d.plot2dwidget import Plot2DWidget
from GUI.viewers import HelpViewer
from GUI.Plots.plot2d.scatterplot import ScatterPlot
from GUI.Plots.plot2d.lineplot import LinePlot
from GUI.Plots.plot2d.histogramplot import HistogramPlot
from GUI.Plots.plot2d.dendrogram import DendrogramPlot
from GUI.Plots.plot2d.imageplot import ImagePlot
from GUI.Plots.plot2d.normalqqplot import NormalQqPlot
from GUI.Plots.plot2d.multilineplot import MultiLinePlot
from GUI.Plots.plot2d.densityplot import DensityPlot
from GUI.Plots.plot2d.gobrowser import GoBrowser, GeneDb
from GUI.Plots.plot2d.ontologygraph import GeneOntology
from GUI.spreadsheet import DataComposer
from Library.Util.Config import Config
from Library.Util.gui import EventReceiver
from Library.Util.lists import flatten
from qtui import QWidgetFactory
from qt import QMainWindow, QWidget, QDockWindow, QListViewItem, QObject
from qt import QPopupMenu, QAction, SIGNAL, QIconSet, QKeySequence, QToolBar, Qt
from qt import QGridLayout, QRect, QPixmap, QMessageBox, QPrinter, QFileDialog
from qt import QPainter, QPaintDeviceMetrics, QTabWidget, QVBoxLayout, QGroupBox
from qt import QListBox, QPushButton, QListViewItemIterator, QLineEdit
from qt import QInputDialog
from sets import Set
import os

connect = QObject.connect

# This package is the part of guiplot2dnode that depends on qwt.
# It has been split into its own module to enable qwt to be optional.
class EPSExportError(Exception):
    pass

# domain exceptions

class NoSuchPlotError(Exception):
    """Tried to create a plot of a non-existing type."""


class GUIPlot2DWindow(QMainWindow, EventReceiver):
    """This window will contain 2d plots."""
    _name = "GUIPlotWindow"
    _window_title = "Plot View"

    def __init__(self, parent, node, event_server=None, initial_pos=None):
        QMainWindow.__init__(self, parent, self._name,
                                Qt.WType_TopLevel | Qt.WStyle_ContextHelp)
        EventReceiver.__init__(self)

        self._parent = parent
        self._node = node
        self._eps_filename = None
        
        ontology_file = os.path.join(Config.datadir, "gene_ontology.obo")
        self._gobrowser = GoBrowser(self, GeneOntology(ontology_file),
                                    GeneDb().lookup)

        self.setCaption(self.tr(self._window_title))

        vert_widgets = self._node.get_vertical_widgets()
        hor_widgets  = self._node.get_horisontal_widgets()

        central_widget = QWidget(self, "main plot window")
        central_widget.setPaletteBackgroundColor(Qt.white)
        self.setCentralWidget(central_widget)
        self._grid_layout = QGridLayout(self.centralWidget(), 1, 1, 5, 5)
        self._init_gui()

        #readying menus
        self._init_bottom_dock_window()
        self._settings_dialog = SettingsDialog(self, self._node, self)
        self._settings_dialog.set_spinners_values(hor_widgets, vert_widgets)

        self._widgets = {}
        self._datacomposers = []
        self._current_x_max = 0
        self._current_y_max = 0
        self.update_plot_widgets(hor_widgets, vert_widgets)
        self._widgets[(0, 0)].set_selected(True)
        self._selected_widget = self._widgets[(0, 0)]

        # Plot map settings must be loaded after all plots and
        # plot_widgets are added to this GUI
        self._node.get_plot_map_handler().load_settings(self)

        self.setGeometry(QRect(self.x(), self.y(), 555, 500))
        if initial_pos:
            self.move(initial_pos)

        if event_server:
            event_server.addEventHandler(self.post_event)

        self.hidden = True

    def show(self):
        #try:

        #except:
            # Catch errors or else we won't be able to open the plot node. 
        QMainWindow.show(self)
        self.hidden = False
        self._bottom_dock_window.show()
        self._bottom_dock_window.update_data()

    def get_go_browser(self):
        """Get the single instance of the gene ontology browser."""
        return self._gobrowser

    def _init_bottom_dock_window(self):
        """Initiates the WindowSettingsDockWindow."""
        self._bottom_dock_window = WindowSettingsDockWindow(self, self._node)
        #shows the dockwindow in the correct place..
        self.moveDockWindow(self._bottom_dock_window, Qt.DockBottom)

    def _init_gui(self):
        # file menu actions
        save_action = QAction(QIconSet(QPixmap(Icons.savePic)),
                              "&Save Custom Plot As...", QKeySequence(), self)
        connect(save_action, SIGNAL('activated()'), self._slot_save_custom_plot)

        print_action = QAction(QIconSet(QPixmap(Icons.printPic)),
                               "&Print Selected...", QKeySequence("CTRL+P"),
                               self)
        connect(print_action, SIGNAL('activated()'), self._slot_print_plot)

        print_all_action = QAction(QIconSet(QPixmap(Icons.printPic)),
                                   "Print &All...", QKeySequence(), self)
        connect(print_all_action, SIGNAL('activated()'),
                self._slot_print_all_plots)

        export_eps_action = QAction('&Selected Plot to EPS...', QKeySequence(),
                                    self)
        connect(export_eps_action, SIGNAL('activated()'),
                self._slot_export_selected_eps)


        export_all_eps_action = QAction('&All Plots to EPS...', QKeySequence(),
                                        self)
        connect(export_all_eps_action, SIGNAL('activated()'),
                self._slot_export_all_eps)

        close_action = QAction('Close &Window', QKeySequence('CTRL+W'), self)
        connect(close_action, SIGNAL('activated()'), self.close)

        # edit menu actions
        settings_action = QAction("&Settings... ", QKeySequence("CTRL+S"),
                                  self)  
        connect(settings_action, SIGNAL("activated()"), self._slot_edit_settings)

        # view menu actions
        toolbar_action = QAction('Toggle &Administrative Toolbar',
                                 QKeySequence(), self)
        toolbar_action.setToggleAction(True)
        toolbar_action.setOn(True)
        connect(toolbar_action, SIGNAL('toggled(bool)'),
                self._slot_toggle_admin_toolbar)

        self._config_action = QAction(QIconSet(QPixmap(Icons.plotControlPic)),
                                      '&Toggle Configuration Window',
                                      QKeySequence('CTRL+T'), self)
        self._config_action.setToggleAction(True)
        self._config_action.setOn(True)
        connect(self._config_action, SIGNAL('toggled(bool)'),
                self.slot_view_plotconfig_toggled)

        refresh_action = QAction('&Refresh', QKeySequence('F5'), self)
        connect(refresh_action, SIGNAL('activated()'), self._slot_refresh)

        # tool menu actions
        gobrowser_action = QAction('&Gene Ontology Browser', QKeySequence(),
                                   self)
        datacomposer_action = QAction('&Data Composer', QKeySequence(),
                                      self)
        connect(gobrowser_action, SIGNAL('activated()'),
                self._slot_launch_gobrowser)
        connect(datacomposer_action, SIGNAL('activated()'),
                self._slot_launch_datacomposer)

        # build file menu structure
        self._file_menu = QPopupMenu()
        self._file_export_menu = QPopupMenu()

        save_action.addTo(self._file_menu)
        self._file_menu.insertSeparator()
        print_action.addTo(self._file_menu)
        print_all_action.addTo(self._file_menu)
        export_eps_action.addTo(self._file_export_menu)

        export_all_eps_action.addTo(self._file_export_menu)
        self._file_menu.insertSeparator()
        self._file_menu.insertItem('&Export', self._file_export_menu)
        self._file_menu.insertSeparator()
        close_action.addTo(self._file_menu)

        # keeps references to menus to keep from getting GC'd
        # build edit menu
        self._edit_menu = QPopupMenu()
        settings_action.addTo(self._edit_menu)

        # build view menu
        self._view_menu = QPopupMenu()
        toolbar_action.addTo(self._view_menu)
        self._config_action.addTo(self._view_menu)
        refresh_action.addTo(self._view_menu)

        # build tools menu
        self._tool_menu = QPopupMenu()
        gobrowser_action.addTo(self._tool_menu)
        datacomposer_action.addTo(self._tool_menu)

        #initializing menubar
        self.menuBar().insertItem('&File', self._file_menu)
        self.menuBar().insertItem("&Edit", self._edit_menu)
        self.menuBar().insertItem('&View', self._view_menu)
        self.menuBar().insertItem('&Tools', self._tool_menu)

        # initialize toolbar
        self.admin_toolbar = QToolBar('Administrative tools', self)
        save_action.addTo(self.admin_toolbar)
        self.admin_toolbar.addSeparator()
        print_all_action.addTo(self.admin_toolbar)
        self.admin_toolbar.addSeparator()
        self._config_action.addTo(self.admin_toolbar)

    def _slot_launch_gobrowser(self):
        self._update_go_browser()
        self._gobrowser.show()

    def _slot_launch_datacomposer(self):
        indices = Set()
        for plot in self._selected_widget.get_plots():
            for aspect in plot.get_mappable_aspects():
                indices.update(plot.get_highlighted_indices_by_aspect(aspect))
        indices = list(indices)
        indices.sort()
        composer = DataComposer(indices)
        composer.show()
        composer.update_ports(self._node.get_all_input_ports())
        self._datacomposers.append(composer)

    def _slot_toggle_admin_toolbar(self, bool_show):
        if bool_show:
            self.admin_toolbar.show()
        else:
            self.admin_toolbar.hide()

    def _slot_save_custom_plot(self):
        q_text, accepted = QInputDialog.getText("Save Custom Plot As...",
                                                "Custom Plot Name:",
                                                QLineEdit.Normal,
                                                "My Custom Plot",
                                                self)
        filename = str(q_text)

        if accepted and filename:
            self._node.nodeRepName = 'Plot.Saved plots:' + filename
            self._node.setName(filename)

            customplotdir = Config.customplot2ddir
            if not os.path.exists(customplotdir):
                os.mkdir(customplotdir)
            if not filename.endswith(".zpl"):
                filename = filename + ".zpl"

            filename_path = os.path.join(customplotdir, filename)

            if not os.path.exists(filename_path):
                self._write_custom_plot_to_file(filename_path)
                self._update_node_tree()
            else:
                ans = QMessageBox.question(self, 'Overwrite Custom Plot?',
                                           'A custom plot with this name already exists.\nWould you like to overwrite it?',
                                           QMessageBox.Yes, QMessageBox.No,
                                           QMessageBox.NoButton)
                if ans == QMessageBox.Yes:
                    self._write_custom_plot_to_file(filename_path)

    def _write_custom_plot_to_file(self, filename_path):
        fd = open(filename_path, 'w')
        taglist = self._node.custom_plot_to_xml()
        for element in taglist:
            fd.write(element + '\n')
        fd.close()

    def _update_node_tree(self):
        node_tree = self._parent._application.get_node_tree()
        node_list = dict(self._parent._application.nodeRepository.getNodeList())
        node_tree.update_node_list(node_list)

    def _get_eps_qprinter(self):
        """Returns a QPrinter object that can be used for printing EPS to.

        The method asks the user for a filename in which to save the
        EPS in, by showing a QFileDialog. None is returned if the user
        canceled the action.
        """
        qprinter = QPrinter()
        filename = str(QFileDialog.getSaveFileName(None, 'Encapsulated PostScript (*.eps)',
                                                      self, None,
                                                      'Export to EPS'))
        if not filename:
            # User pressed cancel 
            return None
        
        if not filename[-4:] == '.eps':
            filename += '.eps'

        qprinter.setOutputFileName(filename)
        qprinter.setColorMode(qprinter.Color)
        return qprinter

    def _get_qpainter(self, qprinter, hor_widgets, vert_widgets):
        """Returns a qpainter using the given qprinter and geometry."""
        qpainter = QPainter(qprinter)
        metrics = QPaintDeviceMetrics(qpainter.device())
        width = metrics.width()
        height = (width / hor_widgets) * vert_widgets
        qpainter.setClipRect(0, 0, width, height, qpainter.CoordPainter)
        return qpainter

    def _print_plots(self, qprinter, hor_widgets, vert_widgets):
        """Prints all plots with the given qprinter.

        If hor_widgets and vert_widgets are 1, the selected plot is
        printed. Else we loop in x-direction and y-direction where
        hor_widgets and vert_widgets are the ends of the loops in each
        direction, and prints the plots.

        Plots are printed with a quadratic size.
        """
        qpainter = self._get_qpainter(qprinter, hor_widgets, vert_widgets)
        # get width and height for each plot 
        metrics = QPaintDeviceMetrics(qpainter.device())
        if hor_widgets >= vert_widgets and metrics.width() < metrics.height():
            # width of plots in x-direction is the largest (wrt. paintdevice)
            width = metrics.width() / hor_widgets
            height = width # quadratically sized plots
        else:
            # height of plots in x-direction is the largest (wrt. paintdevice)
            height = metrics.height() / vert_widgets
            width = height # quadratically sized plots

        if hor_widgets == 1 and vert_widgets == 1:
            # only one plot, selected plot is printed
            self._selected_widget.printPlot(qpainter, QRect(0, 0, width,
                                                            height))
        else:
            # print the plots to their designated slots in the qpainter
            for x in range(hor_widgets):
                for y in range(vert_widgets):
                    self._widgets[(x, y)].printPlot(qpainter,
                                                    QRect(x * width, y * height,
                                                          width, height))
        qpainter.end()

    def _export_to_eps(self, selected_plot=False):
        """Exports plots in this window to EPS.

        If selected_plot is True, only the selected plot is
        exported. Else all plots are exported.
        """
        qprinter = self._get_eps_qprinter()
        if not qprinter:
            # user canceled
            return
        if selected_plot:
            self._print_plots(qprinter, 1, 1)
        else:
            self._print_plots(qprinter, self._current_x_max,
                              self._current_y_max)
        
        # check if successful export
        self._check_eps_file_created(str(qprinter.outputFileName()))

    def _send_to_printer(self, selected_plot=False):
        """Sends plots in this window to the printer.

        If selected_plot is True, only the selected plot is sent to
        the printer. Else all plots are printed.
        """
        qprinter = QPrinter()
        if qprinter.setup(self): # shows a printer setup dialog
            if selected_plot:
                self._print_plots(qprinter, 1, 1)
            else:
                self._print_plots(qprinter, self._current_x_max,
                                  self._current_y_max)

    def _slot_export_selected_eps(self):
        self._export_to_eps(selected_plot=True)

    def _slot_export_all_eps(self):
        self._export_to_eps()

    def _slot_print_plot(self):
        self._send_to_printer(selected_plot=True)
        
    def _slot_print_all_plots(self):
        self._send_to_printer()

    def _slot_edit_settings(self):
        self._settings_dialog.show()

    def slot_view_plotconfig_toggled(self, bool_open):
        if bool_open:
            self._bottom_dock_window.reopen()
        else:
            temp = self.hidden
            self.hidden = False
            self._bottom_dock_window.hide()
            self.hidden = temp

    def _check_eps_file_created(self, filename):
        if not os.path.exists(filename):
            QMessageBox.information(self, 'Error exporting image',
                                       '%s was not created. Do you have a printer installed' % filename)

    def set_plotconfig_toggle_action(self, on):
        self._config_action.setOn(on)

    def replot_all_widgets(self):
        for widget in self._widgets.values():
            widget.replot()

    def update_plot_widgets(self, new_x_max, new_y_max):
        """Readies a plot which has x_max * y_max number of plot widgets.
        x_max is the number of horisontal plot widgets.
        y_max is the number of vertical plot widgets.
        """

        if not new_x_max == self._current_x_max or not new_y_max == self._current_y_max:
            self.centralWidget().hide()

            self._node.update_widgets(new_x_max, new_y_max)

            if self._widgets:
                for x in range(self._current_x_max):
                    for y in range(self._current_y_max):
                        if x >= new_x_max or y >= new_y_max:
                            curr_widget = self._widgets.pop((x, y))
                            self._grid_layout.remove(curr_widget)
                            curr_widget.remove()
                            curr_widget.deleteLater()
                            self._bottom_dock_window.remove_window((x, y))

            self._current_x_max = new_x_max
            self._current_y_max = new_y_max

            for y in range(new_y_max):
                for x in range(new_x_max):
                    if not self._widgets.has_key((x, y)):
                        config = self._node.get_widget_config(x, y)
                        self._widgets[(x, y)] = Plot2DWidget(self.centralWidget(),
                                                             self, config, (x, y))
                        self._grid_layout.addWidget(self._widgets[(x, y)], y, x)
                        self._bottom_dock_window.add_window((x, y), self._widgets[(x, y)])
                    else:
                        self._bottom_dock_window.set_tab_label_from_id((x, y))

            self.update()
            self.centralWidget().updateGeometry()
            self.centralWidget().show()

    def set_selected(self, qwt_plot_widget):
        """Select tab widget belonging to plot widget.

        Called from Plot2DWidget class when a plot window is selected
        (mouse pressed).  This function selects the corresponding tab
        widget for the plot, and a slot is then called.
        """
        for key in self._widgets.keys():
            if self._widgets[key] == qwt_plot_widget:
                return self._bottom_dock_window.select_window(key)

    def update_dock_window(self):
        """Updates the contents of the bottom dock window."""
        self._bottom_dock_window.update_data()

    def update_title_for_plotwidget(self, id_tuple):
        self._bottom_dock_window.set_tab_label_from_id(id_tuple)
    
    def select_plot_widget(self, plot_widget):
        """Called when tab widget changes.
        """
        if self._selected_widget != plot_widget:
            self._selected_widget.set_selected(False)
            plot_widget.set_selected(True)
            self._selected_widget = plot_widget

    def get_dock_window(self):
        return self._bottom_dock_window

    def get_plot_widget(self, plot_id):
        return self._widgets[plot_id]

    def get_plot_widgets(self):
        return self._widgets

    def get_plot_widget_id(self, widget_object):
        for id_ in self._widgets:
            if self._widgets[id_] == widget_object:
                return id_

    def auto_plot_map(self, plot):
        """Automatic map plot to all other plots."""
        handler = self._node.get_plot_map_handler()
        if handler.get_enabled() and handler.get_auto_map_enabled():
            handler.auto_map_plot(self.get_plots(), plot)

    def remove_plot_map(self, plot):
        self._node.get_plot_map_handler().remove_plot_map(plot)

    def get_plots(self):
        plot_list = []
        for widget in self._widgets.values():
            for plot in widget.get_plots(): 
                plot_list.append(plot)
        return plot_list

    def node(self):
        return self._node

    def get_current_x_max(self):
        """Returns the number of widgets in x direction"""
        return self._current_x_max

    def get_current_y_max(self):
        """Returns the number of widgets in y direction"""
        return self._current_y_max

    def hide(self):
        """Overridden from super class."""
        self.hidden = True
        QMainWindow.hide(self)

    def _slot_refresh(self):
        self._bottom_dock_window.update_data()

    def handle_event(self, event):
        """Update diagram when data is updated."""
        if self._node.get_module_diagram().getId() == event.get_module_diagram_id() and event.get_event_class() == "DiagramEvent" and event.getType() == "FINISH":
            # we want to update the data only if the window is open
            if self.isShown():
                self._bottom_dock_window.update_data()
            self._update_go_browser()

    def _update_go_browser(self):
        self._gobrowser.update_ports(self._node.get_all_input_ports())

    def get_mappable_plots(self):
        """Return a set of window-names and the plots in them.

        The returned dictionary has a list of plots for each window
        title key.
        """
        windows = []

        # get all widget configs
        widget_configs = []
        for y in range(self._node.get_vertical_widgets()):
            for x in range(self._node.get_horisontal_widgets()):
                widget_configs.append(self._node.get_widget_config(x, y))

        # get plot2dwidgets for widget configs

        for plot_widget in self._widgets.values():
            x, y = plot_widget.id()
            window_index = y * self._node.get_horisontal_widgets() + x + 1
            
            window_title = None
            if plot_widget.get_settings().has_key("window_title"):
                window_title = plot_widget.get_settings()['window_title']
            if not window_title:
                window_title = "Window %d"%window_index
                
            windows.append((window_title, plot_widget.get_plots()))

        windows.append(("Tools", [self._gobrowser]))

        windows.sort()
        return windows


class WindowSettingsDockWindow(QDockWindow):
    """The DockWindow which contains the window settings for all the different
    qwt-windows in the main window (one tabwidget for each qwt-widget).
    """
    def __init__(self, main_window, node):
        """Constructor.

        Parameters:
        main_window -- the GUIPlot2DWindow
        node        -- the node we retrieve data from
        """
        QDockWindow.__init__(self, QDockWindow.InDock,
                             main_window, 'bottom_dock_window')
        self._node = node
        self._main_window = main_window

        self.setResizeEnabled(True)
        self.setMovingEnabled(True)
        self._bottom_dock_container = QTabWidget(self,
                                                    'bottom_dock_container')
        connect(self._bottom_dock_container,
                     SIGNAL("currentChanged(QWidget *)"),
                     self._slot_widget_changed)
        self.setCloseMode(self.Always)
        self.hidden = True
        self._closed = False

        # the dictionary that contains the TabWidgets. Keys are id's from
        # GUI2DPlotWindow -- (x, y)-tuple.
        self._container_widgets = {}

        self.setWidget(self._bottom_dock_container)

    def show(self):
        """Overridden from super class."""
        if not self._closed:
            QDockWindow.show(self)
            self.hidden = False

    def update_geometry(self):
        m_rect = self._main_window.geometry()
        s_rect = self.geometry()
        if s_rect.width() > m_rect.width():
            m_rect.setWidth(s_rect.width())
        elif m_rect.width() < 555: # this is actually HACKish but it works
            m_rect.setWidth(555)   # since they don't need more than 555 pixels
        else:
            return
        self._main_window.setGeometry(m_rect)

    def update_data(self):
        """Called when input ports have been renamed/added/deleted"""
        # Available input
        for key in self._container_widgets.keys():
            self._container_widgets[key].update_combo_boxes_data()
            self._container_widgets[key].update_data()

    def add_window(self, id, qwt_plot_widget):
        """When a new qwt_plot_widget is added to the main window, this method
        is called to add a TabWidget and give it its reference to the
        qwt_plot_widget.

        Parameters:
        id              -- the id of the plot2dwidget [(x, y)-tuple]
        qwt_plot_widget -- the plot2dwidget the new tab contains settings for
        """
        tab_widget = TabWidget(self, qwt_plot_widget, self._node, id)
        self._container_widgets[id] = tab_widget
        self._bottom_dock_container.insertTab(tab_widget, '',
                                              (id[1] * self._main_window.get_current_x_max()) + id[0])
        self.set_tab_label(tab_widget, tab_widget.get_title())

    def remove_window(self, id):
        """Removes the TabWidget with given id.

        Parameters:
        id -- the id of the plot2dwidget to be removed [(x, y)-tuple]
        """
        self._bottom_dock_container.removePage(self._container_widgets.pop(id))
        for index in range(self._bottom_dock_container.count()):
            tab_widget = self._bottom_dock_container.page(index)
            self.set_tab_label(tab_widget, tab_widget.get_title())

    def set_tab_label_from_id(self, id_tuple):
        tab_widget = self._container_widgets[id_tuple]
        self.set_tab_label(tab_widget, tab_widget.get_title())

    def set_tab_label(self, tab_widget, label=''):
        tid = tab_widget.get_widget_id()
        if not label:
            label = 'Window %d' % (tid[1]*self._main_window.get_current_x_max() + tid[0] + 1)
        self._bottom_dock_container.setTabLabel(tab_widget, label)
        self._bottom_dock_container.update()

    def get_tab_label(self, tab_id):
        try:
            return self._bottom_dock_container.tabLabel(self._container_widgets[tab_id])
        except KeyError:
            return ""

    def select_window(self, id):
        """Selects the TabWidget with given id.

        Parameters:
        id -- the id of the plot2dwidget to be selected [(x, y)-tuple]
        """
        self._bottom_dock_container.setCurrentPage(
            self._bottom_dock_container.indexOf(self._container_widgets[id]))

    def hide(self):
        """Overridden from superclass"""
        QDockWindow.hide(self)
        self.hidden = True
        if self.hidden and not self._main_window.hidden:
            self._main_window.set_plotconfig_toggle_action(False)
            self._closed = True

    def reopen(self):
        """reopens a closed window (x)"""
        self._main_window.set_plotconfig_toggle_action(True)
        self._closed = False
        self.show()

    def is_closed(self):
        return self._closed

    def _slot_widget_changed(self, widget):
        plot_widget_id = widget.get_widget_id()
        plot_widget = self._main_window.get_plot_widget(plot_widget_id)
        self._main_window.select_plot_widget(plot_widget)

class TabWidget(QWidget):
    def __init__(self, parent, qwt_plot_widget, node, widget_id,
                 name='tab_widget'):
        """Constructor

        Parameters:
        parent            -- the qt parent of this widget
        qwt_plot_widget   -- the widget this tab_widget controls
        node              -- the node this tab_widget retrieves data from
        widget_id         -- the (x, y)-tuple is the id of the widget
        name='tab_widget' -- the qt name (default: tab_widget)
        """
        self._node = node # the Library.Nodes.Plot.Plot2DNode instance
        self._qwt_plot_widget = qwt_plot_widget # the plot2dwidget instance
        self._widget_id = widget_id
        self._dock_window = parent

        # plot types available
        self._plot_types = [ScatterPlot, LinePlot, MultiLinePlot, ImagePlot,
                            HistogramPlot, DendrogramPlot, NormalQqPlot,
                            DensityPlot]

        QWidget.__init__(self, parent, name)
        self._layout = QGridLayout(self, 1, 2, 7)
        self._layout.setColStretch(0, 0) # no stretching col 0
        self._layout.setColStretch(1, 1) # do stretch col 1

        #listbox_buttons_container = QWidget(self)
        layout_listbox_buttons_container = QVBoxLayout()
        self._layout.addLayout(layout_listbox_buttons_container, 0, 0)

        # self._group_box_plots contains:
        #   - self._list_box_plot_types (QListBox)
        #   - push_button_add_plot_type (QPushButton)
        #   - self._push_button_delete_plot_type (QPushButton)
        self._group_box_plots = QGroupBox(self, 'group_box_plots')
        self._group_box_plots.setTitle('Plots')
        self._group_box_plots.setColumnLayout(3, Qt.Vertical)

        self._list_box_plot_types = QListBox(self._group_box_plots,
                                                'list_box_plot_types')
        connect(self._list_box_plot_types, SIGNAL('highlighted(int)'),
                self._slot_update_list_box)

        push_button_add_plot_type = QPushButton('Add...',
                                                   self._group_box_plots,
                                                   'push_button_add_plot_type')
        connect(push_button_add_plot_type, SIGNAL('clicked()'),
                self._slot_add_plot)
        self._push_button_delete_plot_type = QPushButton('Delete',
                                                            self._group_box_plots,
                                                            'push_button_delete_plot_type')
        connect(self._push_button_delete_plot_type,
                SIGNAL('clicked()'), self._slot_delete_plot)
        self._push_button_delete_plot_type.setEnabled(False)

        layout_listbox_buttons_container.addWidget(self._group_box_plots)

        push_button_configure_window = QPushButton('Configure Window', self,
                                                      'push_button_configure_window')
        connect(push_button_configure_window, SIGNAL('clicked()'),
                self._slot_configure_window)
        layout_listbox_buttons_container.addWidget(push_button_configure_window)

        # self._group_box_settings: placeholder in GUI when no plots are added
        # to the TabWidget
        self._group_box_settings = QGroupBox(self, 'group_box_settings')
        self._group_box_settings.setTitle('No plot added')

        self._layout.addWidget(self._group_box_settings, 0, 1)

        # list to hold ScatterPlotSettings / LinePlotSettings widgets
        self._plot_settings_widgets = []
        self._current_widget_index = -1 # index in listbox of selected item

        self._add_plots_from_node()

    def _add_plots_from_node(self):
        """Create all plots defined in node for this widget."""
        names = {}
        for plot in self._plot_types:
            names[plot.get_type()] = plot

        number_of_plots = self._node.get_number_of_plots(self._widget_id[0],
                                                         self._widget_id[1])
        for index in range(number_of_plots):
            config = self._node.get_plot_config(self._widget_id[0],
                                                self._widget_id[1], index)
            plot_type = config.get_setting('plot_type')
            if plot_type in names:
                self._add_plot_by_class(names[plot_type])
            else:
                QMessageBox.information(self, 'Invalid ZMD',
                                           'There was an error in your ZMD file:\nNo such plot type: %s' % plot_type)

    def get_title(self):
        """Returns the title of the window."""
        title = self._qwt_plot_widget.title()
        if title:
            return title
        else:
            return self._dock_window.get_tab_label(self._widget_id)

    def set_title(self, title):
        self._dock_window.set_tab_label(self, title)

    def get_widget_id(self):
        """Returns the id of the window this tab controls."""
        return self._widget_id

    def _slot_configure_window(self):
        # must keep reference to make the signals run their slots...
        self.__widget_settings_dialog = WidgetSettingsDialog(self,
                                                             self._qwt_plot_widget)

    def _slot_delete_plot(self):
        """Deletes a plot from the TabWidget. Removes the corresponding curve
        from the qwt_plot_widget."""
        if len(self._plot_settings_widgets):
            curr_idx = self._list_box_plot_types.currentItem()
            self._slot_update_list_box(curr_idx - 1)
            self._remove_list_box_item(curr_idx)
            popped_widget = self._plot_settings_widgets.pop(curr_idx)
            popped_widget.remove_from_plot_widget()
            popped_widget.close()

            for index in range(len(self._plot_settings_widgets)):
                self._plot_settings_widgets[index].set_index(index)
                self._plot_settings_widgets[index].save_settings()
            self._node.update_plots(self._widget_id[0], self._widget_id[1],
                                    len(self._plot_settings_widgets))
            if not len(self._plot_settings_widgets):
                # reset axis settings (i.e. if there is no plots left...)
                self._qwt_plot_widget.reset_axis_settings()
                self._push_button_delete_plot_type.setEnabled(False)

    def _slot_add_plot(self):
        """Allow user to select which new plot to create."""
        add_dlg = QWidgetFactory.create(os.path.join(Config.uidir,
                                                     'dialog_add_plot.ui'),
                                        self, self)
        # fill button group
        plot_list = add_dlg.child('plot_list')

        for plot in self._plot_types:
            # id of each plot is same as list-position
            plot_list.insertItem(plot.get_name())

        plot_list.setSelected(0, True)

        if add_dlg.exec_loop():
            self._node.update_plots(self._widget_id[0], self._widget_id[1],
                                    len(self._plot_settings_widgets)+1)
            self._add_plot_by_class(self._plot_types[plot_list.currentItem()])
            self._plot_settings_widgets[-1].save_settings()

    def _add_plot_by_class(self, plot_class):
        next_index = len(self._plot_settings_widgets)

        config = self._node.get_plot_config(self._widget_id[0],
                                            self._widget_id[1], next_index)

        plot = plot_class(self._qwt_plot_widget, self._node, config)

        settings = plot.create_settings(next_index, self, self._qwt_plot_widget, self._node)
        self._plot_settings_widgets.append(settings)
        self._slot_update_list_box(next_index)
        self._push_button_delete_plot_type.setEnabled(True)
        self._dock_window.update_geometry()

    def _add_plot_by_name(self, plot_type_string):
        """Create new plot based on string."""
        names = {}
        for plot_type in self._plot_types:
            names[plot_type.get_name()] = plot_type

        if plot_type_string in names:
            self._add_plot_by_class(names[plot_type_string])
        else:
            raise NoSuchPlotError, 'No known plottype named "%s".' % plot_type_string

    def _slot_update_list_box(self, index):
        """The list box has activated another item (index). Selects the widget
        corresponding to the item at index index in the ListBox.
        """
        # ignore requests to set a too large index (might occur when
        # slot is triggered before setup is complete)
        if self._current_widget_index >= len(self._plot_settings_widgets):
            return

        if self._current_widget_index == -1:
            self.set_default_group_box_hidden(True)
        else:
            self._plot_settings_widgets[self._current_widget_index].set_hidden(True)
        if index < 0 and len(self._plot_settings_widgets) == 0:
            self.set_default_group_box_hidden(False)
        else:
            if index < 0:
                index = 0
            if self._plot_settings_widgets:
                self._plot_settings_widgets[index].set_hidden(False)
                self._set_highlighted_item(index)
        self._current_widget_index = index

    def set_default_group_box_hidden(self, hidden=True):
        """Hides/unhides the group box shown when no plots have been added."""
        self._group_box_settings.setHidden(hidden)

    def set_list_box_text(self, id_, text):
        """Sets the text at index id in the listbox to text.

        Parameters:
        id_ -- the index in the listbox to modify
        text -- the text to be set on index id in listbox
        """
        self._list_box_plot_types.changeItem(text, id_)

    def add_list_box_text(self, id_=-1, text="<Not Set>"):
        """Adds an item to the listbox at index id with text text.

        Parameters:
        id_=-1            -- index in listbox (default is append)
        text="<Not Set>" -- text to set at the index
        """
        self._list_box_plot_types.insertItem(text, id_)

    def _remove_list_box_item(self, id_):
        """Removes the item at index id in the listbox.

        Parameters:
        id_ -- index in listbox to be removed.
        """
        self._list_box_plot_types.removeItem(id_)

    def _set_highlighted_item(self, id_):
        """Highlights the given index in listbox.

        Parameters
        id -- index to highlight in listbox.
        """
        self._list_box_plot_types.setCurrentItem(id_)

    def update_combo_boxes_data(self):
        """Tells the PlotSettings instances associated with this TabWidget to
        update its comboboxes with input ports.
        """
        for plot_settings in self._plot_settings_widgets:
            plot_settings.update_data_names_combo_boxes()

    def update_data(self):
        for plot_settings in self._plot_settings_widgets:
            plot_settings.update_data()


class SettingsDialog:
    def __init__(self, parent, node, main_window):
        self._widget = QWidgetFactory().create(
            os.path.join(Config.uidir, "plot_2d_node_settings.ui"),
            None, parent, None)
        self._node = node
        self._main_window = main_window
        self._parent = parent        
        self._ready_settings_dialog()
        self._ready_port_settings()
        self._init_plot_map_settings_dialog()
        self._plotmap_dialog = AddPlotMapDialog(self._widget, self, parent)

        self._temp_ports = {}

    def show(self):
        self._ready_port_settings()
        self._ready_plot_map()
        self._widget.show()

    # input ports settings part

    def _save_port_settings(self):
        saved_input_ports = self._node.get_input_port_descriptions().copy()
        if saved_input_ports != self._temp_ports:
            for temp_port_key in self._temp_ports:
                if not saved_input_ports.has_key(temp_port_key):
                    self._node.create_input_port(
                        temp_port_key,
                        self._temp_ports[temp_port_key][0],
                        self._temp_ports[temp_port_key][1])
            for port_key in saved_input_ports:
                if not self._temp_ports.has_key(port_key):
                    self._node.remove_input_port(port_key)

    def _ready_port_settings(self):
        self._temp_ports = self._node.get_input_port_descriptions().copy()
        self._line_edit_name.setText("")
        self._line_edit_type.setText("")
        self._text_edit_description.setText("")
        self._list_box_ports.clear()

        for key in self._temp_ports.keys():
            self._list_box_ports.insertItem(str(key))

        self._line_edit_name.setFocus()

    def _ready_settings_dialog(self):

        dialog = self._widget

        #Getting childs for render window settings
        self._spin_box_x_count = dialog.child("spin_box_horizontal_plot_widgets")
        self._spin_box_y_count = dialog.child("spin_box_vertical_plot_widgets")

        #Getting childs for port settings
        self._line_edit_name = dialog.child("line_edit_name")
        self._line_edit_type = dialog.child("line_edit_type")
        self._text_edit_description = dialog.child("text_edit_description")
        self._list_box_ports = dialog.child("list_box_ports")
        self._push_button_add_port = dialog.child("push_button_add_port")
        self._push_button_update_port = dialog.child("push_button_update_port")
        self._push_button_delete_port = dialog.child("push_button_delete_port")

        #Getting remaining childs
        self._push_button_ok = dialog.child("push_button_ok")
        self._push_button_apply = dialog.child("push_button_apply")
        self._push_button_cancel = dialog.child("push_button_cancel")
        self._push_button_help = dialog.child("push_button_help")

        #Setting some static preferences
        self._push_button_update_port.setEnabled(False)
        self._push_button_delete_port.setEnabled(False)

        #Connecting signals to slots
        connect(self._line_edit_name,
                       SIGNAL("textChanged(const QString &)"),
                       self._slot_line_edit_name_changed)
        connect(self._line_edit_type,
                       SIGNAL("textChanged(const QString &)"),
                       self._slot_line_edit_type_changed)
        connect(self._text_edit_description,
                       SIGNAL("modificationChanged(bool)"),
                       self._slot_text_edit_description_changed)
        connect(self._list_box_ports,
                       SIGNAL("highlighted(int)"),
                       self._slot_list_box_ports_item_selected)
        connect(self._push_button_add_port,
                       SIGNAL("clicked()"),
                       self._slot_add_port_clicked)
        connect(self._push_button_update_port,
                       SIGNAL("clicked()"),
                       self._slot_update_port_clicked)
        connect(self._push_button_delete_port,
                       SIGNAL("clicked()"),
                       self._slot_delete_port_clicked)
        connect(self._push_button_ok,
                       SIGNAL("clicked()"),
                       self._slot_ok_clicked)
        connect(self._push_button_apply,
                       SIGNAL("clicked()"),
                       self._slot_apply_clicked)
        connect(self._push_button_cancel,
                       SIGNAL("clicked()"),
                       self._slot_cancel_clicked)
        connect(self._push_button_help,
                       SIGNAL("clicked()"),
                       self._slot_help_clicked)                       

    def _slot_line_edit_name_changed(self, q_string):
        if str(q_string) and self._list_box_ports.currentText() and str(q_string) != str(self._list_box_ports.currentText()):
            self._push_button_update_port.setEnabled(True)

    def _slot_line_edit_type_changed(self, q_string):
        if str(q_string) and self._list_box_ports.currentText() and str(q_string) != self._temp_ports[str(self._list_box_ports.currentText())][0]:
            self._push_button_update_port.setEnabled(True)

    def _slot_text_edit_description_changed(self, value):
        if self._list_box_ports.currentText() and value:
            self._push_button_update_port.setEnabled(True)

    def _slot_list_box_ports_item_selected(self, curr_index):
        if self._list_box_ports.count():
            currently_selected = self._list_box_ports.currentText()
            self._line_edit_name.setText(str(currently_selected))
            self._line_edit_type.setText(
                self._temp_ports[str(currently_selected)][0])
            self._text_edit_description.setText(
                self._temp_ports[str(currently_selected)][1])
            self._push_button_delete_port.setEnabled(True)
            self._push_button_update_port.setEnabled(False)

    def _slot_add_port_clicked(self):
        if str(self._line_edit_name.text()) and not self._list_box_ports.findItem(self._line_edit_name.text(), Qt.ExactMatch):
            self._list_box_ports.insertItem(self._line_edit_name.text())
            self._temp_ports[str(self._line_edit_name.text())] = (
                str(self._line_edit_type.text()),
                str(self._text_edit_description.text()))
            self._line_edit_name.setText("")
            self._line_edit_type.setText("")
            self._text_edit_description.setText("")
            self._push_button_update_port.setEnabled(False)
            self._push_button_delete_port.setEnabled(False)
            self._line_edit_name.setFocus()

    def _slot_update_port_clicked(self):
        if self._list_box_ports.currentText() and self._line_edit_name.text():
            self._temp_ports[str(self._line_edit_name.text())] = (
                str(self._line_edit_type.text()),
                str(self._text_edit_description.text()))
            if str(self._list_box_ports.currentText()) != str(self._line_edit_name.text()) and not self._list_box_ports.findItem(self._line_edit_name.text(), Qt.ExactMatch):
                self._temp_ports.pop(str(self._list_box_ports.currentText()))
                self._list_box_ports.changeItem(self._line_edit_name.text(), self._list_box_ports.currentItem())
            elif str(self._list_box_ports.currentText()) != str(self._line_edit_name.text()) and self._list_box_ports.findItem(self._line_edit_name.text(), Qt.ExactMatch):
                QMessageBox.information(
                    self._widget,
                    "Update failed",
                    "Port name already exists. Changes will not be saved.",
                    QMessageBox.Ok)
                return
            self._text_edit_description.setModified(False)

            self._push_button_update_port.setEnabled(False)

    def _slot_delete_port_clicked(self):
        if self._list_box_ports.currentText():
            temp_curr_text = self._list_box_ports.currentText()
            self._temp_ports.pop(str(temp_curr_text))
            self._list_box_ports.takeItem(self._list_box_ports.findItem(temp_curr_text))
            self._line_edit_name.setText("")
            self._line_edit_type.setText("")
            self._text_edit_description.setText("")
            self._push_button_delete_port.setEnabled(False)
            self._line_edit_name.setFocus()


    # Plot mapping settings part

    def _init_plot_map_settings_dialog(self):
        """Initialise plot map settings part."""
        dialog = self._widget
        self._add_map_list = []
        self._delete_map_list = []
        self._map_existing_plots = False

        self.handler = self._node.get_plot_map_handler()

        self._enabled = dialog.child("chkbox_enabled")
        self._auto_map_enabled = dialog.child("chkbox_auto_mapping")
        self._listview_mappings = dialog.child("listview_mappings")
        self._button_add_mapping = dialog.child("button_add_mapping")
        self._button_delete_mapping = dialog.child("button_delete_mapping")

        connect(self._button_add_mapping, SIGNAL("clicked()"),
                self._slot_add_mapping)
        connect(self._button_delete_mapping, SIGNAL("clicked()"),
                self._slot_delete_mapping)
        connect(self._auto_map_enabled, SIGNAL("clicked()"),
                self._slot_auto_map_enabled_change)
        connect(self._enabled, SIGNAL("toggled(bool)"),
                self._slot_auto_map_enabled_change)

    def _ready_plot_map(self):
        self._add_map_list = []
        self._delete_map_list = []
        self._map_existing_plots = False

        self._enabled.setChecked(self.handler.get_enabled())
        self._auto_map_enabled.setChecked(self.handler.get_auto_map_enabled())

        self._listview_mappings.clear()
        all_maps = self.handler.get_plot_maps() + self.handler.get_go_browser_maps()
        for plot_map in all_maps:
            plot1, plot2 = plot_map.plots()
            aspect1 = plot_map.get_aspect_for_plot(plot1)
            aspect2 = plot_map.get_aspect_for_plot(plot2)
            PlotAspectMapListViewItem(self._listview_mappings, plot1, plot2,
                                      aspect1, aspect2, False)

    def _save_plot_map_settings(self):
        self.handler.set_enabled(self._enabled.isChecked())
        self.handler.set_auto_map_enabled(self._auto_map_enabled.isChecked())

        for item in self._delete_map_list:
            self.handler.remove_plot_aspect_map(item.plot1(), item.plot2(),
                                                item.aspect1(), item.aspect2())

        for item in self._add_map_list:
            self.handler.add_aspect_map(item.plot1(), item.plot2(),
                                             item.aspect1(), item.aspect2())

        if self._map_existing_plots:
            self._map_existing_plots = False
            self.handler.auto_map_all(self._parent.get_plots())

        self.handler.save_settings()

    def _slot_add_mapping(self):
        self._plotmap_dialog.show()


    def _slot_delete_mapping(self):
        item = self._listview_mappings.selectedItem()
        if item:
            if item.new():
                try:
                    self._add_map_list.remove(item)
                except ValueError:
                    pass
            else:
                self._delete_map_list.append(item)
            self._auto_map_enabled.setChecked(0)                
        self._listview_mappings.takeItem(item)

    def _slot_auto_map_enabled_change(self):
        if self._enabled.isChecked() and \
               self._auto_map_enabled.isChecked() and \
               (len(self._parent.get_plots()) > 1):
            msg = "New plots will now automatically be mapped to existing " +\
                  "plots.\nWould you also like to map all your existing " +\
                  "plots to each other?"
            result = QMessageBox.question(self._widget, "Automatic mapping",
                                             msg, QMessageBox.Yes,
                                             QMessageBox.No)
            if (result == QMessageBox.Yes):
                self._map_existing_plots = True
                self._preview_automap_list()

    def _preview_automap_list(self):
        """Shows the mappings that would be applied when pressing apply or ok.

        Displays a preview of the mappings in the ListView of maps. The changes
        to the mappings is not applied before pressing "apply" or "ok".
        """
        self._listview_mappings.clear()
        plot_widgets = map(self._parent.get_plot_widget,
                           self._parent.get_plot_widgets())

        titles, mappable_plots = zip(*self._parent.get_mappable_plots())
        plots = flatten(mappable_plots)

        for plot1 in plots:
            for plot2 in plots:
                if not plot1 == plot2:
                    for aspect1 in plot1.get_mappable_aspects():
                        for aspect2 in plot2.get_mappable_aspects():
                            if not self._mapping_already_previewed(plot1, plot2,
                                                                   aspect1,
                                                                   aspect2):
                                item = PlotAspectMapListViewItem(
                                    self._listview_mappings, plot1, plot2,
                                    aspect1, aspect2, False)

        self._listview_mappings.triggerUpdate()

    def add_new_mapping(self, plot1, plot2, aspect1, aspect2):
        item = PlotAspectMapListViewItem(self._listview_mappings, plot1, plot2,
                                         aspect1, aspect2)
        self._add_map_list.append(item)

    def _mapping_already_previewed(self, plot1, plot2, aspect1, aspect2):
        """Checks if a mapping is already shown in the ListView.

        When a mapping is previewed in the ListView of maps, this method returns
        true (it doesn't have to be applied yet -- i.e. saved).
        """
        iterator = QListViewItemIterator(self._listview_mappings)
        while iterator.current():
            item = iterator.current()
            if item.contains(plot1) and item.contains(plot2):
                if item.plot1() == plot1:
                    if item.aspect1() == aspect1 and item.aspect2() == aspect2:
                        return True
                else:
                    if item.aspect2() == aspect1 and item.aspect1() == aspect2:
                        return True
            iterator += 1
        return False

    def set_spinners_values(self, hor_val, vert_val):
        self._spin_box_y_count.setValue(vert_val)
        self._spin_box_x_count.setValue(hor_val)

    # Common OK/Apply/Cancel/Help part

    def _slot_ok_clicked(self):
        self._slot_apply_clicked()
        self._widget.hide()

    def _slot_apply_clicked(self):
        self._main_window.update_plot_widgets(self._spin_box_x_count.value(),
                                              self._spin_box_y_count.value())
        self._node.update_widgets(self._spin_box_x_count.value(),
                                  self._spin_box_y_count.value())
        self._save_plot_map_settings()
        self._ready_plot_map()

        self._save_port_settings()
        self._main_window.update_dock_window()

    def _slot_cancel_clicked(self):
        self._widget.hide()

    def _slot_help_clicked(self):
        HelpViewer(self._widget, os.path.join(Config.htmldocpath,
                                              "node_Plotting_in_2D.html"),
                   Config.htmldocpath)


class PlotAspectMapListViewItem(QListViewItem):
    """Custom ListView class, displays plot mappings."""
    def __init__(self, listview, plot1, plot2, aspect1, aspect2,
                 new_plot_map=False):
        self._aspects = { plot1 : aspect1, plot2 : aspect2 }
        QListViewItem.__init__(self, listview, self._item_text(plot1),
                               self._item_text(plot2))

        self._plot1 = plot1
        self._plot2 = plot2

        self.new_plot_map = new_plot_map

    def plot1(self):
        return self._plot1

    def plot2(self):
        return self._plot2

    def contains(self, plot):
        return (plot == self._plot1) or (plot == self._plot2)

    def _item_text(self, plot):
        try:
            dock_window = plot.get_parent_plot().parent().get_dock_window()
            window_id = plot.get_parent_plot().id()
            window_title = str(dock_window.get_tab_label(window_id))
        except Exception:
            return plot.name()

        plot_text = plot.name()
        return "%s - %s - %s" % (plot_text, window_title,
                                 str(self._aspects[plot]))

    def aspect1(self):
        return self._aspects[self.plot1()]

    def aspect2(self):
        return self._aspects[self.plot2()]

    def new(self):
        return self.new_plot_map


class AddPlotMapDialog:
    """'Add plot map' dialog."""

    def __init__(self, parent_widget, parent_object, guiplot2dwindow):
        self._widget = QWidgetFactory().create(
            os.path.join(Config.uidir, "plot_2d_node_plotmap_settings.ui"),
            None, parent_widget, None)
        self._guiplot2dwindow = guiplot2dwindow
        self._parent = parent_object

        self._listview_plot1 = self._widget.child("listview_plot1")
        self._listview_plot2 = self._widget.child("listview_plot2")
        self._button_add_mapping = self._widget.child("button_add_mapping")

        connect(self._button_add_mapping, SIGNAL("clicked()"),
                self._slot_add_mapping)

    def show(self):
        self._ready_dialog()
        self._widget.show()

    def _ready_dialog(self):
        self._listview_plot1.clear()
        self._listview_plot2.clear()

        dock_window = self._guiplot2dwindow.get_dock_window()

        plot_widgets = map(self._guiplot2dwindow.get_plot_widget,
                           self._guiplot2dwindow.get_plot_widgets())

        self._parent._save_plot_map_settings()
        windows = self._guiplot2dwindow.get_mappable_plots()

        # Add each plot to list view with window as parent listitem
        for window_title, plots in windows:
            if plots:
                window_item1 = QListViewItem(self._listview_plot1, window_title)
                window_item2 = QListViewItem(self._listview_plot2, window_title)
                for plot in plots:
                    if plot.name() != 'Unknown':
                        for aspect in plot.get_mappable_aspects():
                            PlotListViewItem(window_item1, plot, aspect)
                            PlotListViewItem(window_item2, plot, aspect)

                self._listview_plot1.insertItem(window_item1)
                self._listview_plot2.insertItem(window_item2)

                window_item1.setOpen(True)
                window_item2.setOpen(True)

    def _slot_add_mapping(self):
        item1 = self._listview_plot1.selectedItem()
        item2 = self._listview_plot2.selectedItem()

        error_message = None

        if item1 and item2:
            if isinstance(item1, PlotListViewItem) and isinstance(item2, PlotListViewItem):
                if item1.plot() == item2.plot():
                    error_message = 'Cannot map plot to itself.'
                elif self._parent._mapping_already_previewed(item1.plot(), item2.plot(),
                                                                item1.aspect(), item2.aspect()):
                    error_message = 'This mapping has already been added.'
                else:
                    self._parent.add_new_mapping(item1.plot(), item2.plot(),
                                                 item1.aspect(), item2.aspect())
            else:
                error_message = 'Windows cannot be mapped.\nPlease select two plots.'
        else:
            error_message = 'Please select two plots.'

        if error_message:
            QMessageBox.information(self._widget, 'Error',
                error_message)


class PlotListViewItem(QListViewItem):
    """Custom ListViewItem class, displays plots."""

    def __init__(self, parent_item, plot, aspect):
        QListViewItem.__init__(self, parent_item, "%s - %s" % (plot.name(),
                                                               str(aspect)))
        self._plot = plot
        self._aspect = aspect

    def plot(self):
        return self._plot

    def aspect(self):
        return self._aspect
