#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from Library.Util import gui
from Library.Util.Config import Config
from Library.Util.unrepr import unrepr

from qt import SIGNAL, Qt, QColorDialog, QObject, QMessageBox
from qt import QProgressDialog, qApp, QListViewItem
from qtui import QWidgetFactory

import os
import qt
import re
import sre
import sys
import types


connect = QObject.connect

# domain exceptions
class NoSuchParameterError(Exception):
    """Tried to get a non-existant parameter."""
    pass


class Settings:
    """Super class for all GUI handling Settings. Provides a dictionary that is
    private, with methods for getting and setting parameters.
    """
    
    def __init__(self):
        self._parameters = {}
        
    def set_parameter(self, key, value=None):
        """Sets a parameter on this instance.

        Parameters:
        key   -- the name of the parameter
        value -- the value of the parameter (default: None)
        """
        if type(value) == types.StringType and key == 'autoupdate':
            value = value == 'True' #sets value to the boolean True or False

        self._parameters[key] = value

    def get_parameter(self, key, *args):
        """Gets the value of the given key.

        When a default argument is given, it is returned when the
        parameter doesn't exist; without it, an exception is raised.
        """
        if self._parameters.has_key(key):
            return self._parameters[key]
        else:
            if len(args):
                return args[0]
            else:
                raise NoSuchParameterError, key

    def get_parameters(self):
        """Returns a list of all keys to the parameters.

        Returns: a list of all keys to the parameters
        """
        return self._parameters.keys()

    def _convert_dict(self, dictionary):
        """Converts string repr. to the correct data type."""
        for key in dictionary.keys():
            if dictionary[key]:
                old = dictionary[key]
                try:
                    dictionary[key] = unrepr(dictionary[key])
                except Exception, e:
                    pass

        return dictionary

# helper functions to WidgetSettingsDialog
_grid_styles = [Qt.SolidLine, Qt.DashLine, Qt.DotLine, Qt.DashDotLine,
                Qt.DashDotDotLine]

def _item_to_style(item):
    """Returns what line-style is on item in style combo box."""
    return _grid_styles[item]

def _style_to_item(style):
    """Returns which item in line-style combo box holds style."""
    return _grid_styles.index(style)


class WidgetSettingsDialog:

    def __init__(self, parent, plot_widget):
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'dialog_window_config.ui'),
                                             None, parent)
        self._plot_widget = plot_widget

        # general settings
        bgcolour = self._plot_widget.get_general_settings()[0]
        self._bgcolour_button = self._widget.child('background_colour')
        self._bgcolour_button.setPaletteBackgroundColor(bgcolour)

        # labels settings
        labelsettings = self._plot_widget.get_labels_settings()
        self._widget.child('line_edit_window_title').setText(labelsettings[0])
        self._widget.child('line_edit_x_axis_label').setText(labelsettings[1])
        self._widget.child('line_edit_y_axis_label').setText(labelsettings[2])

        # grid settings
        gridsettings = self._plot_widget.get_grid_settings()
        self._maj_colour_button = self._widget.child('push_button_maj_grid_colour')
        self._min_colour_button = self._widget.child('push_button_min_grid_colour')
        self._maj_x_enabled = self._widget.child('check_box_x_grid_maj_enable')
        self._maj_y_enabled = self._widget.child('check_box_y_grid_maj_enable')
        self._min_x_enabled = self._widget.child('check_box_x_grid_min_enable')
        self._min_y_enabled = self._widget.child('check_box_y_grid_min_enable')
        self._maj_x_enabled.setChecked(gridsettings[0])
        self._maj_y_enabled.setChecked(gridsettings[1])
        self._maj_colour_button.setPaletteBackgroundColor(gridsettings[2])
        self._widget.child('spin_box_maj_grid_width').setValue(gridsettings[3])
        self._widget.child('combo_box_maj_grid_style').setCurrentItem(_style_to_item(gridsettings[4]))
        self._min_x_enabled.setChecked(gridsettings[5])
        self._min_y_enabled.setChecked(gridsettings[6])
        self._min_colour_button.setPaletteBackgroundColor(gridsettings[7])
        self._widget.child('spin_box_min_grid_width').setValue(gridsettings[8])
        self._widget.child('combo_box_min_grid_style').setCurrentItem(_style_to_item(gridsettings[9]))

        # set correct initial state on gui elements for grid settings
        self._slot_enable_major_grid()
        self._slot_enable_minor_grid() 

        # axes settings
        axessettings = self._plot_widget.get_axes_settings()
        self._widget.child('check_box_x_auto_scale').setChecked(axessettings[0])
        self._x_min = self._widget.child('line_edit_x_min')
        self._x_max = self._widget.child('line_edit_x_max')
        self._y_min = self._widget.child('line_edit_y_min')
        self._y_max = self._widget.child('line_edit_y_max')
        self._ratio = self._widget.child('line_edit_ratio')
        self._x_min.setText(str(axessettings[1]))
        self._x_max.setText(str(axessettings[2]))
        self._widget.child('check_box_y_auto_scale').setChecked(axessettings[3])
        self._y_min.setText(str(axessettings[4]))
        self._y_max.setText(str(axessettings[5]))
        self._widget.child('check_box_ratio').setChecked(axessettings[6])
        self._ratio.setText(str(axessettings[7]))

        self._connect_signals()

        self._widget.show()

    def _connect_signals(self):
        connect(self._widget.child('buttonOk'), SIGNAL('clicked()'),
                self._slot_ok_pressed)
        
        # background colour button
        connect(self._bgcolour_button, SIGNAL('clicked()'),
                self._slot_background_colour)

        # grid connections
        connect(self._maj_colour_button, SIGNAL('clicked()'),
                self._slot_major_grid_colour)
        connect(self._min_colour_button, SIGNAL('clicked()'),
                self._slot_minor_grid_colour)
        connect(self._maj_x_enabled, SIGNAL('clicked()'),
                self._slot_enable_major_grid)
        connect(self._maj_y_enabled, SIGNAL('clicked()'),
                self._slot_enable_major_grid)
        connect(self._min_x_enabled, SIGNAL('clicked()'),
                self._slot_enable_minor_grid)
        connect(self._min_y_enabled, SIGNAL('clicked()'),
                self._slot_enable_minor_grid)

        # axes connections
        connect(self._x_min, SIGNAL('lostFocus()'), self._slot_x_min)
        connect(self._x_max, SIGNAL('lostFocus()'), self._slot_x_max)
        connect(self._y_min, SIGNAL('lostFocus()'), self._slot_y_min)
        connect(self._y_max, SIGNAL('lostFocus()'), self._slot_y_max)
        connect(self._ratio, SIGNAL('lostFocus()'), self._slot_ratio)

    def _slot_ok_pressed(self):
        settings = {}
        
        # apply general settings
        bgcolour = self._bgcolour_button.paletteBackgroundColor()
        settings['background_colour'] = str(bgcolour.name())

        # labels settings
        settings['window_title'] = str(self._widget.child('line_edit_window_title').text())
        settings['xaxis_title'] = str(self._widget.child('line_edit_x_axis_label').text())
        settings['yaxis_title'] = str(self._widget.child('line_edit_y_axis_label').text())

        # grid settings
        settings['x_grid_maj_enable'] = str(int(self._maj_x_enabled.isChecked()))
        settings['x_grid_min_enable'] = str(int(self._min_x_enabled.isChecked()))
        settings['y_grid_maj_enable'] = str(int(self._maj_y_enabled.isChecked()))
        settings['y_grid_min_enable'] = str(int(self._min_y_enabled.isChecked()))
        settings['maj_grid_style'] = str(self._widget.child('combo_box_maj_grid_style').currentItem())
        settings['min_grid_style'] = str(self._widget.child('combo_box_min_grid_style').currentItem())
        settings['maj_grid_width'] = str(self._widget.child('spin_box_maj_grid_width').value())
        settings['min_grid_width'] = str(self._widget.child('spin_box_min_grid_width').value())
        settings['maj_grid_colour'] = str(self._maj_colour_button.paletteBackgroundColor().name())
        settings['min_grid_colour'] = str(self._min_colour_button.paletteBackgroundColor().name())

        # axes settings
        settings['x_axis_auto_scale'] = str(int(self._widget.child('check_box_x_auto_scale').isChecked()))
        settings['y_axis_auto_scale'] = str(int(self._widget.child('check_box_y_auto_scale').isChecked()))
        settings['x_axis_min'] = str(self._x_min.text())
        settings['x_axis_max'] = str(self._x_max.text())
        settings['y_axis_min'] = str(self._y_min.text())
        settings['y_axis_max'] = str(self._y_max.text())
        settings['ratio'] = str(int(self._widget.child('check_box_ratio').isChecked()))
        settings['aspect_ratio'] = str(self._ratio.text())

        # apply settings
        self._plot_widget.set_settings(settings)

    def _set_colour_on_button(self, button):
        """Asks the user to choose a colour, then sets it on button."""
        colour = QColorDialog.getColor(button.paletteBackgroundColor(),
                                       self._widget)
        if colour.isValid():
            button.setPaletteBackgroundColor(colour)

    def _slot_background_colour(self):
        self._set_colour_on_button(self._bgcolour_button)

    def _slot_major_grid_colour(self):
        self._set_colour_on_button(self._maj_colour_button)
    
    def _slot_minor_grid_colour(self):
        self._set_colour_on_button(self._min_colour_button)

    def _set_grid_widgets_enabled(self, mode, enable):
        for child in ['push_button_%s_grid_colour'%mode,
                      'spin_box_%s_grid_width'%mode,
                      'combo_box_%s_grid_style'%mode,
                      'text_label_%s_grid_colour'%mode,
                      'text_label_%s_grid_width'%mode,
                      'text_label_%s_grid_style'%mode]:
            self._widget.child(child).setEnabled(enable)

    def _slot_enable_major_grid(self):
        if self._maj_x_enabled.isChecked() or self._maj_y_enabled.isChecked():
            self._set_grid_widgets_enabled('maj', True)
        else:
            self._set_grid_widgets_enabled('maj', False)

    def _slot_enable_minor_grid(self):
        if self._min_x_enabled.isChecked() or self._min_y_enabled.isChecked():
            self._set_grid_widgets_enabled('min', True)
        else:
            self._set_grid_widgets_enabled('min', False)

    def _ensure_valid_float_strings(self, string):
        """Returns a string which is parseable by float() from the given string"""
        try:
            return str(float(sre.match('[+|-]?[0-9]+[\.[0-9]*]?', string).group()))
        except:
            return '0.0'

    def _slot_x_max(self):
        x_max = str(self._x_max.text())
        self._x_max.setText(self._ensure_valid_float_strings(x_max))

    def _slot_x_min(self):
        x_min = str(self._x_min.text())
        self._x_min.setText(self._ensure_valid_float_strings(x_min))

    def _slot_y_max(self):
        y_max = str(self._y_max.text())
        self._y_max.setText(self._ensure_valid_float_strings(y_max))

    def _slot_y_min(self):
        y_min = str(self._y_min.text())
        self._y_min.setText(self._ensure_valid_float_strings(y_min))

    def _slot_ratio(self):
        ratio = str(self._ratio.text())
        self._ratio.setText(self._ensure_valid_float_strings(ratio))


class PlotSettings(Settings):
    def __init__(self, node, index, tab_widget, qwt_plot_widget, plot,
                 settings_widget=None):
        """Constructor. Is super class, so it should NEVER BE INSTANTIATED.
        Should ONLY (and must) be called from the subclasses.

        Parameters:
        node            -- The node to retrieve data from
        index           -- The index of the item in the listbox
        tab_widget      -- The widget that contains this one and the listbox
        qwt_plot_widget -- The plot_widget we control
        """
        Settings.__init__(self)
        self._tab_widget = tab_widget # parent
        self._index = index # my own index in the listbox of parent
        self._node = node   # the node we get input from
        # the plot_widget in which we draw
        self._qwt_plot_widget = qwt_plot_widget 
        self._plot = plot # the plot using these settings
        self._settings_widget = settings_widget
        
        self._x_hash = None
        self._y_hash = None
        self._indices_hash = None

        tab_widget.add_list_box_text(index, plot.get_name())
        if not plot.get_settings().has_key('name'):
            self.set_parameter('name', plot.get_name()) # the name to display in the listbox
        self.set_parameter('plot_type', plot.get_type())
        self.save_settings()
 
        self._info_pattern = re.compile(" \(.*?\)$")

    def set_hidden(self, hidden):
        """(Un)Hides the widget associated with this PlotSettings instance."""
        self._settings_widget.setHidden(hidden)

    def close(self):
        """Closes the widget assosciated with this instance."""
        self._settings_widget.close()

    def _has_parameter(self, key):
        """Check the existance of a given parameter by key."""
        return self._parameters.has_key(key)

    def set_default_parameter(self, key, value):
        """Set parameter if there is no existing value."""
        if not self._has_parameter(key):
            self.set_parameter(key, value)

    def save_settings(self):
        settings = self._plot.get_settings()
        settings.update(self._parameters)
        self._plot.set_settings(settings)

    def set_index(self, index):
        self._index = index

    def remove_from_plot_widget(self):
        self._plot.remove_from_widget()

    def _get_port_name(self, data_port_name):
        inport = self._node.getInputPortByName(data_port_name)
        # check to see if data is set
        if inport and inport.getData(): 
            port_name = inport.getData().getName() # the name of the Data 
        else:                                        # object on input port
            port_name = None
        return port_name

    def update_data(self):
        """Called whenever data has changed/source is changed.

        This method should be overloaded in subclasses to redraw plot
        after the source of data has changed.
        """
        pass

    def _slot_combo_box_x_axis_data_changed(self, item_text):
        self._set_max_value_spin_boxes()
        x_axis_data = self._info_pattern.sub('', str(item_text))
        self.set_parameter('x_axis_data', x_axis_data)        
        self.save_settings()
        if self.get_parameter('autoupdate'):
            self.update_data()
            

    def _slot_combo_box_y_axis_data_changed(self, item_text):
        self._set_max_value_spin_boxes()
        y_axis_data = self._info_pattern.sub('', str(item_text))
        self.set_parameter('y_axis_data', y_axis_data)
        self.save_settings()
        if self.get_parameter('autoupdate'):
            self.update_data()


    def _slot_combo_box_x_col_row_changed(self, item_index):
        self.set_parameter('x_col_row', int(item_index))
        self.save_settings()
        if self.get_parameter('autoupdate'):
            self.update_data()


    def _slot_combo_box_y_col_row_changed(self, item_index):
        self.set_parameter('y_col_row', int(item_index))
        self.save_settings()
        if self.get_parameter('autoupdate'):
            self.update_data()


    def _slot_spin_box_x_index_changed(self, value):
        self.set_parameter('x_index', int(value)-1)
        self.save_settings()
        if self.get_parameter('autoupdate'):
            self.update_data()
            

    def _slot_spin_box_y_index_changed(self, value):
        self.set_parameter('y_index', int(value)-1)
        self.save_settings()
        if self.get_parameter('autoupdate'):
            self.update_data()
            
            
    def _slot_line_edit_range_changed(self, text):
        self.set_parameter('range', str(self._line_edit_range.text()))
        self.save_settings()

    def _slot_range_return_pressed(self):
        if not str(self._line_edit_range.text()):
            self._line_edit_range.setText("1:end")
        if self.get_parameter('autoupdate'):
            self.update_data()


    def _slot_plot_name_changed(self, name):
        self.set_parameter('name', str(name))
        self._tab_widget.set_list_box_text(self._index, name)
        self.save_settings()

    def _slot_checkbox_changed(self, bool_):
        self.set_parameter('autoupdate', bool(bool_))
        self.save_settings()

    def _load_tooltip_settings(self):
        try:
            chosen = self.get_parameter('label_tooltips')
        except NoSuchParameterError:
            chosen = 'None'
        # if chosen == 'None' do nothing
        if chosen == 'Indexed from 1':
            prefix = self.get_parameter('label_tooltips_prefix')
            self._set_tooltips_indexed(prefix)
        elif chosen == 'From data':
            portname = self.get_parameter('label_tooltips_data')
            # boolean type
            is_col = (self.get_parameter('label_tooltips_colrow') == 'column')
            # integer type
            index = int(self.get_parameter('label_tooltips_index'))
            self._set_tooltips_from_data(portname, is_col, index)
        
    def _slot_set_tooltips(self):
        widget = QWidgetFactory.create(
            os.path.join(Config.uidir, 'tooltip_labels.ui'),
            self._tab_widget, self._tab_widget)
        groupbox_tooltips = widget.child('button_group_tooltip_labels')
        textbox_prefix    = widget.child('text_box_prefix')
        combo_data        = widget.child('combo_box_data')
        combo_colrow      = widget.child('combo_box_col_row')
        spinbox_index     = widget.child('spin_box_index')

        # initiate widget
        for port in self._node.get_all_input_ports():
            combo_data.insertItem(port.getName())
        try:
            chosen = self.get_parameter('label_tooltips')
        except NoSuchParameterError:
            chosen = 'None'
        # if chosen == 'None' we don't have to init
        if chosen == 'Indexed from 1':
            widget.child('choice_indexed').setChecked(True)
            textbox_prefix.setText(self.get_parameter('label_tooltips_prefix'))
        elif chosen == 'From data':
            widget.child('choice_data').setChecked(True)
            combo_data.setCurrentItem(
                gui.get_index(combo_data,
                              self.get_parameter('label_tooltips_data')))
            combo_colrow.setCurrentItem(
                gui.get_index(combo_colrow,
                              self.get_parameter('label_tooltips_colrow')))
            spinbox_index.setValue(
                int(self.get_parameter('label_tooltips_index')))

        # display widget
        if widget.exec_loop():
            # OK button pressed
            chosen = str(groupbox_tooltips.selected().text())
            prefix = str(textbox_prefix.text())
            portname = str(combo_data.currentText())
            is_column = (str(combo_colrow.currentText()) == 'column')
            index = spinbox_index.value()
            if chosen == 'None':
                self._set_tooltips_none()
            elif chosen == '&Indexed from 1':
                self._set_tooltips_indexed(prefix)
            elif chosen == 'From &data':
                self._set_tooltips_from_data(portname, is_column, index - 1)
            self.save_settings()
        else:
            # Cancel button pressed
            pass

    def _set_tooltips_none(self):
        self.set_parameter('label_tooltips', 'None')

    def _set_tooltips_indexed(self, prefix):
        self.set_parameter('label_tooltips', 'Indexed from 1')
        self.set_parameter('label_tooltips_prefix', prefix)

    def _set_tooltips_from_data(self, portname, is_column, index):
        if is_column:
            colrow = 'column'
        else:
            colrow = 'row'
        self.set_parameter('label_tooltips', 'From data')
        self.set_parameter('label_tooltips_data', portname)
        self.set_parameter('label_tooltips_colrow', colrow)
        self.set_parameter('label_tooltips_index', index)

    def _set_max_value_spin_boxes(self):
        try:
            try:
                x_size = self._node.get_data_size(
                    self._info_pattern.sub('', str(self._combo_box_x_axis_data.currentText())),
                    self._combo_box_x_col_row.currentText() == 'column')
            except Exception, e:
                # Exception thrown if x_axis_data is 'Indexed'
                if str(e) == '[Plot2DNode]: no port called Indexed':
                    x_size = 100000
                else:
                    raise e
            y_size = self._node.get_data_size(
                self._info_pattern.sub('', str(self._combo_box_y_axis_data.currentText())),
                self._combo_box_y_col_row.currentText() == 'column')
        except AttributeError:
            # data not set yet
            pass
        except Exception, e:
            # no port called ''
            pass
        else:
            x_too_high = False
            y_too_high = False
            if self._spin_box_x_index.value() > x_size:
                x_too_high = True
            if self._spin_box_y_index.value() > y_size:
                y_too_high = True
            if x_too_high or y_too_high:
                info_string = ''
                if x_too_high:
                    info_string += 'The index in the X data set is too high.\n'
                if y_too_high:
                    info_string += 'The index in the Y data set is too high.\n'
            self._spin_box_x_index.setMaxValue(x_size)
            self._spin_box_y_index.setMaxValue(y_size)

    def __set_current_item(self, combobox, text):
        """Could not use the one in Library.Util.gui because of the dimension
        information.
        """
        if combobox.count() == 0:
            return 

        for index in range(combobox.count()):
            if self._info_pattern.sub('', str(combobox.text(index))) == str(text) \
                   or str(combobox.text(index)) == str(text):
                return combobox.setCurrentItem(index)

    def set_current_item_y_data(self, text):
        self.__set_current_item(self._combo_box_y_axis_data, text)

    def set_current_item_x_data(self, text):
        self.__set_current_item(self._combo_box_x_axis_data, text)
                        
    def update_data_names_combo_boxes(self):
        """updates the combo boxes that contains the names of the inputs."""
        print 'meeep'
        self._set_inports_in_combo_box(self._combo_box_x_axis_data)
        try:
            if self._combo_box_x_axis_data.count():
                self.set_current_item_x_data(self.get_parameter('x_axis_data'))
        except NoSuchParameterError:
            pass #set default current item (top)
        
        self._set_inports_in_combo_box(self._combo_box_y_axis_data)
        try:
            if self._combo_box_y_axis_data.count():
                self.set_current_item_y_data(self.get_parameter('y_axis_data'))
        except NoSuchParameterError:
            pass #set default current item (top)

        try:
            self.set_parameter('x_axis_data',
                               self._info_pattern.sub('',
                                                      str(self._combo_box_x_axis_data.currentText())))
            self.set_parameter('y_axis_data',
                               self._info_pattern.sub('',
                                                      str(self._combo_box_y_axis_data.currentText())))
            self.save_settings()
        except AttributeError:
            pass #not yet initialized
        
    def _set_inports_in_combo_box(self, combo_box):
        """Sets the names of the input ports in the given combo box

        Parameters:
        combo_box -- the combo box to set the input ports on.
        """
        combo_box.clear()
        for inputport in self._node.get_all_input_ports():
            if inputport.datainfo != '':
                dim_string = " (%s)" % inputport.datainfo.replace('\n', '')
            else:
                dim_string = ""
            combo_box.insertItem(inputport.getName() + dim_string)


class LabelSettingsDialog(Settings):

    def __init__(self, parent, node, qwt_plot_widget, plot):
        """Initializes a new LabelSettings dialog."""
        Settings.__init__(self)
        self._node = node
        self._qwt_plot_widget = qwt_plot_widget
        self._label_settings_belong_to = parent.get_parameter("plot_type")
        self._plot_settings = parent
        self._plot = plot

        factory = QWidgetFactory()
        self.labels_dialog = factory.create(
            os.path.join(Config.uidir, "plot_label_settings.ui"),
            None,
            parent._tab_widget)

        if self._label_settings_belong_to == "dendrogram_plot":
            self.choice_indexed = self.labels_dialog.child("choice_indexed")
            self.choice_indexed.setText("Use order indices")

        self.ok_button      = self.labels_dialog.child("ok_button")
        self.cancel_button  = self.labels_dialog.child('cancel_button')
        self.textbox_prefix = self.labels_dialog.child('text_box_prefix')
        self.combobox_data  = self.labels_dialog.child("combo_box_data")
        self.combobox_dir   = self.labels_dialog.child('combo_box_col_row')
        self.spinbox_index  = self.labels_dialog.child('spin_box_index')
        self.button_group   = self.labels_dialog.child("buttonGroup1")
        
        self.spinbox_label_size = self.labels_dialog.child('spinbox_label_font_size')
        
        connect(self.ok_button, SIGNAL('clicked()'), self.slot_ok_pressed)
        connect(self.cancel_button, SIGNAL('clicked()'),
                self.slot_cancel_pressed)
        connect(self.button_group, SIGNAL('clicked(int)'),
                self.slot_change_fokus)

        self.spinbox_index.setMinValue(1)

        self._set_default_values()
        # Set label size spinbox to default value.
        self.spinbox_label_size.setValue(int(self.get_parameter('label_size')))
        self.update_data_names_combo_boxes()
        try:
            self._load_settings()
        except Exception, e:
            QMessageBox.information(parent._tab_widget, 'ZMD error',
                                    'There was an error in your ZMD file.' +\
                                    '\nLabel settings set to default.')
        else:
            self.apply_settings()


    def _set_default_values(self):
        label_size = 12

        self.set_parameter('label_size', label_size)

    def _load_settings(self):
        # we load/save settings in scatterplotsettings...
        for key in self._plot_settings.get_parameters():
            value = self._plot_settings.get_parameter(key)
            if key == 'labels':
                self.button_group.setButton(self.button_group.id(\
                    self.labels_dialog.child(value)))
                self.set_parameter(key, value)
            elif key == 'label_index_prefix':
                self.textbox_prefix.setText(value)
                self.set_parameter(key, value)
            elif key == 'label_data_port':
                self.combobox_data.setCurrentItem(gui.get_index(self.combobox_data,
                                                                value))
                self.set_parameter(key, value)
            elif key == 'label_data_dir':
                self.combobox_dir.setCurrentItem(gui.get_index(self.combobox_dir,
                                                               value))
                self.set_parameter(key, value)
            elif key == 'label_data_index':
                self.spinbox_index.setValue(int(value)+1)
                self.set_parameter(key, value)
            elif key == 'label_size':
                self.spinbox_label_size.setValue(int(value))
                self.set_parameter(key, value)
        if not self.get_parameter('labels'):
            self.set_parameter('labels', 'choice_none')

    def update_data_names_combo_boxes(self):
        """Updates the combo box containing the input ports."""
        self.combobox_data.clear()
        for inputport in self._node.get_all_input_ports():
            self.combobox_data.insertItem(inputport.getName())
        try:
            self.combobox_data.setCurrentItem(gui.get_index(
                self.combobox_data, self.get_parameter('label_data_port')))
        except NoSuchParameterError:
            # the value is not set yet...
            pass

    def slot_none(self):
        """If 'None' radio button is selected in LabelSettings dialog this
        method is called whenever the Ok-button is pressed. Removes all labels
        in ScatterPlot.
        """
        if self._label_settings_belong_to == "dendrogram_plot":
            self._plot.set_labels([])
            self._plot.update()
        self.set_parameter('labels', 'choice_none')
        
    def slot_sequence(self, prefix=None):
        """If 'Indexed from 1 Prefix:' radio-button is selected in LabelSettings
        dialog this method is called whenever the Ok-button is pressed. Creates
        labels with the given prefix and indexed from 1.
        """
        label_list = []
        if self._label_settings_belong_to == "dendrogram_plot":
            if self._plot.get_order():
                order_list = []
                for i in self._plot.get_order():
                    order_list.append(i)
                order_list.sort()
                for element in order_list:
                    if prefix:
                        label_list.append(str(prefix) + " " + str(element))
                    else:
                        label_list.append(str(element))
            self._plot.set_labels(label_list)
            self._plot.update()

        self.set_parameter('labels', 'choice_indexed')
        self.set_parameter('label_index_prefix', prefix)
            
    def slot_label_data(self, portname, is_column, index=2):
        """If 'From data' radio button is selected in LabelSettings dialog this
        method is called whenever the Ok-button is pressed. Creates labels with
        the values of a specified coloum.

        Parameters:
        portname  -- name of the port to get data from
        is_column -- True if we're to collect data from column (False if row)
        index=2   -- Index into data
        """
        if is_column:
            colrow = 'column'
        else:
            colrow = 'row'
        if self._label_settings_belong_to == "dendrogram_plot":
            try:
                label_list = self._node.get_data(portname, is_column, index)
            except AttributeError:
                # not run yet
                return
            except IndexError:
                info_text =  'The index you entered was out of bounds.\n'
                info_text += 'Port "%s" has no %s %s.' % (portname, colrow, str(index+1))
                label_list = []
                QMessageBox.information(
                    self.labels_dialog,
                    'Plot Labels - Index out of bounds',
                    info_text)
            except Exception, e:
                # remove labels... if there were any
                label_list = []
            self._plot.set_labels(label_list)
            self._plot.update()
            
        self.set_parameter('labels', 'choice_data')
        self.set_parameter('label_data_port', portname)
        self.set_parameter('label_data_dir', colrow)
        self.set_parameter('label_data_index', index)

    def slot_label_size(self, size):
        # This method must be called BEFORE the labels are redrawn.
        if size != self.get_parameter('label_size'):
            if self._label_settings_belong_to == "dendrogram_plot":
                self._plot.set_label_size(size)
            self.set_parameter('label_size', size)

    def apply_settings(self):
        # Run the slot_label_size before labels are redrawn.
        self.slot_label_size(self.spinbox_label_size.value())
        selected_name = self.button_group.selected().name()
        if selected_name == 'choice_none':
            self.slot_none()
        elif selected_name == 'choice_indexed':
            self.slot_sequence(str(self.textbox_prefix.text()))
        elif selected_name == 'choice_data':
            self.slot_label_data(self.combobox_data.currentText(),
                                 self.combobox_dir.currentText() == 'column',
                                 self.spinbox_index.value()-1)
        
    def slot_cancel_pressed(self):
        """sets settings back to what they were before the dialog was opened"""
        # The line below is needed for MDs not having the 'label_size'
        # setting set to keep the GUI control in sync when 'Cancel' is
        # pressed.
        self.spinbox_label_size.setValue(int(self.get_parameter('label_size')))
        self._load_settings()

    def slot_ok_pressed(self):
        """Check label settings and update GUI"""
        # saves settings in scatterplotsettings
        self.apply_settings()
        for key in self._parameters.keys():
            self._plot_settings.set_parameter(key, self.get_parameter(key))
        self._plot_settings.save_settings()
        self._plot.update()
        
    def slot_change_fokus(self, ignored):
        """Highlights the Ok-button in case one of the radio buttons are
        selected.
        """
        self.ok_button.setDefault(True)
        
    def display_dialog(self):
        """Shows the dialog box."""
        # state is loaded in __init__
        self.update_data_names_combo_boxes()
        self.labels_dialog.show()


class DatabaseInfoDialog(Settings):
    """Dialog for retrieving data from gene annotation database."""

    def __init__(self, parent, node, qwt_plot_widget, widget_id, index):
        Settings.__init__(self)

        self._plot_widget = qwt_plot_widget
        self._node = node
        self._widget_id = widget_id
        self._index = index
        self._plot = None
        self._db_information = {}
        
        # Setup dialog and tabwidget for both lines and markers.
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'plot_database_info.ui'),
                                             None, parent)
                
        self.init_dialog()

    def init_dialog(self):
        """Initiates the database info dialog."""
    
        # Get child widgets
        self._twidget_plot = self._widget.child('tab_plot_setup')
        
        # Buttons
        self._button_ok = self._widget.child('push_button_ok')
        self._button_apply = self._widget.child('push_button_apply')
        self._button_cancel = self._widget.child('push_button_cancel')

        # Combo boxes
        self._cbox_data_source = self._widget.child('cbox_data')
        self._cbox_col_row = self._widget.child('combo_box_col_row')
        
        # Line edits
        self._line_edit_db_name = self._widget.child('line_edit_db_name')
        self._line_edit_db_key = self._widget.child('line_edit_db_key')
        
        # Others
        self._list_view_info = self._widget.child('list_view_info')
        self._text_edit_description = self._widget.child('text_edit_description')
        self._spin_box_index = self._widget.child('spin_box_index')

        # Connect signals
        connect(self._button_ok, SIGNAL('clicked()'), self._slot_ok)
        connect(self._button_apply, SIGNAL('clicked()'), self._slot_apply)
        connect(self._button_cancel, SIGNAL('clicked()'), self._slot_cancel)
        connect(self._list_view_info, SIGNAL('clicked(QListViewItem*)'),
                self._slot_clicked)

        self._update_data_source_input_ports(self._cbox_data_source, '')

    def display_dialog(self, plot):
        """Displays the database info dialog."""
        
        self._plot = plot
        self._load_settings()
        self._widget.show()

    def close(self):
        """Closes the database info dialog."""
        self._widget.close()
       
    def _load_default_settings(self):
        """Loads default settings."""
        self.set_parameter('db_data_source', 'default')
        self.set_parameter('db_col_row', '0')
        self.set_parameter('db_index', '0')
        self.set_parameter('db_URL', 'http://nova2.idi.ntnu.no/webservice/ADBSearch.wsdl')
        self.set_parameter('db_key', 'W27ADqBQFHKv9c5IRNYXTBxUye+vUAp6')

    def _load_settings(self):
        """Loads saved settings and updates the GUI."""
        # retrieve default settings
        self._load_default_settings()
        
        # retrieve saved settings and merge with default settings
        settings = self._plot.get_settings()
        for key in settings: 
            self.set_parameter(key, settings[key])
        
        # set values in GUI elements
        if self._cbox_data_source.count():
            self._cbox_data_source.setCurrentText(self.get_parameter('db_data_source'))
        self._cbox_col_row.setCurrentItem(int(self.get_parameter('db_col_row')))
        self._spin_box_index.setValue(int(self.get_parameter('db_index'))+1)
        self._line_edit_db_name.setText(self.get_parameter('db_URL'))
        self._line_edit_db_key.setText(self.get_parameter('db_key'))
        
    def _save_settings(self):    
        """Saves current settings."""
    
        # retrieve other plot settings and merge with database info settings
        settings = self._plot.get_settings()
        for key in settings: 
            self.set_parameter(key, settings[key])

        db_data_source = self._cbox_data_source.currentText()
        db_col_row = str(int(str(self._cbox_col_row.currentText()) == 'row'))
        db_index = self._spin_box_index.value() - 1
        db_URL = str(self._line_edit_db_name.text())
        db_key = str(self._line_edit_db_key.text())
    
        self.set_parameter('db_data_source', db_data_source)
        self.set_parameter('db_col_row', db_col_row)
        self.set_parameter('db_index', str(db_index))
        self.set_parameter('db_URL', db_URL)
        self.set_parameter('db_key', db_key)
        
        self._plot.set_settings(self._parameters)

    def _set_max_value(self, spinbox, default_value, is_column, cbox_inputdata):
        port_name = str(cbox_inputdata.currentText())
        max_size = sys.maxint

        try:
            max_size = self._node.get_data_size(port_name, is_column)
        except AttributeError:
            pass
        
        spinbox.setValue(default_value+1)
        spinbox.setMaxValue(max_size)

    def _update_data_source_input_ports(self, cbox, port_param_name):
        """Updates the combo box containing the input ports."""
        cbox.clear()
        for inputport in self._node.get_all_input_ports():
            cbox.insertItem(inputport.getName())

    def _get_database_info(self):
        """Retrieves info from database based on selected indexes in plot."""
        try:
            from System.dbconnect import DbConnect
            dbconnect_import_success = True
                 
        except:
            dbconnect_import_success = False
            QMessageBox.information(self._widget, 'Import error',
                                    'Error importing SOAPpy.'+
                                    'SOAPpy is required for this feature')
        
        if dbconnect_import_success:
            try:
                db_key = self.get_parameter('db_key')
                db_URL = self.get_parameter('db_URL')
                db = DbConnect('test', db_URL, db_key)
            except:
                QMessageBox.information(self._widget, 'Database error',
                                        'Error connecting to Gene Annotation Database.')
        
            marker_indexes = self._plot.get_highlighted_indices_by_aspect(self._plot.get_mappable_aspects()[0])
            port = self._node.getInputPortByName(self._cbox_data_source.currentText())
            
            data_index = self._spin_box_index.value() - 1
            is_column = self._cbox_col_row.currentText() == 'column'

            gene_id = []
            try:
                data = self._node.get_data(self._cbox_data_source.currentText(),
                                           is_column, data_index)
                for index in marker_indexes:
                    gene_id.append(data[index])
                if not gene_id:
                    QMessageBox.information(self._widget, 'Marker warning',
                                            'No points in plot marked')
                
            except IndexError:
                QMessageBox.information(self._widget, 'Index error',
                                        'Error while mapping selected markers to port data.')
            
            db_result = []
            self._db_information = {}
            progress_dialog = QProgressDialog('Retrieving information from database',
                                              'Cancel', len(gene_id),
                                              self._widget, 'Progress', True)
            step = 0
            for id in gene_id:
                #db_result.append(db.get_gene_info(id, 'accno'))
                if re.match('[0-9]', str(id)): #if clone id
                    self._db_information[id] = db.get_gene_info(id, 'clone')
                else:
                    self._db_information[id] = db.get_gene_info(id, 'accno')
                #self._db_information[id] = db.get_gene_info(id, 'accno')
                step += 1
                progress_dialog.setProgress(step)
                qApp.processEvents()
                if progress_dialog.wasCanceled():
                    break
            progress_dialog.setProgress(step)

            counter = 0
            self._list_view_info.clear()

            for gene_id, data in self._db_information.iteritems():
                window_item = QListViewItem(self._list_view_info,
                                            str(gene_id), str(data[0]))
                self._list_view_info.insertItem(window_item)

    # Slots
    def _slot_ok(self):
        self._save_settings()
        self.close() 

    def _slot_apply(self):
        self._save_settings()    
        self._get_database_info()

    def _slot_cancel(self):
        self.close()

    def _slot_clicked(self, item):
        if not item == 0: #if clicked on a QListViewItem
            #the first column of each item contains the gene_id
            gene_info = self._db_information.get(str(item.text(0)))
            if not gene_info == None:
                gene_description = gene_info[1]
            else:
                gene_description = 'No description for this item'                
            self._text_edit_description.setText(gene_description)
    
