# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os
from itertools import izip

from qtui import QWidgetFactory
from qt import QObject, SIGNAL, QComboBox, QLabel, QLineEdit
from qt import QGridLayout, QMessageBox, QSpacerItem

from GUI.guinode import GUINode

from Library.Nodes.Input.UserQueryNode import IllegalValueError
from Library.Util.Config import Config
from System.Data.Data import Data

SETUP_ID = 0 # id's used when raising the widgets inside the widget stack
USE_ID = 1

class GUIUserQueryNode(GUINode):

    def __init__(self, parent, node):

        GUINode.__init__(self, parent, node)
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'user_input_node.ui'),
                                             None, parent)
        # Dont need no stinking statusbar 
        self._widget.statusBar().hide()
        
        self._fetch_widget_children()
        
        self._setup_group_layout = QGridLayout(self._setup_group, 2, 4, 20, 2)

        self._setup_group_layout.addMultiCellWidget(self._setup_number_frame,
                                                    0, 0, 0, 3)

        self._use_group_layout = QGridLayout(self._use_group, 2, 3, 20, 2)
        self._widget.adjustSize()

        self._setup_labels = {}
        self._setup_spacers = {}
        self._setup_labeledits = {}
        self._setup_combo_boxes = {}
        self._use_header_labels = {}
        self._use_row_labels = {}
        self._use_item_edits = {}
        
        self._node = node
        
        self._widget_stack.raiseWidget(SETUP_ID)
        self._use_clear_action.setDisabled(True)
        if self._node.get_number_of_values() != 0:
            self._update_setup_with_saved_values()
        else:
            self._insert_new_combo_box(1)
            
        QObject.connect(self._setup_mode_action, SIGNAL('activated()'),
                        self._slot_setup_mode_activated)
        QObject.connect(self._use_mode_action, SIGNAL('activated()'),
                        self._slot_use_mode_activated)
        QObject.connect(self._use_clear_action, SIGNAL('activated()'),
                        self._slot_clear_use_cells)
        QObject.connect(self._number_spin, SIGNAL('valueChanged(int)'),
                        self._slot_setup_number_changed)
        QObject.connect(self._ok_button, SIGNAL('clicked()'),
                        self._slot_ok_button_clicked)
        QObject.connect(self._apply_button, SIGNAL('clicked()'),
                        self._apply_settings)
        QObject.connect(self._widget_stack, SIGNAL('aboutToShow(int)'),
                        self._slot_stack_changing)
        QObject.connect(self._cancel_button, SIGNAL('clicked()'),
                        self._widget.close)

    def _slot_setup_mode_activated(self):
        self._widget_stack.raiseWidget(SETUP_ID)

    def _slot_use_mode_activated(self):
        self._widget_stack.raiseWidget(USE_ID)

    def _slot_setup_number_changed(self, number):
        if number > len(self._setup_combo_boxes):
            for x in range(len(self._setup_combo_boxes)+1, number+1):
                self._insert_new_combo_box(x)
        else:
            self._remove_combo_boxes_and_labels(number)
            self._remove_use_line_edits(number)
        self._widget.adjustSize()
            
    def _slot_ok_button_clicked(self):
        self._apply_settings()
        self._fetch_values()
        self._widget.close()
        
    def _slot_stack_changing(self, id_):
        if id_ == USE_ID:
            self._update_use_widget()
            self._use_clear_action.setDisabled(False)
        else:
            self._use_clear_action.setDisabled(True)


    def _fetch_widget_children(self):
        self._widget_stack = self._widget.child('widget_stack')
        self._setup_widget = self._widget.child('setup_widget')
        self._setup_number_frame = self._widget.child('setup_number_frame')
        self._setup_group = self._widget.child('setup_group')
        self._use_widget = self._widget.child('use_widget')
        self._use_group = self._widget.child('use_group')
        self._ok_button = self._widget.child('ok_button')
        self._apply_button = self._widget.child('apply_button')
        self._cancel_button = self._widget.child('cancel_button')
        self._number_spin = self._widget.child('setup_number_spin')
        self._setup_mode_action = self._widget.child('setup_mode_action')
        self._use_mode_action = self._widget.child('use_mode_action')
        self._use_clear_action = self._widget.child('clear_use_action')
        self._setup_number_label = self._widget.child('setup_number_label')

    def _update_use_widget(self):

        input_ports = self._node.get_all_input_ports()
        num_inputs = self._node.get_number_of_values()
        port_data = map(lambda port: port.getData(), input_ports)
        
        for data in port_data:
            # Checks if the node has any data or not
            if not isinstance(data, Data):
                return

        # We remove old use widgets to make it a bit more easy for us self.
        self._remove_all_use_widgets()

        self._add_use_labels(num_inputs)
        self._add_use_cells(port_data, num_inputs)
        if self._node.get_values():
            self._insert_saved_use_values()

    def _add_use_labels(self, number):
        if self._use_header_labels:
            for label in self._use_header_labels.values():
                label.hide()                
                
        for i, text in enumerate(self._node.get_value_labels()):
            label = QLabel(text, self._use_group, 'label_%s' % i)
            self._use_group_layout.addWidget(label, 0, i+1)
            label.show()
            label.adjustSize()
            self._use_header_labels[i] = label
        
    def _add_use_cells(self, ports, number):
        index = 1
        for dataset in map(lambda d: d.getData(), ports):
            for value in dataset:
                label = QLabel(str(value), self._use_group)
                self._use_group_layout.addWidget(label, index, 0)
                label.show()
                self._use_row_labels[index] = label
                items = []
                for x in range(number):
                    edit = QLineEdit(self._use_group)
                    items.append(edit)
                    self._use_group_layout.addWidget(edit, index, x+1)
                    edit.show()
                self._use_item_edits[index] = items

                index += 1
            
    def _apply_settings(self):
        spin_value = int(self._number_spin.value())
            
        self._node.set_number_of_values(spin_value)
        types = []
        for combo in self._setup_combo_boxes.values():
            if str(combo.currentText()) ==  'Text':
                types.append('Text')
            elif str(combo.currentText()) == 'Integer':
                types.append('Integer')
            elif str(combo.currentText()) == 'Float':
                types.append('Float')
        labels = []
        for edit in self._setup_labeledits.values():
            labels.append(str(edit.text()))

        self._node.set_value_labels(labels)
        self._node.set_value_types(types)

    def _insert_new_combo_box(self, id_, current_item = 0):
        label = QLabel('Item: %s' % id_, self._setup_group)
        label.show()

        spacer = QSpacerItem(30, 5)        

        labeledit = QLineEdit(self._setup_group, 'setup_label_%s' % id_)
        labeledit.show()

        combo = QComboBox(self._setup_group, 'setup_combo_%s' % id_)
        combo.insertItem('Text', 0)
        combo.insertItem('Integer', 1)
        combo.insertItem('Float', 2)
        combo.setCurrentItem(current_item)
        combo.show()

        self._setup_group_layout.addWidget(label, id_+1, 0)
        self._setup_group_layout.addItem(spacer, id_+1, 1)
        self._setup_group_layout.addWidget(labeledit, id_+1,2)
        self._setup_group_layout.addWidget(combo, id_+1, 3)
        

        self._setup_labels[id_] = label
        self._setup_spacers[id_] = spacer
        self._setup_labeledits[id_] = labeledit
        self._setup_combo_boxes[id_] = combo
        
    def _remove_combo_boxes_and_labels(self, number):
        if number < len(self._setup_combo_boxes):
            key_ids = self._setup_combo_boxes.keys()

            for key in key_ids[number:]: 

                label = self._setup_labels.pop(key)
                spacer = self._setup_spacers.pop(key)
                labeledit = self._setup_labeledits.pop(key)
                combo = self._setup_combo_boxes.pop(key)
                             
                label.hide()
                labeledit.hide()
                combo.hide()
                self._setup_group.removeChild(label)
                self._setup_group_layout.removeItem(spacer)
                self._setup_group.removeChild(labeledit)
                self._setup_group.removeChild(combo)

                
    def nodeDialog(self):
        """ Method that will show the GUI"""
        for port in self._node.get_all_input_ports():
            if port.getData() and self._node.get_number_of_values() > 0:
                self._widget_stack.raiseWidget(USE_ID)
                self._use_clear_action.setDisabled(False)
                self._update_use_widget()
                break
            
        self._widget.show()

    def _fetch_values(self):
        if not self._widget_stack.visibleWidget() == self._use_widget:
            return
        
        self._node._remove_old_values()
        for lineedit in self._use_item_edits.values():
            values = map(lambda edit: str(edit.text()), lineedit)
            try:
                self._node.add_values(values)
            except IllegalValueError, e:
                QMessageBox.warning(self._widget, 'Incorrect value types',
                                  'Some of the values entered are incorrect:\n' + str(e))
                return
                        
    def _update_setup_with_saved_values(self):
        number = self._node.get_number_of_values()
        types = map(lambda type_ : type_[1], self._node.get_input_types())
        self._number_spin.setValue(number)
                            
        for id_, type_ in izip(range(number), types):
            id_ += 1
            if type_ == 'Text':
                current = 0
            elif type_ == 'Integer':
                current = 1
            elif type_ == 'Float':
                current = 2
            self._insert_new_combo_box(id_, current)
        
        keys = self._setup_labeledits.keys()
        keys.sort()

        for label, key in izip(self._node.get_value_labels(),
                               keys):
            self._setup_labeledits[key].setText(str(label))
            
    def _insert_saved_use_values(self):
        values = self._node.get_values()
        
        keys = self._use_item_edits.keys()
        keys.sort()
        
        if self._node.get_number_of_values() == 1:
            for value, key in izip(values, keys):
                line = self._use_item_edits[key][0]
                line.setText(str(value))
            return

        for value_list, key in izip(values, keys):
            edits = self._use_item_edits[key]
            
            if len(value_list) > 1 and not isinstance(value_list, str):
                for i, value in enumerate(value_list):
                    edits[i].setText(str(value))
            else:
                edits[0].setText(value_list)
            
            
    def _remove_use_line_edits(self, number):
        """Removes edit lines that are redundant."""
        if not self._use_item_edits.values():
            return
        len_ = len(self._use_item_edits.values()[0])
        diff = len_ - number
        for edits in self._use_item_edits.values():
            removable_edits = edits[-diff:]
            for edit_line in removable_edits:
                edit_line.hide()
                self._use_group.removeChild(edit_line)
        keys = self._use_header_labels.keys()
        keys.sort()
        removable_keys = keys[-diff:]
        
        for key in removable_keys:
            label = self._use_header_labels.pop(key)
            label.hide()
            self._use_group.removeChild(label)
            
        
    def _slot_clear_use_cells(self):
        for edits in self._use_item_edits.values():
            for edit_line in edits:
                edit_line.setText('')

    def _remove_all_use_widgets(self):
        for label in self._use_row_labels.values():
            label.hide()
            self._use_group.removeChild(label)
        for edits in self._use_item_edits.values():
            for edit_line in edits:
                edit_line.hide()
                self._use_group.removeChild(edit_line)

        self._use_row_labels = {}
        self._use_item_edits = {}
    
