# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os

from itertools import izip
from qt import QListViewItem, QObject, QFileDialog, SIGNAL, QMessageBox, SLOT, QInputDialog
from qt import PYSIGNAL
from qtui import QWidgetFactory

from GUI.guinode import GUINode
from GUI.NodeConfig import NodeConfig
from Library.Util.Config import Config
from Library.Util.gui import get_index
from Library.Util import parsing

connect = QObject.connect

class InputGUINode(GUINode):
    def __init__(self, parent, node):
        GUINode.__init__(self, parent, node)
        self._parent = parent
        self.change_icon()
        self._dialog = None
    
    def nodeDialog(self):
        if not self._dialog:
            self._dialog = InputNodeDialog(self._parent, self)
            self._dialog.show()
        elif not self._dialog.is_shown():
            self._dialog = InputNodeDialog(self._parent, self)
            self._dialog.show()

    def change_icon(self):
        nodename = self.getDiagramNode().get_file_reader_class_name()
        nodeConfig = NodeConfig(nodename)
        
        if nodeConfig.getImage().find("Unknown") > 0:
            nodeConfig = NodeConfig("InputNode")
        
        self.getCanvasNode().setImage(nodeConfig.getImage())


class InputNodeDialog:
    def __init__(self, parent, gui_node, silent_mode=False):
        """Constructs the input node dialog.

        If silent_mode is set, no GUI will flash the screen (intended
        for tests to run smoother and faster).
        """
        self._gui_node = gui_node
        self._node = self._gui_node.getDiagramNode()
        self._silent_mode = silent_mode
        
        self._dialog = QWidgetFactory.create(os.path.join(Config.uidir, "inputnode.ui"), None, parent)
        self._filepath = self._dialog.child("filepath_lineedit")
        self._format_combo = self._dialog.child("fileformat_combo")
        self._output_listview = self._dialog.child("output_listview")
        self._format_options = self._dialog.child("formatoptions_button")

        for format in self._node.get_known_formats():
            self._format_combo.insertItem(format)

        # to be able to cancel actions performed, we keep original values
        self._original_filename = self._node.get_filename()
        self._original_fileformat = self._node.get_fileformat()
        
        self._set_file(self._original_filename)
        
        connect(self._dialog.child("browse_button"), SIGNAL("clicked()"),
                self._slot_browse_and_set_file)
        connect(self._filepath, SIGNAL("lostFocus()"), self._slot_set_file)
        connect(self._filepath, SIGNAL("returnPressed()"), self._slot_set_file)
        connect(self._format_combo, SIGNAL("activated(const QString &)"),
                self._slot_set_file_format)
        connect(self._dialog.child("ok_button"), SIGNAL("clicked()"),
                self._slot_ok_clicked)
        connect(self._dialog.child("cancel_button"), SIGNAL("clicked()"),
                self._slot_cancel_clicked)
        connect(self._format_options, SIGNAL("clicked()"),
                self._slot_configure_separator)

    def _set_file(self, path):
        """Sets path as the file of the filereader node."""
        self._filepath.setText(path)
        if not self._node.get_filename() == path:
            self._node.set_filename(str(path))

        # if new format is unknown, use previous one

        if self._node.get_fileformat()=="Unknown format":
            self. _slot_set_file_format(self._format_combo.currentText())

        self._format_combo.setCurrentItem(get_index(self._format_combo,
                                                    self._node.get_fileformat()))
        self._format_options.setEnabled(self._format_combo.currentText() == "Textfile")
        self._reload_output_listview()

    def _slot_browse_and_set_file(self):
        """Interactively find a file and set that file"""
        filename = QFileDialog.getOpenFileName(Config.last_path_dataset, self._node.get_fileformat_filters(), self._dialog)
        if not filename:
            return
        filename = str(filename)

        self._set_file(filename)

        Config.last_path_dataset = os.path.dirname(filename)
        

        if not self._node.get_available_output() and not self._silent_mode:
            QMessageBox.warning(self._dialog, "No data in file...",
                                "Could not find any data in file %s\n" % repr(str(filename)) +\
                                "Does the file exist and is the correct format chosen?")
            
    def _slot_set_file(self):
        """Set file from text in self._filepath"""
        self._set_file(str(self._filepath.text()))

    def _slot_set_file_format(self, formatstring):
        """Set format on node and reloads outputs."""
        self._node.set_fileformat(formatstring)
        self._node.set_filename(str(self._filepath.text()),
                                force_format=formatstring)
        self._reload_output_listview()
        self._format_options.setEnabled(formatstring == "Textfile")

    def _reload_output_listview(self):
        """Get available output from node and set in listview"""
        self._output_listview.clear()
        for data in self._node.get_available_output():
            item = QListViewItem(self._output_listview, data.getName(),
                                 data.get_data_size_str())
            self._output_listview.insertItem(item)

    def _slot_ok_clicked(self):
        """Apply actions performed on node (i.e., create output ports)"""
        self._node.set_available_output_as_ports()
        self._gui_node.change_icon()
        
    def _slot_cancel_clicked(self):
        """Reset actions performed on node"""
        self._node.set_filename(self._original_filename, force_format=self._original_fileformat)
        
    def show(self):
        if not self._silent_mode:
            self._dialog.show()
            filename = str(self._filepath.text())
            if not self._node.get_available_output() and filename:
                QMessageBox.warning(self._dialog, "No data in file...",
                                    "Could not find any data in file %s\n" % repr(filename) +\
                                    "Does the file exist and is the correct format chosen?")

    def is_shown(self):
        return self._dialog.isShown()

    def _slot_configure_separator(self):
        """Configures node's Separator instance."""
        # keep a reference so that connections work
        self._sepgui = SeparatorGUI(self._node.get_textfilesettings(), self._dialog)
        QObject.connect(self._sepgui.get_widget(), PYSIGNAL('separator_changed()'), self._reload_output_listview)


class SeparatorGUI:

    def __init__(self, textfilesettings, parent):
        self._settings = textfilesettings
        factory = QWidgetFactory()
        self._gui = factory.create(os.path.join(Config.uidir, "sep.ui"), None, parent, None)     
        # variables
        # Last selected box
        self._cur_selected = 'sep'        

        # buttons
        _button_move_left = self._gui.child('pushButton10')
        _button_move_right = self._gui.child('pushButton10_3')
        _button_move_center = self._gui.child('pushButton6')
        _button_add_sep = self._gui.child('pushButton10_3_2')
        _ok = self._gui.child('ok')
        _cancel = self._gui.child('cancel')

        # checkboxes, spinbox
        self._checkbox_disregard = self._gui.child('checkBox18')
        self._spinbox_disregard = self._gui.child('spinBox1')
        self._checkbox_automatic = self._gui.child('checkbox20')
        self._checkbox_rowheaders= self._gui.child('checkBox19')
        self._checkbox_colheaders= self._gui.child('checkBox20')

        # listboxes
        self._list_rows = self._gui.child('listBox1')
        self._list_seps = self._gui.child('listBox1_2_2')
        self._list_cols = self._gui.child('listBox1_2')

        # setting up signals needed
        self._gui.connect(_button_move_left,SIGNAL('clicked()'), self._move_left)
        self._gui.connect(_button_move_right,SIGNAL('clicked()'), self._move_right)
        self._gui.connect(_button_move_center,SIGNAL('clicked()'),self._move_center)

        self._gui.connect(_button_add_sep,SIGNAL('clicked()'), self._add_sep_clicked)
        self._gui.connect(self._list_rows,SIGNAL('clicked( QListBoxItem * )'),self._rowlist_clicked)
        self._gui.connect(self._list_cols,SIGNAL('clicked( QListBoxItem * )'),self._collist_clicked)
        self._gui.connect(self._list_seps,SIGNAL('clicked( QListBoxItem * )'),self._seplist_clicked)
        self._gui.connect(self._list_rows,SIGNAL('selectionChanged ( QListBoxItem * )'),self._selection_changed)
        self._gui.connect(self._list_cols,SIGNAL('selectionChanged ( QListBoxItem * )'),self._selection_changed)
        self._gui.connect(self._list_seps,SIGNAL('selectionChanged ( QListBoxItem * )'),self._selection_changed)
        self._gui.connect(_ok, SIGNAL('clicked()'), self._ok_clicked)
        self._gui.connect(_cancel, SIGNAL('clicked()'), self._cancel_clicked)

        # setting up SLOTs
        # SIGNALS for OK / CANCEL to SLOT ACCEPT/REJECT are set up in UI-file!
        self._gui.connect(self._checkbox_disregard, SIGNAL('toggled(bool)'), self._spinbox_disregard, SLOT('setEnabled(bool)'))
        self._gui.connect(self._checkbox_automatic, SIGNAL('toggled(bool)'), self._gui.child('groupBox4'), SLOT('setDisabled(bool)'))
        self._gui.connect(self._checkbox_automatic, SIGNAL('toggled(bool)'), self._gui.child('groupBox3'), SLOT('setDisabled(bool)'))

        # setting up 
        self._iniate_settings()
        self._gui.show()


    # returns this class' widget
    def get_widget(self):
        return self._gui

    def _iniate_settings(self):
        self._spinbox_disregard.setDisabled(True)        

        settings = self._settings.getSettings()

        to_add = ['Newline','Space','Tab',';',':',',','.','/']
        cols = settings['col_separators']
        rows = settings['row_separators']

        # remove all occurences in rows and cols from to_add (if present)
        [to_add.remove(el) for el in cols if el in to_add]
        [to_add.remove(el) for el in rows if el in to_add]
    
        # adding options
        self._add_items(rows,self._list_rows)
        self._add_items(cols,self._list_cols)
        self._add_items(to_add,self._list_seps)

        # update various settings
        self._checkbox_automatic.setChecked(settings['automatic'])
        self._checkbox_disregard.setChecked(bool(settings['lines_to_remove']))
        self._spinbox_disregard.setValue(int(settings['lines_to_remove']))
        self._checkbox_rowheaders.setChecked(settings['row_headers'])
        self._checkbox_colheaders.setChecked(settings['col_headers'])
        
    # convience function for adding many items to a QListBox
    def _add_items(self,itemlist,qlistbox):
        for item in itemlist:
            qlistbox.insertItem(item)
            
    def _cancel_clicked(self):
        pass

    def _ok_clicked(self):
        # update settings
        at = self._checkbox_automatic.isChecked()
        row_sep = []
        col_sep = []
        for x in xrange(self._list_rows.count()):
            row_sep.append(str(self._list_rows.text(x)))
        for x in xrange(self._list_cols.count()):
            col_sep.append(str(self._list_cols.text(x)))

        col_header = self._checkbox_colheaders.isChecked()
        row_header = self._checkbox_rowheaders.isChecked()

        if not self._checkbox_disregard.isChecked():
            lines = 0
        else:
            lines = parsing.toInt(self._spinbox_disregard.value(),default=0)

        self._settings.set(at, row_sep, col_sep, row_header, col_header,lines)

        self._gui.emit(PYSIGNAL('separator_changed()'), (None, ))


    # moves separatoritems to QListBox selected to the left
    def _move_left(self):
        if self._cur_selected == 'sep' and self._list_seps.currentText():
            self._move(self._list_seps, self._list_rows)
        elif self._cur_selected == 'col' and self._list_cols.currentText():
            self._move(self._list_cols, self._list_rows)

    # mvoes separatoritems to QListBox selected to the right
    def _move_right(self):
        if self._cur_selected == 'sep' and self._list_seps.currentText():
            self._move(self._list_seps, self._list_cols)
        elif self._cur_selected == 'row' and self._list_rows.currentText():
            self._move(self._list_rows, self._list_cols)

    def _move_center(self):
        if self._cur_selected == 'row' and self._list_rows.currentText():
            self._move(self._list_rows, self._list_seps)
        elif self._cur_selected == 'col' and self._list_cols.currentText():
            self._move(self._list_cols, self._list_seps)

    # convience function for moving between two supplied QListBoxes
    def _move(self,from_,to_):
        index = from_.currentItem()
        to_.insertItem(from_.text(index),0)
        to_.setCurrentItem(-1)
        from_.removeItem(index)
        from_.setCurrentItem(index)
        self._set_selected(index,from_)


    # set selected index to selected and clears selection on other items
    def _set_selected(self,index,item):
        if item != self._list_seps:
            self._list_seps.clearSelection()
        if item != self._list_rows:
            self._list_rows.clearSelection()
        if item != self._list_cols:
            self._list_cols.clearSelection()
        if not item.isSelected(index):
            item.setSelected(index,True) 

    # called from all ListBoxes when selection changes, calls _set_selected to clear other boxes
    def _selection_changed(self,listboxitem):
        listview = listboxitem.listBox()
        self._set_selected(listview.index(listboxitem),listview)


    # Method for keeping tracking of last active QListBox
    def _rowlist_clicked(self,index):
        if index:
            self._cur_selected = 'row'
    def _seplist_clicked(self,index):
        if index:
            self._cur_selected = 'sep'
    def _collist_clicked(self,index):
        if index:
            self._cur_selected = 'col'      


    def _add_sep_clicked(self):
        text, ok = QInputDialog.getText(
            "New Separator", "Please enter wanted char / string:\n(Please note, strings 'Newline','Return','Tab'\n and 'Space' will be converted into the actual chars)")
        if ok and text:
            self._list_seps.insertItem(text)
            
