# -*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.guinode import GUINode
from qtui import QWidgetFactory
import os
from Library.Util.Config import Config
from GUI.NodeConfig import NodeConfig
from qt import QFileDialog, QInputDialog, QLineEdit, QListViewItem, QObject
from qt import SIGNAL


connect = QObject.connect


class FileWriterGUINode(GUINode):
    """MDView-represenation of OutputNode."""
    def __init__(self, parent, node):
        GUINode.__init__(self, parent, node)
        self._writers = []
        self._data_names = {}
        # keep connection between list view items and data sources
        self._item_ports = {}
        self._update_icon_to_writer()

        self._parent = parent

        # widgets
        self._filename = None
        self._file_type_list = None
        self._storage_list = None
        self._widget = None

    def _query_rename(self, dummy=None):
        """Show dialog to user to rename a data source entry."""
        item = self._storage_list.selectedItem()
        if not item:
            return
        name = self._data_names[item]
        output_port = self._item_ports[item]
        
        text, changed = QInputDialog.getText("Renaming %s"%name ,
                                             "Enter new name for data:",
                                             QLineEdit.Normal, name,
                                             self._widget)
        if changed:
            self._data_names[item] = str(text)
            item.setText(0, text)

    def _update_widgets(self):
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          "outputnode.ui"),
                                             None, self._parent)        

        self._filename = self._widget.child('filename')
        self._file_type_list = self._widget.child('filetype_list')
        self._storage_list = self._widget.child('storage_list')
        browse_button = self._widget.child('browse_button')
        cancel_button = self._widget.child('cancel_button')
        ok_button = self._widget.child('ok_button')
        rename_button = self._widget.child('rename_button')

        connect(browse_button, SIGNAL('clicked()'), self._browse)
        connect(cancel_button, SIGNAL('clicked()'), self._widget.close)
        connect(ok_button, SIGNAL('clicked()'), self._ok)
        connect(rename_button, SIGNAL('clicked()'), self._query_rename)
        connect(self._storage_list, SIGNAL('doubleClicked(QListViewItem*)'),
                self._query_rename)
        connect(self._storage_list, SIGNAL('spacePressed(QListViewItem*)'),
                self._query_rename)
        connect(self._filename, SIGNAL("textChanged(const QString &)"),
                self._filename_changed)

        for writer in self.node.get_file_writers():
            self._file_type_list.insertItem(writer.get_description())
            # keep score of which writer is which in list
            self._writers.append(writer.get_filetype())

        # fill with info from node
        self._filename.setText(self.node.get_filename())

        # select correct value in combobox
        try:
            index = self._writers.index(self.node.get_file_writer())
        except ValueError:
            index = 0
        self._file_type_list.setCurrentItem(index)

        # fill data source info from connected ports
        self._storage_list.clear()
        self._data_names = {}
        self._item_ports = {}
        for name, port in self.node.get_storage_names():
            item = QListViewItem(self._storage_list, name, port.getName())
            self._item_ports[item] = port.getDataSource()
            self._data_names[item] = name
        
    def nodeDialog(self):
        self._update_widgets()
        self._widget.exec_loop()
        
    def _browse(self):
        output_path = QFileDialog.getSaveFileName("", self.node.get_filters(),
                                                  self._widget,
                                                  "file_writer_browser",
                                                  "Choose a file to save...")
        if output_path:
            self._filename.setText(output_path)
                                                  
    def _rename_source(self, list_view_item, new_name):
        self._data_names[list_view_item] = new_name
        list_view_item.setText(0, new_name)

    def _ok(self):
        self.node.set_filename(str(self._filename.text()))
        self.node.set_file_writer(self._writers[self._file_type_list.currentItem()])
        for item, name in self._data_names.items():
            port = self._item_ports[item]
            self.node.set_stored_name(port, name)
        self._widget.close()
        self._update_icon_to_writer()

    def _filename_changed(self, new_name):
        """Allow extension of filename to adjust writer used."""
        path, extension = os.path.splitext(str(new_name))
        # remove . from extension
        extension = extension[1:]
        extension_table = self.node.get_writers_by_extension()
        if extension in extension_table:
            writer = extension_table[extension]
            index = self._writers.index(writer.get_filetype())
            self._file_type_list.setCurrentItem(index)

    def _update_icon_to_writer(self):
        nodeConfig = NodeConfig(self.node.get_file_writer_class())
        
        if nodeConfig.getImage().find("Unknown") > 0:
            nodeConfig = NodeConfig("OutputNode")
        
        self.getCanvasNode().setImage(nodeConfig.getImage()) 
