# -*- coding: utf-8 -*-

# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QApplication, QGridLayout, QGroupBox, QHBoxLayout, QListBox
from qt import QMessageBox, QPushButton, QSize, QSizePolicy, QSpacerItem
from qt import QToolTip, QVBoxLayout, QWidget, Qt, SIGNAL, SLOT, qApp

class AliasWidget(QWidget):
    """A widget containing gui and logic for setting alias names on input ports."""
    def __init__(self, setupWidget, name = None, fl = 0):
        """
        Params:
               - setupWidget : PlotWindowSetupWidget - The setup widget where the
               alias button resides.
        Consturctor. Initialises the widget"""
        
        QWidget.__init__(self, setupWidget, name, fl)

        if not name:
            self.setName("AliasWidget")

        self.setupWidget = setupWidget
        self.windowController = self.setupWidget.windowController
        
        self.setCaption(self.__tr("Set aliases"))
        # main Layout
        AliasWidgetLayout = QGridLayout(self,1,1,11,6,"AliasWidgetLayout")

        # Ports layout for list and label
        
        self.portsGroup = QGroupBox(self.__tr("Input ports"), self)
        self.portsGroup.setColumnLayout(0, Qt.Vertical)
        portsLayout = QVBoxLayout(self.portsGroup.layout(), 0)
        
        self.portList = QListBox(self.portsGroup, "portList")
        self.portList.clear()
        self.connect(self.portList, SIGNAL("highlighted(const QString &)"), self.slotShowAliases)
        QToolTip.add(self.portList,self.__tr("Available input ports."))
        
        portsLayout.addWidget(self.portList)
        AliasWidgetLayout.addWidget(self.portsGroup,0,0)

        # Alias layout for alias list and label
        
        self.portAliasGroup = QGroupBox(self.__tr("Port aliases"), self)
        self.portAliasGroup.setColumnLayout(0, Qt.Vertical)
        portAliasLayout = QVBoxLayout(self.portAliasGroup.layout(), 0)

        self.portAliasList = QListBox(self.portAliasGroup,"portAliasList")
        self.portAliasList.clear()
        QToolTip.add(self.portAliasList,self.__tr("Assigned aliases to selected port."))

        portAliasLayout.addWidget(self.portAliasList)
        AliasWidgetLayout.addWidget(self.portAliasGroup, 0, 1)

        # Available layout for avaialable list and label
        self.availableGroup = QGroupBox(self.__tr("Available aliases"), self)
        self.availableGroup.setColumnLayout(0, Qt.Vertical)
        availableLayout = QVBoxLayout(self.availableGroup.layout(), 0)

        self.availableList = QListBox(self.availableGroup, "availableList")
        self.availableList.clear()
        QToolTip.add(self.availableList,self.__tr("Available aliases that can be assigned."))
        

        availableLayout.addWidget(self.availableList)
        AliasWidgetLayout.addWidget(self.availableGroup, 0, 3)

        # Layout for add/remove buttons
        addLayout = QVBoxLayout(None,0,6,"addLayout")
        spacer = QSpacerItem(20,20,QSizePolicy.Minimum,QSizePolicy.MinimumExpanding)
        addLayout.addItem(spacer)
        
        # Add button
        self.addButton = QPushButton(self,"addButton")
        self.addButton.setText(self.__tr("<-- Add Alias"))
        self.connect(self.addButton, SIGNAL("clicked()"), self.slotAddAlias)
        # Remove button
        self.removeButton = QPushButton(self,"removeButton")
        self.removeButton.setText(self.__tr("Remove Alias -->"))
        self.connect(self.removeButton, SIGNAL("clicked()"), self.slotRemoveAlias)
        
        spacer_2 = QSpacerItem(20,10,QSizePolicy.Minimum,QSizePolicy.MinimumExpanding)

        addLayout.addWidget(self.addButton)
        addLayout.addWidget(self.removeButton)
        addLayout.addItem(spacer_2)
        AliasWidgetLayout.addLayout(addLayout,0,2)

        # Button layout for ok button etc.
        buttonLayout = QHBoxLayout(None,0,6,"buttonLayout")
        spacer_3 = QSpacerItem(180,20,QSizePolicy.Expanding,QSizePolicy.Minimum)
        buttonLayout.addItem(spacer_3)

        self.doneButton = QPushButton(self,"doneButton")
        self.doneButton.setText(self.__tr("Done"))
        self.connect(self.doneButton, SIGNAL("clicked()"), self, SLOT("close()"))
        buttonLayout.addWidget(self.doneButton)

        AliasWidgetLayout.addMultiCellLayout(buttonLayout,1,1,0,3)

        
        self.resize(QSize(459,162).expandedTo(self.minimumSizeHint()))

        self.ports = {}
        # self.show()

                    
    def initAliasList(self):
        """
        Initialises the list of availbale aliases. If no alias is available, the method will
        disable the buttons for adding and removing aliases, and display an informational
        dialog box."""
        available = self.windowController.getAvailableAliases()

        if available:
            for port in available:
                self.availableList.insertItem(str(port))
            return 1
        
        else:
            # got no aliases, so setting them is disabled
            self.addButton.setEnabled(0)
            self.removeButton.setEnabled(0)
                                    
            retval = QMessageBox.information(self,
                                             "SciCraft",
                                             "The add and remove buttons have been disabled\n" +
                                             "as this plot node does not have any alias data.\n" +
                                             "To set aliases, make sure you have an input port\n" +
                                             "called \"alias\" and an edit node connected to\nit. " +
                                             "Then rerun the module diagram.",
                                             QMessageBox.Ok)
            
            return 0
               
    def setPorts(self, ports):
        """
        Params:
               - The ports that should be added to the ports list.
        Adds the specified ports to the list of available ports. If a port is named
        alias, it will be omitted from the list."""
        
        for port in ports:
            if not (port == "Alias" or port == "Aliases" or port == "alias" or port == "aliases"): 
                self.ports[port] = []
                self.portList.insertItem(str(port))

    def slotShowAliases(self, text):
        """
        Params:
               - text : String - The name of the port that has been selected.
        Displays the aliases of the specified port name in the list of added aliases."""
        
        aliases = self.setupWidget.getPortAliases(str(text))

        self.portAliasList.clear()

        for alias in aliases:
            self.portAliasList.insertItem(str(alias))

    def slotAddAlias(self):
        """
        Slot fired when a user push the add button. Adds the alias to the port."""
        
        alias = str(self.availableList.currentText())
        port = str(self.portList.currentText())

        if alias and port:
            possibleOther = self.portAliasList.findItem(alias, Qt.ExactMatch)

            if not possibleOther:
                self.setupWidget.addPortAlias(port, alias)
                self.portAliasList.insertItem(alias)
            else:
                retval = QMessageBox.warning(self,
                                             "SciCraft",
                                             "Port " + port + " alreday has an alias " + alias + ".",
                                             QMessageBox.Ok,
                                             QMessageBox.NoButton)
        
        
    def slotRemoveAlias(self):
        """
        Deletes the selected alias from the selected port."""
        
        alias = str(self.portAliasList.currentText())
        index = self.portAliasList.currentItem()
        port = str(self.portList.currentText())

        if alias and port:
            self.setupWidget.removePortAlias(port, alias)
            self.portAliasList.removeItem(index)

    
    def __tr(self,s,c = None):
        return qApp.translate("AliasWidget",s,c)

if __name__=="__main__":

    app = QApplication([])

    win = AliasWidget(None)
    win.show()

    app.connect(app, SIGNAL("lastWindowClosed()"),
                app, SLOT("quit()"))

    app.exec_loop()
