# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QColor, QColorDialog, QGridLayout, QGroupBox, QIntValidator
from qt import QLabel, QLineEdit, QPushButton, QSize, QSizePolicy, QSlider
from qt import QSpacerItem, QString, QToolTip, QVBoxLayout, QWhatsThis, QWidget
from qt import Qt, SIGNAL
from GUI.Plots.Options import Options
from GUI.Plots import PlotConfig

class ColourOptions(Options.Options):
    """Options tied to the generic plot class. This is a widget that
    contains the GUI controls and the values associated with the available
    and user changeable configuraton options common to all plots."""
    
    def __init__(self, optionsWindow, name = None):
        """
        Parameters:
                   - parent : QWidget - The parent widget of this widget
                   - name : String - The name of this widget. Not used.

        Constuctor. Initialises all the values that a user
        can change. The constructor does not build the widget itself, this
        is the task of buildWidget()"""

        if not name:
            name = "Colour"

        Options.Options.__init__(self, optionsWindow, name)
        # Just in case, limit to max 255 and int value
        cl = self.renderWindow.renderer.GetBackground()
        self.bgRed = int(cl[0]*255)
        
        self.bgGreen = int(cl[1]*255)
                
        self.bgBlue = int(cl[2]*255)

        self.buildWidget()
        
    def buildWidget(self):
        """Creates the widget containg the controls"""
        Options.Options.buildWidget(self)
        
        self.setCaption(self.tr("Colour Options"))

        # Top level layout between the widget and the rest of the layots
        PlotOptionsLayout = QVBoxLayout(self, 11, 6,"PlotOptionsLayout")

        # The group box for changing background color of the renderer
        self.bgColourGroup = QGroupBox(self,"bgColourGroup")
        self.bgColourGroup.setColumnLayout(0,Qt.Vertical)
        self.bgColourGroup.layout().setSpacing(6)
        self.bgColourGroup.layout().setMargin(11)
        self.bgColourGroup.setTitle(self.tr("Background Colour"))
        QToolTip.add(self.bgColourGroup,self.tr("Background colour of plot"))
        QWhatsThis.add(self.bgColourGroup,self.tr("Sets the background colour of the plot. Specify the background colour with the sliders or enter the RGB values directly into the boxes. You may also click the \"Palette\" button to choose a colour from a palette."))

        bgColourGroupLayout = QGridLayout(self.bgColourGroup.layout())
        bgColourGroupLayout.setAlignment(Qt.AlignTop)
        
        # palette button
        self.moreButton = QPushButton(self.bgColourGroup)
        self.moreButton.setText(self.tr("Palette..."))
        self.moreButton.setMaximumSize(QSize(20000, 23))
        QToolTip.add(self.moreButton, self.tr("Choose colour from palette..."))
        QWhatsThis.add(self.moreButton, self.tr("Opens a window containing a colour palette to choose from"))
                        
        bgColourGroupLayout.addWidget(self.moreButton,2,3)

        # Sliders
        self.blueSlider = QSlider(self.bgColourGroup,"blueSlider")
        self.blueSlider.setMaxValue(255)
        self.blueSlider.setValue(self.bgBlue)
        self.blueSlider.setOrientation(QSlider.Horizontal)
        
        bgColourGroupLayout.addWidget(self.blueSlider,2,1)

        self.greenSlider = QSlider(self.bgColourGroup,"greenSlider")
        self.greenSlider.setMaxValue(255)
        self.greenSlider.setValue(self.bgGreen)
        self.greenSlider.setOrientation(QSlider.Horizontal)

        bgColourGroupLayout.addWidget(self.greenSlider,1,1)
        
        self.greenLabel = QLabel(self.bgColourGroup,"greenLabel")
        self.greenLabel.setText(self.tr("Green"))

        bgColourGroupLayout.addWidget(self.greenLabel,1,0)

        self.blueLabel = QLabel(self.bgColourGroup,"blueLabel")
        self.blueLabel.setText(self.tr("Blue"))

        bgColourGroupLayout.addWidget(self.blueLabel,2,0)

        self.redSlider = QSlider(self.bgColourGroup,"redSlider")
        self.redSlider.setMaxValue(255)
        self.redSlider.setValue(self.bgRed)
        self.redSlider.setOrientation(QSlider.Horizontal)

        bgColourGroupLayout.addWidget(self.redSlider,0,1)

        self.redLabel = QLabel(self.bgColourGroup,"redLabel")
        self.redLabel.setText(self.tr("Red"))

        bgColourGroupLayout.addWidget(self.redLabel,0,0)

        # Should only enter ints between 0 and 255
        validator = QIntValidator(0, 255, self)

        # Line Edit boxes
        self.redEdit = QLineEdit(self.bgColourGroup,"redEdit")
        self.redEdit.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed,0,0,self.redEdit.sizePolicy().hasHeightForWidth()))
        self.redEdit.setMaximumSize(QSize(31,32767))
        self.redEdit.setReadOnly(0)
        self.redEdit.setValidator(validator)
        self.redEdit.setText(str(self.bgRed))

        bgColourGroupLayout.addWidget(self.redEdit,0,2)

        self.greenEdit = QLineEdit(self.bgColourGroup,"greenEdit")
        self.greenEdit.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed,0,0,self.greenEdit.sizePolicy().hasHeightForWidth()))
        self.greenEdit.setMaximumSize(QSize(31,32767))
        self.greenEdit.setReadOnly(0)
        self.greenEdit.setValidator(validator)
        self.greenEdit.setText(str(self.bgGreen))

        bgColourGroupLayout.addWidget(self.greenEdit,1,2)

        self.blueEdit = QLineEdit(self.bgColourGroup,"blueEdit")
        self.blueEdit.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed,0,0,self.blueEdit.sizePolicy().hasHeightForWidth()))
        self.blueEdit.setMaximumSize(QSize(31,32767))
        self.blueEdit.setValidator(validator)
        self.blueEdit.setText(str(self.bgBlue))
        bgColourGroupLayout.addWidget(self.blueEdit,2,2)

        self.previewLabel = QLabel(self.bgColourGroup,"previewLabel")
        self.previewLabel.setPaletteBackgroundColor(QColor(self.bgRed, self.bgGreen, self.bgBlue))
        self.previewLabel.setFrameShape(QLabel.Box)
        self.previewLabel.setText(QString.null)

        bgColourGroupLayout.addMultiCellWidget(self.previewLabel,0,1,3,3)
        PlotOptionsLayout.addWidget(self.bgColourGroup)
        spacer = QSpacerItem(5, 2, QSizePolicy.Minimum, QSizePolicy.MinimumExpanding)
        PlotOptionsLayout.addItem(spacer)

        self.resize(QSize(512,151).expandedTo(self.minimumSizeHint()))

        # Connect buttons, sliders and edit boxes with slots
        self.connect(self.moreButton, SIGNAL("clicked()"), self.showPalette)
        self.connect(self.redSlider,SIGNAL("valueChanged(int)"),self.slotSliderChanged)
        self.connect(self.greenSlider,SIGNAL("valueChanged(int)"),self.slotSliderChanged)
        self.connect(self.blueSlider,SIGNAL("valueChanged(int)"),self.slotSliderChanged)
        self.connect(self.redEdit,SIGNAL("textChanged(const QString&)"),self.slotTextChanged)
        self.connect(self.greenEdit,SIGNAL("textChanged(const QString&)"),self.slotTextChanged)
        self.connect(self.blueEdit,SIGNAL("textChanged(const QString&)"),self.slotTextChanged)
        

    def slotSliderChanged(self, value):
        """
        Parameters:
                   - value : int - The new value of the slider that has
                   changed. Surprisingly this parameter is actually not used!

        Slot that is fired when one of the colour sliders are dragged.
        All relevant controls are updated with the new value"""

        self.bgRed = self.redSlider.value()
        self.bgGreen = self.greenSlider.value()
        self.bgBlue = self.blueSlider.value()

        self.redEdit.setText(str(self.bgRed))
        self.greenEdit.setText(str(self.bgGreen))
        self.blueEdit.setText(str(self.bgBlue))

        self.previewLabel.setPaletteBackgroundColor(QColor(self.bgRed, self.bgGreen, self.bgBlue))
        

    def slotTextChanged(self, value):
        """
        Parameters:
                   - value : String - The new value of the box that
                   has been edited (formatted as a String)
                   changed. Surprisingly this parameter is actually not used!

        Slot that is fired when one of the edit boxes has changed its
        value. All relevant controls are updated with the new value"""
        
        # One try around each colour to find which coulur is
        # failing. 
        try:
            t = str(self.redEdit.text())
            if not t:
                t = "0"
            self.bgRed = int(t)

        except (TypeError, ValueError), e:
            # If user enters text or spaces
            self.bgRed = 0

        try:
            t = str(self.greenEdit.text())
            if not t:
                t = "0"
            self.bgGreen = int(t)

        except (TypeError, ValueError), e:
            self.bgGreen = 0
                        
        try:
            t = str(self.blueEdit.text())
            if not t:
                t = "0"
            self.bgBlue = int(t)

        except (TypeError, ValueError), e:
            self.bgBlue = 0
            
        self.redSlider.setValue(self.bgRed)
        self.greenSlider.setValue(self.bgGreen)
        self.blueSlider.setValue(self.bgBlue)
                
        self.previewLabel.setPaletteBackgroundColor(QColor(self.bgRed, self.bgGreen, self.bgBlue))

    def showPalette(self):
        """Opens a modal QColorDialog and lets the user choose a colour
        more elegantly"""
        color = QColorDialog.getColor(QColor(self.bgRed, self.bgGreen, self.bgBlue), self)

        if color.isValid():
            # Do not alter the order of these lines !!!!
            self.bgRed = color.red()
            self.redSlider.setValue(self.bgRed)
            self.redEdit.setText(str(self.bgRed))

            self.bgGreen = color.green()
            self.greenSlider.setValue(self.bgGreen)
            self.greenEdit.setText(str(self.bgGreen))

            self.bgBlue = color.blue()
            self.blueSlider.setValue(self.bgBlue)
            self.blueEdit.setText(str(self.bgBlue))
            # !!!!

            self.previewLabel.setPaletteBackgroundColor(QColor(self.bgRed, self.bgGreen, self.bgBlue))
        
    def dummy(self):
        """You silly monkey """
        print "I am a PlotOptions Widget"

    def saveOptions(self):
        """Stores the configurable values to a permanent place.
        More specifically in the PlotConfig class."""
        
        PlotConfig.PlotConfig.bgR = self.bgRed
        PlotConfig.PlotConfig.bgG = self.bgGreen
        PlotConfig.PlotConfig.bgB = self.bgBlue

    def applyOptions(self):
        """Applies the current configration settings to the plot related
        to this specific options widget"""
        self.renderWindow.renderer.SetBackground((float(self.bgRed)/255,
                                          float(self.bgGreen)/255,
                                          float(self.bgBlue)/255))
        
        self.renderWindow.update()

        self.optionsWindow.renderSetup.apply_all_plot_settings()

    def allApplyOptions(self, renderWindows):
        for renWin in renderWindows:
            renWin.renderer.SetBackground((float(self.bgRed)/255,
                                           float(self.bgGreen)/255,
                                           float(self.bgBlue)/255))
            renWin.setupWidget.apply_all_plot_settings()
            

        
