# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QGridLayout, QGroupBox, QHBoxLayout, QListBox, QSize, QSizePolicy
from qt import QSpacerItem, QSplitter, QTextEdit, QToolTip, QVBoxLayout
from qt import QWhatsThis, QWidget, Qt, SIGNAL
from GUI.Plots.Options import Options
from GUI.Plots import PlotConfig

# from GUI.Plots import PlotOptions
class InteractionOptions(Options.Options):
    """Options tied to the generic plot3D class. This is a widget that
    contains the GUI controls and the values associated with the available
    and user changeable configuraton options common to all 3D plots."""

    def __init__(self, optionsWindow, name = None):
        """
        Parameters:
                   - parent : QWidget - The parent widget of this widget
                   - name : String - The name of this widget. Not used.

        Constuctor. Initialises all the values that a user
        can change. The constructor does not build the widget itself, this
        is the task of buildWidget()"""

        
        if not name:
            name = "Interaction"
        
        Options.Options.__init__(self, optionsWindow, name)

        self.interactionMode = PlotConfig.PlotConfig.interaction3DMode
        
        self.descriptionList = []

        self.buildWidget()
        item = self.modeList.findItem(self.renderSetup.getInteractorStyleName(), Qt.ExactMatch)
        self.modeList.setCurrentItem(item)


    def buildWidget(self):
        """Creates the widget containg the controls"""
        Options.Options.buildWidget(self)

        InteractionOptionsWidgetLayout = QVBoxLayout(self,11,6,"InteractionOptionsWidgetLayout")

        modeLayout = QHBoxLayout(None,0,6,"modeLayout")
        
        self.modeGroup = QGroupBox(self,"modeGroup")
        self.modeGroup.setTitle(self.tr("Interaction Mode"))
        QToolTip.add(self.modeGroup,self.tr("Choose how the interaction with the plot should work"))
        QWhatsThis.add(self.modeGroup,self.tr("Choose the mode in which the interaction with the plot should be."))
        
        self.modeGroup.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding,20,100,self.modeGroup.sizePolicy().hasHeightForWidth()))
        self.modeGroup.setColumnLayout(0,Qt.Vertical)
        self.modeGroup.layout().setSpacing(6)
        self.modeGroup.layout().setMargin(11)
        modeGroupLayout = QGridLayout(self.modeGroup.layout())
        modeGroupLayout.setAlignment(Qt.AlignTop)

        self.modeSplitter = QSplitter(self.modeGroup,"modeSplitter")
        self.modeSplitter.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding,0,100,self.modeSplitter.sizePolicy().hasHeightForWidth()))
        self.modeSplitter.setOrientation(QSplitter.Horizontal)

        self.modeList = QListBox(self.modeSplitter,"modeList")
        self.modeList.clear()
        self.modeList.insertItem(self.tr("Flight"))
        self.descriptionList.append(self.tr("""Flight motion:<br>
        Left mouse button press produces forward motion. Right mouse button press produces reverse motion. Moving mouse during motion steers user in desired direction. Keyboard controls are: Left/Right/Up/Down Arrows for steering direction 'A' forward, 'Z' reverse motion Ctrl Key causes sidestep instead of steering in mouse and key modes Shift key is accelerator in mouse and key modes Ctrl and Shift together causes Roll in mouse and key modes """))
        
        # self.modeList.insertItem(self.tr("Joystick Actor"))
        # self.descriptionList.append(self.tr("""Joystick Actor: <br> This mode allows the user to interact with (rotate, zoom, etc.) separate objects in the scene independent of each other. The position of the mouse relative to the center of the object determines the speed of the object's motion. The mouse's velocity detemines the acceleration of the object's motion, so the object will continue moving even when the mouse is not moving. For a 3-button mouse, the left button is for rotation, the right button for zooming, the middle button for panning, and ctrl + left button for spinning. (With fewer mouse buttons, ctrl + shift + left button is for zooming, and shift + left button is for panning.) """))
        
        self.modeList.insertItem(self.tr("Joystick Camera"))
        self.descriptionList.append(self.tr("""Joystick camera: <br>
        This mode allows the user to move (rotate, pan, etc.) the camera, the point of view for the scene. The position of the mouse relative to the center of the scene determines the speed at which the camera moves, and the speed of the mouse movement determines the acceleration of the camera, so the camera continues to move even if the mouse if not moving. For a 3-button mouse, the left button is for rotation, the right button for zooming, the middle button for panning, and ctrl + left button for spinning. (With fewer mouse buttons, ctrl + shift + left button is for zooming, and shift + left button is for panning.) """))
        
        # self.modeList.insertItem(self.tr("Trackball Actor"))
        # self.descriptionList.append(self.tr("""Trackball actor: <br> This mode allows the user to interact with (rotate, pan, etc.) objects in the scene indendent of each other. In trackball interaction, the magnitude of the mouse motion is proportional to the actor motion associated with a particular mouse binding. For example, small left-button motions cause small changes in the rotation of the actor around its center point. <br>The mouse bindings are as follows. For a 3-button mouse, the left button is for rotation, the right button for zooming, the middle button for panning, and ctrl + left button for spinning. (With fewer mouse buttons, ctrl + shift + left button is for zooming, and shift + left button is for panning.) """))
        
        self.modeList.insertItem(self.tr("Trackball Camera"))
        self.descriptionList.append(self.tr("""Trackball camera: <br>
        This mode allows the user to interactively manipulate (rotate, pan, etc.) the camera, the viewpoint of the scene. In trackball interaction, the magnitude of the mouse motion is proportional to the camera motion associated with a particular mouse binding. For example, small left-button motions cause small changes in the rotation of the camera around its focal point. For a 3-button mouse, the left button is for rotation, the right button for zooming, the middle button for panning, and ctrl + left button for spinning. (With fewer mouse buttons, ctrl + shift + left button is for zooming, and shift + left button is for panning.) """))
        
        self.modeList.insertItem(self.tr("Unicam"))
        self.descriptionList.append(self.tr("""Unicam: <br>
        This mode is not available yet. Do not select this mode!"""))
        
        QToolTip.add(self.modeList,self.tr("The available interaction modes"))
        
        self.modeList.setSizePolicy(QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding,20,0,self.modeList.sizePolicy().hasHeightForWidth()))
        self.modeList.setMinimumSize(QSize(110,0))

        self.modeDescription = QTextEdit(self.modeSplitter,"modeDescription")
        self.modeDescription.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding,40,0,self.modeDescription.sizePolicy().hasHeightForWidth()))
        QToolTip.add(self.modeDescription,self.tr("A description of the selected interaction mode"))
        self.modeDescription.setReadOnly(1)

        modeGroupLayout.addWidget(self.modeSplitter,0,0)
        modeLayout.addWidget(self.modeGroup)
        InteractionOptionsWidgetLayout.addLayout(modeLayout)
        spacer_3 = QSpacerItem(20,10,QSizePolicy.Minimum,QSizePolicy.Minimum)
        InteractionOptionsWidgetLayout.addItem(spacer_3)
        

        self.resize(QSize(556,293).expandedTo(self.minimumSizeHint()))

        self.connect(self.modeList,SIGNAL("highlighted(int)"),self.slotModeItemSelected)

        item = self.modeList.findItem(self.interactionMode)

        # If item is not found, the config-file contains a faulty default mode.
        if item:
            self.modeList.setSelected(item, 1)

            
    def slotModeItemSelected(self, index):
        """
        Parameters:
                   - index : int - The index of the item selected
        
        Slot fired every time the user selects another interaction mode
        in the list of modes. Displays the corresponding description string in the
        description field"""
        
        self.interactionMode = self.modeList.text(index)
        self.modeDescription.setText(self.descriptionList[index])
    

    def saveOptions(self):
        """Stores the configurable values to a permanent place.
        More specifically in the PlotConfig class."""
        PlotConfig.PlotConfig.interaction3DMode = self.interactionMode

    def applyOptions(self):
        """Applies the current configration settings to the plot related
        to this specific options widget"""
        self.renderWindow._Iren.SetInteractorStyle(self.renderSetup.interactionStyles[str(self.interactionMode)])
        

    def allApplyOptions(self, renderWindows):
        for renWin in renderWindows:

            renWin._Iren.SetInteractorStyle(renWin.setupWidget.interactionStyles[str(self.interactionMode)])
            
