# -*- coding: utf-8 -*-

# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QDoubleValidator, QGridLayout, QGroupBox, QLabel, QLineEdit
from qt import QListBox, QSize, QSizePolicy, QSlider, QTextEdit, QToolTip
from qt import QVBoxLayout, QWhatsThis, Qt, SIGNAL, qApp
from GUI.Plots.Options import Options
from GUI.Plots import PlotConfig

class StereoOptions(Options.Options):
    def __init__(self, optionsWindow, name = None):
        
        
        if not name:
            name = "Stereo"

        Options.Options.__init__(self, optionsWindow, name)
        self._mode_col = StereoModeCollection()

        self.buildWidget()
        

    def buildWidget(self):
        
        self.setCaption(self.__tr("Stereo options"))
        currentEyeAngle = float(self.renderWindow.renderer.GetActiveCamera().GetEyeAngle())
        
        StereoOptionsLayout = QVBoxLayout(self,11,6,"StereoOptionsLayout")
        # Stereo type group

        self.stereo_group = QGroupBox(self,"stereo_group")
        self.stereo_group.setColumnLayout(0,Qt.Vertical)
        self.stereo_group.layout().setSpacing(6)
        self.stereo_group.layout().setMargin(11)
        stereo_group_layout = QGridLayout(self.stereo_group.layout())
        stereo_group_layout.setAlignment(Qt.AlignTop)
        self.stereo_group.setTitle(self.__tr("Stereo type"))

        # Type list
        self.type_list = QListBox(self.stereo_group,"type_list")
        self.type_list.clear()
        self.type_list.setSizePolicy(QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding,0,0,
                                                 self.type_list.sizePolicy().hasHeightForWidth()))
        self.connect(self.type_list, SIGNAL("highlighted(const QString &)"), self.slot_mode_changed)
        stereo_group_layout.addWidget(self.type_list,1,0)

        self.type_label = QLabel(self.stereo_group,"type_label")
        self.type_label.setText(self.__tr("Type:"))
        stereo_group_layout.addWidget(self.type_label,0,0)

        # Description field
        self.description_edit = QTextEdit(self.stereo_group,"description_edit")
        self.description_edit.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding,0,0,self.description_edit.sizePolicy().hasHeightForWidth()))
        stereo_group_layout.addWidget(self.description_edit,1,1)
        
        self.description_label = QLabel(self.stereo_group,"description_label")
        self.description_label.setText(self.__tr("Description:"))
        stereo_group_layout.addWidget(self.description_label,0,1)        

        # Note label
        self.note_label = QLabel(self.stereo_group,"note_label")
        self.note_label.setText(self.__tr("<b>Note!</b> This setting affects all render windows."))
        self.note_label.setAlignment(QLabel.WordBreak | QLabel.AlignCenter)
        stereo_group_layout.addMultiCellWidget(self.note_label,2,2,0,1)

        StereoOptionsLayout.addWidget(self.stereo_group)

        modes = self._mode_col.get_modes()
        for mode in modes:            
            self.type_list.insertItem(mode.name)
        self.type_list.setCurrentItem(int(self.renderWindow.GetRenderWindow().GetStereoType()) - 1)
        
        # eye angle group
        self.eye_group = QGroupBox(self,"eye_group")
        self.eye_group.setTitle(self.__tr("Eye separation"))
        self.eye_group.setColumnLayout(0,Qt.Vertical)
        self.eye_group.layout().setSpacing(6)
        self.eye_group.layout().setMargin(11)
        QToolTip.add(self.eye_group,self.__tr("Set the eye separation in stereo view."))
        QWhatsThis.add(self.eye_group,self.__tr("Adjusts the eye separation (in degrees). This is useful in stereo mode only."))
        
        eye_groupLayout = QGridLayout(self.eye_group.layout())
        eye_groupLayout.setAlignment(Qt.AlignTop)

        self.eyeSlider = QSlider(self.eye_group,"eyeSlider")
        self.eyeSlider.setSizePolicy(QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed,1,0,self.eyeSlider.sizePolicy().hasHeightForWidth()))
        self.eyeSlider.setMaxValue(100)
        self.eyeSlider.setLineStep(1)
        self.eyeSlider.setValue(currentEyeAngle*10)
        self.eyeSlider.setOrientation(QSlider.Horizontal)
        self.connect(self.eyeSlider, SIGNAL("valueChanged(int)"), self.slotSliderMoved)

        eye_groupLayout.addWidget(self.eyeSlider,0,0)

        self.eyeEdit = QLineEdit(self.eye_group,"eyeEdit")
        self.eyeEdit.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed,0,0,self.eyeEdit.sizePolicy().hasHeightForWidth()))
        self.eyeEdit.setMaximumSize(QSize(50,32767))
        self.eyeEdit.setText(str(currentEyeAngle))
        self.eyeEdit.setValidator(QDoubleValidator(0.0, 90.0, 1, self))
        self.connect(self.eyeEdit, SIGNAL("textChanged(const QString &)"), self.slotEditChanged)

        eye_groupLayout.addWidget(self.eyeEdit,0,1)

        self.eyeLabel = QLabel(self.eye_group,"eyeLabel")
        self.eyeLabel.setText(self.__tr("degrees"))

        eye_groupLayout.addWidget(self.eyeLabel,0,2)
        StereoOptionsLayout.addWidget(self.eye_group)

        

        self.resize(QSize(284,76).expandedTo(self.minimumSizeHint()))
    
    def getEyeAngle(self):

        return float(self.eyeSlider.value()/10.0)
        
    def __tr(self,s,c = None):
        return qApp.translate("StereoOptions",s,c)

    def slot_mode_changed(self, current_text):
        current_text = str(current_text)

        description = self._mode_col.get_mode_description(current_text)
        self.description_edit.setText(description)

    def slotSliderMoved(self, value):
        self.eyeEdit.setText(str(value/10.0))

    def slotEditChanged(self, value):
        val = 0.0
        try:
            val = float(str(value))
        except:
            print "oops"

        val = int(val * 10.0)
        self.eyeSlider.setValue(val)

    def save_stereo(self):

        stereo_type = self._mode_col.get_mode_number(self.type_list.currentText())
        PlotConfig.PlotConfig.stereo_mode = stereo_type
        controller = self.optionsWindow.renderSetup.windowController

        render_windows = controller.getRenderWindows()

        for ren_win in render_windows:
            ren_win.GetRenderWindow().SetStereoType(stereo_type)        
    
    def saveOptions(self):
        PlotConfig.PlotConfig.eyeAngle = self.getEyeAngle()
        self.save_stereo()
        
    def applyOptions(self):
        self.renderWindow.renderer.GetActiveCamera().SetEyeAngle(self.getEyeAngle())
        self.save_stereo()
        
    def allApplyOptions(self, renderWindows):
        self.save_stereo()
        for renWin in renderWindows:
            renWin.renderer.GetActiveCamera().SetEyeAngle(self.getEyeAngle())

class StereoMode:

    def __init__(self, number, name, description="No description."):

        self.number = number
        self.name = name
        self.description = description

class StereoModeCollection:

    def __init__(self):

        self._modes = []

        self._modes.append(StereoMode(1, "Crystal eyes", "crystal"))
        self._modes.append(StereoMode(2, "Red/Blue"))
        self._modes.append(StereoMode(3, "Interlaced"))
        self._modes.append(StereoMode(4, "Left"))
        self._modes.append(StereoMode(5, "Right"))
        self._modes.append(StereoMode(6, "Dresden"))

    def get_modes(self):
        return self._modes

    def get_mode_number(self, mode_text):
        for mode in self._modes:
            if mode.name == mode_text:
                return mode.number
        return 0

    def get_mode_description(self, mode_text):

        for mode in self._modes:
            if mode.name == mode_text:
                return mode.description

        return ""

    def get_mode_name(self, number):
        for mode in self._modes:
            if mode.number == number:
                return mode.name

        return ""

    


    
