# -*- coding: utf-8 -*-

# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Plots import PlotConfig
from GUI.Plots import PlotDataFilter

class Plot:
    """Superclass for all plots """
    def __init__(self):
        """
        Constructor. Initializes some variables. The plot itself will not be created
        here, but in the "create" method"""
        self.plotType = ""
        self.options = None
        self.data = {}
        self.dataFilter = PlotDataFilter.PlotDataFilter(None, self)
        # self.dataFilter = BufferedPlotDataFilter.BufferedPlotDataFilter(None, self)
        self.actorList = {}
        self.actorList["pos"] = []
        self.actorList["picked"] = []
        self.actorList["deleted"] = []
        self.actorList["axis"] = []
        
        self.renderWindow = None
        self.renderer = None
        self.picker = None

        
    def destructSelf(self):

        self.data = {}
        self.dataFilter = None
        
        self.removeActors()

        if self.picker:
            self.picker.RemoveObservers("EndPickEvent")
        
        self.actorList = {}
        self.renderWindow = None
        self.renderer = None

    def removeActors(self):

        for actorGroup in self.actorList:
            for actor in self.actorList[actorGroup]:
                self.renderer.RemoveActor(actor)
            self.actorList[actorGroup] = []
        # self.actorList = {}

    
    def getDataFilter(self):
        return self.dataFilter

    
    def addData(self, data):

        pass

    
    def createSelf(self, renderWindow = None):
        """
        Creates the plot itself by calling the superclass constructor """
        
        # QVTKRenderWindowInteractor.QVTKRenderWindowInteractor.__init__(self, parent)

        if not renderWindow:
            return
        
        self.isPickable = 0
        
        self.menus = {}
        self.toolbarItems = []
        self.renderWindow = renderWindow
        self.picker = self.renderWindow.GetPicker()
        self.picker.AddObserver("PickEvent", lambda a,b : self.annotatePick())
        self.picker.AddObserver("EndPickEvent", self.endPickEvent)
        self.renderer = self.renderWindow.renderer

        self.createBarItems()
        self.initializeOptions()
        

    def createBarItems(self):
        """Creates all the necessary menu and toolbar items
        in the form of QActions, and adds the items to the
        toolbar list and/or the menu list """
        return
        
    
    def getMenus(self):
        """Empty method in this class --- should be redefined in
        specific plot classes. Returns a list of all the QActions
        that should be present in a menu, associated with this
        kind of plot"""
        return self.menus

    def getToolbarItems(self):
        """Empty method in this class --- should be redifined in
        specific plot classes. Returns a list of all the QActions
        that should be prsent in both menus and the toolbar."""
        return self.toolbarItems

    def getOptions(self, parent = None):
        """
        Params:
               - parent : QWidget - the parent widget of the options widget to come
        Return:
               - Options - The options widget of this plot"""
        
        if self.options:
            return self.options
        
        self.options = PlotOptions.PlotOptions(parent, "Plot Options", self)
        return self.options

    def initializeOptions(self):

        self.renderColor = (float(PlotConfig.PlotConfig.bgR)/255,
                            float(PlotConfig.PlotConfig.bgG)/255,
                            float(PlotConfig.PlotConfig.bgB)/255)
        
        self.renderer.SetBackground(self.renderColor)

    def setDataSource(self, dataDict):
        """
        Params:
               - dataDict : Dict - A dictionary with variable names as keys and
               data as records.

        Set the data source to this plot."""

        self.data = dataDict

    def addDataSource(self, name, data):
        """
        Params:
               - name : String - The name of the variable to be added
               - data : List - The data of the variable
        Adds a data source to the dictionary of all data sources to this plot"""
        
        self.data[name] = data

    def annotatePick(self, actor = None):
        """
        Method called when an picking operation is perfomed. If
        we use select box mode picking, the actor parameter will be
        set, otherwise this method will find the picked actor itself.
        """
        if not actor:
            actor = self.picker.GetActor()
        

        if actor in self.actorList["pos"]:
            
            objectIndex = self.actorList["pos"].index(actor)
            
            if actor in self.actorList["picked"]:
                self.setupWidget.renderSetup.windowController.filterControl.unRegisterPick(objectIndex, self.dataFilter)

            else:
                self.setupWidget.renderSetup.windowController.filterControl.registerPick(objectIndex, self.dataFilter)
        
        # print self.setupWidget
    def endPickEvent(self, object, event):
        pass
        ###### Continue here
    def pickActor(self, index):
        print "Plot: pickActor " + str(index)

    def unPickActor(self, index):
        print "Plot: unPickActor" + str(index)

    def deletePickedElements(self):
        retVal = []
        for actor in self.actorList["picked"]:

            self.renderer.RemoveActor(actor)
            retVal += self.dataFilter.getRemoveStatement(self.actorList["pos"].index(actor))
            self.actorList["deleted"].append(actor)

        # Clear the picked list, now all the actors are moved to the
        # deleted list
        self.actorList["picked"] = []
        self.renderWindow.update()
        return retVal
        # return []

    def undoPicking(self):
        """
        Puts all the actors that have been removed from the plot back again"""
        for actor in self.actorList["deleted"]:
            actor.GetProperty().SetColor(self.negateColor(actor.GetProperty().GetColor()))
            self.renderer.AddActor(actor)

        self.actorList["deleted"] = []
        self.renderWindow.update()

    def unPickAllActors(self):
        """Unpicks all the actors that have been picked, but not yet removed"""

    def selectWithinBox(self, x, y, width, height):
        """Params:
                   - x : int - x coordinate of the box (left to right)
                   - y : int - y coordinate of the box (top to bottom). Note that
                   VTK uses bottom-to-top so this value must be subtracted from the
                   height of the PlotRenderWindow if the value should make sense to VTK.
                   - width : int - The width of the box
                   - height : int - The height of the box
                   
        Picks all actors inside the defined bounding box."""
        pass
            

    def redrawPlot(self, renderWindow = None):
        """

        Exception:
                  - IndexError - 
        """
        # Torbjorn stuff
        if not self.dataFilter.getPlotData():
            return

        
        if not self.renderer:
            self.createSelf(renderWindow)
            
        self.removeActors()

        # self.actorList["pos"] = []
        # self.actorList["axis"] = []
        # self.actorList["axistitle"] = []

        self.dataFilter.updateRangeToData()
                
        self.addData(self.dataFilter.getPlotData())
        # Torbjrn stuff
        try:
            self.addAxes()
        except:
            pass
                
        self.renderWindow.update()
        
    def getSetupWidget(self, renderSetup):
        """
        Params:
               - renderSetup : RenderSetupWidget - The render setup widget
               that controlls all things related to the render window the
               plot belongs to.
               
        Returns:
               - None - This methos will be redefined in the plot subclasses.
               The generic Plot setup widget is not useful to anybody, so it
               will not be returned
        Returns the setup widget of this plot. If no exists, one will be
        created before is is returned"""

        return None

    def negateColor(self, color):
        
        newColor = []
        newColor.append(1 - color[0])
        newColor.append(1 - color[1])
        newColor.append(1 - color[2])
        return newColor

    def getPreferredTextColor(self):

        if not self.renderer:
            return [0.0,0.0,0.0]
        
        textColor = self.renderer.GetBackground()
        sum = 0.0
        for num in textColor:
            sum += float(num)
        
        if float(sum) > 1.5:
            textColor = [0.0, 0.0, 0.0]
        else:
            textColor = [1.0, 1.0, 1.0]

        return textColor
