# -*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QHBoxLayout, QListView, QListViewItem, QPushButton, QSize
from qt import QSizePolicy, QSpacerItem, QVBoxLayout, QWidget, Qt, SIGNAL, SLOT
from qt import qApp
from GUI.Plots import PlotRepository

class PlotAddWindow(QWidget):
    """This is a window from which a user can select plots to add into a plot window.
    This window is opened from a RenderSetupWidget whenever a user clicks the Add button."""
    def __init__(self, setupWidget, name = None, fl = 0):
        """
        Params:
               - setupWidget : RenderSetupWidget - the setup widget that wants to
               add a new plot.
               - parent : QWidget - This widget's parent. Ususally None because this is
               a stand-alonme window.

        Constructs a list view and fills is with all the plots available in the plot
        repository"""

        # We want this dialog to be modal
        QWidget.__init__(self, setupWidget, name, fl|Qt.WShowModal|Qt.WType_Dialog)

        if not name:
            self.setName("PlotAddWindow")

        # Setup a repository
        self.plotRepository = PlotRepository.PlotRepository()
        
        self.setupWidget = setupWidget

        self.setCaption(self.__tr("Available plots"))
        PlotAddWindowLayout = QVBoxLayout(self, 11, 6, "PlotAddWindowLayout")

        self.plotList = QListView(self,"plotList")
        self.plotList.addColumn(self.__tr("Plot"))
        self.plotList.header().setClickEnabled(0, self.plotList.header().count() - 1)
        self.plotList.addColumn(self.__tr("Description"))
        self.plotList.header().setClickEnabled(0, self.plotList.header().count() - 1)
        self.plotList.setResizeMode(QListView.AllColumns)
        self.connect(self.plotList, SIGNAL("doubleClicked(QListViewItem *)"), self.slotDoubleClicked)
        
        
        self.plotList.setRootIsDecorated(1)
        self.plotList.clear()

        PlotAddWindowLayout.addWidget(self.plotList)

        buttonLayout = QHBoxLayout(None,0,6,"buttonLayout")
        spacer = QSpacerItem(300,20,QSizePolicy.Expanding,QSizePolicy.Minimum)
        buttonLayout.addItem(spacer)

        self.addButton = QPushButton(self,"addButton")
        self.addButton.setText(self.__tr("Add"))
        self.connect(self.addButton, SIGNAL("clicked()"), self.slotAddPlot)
        buttonLayout.addWidget(self.addButton)

        self.doneButton = QPushButton(self,"doneButton")
        self.doneButton.setText(self.__tr("Done"))
        buttonLayout.addWidget(self.doneButton)
        self.connect(self.doneButton, SIGNAL("clicked()"), self, SLOT("close()"))

        
        PlotAddWindowLayout.addLayout(buttonLayout)

        self.resize(QSize(350,200).expandedTo(self.minimumSizeHint()))
        self.buildList()

    def buildList(self):
        """Builds the list view with all the plots received from the plot repository"""
        plotDict = self.plotRepository.getPlotTypesAsDict()

        for key in plotDict:
            levelList = key.split(".")
            lastItem = levelList[-1]
            del levelList[-1]
            parentItem = self.plotList
            for level in levelList:
                item = self.plotList.findItem(level, 0)
                if item:
                    parentItem = item
                else:
                    newItem = QListViewItem(parentItem, level)
                    newItem.setSelectable(0)
                    parentItem.insertItem(newItem)
                    parentItem = newItem
            plotItem = QListViewItem(parentItem, lastItem)
            parentItem.insertItem(plotItem)

    def slotDoubleClicked(self, item):
        self.slotAddPlot()
    
    def slotAddPlot(self):
        """Gets the plot corresponding to the selected list view item and adds it to the
        render setup widget"""
        listItem = self.plotList.selectedItem()
        if listItem:
            try:
                plot = self.plotRepository.getPlot(listItem.text(0))
                self.setupWidget.addPlot(plot)

            except ValueError, e:
                print e
    
    def __tr(self,s,c = None):
        return qApp.translate("PlotAddWindow",s,c)
