# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QHBoxLayout, QMainWindow, QPushButton, QSizePolicy, QSpacerItem
from qt import QTabWidget, QToolTip, QVBoxLayout, QWhatsThis, QWidget, Qt
from qt import SIGNAL, SLOT
from GUI.Plots import PlotConfig
from GUI.Plots.Options import InteractionOptions
from GUI.Plots.Options import ColourOptions
from GUI.Plots.Options import StereoOptions

class PlotOptionsWindow(QMainWindow):


    def __init__(self, renderSetup, name = None, fl = 0):

        QMainWindow.__init__(self, renderSetup, name,
                             fl |
                             Qt.WType_Dialog |
                             Qt.WShowModal |
                             Qt.WStyle_ContextHelp)

        if name:

            self.setCaption(name)
        
        self.setCentralWidget(QWidget(self))

        self.renderSetup = renderSetup
        self.renderWindow = renderSetup.renderWindow
        

        # What expands all the widgets to the central widget
        centralLayout = QVBoxLayout(self.centralWidget(),11,6)

        # The layout that binds the buttons and tab bar together
        mainLayout = QVBoxLayout(None,0,6)

        # The layout for organising the buttons
        buttonLayout = QHBoxLayout(None,0,6)

        spacer = QSpacerItem(1,20,QSizePolicy.Expanding,QSizePolicy.Minimum)
        buttonLayout.addItem(spacer)

        # Save settings for all plots
        self.saveButton = QPushButton(self.centralWidget())
        self.saveButton.setText(self.tr("Make default"))
        self.connect(self.saveButton, SIGNAL("clicked()"), self.slotSaveOptions)
        QToolTip.add(self.saveButton,self.tr("Save the settings and use them on all render windows in the future"))
        QWhatsThis.add(self.saveButton,self.tr("Save the settings and use them on all render windows. This will save the settings permanently and they will be applied to all plot windows in the future as well"))
        buttonLayout.addWidget(self.saveButton)

        # Save settings on all open plots
        self.openButton = QPushButton(self.centralWidget())
        self.openButton.setText(self.tr("Use on all render windows"))
        self.connect(self.openButton, SIGNAL("clicked()"), self.slotAllApplyOptions)
        QToolTip.add(self.openButton, self.tr("Apply the changes to all render windows in this node."))
        QWhatsThis.add(self.openButton, self.tr("Apply the changes to all the plot windows in this plot node. The settings will not be saved for other nodes. This window will be closed after pressing this button."))
        buttonLayout.addWidget(self.openButton)

        # Save settings for this plot only
        self.okButton = QPushButton(self.centralWidget())
        self.okButton.setText(self.tr("Use on this render window"))
        self.connect(self.okButton, SIGNAL("clicked()"), self.slotApplyOptions)
        QToolTip.add(self.okButton, self.tr("Apply the changes on this plot window only"))
        QWhatsThis.add(self.okButton, self.tr("Apply the changes on this plot window only. The settings will not be applied to other plot windows. This window will be closed after pressing this button."))
        buttonLayout.addWidget(self.okButton)

        
        # Cancel, no saving
        self.cancelButton = QPushButton(self.centralWidget())
        self.cancelButton.setText(self.tr("Cancel"))
        self.connect(self.cancelButton, SIGNAL("clicked()"), self, SLOT("close()"))
        QToolTip.add(self.cancelButton, self.tr("Cancel all changes and revert to the previous settings"))
        QWhatsThis.add(self.cancelButton, self.tr("Cancel all changes and revert to the previous settings. The window will be closed after pressing this button"))
        buttonLayout.addWidget(self.cancelButton)

        self.tabWidget = QTabWidget(self.centralWidget())
        self.tabWidget.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding,0,0,self.tabWidget.sizePolicy().hasHeightForWidth()))

        mainLayout.addWidget(self.tabWidget)
        mainLayout.addLayout(buttonLayout)
        centralLayout.addLayout(mainLayout)

        colour = ColourOptions.ColourOptions(self)
        interaction = InteractionOptions.InteractionOptions(self)
        stereo = StereoOptions.StereoOptions(self)

        self.addTab(colour, colour.name)
        self.addTab(interaction, interaction.name)
        self.addTab(stereo, stereo.name)

    def addTab(self, widget, text):
        self.tabWidget.insertTab(widget, text)
        widget.show()

        
    def slotSaveOptions(self):

        for tabIndex in range(0, self.tabWidget.count()):
            self.tabWidget.page(tabIndex).saveOptions()
        self.slotAllApplyOptions()

        PlotConfig.writeConfig()
        self.close()
        
    def slotApplyOptions(self):

        for tabIndex in range(0, self.tabWidget.count()):

            self.tabWidget.page(tabIndex).applyOptions()

        self.close()

    def slotAllApplyOptions(self):

        renderWindows = self.renderSetup.windowController.getRenderWindows()
        for tabIndex in range(self.tabWidget.count()):

            self.tabWidget.page(tabIndex).allApplyOptions(renderWindows)

        self.close()
    
            
        

