
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Plots.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
from vtk import vtkRenderer
from qt import QKeyEvent, QSizePolicy
from GUI.Plots import RenderSetupWidget
from GUI.Plots.interaction.NullMode import NullMode
from GUI.Plots import PlotConfig

class PlotRenderWindow(QVTKRenderWindowInteractor):
    """
    This is the actual window in which plots are drawn. It contains
    a reference to a vtkRenderer, responsible for the actual rendering done
    by VTK
    """

    def __init__(self, plotWindow, windowController, parent = None, name = None):

        QVTKRenderWindowInteractor.__init__(self, parent, name)
        self.parent = parent
        self.plotWindow = plotWindow
        self.windowControl = self.plotWindow.controller
        self.renderer = vtkRenderer()
        
        self.activeMode = NullMode(self)

        # Try to make text and lines smoother
        self.GetRenderWindow().SetLineSmoothing(1)
        self.GetRenderWindow().AddRenderer(self.renderer)

        # Set default behaviour from PlotConfig
        self.renderer.GetActiveCamera().SetEyeAngle(float(PlotConfig.PlotConfig.eyeAngle))
        self.GetRenderWindow().SetStereoType(int(PlotConfig.PlotConfig.stereo_mode))

        # self.resize(QSize(150,150))
        # self.setMinimumSize(QSize(30, 30))
        self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))
        
        self.setupWidget = RenderSetupWidget.RenderSetupWidget(self, windowController)
        # self.hideBoundingBox()

    def hidePlotList(self):
        """
        Method used by the molecule editor,
        as it craves no such thing as a list of several plots.
        """
        self.setupWidget.hidePlotList()

    def hideBoundingBox(self):
        interactorStyle = self._Iren.GetInteractorStyle()
        interactorStyle.HighlightProp(None)

    def renewRenderer(self):
        """
        Clears the vtkRenderer attached to this render window
        in order to delete all vtk elements that no longer are in use.
        Creates a new vtkRenderer and adds it to this render window
        afterwards.
        """
        bgColor = self.renderer.GetBackground()
        # inter = self.
        self.GetRenderWindow().RemoveRenderer(self.renderer)
        self.renderer = vtkRenderer()
        self.renderer.SetBackground(bgColor)
        self.GetRenderWindow().AddRenderer(self.renderer)
        
    def getSetupWidget(self):
        """
        Return:
               - RenderSetupWidget - the render setup widget governing
               the contents of this render window.
        """
        return self.setupWidget

    

    def toggle3D(self, on):
        """
        Params:
              - on : bool - whether the 3d mode is activated or deactivated.

        Toggles this render window in and out of 3 mode. This method really
        does not care about the value of the parameter on, it just sends the
        same key event anyway, hoping that the start state og the 3D button
        yields the start state of the render window. If it does, we have no
        worries.
        """
        
        pressEvent = QKeyEvent(QKeyEvent.KeyPress, QKeyEvent.Key_3, QKeyEvent.Key_3, 0)
        self.keyPressEvent(pressEvent)

    def mousePressEvent(self, ev):
        self.plotWindow.setActiveRenderer(self)
        fallthrough = self.activeMode.mousePressed(ev.x(), ev.y())
        if fallthrough:
            QVTKRenderWindowInteractor.mousePressEvent(self, ev)
        
    def mouseMoveEvent(self, ev):
        fallthrough = self.activeMode.mouseMoved(ev.x(), ev.y())
        if fallthrough:
            QVTKRenderWindowInteractor.mouseMoveEvent(self, ev)

    def mouseReleaseEvent(self, ev):
        fallthrough = self.activeMode.mouseReleased(ev.x(), ev.y())
        if fallthrough:
            QVTKRenderWindowInteractor.mouseReleaseEvent(self, ev)    

    def paintEvent(self, ev):
        QVTKRenderWindowInteractor.paintEvent(self, ev)
        self.activeMode.postPaintEvent()

    def startSelection(self, callback = None):
        self.selectionCallBack = callback
        self.selectMode = True

    def getActiveMode(self):
        return self.activeMode

    def setActiveMode(self, mode):
        self.activeMode.endMode()
        self.activeMode = mode
        self.activeMode.startMode()
        
