# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.


import os, sys
from qt import QPopupMenu

from Library.Util.Config import Config


class PlotRepository:
    """A PlotRepository class is instanciated at the startup of SciCraft.
    The repository searches through the plot directories and builds a
    dictionary of the plots present. It also builds a QPopupMenu that can
    be retrieved and inserted into the GUI. In this way, the user will get
    access to all the plots present in the system."""
    def __init__(self, workspace = None):
        """Constructor"""
        
        self.workspace = workspace
        # list of potential plots
        self.candidateList = {}

        # list of valid plots
        self.plots = {}
        self.customPlots = {}

        # Index of plots and their names
        self.indexCounter = 0
        self.indexedPlots = []

        # The base directory of all regular plots
        self.baseDirectory = "Plots"
        self.customBaseDirectory = "Custom"
        
        # The (relative) path from the main program directory
        # to the base directory. Will probably be GUI
        # 
        self.pathToBaseDirectory = os.path.join(Config.scicrafthome, "GUI")
        self.basePackage = "GUI"
        
        # Valid file suffices for plot files and custom made plots
        self.validSuffix = [".py"]
        self.validCustomSuffix = [".zpl"]

        # Files that should not be checked 
        self.invalidFilenames = ["__init__", "PlotRepository", "Plot", "Plot2D","plot2d", "Plot3D", "Options"]

        # find path to .scicraft/Customplots (or equivalent)
        custompath = os.path.join(Config.userconfigdir, "Plots")
        if not os.path.isdir(custompath) :
            os.mkdir(custompath, 0755)

        # build the plot list
        self.buildPlotTypes(os.path.join(self.pathToBaseDirectory, self.baseDirectory))
        self.buildCustomPlots(os.path.join(custompath, self.customBaseDirectory))
        self.buildCustomPlots(Config.customplotdir)
        

    def getPlot(self, name):
        """
        Params:
               - name : String - A string representing the plot type requested

        Exception:
               - ValueError - if a plot with the given name can't be found

        Returns:
               - Plot - An instance of the plot type requested"""

        # make name a string in case it was a QString (may happen i think)
        realName = str(name).split(".")[-1]

        
        if self.plots.has_key(realName):
            # retrieve the import statement from the dictionary
            execString = self.plots[realName][1]

            # execute the import statement
            exec(execString)

            # return an instance
            return mod()
        else:
            raise ValueError, "No plots found matching the name: " + str(name)

    def getCustomPlot(self, name):
        """
        Params:
               - name : String - A string representing the plot type requested

        Exception:
               - ValueError - if a plot with the given name can't be found

        Returns:
               - Plot - Path to the .zpl file containing the custom plot"""
        realName = str(name).split(".")[-1]

        if self.customPlots.has_key(realName):
            path = ""
            for part in self.customPlots[realName][0]:
                path += part + os.sep
            path += realName
            path += self.validCustomSuffix[0]
            return path
        else:
            raise ValueError, "No plots found matching the name: " + str(name)
        
    def buildCustomPlots(self, path):
        """
        Params:
               - path : String - The path to the root of the directory that
               holds all the plots

        Builds an internal list of all the plots found in the directory
        structure below 'path'"""
        
        # walks through directories and builds the candidateList
        # For each of the directories entered the method self.checkDir
        # will be called with the directory as its parameter.
        os.path.walk(path, self.checkCustomDir, "")

    def buildPlotTypes(self, path):
        """
        Params:
               - path : String - The path to the root of the directory that
               holds all the plots

        Builds an internal list of all the plots found in the directory
        structure below 'path'"""
        
        # walks through directories and builds the candidateList
        # For each of the directories entered the method self.checkDir
        # will be called with the directory as its parameter.
        os.path.walk(path, self.checkDir, "")

    def checkDir(self, arg, dirname, names):
        """
        Params:
               - arg : List - Bogus variable not used
               - dirname : String - The path of the directory to be checked
               - names : List[String] - The names of all the files inside 'dirname'

        Finds all the valid files in the given directory and adds the names to a
        candidate list of all the potential plots"""

        # We will skip the checking if we are in a .svn directory.
        if ".svn" in dirname:
            return
        
        for file in names:

            # if file is a valid file, strip the .py suffix from
            # the file and add the file with its path
            # in the candidateList
            if self.isValidFile(dirname,file):
                
                splitList = os.path.splitext(file)
                if dirname.startswith(self.pathToBaseDirectory):
                    dirname = dirname[len(self.pathToBaseDirectory):]

                strippedPackagePath = dirname.split(os.sep)

                if strippedPackagePath[0] == ".":
                    del strippedPackagePath[0]

                # Start constructing the import statement.
                importStatement = "from " + self.basePackage

                for pathElem in strippedPackagePath:
                    importStatement += pathElem + "."
                
                # Final part of the import Statement
                importStatement += splitList[0] + " import " + splitList[0] + " as mod"

                self.plots[splitList[0]] = [strippedPackagePath, importStatement]

                # self.candidateList[splitList[0]] = dirname
            
    def checkCustomDir(self, arg, dirname, files):
        if ".svn" in dirname:
            return
        for file in files:

            # if file is a valid file, strip the .py suffix from
            # the file and add the file with its path
            # in the candidateList
            if self.isValidCustomFile(dirname,file):
                
                splitList = os.path.splitext(file)
                strippedPackagePath = dirname.split(os.sep)

                if strippedPackagePath[0] == ".":
                    del strippedPackagePath[0]

                # Start constructing the import statement.
#                importStatement = "from "

#                for pathElem in strippedPackagePath:
#                    importStatement += pathElem + "."
                
                # Final part of the import Statement
#                importStatement += splitList[0] + " import " + splitList[0] + " as mod"

                self.customPlots[splitList[0]] = [strippedPackagePath]

                # self.candidateList[splitList[0]] = dirname
            
    def isValidFile(self, path, filename):
        """
        Params:
               - path : String - The path to the file to be checked
               - filename : String - The file to be checked for validity
        Checks that a given file could be a plot file by determining that it is
        actually a file (and not a directory), it has the same name as the directory it lies in
        and that it has the right filename
        prefix and suffix"""
        # check if it is a file, and split away the suffix.
        if os.path.isfile(os.path.join(path, filename)):
            splitList = os.path.splitext(filename)

            directory = os.path.basename(path)
            
            # check if the file has both a valid suffix, and not an
            # invalid prefix (like __init__.py etc.)
            if splitList[0] == directory and not(splitList[0] in self.invalidFilenames) and splitList[1] in self.validSuffix:
                return 1
        return 0

    def isValidCustomFile(self, path, filename):
        """
        Params:
               - path : String - The path to the file to be checked
               - filename : String - The file to be checked for validity
        Checks that a given file could be a custom plot file by determining that it is
        actually a file (and not a directory)
        and that it has the right filename
        prefix and suffix"""
        # check if it is a file, and split away the suffix.
        if os.path.isfile(os.path.join(path, filename)):
            splitList = os.path.splitext(filename)

            directory = os.path.basename(path)
            
            # check if the file has both a valid suffix, and not an
            # invalid prefix (like __init__.py etc.)
            if not(splitList[0] in self.invalidFilenames) and splitList[1] in self.validCustomSuffix:
                return 1
        return 0

    def getPlotTypesAsDict(self):
        """
        Return:
               - dict - Adictinary of all the different plot types in the repository.
        Returns all the plot types in a way that the gui can handle. See also documnetation
        of getPlotsAsDict"""

        plotDict = {}
        # Go through all the plots
        for plot in self.plots:

            # Not sure if this should be done, but it sure looks nice!
            baseIndex = self.findBaseIndex(self.plots[plot][0])

            #Build the key string
            menustring = ""

            # Find the index and append the key string with the directory
            # at the index.
            for directoryindex in range(baseIndex + 1, len(self.plots[plot][0])):
                menustring += self.plots[plot][0][directoryindex] + "."

            # Remove trailing dot
            if menustring[-1] == ".":
                
                menustring = menustring[0:-1]
                
            # This is commented out. Could cause problems with funny directory structures.
            # menustring += plot

            # Add to return dict
            plotDict[menustring] = plot

        return plotDict
            
        
    def getPlotsAsDict(self):
        """
        Return:
                   - dict - A dictionary of all the plots in the repository.
        Returns all the plots in a way that the gui can handle. This means that the dictionary's
        keys consists of a dot separated string of menu desired menu structure of the plot and the
        values are strings with the text that the plot node should display in the GUI. E.g. a
        Scatter Plot 3D will be returned as {Plots.Plots3D.Scatter Plot 3D : "Scatter Plot 3D"}"""
        
        plotDict = {}
        plotDict["Plot3D"] = "Plot3D"
                
        # Get the custom plots
        for plot in self.customPlots:
            # Not sure if this should be done, but it sure looks nice!
            baseIndex = self.findBaseIndex(self.customPlots[plot][0])

            #Build the key string
            menustring = ""

            # Find the index and append the key string with the directory
            # at the index.
            for directoryindex in range(baseIndex + 1, len(self.customPlots[plot][0])):
                menustring += self.customPlots[plot][0][directoryindex] + "."
            menustring += plot

            # Add to return dict
            plotDict[menustring] = plot
        
        return plotDict
    
    def findBaseIndex(self, plot):
        """Return:
                   - int : The index of tha base directory for the plots
                   in the path list for a plot"""
        for index in range(0, len(plot)):

            custompathbase = Config.customplotdir
            if custompathbase.endswith(os.sep):
                custompathbase = custompathbase[:-1]
            custompathbase = os.path.basename(custompathbase)
            
            if plot[index] in (self.baseDirectory, self.customBaseDirectory, custompathbase):
                return index
            
        raise ValueError, "Plot not inside base directory, this is a MAJOR BUG!"
    
    def slotTest(self):
        print "Plot clicked: " + self.indexedPlots

    
