# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QAction, QActionGroup, QCloseEvent, QColor, QDialog, QFileDialog
from qt import QFrame, QGridLayout, QIconSet, QKeySequence, QMainWindow
from qt import QMenuBar, QMessageBox, QPixmap, QPoint, QPopupMenu, QSize
from qt import QSizePolicy, QSlider, QSplitter, QString, QToolBar, QToolTip
from qt import QWhatsThis, QWidget, Qt, SIGNAL, SLOT
import os
from vtk import vtkBMPWriter, vtkJPEGWriter, vtkPNGWriter, vtkPNMWriter
from vtk import vtkPostScriptWriter, vtkTIFFWriter, vtkWindowToImageFilter
from GUI.Pixmaps.Icons import Icons
from GUI.Plots import Plot
from GUI.Plots import PlotConfig
from GUI.Plots import PlotOptionsWindow
from GUI.viewers import HelpViewer
from GUI.Plots import PlotRenderWindow
from Library.Util.Config import Config

from GUI.Plots import PlotWindowControl
# Testing testing
import PlotRepository

class PlotWindow(QMainWindow):
    """This class will handle all the common elements of a plotting window GUI """

    def __init__(self, data, eventServer, node, type=None, parent = None):
        """
        Params:
               - data : List - the data that should be plotted, this parameter may be changed
               - parent : QWidget - This window's parent. Can be None

        Opens a plot window and displays a plot of the given data."""
    
        QMainWindow.__init__(self,
                             parent,
                             "plotwindow",
                             Qt.WType_TopLevel | Qt.WStyle_ContextHelp)
        
        self.setCaption(self.tr("Plot Window"))
        
        # a timer for the duration of status messages.
        # Could be ignored, but holds a reasonable value for all
        # time dependent messages.
        self.statusTime = 3000

        # Add a callback to the eventsystem.
        eventServer.addEventHandler(self.dataUpdatedCallback)

        self.node = node
        # self.plotType = plotType
        # Storing the data to be plotted
        self.data = data
        self.dataUpdated = False
        # self.resize(500, 400)
        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        self.resize(QSize(500, 400).expandedTo(self.minimumSizeHint()))
        self.setCentralWidget(QWidget(self, "centralWidget"))
                
        self.controller = None
        
        self.renderWindowList = []
        self.plotList = []
        self.newRenderPosition = (0,0)
        self.activeRenderer = None

        self.plotLayout = QGridLayout(self.centralWidget(), 0, 0, 0, 0, "plotLayout")

        self.lineSplitter = QSplitter(self.centralWidget(), "lineSplitter")
        # More resizing problems.
        self.lineSplitter.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)

        self.lineSplitter.setOrientation(QSplitter.Vertical)
        self.activeColSplitter = QSplitter(self.lineSplitter)
        self.activeColSplitter.setOrientation(QSplitter.Horizontal)

        self.plotLayout.addWidget(self.lineSplitter,0,0)
        
        self.statusBar = self.statusBar()

        # Add general and specific menus
        self.addWindowController()
        self.initToolbars()

        # A compability fix, after the generic Plot was renamed to Plot3D
        # With this, old MD's will still work, but the plot nodes will come
        # up with their old names ("Plot"). This is fixed by saving the
        # diagram.
        if type == "Plot":
            type = "Plot3D"
        if type and not type == "Plot3D":
            plotRepository = PlotRepository.PlotRepository()
            path = plotRepository.getCustomPlot(type)
            self.loadConfig(path)
        
        self.controller.loadNodeDiagram()

    def isDataUpdated(self):
        """
        Return:
              - bool - True if the data is updated and there is no need
              to update this window's data. Returns False if the module diagram
              has been executed after the first opening of the window.
        """
        return self.dataUpdated

    def setDataUpdated(self, up):
        """
        Params:
              - up : bool - If true the data from middleware is updated and plots need
              to be redrawn.
        """
        
        self.dataUpdated = up

    def dataUpdatedCallback(self, event):
        """
        Params:
              - event : System.events - The event that just occured.
              
        Event callback that listens on events sent from EventServer. If
        the event is a start event of the module diagram the dataUpdated
        paramater will be set to false, and plots will have to be redrawn when
        reopening this plot window.        
        """
        if event.get_event_class() == "DiagramEvent" and \
           event.getType() == "START":
            self.dataUpdated = False
    
    def show(self):
        QMainWindow.show(self)

        if self.controller.isFinished():
            self.viewControllerAction.setOn(0)
        else:
            self.viewControllerAction.setOn(1)

    def setData(self, data):
        self.data = data

    def getData(self):
        return self.data
    
    def addPlot(self, plotname, renderWindow = None):

        plotRep = PlotRepository.PlotRepository()
        plot = plotRep.getPlot(plotname)
        if not renderWindow:
            # May be changed!
            self.addRenderWindows(1,1)
            renderWindow = self.renderWindowList[0]
            plot.createSelf(renderWindow)
        
        else:
            plot.createSelf(renderWindow)
        
        return plot
            

    def addRenderWindows(self, rows = 1, columns = 1):
        """
        Params:
               - rows : int - Number of rows with plot renderers in this window
               - columns : int - Number of colums in each row

        Creates a matrix, with the specified dimension, with render windows."""
        
        amount = rows * columns

        # Remove the active renderer, or we will get a hassle with C++
        self.setActiveRenderer(None)

        # Check if we are removing windows 
        if amount < len(self.renderWindowList):

            # The number of windows that will be removed
            diff = len(self.renderWindowList) - amount

            # Open a message box to warn the user of the upcoming removal
            retval = QMessageBox.warning(self,
                                     "SciCraft",
                                     "This operation will delete the last " +
                                     str(diff) +
                                     " render window(s)!",
                                     QMessageBox.Ok,
                                     QMessageBox.Cancel)

            # User cancelled
            if retval == QMessageBox.Cancel:
                return

            # User wants to remove windows
            elif retval == QMessageBox.Ok:
                # Get the windows that should be removed and tell the controller to
                # get rid of them
                for win in self.renderWindowList[-diff:]:
                    # Delete render windows
                    self.controller.removeRenderer(win)

        
        # Beyond the check if we are going to remove windows.
        # Now we want to add the remaining renderers again, and maybe some new ones.
        
        for win in self.renderWindowList:
            # This pathetic little point is necessary to make the reparent method complete
            point = QPoint()
            
            # Take renderers out og the splitterlayout by reparenting the frame underneath
            # each renderer.
            win.parent.reparent(self, point) 
            win.parent.hide()


        # Crude hack to make the existing spiltters go away.
        for splitter in self.lineSplitter.children():
            if isinstance(splitter,QSplitter):
                point = QPoint()
                splitter.reparent(None, point)

        # Add new renderers to our renderWindowList
        while(len(self.renderWindowList) < amount):
            
            frame = QFrame(self, "frame" + str(len(self.renderWindowList) + 1))
            framelayout = QGridLayout(frame, 0, 0, 2, 0, "framelayout" + str(len(self.renderWindowList) + 1))
            render = PlotRenderWindow.PlotRenderWindow(self, self.controller, frame, "renderWindow" + str(len(self.renderWindowList) + 1))
        
            self.renderWindowList.append(render)
            framelayout.addWidget(render, 0, 0)

            # Add a new setup widget for each new renderer
            self.controller.addRenderTab(render.getSetupWidget())

        # Go through the new matrix
        for row in range(rows):

            # Make a colplitter for each row
            self.activeColSplitter = QSplitter(self.lineSplitter)

            # Trying to solve the resizing problem
            # self.lineSplitter.setResizeMode(self.activeColSplitter, QSplitter.Stretch)
            self.activeColSplitter.setOrientation(QSplitter.Horizontal)
            self.activeColSplitter.show()

            # Add the renderers to the new layout
            for col in range(columns):
                point = QPoint()
                render = self.renderWindowList[row*columns+col]
                                
                render.parent.reparent(self.activeColSplitter, point)
                render.parent.show()

        # Move the controller to the bottom of the window.
        self.lineSplitter.moveToLast(self.controller)
       
        # Try not to grow the controller, it is very annoying
        self.lineSplitter.setResizeMode(self.controller, QSplitter.KeepSize)
        # self.lineSplitter.setResizeMode(self.controller, QSplitter.Auto)
        

    
    def addWindowController(self, data = None):
        # The controller needs a reference to this window
        self.controller = PlotWindowControl.PlotWindowControl(self,  self.lineSplitter, self.node, "WindowControl")
        self.controller.mainSetupWidget.page(0).slotAddRenderWindows()
            
    def setActiveRenderer(self, plotRenderer):
        """
        Params:
               - plotRenderer : PlotWindowRenderer - The render windows that should
               be set as the active one.

        This method i fired at every mousePressEvent of the render windows. It will update
        the active renderer to the one in which the mouse is loated.
        """
        if not self.activeRenderer == plotRenderer:
            try:
                self.activeRenderer.parentWidget().unsetPalette()
            except AttributeError:
                pass
            
            self.activeRenderer = plotRenderer

            if not plotRenderer == None:
                plotRenderer.parentWidget().setPaletteBackgroundColor(QColor(255,0,0))

        if not plotRenderer == None:
            self.controller.setActiveTab(self.activeRenderer.getSetupWidget())
        else:
            self.controller.setActiveTab(self.controller.mainSetupWidget.page(0))

    def dummy(self):
        print "You dummy!"

    def initToolbars(self):
        """
        Creates the default menu and toolbaritems for this window. This includes
        actions for saving and printing of the plot."""
        
        # Create a toolbar and a menubar for the window
        self.toolBar = QToolBar(self)
        self.toolBar.setLabel("File actions")
        self.menuBar = QMenuBar(self)
        self.plotToolBar = QToolBar(self)
        self.plotToolBar.setLabel("Plot control actions")
        self.modeToolBar = QToolBar(self)
        self.modeToolBar.setLabel("Interaction")
        self.animToolBar = QToolBar(self)
        self.animToolBar.setLabel("Animation actions")
        self.animToolBar.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.animToolBar.setEnabled(0)
        self.animToolBar.hide()


        # Fill the menubar with popup menus
        self.fileMenu = QPopupMenu(self.menuBar)
        self.menuBar.insertItem(self.tr("&File"),self.fileMenu)

        self.editMenu = QPopupMenu(self.menuBar)
        self.menuBar.insertItem(self.tr("&Edit"), self.editMenu)

        self.viewMenu = QPopupMenu(self.menuBar)
        self.menuBar.insertItem(self.tr("&View"), self.viewMenu)

        self.modeMenu = QPopupMenu(self.menuBar)
        self.menuBar.insertItem(self.tr("&Mode"), self.modeMenu)
        
        self.optionsMenu = QPopupMenu(self.menuBar)
        # self.menuBar.insertItem(self.tr("&Options"), self.optionsMenu)
        
        # Create the actions for the file menu
        self.optionsAction = QAction(self)
        self.optionsAction.setText(self.tr("Options ..."))
        self.optionsAction.setMenuText(self.tr("Options ..."))
        # self.optionsAction.setOn(0)
        self.optionsAction.setToolTip(self.tr("Change the options"))
        self.optionsAction.setWhatsThis(self.tr("Opens a window with controls to adjust the settings of the plot"))
        self.optionsAction.addTo(self.optionsMenu)
        self.connect(self.optionsAction, SIGNAL("activated()"), self.slotOpenOptions)


        self.saveDataAction = QAction(self)
        self.saveDataAction.setIconSet(QIconSet(QPixmap(Icons.savePic)))
        self.saveDataAction.setText(self.tr("Save Plot Configuration ..."))
        self.saveDataAction.setMenuText(self.tr("&Save Plot Configuration ..."))
        # self.saveDataAction.setOn(0)
        # self.saveDataAction.setEnabled(0)
        self.saveDataAction.setEnabled(1)
        self.saveDataAction.setToolTip(self.tr("Save the configuration of the plot window to file..."))
        self.saveDataAction.setWhatsThis(self.tr("Displays a file dialogue and lets you choose a filename to save the plot configuration to. When you restart SciCraft, a new plot will appear in the \"Custom\" menu, and you can use the setup you saved instantly."))
        self.saveDataAction.setAccel(QKeySequence("CTRL+S"))
        self.connect(self.saveDataAction, SIGNAL("activated()"), self.saveConfig)
        self.saveDataAction.addTo(self.fileMenu)
        self.saveDataAction.addTo(self.toolBar)

        self.loadDataAction = QAction(self)
        self.loadDataAction.setIconSet(QIconSet(QPixmap(Icons.openPic)))
        self.loadDataAction.setText(self.tr("Load Plot Configuration ..."))
        self.loadDataAction.setMenuText(self.tr("&Load Plot Configuration ..."))
        self.loadDataAction.setEnabled(1)
        self.loadDataAction.setToolTip(self.tr("Load a plot configuration from file"))
        self.loadDataAction.setWhatsThis(self.tr("Displays a file dialogue and lets you choose a filename to load the plot configuration from."))
        self.loadDataAction.setAccel(QKeySequence("CTRL+L"))
        self.connect(self.loadDataAction, SIGNAL("activated()"), self.loadConfig)
        self.loadDataAction.addTo(self.fileMenu)
        self.loadDataAction.addTo(self.toolBar)

        self.savePlotAction = QAction(self)
        self.savePlotAction.setIconSet(QIconSet(QPixmap(Icons.plotSavePicturePic)))
        self.savePlotAction.setText(self.tr("Save Plot ..."))
        self.savePlotAction.setMenuText(self.tr("S&ave Plot ..."))
        # self.savePlotAction.setOn(0)
        self.savePlotAction.setToolTip(self.tr("Save the plot(s) in the active render window as a picture"))
        self.savePlotAction.setWhatsThis(self.tr("Displays a file dialogue and lets you choose a filename and an image type for the plot to be saved in."))
        self.savePlotAction.setAccel(QKeySequence("CTRL+G"))
        self.connect(self.savePlotAction, SIGNAL("activated()"), self.savePicture)
        self.savePlotAction.addTo(self.fileMenu)
        self.savePlotAction.addTo(self.toolBar)

        self.printPlotAction = ppa = QAction(self)
        ppa.setIconSet(QIconSet(QPixmap(Icons.printPic)))
        ppa.setText(self.tr("Print ..."))
        ppa.setMenuText(self.tr("&Print ..."))
        # ppa.setOn(0)
        ppa.setEnabled(0)
        ppa.setToolTip(self.tr("Print the plot to a printer"))
        ppa.setWhatsThis(self.tr("Displays a print dialogue. This feature is not implemented yet, so if you want to print a plot, save it as a postscript image and print the image from elsewhere (e.g. the command line)."))
        ppa.setAccel(QKeySequence("CTRL+P"))
        ppa.connect(self.printPlotAction, SIGNAL("activated()"), self.dummy)
        ppa.addTo(self.fileMenu)
        ppa.addTo(self.toolBar)

        # Fullscreen
        self.fullscreenAction = fsa = QAction(self)
        fsa.setIconSet(QIconSet(QPixmap(Icons.fullscreenPic)))
        fsa.setText(self.tr("Fullscreen Mode"))
        fsa.setMenuText(self.tr("Fullscreen"))
        fsa.setToggleAction(1)
        fsa.setOn(0)
        
        fsa.setToolTip(self.tr("Fullscreen. Toggle with ALT+Return"))
        fsa.setWhatsThis(self.tr("Expands the plots to cover the whole screen (Removing most of the window decorations). Toggle this mode by clicking ALT+Return"))
        fsa.setAccel(QKeySequence("Alt+Return"))
        fsa.connect(self.fullscreenAction, SIGNAL("toggled(bool)"), self.slotToggleFullscreen)
        fsa.addTo(self.plotToolBar)
        fsa.addTo(self.viewMenu)

        self.view3dAction = v3a = QAction(self)
        v3a.setIconSet(QIconSet(QPixmap(Icons.view3dPic)))
        v3a.setText(self.tr("Stereo Mode"))
        v3a.setMenuText(self.tr("View in stero"))
        v3a.setToggleAction(1)
        v3a.setOn(0)
        v3a.setToolTip(self.tr("Toggle stereo mode"))
        v3a.setWhatsThis(self.tr("Toggle between normal view and 3D view of the plots." +
                         "You need stereo glasses to benefit from this feature."))
        v3a.setAccel(QKeySequence("Ctrl+3"))
        v3a.connect(self.view3dAction, SIGNAL("toggled(bool)"), self.slotToggle3D)

        v3a.addTo(self.plotToolBar)
        v3a.addTo(self.viewMenu)

        self.viewControllerAction = vc = QAction(self)
        vc.setIconSet(QIconSet(QPixmap(Icons.plotControlPic)))
        vc.setText(self.tr("Window Controller"))
        vc.setToggleAction(1)
        vc.setOn(1)
        vc.setToolTip(self.tr("Show/hide window controller."))
        vc.setWhatsThis(self.tr("Shows and hides the plot window controller in the bottom of this window."))
        vc.addTo(self.plotToolBar)
        vc.addTo(self.viewMenu)
        vc.connect(self.viewControllerAction, SIGNAL("toggled(bool)"), self.slotToggleController)
        
        
        
        # Exit window action should be separated
        self.fileMenu.insertSeparator()
        
        self.closeAction = QAction(self)
        self.closeAction.setText("Close Window")
        self.closeAction.setMenuText(self.tr("Close"))
        # self.closeAction.setOn(0)
        self.closeAction.setToolTip(self.tr("Close the window"))
        self.closeAction.setWhatsThis(self.tr("Close the window (obviously)"))
        self.closeAction.setAccel(QKeySequence("CTRL+Q"))
        self.connect(self.closeAction, SIGNAL("activated()"), self, SLOT("close()"))
        
        self.closeAction.addTo(self.fileMenu)

        # Picking actions:

        self.deletePickAction = dpa = QAction(self)
        if not self.controller.isFinished():
            dpa.setEnabled(0)
        self.deletePickAction.setIconSet(QIconSet(QPixmap(Icons.plotDeletePic)))
        dpa.setText("Delete picked elements")
        dpa.setMenuText("Delete picked elements")
        dpa.setToolTip(self.tr("Delete the picked elements in the plots, and set the deletion on the output port."))
        dpa.setWhatsThis(self.tr("Delete all the elements you have picked, and record their positions on the output port of this node. Next time the module diagram is run, the command on the output port will instruct the picked data elements to be removed from the source data if the port is connected to an edit node."))
        dpa.setAccel(QKeySequence("CTRL+X"))

        self.connect(self.deletePickAction, SIGNAL("activated()"), self.controller.deletePickedElements)
        self.plotToolBar.addSeparator()
        self.deletePickAction.addTo(self.plotToolBar)
        self.deletePickAction.addTo(self.editMenu)

        self.undoDeleteAction = uda = QAction(self)
        self.undoDeleteAction.setIconSet(QIconSet(QPixmap(Icons.plotUndoPic)))
        uda.setEnabled(0)
        uda.setText("Undo all picking")
        uda.setMenuText("Undo all picking")
        uda.setToolTip(self.tr("Undo all picking, and put all removed elements back into the plots"))
        uda.setWhatsThis(self.tr("Undo all picking that is done, and put the elements that has been picked and removed back into their plots."))
        uda.setAccel(QKeySequence("CTRL+Z"))
        self.connect(self.undoDeleteAction, SIGNAL("activated()"), self.controller.undoPicking)
        self.undoDeleteAction.addTo(self.plotToolBar)
        self.undoDeleteAction.addTo(self.editMenu)


        # Exclusive mode Actions. Only one of them can be selected at
        # any time, therefore thy are added to a QActionGroup. 
        self.modeGroup = QActionGroup(self, "modeGroup", True)
        
        self.cameraModeAction = cma = QAction(self.modeGroup)
        cma.setIconSet(QIconSet(QPixmap(Icons.viewPic)))
        cma.setEnabled(1)
        cma.setToggleAction(1)
        
        cma.setText(self.tr("Camera mode"))
        cma.setToolTip(self.tr("Rotate, pan and zoom plots."))
        cma.setWhatsThis(self.tr("This mode enables you to rotate, pan and zoom the plots." +
                               "How these actions are performed can be redefined in" +
                               "\"Renderer Options\"."))
        self.connect(self.cameraModeAction,
                     SIGNAL("toggled(bool)"),
                     self.controller.slotCameraModeSelected)
        cma.toggle()
        


        self.selectModeAction = sma = QAction(self.modeGroup)
        sma.setIconSet(QIconSet(QPixmap(Icons.selectBoxPic)))
        sma.setEnabled(1)
        sma.setToggleAction(1)
        sma.setText(self.tr("Select mode"))
        sma.setToolTip(self.tr("Select objects with bounding box."))
        sma.setWhatsThis(self.tr("In this mode the mouse creates a bounding box." +
                               "All objects inside the box will be picked."))
        self.connect(self.selectModeAction,
                     SIGNAL("toggled(bool)"),
                     self.controller.slotSelectModeSelected)
        

        self.pickModeAction = pma = QAction(self.modeGroup)
        pma.setIconSet(QIconSet(QPixmap(Icons.pickPic)))
        pma.setEnabled(1)
        pma.setToggleAction(1)
        pma.setText(self.tr("Pick mode"))
        pma.setToolTip(self.tr("Select individual objects."))
        pma.setWhatsThis(self.tr("This mode lets you select individual objects in " +
                                 "the plots by clicking on them. The same action can " +
                                 "be perfomed outside this mode by pressing the \"P\" key."))
        self.connect(self.pickModeAction,
                     SIGNAL("toggled(bool)"),
                     self.controller.slotPickModeSelected)
        
        self.modeGroup.addTo(self.modeToolBar)
        self.modeGroup.addTo(self.modeMenu)


        
        # self.animSlider.show()

        self.animBackAction = aba = QAction(self)
        aba.setIconSet(QIconSet(QPixmap(Icons.arrowBackPic)))
        aba.setEnabled(1)
        aba.setText(self.tr("Backward"))
        aba.setToolTip(self.tr("Move backwards to previous element i data buffer."))
        aba.setWhatsThis(self.tr("If you have enabled data buffering, you can" +
                                 " use this button to scroll backwards in the history of" +
                                 " this plot node's data."))

        # Connect to buffer pointer
        self.connect(aba, SIGNAL("activated()"), self.controller.filterControl.decreaseBufferIndex)
        
        aba.addTo(self.animToolBar)
        

        self.animForwardAction = afa = QAction(self)
        afa.setIconSet(QIconSet(QPixmap(Icons.arrowForwardPic)))
        afa.setEnabled(1)
        afa.setText(self.tr("Forward"))
        afa.setToolTip(self.tr("Move forwards to next element i data buffer."))
        afa.setWhatsThis(self.tr("If you have enabled data buffering, you can" +
                                 " use this button to scroll forwards to more recent" +
                                 " versions of this plot node's data."))

        # Connect to buffer pointer
        self.connect(afa, SIGNAL("activated()"), self.controller.filterControl.increaseBufferIndex)

        afa.addTo(self.animToolBar)
        self.animSlider = QSlider(Qt.Horizontal, self.animToolBar)
        # self.animSlider.setTickInterval(10)
        self.animSlider.setTickmarks(QSlider.Below)
        # self.animSlider.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        QToolTip.add(self.animSlider, self.tr("Animate data set translation"))
        QWhatsThis.add(self.animSlider, self.tr("Animate the translation between " +
                                                "two data sets by dragging the slider. " +
                                                "By dragging it to the right you will animate" +
                                                " the translation of objects from a " +
                                                " previous data set to the current data set."+
                                                " By dragging it to the left, you will "+
                                                "go from the current data set to the previous one " +
                                                "in the data buffer."))        
        

        # Add the specific plot items before the Help menu
        self.addCustomItems()

        self.helpMenu = QPopupMenu(self.menuBar)
        self.helpAction = QAction(self)
        self.helpAction.setText("Help")
        self.helpAction.setMenuText("Help")
        self.helpAction.setToolTip("Display help pages for plotting")
        self.helpAction.setAccel(QKeySequence("F1"))
        self.connect(self.helpAction, SIGNAL("activated()"), self.slotShowHelp)
        self.helpAction.addTo(self.helpMenu)
        self.menuBar.insertItem("&Help", self.helpMenu)

    def addCustomItems(self):
        """
        Adds the menu and toolbar items specific to the plot
        currently being displayed in this plot window"""

        for plot in self.plotList:
            menus = plot.getMenus()
            items = plot.getToolbarItems()
            toolbar = QToolBar(self)
        
            for tool in items:
                tool.addTo(toolbar)
                        
                for entry in menus:
                    self.menuBar.insertItem(str(entry), menus[entry])

    def savePicture(self, defaultFileName = ""):
        """
        Params:
               - defaultFileName : String - The filename to be secified when the user
               is presentet to the file save dialog. This parameter can very well be left
               unspecified, it has some value when this method is called recurively (rare)

        This method is connected to the "save as picture" button in the plot
        window. It will open a file dialog and try to save the current plot to
        a file specified by the user. The user can choose from several different image formats,
        as well as "by extension". When the latter is chosen, the method will find the image format
        by looking at the extension (file suffix) given by the user. This method is fairly
        robust, and will try to recover from any mistake made by the user. It will a.o. check
        for file name validity and prompt the user before overwriting anything."""

        # The the last directory the user visited, usually the user would
        # want to save here
        lastdir = Config.homedir

        # Check if the last used dir still exists,
        # otherwise continue to use the home directory
        if os.path.isdir(PlotConfig.PlotConfig.lastsavedir):
            lastdir = PlotConfig.PlotConfig.lastsavedir
        
        # The valid image formats gif may be added later
        validSuffices = {".jpeg": "JPEG Image (*.jpg *.jpeg)",
                         ".jpg" : "JPEG Image (*.jpg *.jpeg)",
                         ".bmp" : "Bitmap Picture (*.bmp)",
                         ".png" : "Portable Network Graphics (*.png)",
                         ".pnm" : "Portable Anymap (*.pnm)",
                         ".ps"  : "Encapsulated Postscript (*.ps *.eps)",
                         ".eps" : "Encapsulated Postscript (*.ps *.eps)",
                         ".tif" : "Tag Image File Format (*.tif *.tiff)",
                         ".tiff": "Tag Image File Format (*.tif *.tiff)"}
                        

        # Remember the file selected
        selectedFile = defaultFileName
        
        # The file save dialog
        fileDialog = QFileDialog(lastdir,
                                 str(self.tr("By extension (*)")),
                                 self,
                                 "fileDialog",
                                 1)

        # Add the rest of the filters
                
        fileDialog.addFilter(validSuffices[".bmp"])
        fileDialog.addFilter(validSuffices[".jpg"])
        fileDialog.addFilter(validSuffices[".png"])
        fileDialog.addFilter(validSuffices[".pnm"])
        fileDialog.addFilter(validSuffices[".eps"])
        fileDialog.addFilter(validSuffices[".tif"])
        
        # Set the selected file to the previous seleced file
        fileDialog.setSelection(selectedFile)
        fileDialog.setMode(QFileDialog.AnyFile)
        
        # Remember the last file filter used
        fileDialog.setSelectedFilter(PlotConfig.PlotConfig.plotPictureSaveFilter)
        fileDialog.setCaption("Save plot as picture")
        
        # User has pushed the save button
        if fileDialog.exec_loop() == QDialog.Accepted:
            
            fileFilter = fileDialog.selectedFilter()
            PlotConfig.PlotConfig.plotPictureSaveFilter = fileFilter
            fullFileName = fileDialog.selectedFile()
                        
                
        # User has cancelled saving
        else:
            self.setStatusMessage(self.tr("Save to picture cancelled"), self.statusTime)
            return
        
        # Strip the name from the file path
        filePath, fileName = os.path.split(str(fullFileName))
        PlotConfig.PlotConfig.lastsavedir = filePath
        PlotConfig.writeConfig()

        # Strip away the suffix (if any). Will be used for testing
        prefix, suffix = os.path.splitext(fileName)

        # if user has entered an invalid suffix, stuff
        # it together with the file prefix and let the
        # file filter decide the format
        if not suffix in validSuffices:
            prefix = prefix + suffix
            suffix = ""

        # if the user has not enterd a valid suffix but want to save by an extension,
        # send message to user
        if str(fileFilter) == str(self.tr("By extension (*)")) and not suffix:

            # I'm bored
            message = str(self.tr(" is an invalid image file type."))
            if os.path.splitext(prefix)[1] == ".gif":
                message += str(self.tr(" Future GIF image support is possible and probable."))
            message += "\n"
                        
            retval = QMessageBox.warning(self,
                                    "SciCraft",
                                    fileName +
                                    message +
                                    str(self.tr("Do you want to abort the save or specify a file with a valid type?")),
                                    QMessageBox.Retry,
                                    QMessageBox.Abort,
                                    QMessageBox.NoButton)

            # User aborts the save
            if retval == QMessageBox.Abort:
                self.setStatusMessage(self.tr("Save to picture cancelled"), self.statusTime)
                return
            # User retries. Call this method again
            else:
                self.savePicture(prefix)
                return
                
        # User has entered a valid suffix and save by extension.
        # Find the right filter.
        elif str(fileFilter) == str(self.tr("By extension (*)")):

            # Not necessary to make i QString, but lets keep
            # the type consistent throughout the method
            fileFilter = QString(validSuffices[suffix])

        # Check for the chosen filter, create the appropriate writer
        # and add the correct file suffix to the file
        if str(fileFilter) == validSuffices[".bmp"]:
            writer = vtkBMPWriter()
            if not suffix == ".bmp":
                prefix = prefix + suffix
                suffix = ".bmp"

        elif str(fileFilter) == validSuffices[".jpg"]:
            writer = vtkJPEGWriter()
            if not (suffix == ".jpg" or suffix == ".jpeg"):
                prefix = prefix + suffix
                suffix = ".jpg"

        elif str(fileFilter) == validSuffices[".png"]:
            writer = vtkPNGWriter()
            if not suffix == ".png":
                prefix = prefix + suffix
                suffix = ".png"

        elif str(fileFilter) == validSuffices[".pnm"]:
            writer = vtkPNMWriter()
            if not suffix == ".pnm":
                prefix = prefix + suffix
                suffix = ".pnm"

        elif str(fileFilter) == validSuffices[".eps"]:
            writer = vtkPostScriptWriter()
            if not (suffix == ".eps" or suffix == ".ps"):
                prefix = prefix + suffix
                suffix = ".eps"

        elif str(fileFilter) == validSuffices[".tif"]:
            writer = vtkTIFFWriter()
            if not (suffix == ".tif" or suffix == ".tiff"):
                prefix = prefix + suffix
                suffix = ".tif"
        
        # This should not occur at all
        else:
            QMessageBox.error(fileDialog,
                              "SciCraft",
                              self.tr("Found no matching file filter. This is a BUG! Please file a bug report. The image will not be saved."),
                              QMessageBox.Ok,
                              QMessageBox.NoButton,
                              QMessageBox.NoButton)
            
            return

        
        # Check if file already exists. If it does, prompt user
        # for advice
        if os.path.isfile(os.path.join(filePath, prefix + suffix)):
            retval = QMessageBox.warning(fileDialog,
                                         "SciCraft",
                                         str(self.tr("Overwrite ")) + prefix + suffix + "?",
                                         QMessageBox.Yes,
                                         QMessageBox.No,
                                         QMessageBox.NoButton)


            # User won't overwrite, recall this method
            if retval == QMessageBox.No:
                self.savePicture(prefix + suffix)
                return

        # self.repaint()

        imageFilter = vtkWindowToImageFilter()

        # This line makes the plot renderer rerender the plot before
        # saving it to file. If this call is not made, a picture of the
        # save dialog itself is written to file.
        self.activeRenderer._RenderWindow.Render()

        imageFilter.SetInput(self.activeRenderer._RenderWindow)

        # Just make sure
        imageFilter.Update()
        
        writer.SetInput(imageFilter.GetOutput())
        writer.SetFileName(os.path.join(filePath, prefix + suffix))

        self.activeRenderer.Render()

        # Things could go wrong when writing the file
        try:

            # VTK is not very good at throwing exceptions. To catch some of the
            # errors that could occur when trying to write a file, a test file
            # is first written, by python, in the place of the image file.
            # If this goes allright, vtk will overwrite the file with
            # the actual image file. This method is not bullet proof, and some
            # underlying assumptions must hold if the program should be able to
            # avoid a crash.
            testFile = open(os.path.join(filePath, prefix + suffix), "w")
            testFile.write("SciCraft: Should contain image data! Something is very wrong :(")
            testFile.close()

            # VTK
            writer.Write()

        # This is the exception that the python file handlig would throw
        except IOError, e:

            # Give the user a new set of choices
            retval = QMessageBox.critical(self,
                                         "SciCraft",
                                         "Could not save " + str(fullFileName) + ":" +
                                         e.strerror + "!\n"+
                                         "Do you want to try another location or abort the save?" ,
                                         QMessageBox.Retry,
                                         QMessageBox.Abort,
                                         QMessageBox.NoButton)
            
            # Retry the save, call this method again
            if retval == QMessageBox.Retry:
                # Recursive call
                self.savePicture(prefix + suffix)
                return

            # User gives up 
            else:
                self.setStatusMessage(self.tr("Save to picture cancelled"), self.statusTime)
                return

        self.setStatusMessage(self.tr("Plot saved to " + os.path.join(filePath, prefix + suffix)),
                              self.statusTime)

    def setStatusMessage(self, text, time = 0):
        """
        Params:
               - text : String - The message to be displayed
               - time : int - An optional time limit for the message to be shown.
               If 0 is specified (default) the message will be displayed until another
               message is displayed

        Displays a text message in the plot windows statusbar."""
        
        if time:
            self.statusBar.message(text, time)
        else:
            self.statusBar.message(text)
    
    def slotAddPlotNode(self, index):

        print "Halla!"

    def slotOpenOptions(self):
        """
        Opens the options dialog corresponding to the current plot"""

        # Get the window
        self.window = window = PlotOptionsWindow.PlotOptionsWindow(self, "Options")

        # get the options widget corresponding to the current plot
        optionsWidget = self.plot.getOptions(self)

        # Get all the parent widgets too
        for wid in optionsWidget.getParentWidgets():
            wid.buildWidget()
            window.addTab(wid, wid.name)

        optionsWidget.buildWidget()
        window.addTab(optionsWidget, optionsWidget.name)
        window.show()

    def slotToggleController(self, on):
        """
        Params:
               - on : bool - Whether the controller should be toggled on or off.
        Toggles the visibility of the plot window controller at the bottom
        of the window.
        """
        if on:
            self.controller.show()
        else:
            self.controller.hide()

    def slotToggle3D(self, on):

        for rw in self.renderWindowList:
            rw.toggle3D(on)

    def slotToggleFullscreen(self, on):
        """
        Params:
               - on : bool - whether the fullscreen was toggled on or off
        Toggles the display to be viewed in fullscreen or in normal mode.
        """
        if on:

            # Note that the controller will
            # show up anyway if the MD has not run yet,
            # as there is no point in looking at a blank
            # fullscreen.
            self.viewControllerAction.setOn(0)
            self.toolBar.hide()
            self.menuBar.hide()
            self.statusBar.hide()

            # From qt
            self.showFullScreen()
        else:            
            self.toolBar.show()
            self.menuBar.show()
            self.statusBar.show()

            # From qt
            self.showNormal()

    def closeEvent(self, event):
        """
        Params:
               - event : QCloseEvent? - Event sent from Qt when user clicks the close
               button on the window frame.
        Redefines the closeEvent method in the QMainWindow class. A plot window
        should never be closed as such, but just hidden. This method therefore
        only hides the window when it is instructed to close."""
        
        self.hide()
        
        # QMainWindow.closeEvent(self, event)
        
    def saveConfig(self):
        self.controller.saveConfig()

    def loadConfig(self, path=None):
        self.controller.loadConfig(path)

    def saveNodeDiagram(self):
        self.controller.saveNodeDiagram()

    def slotShowHelp(self):
        viewer = HelpViewer(self, os.path.join(Config.htmldocpath, "node_Plotting.html"), Config.htmldocpath)
        
    def connectAnimator(self,animator,slot):
        self.connect(self.animSlider, SIGNAL("valueChanged(int)"), slot)
