# -*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import PYSIGNAL, QFileDialog, QMessageBox, QSize, QSizePolicy
from qt import QTabWidget, QVBoxLayout, QWidget, SIGNAL, qApp
from GUI.Plots import PlotWindowSetupWidget
from GUI.Plots import DataFilterController
from Library.Util.Config import Config
from GUI.Plots.interaction.SelectBoxMode import SelectBoxMode
from GUI.Plots.interaction.CameraMode import CameraMode
from GUI.Plots.interaction.PickMode import PickMode
import types
import os
import os.path

class PlotWindowControl(QWidget):
    """Class for setting up the plot window to the users specification"""
    def __init__(self, gui, parent, node, name = None):
        QWidget.__init__(self, parent, name, 0)

        if not name:
            self.setName("plotControl")

        #self.setResizeEnabled(1)
        #self.setMovingEnabled(0)
        
        # self.centralLayout = QVBoxLayout(self.widget(), 11, 6, "centralLayout")
        
        self.gui = gui
        self.finished = 0
        self.filterControl = DataFilterController.DataFilterController(self)
        self.outputData = []
        self.node = node
        
        self.plotControlLayout = QVBoxLayout(self,6,6,"plotControlLayout")

        # users this to remember what were removed from last plot
        # Torbjørn stuff.
        self.lastRemoved = []
        
        # Tab widget containing the control to all plot renderes
        # tab widgets
        self.mainSetupWidget = QTabWidget(self, "mainSetupWidget")
        
        self.plotControlLayout.addWidget(self.mainSetupWidget)

        self.connect(self.mainSetupWidget, SIGNAL("currentChanged(QWidget *)"), self.slotUpdateActiveRenderer)
        self.addSetupTab()
        #                                                 Fixed before
        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        self.resize(QSize(0,0).expandedTo(self.minimumSizeHint()))
        # self.show()
        #self.loadNodeDiagram() #do this in gui instead

    def sendOutputData(self):
        """Sends the output data to the Edit GUINode connected to this node
        I can't count all the pitfalls in this method..."""

        portList = []

        inputPorts = self.node.getInputPorts()

        allnodes = self.node.get_module_diagram().get_nodes()

        for node in allnodes:
            typ = node.getNodeType()
            if typ == "EditNode":

                connectionPairs = node.get_connections_to_node(self.node)
                if connectionPairs:
                    parameterList = node.getParameters()

                    for param in parameterList:
                        if str(param.getName()) == "ignoreThisParameter4ever":
                            param.setValue(self.outputData)


        # Torbjørn stuff
        for thing in self.outputData:
            self.lastRemoved.append(thing)
        
        # tells anyone interested that pick indeces are available
        self.emit(PYSIGNAL("pickComplete"),())

        self.outputData = []

    # Torbjørn stuff
    def getLastRemoved(self):
        
        return self.lastRemoved
    
    def setFinished(self, finished):
        self.finished = finished

        # If plot has run, the ports can no longer be manipulated
        # This operation may be dangerous. The user has no way of doing anything
        # with the ports after the MD has run.
        if self.isFinished():
            self.mainSetupWidget.page(0).enablePorts(0)
            # self.sendOutputData()
            self.gui.deletePickAction.setEnabled(1)
        else:
            self.mainSetupWidget.page(0).enablePorts(1)
            self.gui.deletePickAction.setEnabled(0)

    def isFinished(self):
        return self.finished
    
    def addSetupTab(self):
        """Adds a control tab at the beginning"""

        # ctrl = QWidget(self.mainSetupWidget)
        ctrl = PlotWindowSetupWidget.PlotWindowSetupWidget(self, self.mainSetupWidget, self.node)
        self.mainSetupWidget.addTab(ctrl, "Plot Setup")
        # ctrl.slotAddRenderWindows()

    def getSetupTab(self):

        return self.mainSetupWidget.page(0)

    def addRenderTab(self, renderTab):
        """Inserts a setup tab to a newly created plot render window """
        self.mainSetupWidget.insertTab(renderTab, self.tr("Window " + str(len(self.gui.renderWindowList))))

    def setActiveTab(self, renderTab):
        """Updates the selected tab to reflect the slected plot render window """
        self.mainSetupWidget.showPage(renderTab)

    def showErrorMessage(self, message):

        retval = QMessageBox.warning(self,
                                     "SciCraft",
                                     str(message),
                                     QMessageBox.Ok,
                                     QMessageBox.NoButton)

    def resizeController(self, glob = False):

        if glob:
            # print "glob glob glob..."
            
            self.resize(QSize(10, 10).expandedTo(self.minimumSizeHint()))
            # self.gui.setMinimumSize(self.gui.
            self.gui.resize(QSize(10, 10).expandedTo(self.gui.minimumSizeHint()))
        else:
            self.resize(QSize(10, 10).expandedTo(self.minimumSizeHint()))
            

    def slotUpdateActiveRenderer(self, tab):
        """Updates the active plot render window when a tab is selected """

        if not tab == self.mainSetupWidget.page(0):
            self.gui.setActiveRenderer(tab.renderWindow)

    def removeRenderer(self, renderWindow):

        # Remove setup tab from tab widget
        self.mainSetupWidget.removePage(renderWindow.getSetupWidget())

        
        self.gui.renderWindowList.remove(renderWindow)

        renderWindow.parent.close(1)

    def getInputPorts(self):

        return self.mainSetupWidget.page(0).guiPorts.keys()
    
    def updateInputPorts(self, ports):

        for setupWidget in range(1, self.mainSetupWidget.count()):
            self.mainSetupWidget.page(setupWidget).updateInputPorts(ports)

    def apply_all_plot_settings(self):
        for setupWidget in range(1, self.mainSetupWidget.count()):
            self.mainSetupWidget.page(setupWidget).apply_all_plot_settings()

    def redrawPlots(self):
        self.filterControl.updateData(self.gui.getData())
        for setupWidget in range(1, self.mainSetupWidget.count()):
            self.mainSetupWidget.page(setupWidget).redrawPlots()

    def saveConfig(self):
        # legg til info om porter og nxn oppsett
        config = ""
        config += "<plotconfig>\n"
        setupWidget = self.mainSetupWidget.page(0)
        config += "<rows>" + str(setupWidget.rows) + "</rows>\n"
        config += "<cols>" + str(setupWidget.cols) + "</cols>\n"
        ports = setupWidget.guiPorts
        for portname in ports:
            config += "<port>\n"
            config += "\t<name>" + str(portname) + "</name>\n"
            config += "\t<type>" + str((ports[portname])[1]) + "</type>\n"
            config += "\t<desc>" + str((ports[portname])[2]) + "</desc>\n"
            config += "</port>\n"
        for render in self.gui.renderWindowList:
            setupWidget = render.getSetupWidget()
            config += setupWidget.getConfig()
        config += "</plotconfig>\n"

        custompath = os.path.join(Config.userconfigdir,"Plots")
        if not os.path.isdir(custompath) :
            os.mkdir(custompath, 0755)
        custompath += os.sep + "Custom"
        if not os.path.isdir(custompath) :
            os.mkdir(custompath, 0755)
        
        filename = QFileDialog.getSaveFileName(custompath,
                                               "SciCraft plot files (*.zpl)",
                                               self,
                                               "save file dialog",
                                               "Choose a file").latin1()
        if filename:
            if filename.rfind(".zpl") < 0:
                filename += ".zpl"
            file = open(filename, 'w+')
            file.write(config)
            self.gui.setStatusMessage("Config written to " + filename, self.gui.statusTime)

            # Added by oyvindkn. Try to save a pixmap file also
            # Won't work in most cases, but I don't care.
            try:                
                pixfile = os.path.basename(filename)
                pixfile = os.path.splitext(pixfile)[0] + ".txt"
                f = open(os.path.join("GUI",
                                      "conf",
                                      pixfile),
                         'w')
                f.write("type=" + str(os.path.splitext(pixfile)[0]) + "\n")
                f.write("image=3D_in\n")
                f.write("point1=[0,20]:[-10,20] \n")
                f.close()
            except Exception, e:
                pass
        else:
            self.gui.setStatusMessage("Save operation cancelled", self.gui.statusTime)

    def getAvailableAliases(self):
        return self.filterControl.getAliasData()

    def getRenderWindows(self):

        renList = []

        for index in range(1, self.mainSetupWidget.count()):
            
            renList.append(self.mainSetupWidget.page(index).renderWindow)

        return renList

    def getRenderSetupWidgets(self):

        widList = []
        for index in range(1, self.mainSetupWidget.count()):
            widList.append(self.mainSetupWidget.page(index))

        return widList
        
        
    def slotCameraModeSelected(self, activated):
        """
        Params:
               - activated : boolean - whetet the action is activated or not.
        Slot fired when user selects Camera mode."""
        if activated:
            
            for setupW in self.getRenderSetupWidgets():
                rw = setupW.renderWindow
                rw.setActiveMode(CameraMode(rw))

    def slotSelectModeSelected(self, activated):
        if activated:
            for setupW in self.getRenderSetupWidgets():
                rw = setupW.renderWindow
                rw.setActiveMode(SelectBoxMode(rw, setupW.pickWithSelectBox))

    def slotPickModeSelected(self, activated):
        if activated:
            for setupW in self.getRenderSetupWidgets():
                rw = setupW.renderWindow
                rw.setActiveMode(PickMode(rw))

    
        
    def loadConfig(self, filename=None):
        custompath = os.path.join(Config.userconfigdir, "Plots")
        if not os.path.isdir(custompath) :
            os.mkdir(custompath, 0755)
        custompath += os.sep + "Custom"
        if not os.path.isdir(custompath) :
            os.mkdir(custompath, 0755)

        if not filename:
            filename = str(QFileDialog.getOpenFileName(custompath,
                                                       "SciCraft plot files (*.zpl)",
                                                       self,
                                                       "load file dialog",
                                                       "Choose a file"))
        
        if len(filename) != 0 and filename.rfind(".zpl") > 0:
            from xml.dom import minidom
            dom = minidom.parse(filename)
            lamb = xml2lambda([dom]).plotconfig

            setupWidget = self.mainSetupWidget.page(0)
            self.gui.addRenderWindows(int(lamb.rows), int(lamb.cols))
            setupWidget.rows = int(lamb.rows)
            setupWidget.cols = int(lamb.cols)
            setupWidget.rowCombo.setCurrentItem(int(lamb.rows) - 1)
            setupWidget.colCombo.setCurrentItem(int(lamb.cols) - 1)
            
            if lamb.__dict__.has_key("port"):
                if not type(lamb.port) == types.ListType:
                    lamb.port = [lamb.port]
                for port in lamb.port:
                    setupWidget.nameEdit.setText(str(port.name))
                    setupWidget.descriptionEdit.setText(str(port.desc))
                    setupWidget.typeEdit.setText(str(port.type))
                    setupWidget.slotAddPort()

            i = 0
            if lamb.__dict__.has_key("renderer"):
                if not type(lamb.renderer) == types.ListType:
                    lamb.renderer = [lamb.renderer]
                for renderer in lamb.renderer:
                    setupWidget = self.gui.renderWindowList[i].getSetupWidget()
                    setupWidget.setConfig(renderer)
                    i+=1
        
            
            self.gui.setStatusMessage("Config loaded from " + filename, self.gui.statusTime)
        else:
            self.gui.setStatusMessage("Load operation cancelled", self.gui.statusTime)
        

    def deletePickedElements(self):

        if not self.mainSetupWidget.page(0).hasSetAlias():

            retval = QMessageBox.warning(self,
                                         "SciCraft",
                                         "You have not set any aliases! Deleting picked elements will\n" +
                                         "have no effect on any data. The elements will just be removed \n" +
                                         "from the plot. Set up aliases in the \"Plot Setup\" tab in the\n" +
                                         "controller if you want to remove the picked elements from a data\n set. " +
                                         "Do you still want to remove the picked elements from the plot?",
                                         QMessageBox.Yes,
                                         QMessageBox.No)

            if retval == QMessageBox.No:
                return

        self.gui.undoDeleteAction.setEnabled(1)

        for index in xrange(1, self.mainSetupWidget.count()):

            setup = self.mainSetupWidget.page(index)
            self.outputData += setup.deletePickedElements()

        self.sendOutputData()
        
        

    def undoPicking(self):
        self.gui.undoDeleteAction.setEnabled(0)
        for index in xrange(1, self.mainSetupWidget.count()):
            
            setup = self.mainSetupWidget.page(index)
            setup.undoPicking()
        self.outputData = []
        self.lastRemoved = []
        self.sendOutputData()
        self.emit(PYSIGNAL("pickReset"),())

    def toggleBuffering(self, on):
        onText = "on"
        if not on:
            self.gui.animToolBar.setEnabled(0)
            self.gui.animToolBar.hide()
            onText = "off"
        else:
            self.gui.animToolBar.setEnabled(1)
            self.gui.animToolBar.show()
        
    def changeBufferSize(self, newSize):
        self.filterControl.setBufferSize(newSize)
    
    def __tr(self,s,c = None):
        return qApp.translate("plotControl",s,c)

    def saveNodeDiagram(self):
        config = ""
        config += "<plotconfig>\n"
        setupWidget = self.mainSetupWidget.page(0)
        config += "<rows>" + str(setupWidget.rows) + "</rows>\n"
        config += "<cols>" + str(setupWidget.cols) + "</cols>\n"
        # ports aren't needed when we save the whole module diagram
        for render in self.gui.renderWindowList:
            setupWidget = render.getSetupWidget()
            config += setupWidget.getConfig()
        config += "</plotconfig>\n"
        #print "Setting custom XML:", config
        if not self.node.setCustomXML(config):
            print "Failed setting custom XML", config
        else:
            pass

    def loadNodeDiagram(self):
        lambdaish = self.node.getCustomXMLasLambda()
        
        if lambdaish:
            lamb = lambdaish.plotconfig
            setupWidget = self.mainSetupWidget.page(0)
            self.gui.addRenderWindows(int(lamb.rows), int(lamb.cols))
            setupWidget.rows = int(lamb.rows)
            setupWidget.cols = int(lamb.cols)
            setupWidget.rowCombo.setCurrentItem(int(lamb.rows) - 1)
            setupWidget.colCombo.setCurrentItem(int(lamb.cols) - 1)
        
            i = 0
            if lamb.__dict__.has_key("renderer"):
                if not type(lamb.renderer) == types.ListType:
                    lamb.renderer = [lamb.renderer]
                for renderer in lamb.renderer:
                    setupWidget = self.gui.renderWindowList[i].getSetupWidget()
                    setupWidget.setConfig(renderer)
                    i+=1
        
            self.gui.setStatusMessage("Config loaded" , self.gui.statusTime)
        else:
            #print "Could not load custom xml!"
            pass
        
    
        
def xml2lambda(elementlist):

    def handleElement(element):
            
        
            if hasattr(element,"data"):
                # element is a textnode. Return text right away
                return element.data

            stringValue = ""
            returnvalue = lambda : None

            # check each child this element has
            for child in element.childNodes:

                # iterate this over this element                
                value = handleElement(child)

                # check if checked node was a textNode
                if child.nodeType == child.TEXT_NODE:
                    stringValue = value

                elif hasattr(returnvalue,str(child.localName)):
                    # the node already has checked an occurence of this childNode

                    # get the attribute
                    attr = getattr(returnvalue,str(child.localName))

                    # if not list, make it a list
                    if not isinstance(attr,list):
                        attr = [attr]

                    # add the value, and put it back on the returnvalue object
                    attr.append(value)
                    setattr(returnvalue,str(child.localName),attr)

                else:
                    # new node, just adding the variable to the dict
                    setattr(returnvalue,str(child.localName),value)

            if returnvalue.__dict__:
                # object had non-text childNodes. Return parsed input from these
                return returnvalue
            else:
                # object had no non-text childNodes, return text node if any, else return empty string
                return stringValue


    # if element exists, parse its first element
    if elementlist:
                
            # call recursive method.
            return handleElement(elementlist[0])

    # no object to play with, returing nothing :(
    return None     
 

