# -*- coding: utf-8 -*-

# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QButtonGroup, QCheckBox, QComboBox, QGridLayout, QGroupBox
from qt import QHBoxLayout, QKeySequence, QLabel, QLineEdit, QListBox
from qt import QMessageBox, QPushButton, QSize, QSizePolicy, QSpacerItem
from qt import QSpinBox, QTextEdit, QToolTip, QWhatsThis, QWidget, Qt, SIGNAL
from qt import qApp
from GUI.Plots import AliasWidget

class PlotWindowSetupWidget(QWidget):
    def __init__(self, windowController, parent, node, name = None):
        QWidget.__init__(self,parent,name,0)

        if not name: 
           self.setName("PlotWindowSetupWidget")

        #                                             row col margin space
        PlotWindowSetupWidgetLayout = QGridLayout(self, 2, 2, 3, 3, "PlotWindowSetupWidgetLayout")
        PlotWindowSetupWidgetLayout.setColStretch(0, 10)
        self.windowController = windowController
        self.node = node
        
        # Group box for ports
        self.portsGroupBox = QGroupBox(self,"portsGroupBox")
        self.portsGroupBox.setTitle(self.__tr("Ports-Input"))
        self.portsGroupBox.setEnabled(1)
        self.portsGroupBox.setColumnLayout(0,Qt.Vertical)
        self.portsGroupBox.layout().setSpacing(6)
        self.portsGroupBox.layout().setMargin(6)
        self.portsGroupBox.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        
        portsGroupBoxLayout = QGridLayout(self.portsGroupBox.layout())
        portsGroupBoxLayout.setAlignment(Qt.AlignTop)

        # Port list box
        self.portList = QListBox(self.portsGroupBox,"portList")
        self.connect(self.portList, SIGNAL("highlighted(const QString &)"), self.slotPortSelected)
        portsGroupBoxLayout.addMultiCellWidget(self.portList,0,2,2,2)

        # Layout for buttons
        portsButtonLayout = QHBoxLayout(None,0,6,"portsButtonLayout")
        

        # Ports buttons
        self.aliasButton = QPushButton(self.portsGroupBox, "aliasButton")
        self.aliasButton.setText(self.__tr("Set Aliases..."))
        portsButtonLayout.addWidget(self.aliasButton)
        self.connect(self.aliasButton, SIGNAL("clicked()"), self.slotOpenAliases)

        spacer = QSpacerItem(38,20,QSizePolicy.Expanding,QSizePolicy.Minimum)
        portsButtonLayout.addItem(spacer)
        
        self.addButton = QPushButton(self.portsGroupBox,"addButton")
        self.addButton.setText(self.__tr("Add"))
        QToolTip.add(self.addButton, self.__tr("Add a new port with the specified name, type and description."))
        self.addButton.setAccel(QKeySequence("Return"))
        portsButtonLayout.addWidget(self.addButton)
        self.connect(self.addButton, SIGNAL("clicked()"), self.slotAddPort)

        self.updateButton = QPushButton(self.portsGroupBox,"updateButton")
        self.updateButton.setText(self.__tr("Update"))
        QToolTip.add(self.updateButton, self.__tr("Update the selected port to the specified name, type and description."))
        portsButtonLayout.addWidget(self.updateButton)
        self.connect(self.updateButton, SIGNAL("clicked()"), self.slotUpdatePort)

        self.deleteButton = QPushButton(self.portsGroupBox,"deleteButton")
        self.deleteButton.setText(self.__tr("Delete"))
        QToolTip.add(self.deleteButton, self.__tr("Delete the selected port."))
        self.deleteButton.setAccel(QKeySequence("Del"))
        portsButtonLayout.addWidget(self.deleteButton)
        self.connect(self.deleteButton, SIGNAL("clicked()"), self.slotDeletePort)

        portsGroupBoxLayout.addMultiCellLayout(portsButtonLayout,3,3,0,2)

        # Edits for ports
        self.nameLabel = QLabel(self.portsGroupBox,"nameLabel")
        self.nameLabel.setText(self.__tr("Name:"))
        portsGroupBoxLayout.addWidget(self.nameLabel,0,0)
        self.nameEdit = QLineEdit(self.portsGroupBox,"nameEdit")
        portsGroupBoxLayout.addWidget(self.nameEdit,0,1)
                
        self.typeLabel = QLabel(self.portsGroupBox,"typeLabel")
        self.typeLabel.setText(self.__tr("Type:"))
        portsGroupBoxLayout.addWidget(self.typeLabel,1,0)
        self.typeEdit = QLineEdit(self.portsGroupBox,"typeEdit")
        portsGroupBoxLayout.addWidget(self.typeEdit,1,1)

        self.descriptionLabel = QLabel(self.portsGroupBox,"descriptionLabel")
        self.descriptionLabel.setText(self.__tr("Description:"))
        self.descriptionLabel.setAlignment(QLabel.AlignTop)
        portsGroupBoxLayout.addWidget(self.descriptionLabel,2,0)
        self.descriptionEdit = QTextEdit(self.portsGroupBox,"descriptionEdit")
        self.descriptionEdit.setSizePolicy(QSizePolicy(QSizePolicy.Preferred,
                                                       QSizePolicy.Expanding, 0, 0,
                                                       self.descriptionEdit.sizePolicy().hasHeightForWidth()))
        self.descriptionEdit.setWrapColumnOrWidth(90)
        portsGroupBoxLayout.addWidget(self.descriptionEdit,2,1)
        
        PlotWindowSetupWidgetLayout.addMultiCellWidget(self.portsGroupBox, 0, 1, 0, 0)

        # Render windows group box
        self.renderGroupBox = QButtonGroup(self,"renderGroupBox")
        self.renderGroupBox.setTitle(self.__tr("Render windows"))
        # QToolTip.add(self.renderGroupBox,self.__tr("Number of render windows"))
        QWhatsThis.add(self.renderGroupBox,self.__tr("Specifies the number of rows and columns of render windows in this plot window. E.g. with rows=2 and colums=2 you will get 2×2=4 render windows to put your plots into."))
        self.renderGroupBox.setColumnLayout(0,Qt.Vertical)
        self.renderGroupBox.layout().setSpacing(6)
        self.renderGroupBox.layout().setMargin(6)
        self.renderGroupBox.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        renderGroupBoxLayout = QGridLayout(self.renderGroupBox.layout())
        renderGroupBoxLayout.setAlignment(Qt.AlignTop)
        
        self.colLabel = QLabel(self.renderGroupBox,"colLabel")
        self.colLabel.setText(self.__tr("Columns:"))        
        renderGroupBoxLayout.addMultiCellWidget(self.colLabel,1,1,0,1)

        self.colCombo = QComboBox(0,self.renderGroupBox,"colCombo")
        QToolTip.add(self.colCombo,self.__tr("Columns of render windows"))
        self.colCombo.clear()
        self.colCombo.insertItem(self.__tr("1"))
        self.colCombo.insertItem(self.__tr("2"))
        self.colCombo.insertItem(self.__tr("3"))
        renderGroupBoxLayout.addWidget(self.colCombo,1,2)

        self.rowLabel = QLabel(self.renderGroupBox,"rowLabel")
        self.rowLabel.setText(self.__tr("Rows:"))
        renderGroupBoxLayout.addMultiCellWidget(self.rowLabel,0,0,0,1)

        self.rowCombo = QComboBox(0,self.renderGroupBox,"rowCombo")
        self.rowCombo.clear()
        self.rowCombo.insertItem(self.__tr("1"))
        self.rowCombo.insertItem(self.__tr("2"))
        self.rowCombo.insertItem(self.__tr("3"))
        QToolTip.add(self.rowCombo,self.__tr("Rows of render windows"))

        renderGroupBoxLayout.addWidget(self.rowCombo,0,2)
        
        spacer_2 = QSpacerItem(20,10,QSizePolicy.Minimum,QSizePolicy.Expanding)
        renderGroupBoxLayout.addItem(spacer_2,2,2)

        self.applyButton = QPushButton(self.renderGroupBox,"applyButton")
        self.applyButton.setText(self.__tr("Update"))
        QToolTip.add(self.applyButton, self.__tr("Update this plot window to contain the specified number of render windows."))
        self.connect(self.applyButton, SIGNAL("clicked()"), self.slotAddRenderWindows)

        renderGroupBoxLayout.addMultiCellWidget(self.applyButton,3,3,1,2)
        spacer_3 = QSpacerItem(42,20,QSizePolicy.Expanding,QSizePolicy.Minimum)
        renderGroupBoxLayout.addItem(spacer_3,3,0)
        PlotWindowSetupWidgetLayout.addWidget(self.renderGroupBox, 0, 1)

        ## Buffer group

        self.bufferGroupBox = QGroupBox(self)
        self.bufferGroupBox.setEnabled(0)
        self.bufferGroupBox.setTitle("Buffer data")        
        self.bufferGroupBox.setColumnLayout(0, Qt.Horizontal)
        self.bufferGroupBox.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        
        self.bufferLayout = QGridLayout(self.bufferGroupBox.layout(), 2, 2, 3)

        self.bufferLabel = QLabel(self.bufferGroupBox)
        self.bufferLabel.setText("Buffer incoming data:")
        QToolTip.add(self.bufferLabel, self.__tr("Stores the data between runs so that you can scroll forwards and backwards in history."))
        QWhatsThis.add(self.bufferLabel, self.__tr("Stores all the data coming to the input ports og this node in a buffer. This enables you to scroll forwards and backwards in history and view the data recieved in prevoius runs."))
        self.bufferLayout.addWidget(self.bufferLabel, 0, 0)

        self.bufferCheck = QCheckBox(self.bufferGroupBox)
        self.bufferCheck.setText("")
        self.bufferCheck.setDown(False)
        self.connect(self.bufferCheck,
                     SIGNAL("toggled(bool)"),
                     self.slotToggleBuffering)
        
        self.bufferLayout.addWidget(self.bufferCheck, 0, 1)

        self.limitLabel = QLabel(self.bufferGroupBox)
        self.limitLabel.setEnabled(0)
        self.limitLabel.setText("Buffer limit:")
        QToolTip.add(self.limitLabel, self.__tr("The maximum length of the buffer."))
        QWhatsThis.add(self.limitLabel, self.__tr("Sets the maximum length of the buffer. If you run the module diagram more times than the length of the buffer, the oldest data in the buffer will be delted to give space to the data coming from the most recent runs."))
        self.bufferLayout.addWidget(self.limitLabel, 1, 0)

        self.limitSpin = QSpinBox(self.bufferGroupBox)
        self.limitSpin.setEnabled(0)
        self.limitSpin.setMinValue(2)
        self.limitSpin.setMaxValue(10)
        self.connect(self.limitSpin,
                     SIGNAL("valueChanged(int)"),
                     self.slotChangeBufferSize)
        
        self.bufferLayout.addWidget(self.limitSpin, 1, 1)

        #Torbjørn stuff
        # self.bufferCheck.animateClick()
        PlotWindowSetupWidgetLayout.addWidget(self.bufferGroupBox, 1, 1)

        self.resize(QSize(10,10).expandedTo(self.minimumSizeHint()))

        # {nodename : [port, type, description, new?]}
        self.guiPorts = {}
        self.aliases = {}

        # Get the ports from middleware
        self.initPorts()
        # self.disablePorts()

    def slotToggleBuffering(self, on):

        if on:
            self.limitLabel.setEnabled(1)
            self.limitSpin.setEnabled(1)
        else:
            self.limitLabel.setEnabled(0)
            self.limitSpin.setEnabled(0)
        
        self.windowController.toggleBuffering(on)

    def slotChangeBufferSize(self, newSize):

        self.windowController.changeBufferSize(newSize)
        
    def __tr(self,s,c = None):
        return qApp.translate("PlotWindowSetupWidget",s,c)

    def initPorts(self):
        """
        Initializes the ports of this node by fetching the nodeports from
        the middleware and inserting them into the list of ports"""

        inPorts = self.node.getInputPorts()

        for port in inPorts:
            name = port.getName()
            self.guiPorts[name] = [port, port.getDataType(), port.getDescription(), 0]

            # Init the aliases ports list.
            self.aliases[name] = []
        for name in self.guiPorts:
            self.portList.insertItem(name)
            
            
    
    def slotAddRenderWindows(self):
        
        self.rows = 1
        self.cols = 1

        try:
            self.rows = int(str(self.rowCombo.currentText()))
        except ValueError:
            self.rows = 1

        try:
            self.cols = int(str(self.colCombo.currentText()))
        except ValueError:
            self.cols = 1

        
        self.windowController.gui.addRenderWindows(self.rows, self.cols)
        # self.resize(QSize(460,205).expandedTo(self.minimumSizeHint()))

    def cleanPortFields(self):
        """
        Sets the text of all the fields in the ports group box to nothing"""
                
        self.nameEdit.setText("")
        self.typeEdit.setText("")
        self.descriptionEdit.setText("")
        
    def slotAddPort(self):
        """
        Adds a user specified input port to the plot node."""

        # Remove whitespace as it only causes trouble
        portName = self.nameEdit.text().stripWhiteSpace()

        # If the name of the port is empty do nothing interesting
        if portName.isEmpty():
            self.cleanPortFields()
            
        else:
            #### Add the port to the plot node in middleware here!!!!
            
            listItem = self.portList.findItem(portName, Qt.ExactMatch)
            if listItem:
                retval = QMessageBox.warning(self,
                                             "SciCraft",
                                             "There exists a port \"" +
                                             str(portName) + "\" already. Enter a new name and press \"Add\" again.",
                                             QMessageBox.Ok)
                return

            # Going to get the new port, note the type is not sent down to the middleware at this point!
            port = self.node.create_input_port(str(portName), "", str(self.descriptionEdit.text()))
            port.setDataType(str(self.typeEdit.text()))
            
            self.guiPorts[str(portName)] = [port, str(self.typeEdit.text()), str(self.descriptionEdit.text()),1]
            self.aliases[str(portName)] = []
            self.portList.insertItem(portName)
            self.windowController.updateInputPorts(self.guiPorts.keys())
            self.cleanPortFields()
            # self.windowController.setFinished(0)
        
    def slotUpdatePort(self):
        """
        Slot fired when the user presses the update button. Updates the port selected
        in the port list with the name and attruibuts currently in the edit fields. The canges are
        communicated both to the middleware and the window controller in order to update all objects
        dependent upon up to date information about all available ports"""
        
        oldPortName = self.portList.currentText()
        
        portName = self.nameEdit.text().stripWhiteSpace()

        if not portName.isEmpty() and not oldPortName.isEmpty():
                        
            oldPortName = str(oldPortName)

            listItem = self.portList.findItem(str(portName), Qt.ExactMatch)
            
            if listItem and not listItem == self.portList.item(self.portList.currentItem()):
                retval = QMessageBox.warning(self,
                                             "SciCraft",
                                             "There exists a port \"" +
                                             str(portName) + "\" already. Enter a new name and press \"Update\" again.",
                                             QMessageBox.Ok)
                return

            # The attribute list of the port
            attrList = self.guiPorts[oldPortName]
            aliasList = self.guiPorts[oldPortName]
            
            # Updating the middleware
            attrList[0].setName(str(portName))
            attrList[0].setDataType(str(self.typeEdit.text()))
            attrList[0].setDescription(str(self.descriptionEdit.text()))

            # Add the new one to the guiPorts dictionary
            self.guiPorts[str(portName)] = [attrList[0],
                                            str(self.typeEdit.text()),
                                            str(self.descriptionEdit.text()),
                                            0]

            self.aliases[str(portName)] = aliasList
            
            self.portList.changeItem(portName, self.portList.currentItem())
                        
            # self.cleanPortFields()
            
            # Delete the old key
            if not oldPortName == str(portName):
                del self.guiPorts[oldPortName]
                del self.aliases[oldPortName]
            
            # make sure all the setup widgets that use ports can update the changes
            self.windowController.updateInputPorts(self.guiPorts.keys())
        
    def slotDeletePort(self):
        """Deletes a port from the list of ports, the middleware and lists available
        to all the setupWidgets """
        
        portName = str(self.portList.currentText())

        if portName: 
            self.guiPorts[portName][0].remove()
            
            del self.guiPorts[portName]
            del self.aliases[portName]
            self.portList.removeItem(self.portList.currentItem())

            self.cleanPortFields()

            self.windowController.updateInputPorts(self.guiPorts.keys())

    def slotPortSelected(self, text):
        """Slot fired when a user clicks a port in the plot list. The edit fields is updated with
        the port's atrributes (name, type etc.)"""
        
        portName = str(text)
        attr = self.guiPorts[portName]

        self.nameEdit.setText(portName)
        self.typeEdit.setText(attr[1])
        self.descriptionEdit.setText(attr[2])

    def hasSetAlias(self):
        for key in self.aliases:

            if self.aliases[key]:
                return 1

        return 0
        

    def slotOpenAliases(self):
        al = AliasWidget.AliasWidget(self,
                                     None,
                                     Qt.WType_Dialog|
                                     Qt.WShowModal)
                
        al.setPorts(self.aliases.keys())
        al.show()
        # al.update()
        al.initAliasList()

    def getPortAliases(self, portName):

        return self.aliases[portName]

    def addPortAlias(self, portName, aliasName):
        
        if self.aliases.has_key(portName):
            self.aliases[portName].append(aliasName)

    def removePortAlias(self, portName, aliasName):        
        if self.aliases.has_key(portName):
            self.aliases[portName].remove(aliasName)
            
    def enablePorts(self, enable = 1):
        pass
        # self.addButton.setEnabled(enable)
        # self.updateButton.setEnabled(enable)
        # self.deleteButton.setEnabled(enable)
        # self.aliasButton.setEnabled(not enable)
