# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from vtk import vtkCoordinate, vtkMath, vtkTransform
from qt import QWidget
from GUI.Plots.Plots3D.Plot3D import Plot3D
from GUI.Plots.interaction.NullMode import NullMode
from GUI.Plots.interaction.SelectBoxMode import SelectBoxMode
from MolPlot3DSetup import MolPlot3DSetup
import time
import copy
import math
from GUI.Plots.SimplePlotDataFilter import SimplePlotDataFilter

class MolPlot3D(Plot3D):

    BNSPLOT = 0
    SFPLOT = 1
    LPLOT = 2

    def __init__(self, setupWidgetClass=MolPlot3DSetup):
        Plot3D.__init__(self)

        self.plotType = "Molecule"
        self.setupWidget = None
        self.dataFilter = SimplePlotDataFilter()
        self.molPlotType = self.BNSPLOT
        self.picked = []
        self.showIndex = False
        self.showName = False
        self.hiddenActors = []
        self.models = []
        self.bondActors = {}
        self.setupWidgetClass = setupWidgetClass
        self.originalPositions = {} # used in moving
        self.atomIDActorMap = {}
        self.atomIDMoleculeMap = {}

    def getOptions(self, parent = None):

        if self.options:
            return self.options

        return MolPlot3DOptions.MolPlot3DOptions(parent,
                                                         "Molecule Plot 3D Options",
                                                         self)
    
    def resetModelSettings(self, models):
        self.models = models
        self.hiddenActors = []
        self.hiddenAtoms = []
        self.picked = []

    def updateAnnotation(self):
        # imported here to be able to instantiate this class without pymol
        from VtkGenerator import VtkGenerator
        
        for actor in self.actorList['text']:
            self.renderer.RemoveActor(actor)
        self.actorList['text'] = []
        camera = self.renderer.GetActiveCamera()
        for model in self.models:
            gen = VtkGenerator(model)
            if self.showName or self.showIndex:
                (textActors,textHidden) = gen.getAnnotateActors(camera,self.hiddenAtoms, self.showIndex,self.showName)
                for actor in textActors:
                    self.actorList['text'].append(actor)
                    self.renderer.AddActor(actor)
        
    def addData(self, data):
        """Adds the data to the plot"""
        # imported here to be able to instantiate this class without pymol
        from VtkGenerator import VtkGenerator
        
        self.picked = []

        newMoleculesReceived = False
        if self.models != data:
            self.resetModelSettings(data)
            newMoleculesReceived = True

        camera = self.renderer.GetActiveCamera()

        self.bondActors = {}
        
        for model in data:
            gen = VtkGenerator(model)
            if self.molPlotType == self.BNSPLOT:
                (actors,self.hiddenActors,bondActors) = gen.getBallAndStickActors(self.hiddenAtoms)
            elif self.molPlotType == self.SFPLOT:
                (actors,self.hiddenActors,bondActors) = gen.getSpaceFillingActors(self.hiddenAtoms)
            else:
                (actors,self.hiddenActors,bondActors) = gen.getLineActors(self.hiddenAtoms)

            for actor in actors:
                self.actorList['molecules'].append(actor)
                self.renderer.AddActor(actor)

            self.bondActors.update(bondActors)

        self.updateAnnotation()
        self.updateAtomIDActorMap()
        self.updateAtomIDMoleculeMap()

        if newMoleculesReceived:
            self.renderer.ResetCamera()

        self.picker = self.renderWindow.GetPicker()      
        self.picker.AddObserver("PickEvent", lambda a,b : self.annotatePick())

    def getActorForBond(self, bond):
        return self.bondActors[bond]

    def getActorForAtom(self, atom):
        return self.atomIDActorMap[id(atom)]

    def updateAtomIDActorMap(self):
        self.atomIDActorMap = {}
        for actor in self.actorList['molecules']:
            if hasattr(actor, "atom"):
                self.atomIDActorMap[id(actor.atom)] = actor

    def startSelection(self):
        callback = self.selectCallback
        selectBoxMode = SelectBoxMode(self.renderWindow, callback)
        self.renderWindow.setActiveMode(selectBoxMode)

    def selectCallback(self, x, y, width, height):
        """Select atoms within area and at the same time remove the
        selection mode from render window."""
        self.selectWithinBox(x, y, width, height)
        self.renderWindow.setActiveMode(NullMode(self.renderWindow))

    def endPick(self):

        act = self.picker.GetActor()
        self.negateColor(act)

    def interactiveRotate(self, xpos, ypos, lastXpos, lastYpos):
        """Update intermediate positions while interactively rotating."""

        cam = self.renderer.GetActiveCamera()

        if not self.originalPositions:
            self.originalPositions = {}
            for a in self.picked:
                self.originalPositions[a] = a.GetCenter()

        # find mass center
        center = [0.0,0.0,0.0]
        totalmass = 0.0
        bounds = list(self.originalPositions.keys()[0].GetBounds())
        for a in self.originalPositions:
            center[0] += self.originalPositions[a][0] * a.atom.get_mass()
            center[1] += self.originalPositions[a][1] * a.atom.get_mass()
            center[2] += self.originalPositions[a][2] * a.atom.get_mass()
            totalmass += a.atom.get_mass()
            aBounds = a.GetBounds()
            if aBounds[0] < bounds[0]:
                bounds[0] = aBounds[0]
            if aBounds[1] > bounds[1]:
                bounds[1] = aBounds[1]
            if aBounds[2] < bounds[2]:
                bounds[2] = aBounds[2]
            if aBounds[3] > bounds[3]:
                bounds[3] = aBounds[3]
            if aBounds[4] < bounds[4]:
                bounds[4] = aBounds[4]
            if aBounds[5] > bounds[5]:
                bounds[5] = aBounds[5]

        boundDiag = 0.5 * math.sqrt(vtkMath.Distance2BetweenPoints((bounds[0],bounds[2],bounds[4]),
                                                                   (bounds[1],bounds[3],bounds[5]))) 

        center[0] /= float(totalmass)
        center[1] /= float(totalmass)
        center[2] /= float(totalmass)

        # create axis as seen by camera:

        cam.OrthogonalizeViewUp()
        cam.ComputeViewPlaneNormal()
        viewUp = list(cam.GetViewUp())
        vtkMath.Normalize(viewUp)
        viewLook = cam.GetViewPlaneNormal()
        viewRight = [0.0,0.0,0.0]
        vtkMath.Cross(viewUp, viewLook, viewRight)
        vtkMath.Normalize(viewRight)

        # furthest possible point from position + origin
        outsidePt = [0.0,0.0,0.0]

        outsidePt[0] = center[0] + viewRight[0] * boundDiag
        outsidePt[1] = center[1] + viewRight[1] * boundDiag
        outsidePt[2] = center[2] + viewRight[2] * boundDiag

        # convert to display coordinates
        dispObjCenter = [0.0, 0.0, 0.0]
        vc = vtkCoordinate()
        vc.SetCoordinateSystemToWorld()
        vc.SetValue(center)
        dispObjCenter = vc.GetComputedDisplayValue(self.renderer)
        vc.SetCoordinateSystemToWorld()
        vc.SetValue(outsidePt)
        outsidePt = vc.GetComputedDisplayValue(self.renderer)

        radius = math.sqrt(vtkMath.Distance2BetweenPoints((dispObjCenter[0], dispObjCenter[0], 0),
                                                          (outsidePt[0], outsidePt[1], 0)))

        nxf = (xpos - dispObjCenter[0]) / radius
        nyf = (ypos - dispObjCenter[1]) / radius

        oxf = (lastXpos - dispObjCenter[0]) / radius
        oyf = (lastYpos - dispObjCenter[1]) / radius

        
        newXAngle = nxf * vtkMath.RadiansToDegrees()
        newYAngle = nyf * vtkMath.RadiansToDegrees()
        oldXAngle = oxf * vtkMath.RadiansToDegrees()
        oldYAngle = oyf * vtkMath.RadiansToDegrees()
        
        scale = [1.0, 1.0, 1.0]
        xRotate = [newXAngle - oldXAngle,
                   viewUp[0],
                   viewUp[1],
                   viewUp[2]]
        yRotate = [newYAngle - oldYAngle,
                   viewRight[0],
                   viewRight[1],
                   viewRight[2]]

        for a in self.picked:
            self.prop3DTransform(a,
                                 center,
                                 xRotate,
                                 yRotate)
            
        self.renderer.ResetCameraClippingRange()
        self.renderWindow.Render()

    def prop3DTransform(self, actor, center, xRotate, yRotate):
        # only xRotate and yRotate implemented for now, scale ignored

        oldMatrix = actor.GetMatrix()
        orig = actor.GetOrigin()
        newTransform = vtkTransform()
        newTransform.PostMultiply()

        newTransform.SetMatrix(oldMatrix)

        newTransform.Translate(-center[0],
                               -center[1],
                               -center[2])

        newTransform.RotateWXYZ(xRotate[0],
                                xRotate[1],
                                xRotate[2],
                                xRotate[3])
        
        newTransform.RotateWXYZ(yRotate[0],
                                yRotate[1],
                                yRotate[2],
                                yRotate[3])

        newTransform.Translate(center)
        
        newTransform.Translate(-orig[0],
                               -orig[1],
                               -orig[2])
        newTransform.PreMultiply()
        newTransform.Translate(orig)

        actor.SetPosition(newTransform.GetPosition())
        actor.SetOrientation(newTransform.GetOrientation())
        
        
    def interactiveMove(self, movedDistance):
        """Update intermediate positions while interactively moving."""
        # Move all picked actors the remaining amount (they probably are already moved some part of the way)
        
        if not self.originalPositions:
            self.originalPositions = {}
            for a in self.picked:
                self.originalPositions[a] = a.GetCenter()
        
        for a in self.picked:
                currentPos = a.GetCenter()
                changePos = (movedDistance[0] + self.originalPositions[a][0] - currentPos[0],
                             movedDistance[1] + self.originalPositions[a][1] - currentPos[1],
                             movedDistance[2] + self.originalPositions[a][2] - currentPos[2])
                a.AddPosition(changePos)
        self.renderWindow.repaint()
            

    def selectWithinBox(self, x, y, width, height):
        """Select all atoms within the pixel-rectangle x,y,x+width-1,y+height-1."""
        self.unselectAll()
        for item in self.actorList['molecules']:
            vtm = self.renderer.GetActiveCamera().GetCompositePerspectiveTransformMatrix(self.renderer.GetAspect()[0], 0, 128)
            size = self.renderer.GetSize()

            if hasattr(item, "atom"):
                atom = item.atom
                vc = vtkCoordinate()
                vc.SetCoordinateSystemToWorld()
                vc.SetValue(atom.coord)
                hcoord = (atom.coord[0], atom.coord[1], atom.coord[2], 1)
                dispvalue = vc.GetComputedDisplayValue(self.renderer)
                windowvalue = (dispvalue[0], self.renderWindow.height() - dispvalue[1])
                multcoord = vtm.MultiplyPoint(hcoord)
                if windowvalue[0] > x and windowvalue[0] < (width+x) and windowvalue[1] > y and windowvalue[1] < (height+y) and multcoord[3] > 0:
                    self.picked.append(item)
                    self.negateColor(item)
        self.renderWindow.update()
        self.renderer.Render()

    def annotatePick(self):
        pos = self.picker.GetPickPosition()
        
        act = self.picker.GetActor()

        # only pick atoms
        if act and hasattr(act, "atom"):
            self.negateColor(act)
        
            if (self.picked.count(act)):
                self.picked.remove(act)
            else:
                self.picked.append(act)

    def updateAtomIDMoleculeMap(self):
        self.atomIDMoleculeMap = {}
        for m in self.models:
            for a in m.atom:
                self.atomIDMoleculeMap[id(a)] = m

    def getMoleculeForAtom(self, atom):
        return self.atomIDMoleculeMap[id(atom)]

    def extendSelectionAlongBonds(self):
        # imported here to be able to instantiate this class without pymol
        from Library.Util import molecular
        unhandled = {}
        handled = {}
        for actor in self.picked:
            if hasattr(actor, "atom"):
                unhandled[id(actor.atom)] = actor.atom

        self.unselectAll()

        while unhandled:
            atoms = copy.copy(unhandled)
            for atomid in atoms:
                atom = atoms[atomid]
                unhandled.pop(atomid)
                handled[atomid] = atom
                try:
                    actor = self.getActorForAtom(atom)
                    self.negateColor(actor)
                    self.picked.append(actor)

                    model = self.getMoleculeForAtom(atom)
                    index = model.index_atom(atom)
                    
                    attached = molecular.getAttached(model, index)
                    for i in attached:
                        a = model.atom[i]
                        if not handled.get(id(a)):
                            unhandled[id(a)] = a
                except KeyError:
                    pass
                    
        self.renderWindow.update()
                    

    def extendSelectionToModel(self):
        """Pick all atoms in same models as already picked atoms."""
        
        #find out which models have picked atoms
        selectedModels = {}
        for p in self.picked:
            if hasattr(p, "atom"):
                selectedModels[self.getMoleculeForAtom(p.atom)] = True

        self.unselectAll()

        for model in selectedModels:
            for a in model.atom:
                actor = self.getActorForAtom(a)
                self.negateColor(actor)
                self.picked.append(actor)

        self.renderWindow.update()

    def selectAll(self):
        self.unselectAll()
        self.invertPicked()

    def unselectAll(self):
        for item in self.picked:
            self.negateColor(item)

        self.picked = []
        

    def invertPicked(self):
        for item in self.actorList['molecules']:
            if hasattr(item, "atom"):
                self.negateColor(item)
                if self.picked.count(item):
                    self.picked.remove(item)
                else:
                    self.picked.append(item)
        self.renderWindow.update()
        self.renderer.Render()
        
    def negateColor(self, actor):

        color = actor.GetProperty().GetColor()
        newColor = []
        newColor.append(1 - color[0])
        newColor.append(1 - color[1])
        newColor.append(1 - color[2])

        actor.GetProperty().SetColor(newColor)

    def addAxes(self):
        pass


    def getSetupWidget(self, parent):
        """
        Params:
               - renderSetup : RenderSetupWidget - The render setup widget that controlls all things
               related to the render window the plot belongs to.
               - parent : QWidget - The parent widget of the setup widget. Should not be
               None, otehrwise we will get a lot of unwanted top level windows.

        Returns:
               - MolPlot3DSetup - This plot's setup widget.
        Returns the setup widget of this plot. If no exists, one will be created before is is
        returned"""
        
        if not self.setupWidget:
            self.setupWidget = self.setupWidgetClass(parent)
            self.setupWidget.setPlot(self)
        return self.setupWidget

    def redrawPlot(self, renderWindow = None):
        if not self.renderer:
            self.createSelf(renderWindow)

        self.removeActors()

        self.actorList['molecules'] = []
        self.actorList['text'] = []

        self.addData(self.dataFilter.getPlotData())

        self.renderWindow.hideBoundingBox()
        self.renderWindow.update()


    def hidePicked(self):
        for p in self.picked:
            atom = p.atom
            self.renderer.RemoveActor(p)
            self.actorList['molecules'].remove(p)
            # set color back
            self.negateColor(p)
            self.hiddenActors.append(p)
            self.hiddenAtoms.append(atom)

            for model in self.models:
                index = model.index_atom(atom)
                if index != -1:
                    for b in model.bond:
                        if index in b.index:
                            bondActor = self.getActorForBond(b)
                            self.renderer.RemoveActor(bondActor)
                            self.hiddenActors.append(bondActor)
        self.picked = []
        self.redrawPlot()
 
    def unhideAll(self):
        for p in self.hiddenActors:
            self.renderer.AddActor(p)
            self.actorList['molecules'].append(p)

        self.hiddenActors = []
        self.hiddenAtoms = []
        self.updateAnnotation()
        self.redrawPlot()

    def setSpaceFillingPlot(self):
        self.molPlotType = self.SFPLOT
        if self.renderer and self.renderWindow:
            self.redrawPlot()

    def setBallNStickPlot(self):
        self.molPlotType = self.BNSPLOT
        if self.renderer and self.renderWindow:
            self.redrawPlot()
        
    def setLinePlot(self):
        self.molPlotType = self.LPLOT
        if self.renderer and self.renderWindow:
            self.redrawPlot()

    def printPlaneAngleForSelectedFour(self):
        # imported here to be able to instantiate this class without pymol
        from Library.Util import molecular
        if len(self.picked) != 4:
            "Must pick four atoms to calculate bondangle."
        else:
            a = Point3D(self.picked[0].atom.coord)
            b = Point3D(self.picked[1].atom.coord)
            c = Point3D(self.picked[2].atom.coord)
            d = Point3D(self.picked[3].atom.coord)
            print "Angle:",molecular.getAngleBetweenPlanes(a,b,c,d)
