# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from vtk import vtkActor, vtkAssembly, vtkFollower, vtkLineSource
from vtk import vtkPolyDataMapper, vtkSphereSource, vtkTubeFilter, vtkVectorText
from Library.Util.molecular import getAttached
from Library.Util.math3d.Point3D import Point3D
from Library.Util.math3d.Vector3D import Vector3D

#hack for redhat-installed pymol
import sys
import os.path
for path in sys.path:
    combopath = os.path.join(path, 'pymol/modules')
    if os.path.exists(combopath):
        if sys.path.count(combopath) == 0:
            sys.path.append(combopath)

colors = {
    'C' : [200, 200, 200],
    'O' : [240, 0, 0],
    'H' : [255, 255, 255],
    'N' : [143, 143, 255],
    'S' : [255, 200, 50],
    'P' : [255, 165, 0],
    'Cl' : [0, 255, 0],
    'Zn' : [165, 42, 42],
    'Br' : [165, 42, 42],
    'Na' : [0, 0, 255],
    'Fe' : [255, 165, 0],
    'Mn' : [42, 128, 42],
    'Ca' : [128, 128, 128],
    'Unknown' : [255, 20, 147]
    }

vdwradii = {
    'Ag' : 1.72,
    'Ar' : 1.88,
    'As' : 1.85,
    'Au' : 1.66,
    'Br' : 1.85,
    'C' : 1.70,
    'Cd' : 1.58,
    'Cl' : 1.75,
    'Cu' : 1.40,
    'F' : 1.47,
    'Ga' : 1.87,
    'H' : 1.20,
    'He' : 1.40,
    'Hg' : 1.55,
    'I' : 1.98,
    'In' : 1.93,
    'K' : 2.75,
    'Kr' : 2.02,
    'Li' : 1.82,
    'Mg' : 1.73,
    'N' : 1.55,
    'Na' : 2.27,
    'Ne' : 1.54,
    'Ni' : 1.63,
    'O' : 1.52,
    'P' : 1.80,
    'Pb' : 2.02,
    'Pd' : 1.63,
    'Pt' : 1.72,
    'S' : 1.80,
    'Se' : 1.90,
    'Si' : 2.10,
    'Sn' : 2.17,
    'Te' : 2.06,
    'Tl' : 1.96,
    'U' : 1.86,
    'Xe' : 2.16,
    'Zn' : 1.39,
    'Unknown' : 2.0
    }


class VtkGenerator:

    def __init__(self, model):
        self.model = model
        self.colors = {}
        for i in colors:
            self.colors[i] = self.__toFloats(colors[i])

    def __toFloats(self, (a, b, c)):
        max = 255.0
        return [a/max, b/max, c/max]
        
    def getAnnotateActors(self,camera,hiddenAtoms, showIndex=True,showName=True):
        atoms = self.model.atom
        actors = []
        hiddenActors = []

        for a in atoms:

            indexString = str(atoms.index(a) + 1) # 1-indexed for display
            nameString = a.name or a.symbol

            string = ""

            if showIndex and showName:
                string = indexString + "," + nameString
            elif showIndex:
                string = indexString
            elif showName:
                string = nameString
                
            text = vtkVectorText()
            text.SetText(string)
            textMapper = vtkPolyDataMapper()
            textMapper.SetInput(text.GetOutput())
            textActor = vtkFollower()
            textActor.SetMapper(textMapper)
            textActor.SetScale(0.4, 0.4, 0.4)
#            textActor.SetOrigin(a.coord)
            textActor.AddPosition(a.coord)
            textActor.AddPosition([0.0, -0.5, 0.0])
            textActor.GetProperty().SetColor([1.0, 0.0, 0.0])
            
            textActor.SetCamera(camera)

            if a in hiddenAtoms:
                hiddenActors.append(textActor)
            else:
                actors.append(textActor)

        return (actors, hiddenActors)

    def __createStickActor(self, start, end, radius, color):
        bondSource = vtkLineSource()
        bondSource.SetPoint1(start[0:3])
        bondSource.SetPoint2(end[0:3])
        tubeFilter = vtkTubeFilter()
        tubeFilter.SetInput(bondSource.GetOutput())
        tubeFilter.SetRadius(radius)
        tubeFilter.SetNumberOfSides(10)
        bondMapper = vtkPolyDataMapper()
        bondMapper.SetInput(tubeFilter.GetOutput())
        bondActor = vtkActor()
        bondActor.GetProperty().SetColor(color)
        bondActor.SetMapper(bondMapper)
        return bondActor

    def __createBallActor(self, position, radius, color):
        atomSource = vtkSphereSource()
        atomSource.SetRadius(radius)
        #        atomSource.SetRadius(0.1)
        atomSource.SetCenter(position[0:3])
        atomSource.SetPhiResolution(16)
        atomSource.SetThetaResolution(16)
        atomMapper = vtkPolyDataMapper()
        atomMapper.SetInput(atomSource.GetOutput())
        atomActor = vtkActor()
        atomActor.SetMapper(atomMapper)
        atomActor.GetProperty().SetColor(color)
        #    atomActor.GetProperty().SetOpacity(0.5)
        return atomActor

    def __createLineActor(self, start, end, color):
            bondSource = vtkLineSource()
            bondSource.SetPoint1(start[0:3])
            bondSource.SetPoint2(end[0:3])
            bondMapper = vtkPolyDataMapper()
            bondMapper.SetInput(bondSource.GetOutput())
            bondActor = vtkActor()
            bondActor.GetProperty().SetColor(color)
            bondActor.SetMapper(bondMapper)
            return bondActor

    def __getOffsetVector(self, avec,bvec):
        """Find vector in the same plane as avec and bvec, perpendicular to avec."""
        kvec = avec.cross(bvec)
        result = kvec.cross(avec)
        return result
        
    def getBallAndStickActors(self, hiddenAtoms):
        atoms = self.model.atom
        bonds = self.model.bond

        actors = []
        hiddenActors = []
        bondActors = {}

        for a in atoms:
            atomActor = self.__createBallActor(a.coord,
                                               0.2*vdwradii[a.symbol],
                                               self.colors[a.symbol])
            atomActor.atom = a

            # GaussDal stuff.
            # The GaussDal interface depends on a consistent atom id across instances of the same
            # molecule.
            if hasattr(a, 'id'):
                atomActor.id = a.id

            if a in hiddenAtoms:
                hiddenActors.append(atomActor)
            else:
                actors.append(atomActor)
        
        for b in bonds:
            bondAssembly = vtkAssembly()
            first = atoms[b.index[0]]
            second = atoms[b.index[1]]
            firstPos = Point3D(first.coord)
            secondPos = Point3D(second.coord)
            bondVector = Vector3D(startPoint=firstPos,
                                   endPoint=secondPos)
            bondVector.multiply(0.5)
            centerPos = Point3D(firstPos)
            centerPos.add(bondVector)

            bondActor1 = self.__createStickActor(firstPos, centerPos, 0.1, self.colors[first.symbol])
            bondActor2 = self.__createStickActor(centerPos, secondPos, 0.1, self.colors[second.symbol])

            if b.order == 2:
                length = 0.1
                
                # find other connected atom to get the plane
                connectedAtoms = getAttached(self.model, b.index[0])
                connectedAtoms.extend(getAttached(self.model, b.index[1]))
                if (connectedAtoms.count(b.index[0])):
                    connectedAtoms.remove(b.index[0])
                if (connectedAtoms.count(b.index[1])):
                    connectedAtoms.remove(b.index[1])
                    
                if connectedAtoms:
                    # create in same plane as existing atom
                    planeAtomPos = Point3D(atoms[connectedAtoms[0]].coord)
                    bvec = Vector3D(firstPos, planeAtomPos)
                    offsetVector = self.__getOffsetVector(bondVector, bvec)
                else:
                    # create in arbitrary direction
                    offsetVector = bondVector.getPerpendicular()
                offsetVector.setLength(length)

                bondActor3 = self.__createStickActor(firstPos, centerPos, 0.1, self.colors[first.symbol])
                bondActor4 = self.__createStickActor(centerPos, secondPos, 0.1, self.colors[second.symbol])

                # offset the two two-part bond-cylinders
                bondActor1.AddPosition(offsetVector[0:3])
                bondActor2.AddPosition(offsetVector[0:3])
                offsetVector.negate()
                bondActor3.AddPosition(offsetVector[0:3])
                bondActor4.AddPosition(offsetVector[0:3])
                
                bondAssembly.AddPart(bondActor3)
                bondAssembly.AddPart(bondActor4)
            elif b.order == 3:
                length = 0.1

                offsetVector = bondVector.getPerpendicular()
                offsetVector.setLength(length)

                bondActor3 = self.__createStickActor(firstPos, centerPos, 0.1, self.colors[first.symbol])
                bondActor4 = self.__createStickActor(centerPos, secondPos, 0.1, self.colors[second.symbol])
                bondActor5 = self.__createStickActor(firstPos, centerPos, 0.1, self.colors[first.symbol])
                bondActor6 = self.__createStickActor(centerPos, secondPos, 0.1, self.colors[second.symbol])

                bondActor3.AddPosition(offsetVector[0:3])
                bondActor4.AddPosition(offsetVector[0:3])
                offsetVector.negate()
                bondActor5.AddPosition(offsetVector[0:3])
                bondActor6.AddPosition(offsetVector[0:3])
                
                bondAssembly.AddPart(bondActor3)
                bondAssembly.AddPart(bondActor4)
                bondAssembly.AddPart(bondActor5)
                bondAssembly.AddPart(bondActor6)

                bondActor3.bond = b
                bondActor4.bond = b
                bondActor5.bond = b
                bondActor6.bond = b
            
            bondActor1.bond = b
            bondActor2.bond = b
            bondAssembly.AddPart(bondActor1)
            bondAssembly.AddPart(bondActor2)

            bondAssembly.bond = b

            bondActors[b] = bondAssembly

            if first in hiddenAtoms or second in hiddenAtoms:
                hiddenActors.append(bondAssembly)
            else:
                actors.append(bondAssembly)

        return (actors, hiddenActors, bondActors)

    def getSpaceFillingActors(self, hiddenAtoms):
        atoms = self.model.atom
        actors = []
        hiddenActors = []
        bondActors = {}
        
        for a in atoms:
            atomSource = vtkSphereSource()
            atomSource.SetRadius(vdwradii[a.symbol])
            atomSource.SetCenter(a.coord)
            atomSource.SetPhiResolution(16)
            atomSource.SetThetaResolution(16)
            atomMapper = vtkPolyDataMapper()
            atomMapper.SetInput(atomSource.GetOutput())
            atomActor = vtkActor()
            atomActor.SetMapper(atomMapper)
            atomActor.GetProperty().SetColor(self.colors[a.symbol])

            atomActor.atom = a

            if a in hiddenAtoms:
                hiddenActors.append(atomActor)
            else:
                actors.append(atomActor)

        return (actors, hiddenActors, bondActors)

    def getLineActors(self, hiddenAtoms):
        atoms = self.model.atom
        bonds = self.model.bond
        
        actors = []
        hiddenActors = []
        bondActors = {}
        assemblies = {}
        
        for a in atoms:
            assembly = vtkAssembly()
            assemblies[atoms.index(a)] = assembly
            if a in hiddenAtoms:
                hiddenActors.append(assemblies[atoms.index(a)])
            else:
                actors.append(assemblies[atoms.index(a)])
                
            
        for b in bonds:
            first = atoms[b.index[0]]
            second = atoms[b.index[1]]
            firstPos = Point3D(first.coord)
            secondPos = Point3D(second.coord)
            bondVector = Vector3D(startPoint=firstPos,
                                  endPoint=secondPos)
            bondVector.multiply(0.5)
            centerPos = Point3D(firstPos)
            centerPos.add(bondVector)
                                
            bondActor1 = self.__createLineActor(firstPos, centerPos, self.colors[first.symbol])
            bondActor2 = self.__createLineActor(centerPos, secondPos, self.colors[second.symbol])

            if b.order == 2:
                length = 0.1
                
                # find other connected atom to get the plane
                connectedAtoms = getAttached(self.model, b.index[0])
                connectedAtoms.extend(getAttached(self.model, b.index[1]))
                if (connectedAtoms.count(b.index[0])):
                    connectedAtoms.remove(b.index[0])
                if (connectedAtoms.count(b.index[1])):
                    connectedAtoms.remove(b.index[1])
                    
                if connectedAtoms:
                    planeAtomPos = Point3D(atoms[connectedAtoms[0]].coord)
                    bvec = Vector3D(firstPos, planeAtomPos)
                    offsetVector = self.__getOffsetVector(bondVector, bvec)
                else:
                    offsetVector = bondVector.getPerpendicular()
                offsetVector.setLength(length)

                bondActor3 = self.__createLineActor(firstPos, centerPos, self.colors[first.symbol])
                bondActor4 = self.__createLineActor(centerPos, secondPos, self.colors[second.symbol])

                # offset the two two-part bond-cylinders
                bondActor1.AddPosition(offsetVector[0:3])
                bondActor2.AddPosition(offsetVector[0:3])
                offsetVector.negate()
                bondActor3.AddPosition(offsetVector[0:3])
                bondActor4.AddPosition(offsetVector[0:3])
                
                assemblies[b.index[0]].AddPart(bondActor3)
                assemblies[b.index[1]].AddPart(bondActor4)

            elif b.order == 3:
                length = 0.1

                offsetVector = bondVector.getPerpendicular()
                offsetVector.setLength(length)

                bondActor3 = self.__createLineActor(firstPos, centerPos, self.colors[first.symbol])
                bondActor4 = self.__createLineActor(centerPos, secondPos, self.colors[second.symbol])
                bondActor5 = self.__createLineActor(firstPos, centerPos, self.colors[first.symbol])
                bondActor6 = self.__createLineActor(centerPos, secondPos, self.colors[second.symbol])

                bondActor3.AddPosition(offsetVector[0:3])
                bondActor4.AddPosition(offsetVector[0:3])
                offsetVector.negate()
                bondActor5.AddPosition(offsetVector[0:3])
                bondActor6.AddPosition(offsetVector[0:3])
                
                assemblies[b.index[0]].AddPart(bondActor3)
                assemblies[b.index[1]].AddPart(bondActor4)
                assemblies[b.index[0]].AddPart(bondActor5)
                assemblies[b.index[1]].AddPart(bondActor6)

                bondActor3.bond = b
                bondActor4.bond = b
                bondActor5.bond = b
                bondActor6.bond = b

            assemblies[b.index[0]].AddPart(bondActor1)
            assemblies[b.index[1]].AddPart(bondActor2)


        return (actors, hiddenActors, bondActors)

