# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from vtk import vtkCubeAxesActor2D, vtkDoubleArray, vtkImageDataGeometryFilter
from vtk import vtkLODActor, vtkMergeFilter, vtkPolyDataMapper, vtkScalarBarActor
from vtk import vtkSmoothPolyDataFilter, vtkStructuredPoints, vtkWarpScalar

from GUI.Plots.Plots3D import Plot3D
from GUI.Plots.Plots3D.SurfacePlot3D import SurfacePlot3DSetup
from GUI.Plots import SimplePlotDataFilter
class SurfacePlot3D(Plot3D.Plot3D):

    def __init__(self):
        Plot3D.Plot3D.__init__(self)

        # We want a simple filter here
        self.dataFilter = SimplePlotDataFilter.SimplePlotDataFilter()
        #self.plotType = "Surface Plot 3D"
        self.plotType = "3D-Surface"
        self.isPickable = 0
        self.setupWidget = None

    def createSelf(self, renderWindow = None):
        Plot3D.Plot3D.createSelf(self, renderWindow)
        # self.renderer.SetBackground(0.8, 0.8, 0.8)
        # self.addAxes()

    def createBarItems(self):

        Plot3D.Plot3D.createBarItems(self)

    def addAxes(self):
        pass

    def addData(self, data):

        self.data = data
        self.initDataToVTK()

    def initDataToVTK(self):
        self.maxData = self.data[0][0]
        self.minData = self.data[0][0]

        
        spData = vtkStructuredPoints()
        
        spData.SetDimensions(len(self.data[0]), len(self.data), 1)
        spData.SetOrigin(0.0, 0.0, 0.0)
        spData.SetSpacing(1.0, 1.0, 1.0)

        # print "Empty spData is now:\n", spData
        pData = spData.GetPointData()

        dArray = vtkDoubleArray()
        dArray.SetName("point_data")

        for row in self.data:
            for num in row:
                if num < self.minData:
                    self.minData = num

                if num > self.maxData:
                    self.maxData = num

                dArray.InsertNextValue(num)

        pData.SetScalars(dArray)

        self.maxXY = max(len(self.data),len(self.data[0]))
        self.rangeData = self.maxData - self.minData
        self.warpFactor = self.maxXY/self.rangeData


        
        # Converts structured points to polyData
        geometry = vtkImageDataGeometryFilter()
        # print "getting sp output"
        # print str(spData.GetPointData().GetScalars())
                
        # geometry.SetInput(reader.GetOutput())
        geometry.SetInput(spData)
        
        warp = vtkWarpScalar()
        warp.SetInput(geometry.GetOutput())
        warp.XYPlaneOff()
        #warp.SetScaleFactor(self.rangeData)
        warp.SetScaleFactor(100)

        merge = vtkMergeFilter()
        merge.SetGeometry(warp.GetOutput())
        
        # merge.SetScalars(reader.GetOutput())
        merge.SetScalars(spData)

        
        smoother = vtkSmoothPolyDataFilter()
        smoother.SetInput(merge.GetOutput())
        smoother.SetNumberOfIterations(50)

        mapper = vtkPolyDataMapper()
        mapper.SetInput(smoother.GetOutput())
        mapper.SetScalarRange(self.minData,self.maxData)
        mapper.ImmediateModeRenderingOff()

        surface = vtkLODActor()
        surface.SetMapper(mapper)
        surface.SetPickable(False)

        self.actorList["pos"].append(surface)
        self.renderer.AddActor(surface)

        axes = vtkCubeAxesActor2D()
        axes.SetInput(merge.GetOutput())
        axes.SetCamera(self.renderer.GetActiveCamera())
        axes.SetFlyModeToOuterEdges()
        axes.GetProperty().SetColor(0.0, 0.0, 0.0)
        axes.GetAxisLabelTextProperty().BoldOn()
        axes.GetAxisLabelTextProperty().ShadowOn()
        axes.SetPickable(False)
        self.renderer.AddActor(axes)
        self.actorList["axis"].append(axes)

        scalarbar = vtkScalarBarActor()
        scalarbar.SetPickable(False)
        scalarbar.SetLookupTable(mapper.GetLookupTable())
        scalarbar.SetWidth(0.1)
        scalarbar.SetNumberOfLabels(3)
        scalarbar.SetOrientation(3)
        scalarbar.GetProperty().SetColor(0,0,0)
        self.renderer.AddActor(scalarbar)
        self.actorList["axis"].append(scalarbar)

        camera = self.renderer.GetActiveCamera()
        camera.SetFocalPoint(0,0,0)
        camera.SetPosition(1,1,1)
        camera.ComputeViewPlaneNormal()
        camera.SetViewUp(0.0,0.0,0.5)
        camera.OrthogonalizeViewUp()

        self.renderer.SetActiveCamera(camera)
        self.renderer.ResetCamera()

    def getSetupWidget(self, renderSetup):
        if not self.setupWidget:
            self.setupWidget = SurfacePlot3DSetup.SurfacePlot3DSetup(renderSetup)
            self.setupWidget.setPlot(self)
        return self.setupWidget
