# -*- coding: utf-8 -*-

# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.



import vtk
from qt import QBoxLayout, QFrame, QGroupBox, QHBoxLayout, QListBox, QMessageBox
from qt import QPoint, QPushButton, QScrollView, QSize, QSizePolicy, QSpacerItem
from qt import QToolTip, QVBoxLayout, QWidget, Qt, SIGNAL, qApp
from GUI.Plots import PlotAddWindow
from GUI.Plots import PlotOptionsWindow
from GUI.Plots import PlotConfig
from GUI.Plots import PlotRepository
import types

class RenderSetupWidget(QWidget):
    """
    This class governs the contents and behaviour of a specific PlotRenderWindow.
    That is, it knows what plots are going into the render window as well as providing
    GUI space for each individual plot's own setup widget.
    """
    
    def __init__(self, renderWindow, windowController, parent = None, name = None, fl = 0):
        QWidget.__init__(self, parent, name, fl)
        
        if not name:
            self.setName("RenderSetupWidget")

        self.windowController = windowController
        self.renderWindow = renderWindow
        self.emptySetupText = "No plots added"

        # Interaction styles that can be used...
        self.interactionStyles = {
            "Flight" : vtk.vtkInteractorStyleFlight(),
            "Joystick Actor" : vtk.vtkInteractorStyleJoystickActor(),
            "Joystick Camera" : vtk.vtkInteractorStyleJoystickCamera(),
            "Trackball Actor" : vtk.vtkInteractorStyleTrackballActor(),
            "Trackball Camera" : vtk.vtkInteractorStyleTrackballCamera(),
            "Unicam" : vtk.vtkInteractorStyleUnicam()}

        self.renderWindow.renderer.SetBackground(float(PlotConfig.PlotConfig.bgR)/255,
                                                 float(PlotConfig.PlotConfig.bgG)/255,
                                                 float(PlotConfig.PlotConfig.bgB)/255)
        
        self.renderWindow._Iren.SetInteractorStyle(
            self.interactionStyles[str(PlotConfig.PlotConfig.interaction3DMode)])
        
        self.plots = list()
        RenderSetupWidgetLayout = QHBoxLayout(self,4,4,"RenderSetupWidgetLayout")

        controlLayout = QVBoxLayout(None,0,2,"controlLayout")

        self.groupBox1 = QGroupBox(self,"groupBox1")
        self.groupBox1.setSizePolicy(QSizePolicy(QSizePolicy.Minimum,
                                                 QSizePolicy.Preferred, 0, 0,
                                                 self.groupBox1.sizePolicy().hasHeightForWidth()))
        self.groupBox1.setTitle(self.__tr("Plots"))
        self.groupBox1.setColumnLayout(0,Qt.Vertical)
        self.groupBox1.layout().setSpacing(6)
        self.groupBox1.layout().setMargin(11)
        groupBox1Layout = QVBoxLayout(self.groupBox1.layout())
        groupBox1Layout.setAlignment(Qt.AlignTop)
        groupBox1Layout.setMargin(2)
        groupBox1Layout.setSpacing(2)

        self.plotList = QListBox(self.groupBox1,"plotList")
        self.plotList.setSizePolicy(QSizePolicy(QSizePolicy.Preferred,
                                                QSizePolicy.Expanding, 0, 0,
                                                self.plotList.sizePolicy().hasHeightForWidth()))
        self.connect(self.plotList, SIGNAL("highlighted(int)"), self.slotPlotHighlighted)
        groupBox1Layout.addWidget(self.plotList)

        self.plotDeleteButton = QPushButton(self.groupBox1,"plotDeleteButton")
        self.plotDeleteButton.setText(self.__tr("Delete"))
        QToolTip.add(self.plotDeleteButton, self.__tr("Delete the selected plot from this render window."))
        self.connect(self.plotDeleteButton, SIGNAL("clicked()"), self.slotDeletePlot)
        groupBox1Layout.addWidget(self.plotDeleteButton)

        self.plotAddButton = QPushButton(self.groupBox1,"plotAddButton")
        groupBox1Layout.addWidget(self.plotAddButton)
        self.plotAddButton.setText(self.__tr("Add..."))
        QToolTip.add(self.plotAddButton, self.__tr("Add a new plot to this render window."))
        self.connect(self.plotAddButton, SIGNAL("clicked()"), self.slotOpenAddPlots)
        
        controlLayout.addWidget(self.groupBox1)
        spacer = QSpacerItem(20,0,QSizePolicy.Minimum,QSizePolicy.Minimum)
        controlLayout.addItem(spacer)

        self.optionsButton = QPushButton(self,"optionsButton")
        self.optionsButton.setSizePolicy(QSizePolicy(QSizePolicy.Fixed,
                                                     QSizePolicy.Fixed, 0, 0,
                                                     self.optionsButton.sizePolicy().hasHeightForWidth()))
        self.optionsButton.setText(self.__tr("Renderer Options..."))
        QToolTip.add(self.optionsButton, self.__tr("Change the settings of the render window, such as background colour etc."))
        self.connect(self.optionsButton, SIGNAL("clicked()"), self.slotOpenOptions)
        controlLayout.addWidget(self.optionsButton)

        
        RenderSetupWidgetLayout.addLayout(controlLayout)

        self.plotSetupFrame = QGroupBox(self,"plotSetupFrame")
        # self.plotSetupFrame.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred,0,0,self.plotSetupFrame.sizePolicy().hasHeightForWidth()))
        #                                             Horizontal             Vertical
        self.plotSetupFrame.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding, 0, 0))
        self.plotSetupFrame.setMinimumSize(QSize(100,0))
        self.plotSetupFrame.setColumnLayout(0, Qt.Vertical)
        self.plotSetupFrame.setInsideMargin(0)
        self.plotSetupFrame.setTitle(self.tr(self.emptySetupText))
        RenderSetupWidgetLayout.addWidget(self.plotSetupFrame)
        self.plotSetupFrameLayout = QBoxLayout(self.plotSetupFrame.layout(), QBoxLayout.Down, 0, "plotSetupFrameLayout")
        # Let the scrollview have a lot of space 
        self.plotSetupFrameLayout.setMargin(3)

        # Added as a reslolution to the resize problems
        self.plotSetupView = QScrollView(self.plotSetupFrame)
        self.plotSetupView.setFrameShape(QFrame.NoFrame)
        self.plotSetupView.viewport().setPaletteBackgroundColor(self.plotSetupFrame.paletteBackgroundColor())
        # This will make the viewport act as a BoxLayout
        self.plotSetupView.setResizePolicy(QScrollView.AutoOneFit)
                
        self.plotSetupFrameLayout.addWidget(self.plotSetupView)

        self.plotSetupWidget = None

    def hidePlotList(self):
        self.groupBox1.hide()
        self.optionsButton.hide()

    def getInteractorStyleName(self):
        """
        Return:
               - String - the name of the current active interactor style
        """
        style = self.renderWindow._Iren.GetInteractorStyle()

        for kw in self.interactionStyles:

            if (str(style.__class__).split(".")[-1]) == (str(self.interactionStyles[kw].__class__).split(".")[-1]):
                return kw
        return "Trackball Camera"

    def slotOpenAddPlots(self):
        """
        Opens a modal window for adding plots into the PlotRenderWindow
        related to to this setup widget
        """
        self.addWin = PlotAddWindow.PlotAddWindow(self)
        self.addWin.show()

    def slotOpenOptions(self):
        
        self.optionsWin = PlotOptionsWindow.PlotOptionsWindow(self)
        self.optionsWin.show()

    def slotDeletePlot(self):

        ##################
        index = self.plotList.currentItem()
        
        text = self.plotList.currentText()

        retval = QMessageBox.warning(self,
                                     "SciCraft",
                                     "This will remove " +
                                     str(text) + " from the renderer!",
                                     QMessageBox.Ok,
                                     QMessageBox.Cancel)

        if retval == QMessageBox.Ok and text:

            self.plotList.removeItem(index)

            plot = self.plots.pop(index)

            setupWidget = plot.getSetupWidget(self)
            # Explicit attempt to delete the object. Same strategy used in PlotWindow with success.
            point = QPoint()
            setupWidget.hide()
            setupWidget.reparent(None, point)

            self.windowController.filterControl.unregisterFilter(plot.getDataFilter())
            plot.destructSelf()
            # plot.redrawPlot()
            self.renderWindow.update()
            del plot
            
            if self.plotList.count() == 0:
                self.renderWindow.renewRenderer()
                self.plotSetupFrame.setTitle(self.emptySetupText)

            else:
                self.plotList.setCurrentItem(0)
                self.setActivePlot(self.plots[0])
            

    def __tr(self,s,c = None):
        return qApp.translate("RenderSetupWidget",s,c)

    def addPlot(self,plot):
        self.plots.append(plot) # append object to internal list of plots

        # Get the data filter of the plot and register it at the filter controller
        self.windowController.filterControl.registerFilter(plot.getDataFilter())
        
        self.setActivePlot(plot, self.windowController.getInputPorts())
        self.plotList.insertItem(plot.plotType) # add item to list
        self.plotList.setSelected(len(self.plots)-1, 1) # select the right item in the list

    def renameSelectedPlot(self, name):
        """
        Param:
              - name : String - The new name of the plot in the plot list.

        Renames the entry for the current selected plot
        in the plot list with a new name. """
        
        self.plotList.changeItem(name, self.plotList.currentItem())
        
    def showPlots(self):

        for plot in self.plots:

            plot.createSelf(self.renderWindow)
    

    def setActivePlot(self, plot, ports = None):
        if self.plotSetupWidget:
            
            self.plotSetupWidget.hide()
            # self.testLayout.remove(self.plotSetupWidget)
            self.plotSetupView.removeChild(self.plotSetupWidget)
            
        self.plotSetupWidget = plot.getSetupWidget(self)
        if ports:
            self.plotSetupWidget.updateInputPorts(ports)

        
        # self.testLayout.addWidget(self.plotSetupWidget)
        self.plotSetupView.addChild(self.plotSetupWidget)
        self.plotSetupFrame.setTitle(str(self.plotSetupWidget.typeLabel.text()))
        self.plotSetupWidget.show()

        # Trying to resize controller so the setup widget
        # will fit.
        # self.windowController.resizeController()

        # print self.plotSetupWidget.parent()
        

    def remPlot(self,plot):
        index = self.plots.index(plot)
        self.remPlotIndex(index)

    def remPlotIndex(self,index):
        self.plots.pop(index)
        self.plotList.removeItem(index)

    def slotPlotHighlighted(self, index):
        self.setActivePlot(self.plots[index])
        # pass
        
    def updateInputPorts(self, ports):

        for plot in self.plots:
            plot.getSetupWidget(self).updateInputPorts(ports)

    def redrawPlots(self):
        # Not necessary me thinks. Bollocks actually...
        # data = self.windowController.gui.getData()
        try:
            for plot in self.plots:
                plot.redrawPlot(self.renderWindow)
        except IndexError, e:
            self.windowController.showErrorMessage(e)
    def apply_all_plot_settings(self):

        try:
            for plot in self.plots:
                setupWidget = plot.getSetupWidget(self)
                setupWidget.slotApply()
        except IndexError, e:
            self.windowController.showErrorMessage(e)
        
    def getConfig(self):
        xmldata = ""
        xmldata += "<renderer>\n"
        for plot in self.plots:
            plottype = str(plot.__class__).split('.')[-1]
            xmldata += "\t<plot>\n"
            xmldata += "\t\t<type>" + plottype + "</type>\n"
            xmldata += "\t\t<config>\n"
            plotconfig = plot.getSetupWidget(self).getConfig()
            for key in plotconfig:
                xmldata += "\t\t\t<" + str(key) + ">"
                xmldata += str(plotconfig[key])
                xmldata += "</" + str(key) + ">\n"
            xmldata += "\t\t</config>\n"
            xmldata += "\t</plot>\n"
        xmldata += "</renderer>\n"
        return xmldata
          
    def setConfig(self, config):
        plotRepository = PlotRepository.PlotRepository()
        # if only one plot, convert it to a list for convenience
        if type(config) == types.LambdaType and config.__dict__.has_key("plot"):
            if not type(config.plot) == types.ListType:
                config.plot = [config.plot]
            # traverse the list of plots and create them
            for plot in config.plot:
                newplot = plotRepository.getPlot(plot.type)
                self.addPlot(newplot)

                # make dictionary from the lambda expression
                dict = {}
                for key in plot.config.__dict__:
                    dict[key] = getattr(plot.config,key)

                # set the plots configuration
                newplot.getSetupWidget(self).setConfig(dict)
                newplot.getSetupWidget(self).slotApply()

        else: #no plots for this renderer
            pass

    def pickWithSelectBox(self, x, y, width, height):
        """Callback function made by the render window when the user has
        drawn a bounding box to pick. See Plot.py for details o method."""
        for plot in self.plots:
            plot.selectWithinBox(x, y, width, height)
        
    def deletePickedElements(self):
        retval = []
        for plot in self.plots:
            retval += plot.deletePickedElements()

        return retval

    def undoPicking(self):

        for plot in self.plots:

            plot.undoPicking()
