# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Plots.interaction.InteractionMode import InteractionMode
from vtk import vtkCoordinate

class MoveMode(InteractionMode):
    """Creates a mode for moving actors. The actual moving is done by the defined callbacks."""

    def __init__(self, renderWindow, interactiveCallback, endCallback):
        """
        interactiveCallback and endCallback both must take one parameter:
        movedDistance - tuple of (x,y,z) distance moved since start

        interactiveCallback is called repeatedly while moving the mouse
        with mouse button down.
        
        endCallback is called when releasing the button, making the move
        final."""
        InteractionMode.__init__(self, renderWindow)
        
        self.endCallback = endCallback
        self.interactiveCallback = interactiveCallback
        self.moving = False

    def mousePressed(self, x, y):
        """Only try to move actors if mouse is pressed on an actor."""
        InteractionMode.mousePressed(self, x, y)
        picker = self.renderWindow.GetPicker()
        # convert coordinates
        pickY = self.renderWindow.height() - y

        # object hit?
        if picker.Pick((x, pickY, 0.0), self.renderWindow.renderer):
            self.startPosition = picker.GetPickPosition()

            self.moving = True
            self.moved = lambda : None
            self.moved.startx = x
            self.moved.starty = y
            self.moved.dx = 0 # change in x
            self.moved.dy = 0 # change in y

            self.movedDistance = (0,0,0)

            return False
        else:
            return True
            

    def mouseReleased(self, x, y):
        """Mouse released at pixel position x,y. Returns True if event should be passed on."""
        # skip SelectBoxMode's mouseReleased and run InteractionMode's instead.
        InteractionMode.mouseReleased(self, x, y)
        if self.moving:
            # only call callback if there was movement
            if self.movedDistance == (0,0,0):
                pass
            else:
                self.endCallback(self.movedDistance)
            self.moving = False
            return False
        else:
            return True

    def mouseMoved(self, x, y):
        """Mouse moved, currently at pixel position x,y. Returns True if event should be passed on."""
        InteractionMode.mouseMoved(self, x, y)
        if self.isMouseDown():
            if self.moving:
                self.moved.dx = x - self.moved.startx
                self.moved.dy = y - self.moved.starty

                screenCoord = self.convert3dToScreen(self.startPosition)

                screenCoord[0] += self.moved.dx
                screenCoord[1] += self.moved.dy

                worldCoord = self.convertScreenTo3d(screenCoord)

                self.movedDistance = (worldCoord[0] - self.startPosition[0],
                                      worldCoord[1] - self.startPosition[1],
                                      worldCoord[2] - self.startPosition[2])

                self.interactiveCallback(self.movedDistance)

            return False
        else:
            return True

    def convert3dToScreen(self, (x, y, z)):
        vtm = self.renderWindow.renderer.GetActiveCamera().GetCompositePerspectiveTransformMatrix(self.renderWindow.renderer.GetAspect()[0], 0, 1)

        hcoord = [x,y,z,1]
        multcoord = vtm.MultiplyPoint(hcoord)

        # convert coordinate to 2d(screen) and adjust according to dx,dy
        
        vc = vtkCoordinate()
        vc.SetCoordinateSystemToWorld()
        vc.SetValue(hcoord[:3])
        dispvalue = list(vc.GetComputedDisplayValue(self.renderWindow.renderer))
        return [dispvalue[0],
                self.renderWindow.height() - dispvalue[1],
                multcoord[2]/multcoord[3]]

    def convertScreenTo3d(self, (x, y, z)):
        vc = vtkCoordinate()
        vc.SetCoordinateSystemToDisplay()
        vc.SetValue(x, self.renderWindow.height() - y, z)
        return list(vc.GetComputedWorldValue(self.renderWindow.renderer))
