# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Plots.plot2d.plot2d import Plot2D
from Library.Util.plot.cluster import ClusterGenerator, Cluster
from Library.Util.Config import Config
from Library.Util.unrepr import unrepr
from GUI.Plots.plot2d.plot2d import PlotMarker
from GUI.Nodes.guiplot2dnodesettings import PlotSettings, LabelSettingsDialog

from qwt import QwtSymbol, QwtScaleDraw, QwtPlot
from qwt import QwtPlotCurve

from qt import Qt, QColor, QSize, QBrush, SIGNAL, QMessageBox
from qtui import QWidgetFactory

from copy import copy
from itertools import izip
import os

class DendrogramPlot(Plot2D):
    """Dendrogram plot with toggleable clusters.."""

    # Constants telling how much the plot must be transformed when inserting
    # labels
    LABEL_X_SPACE = 20
    _map_aspect = "objects"
    _type = "dendrogram_plot"
    _name = "Dendrogram Plot"
    
    def __init__(self, parent_plot, node, config):
        Plot2D.__init__(self, parent_plot, config, [self._map_aspect])
        self._node = node
        self._matrix = None
        self._order = None
        self._cluster_lines = {}
        self._cluster_generator = ClusterGenerator()
        self._cluster_height = None
        self._label_list = []
        self._enabled_labels = {}
        self._selected_marker_list = []
        self.marker_list = []

        # should only be used in methods set_data(...) and _draw_cluster(...)
        self._tree_clusters = []
        self.xaxis = None
        self._parent_plot.enableYLeftAxis(False)

    def create_settings(self, index, tab_widget, plot_widget, node):
        return DendrogramPlotSettings(node, index, tab_widget, plot_widget,
                                      self)
        
    def remove_from_widget(self):

        for line in self._cluster_lines.itervalues():
            line.remove()

        self._parent_plot.removeMarkers()

        # Remove our implementation of scaledraw
        self._parent_plot.setAxisScaleDraw(QwtPlot.xBottom, QwtScaleDraw())
        Plot2D.remove_from_widget(self)
        
    def get_order(self):
        return self._order

    def get_matrix(self):
        return self._matrix

    def deselect_self(self):
        Plot2D.deselect_self(self)
        for m in self._selected_marker_list:
            m.highlight(False)
        
        self._selected_marker_list = []

    def set_data(self, matrix, order=None, indices=None):
        """Method used by classes using this class to set data from input ports

        WARNING: any exceptions in this method is caught in 'guiplot2dnodesettings'
        """
        from Numeric import concatenate, reshape, array

        height = matrix['height'][:]
        height.sort()

        self._matrix = concatenate([array(matrix['merge']),
                                    reshape(height,
                                            [len(matrix['height']), 1])], 1)
        self._order = matrix['order']

        self._clear_plot()
        
        self._tree_clusters = self._cluster_generator.cluster_builder(copy(self._matrix),
                                                                      self._order)
        self.xaxis = DendrogramXAxis()
        self.xaxis.set_max_value(self._tree_clusters[0].get_height())
        
        self._draw_tree()
        
    def set_cutoff(self, level):
        self._cluster_generator.set_cutoff_level(level)
        self._draw_tree()

    def reset_cutoff(self):
        self._cluster_generator.reset_cutoff()
        self._draw_tree()

    def _draw_tree(self):
        """Traverse through all clusters found within the original cluster."""
        
        for line in self._cluster_lines.itervalues():
            line.remove()
        self._cluster_lines = {}
        for cluster in self._tree_clusters:
            self._cluster_height = cluster.get_height()
            cluster_list = [cluster]
            while cluster_list:
                cluster = cluster_list[0]
                cluster_mode = cluster.get_cluster_mode()

                if cluster_mode == Cluster.OPEN:
                    if cluster.get_left_cluster():
                        cluster_list.append(cluster.get_left_cluster())
                    if cluster.get_right_cluster():
                        cluster_list.append(cluster.get_right_cluster())
                    self._draw_cluster(cluster)

                elif cluster_mode == Cluster.BAR:
                    self._draw_bar_cluster(cluster)

                elif cluster_mode == Cluster.TRIANGLE:
                    self._draw_triangle_cluster(cluster)
                
                cluster_list.remove(cluster)

        if self._label_list:
            self._insert_space_for_labels()
        else:
            self._parent_plot.set_x_axis_autoscale()

    def _draw_cluster(self, cluster):
        """Find and draw cluster lines.
        
        Find which lines are needed to draw a cluster and use
        DendrogramLine to show these lines. Each corner is defined by
        a tuple.
        
        These lines will look something like this


        (left_x, top_y)-----------------(right_x, top_y)
                |                              |
                |                              |
                |                              |
        (left_x, bottom_left_y       (right_x, bottom_right_y)
        """

        left_x = cluster.get_left_index() #left x index
        right_x = cluster.get_right_index() # right x index

        # find bottom left y index
        if cluster.get_left_cluster():
            bottom_left_y =  cluster.get_left_cluster().get_height()
        else:
            bottom_left_y = 0
            
        top_y = cluster.get_height() # top y index

        #find bottom right y index
        if cluster.get_right_cluster():
            bottom_right_y = cluster.get_right_cluster().get_height()
        else:
            bottom_right_y = 0            

        
        bottom_left_y = self._cluster_height - bottom_left_y
        bottom_right_y = self._cluster_height - bottom_right_y
        top_y = self._cluster_height - top_y
        
        x_coords = [left_x, left_x, right_x, right_x]
        y_coords = [bottom_left_y, top_y, top_y, bottom_right_y]

        
        line = _DendrogramLine(self._parent_plot)

        # change x- and y-coordinates in order to rotate
        # dendrogram 90 degrees
        line.set_data(y_coords, x_coords)
        
        # labels will not be set on markers, unless self.marker_list has markers
        if len(self.marker_list) == 0 or \
               len(self.marker_list) != len(self._label_list):
            if not cluster.get_left_cluster():            
                # set dendrogram_plot data index instead of line_plot
                source_marker = line.marker_list[0]
                marker = PlotMarker(self._parent_plot, 0, QwtSymbol.None,
                                    QColor(0, 0, 0), QSize(8, 8), 1)
                marker.setXValue(source_marker.xValue())
                marker.setYValue(source_marker.yValue())
                marker.set_data_index(abs(int(cluster.get_left_object())) - 1)
                marker.set_key(self._parent_plot.insertMarker(marker))
                # store leaf clusters marker for later use
                self.marker_list.append(marker)

            if not cluster.get_right_cluster():
                # set dendrogram_plot data index instead of line_plot
                # store leaf clusters marker for later use
                source_marker = line.marker_list[3]
                marker = PlotMarker(self._parent_plot, 3, QwtSymbol.None,
                                    QColor(0, 0, 0), QSize(8, 8), 1)
                marker.setXValue(source_marker.xValue())
                marker.setYValue(source_marker.yValue())
                marker.set_data_index(abs(int(cluster.get_right_object())) - 1)
                marker.set_key(self._parent_plot.insertMarker(marker))
                # store leaf clusters marker for later use
                self.marker_list.append(marker)

        self._parent_plot.setAxisScaleDraw(2, self.xaxis)
        self._cluster_lines[cluster.get_cluster_number()] = line


    def _draw_bar_cluster(self, cluster):
        """Draw lines needed to show clusters below a cutoff level.

        These lines will look something like this:
            
               (left_x, top_y)-----------------(right_x, top_y)
                |                              |
                |                              |
                |                              |
        (left_x, bottom_left_y       (right_x, bottom_right_y)

        The width of the cluster is predefined."""

        top_y = cluster.get_height()
        bottom_left_y = 0 #bottom_left_y
        bottom_right_y = 0 #bottom_left_y
        left_x  = cluster.get_left_index()
        right_x = cluster.get_right_index()
        middle_x = left_x +((right_x -left_x)/2.0) #middle index
        left_x = middle_x - 0.5 #left_x
        right_x = middle_x + 0.5 #right_x

        # changing coordinates in order to rotate
        # 90 degrees
        bottom_left_y = self._cluster_height - bottom_left_y
        bottom_right_y = self._cluster_height - bottom_right_y
        top_y = self._cluster_height - top_y

        x_coords = [left_x, left_x, right_x, right_x]
        y_coords = [bottom_left_y, top_y, top_y, bottom_right_y]

        # the true argument indicates that the bar will be filled
        line = _DendrogramLine(self._parent_plot, True)

        # changed x- and y-coordinates
        line.set_data(y_coords, x_coords)

        self._parent_plot.setAxisScaleDraw(2, self.xaxis)
        self._cluster_lines[cluster.get_cluster_number()] = line


    def _draw_triangle_cluster(self, cluster):
        """Draw lines are needed to show clusters below a cutoff level.

        These lines will look something like this:

                    middle_x, top_y
                             *
                            / \
                           /   \
                          /     \
                         *       *
                    left_x       right_x
             bottom_left_y       bottom_right_y

             
        Bottom_left_y and bottom_right_y will have predetermened values
        and is not depending on the width of the clusters it is representing.
        """

        top_y = cluster.get_height()#top y
        left_x  = cluster.get_left_index()#left x index
        right_x = cluster.get_right_index() #right x index
        bottom_left_y = 0
        bottom_right_y = 0

        bottom_left_y = self._cluster_height - bottom_left_y
        bottom_right_y = self._cluster_height - bottom_right_y
        top_y = self._cluster_height - top_y
        

        left_x = cluster.get_left_index()
        right_x = cluster.get_right_index()
        middle_x = left_x +((right_x -left_x)/2.0) #middle index

        #draw cluster
        x_coords = [left_x, middle_x, right_x]
        y_coords = [bottom_left_y, top_y, bottom_right_y]

        line = _DendrogramLine(self._parent_plot, True)

        # changed x- and y-coordinates in order
        # to rotate 90 degrees
        line.set_data(y_coords, x_coords)
        
        self._parent_plot.setAxisScaleDraw(2, self.xaxis)
        self._cluster_lines[cluster.get_cluster_number()]= line


    def select_point(self, point):
        for cluster in self._tree_clusters:
            self._cluster_height = cluster.get_height()
            cluster_list = [(cluster, None)]
            while cluster_list:
                cluster = cluster_list[0][0]
                parent =  cluster_list[0][1]
                if cluster.cluster_mode == 0:
                    if cluster.get_left_cluster():
                        cluster_list.append((cluster.get_left_cluster(),
                                             cluster))
                    if cluster.get_right_cluster():
                        cluster_list.append((cluster.get_right_cluster(),
                                             cluster))

                if self._branch_hit(point, cluster):                        
                    cluster.change_cluster_mode(parent)
                    self._update_labels_on_markers(cluster, None, True)
                    self._update_labels()
                    self._draw_tree()
                    self.select_cluster_points(cluster)
                    return True

                elif self._cluster_is_triangle(cluster):
                    if self._triangle_hit(point, cluster):
                        self.select_cluster_points(cluster)
                        return True

                elif self._cluster_is_bar(cluster):
                    if self._bar_hit(point, cluster):
                        self.select_cluster_points(cluster)
                        return True

                cluster_list.remove(cluster_list[0])
        return False


    def _update_labels_on_markers(self, cluster, parent_is_open,
                                  first_run=False):
        if first_run:
            if cluster.is_open():
                parent_is_open = True
            else:
                parent_is_open = False

        if not cluster.is_open():
            parent_is_open = False

        if cluster.get_right_cluster():
            self._update_labels_on_markers(cluster.get_right_cluster(),
                                           parent_is_open)
        else:
            index = abs(int(cluster.get_right_object())) - 1
            if parent_is_open and cluster.is_open():
                self.enable_label(index)
            else:
                self.disable_label(index)

        if cluster.get_left_cluster():
            self._update_labels_on_markers(cluster.get_left_cluster(),
                                           parent_is_open)
        else:
            index = abs(int(cluster.get_left_object())) - 1
            if parent_is_open and cluster.is_open():
                self.enable_label(index)
            else:
                self.disable_label(index)                
    

    def _branch_hit(self, point, cluster):
        """Check if the mouse (point) hits a branch in the dendrogram."""

        # The size of the area around a branching point which is
        # handled as a branch hit.  If a branch is located at (x, y),
        # then a hit is reported if point lies inside the rectangle
        # (x-tolerance, y-tolerance, x+tolerance, y+tolerance).
        tolerance = 8

        cluster_line = self._cluster_lines[cluster.get_cluster_number()]
        left_x = cluster_line.marker_list[1].xValue()
        left_y = cluster_line.marker_list[1].yValue()
        right_x = cluster_line.marker_list[2].xValue()
        right_y = cluster_line.marker_list[2].yValue()

        # if triangle, the middle point is the left one
        if not right_x == left_x:
            middle_x = left_x
            middle_y = left_y
        else:
            middle_x = left_x + abs(left_x - right_x)/2.0
            middle_y = left_y + abs(left_y - right_y)/2.0

        middle_x_transformed = self._parent_plot.transform(QwtPlot.xBottom,
                                                           middle_x)
        middle_y_transformed = self._parent_plot.transform(QwtPlot.yLeft,
                                                           middle_y)

        middle_x2_transformed = copy(middle_x_transformed) + tolerance
        middle_y2_transformed = copy(middle_y_transformed) - tolerance
        
        middle_x1_transformed = copy(middle_x_transformed) - tolerance
        middle_y1_transformed = copy(middle_y_transformed) + tolerance

        middle_x1 = self._parent_plot.invTransform(QwtPlot.xBottom,
                                                   middle_x1_transformed)
        middle_y1 = self._parent_plot.invTransform(QwtPlot.yLeft,
                                                   middle_y1_transformed)
        middle_x2 = self._parent_plot.invTransform(QwtPlot.xBottom,
                                                   middle_x2_transformed)
        middle_y2 = self._parent_plot.invTransform(QwtPlot.yLeft,
                                                   middle_y2_transformed)
        
        return point.x() >= middle_x1 and point.x() <= middle_x2 and point.y() >= middle_y1 and point.y() <= middle_y2

    def _get_marker_data(self, cluster):
        x = []
        y = []
        cluster_line = self._cluster_lines[cluster.get_cluster_number()]
        for marker in cluster_line.marker_list:
            x.append(marker.xValue())
            y.append(marker.yValue())
        return (x, y)

    def _triangle_hit(self, point, cluster):
        """Check if the mouse (point) hits a triangle in the dendrogram."""
        x_coords, y_coords = self._get_marker_data(cluster)
        return point.x() <= x_coords[0] and point.x() >= x_coords[1] and point.y() >= y_coords[0] and point.y() <= y_coords[2]

    def _bar_hit(self, point, cluster):
        """Check if the mouse (point) hits a bar in the dendrogram."""
        x_coords, y_coords = self._get_marker_data(cluster)
        return point.x() >= x_coords[1] and point.x() <= x_coords[3] and point.y() >= y_coords[0] and point.y() <= y_coords[2]

    def _cluster_is_triangle(self, cluster):
        if cluster.get_cluster_mode() == cluster.TRIANGLE:
            return True
        else:
            return False

    def _cluster_is_bar(self, cluster):
        if cluster.get_cluster_mode() == cluster.BAR:
            return True
        else:
            return False

    def select_cluster_points(self, cluster):
        object_list = []
        cluster_list = [cluster]
        while cluster_list:
            cluster = cluster_list[0]
            if cluster.get_left_cluster():
                cluster_list.append(cluster.get_left_cluster())
            else:
                object_list.append(int(cluster.get_left_object())*(-1)-1)
            if cluster.get_right_cluster():
                cluster_list.append(cluster.get_right_cluster())
            else:
                object_list.append(int(cluster.get_right_object())*(-1)-1)

            cluster_list.remove(cluster)

        self.select_values(object_list, self._map_aspect)

    def _clear_plot(self):
        """Removes data from lines used to draw the old dendrogram"""
        for line in self._cluster_lines.itervalues():
            line.remove()
        self.marker_list = []
        self._parent_plot.removeMarkers()
        self._cluster_lines = {}

    def set_labels(self, label_list):
        self._label_list = label_list
        self._parent_plot.replot()

        self._enabled_labels = {}
        for marker in self.marker_list:
            self._enabled_labels[marker.data_index()] = True

    def _insert_space_for_labels(self):
        self._parent_plot.set_x_axis_autoscale()
        x_min = self._parent_plot.get_x_axis_min()        
        x_max = self._parent_plot.get_x_axis_max()
        x_max_transformed = self._parent_plot.transform(QwtPlot.xBottom, x_max)
        x_max_transformed += self.LABEL_X_SPACE
        x_max = self._parent_plot.invTransform(QwtPlot.xBottom,
                                               x_max_transformed)
        self._parent_plot.set_x_axis_scale(x_min, x_max)
        self._parent_plot.replot()

    def _update_labels(self):
        """Sets labels from label_list on the markers."""
        if len(self.marker_list):
            # Get the font currently use, then change point size.
            label_font = self.marker_list[0].font()
            label_font.setPointSize(self.get_label_size())
        for marker in self.marker_list:
            if self._enabled_labels.has_key(marker.data_index()):
                if self._enabled_labels[marker.data_index()]:
                    try:
                        text = str(self._label_list[marker.data_index()])
                    except IndexError:
                        text = ""
                        
                    marker.setLabel(text)
                    marker.setLabelAlignment(Qt.AlignRight | Qt.AlignCenter)
                    marker.setFont(label_font)
                    marker.setEnabled(True)
                else:
                    marker.setEnabled(False)

        self._insert_space_for_labels()
        
    def enable_label(self, index):
        try:
            self._enabled_labels[index] = True
        except KeyError, e:
            print "KeyError", e

    def disable_label(self, index):
        try:
            self._enabled_labels[index] = False
        except KeyError, e:
            print "KeyError", e

    def reset_axis(self):
        self._parent_plot.setAxisScaleDraw(QwtPlot.xBottom, self._old_scale_draw)


    def highlight_indices_by_aspect(self, index_list, aspect):
        """ Highlights markers, labels in this plot, that has been selected"""
        Plot2D.highlight_indices_by_aspect(self, index_list, self._map_aspect)

        if aspect != self._map_aspect:
            return

        markers = {}
        for m in self.marker_list:
            markers[m.data_index()] = m

        for i in index_list:
            try:
                self._selected_marker_list.append(markers[i])
                markers[i].highlight(True)
            except KeyError, e:
                pass

        # need to update plot widget if not called within this window
        self._parent_plot.replot()

    def update(self):
        self._update_labels()

class _DendrogramLine:
    """Dendrogram line which make up a cluster in a dendrogram plot."""
    def __init__(self, parent_plot, fill = False):
        self._curve = QwtPlotCurve(parent_plot)
        self.marker_list = []
        self._parent_plot = parent_plot
        
        pen = self._curve.pen()
        pen.setColor(QColor(0, 0, 255))
        pen.setWidth(2)
        pen.setStyle(Qt.SolidLine)

        if fill:
            self.get_curve().setBrush(QBrush(QColor(0, 0, 255)))

        self._curve_key = parent_plot.insertCurve(self.get_curve())            

    def get_curve(self):
        return self._curve

    def set_data(self, matrix, order):
        """Set subset of x,y as data, as described by indices (0-base)."""
        for m in self.marker_list:
            self._parent_plot.removeMarker(m.key())
        self.marker_list = []

        self.get_curve().setData(matrix, order)

        for i, (x, y) in enumerate(izip(matrix, order)):
            marker = PlotMarker(self._parent_plot, i, QwtSymbol.None,
                                QColor(0, 0, 0), QSize(8, 8), 1)
            marker.setXValue(x)
            marker.setYValue(y)
            marker.set_key(self._parent_plot.insertMarker(marker))
            self.marker_list.append(marker)

        self.get_curve().setStyle(Qt.SolidLine)

    def remove(self):
        for m in self.marker_list:
            self._parent_plot.removeMarker(m.key())
        self._parent_plot.removeCurve(self._curve_key)

class DendrogramXAxis(QwtScaleDraw):
    """Class responsible for setting axis labels"""
    
    def __init__(self):
        QwtScaleDraw.__init__(self)
        self.max_value = 0
    
    def set_max_value(self, max_value): 
        self.max_value = max_value
    
    def label(self, number):
        """This method is responsible for returning a string
        that will be used as lable.
        
        IMPORTANT: scicraft will segfault if this method does not return a string
        """
        number = self.max_value - number
        if number >= 0:
            if isinstance(number, float):
                number = round(number,2)
            return str(number)
        else:
            return ""


class DendrogramPlotSettings(PlotSettings):
    def __init__(self, node, index, tab_widget, qwt_plot_widget, plot):
        """Constructor.

        Parameters:
        node            -- the node to retrieve data from
        index           -- the index in the listbox of the tab_widget
        tab_widget      -- the tab_widget the associated widget is drawn in
        qwt_plot_widget -- the plot_widget in which the curve we control is drawn
        """
        self._qwt_plot_widget = qwt_plot_widget
        self._plot = plot
                
        PlotSettings.__init__(self, node, index, tab_widget, qwt_plot_widget,
                              plot)
        
        self._load_settings()
        self._dendrogram_settings_dialog = DendrogramSettingsDialog(self,
                                                                    tab_widget,
                                                                    plot,
                                                                    qwt_plot_widget)
                                                                    
        self._label_settings_dialog = LabelSettingsDialog(self, node,
                                                          qwt_plot_widget,
                                                          plot)
        
        self._init_widget()
        self._set_widget_values()
        self.update_data_names_combo_boxes()
        self._connect_signals()

    def _init_widget(self):
        """Initiates the associated widget."""
        self._settings_widget = QWidgetFactory.create(
            os.path.join(Config.uidir, 'form_dendrogram_plot_setup.ui'),
            self._tab_widget, self._tab_widget)
        self._line_edit_plot_name   = self._settings_widget.child('line_edit_plot_name')
        self._combo_box_x_axis_data = self._settings_widget.child('combo_box_x_axis_data')
        self._autoupdate_checkbox   = self._settings_widget.child('autoupdate_checkbox')
        self._update_button         = self._settings_widget.child('update_button')
        self._dendrogram_settings_button = self._settings_widget.child('dendrogram_settings')
        self._push_button_label_settings = self._settings_widget.child('push_button_label_settings')


        self._set_inports_in_combo_box(self._combo_box_x_axis_data)

        self._autoupdate_checkbox.setChecked(unrepr(str(self.get_parameter('autoupdate'))))

        self._tab_widget.layout().addWidget(self._settings_widget, 0, 1)

        self._tooltips_button = self._settings_widget.child('push_button_tooltips')
        self._tooltips_button.hide()

    def _load_settings(self):
        # set defaults
        default_input_port = ''
        input_ports = self._node.get_all_input_ports()
        if input_ports:
            default_input_port = input_ports[0].getName()            
        self.set_parameter('x_axis_data', default_input_port)
        self.set_parameter('y_axis_data', default_input_port)
        self.set_parameter('y_col_row', '0')
        self.set_parameter('y_index', '0')
        self.set_parameter('range', '1:end')
        self.set_parameter('name', 'Dendrogram Plot')
        self.set_parameter('autoupdate', False)
        self.set_parameter('labels', '0')
        self.set_parameter('dendrogram_cutoff_enabled', 0)
        self.set_parameter('dendrogram_cutoff_level', 0)

        try:
            params = self._plot.get_settings()
            for key, value in params.iteritems():
                if key == 'dendrogram_cutoff_level' and value:
                    self.set_parameter(key, unrepr(value))
                else:                                       
                    self.set_parameter(key, value)
        except AttributeError, e:
            print "AttributeError:", e

    def _set_widget_values(self):
        if self._combo_box_x_axis_data.count():
            self.set_current_item_x_data(self.get_parameter('x_axis_data'))
        self._line_edit_plot_name.setText(self.get_parameter('name'))
        if self.get_parameter('autoupdate'):
            self.update_data()

    def _connect_signals(self):
        self._settings_widget.connect(self._combo_box_x_axis_data,
                                      SIGNAL('activated(const QString &)'),
                                      self._slot_combo_box_x_axis_data_changed)
        self._settings_widget.connect(self._line_edit_plot_name,
                                      SIGNAL('textChanged(const QString &)'),
                                      self._slot_plot_name_changed)
        self._settings_widget.connect(self._autoupdate_checkbox,
                                      SIGNAL('toggled(bool)'),
                                      self._slot_checkbox_changed)
        self._settings_widget.connect(self._update_button,
                                      SIGNAL('clicked()'),
                                      self.update_data)
        self._settings_widget.connect(self._dendrogram_settings_button,
                                      SIGNAL('clicked()'),
                                      self._slot_dendrogram_settings_clicked)
        self._settings_widget.connect(self._push_button_label_settings,
                                      SIGNAL('clicked()'),
                                      self._slot_label_settings_clicked)

    def _slot_line_setup_clicked(self):
        self._dendrogram_settings_dialog.display_dialog()
                                     
    def _slot_dendrogram_settings_clicked(self):
        self._dendrogram_settings_dialog.display_dialog()

    def _slot_label_settings_clicked(self):
        self._label_settings_dialog.display_dialog()

    def _set_clusters_as_inports(self):
        """Sets the names of the input ports in the given combo box

        Parameters:
        combo_box -- the combo box to set the input ports on.
        """

        def sortitems(d):
            keys = d.keys()[:]
            keys.sort()
            return [(key, d[key]) for key in keys]
        
        self._combo_box_x_axis_data.clear()

        ports = dict([(port.getName(), port) for port in self._node.get_all_input_ports()])
            
        for name, inputport in sortitems(ports):
            if inputport.getData():
                data = inputport.getData().getData()
                if not (isinstance(data, dict) and data.has_key("height") and \
                   data.has_key("order") and data.has_key("merge")):
                    continue
                else:
                    dim_string = " (%d objects)" % len(data['order'])
            else:
                dim_string = ""

            self._combo_box_x_axis_data.insertItem(inputport.getName() + dim_string)

    def update_data_names_combo_boxes(self):
        """updates the combo boxes that contains the names of the inputs."""
        self._set_clusters_as_inports()
        try:
            if self._combo_box_x_axis_data.count():
                self.set_current_item_x_data(self.get_parameter('x_axis_data'))
        except NoSuchParameterError:
            pass #set default current item (top)
        
        try:
            self.set_parameter('x_axis_data',
                               self._info_pattern.sub('',
                                                      str(self._combo_box_x_axis_data.currentText())))
            self.save_settings()
        except AttributeError:
            pass #not yet initialized        
        
    def update_data(self):
        """Updates the plot and sets the data in the plot_widget"""
        PlotSettings.update_data(self)
        
        if not self._combo_box_x_axis_data.count():
            return

        msgheader = None
        info = None
        matrix_port_name = None
        matrix_data = None
        
        matrix_data_name = self.get_parameter('x_axis_data')
        if matrix_data_name.count("("):
            matrix_data_name = matrix_data_name[:matrix_data_name.index("(")-1]
        
        if not self._node.getInputPortByName(matrix_data_name):
            return

        try:
            data = self._node.getInputPortByName(matrix_data_name).getData()
            if not data:
                return
            matrix_data = data.getData()
        except IndexError, e:
            msgheader, info = ('Plot - Index out of bounds', 'One of the indexes on the data is out'+\
                               ' of bounds.\n%s' % e)

        if not matrix_data:
            return
        try:
            self._qwt_plot_widget.set_data_by_plot(self._plot, matrix_data, None, indices=None)
            if self.get_parameter('dendrogram_cutoff_enabled') == '1':
                self._plot.set_cutoff(float(self.get_parameter('dendrogram_cutoff_level')))
            self._label_settings_dialog.apply_settings()
            self._qwt_plot_widget.replot()
        except TypeError, e:
            self._qwt_plot_widget.replot()
            msgheader, info = ('Plot - Wrong type of data', 'The data type coming from the input ' + \
                               'port(s) are invalid.\n %s' %e)
        except Exception, e:
            self._qwt_plot_widget.replot()
            msgheader, info = ("Plot - Invalid data", "Something is wrong with your data set. " +\
                               "Unable to create plot.\n%s" %e)    
       
        if msgheader and info:
            QMessageBox.information(self._tab_widget, msgheader, info)
        self._set_max_value_spin_boxes()
        self.update_data_names_combo_boxes()

class DendrogramSettingsDialog:

    def __init__(self, dendrogram_settings,parent,plot, qwt_plot_widget):
 
        self._dendrogram_settings = dendrogram_settings
        self._plot = plot
        self._qwt_plot_widget = qwt_plot_widget
        self._dendrogram_settings_dialog = QWidgetFactory.create(
            os.path.join(Config.uidir, 'dendrogram_settings.ui'), parent)
        self.init_dialog()
        
    def init_dialog(self):
        self._cutoff_checkbox = self._dendrogram_settings_dialog.child('cutoff_checkbox')
        self._cutoff_level_textbox = self._dendrogram_settings_dialog.child('cutoff_level_textbox')

        # Buttons
        self._button_ok = self._dendrogram_settings_dialog.child('pbutton_ok')
        self._button_cancel = self._dendrogram_settings_dialog.child('pbutton_cancel')
        self._button_apply = self._dendrogram_settings_dialog.child('pbutton_apply')

        # Connect signals
        self._dendrogram_settings_dialog.connect(self._button_ok,
                                                 SIGNAL('clicked()'),
                                                 self._slot_ok)
        self._dendrogram_settings_dialog.connect(self._button_cancel,
                                                 SIGNAL('clicked()'),
                                                 self._slot_cancel)
        self._dendrogram_settings_dialog.connect(self._button_apply,
                                                 SIGNAL('clicked()'),
                                                 self._slot_apply)

    def display_dialog(self):
        """Displays the database info dialog."""

        self._load_settings()
        self._dendrogram_settings_dialog.show()
        
    def _load_settings(self):
        """Loads saved settings and updates the GUI."""
        
        # retrieve default settings
        self._set_default_settings()

        # retrieve saved settings and merge with default settings
        settings = self._plot.get_settings()
        for key in settings: 
            self._dendrogram_settings.set_parameter(key, settings[key])
        
        # set values in GUI elements
        self._cutoff_checkbox.setChecked(int(self._dendrogram_settings.get_parameter('dendrogram_cutoff_enabled')))
        if self._cutoff_checkbox.isOn():
            self._cutoff_level_textbox.setText(str(self._dendrogram_settings.get_parameter('dendrogram_cutoff_level')))

    def _save_settings(self):
        """Saves current settings."""
        # retrieve other plot settings and merge with database info settings
        settings = self._plot.get_settings()
        for key in settings: 
            self._dendrogram_settings.set_parameter(key, settings[key])

        if self._cutoff_checkbox.isOn():
            self._dendrogram_settings.set_parameter('dendrogram_cutoff_enabled', '1')
            value = self._cutoff_level_textbox.text()
            self._dendrogram_settings.set_parameter('dendrogram_cutoff_level', str(value))
        else:
            self._dendrogram_settings.set_parameter('dendrogram_cutoff_enabled', '0')
        self._dendrogram_settings.save_settings()
    
    def _set_default_settings(self):
        self._dendrogram_settings.set_parameter('dendrogram_cutoff_enabled', '0')
        self._dendrogram_settings.set_parameter('dendrogram_cutoff_level', '0')
        #self._cutoff_level_textbox.setText('')
        self._cutoff_checkbox.setChecked(0)

    def _apply_changes(self):
        if self._cutoff_checkbox.isOn():
            self._dendrogram_settings.set_parameter('dendrogram_cutoff_enabled', '1')
            value =  self._cutoff_level_textbox.text()
            self._dendrogram_settings.set_parameter('dendrogram_cutoff_level', str(value))
            self._plot.set_cutoff(float(str(value)))
        else:
            self._dendrogram_settings.set_parameter('dendrogram_cutoff_enabled', '0')
            self._dendrogram_settings.set_parameter('dendrogram_cutoff_level', '0')
            y_max = float(self._qwt_plot_widget.get_y_axis_max())
            self._plot.reset_cutoff()
        self._dendrogram_settings.save_settings()

    def _check_input_level_format(self):
        value =  self._cutoff_level_textbox.text()
        if value == '':
            return True
        try:
            nmb_value = float(str(value))
            self._dendrogram_settings.set_parameter('dendrogram_cutoff_level', value)
            self._plot.set_cutoff(nmb_value)
            return True
        except Exception:
            QMessageBox.information(self._dendrogram_settings_dialog,
                                       'Invalid input format',
                                       'Input was not a number')

    def _slot_ok(self):
        if self._check_input_level_format():
            self._apply_changes()
            self._save_settings()
            self._dendrogram_settings_dialog.close() 

    def _slot_cancel( self):
        self._dendrogram_settings_dialog.close()

    def _slot_apply(self):
        if self._check_input_level_format():
            self._apply_changes()
            

