# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Plots.plot2d.plot2d import Plot2D
from GUI.Plots.plot2d.dataselector import DataSelector
from GUI.Plots.plot2d.imageplot import matrix_to_qimage, ImageRect, blow_up_region
from GUI.Nodes.guiplot2dnodesettings import PlotSettings
from Library.Util.Config import Config
from Library.Util.unrepr import unrepr
from Library.Util import numerical
from Library.Util import lists
from Library.Util.rtester import RTester, RInstallError

from qwt import QwtDiMap, QwtPlotCurve
from qtui import QWidgetFactory
from qt import QObject, SIGNAL, QBrush, QColor, PYSIGNAL
import os
import Numeric
from itertools import izip

connect = QObject.connect


def sub_area_calculation((bounds_x1, bounds_y1, bounds_x2, bounds_y2),
                         (area_x1, area_y1, area_x2, area_y2),
                         (image_width, image_height)):
    """Get coordinates to use with blow_up_region.

    bounds specify entire area used to generate density plot image,
    while area specifies area to blow up. image_size is the size of
    the original image (input to blow_up_region).
    """
    mx = image_width / (bounds_x2 - bounds_x1)
    my = image_height / (bounds_y2 - bounds_y1)

    return ((area_x1 - bounds_x1)*mx + 0.5,
            image_height - (area_y2 - bounds_y1)*my + 0.5,
            (area_x2 - bounds_x1)*mx + 0.5,
            image_height - (area_y1 - bounds_y1)*my + 0.5)


class DensityPlot(Plot2D):

    _map_aspect = 'bin'
    _type = "density_plot"
    _name = "Density Plot"

    def __init__(self, parent_plot, node, config):
        Plot2D.__init__(self, parent_plot, config, [self._map_aspect])
        self._curve = None
        # default resolution
        self._resolution = (64, 64)
        self._x_values = self._y_values = None
        self._image = None
        self._blur_maskwidth = 0
        self._nan_indices = []

    def create_settings(self, index, tab_widget, plot_widget, node):
        return DensityPlotSettings(node, index, tab_widget, plot_widget, self)

    def set_values(self, x_values, y_values):
        if isinstance(x_values,Numeric.ArrayType):
            x_values = lists.flatten(x_values)
        if isinstance(y_values,Numeric.ArrayType):
            y_values = lists.flatten(y_values)
        indices = range(min(len(x_values), len(y_values)))
        nan_indices = []
        if x_values and y_values:
            nanx,nany = numerical.find_all_nan_indices(x_values,y_values)
            if nanx:
                indices = filter(lambda i: not nanx[i], indices)
            if nany:
                indices = filter(lambda i: not nany[i], indices)
                
            nan_indices = numerical.find_unique_nan_indices(nanx,x_values,nany,y_values)
        self._nan_indices= nan_indices
        self._x_values, self._y_values = x_values, y_values
        x_data = map(lambda i : x_values[i], indices)
        y_data = map(lambda i : y_values[i], indices)
        self._indices = indices
        
        self.get_parent_plot().removeCurves()
        curve = QwtPlotCurve(self.get_parent_plot())
        curve.setData(x_data, y_data)
        curve.setStyle(QwtPlotCurve.NoCurve)
        self._curve = self.get_parent_plot().insertCurve(curve)
        if self.should_blur():
             try:
                  RTester.r_install_test()
                  import rpy
                  rpy.r('library(MASS)')
                  count_matrix = rpy.r.kde2d(x_data, y_data, n=self._blur_maskwidth)['z']
                  image = matrix_to_qimage(count_matrix, None, None)
                  # Need to flip the image differently on Windows...
                  if os.name == 'nt':
                      self._image = image.mirror(False, True)
                  else:
                      self._image = image.mirror(True, False)
             except RInstallError, e:
                 pass
        else:
            count_matrix = Numeric.log(self._grid_count(0.9))
            image = matrix_to_qimage(count_matrix, None, None)
            # mirror vertically
            self._image = image.mirror(False, True)

        left, top, right, bottom = self._get_value_area()
        self._parent_plot.set_x_axis_scale(left, right)
        self._parent_plot.set_y_axis_scale(top, bottom)
        

    def should_blur(self):
        return self._blur_maskwidth > 0
    
    def set_blur_maskwidth(self, should_blur):
        self._blur_maskwidth = should_blur

    def remove_from_widget(self):
        Plot2D.remove_from_widget(self)
        if self._curve:
            self.get_parent_plot().removeCurve(self._curve)

    def set_x_resolution(self, x_res):
        """Set how many columns the plot should be divided into."""
        self._resolution = (x_res, self._resolution[1])
        self._plot_config.set_setting("x_res", x_res)

    def set_y_resolution(self, y_res):
        """Set how many rows the plot should be divided into."""
        self._resolution = (self._resolution[0], y_res)
        self._plot_config.set_setting("y_res", y_res)

    def get_resolution(self):
        """Get how many columns and rows are used for plotting."""
        return self._resolution

    def _get_image(self):
        return self._image

    def _get_value_area(self):
        """Get area surrounding counted values, expanded by 5 percent.

        Returns (xmin, ymin, xmax, ymax).
        """
        if not self._x_values:
            return (-1, -1, 1, 1)
        
        if len(self._x_values) == 1:
            x, = self._x_values
            y, = self._y_values
            return (x-1, y-1, x+1, y+1)

        xmin = min(self._x_values)
        xmax = max(self._x_values)
        ymin = min(self._y_values)
        ymax = max(self._y_values)

        return (xmin - (xmax-xmin)*0.05, ymin - (ymax-ymin)*0.05,
                xmax + (xmax-xmin)*0.05, ymax + (ymax-ymin)*0.05)

    def draw(self, painter, rect, x_map, y_map, plot_widget):
        if not self._image:
            return
        
        screen_rect = (rect.left(), rect.top(), rect.right(), rect.bottom())
        screen_area = ImageRect(x_map.invTransform(rect.left()),
                                y_map.invTransform(rect.bottom()),
                                x_map.invTransform(rect.right()),
                                y_map.invTransform(rect.top()))
                                                   
        value_area = ImageRect(*self._get_value_area())
        sub_area = value_area.intersect(screen_area)

        region = sub_area_calculation(value_area.tuple(), screen_area.tuple(),
                                      (self._image.width(),
                                       self._image.height()))

        size = rect.width(), rect.height()
        image = blow_up_region(self._image, region, size)
        pos = x_map.transform(sub_area.x1), y_map.transform(sub_area.y2)

        painter.drawImage(pos[0], pos[1], image)
        self._draw_selection(painter, image, x_map, y_map)

    def _draw_selection(self, painter, image, x_map, y_map):
        brush = QBrush(QColor(255, 0, 0))
        indices = self.get_highlighted_indices_by_aspect(self._map_aspect)

        for i in indices:
            x = x_map.transform(self._x_values[i])
            y = y_map.transform(self._y_values[i])

            painter.fillRect(x, y, 1, 1, brush)

    def _grid_count(self, empty_value=0):
        """Count values within grid-area within bounding rectangle.

        Size of grid-area is size of rectangle divided by resolution.

        This is really a two-dimensional histogram, providing counts of
        values within a grid with the given resolution.
        """
        x_values, y_values = self._x_values, self._y_values
        xmin, ymin, xmax, ymax = self._get_value_area()
        x_resolution, y_resolution = self.get_resolution()
        
        matrix = Numeric.zeros([y_resolution, x_resolution])
        matrix = matrix + empty_value

        # create converter between positions and matrix-indices
        grid_width = (xmax-xmin) / float(x_resolution)
        grid_height = (ymax-ymin) / float(y_resolution)
        x_di_map = QwtDiMap(0, x_resolution-1, xmin+grid_width / 2,
                            xmin+(x_resolution*2 - 1)*grid_width / 2)
        y_di_map = QwtDiMap(0, y_resolution-1, ymin+grid_height / 2,
                            ymin+(y_resolution*2 - 1)*grid_height / 2)

        for (x, y) in zip(x_values, y_values):
            col = x_di_map.transform(x)
            if col < 0 or col >= x_resolution: continue

            row = y_di_map.transform(y)
            if row < 0 or row >= y_resolution: continue

            matrix[row][col] += 1

        return matrix

    def _get_matrix_position(self, x, y):
        # much of this code is still duplicated from _grid_count
        # because of optimalization issues (having _grid_count call
        # this method for each value in the dataset would take
        # forever). feel free to refactor as long as the result is
        # profiled

        xmin, ymin, xmax, ymax = self._get_value_area()
        x_resolution, y_resolution = self.get_resolution()
        
        # create converter between positions and matrix-indices
        grid_width = (xmax-xmin) / float(x_resolution)
        grid_height = (ymax-ymin) / float(y_resolution)
        x_di_map = QwtDiMap(0, x_resolution-1, xmin+grid_width / 2,
                            xmin+(x_resolution*2 - 1)*grid_width / 2)
        y_di_map = QwtDiMap(0, y_resolution-1, ymin+grid_height / 2,
                            ymin+(y_resolution*2 - 1)*grid_height / 2)

        col = x_di_map.transform(x)
        row = y_di_map.transform(y)

        return row, col

    def select_point(self, point):
        if not self._x_values:
            return False

        row, col = self._get_matrix_position(point.x(), point.y())
        indices = self._get_indices_in_position(row, col)

        if indices:
            self.select_values(indices, self._map_aspect)
            Plot2D.select_point(self, point)
            return True

        Plot2D.select_point(self, point)
        return False

    def _get_indices_in_position(self, row, col):
        """Get indices of data in a row, column of the density matrix."""

        x_values, y_values = self._x_values, self._y_values
        xmin, ymin, xmax, ymax = self._get_value_area()
        x_resolution, y_resolution = self.get_resolution()
        
        # create converter between positions and matrix-indices
        grid_width = (xmax-xmin) / float(x_resolution)
        grid_height = (ymax-ymin) / float(y_resolution)
        x_di_map = QwtDiMap(0, x_resolution-1, xmin+grid_width / 2,
                            xmin+(x_resolution*2 - 1)*grid_width / 2)
        y_di_map = QwtDiMap(0, y_resolution-1, ymin+grid_height / 2,
                            ymin+(y_resolution*2 - 1)*grid_height / 2)

        indices = []

        for i, (x, y) in enumerate(zip(x_values, y_values)):
            if x_di_map.transform(x) == col and y_di_map.transform(y) == row:
                indices.append(i)

        return indices

    def select_area(self, rect):
        """Search for points inside given rectangle."""
        Plot2D.select_area(self, rect)

        index_list = self._get_points_within_area(rect.x1(), rect.x2(),
                                                  rect.y1(), rect.y2())
        self.select_values(index_list, self._map_aspect)

    def _get_points_within_area(self, x1, x2, y1, y2):
        index_list = []
        for i, (x, y) in enumerate(izip(self._x_values, self._y_values)):
            if x >= x1 and x <= x2 and y >= y1 and y <= y2:
                index_list.append(i)

        return map(self._indices.__getitem__, index_list)

class DensityPlotSettings(PlotSettings):
    __blur_plot_index_to_mask_width = {0: 0, 1: 2, 2: 4, 3: 8, 4: 16, 5: 32, 6: 64,
                                       7: 128, 8: 256, 9: 512, 10: 1024 }

    def __init__(self, node, index, tab_widget, plot_widget, plot):
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'form_density_plot_setup.ui'),
                                             tab_widget, tab_widget)
        PlotSettings.__init__(self, node, index, tab_widget, plot_widget, plot,
                              self._widget)
        tab_widget.layout().addWidget(self._widget, 0, 1)

        # getting widget childs
        self._name_line_edit = self._widget.child('plot_name_edit')
        self._x_data_combo_box = self._widget.child('x_data_combo_box')
        self._y_data_combo_box = self._widget.child('y_data_combo_box')
        update_button = self._widget.child('update_button')
        self._autoupdate_checkbox = self._widget.child('autoupdate_checkbox')
        self._x_index_spin_box = self._widget.child('x_index_spin_box')
        self._y_index_spin_box = self._widget.child('y_index_spin_box')
        self._x_col_row_combo_box = self._widget.child('x_col_row_combo_box')
        self._y_col_row_combo_box = self._widget.child('y_col_row_combo_box')
        self._x_res_spin_box = self._widget.child('x_res_spin_box')
        self._y_res_spin_box = self._widget.child('y_res_spin_box')
        self._blur_plot_combobox = self._widget.child('blur_plot_combobox')

        # initializing connections
        connect(self._autoupdate_checkbox, SIGNAL("toggled(bool)"),
                self._set_autoupdate)
        connect(update_button, SIGNAL("clicked()"), self.update_data)
        connect(self._name_line_edit, SIGNAL("textChanged(const QString &)"),
                self._slot_plot_name_changed)
        connect(self._x_res_spin_box, SIGNAL("valueChanged(int)"),
                self._x_res_changed)
        connect(self._y_res_spin_box, SIGNAL("valueChanged(int)"),
                self._y_res_changed)
        connect(self._blur_plot_combobox, SIGNAL("activated(int)"),
                self._set_blur_plot)

        # set source from settings
        settings = plot.get_settings()
        x_source_id = settings.get('x_data_source', None)
        y_source_id = settings.get('y_data_source', None)
        x_is_column = not bool(int(settings.get('x_is_row', 0)))
        y_is_column = not bool(int(settings.get('y_is_row', 0)))
        x_index = int(settings.get('x_index', 1))
        y_index = int(settings.get('y_index', 1))

        self._x_data_selector = DataSelector(x_source_id, x_is_column, x_index)
        self._y_data_selector = DataSelector(y_source_id, y_is_column, y_index)

        value = None
        if settings.has_key('name'):
            value = str(settings['name'])
        else:
            value = "Density Plot"

        self.set_parameter('name', value)
        self._name_line_edit.setText(value)

        if settings.has_key('autoupdate'):
            value = unrepr(str(settings['autoupdate']))
            self.set_parameter("autoupdate", value)
            self._autoupdate_checkbox.setChecked(value)
        else:
            self.set_parameter("autoupdate", False)

        value = None
        if settings.has_key('x_res'):
            value = unrepr(str(settings['x_res']))
        else:
            value = 64

        self.set_parameter('x_res', value)
        self._x_res_spin_box.setValue(value)
        self._plot.set_x_resolution(value)

        value = None
        if settings.has_key('y_res'):
            value = unrepr(str(settings['y_res']))
        else:
            value = 64

        self.set_parameter('y_res', value)
        self._y_res_spin_box.setValue(value)
        self._plot.set_y_resolution(value)

        if settings.has_key('blur_plot'):
            value = unrepr(str(settings['blur_plot']))
            self.set_parameter("blur_plot", value)
            index = filter(lambda x : self.__blur_plot_index_to_mask_width[x] == value,
                           self.__blur_plot_index_to_mask_width)
            if index:
                self._blur_plot_combobox.setCurrentItem(index[0])
                #else use default (which is 0)
            
            self._plot.set_blur_maskwidth(value)
        else:
            self.set_parameter("blur_plot", self.__blur_plot_index_to_mask_width[0])

        self._x_data_selector.set_source_combo(self._x_data_combo_box)
        self._x_data_selector.set_mode_combo(self._x_col_row_combo_box)
        self._x_data_selector.set_index_spinbox(self._x_index_spin_box)
        self._y_data_selector.set_source_combo(self._y_data_combo_box)
        self._y_data_selector.set_mode_combo(self._y_col_row_combo_box)
        self._y_data_selector.set_index_spinbox(self._y_index_spin_box)

        connect(self._x_data_selector, PYSIGNAL("dataChanged()"), self._x_settings_changed)
        connect(self._y_data_selector, PYSIGNAL("dataChanged()"), self._y_settings_changed)

        self.update_data_names_combo_boxes()

    def _x_res_changed(self, value):
        self._plot.set_x_resolution(value)
        self.set_parameter('x_res', value)
        if self.get_parameter("autoupdate"):
            self.update_data()

    def _y_res_changed(self, value):
        self._plot.set_y_resolution(value)
        self.set_parameter('y_res', value)
        if self.get_parameter("autoupdate"):
            self.update_data()

    def _x_settings_changed(self):
        self.set_parameter('x_data_source', self._x_data_selector.get_source_id())
        if self._x_data_selector.is_column_mode():
            self.set_parameter('x_is_row', 0)
        else:
            self.set_parameter('x_is_row', 1)
        self.set_parameter('x_index', self._x_data_selector.get_index())
            
        self.save_settings()
        if self.get_parameter("autoupdate"):
            self.update_data()

    def _y_settings_changed(self):
        self.set_parameter('y_data_source', self._y_data_selector.get_source_id())
        if self._y_data_selector.is_column_mode():
            self.set_parameter('y_is_row', 0)
        else:
            self.set_parameter('y_is_row', 1)
        self.set_parameter('y_index', self._y_data_selector.get_index())
            
        self.save_settings()
        if self.get_parameter("autoupdate"):
            self.update_data()

    def _set_autoupdate(self, value):
        self.set_parameter("autoupdate", value)
        self.save_settings()

    def _set_blur_plot(self, index):
        mask_width = self.__blur_plot_index_to_mask_width[index]
        self.set_parameter("blur_plot", mask_width)
        self._plot.set_blur_maskwidth(mask_width)
        self.save_settings()
        if self.get_parameter("autoupdate"):
            self.update_data()

    def _get_port_by_id(self, port_id, ports):
        for port in ports:
            if port.getId() == port_id:
                return port

    def update_data_names_combo_boxes(self):
        self._x_data_selector.set_active_sources(self._node.get_all_input_ports())
        self._y_data_selector.set_active_sources(self._node.get_all_input_ports())

        self.save_settings()

    def update_data(self):
        x_values = self._x_data_selector.get_list()
        y_values = self._y_data_selector.get_list()
        
        if not x_values or not y_values:
            return

        length = min(len(x_values), len(y_values))
        self._plot.set_values(x_values[0:length], y_values[0:length])
