# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Pixmaps.Icons import Icons
from GUI.Plots.plot2d.dataselector import DataSelector
from Library.Util.Config import Config
from Library.Util.gui import ValueListViewItem
from GUI.Plots.plot2d.plotmap import MappableSelection
from Library.Util.rtester import RTester, RInstallError

from itertools import izip
from qt import QObject, SIGNAL, QListView, QListViewItem, PYSIGNAL, Qt, QPixmap
from qt import qApp, QThread, QFileDialog
from qtui import QWidgetFactory
from sets import Set
from threading import Thread
from xml.sax._exceptions import SAXParseException

import os
import math
import cPickle

connect = QObject.connect

class GoBrowser(MappableSelection):
    """Graphical tree-browser for Gene Ontology mapped to plots."""
    def __init__(self, parent, ontology, lookup_function):
        """Create browser for Gene Ontology using lookup_function for lookups.

        ontology is a GeneOntology providing named/structured
        classification of genes given GOIDs.  lookup_function is a
        callback supporting being called as lookup_function(type_id,
        id_list) and return list of [(id, genename, [GO-term ID,
        ...]), ...].
        """
        self._aspect = 'gene indices'
        MappableSelection.__init__(self, [self._aspect])
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          "gobrowser.ui"),
                                             None, parent)
        self._dbwidget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                            "gobrowser_dataselection.ui"),
                                               None, self._widget)

        # create listviews (_GoBrowserListView)
        tabwidget = self._widget.child('go_tree_tabwidget')
        self._molecular_list_view = _GoBrowserListView(tabwidget.page(0),
                                                  'molecular_tree')
        self._biological_list_view = _GoBrowserListView(tabwidget.page(1),
                                                   'biological_tree')
        self._cellular_list_view = _GoBrowserListView(tabwidget.page(2),
                                                 'cellular_tree')
        tabwidget = self._widget.child('go_tree_tabwidget')
        tabwidget.page(0).layout().addWidget(self._molecular_list_view, 0, 0)
        tabwidget.page(1).layout().addWidget(self._biological_list_view, 0, 0)
        tabwidget.page(2).layout().addWidget(self._cellular_list_view, 0, 0)

        self._target_type_combo = self._dbwidget.child('target_type_combo')
        self._master_type_combo = self._dbwidget.child('master_type_combo')
        self._master_available_genes = self._widget.child('master_available_genes')
        self._master_selected_genes = self._widget.child('master_selected_genes')
        self._master_tree_genes = self._widget.child('master_tree_genes')
        self._target_available_genes = self._widget.child('target_available_genes')
        self._target_selected_genes = self._widget.child('target_selected_genes')
        self._target_tree_genes = self._widget.child('target_tree_genes')
        # last looked up gene data - list of tuples (gene_id,
        # gene_name, [goid1, goid2, ...])
        self._target_gene_data = {}
        self._master_gene_data = {}

        #attrib used to check which data is used to generate the trees
        self._tree_data = None
        
        # holds _OntologyGraph instances for master input
        self._master_bio_subgraph = None
        self._master_mol_subgraph = None
        self._master_cel_subgraph = None
        self._master_input_changed = False
        self._master_type = ""

        self._ontology = ontology


        db_connect_button = self._dbwidget.child('download_button')
        connect(db_connect_button, SIGNAL("clicked()"),
                self._lookup_db)

        self._master_selected_action = self._widget.child('master_selected_action')
        connect(self._master_selected_action, SIGNAL('activated()'),
                self._slot_master_selection_tree)
        self._master_all_action = self._widget.child('master_all_action')
        connect(self._master_all_action, SIGNAL("activated()"),
                self._slot_master_full_tree)
        
        self._target_selected_action = self._widget.child('target_selected_action')
        connect(self._target_selected_action, SIGNAL('activated()'),
                self._slot_target_selection_tree)
        self._target_all_action = self._widget.child('target_all_action')
        connect(self._target_all_action, SIGNAL("activated()"),
                self._slot_target_full_tree)

        self._add_selected_genes_action = self._widget.child('add_to_target_action')
        connect(self._add_selected_genes_action, SIGNAL('activated()'),
                self._slot_add_selected_to_target_list)
        self._add_selected_genes_action.setEnabled(False) # disable for now
        
        connect(self._widget.child('cutoff_action'),
                SIGNAL('activated()'),
                self._slot_cutoff_dialog)
        connect(self._widget.child('download_action'), SIGNAL('activated()'),
                self._show_download_selector)
        connect(self._widget.child('save_action'), SIGNAL('activated()'),
                self._slot_save_genedata)
        connect(self._widget.child('load_action'), SIGNAL('activated()'),
                self._slot_load_genedata)

        for list_view in (self._biological_list_view,
                          self._molecular_list_view,
                          self._cellular_list_view):
            connect(list_view, SIGNAL('selectionChanged()'),
                    self._slot_selection_changed)
            connect(list_view, SIGNAL('doubleClicked(QListViewItem*, const QPoint&, int)'),
                    self._slot_item_doubleclicked)
       
        self._lookup_function = lookup_function

        self._target_data_selector = DataSelector()
        self._target_data_selector.set_source_combo(self._dbwidget.child('target_source_combo'))
        self._target_data_selector.set_mode_combo(self._dbwidget.child('target_mode_combo'))
        self._target_data_selector.set_index_spinbox(self._dbwidget.child('target_index_spinbox'))

        self._master_data_selector = DataSelector()
        self._master_data_selector.set_source_combo(self._dbwidget.child('master_source_combo'))
        self._master_data_selector.set_mode_combo(self._dbwidget.child('master_mode_combo'))
        self._master_data_selector.set_index_spinbox(self._dbwidget.child('master_index_spinbox'))

        connect(self._master_data_selector, PYSIGNAL('dataChanged()'),
                self._slot_master_input_changed)
        connect(self._master_type_combo, SIGNAL('activated(const QString &)'),
                self._slot_master_type_changed)

        self._pvalue_cutoff = False
        self._depth_cutoff = False
        self._cutoff_value = 0.0

    def _slot_item_doubleclicked(self, item, point, column):
        """Display information about clicked gene."""
        gene = item.get_value()
        if isinstance(gene, Gene):
            dialog = QWidgetFactory.create(os.path.join(Config.uidir,
                                                        "gene_info.ui"),
                                           None, self._widget)
            dialog.child('id').setText(gene.gene_id())
            dialog.child('name').setText(gene.name())
            dialog.child('locuslink').setText(str(gene.locuslink_id()))
            dialog.exec_loop()

    def _slot_master_input_changed(self):
        self._master_input_changed = True

    def _slot_master_type_changed(self, master_type):
        if master_type != self._master_type:
            self._master_input_changed = True
            self._master_type = str(master_type)

    def _slot_cutoff_dialog(self):
        self._cutoff_dialog = CutOffDialog(self._widget,
                                           (self._depth_cutoff or self._pvalue_cutoff),
                                           self._pvalue_cutoff,
                                           self._cutoff_value)
        result_ = self._cutoff_dialog.show()
        if result_:
            self._pvalue_cutoff = self._cutoff_dialog.is_pvalue()
            self._depth_cutoff = self._cutoff_dialog.is_depth()
            self._cutoff_value = self._cutoff_dialog.get_cutoff_value()

        self._update_cutoff()

    def _slot_master_selection_tree(self):
        indices = self.get_highlighted_indices_by_aspect(self._aspect)
        self._generate_master_list_views(indices)

    def _slot_target_selection_tree(self):
        indices = self.get_highlighted_indices_by_aspect(self._aspect)
        self._generate_target_list_views(indices)

    def _slot_master_full_tree(self):
        self._generate_master_list_views()

    def _slot_target_full_tree(self):
        self._generate_target_list_views()

    def _gene_items(self):
        gene_items = self._biological_list_view._gene_items
        gene_items.extend(self._molecular_list_view._gene_items)
        gene_items.extend(self._cellular_list_view._gene_items)
        return gene_items

    def _slot_selection_changed(self):
        selected_items = self._biological_list_view.get_gene_selection()
        selected_items.extend(self._molecular_list_view.get_gene_selection())
        selected_items.extend(self._cellular_list_view.get_gene_selection())

        # extract indices from items
        selected_ids = [item.get_value().gene_id() for item in selected_items]
        selected_indices = []
        for key, data in self._tree_data.iteritems():
            if data[0] in selected_ids:
                selected_indices.append(key)

        self.deselect_with_mapped()
        self.select_values(selected_indices, self._aspect)

    def _update_cutoff(self):
        listviews = [self._biological_list_view, self._molecular_list_view,
                     self._cellular_list_view]
        if self._pvalue_cutoff:
            for tree in listviews:
                tree.pvalue_cutoff(self._cutoff_value)
        elif self._depth_cutoff:
            for tree in listviews:
                tree.depth_cutoff(self._cutoff_value)
        else:
            for tree in listviews:
                tree.open_all_items()

    def name(self):
        """Name used for display when adding/showing plot mappings."""
        return "Gene Ontology browser"

    def update_ports(self, port_list):
        """Inform browser about new data available in ports.

        Required to update database lookup.
        """
        self._target_data_selector.set_active_sources(port_list)
        self._master_data_selector.set_active_sources(port_list)

    def _lookup_db(self, show_gui=True):
        """Perform database lookup based on selected data source and ID type.

        If show_gui is False the LookupDialog will not be shown (for tests).
        """
        target_list = self._target_data_selector.get_list()
        if not target_list:
            target_list = []
        else:
            target_list = list(target_list) # make sure it is not a tuple (!)
        master_list = self._master_data_selector.get_list()
        if not master_list:
            master_list = []
        else:
            master_list = list(master_list) # make sure it is not a tuple (!)
        try:
            lookup = LookupDialog(self._dbwidget, self._lookup_function,
                                  str(self._target_type_combo.currentText()),
                                  target_list,
                                  str(self._master_type_combo.currentText()),
                                  master_list, silent_startup=not show_gui)
        except FunctionCancelled:
            return

        self._target_gene_data = {}
        if target_list:
            gene_data = lookup.get_result("Target")
            for gene_id, gene_name, ids, llid in gene_data:
                index = target_list.index(gene_id)
                self._target_gene_data[index] = gene_id, gene_name, ids, llid

        if master_list and self._master_input_changed:
            gene_data = lookup.get_result("Master")
            self._master_gene_data = {}
            for gene_id, gene_name, ids, llid in gene_data:
                index = master_list.index(gene_id)
                self._master_gene_data[index] = gene_id, gene_name, ids, llid

            self._master_input_changed = False

        self._update_genes()

    def _update_genes(self):
        self._generate_master_graphs()
        self._update_master_available_genes()
        self._update_target_available_genes()

    def _update_target_available_genes(self):
        self._target_available_genes.setText('<p align="right"><b>%d</b></p>' \
                                      % len(self._target_gene_data))
        self._target_all_action.setEnabled(bool(self._target_gene_data))

    def _update_master_available_genes(self):
        self._master_available_genes.setText('<p align="right"><b>%d</b></p>' \
                                      % len(self._master_gene_data))
        self._master_all_action.setEnabled(bool(self._master_gene_data))        

    def _generate_pvalue_inputs(self, target_bio_graph, target_mol_graph,
                                target_cel_graph):
        """Returns a list containing input for pvalue function.

        The result is a list of 3 tuples in a specific order. Each tuple
        contains:
         * list of goids
         * list of numbers of gene_counts found on and below each term for
           master
         * list of numbers of gene_counts found on and below each term for
           target
        """

        if not self._master_bio_subgraph.get_terms() or \
           not self._master_mol_subgraph.get_terms() or \
           not self._master_cel_subgraph.get_terms():
            return []


        results = []        

        for target, master in [(target_bio_graph, self._master_bio_subgraph),
                                (target_mol_graph, self._master_mol_subgraph),
                                (target_cel_graph, self._master_cel_subgraph)]:
            target_go_terms = list(Set(target.get_terms()))
            goids = [term.get_id() for term in target_go_terms]
            master_genes = []
            target_genes = []
            for term in target_go_terms:
                target_genes.append(term.get_gene_count())
                master_genes.append(master.lookup(term.get_id()).get_gene_count())

            results.append((goids, master_genes, target_genes))

        return results

    def _generate_master_graphs(self):
        allgenes = self._master_gene_data.values()

        graphs = self._generate_graphs(allgenes)
        
        self._master_bio_subgraph = graphs[0]
        self._master_mol_subgraph = graphs[1]
        self._master_cel_subgraph = graphs[2]

    def _generate_master_list_views(self, indices=None):
        goids = Set() 

        if indices == None:
            selected_genes = self._master_gene_data.values()
        else:
            selected_genes = []
            for i in indices:
                if self._master_gene_data.has_key(i):
                    selected_genes.append(self._master_gene_data[i])

        graphs = self._generate_graphs(selected_genes)

        self._biological_list_view.populate(graphs[0])
        self._molecular_list_view.populate(graphs[1])
        self._cellular_list_view.populate(graphs[2])

        # number of genes in all trees should be same as number of
        # selected entries
        self._master_tree_genes.setText('<p align="right"><b>%d</b></p>' \
                                        % len(selected_genes))
        self._target_tree_genes.setText('<p align="right"><b>0</b></p>')
        self._tree_data = self._master_gene_data
        
        self._add_selected_genes_action.setEnabled(True)
        self._master_selected_action.setEnabled(True)
        self._target_selected_action.setEnabled(False)
        
    def _generate_target_list_views(self, indices=None):
        """Generate trees based on indices and last lookup.

        If indices is None, all ids are looked up.
        """
        goids = Set()

        if indices == None:
            selected_genes = self._target_gene_data.values()
        else:
            selected_genes = []
            for i in indices:
                if self._target_gene_data.has_key(i):
                    selected_genes.append(self._target_gene_data[i])

        graphs = self._generate_graphs(selected_genes)

        inputs = self._generate_pvalue_inputs(*graphs)

        if inputs:
            master_gene_counts = [self._master_bio_subgraph.get_root().get_gene_count(),
                                  self._master_mol_subgraph.get_root().get_gene_count(),
                                  self._master_cel_subgraph.get_root().get_gene_count()]

            # add p-values to graphs
            for pvalue_input, graph, gene_count in izip(inputs, graphs, master_gene_counts):
                if graph.get_root():
                    goids, master_count, target_count = pvalue_input
                    pvalues = pvalue_function(gene_count,
                                              graph.get_root().get_gene_count(),
                                              master_count, target_count)
                    graph.add_pvalues(pvalues, goids)
        
        self._biological_list_view.populate(graphs[0])
        self._molecular_list_view.populate(graphs[1])
        self._cellular_list_view.populate(graphs[2])

        # number of genes in all trees should be same as number of
        # selected entries
        self._target_tree_genes.setText('<p align="right"><b>%d</b></p>' \
                                        % len(selected_genes))
        self._master_tree_genes.setText('<p align="right"><b>0</b></p>')
        self._tree_data = self._target_gene_data
        
        self._add_selected_genes_action.setEnabled(False)        
        self._master_selected_action.setEnabled(False)
        self._target_selected_action.setEnabled(True)        
        
    def _generate_graphs(self, genes):
        goids = Set()
        for gene_id, gene_name, ids, llid in genes:
            goids.union_update(ids)

        bio_subgraph = self._ontology.get_biological_process_subgraph(goids)
        mol_subgraph = self._ontology.get_molecular_function_subgraph(goids)
        cel_subgraph = self._ontology.get_cellular_component_subgraph(goids)
        
        self._add_genes_to_graph(genes, bio_subgraph)
        self._add_genes_to_graph(genes, mol_subgraph)
        self._add_genes_to_graph(genes, cel_subgraph)

        return bio_subgraph, mol_subgraph, cel_subgraph        

    def _add_genes_to_graph(self, gene_data, graph):
        for gene_id, gene_name, ids, llid in gene_data:
            gene = Gene(gene_id, gene_name, llid)
            for goid in ids:
                term = graph.lookup(goid)
                if term:
                    term.add_gene(gene)

    def show(self):
        self._widget.show()
        self._widget.raiseW()

    def _show_download_selector(self):
        self._dbwidget.show()
        self._dbwidget.raiseW()

    def _update_target_selected_genes(self):
        indices = self.get_highlighted_indices_by_aspect(self._aspect)
        self._target_selected_genes.setText('<p align="right"><b>%d</b></p>' \
                                     % len(indices))
        self._target_selected_action.setEnabled(bool(indices) and
                                               bool(self._target_gene_data))

    def _update_master_selected_genes(self):
        indices = self.get_highlighted_indices_by_aspect(self._aspect)
        self._master_selected_genes.setText('<p align="right"><b>%d</b></p>' \
                                     % len(indices))
        self._master_selected_action.setEnabled(bool(indices) and
                                               bool(self._master_gene_data))

    def highlight_indices_by_aspect(self, indices, aspect):
        MappableSelection.highlight_indices_by_aspect(self, indices, aspect)
        if self._tree_data == self._master_gene_data:
            self._update_master_selected_genes()
        else:
            self._update_target_selected_genes()
        self._highlight_gene_indices(indices)

    def _clear_selection(self):
        for gene_item in self._gene_items():
            gene_item.set_selected(False)

        self._biological_list_view.selectAll(False)
        self._cellular_list_view.selectAll(False)
        self._molecular_list_view.selectAll(False)

    def _highlight_gene_indices(self, indices):
        if not self._tree_data:
            # no genes has yet been looked up through _lookup_db()
            return

        self._clear_selection()
        
        # filter out id's we don't have        
        known_ids = filter(lambda index: self._tree_data.has_key(index), indices)

        selected_ids = [self._tree_data[index][0] for index in known_ids]
        for item in self._gene_items():
            if item.get_value().gene_id() in selected_ids:
                item.set_selected(True)

    def deselect_self(self):
        MappableSelection.deselect_self(self)
        self._clear_selection()
        self._update_target_selected_genes()
        self._update_master_selected_genes()

    def _save_genedata(self, filename):

        dict_ = {'target_data': self._target_gene_data,
                 'master_data': self._master_gene_data}

        dumpfile = open(filename, 'w')
        cPickle.dump(dict_, dumpfile)
        dumpfile.close()

    def _load_genedata(self, filename):

        loadfile = open(filename)
        dict_ = cPickle.load(loadfile)
        loadfile.close()

        self._target_gene_data = dict_['target_data']
        self._master_gene_data = dict_['master_data']

        self._update_genes()
        status_bar = self._widget.statusBar()
        status_bar.message('%s target and %s master genes loaded' % \
                          (len(self._target_gene_data),
                           len(self._master_gene_data)),
                           3000)

    def _slot_save_genedata(self):
        filename = QFileDialog.getSaveFileName('',
                                               "Gene Data (*.sgd)",
                                               self._widget,
                                               "Save File",
                                               "Save File").latin1()
        
        if filename:
            if not filename.endswith('.sgd'):
                filename += '.sgd'
            self._save_genedata(filename)
            
    def _slot_load_genedata(self):
        filename = QFileDialog.getOpenFileName('',
                                               "Gene Data (*.sgd);;All files (*)",
                                               self._widget, "Open Gene Data",
                                               "Choose a file").latin1()
        if filename:
            self._load_genedata(filename)

    def _slot_add_selected_to_target_list(self):
        indices = self.get_highlighted_indices_by_aspect(self._aspect)

        for index in indices:
            self._target_gene_data[index] = self._master_gene_data[index]

        self._update_genes()        
        
        
class _GoBrowserListView(QListView):

    def __init__(self, parent, name='0'):
        QListView.__init__(self, parent, name)
        self.setSelectionMode(QListView.Extended)
        self.setRootIsDecorated(True)
        self.setResizeMode(QListView.LastColumn)
        self.addColumn('')
        # make sure columnheaders are hidden
        self.header().hide()

        self._gene_items = [] # all listviewitems holding Genes
        self._items = [] # all listviewitems in listview

        self._gene_icon = QPixmap(Icons.gene_icon)
        self._term_icon = QPixmap(Icons.term_icon)

        self._graph = None # the graph that this listview is populated from
        
        # tells whether or not the 'Ctrl' button was pressed during
        # the last mouse press event
        self._control_pressed = False 

    def get_go_terms(self):
        """Return all distinct _OntologyTerms in the listview tree."""
        # because of alternative ids we need to filter out duplicated terms
        return list(Set(self._graph.get_terms()))

    def get_selection(self):
        """Get the selected listviewitems of this listview."""
        selection = [item for item in self._items if item.isSelected()]
        return selection

    def get_gene_selection(self):
        """Get the selected listviewitems which are holding genes.

        If an internal node (i.e., a term listviewitem) is selected
        all of its descendant listviewitems which hold genes, are
        returned.
        """
        selection = []
        for item in self.get_selection():
            if item.childCount() or self._control_pressed:
                # do not add leafnodes which were in the previous
                # selection, unless 'Ctrl' was pressed at the same
                # time as selection was set
                selection.extend(self._find_all_leaf_nodes(item))

        return selection + [item for item in self._gene_items
                            if QListViewItem.isSelected(item)]

    def _find_all_leaf_nodes(self, item):
        if not item.childCount():
            return [item]

        leafs = []
        child = item.firstChild()
        while child:
            leafs.extend(self._find_all_leaf_nodes(child))
            child = child.nextSibling()

        return leafs

    def contentsMousePressEvent(self, mouseEvent):
        if mouseEvent.state() & Qt.ControlButton:
            self._control_pressed = True
            
        QListView.contentsMousePressEvent(self, mouseEvent)

    def contentsMouseReleaseEvent(self, mouseEvent):
        self._control_pressed = False
        QListView.contentsMouseReleaseEvent(self, mouseEvent)

    def populate(self, graph):
        """Populate listview with all terms in ontology graph."""
        # make sure sorting is off, so genes are placed below terms
        self.clear()
        self.setSorting(-1)
        self._gene_items = []
        self._items = []

        root = graph.get_root()
        if root:
            item_text = "%s (%d)" % (root.name(), root.get_gene_count())
            if root.get_p_value():
                item_text += " p=%0.4f" % root.get_p_value()
            root_item = ValueListViewItem(self, root, item_text)
            self._populate_from_term(root_item, root)
            root_item.setOpen(True)
            self._items.append(root_item)

        self._graph = graph
        
    def _populate_from_term(self, parent_item, term):
        """Create ListViewItems recursively for children of term."""
        last_child_item = None
        for child in term.children():
            item_text = "%s (%d)" % (child.name(), child.get_gene_count())
            if child.get_p_value():
                item_text += " p=%0.4f" % child.get_p_value()
            if last_child_item:
                item = ValueListViewItem(parent_item, child, last_child_item,
                                         item_text)
            else:
                item = ValueListViewItem(parent_item, child, item_text)
            self._populate_from_term(item, child)
            last_child_item = item
            # make sure term is open on creation, showing all subitems
            item.setOpen(True)
            item.setPixmap(0, self._term_icon)
            self._items.append(item)
            
        for gene in term.get_genes():
            if last_child_item:
                item = BoldTextValueListViewItem(parent_item, gene,
                                                 last_child_item, gene.name())
            else:
                item = BoldTextValueListViewItem(parent_item, gene, gene.name())
            item.setPixmap(0, self._gene_icon)
            self._gene_items.append(item)
            self._items.append(item)

    def depth_cutoff(self, depth):
        """Closes all items with item.depth() greater than or equal to depth."""
        for item in self._items:
            if item.depth() >= depth:
                item.setOpen(False)
            else:
                item.setOpen(True)

    def pvalue_cutoff(self, pvalue, item=None):
        """Hides items with p-value greater than pvalue.

        If all descendants of item has p-value greater than pvalue,
        the item will be hidden.
        """
        if not item:
            item = self.firstChild()

        if not item.childCount():
            # item is a gene item with no p-value
            return False

        item_pvalue = item.get_value().get_p_value()

        child = item.firstChild()
        children_visible = False
        while child:
            if self.pvalue_cutoff(pvalue, child):
                children_visible = True
            child = child.nextSibling()

        if children_visible or item_pvalue <= pvalue:
            return True
        else:
            item.setVisible(False)
            return False

    def open_all_items(self):
        for item in self._items:
            item.setOpen(True)
            item.setVisible(True)
        
            
            
class BoldTextValueListViewItem(ValueListViewItem):
    def __init__(self, parent, value, *text):
        ValueListViewItem.__init__(self, parent, value, *text)
        self._is_selected = False

    def set_selected(self, bool_):
        self._is_selected = bool_
        QListViewItem.repaint(self)

    def isSelected(self):
        return self._is_selected

    def paintCell(self, painter, *args):
        if self._is_selected:
            painter.font().setBold(True) 
        else:
            painter.font().setBold(False)

        ValueListViewItem.paintCell(self, painter, *args)
            

class Gene:
    def __init__(self, gene_id, name, locuslink_id):
        """Name is the name of the gene returned from eGOn, gene_id is
        the id that is looked up in eGOn"""
        self._gene_id = gene_id
        self._name = name
        self._locuslink_id = locuslink_id
        
    def name(self):
        return self._name

    def gene_id(self):
        return self._gene_id

    def locuslink_id(self):
        return self._locuslink_id

class CutOffDialog:
    def __init__(self, parent, enable_radio = False, pvalue_on = False, cutoff_value = 0.0):
        """Dialog widget used to enable/disable cutoff and enter the cutoff value

        enable_radio - bool, used to enable the buttongroup used for the radiobuttons
        pvalue_on - bool, if the pvalue radiobutton is on or not
        
        This dialog box is modal and exec_loop() will return 1 if OK is clicked
        and 0 if Cancel is clicked. This is defined in the ui file.
        """
        
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'gobrowser_cutoff.ui'),
                                             None, parent)

        self._widget.child('cutoff_check').setChecked(enable_radio)
        if enable_radio:
            self._widget.child('cutoff_radiobuttons_group').setEnabled(enable_radio)
            self._widget.child('cutoff_depth').setChecked(not pvalue_on)
            self._widget.child('cutoff_pvalue').setChecked(pvalue_on) 
            self._widget.child('cutoff_valueedit').setText(str(cutoff_value))

        connect(self._widget.child('cutoff_check'), SIGNAL('toggled(bool)'),
                self._slot_disable_radio)

    def _slot_disable_radio(self, bool_):
        if not bool_:
            self._widget.child('cutoff_depth').setChecked(False)
            self._widget.child('cutoff_pvalue').setChecked(False) 
                
    def show(self):
        return self._widget.exec_loop()
        
    def is_pvalue(self):
        return self._widget.child('cutoff_pvalue').isOn()

    def is_depth(self):
        return self._widget.child('cutoff_depth').isOn()
    
    def get_cutoff_value(self):
        try:
            if self.is_pvalue():
                return float(str(self._widget.child('cutoff_valueedit').text()))
            elif self.is_depth():
                # The string needs to be converted to a float before it can be
                # converted to an integer. 
                return int(float(str(self._widget.child('cutoff_valueedit').text())))
        except ValueError:
            return 0.0


class GeneDb:
    
    BATCH_LIMIT = float(1000) # has to be a float
    
    def lookup(self, id_type, ids, callback = None):
        """Lookup list of gene IDs in eGOn.
        
        Returns list of [(id, genename, [GO-term ID, ...]), ...] for all
        IDs. IDs not known to the database will not be present in the
        result.
        """
        types = {"unigene" : "cluster",
                 "locuslink id" : "locus",
                 "clone id" : "clone",
                 "accession number" : "accno",
                 "affymetrix" : "affy",
                 "operon oligo" : "oligo",
                 "agilent oligo" : "agilent",
                 "image clones" : "image_clone"}
        
        if id_type.lower() in types:
            id_type = types[id_type.lower()]
            
        result = []
        from SOAPpy import WSDL
        server = WSDL.Proxy("http://www.microarray.ntnu.no/ws/server/?wsdl")
        key = "W27ADqBQFHKv9c5IRNYXTBxUye+vUAp6"

        runs = int(math.ceil(len(ids)/GeneDb.BATCH_LIMIT))
        for x in [x*1000 for x in range(runs)]:
            for gene in server.search(key, ids[x:x+1000], id_type):
                # go_terms is magic, so we can't do "if gene.go_terms:"
                if len(gene.go_terms) > 0:
                    go_ids = [term.go_id for term in gene.go_terms]
                    result.append((gene.id, gene.name, go_ids, gene.locuslink))
            if callback:
                callback()
                    

        return result
    

def pvalue_function(master_number, target_number, master_genes, target_genes):
    eGOn_pvalue_R_function = """
    egonP <- function(Narray,Nsig,GenesInMaster,GenesInTarget){                     
    
    GenesInMaster <- as.vector(GenesInMaster)                                        
    GenesInTarget <- as.vector(GenesInTarget)                                       
    
    a <- vector(mode='numeric',length=length(GenesInTarget))                        
    for (gen in 1:length(GenesInTarget)){                                            
    
    M <- matrix(c(Narray,Nsig,GenesInMaster[gen],GenesInTarget[gen]),nrow =        
    2,ncol=2)                                                                       
    ta <- fisher.test(M)                                                           
    a[gen] <- ta$p.value                                                           
    
    }                                                                              
    
    return(list(pval=a))                                                            
    
    } 
    """

    try:
        RTester.r_install_test()
        import rpy
        rpy.r(eGOn_pvalue_R_function)
        result = rpy.r.egonP(master_number, target_number, master_genes, target_genes)
        return result['pval']
    except RInstallError:
        # No result returned
        return [0 for x in target_genes]

class FunctionCancelled(Exception):
    """Raised when a FunctionThread was terminated, and get_return_value called"""


class FunctionThread(QThread):
    """Runs a given function in another thread."""
            
    def __init__(self, function, function_args):
        """Initiate the thread.

        function will be run in this thread, when start is called on this
        thread. The method is finished when running() is False or finished() is
        True. The thread may be terminated by calling terminate.
        """
        QThread.__init__(self)
        self._function = function
        self._function_args = function_args        
        self._return_value = []
        self._was_terminated = False
                
    def run(self):
        try:
            self._return_value = self._function(*self._function_args)
        except SAXParseException:
            self._return_value = []

    def terminate(self):
        QThread.terminate(self)
        self._was_terminated = True

    def get_return_value(self):
        if self._was_terminated:
            raise FunctionCancelled, "Thread was terminated before function returned"
        return self._return_value


class LookupDialog:
    """Display information to user while downloading gene data.

    Also allow user to abort operation.
    """
    def __init__(self, parent, lookup_function, target_type, target_ids,
                 master_type, master_ids, silent_startup=False):
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          "gobrowser_busydialog.ui"),
                                             None, parent)

        connect(self._widget.child("abort"), SIGNAL("clicked()"),
                self.abort_download)
        if not silent_startup:
            self._widget.show()

        self._user_info = self._widget.child("text")
        self._progress_bar = self._widget.child('go_progressbar')
        self._lookup_function  = lookup_function

        self._function_thread = None # only set when it is active

        self._results = {} # holds results from calls to lookup_function

        found_something = False
        aborted = False
        
        master_runs = int(math.ceil(len(master_ids)/GeneDb.BATCH_LIMIT))
        target_runs = int(math.ceil(len(target_ids)/GeneDb.BATCH_LIMIT))

        self._progress_bar.setTotalSteps(master_runs + target_runs)
        self._progress_bar.setProgress(0)
        

        for info_type, type_, ids in [("Master", master_type, master_ids),
                                      ("Target", target_type, target_ids)]:
            if ids:
                try:
                    result = self._threaded_lookup(type_, ids)
                except FunctionCancelled:
                    aborted = True
                    break

                self._results[info_type] = result

                if len(result) == 1:
                    self.add_to_user_info("\n1 %s gene" % info_type)
                elif not result:
                    self.add_to_user_info("\nNo %s genes" % info_type)
                else:
                    self.add_to_user_info("\n%d %s genes" % (len(result), info_type))

                if result and not found_something:
                    found_something = True

        if aborted:
            self.add_to_user_info("\n\nDownload aborted.")
        elif found_something:
            self.add_to_user_info("\n\nDownload complete.")
        elif not master_ids and not target_ids:
            self.add_to_user_info("""

No genes were downloaded as no gene IDs were specified.
Is there data on the port(s) you specified?""")
        else:
            self.add_to_user_info("""

No genes were found in the database.
Did you choose the correct port and ID type?""")

        self._widget.child("abort").setEnabled(False)
        self._widget.child("close").setEnabled(True)

        if aborted:
            raise FunctionCancelled, "Download was aborted by user"

    def add_to_user_info(self, text):
        """Adds text to the user info text field."""
        current_text = str(self._user_info.text())
        self._user_info.setText(current_text + text)

    def _threaded_lookup(self, type_, ids):
        self._function_thread = FunctionThread(self._lookup_function,
                                               (type_,
                                                ids,
                                                self.increment_progress))
        self._function_thread.start()
        while self._function_thread.running():
            qApp.processEvents()
            QThread.msleep(100)

        return_value = self._function_thread.get_return_value()
        self._function_thread = None
        return return_value

    def abort_download(self):
        if self._function_thread:
            self._function_thread.terminate()

    def get_result(self, listtype):
        """Get the result from the call to lookup_function.

        listtype is either the string "Master" or "Target".
        """
        return self._results[listtype]

    def increment_progress(self):
        progress = self._progress_bar.progress() + 1
        self._progress_bar.setProgress(progress)
