# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Plots.plot2d.plot2d import Plot2D
from GUI.Nodes.guiplot2dnodesettings import PlotSettings
from qtui import QWidgetFactory
from Library.Util.Config import Config
from Library.Util.plot import histogramutils
from Library.Util import lists
from Library.Util.rtester import RTester, RInstallError
from GUI.Plots.plot2d.dataselector import DataSelector

import os
import Numeric
from qwt import QwtPlotCurve
from qt import QColor, Qt, QObject, SIGNAL, QBrush, PYSIGNAL


class HistogramPlot(Plot2D):
    _type = "histogram_plot"
    _name = "Histogram Plot"

    def __init__(self, plot_widget, node, config):
        Plot2D.__init__(self, plot_widget, config)
        self._plot_widget = plot_widget
        self._histogram_data = []
        self._bin_range = 0
        self._bin_list = []
        self._bins = 0
        self._bars = []
        
    def set_histogram_data(self, vector, bins):
        """Sets correct data after counting

        vector - data set
        bins - scalar with number of bins that should be used
        """

        
        self._clear_plot()
        if isinstance(vector[0],str):
            return
        if type(vector) == Numeric.ArrayType: 
            self._vector = lists.flatten(vector)
        else:
            self._vector = vector
        self._bins = bins
        try:
            RTester.r_install_test()
            import rpy
            nans = rpy.r.is_na(self._vector)
            if isinstance(nans,list):
                self._vector =  map(lambda (isnan, value) : value,
                                    filter(lambda (i, value) : not nans[i],
                                           enumerate(self._vector)))
            if nans == 1:
                self._vector = None
        except RInstallError, e:
            pass
        if not self._vector: 
            return
        self._bin_range = round((max(self._vector)-min(self._vector))/float(self._bins), 10)
        self._histogram_data = histogramutils.count_data(self._vector[:], bins)

        first = min(self._vector)
        for x in self._histogram_data:
            self._bin_list.append([(first, 0), (first, x), (first+self._bin_range, x),
                                   (first+self._bin_range, 0)])
            first += self._bin_range

        self._bars = []
        for bar_bin in self._bin_list:
            if bar_bin[2][1] == 0 and bar_bin[3][1] == 0:
                continue #skip empty bins
            bar = BarPlot(self._parent_plot)
            bar.set_data(bar_bin)
            self._bars.append(bar)

        max_y = max(self._histogram_data)
        min_x = min(self._vector) 
        max_x = max(self._vector) 

        self._plot_widget.set_y_axis_scale(0, max_y)
        self._plot_widget.set_x_axis_scale(min_x, max_x)

    def create_settings(self, index, tab_widget, plot_widget, node):
        return HistogramPlotSettings(node, index, tab_widget, plot_widget, self)

    def _clear_plot(self):
        for bar in self._bars:
            bar.remove()
            
        self._histogram_data = []
        self._bin_range = 0
        self._bin_list = []
        self._bins = 0
        self._bars = []


    def remove_from_widget(self):
        """Main delete method"""
        self._clear_plot()
        self._parent_plot.remove_plot(self)

class BarPlot:
    
    def __init__(self, parent_plot):
        self._curve = QwtPlotCurve(parent_plot)
        self._parent_plot = parent_plot

        pen = self._curve.pen()
        pen.setColor(QColor(0, 0, 255))
        pen.setWidth(2)
        pen.setStyle(Qt.SolidLine)

        # fill
        self.get_curve().setBrush(QBrush(QColor(0, 0, 127)))

        self._curve_key = parent_plot.insertCurve(self.get_curve())


    def set_data(self, coords):
        # coords is a list with 4 coordinates, bottom left, top left,
        # top right, bottom right

        x_data = []
        y_data = []
        for i in coords:
            x_data.append(i[0])
            y_data.append(i[1])
        self.get_curve().setData(x_data,y_data)
                    
        self.get_curve().setStyle(Qt.SolidLine)

    def remove(self):
        self._parent_plot.removeCurve(self._curve_key)

    def get_curve(self):
        return self._curve


class HistogramPlotSettings(PlotSettings):

    def __init__(self, node, index, tab_widget, qwt_plot_widget, plot):
        self._settings_widget = QWidgetFactory.create(
            os.path.join(Config.uidir, 'form_histogram_plot_setup.ui'),
            tab_widget, tab_widget)
        
        PlotSettings.__init__(self, node, index, tab_widget,
                              qwt_plot_widget, plot, self._settings_widget)
        
        self._tab_widget.layout().addWidget(self._settings_widget, 0, 1)

        x, y = tab_widget.get_widget_id()
        self._settings = plot.get_settings()

        self._name_line_edit = None
        self._input_combo_box = None
        self._column_combo_box = None
        self._index_spin_box = None
        self._bins_spin_box = None
        self._update_button = None

        # set source from settings
        settings = plot.get_settings()
        source_id = settings.get('data_source', None)
        column_mode = not int(settings.get('data_col_row', 0))
        index = int(settings.get('col_row_index', 0)) + 1
        self._data_selector = DataSelector(source_id, column_mode, index)

        self._init_settings()
        self._init_widget()

        self.update_data_names_combo_boxes()
        self.update_data()

    def _init_widget(self):
        """Method that fetches all needed GUI items and sets their values."""
        self._name_line_edit = self._settings_widget.child('line_edit_plot_name')
        self._input_combo_box = self._settings_widget.child('combo_box_data')
        self._column_combo_box = self._settings_widget.child('combo_box_data_col_row')
        self._index_spin_box = self._settings_widget.child('spin_box_data_index')
        self._bins_spin_box = self._settings_widget.child('spin_box_bins')
        self._update_button = self._settings_widget.child('update_button')
        self._auto_check_box = self._settings_widget.child('autoupdate_checkbox')
        
        self._bins_spin_box.setValue(int(self._num_bins))
        self._name_line_edit.setText(str(self._name))
        self._auto_check_box.setChecked(self._autoupdate)
        
        self._bins_spin_box.setMinValue(1)
        self._bins_spin_box.setMaxValue(999999)
        
        QObject.connect(self._update_button, SIGNAL('clicked()'),
                        self.update_data)
        QObject.connect(self._bins_spin_box, SIGNAL('valueChanged(int)'),
                        self._slot_bins_value_changed)
        QObject.connect(self._name_line_edit,
                        SIGNAL('textChanged(const QString &)'),
                        self._slot_plot_name_changed) 
        QObject.connect(self._auto_check_box, SIGNAL('toggled(bool)'),
                        self._slot_autoupdate_changed)

        self._data_selector.set_source_combo(self._input_combo_box)        
        self._data_selector.set_mode_combo(self._column_combo_box)
        self._data_selector.set_index_spinbox(self._index_spin_box)

        QObject.connect(self._data_selector, PYSIGNAL("dataChanged()"), self._update_when_autoupdate)

    def _update_when_autoupdate(self):
        self._save_settings()
        if self.get_parameter('autoupdate'):
            self.update_data()
            
    def _init_settings(self):
        """Method that sets defualt values and possible saved values."""
        self._num_bins = 10
        self._name = 'Histogram Plot'
        self._autoupdate = False

        for key, value in self._settings.iteritems():
            if key == 'num_bins':
                self._num_bins = int(value)
            elif key == 'name':
                self._name = str(value)
            elif key == 'autoupdate':
                if isinstance(value, bool):
                    self._autoupdate = value
                elif value == 'True':
                    self._autoupdate = True
                elif value == 'False':
                    self._autoupdate = False

        if self._settings:
            self._save_settings()

    def _save_settings(self):
        self.set_parameter('num_bins', self._num_bins)
        self.set_parameter('autoupdate', self._autoupdate)
        self.set_parameter('name', self._name)
        self.set_parameter('plot_type', 'histogram_plot')
        self.set_parameter('data_source', self._data_selector.get_source_id())
        self.set_parameter('col_row_index', self._data_selector.get_index()-1)

        if self._data_selector.is_column_mode():
            self.set_parameter('data_col_row', 0)
        else:
            self.set_parameter('data_col_row', 1)

        self.save_settings()        
            
    def update_data_names_combo_boxes(self):
        """Updates data combo box

        Updates the data combo box with the ports names and info
        that can be found in the plot node,
        """
        self._data_selector.set_active_sources(self._node.get_all_input_ports())

    def update_data(self):
        PlotSettings.update_data(self)
        
        data = self._data_selector.get_list()
        if data:
            self._plot.set_histogram_data(data, self._num_bins)

    def _slot_bins_value_changed(self, value):
        self._num_bins = value
        self._save_settings()
        if self.get_parameter('autoupdate'):
            self.update_data()
        
    def _slot_plot_name_changed(self, value):
        self._name = str(value)
        self._save_settings()

    def _slot_autoupdate_changed(self, value):
        self._autoupdate = value
        self._save_settings()

