# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QImage, qRgb, QObject, SIGNAL, QColor, Qt, QBrush
from qtui import QWidgetFactory
from qwt import QwtPlot, QwtAutoScale, QwtDoubleRect, toQImage
from Numeric import array, ravel, transpose, UnsignedInt8,ArrayType
from GUI.Plots.plot2d.plot2d import Plot2D, NoSuchAspectError
from GUI.Nodes.guiplot2dnodesettings import PlotSettings
from math import ceil, floor
from Library.Util import colormap
from Library.Util.Config import Config
from Library.Util.unrepr import unrepr
from Library.Util.rtester import RTester, RInstallError

import os
import Numeric


# helper class
class ImageRect:
    def __init__(self, x1, y1, x2, y2):
        self.x1, self.y1, self.x2, self.y2 = x1, y1, x2, y2

    def __repr__(self):
        return "<ImageRect %s,%s,%s,%s>" % (self.x1, self.y1, self.x2, self.y2)

    def __str__(self):
        return "<ImageRect %s,%s,%s,%s>" % (self.x1, self.y1, self.x2, self.y2)

    def __eq__(self, other_rect):
        if not isinstance(other_rect, ImageRect): return False
        
        return self.x1 == other_rect.x1 and \
               self.y1 == other_rect.y1 and \
               self.x2 == other_rect.x2 and \
               self.y2 == other_rect.y2

    def intersect(self, other_rect):
        intersection = ImageRect(max(other_rect.x1, self.x1),
                                 max(other_rect.y1, self.y1),
                                 min(other_rect.x2, self.x2),
                                 min(other_rect.y2, self.y2))
        return intersection

    def offset(self, offset):
        return ImageRect(self.x1 + offset,
                         self.y1 + offset,
                         self.x2 + offset,
                         self.y2 + offset)
    def copy(self):
        return self.offset(0)

    def expand_to_ints(self):
        """Return closest enclosing rectangle with integer-values."""
        return ImageRect(floor(self.x1), floor(self.y1), ceil(self.x2),
                         ceil(self.y2))

    def scale(self, x_scale, y_scale):
        """Scale rectangle with regards to 0,0."""
        return ImageRect(self.x1 * x_scale, self.y1 * y_scale,
                         self.x2 * x_scale, self.y2 * y_scale)

    def width(self):
        return self.x2 - self.x1

    def height(self):
        return self.y2 - self.y1

    def move(self, dx, dy):
        return ImageRect(self.x1 + dx, self.y1 + dy, self.x2 + dx,
                         self.y2 + dy)

    def tuple(self):
        return self.x1, self.y1, self.x2, self.y2

# helper functions

def matrix_to_qimage(matrix, color_map=None, removed_pixels=None):
    """Convert matrix to 8bit image with grayscale or provided colormap."""

    # convert to Numeric format
    matrix = array(matrix)
    height, width = Numeric.shape(matrix)
    image = QImage(width, height, 8, 256)
    max_value = max(ravel(matrix))
    min_value = min(ravel(matrix))
    
    factor = 255.0 / ((max_value - min_value) or 1)
    offset = min_value
    
    if not color_map:
        color_map = map(lambda i : qRgb(i, i, i), range(256))

    matrix = matrix - offset
    matrix = matrix * factor

    image = toQImage(transpose(matrix.astype(UnsignedInt8)))
    
    for i in range(256):
        image.setColor(i, color_map[i])
    if removed_pixels:
        col = Qt.color0
        for x,y in removed_pixels:
            pindex = image.pixelIndex(y,x)
            image.setColor(pindex,col.rgb())
    return image


def blow_up_region(image, crop_rect, new_size):
    """Return a resized region of an image.

    Select a crop_rect (x1, y1, x2, y2) of an image, using real
    values, and resize it to become new_size (width, height). When
    specifying the crop_rect, pixels in the original image start at
    0.5 and end at n+0.5, where n is width or height, so that the
    center of the topleft pixel is at 1,1.

    If the region of crop_rect is outside the image, only the valid
    image area is returned with a possibly smaller size than specified
    as the new_size variable sets the size of the entire crop_rect
    area.
    """
            
    # change border between pixels to be at integer values (easier to
    # round)

    image_rect = ImageRect(0, 0, image.width(), image.height())
    adjusted = ImageRect(*crop_rect).offset(-0.5)
    # shrink crop rectangle to not go outside original image
    shrunk = adjusted.intersect(image_rect)
    # expand crop rectangle to encompass entire pixels
    expanded = shrunk.expand_to_ints()
    # find the scale to use when resizing
    width, height = new_size

    if 0 in (adjusted.width(), adjusted.height()):
        return image.copy(0,0,0,0)
    
    x_scale = 1.0*width/adjusted.width()
    y_scale = 1.0*height/adjusted.height()

    sub_image = image.copy(expanded.x1, expanded.y1,
                           expanded.width(),
                           expanded.height())

    if sub_image.isNull(): return sub_image

    scaled_image = sub_image.smoothScale(sub_image.width()*x_scale,
                                   sub_image.height()*y_scale)

    # find how much the crop-rectangle must move to reflect change in
    # coordinate system
    scaled_expanded = expanded.scale(x_scale, y_scale)
    
    # scale the crop rectangle
    scaled = shrunk.scale(x_scale, y_scale).move(-scaled_expanded.x1,
                                                 -scaled_expanded.y1)
    return scaled_image.copy(scaled.x1, scaled.y1, scaled.width(),
                             scaled.height())


class ImagePlot(Plot2D):

    _type = "image_plot"
    _name = "Image Plot"
    
    def __init__(self, parent_plot, node, config):
        Plot2D.__init__(self, parent_plot, config, ['rows', 'columns',
                                                    'pixels (by column)',
                                                    'pixels (by row)'])
        self._node = node
        self._image = None

        self._color_map = None
        self._redraws = 0
        self._pixel_removed = []

        self._last_data_size = (0, 0)
        self._selection_mode = 'rows'

    def create_settings(self, index, tab_widget, plot_widget, node):
        return ImagePlotSettings(node, index, tab_widget, plot_widget, self)

    def _get_columns(self):
        return self._last_data_size[1]

    def _get_rows(self):
        return self._last_data_size[0]

    def get_selection_mode(self):
        """Get active selection mode for this plot."""
        return self._selection_mode

    def set_selection_mode(self, mode):
        """Set active selection mode of available aspects."""
        if mode not in self.get_mappable_aspects():
            raise NoSuchAspectError, mode
        self._selection_mode = mode

    def select_area(self, rect):
        mode = self.get_selection_mode()

        index_list = []
        rows = range(int(round(rect.y1())) - 1, int(round(rect.y2())))
        columns = range(int(round(rect.x1())) - 1, int(round(rect.x2())))

        # remove negative values
        rows = filter(lambda value : value >= 0, rows)
        columns = filter(lambda value : value >= 0, columns)
        
        if mode == 'rows':
            index_list = rows
        elif mode == 'columns':
            index_list = columns
        elif mode == 'pixels (by column)':
            for row in rows:
                for column in columns:
                    index_list.append(column*self._get_rows() + row)
        elif mode == 'pixels (by row)':
            for row in rows:
                for column in columns:
                    index_list.append(row*self._get_columns() + column)

        self.select_values(index_list, mode)

        Plot2D.select_area(self, rect)

    def set_color_map(self, color_map):
        self._color_map = color_map
        # replace colors in existing image
        if self._image:
            for i in range(256):
                self._image.setColor(i, color_map[i])
            
    def set_matrix_data(self, matrix, plot_widget):
        """Set current data to be shown.

        The area of the current plot widget will also be
        rescaled/resized to fit plot data, so that the plot area is
        from (0.5, 0.5) to (columns+0.5, rows+0.5).
        """
        if isinstance(matrix[0],list) or isinstance(matrix[0],ArrayType):
            columns = len(matrix[0])
        else:
            matrix = [matrix]
            columns = len(matrix[0])

        if Numeric.array(matrix).typecode() == Numeric.typecodes['Character']:
            return
            
        rows = len(matrix)
        correct_zoom_base = QwtDoubleRect(0.5, 0.5 + columns, 0.5, 0.5 + rows)
        zoom_base_changed =  correct_zoom_base != plot_widget.get_zoom_base()
        nan_matrix = [[0]*columns]*rows
        amatrix = []
        for row in range(rows):
            amatrix.append(matrix[row][:])
        if (rows, columns) != self._last_data_size or zoom_base_changed:
            plot_widget.set_x_axis_scale(0.5, columns + 0.5)
            plot_widget.set_y_axis_scale(0.5, rows + 0.5)

            # Invert Y-axis to make origin top-left
            plot_widget.changeAxisOptions(QwtPlot.yLeft, QwtAutoScale.Inverted, True)            
        try:
            RTester.r_install_test()
            import rpy
            for row in range(rows):
                nan_matrix[row] = rpy.r.is_na(matrix[row])
                pixel_removed = []
            for (row,colvalues) in enumerate(nan_matrix):
                if isinstance(colvalues,list):
                    for col in range(len(colvalues)):
                        if nan_matrix[row][col]==1:
                            pixel_removed.append((row,col) )
                            amatrix[row][col] = 1
        except RInstallError, e:
            pass

        self._image = matrix_to_qimage(amatrix, self._color_map, pixel_removed)
        self._last_data_size = rows, columns

    def highlight_indices_by_aspect(self, index_list, aspect):
        Plot2D.highlight_indices_by_aspect(self, index_list, aspect)
        # we need to replot when selection is changed in other window
        # (e.g., GoBrowser)
        self._parent_plot.replot()
        
    def draw(self, painter, rect, x_map, y_map, plot_widget):
        if y_map.d2() > y_map.d1() and not self._redraws:
            plot_widget.changeAxisOptions(QwtPlot.yLeft, QwtAutoScale.Inverted,
                                          True)
            self._redraws += 1
            plot_widget.replot()
            return
        else:
            self._redraws = 0
            
        if self._image:
            x_d1, x_d2 = map(x_map.invTransform, (rect.left(), rect.right()))
            y_d1, y_d2 = map(y_map.invTransform, (rect.bottom(), rect.top()))
            image = blow_up_region(self._image, (x_d1, y_d2, x_d2, y_d1),
                                   (rect.width(), rect.height()))
            if not image.isNull():
                left, top = rect.left(), rect.top()
                if x_map.invTransform(rect.left()) < 0.5:
                    left = x_map.transform(0.5)
                if y_map.invTransform(rect.top()) < 0.5:
                    top = y_map.transform(0.5)
                painter.drawImage(left, top, image)
                self._draw_selection(painter, image, x_map, y_map)

    def _draw_selection(self, painter, image, x_map, y_map):
        # highlights selection by XOR-ing chosen columns and rows
        painter.setRasterOp(Qt.XorROP)
        brush = QBrush(QColor(255, 255, 255))
        rows, columns, pixels_by_row, pixels_by_column = map(self.get_highlighted_indices_by_aspect,
                                                             ('rows', 'columns',
                                                              'pixels (by row)',
                                                              'pixels (by column)'))
        for row in [row + 1 for row in rows]:
            # draw row selection
            painter.fillRect(x_map.i1(),
                             y_map.transform(row-0.5),
                             x_map.i2() - x_map.i1() + 1,
                             y_map.transform(row+0.5) - y_map.transform(row-0.5),
                             brush)
            
        for column in [column + 1 for column in columns]:
            # draw col selection
            painter.fillRect(x_map.transform(column-0.5),
                             y_map.i2(),
                             x_map.transform(column+0.5) - x_map.transform(column-0.5),
                             y_map.i1() - y_map.i2() + 1, brush)

        for pixel in pixels_by_row:
            row = 1 + pixel / self._get_columns()
            column = 1 + pixel % self._get_columns()
            painter.fillRect(x_map.transform(column-0.5),
                             y_map.transform(row-0.5),
                             x_map.transform(column+0.5) - x_map.transform(column-0.5),
                             y_map.transform(row+0.5) - y_map.transform(row-0.5),
                             brush)
            
        for pixel in pixels_by_column:
            row = 1 + pixel % self._get_rows()
            column = 1 + pixel / self._get_rows()
            painter.fillRect(x_map.transform(column-0.5),
                             y_map.transform(row-0.5),
                             x_map.transform(column+0.5) - x_map.transform(column-0.5),
                             y_map.transform(row+0.5) - y_map.transform(row-0.5),
                             brush)
            
        painter.setRasterOp(Qt.CopyROP)

    def _get_crop_region(self, image_width, image_height, x_map, y_map):
        """Get (x,y,w,h)-rectangle of sub-area, defined by maps."""
        x = int(x_map.d1() - 0.5)
        w = int(x_map.d2() - 0.5) - x
        y = int(y_map.d2() - 0.5)
        h = int(y_map.d1() - 0.5) - y
        return x, y, w, h

    def remove_from_widget(self):
        # need to reset the Y-axis
        self._parent_plot.changeAxisOptions(QwtPlot.yLeft, QwtAutoScale.Inverted, False)
        self._parent_plot.remove_plot(self)
    
        
class ImagePlotSettings(PlotSettings):
    def __init__(self, node, index, tab_widget, plot_widget, plot):

        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'form_image_plot_setup.ui'),
                                             tab_widget, tab_widget)
        PlotSettings.__init__(self, node, index, tab_widget, plot_widget, plot,
                              self._widget)

        tab_widget.layout().addWidget(self._widget, 0, 1)

        self._data_source = None
        self._color_map_name = None
        self._color_map = None
        self._data_sources = []
        self._selection_modes = {}

        self._input_combo_box = self._widget.child('input_data_combo_box')
        self._color_map_combo_box = self._widget.child('color_map_combo_box')
        self._selection_mode_combo_box = self._widget.child('selection_mode_combo_box')
        self._autoupdate_checkbox = self._widget.child('autoupdate_checkbox')
        update_button = self._widget.child('update_button')
        self._plot_name_edit = self._widget.child('plot_name_edit')

        QObject.connect(self._input_combo_box, SIGNAL("activated(int)"),
                        self._set_data_source_by_index)
        QObject.connect(self._color_map_combo_box,
                        SIGNAL("activated(const QString&)"),
                        self.set_color_map)
        QObject.connect(self._selection_mode_combo_box,
                        SIGNAL("activated(const QString&)"),
                        self._set_selection_mode)
        QObject.connect(self._autoupdate_checkbox, SIGNAL("toggled(bool)"),
                        self._set_autoupdate)
        QObject.connect(update_button, SIGNAL("clicked()"), self.update_data)
        
        settings = self._plot.get_settings()
        if settings.has_key('data_source'):
            source_id = settings['data_source']
            for port in self._node.get_all_input_ports():
                if port.getId() == source_id:
                    self._set_data_source(port)

        # set selection modes
        for i, aspect in enumerate(plot.get_mappable_aspects()):
            self._selection_modes[aspect] = i
            self._selection_mode_combo_box.insertItem(aspect, i)
            
        if settings.has_key('selection_mode'):
            mode = settings['selection_mode']
            self._set_selection_mode(mode)
            self._selection_mode_combo_box.setCurrentItem(\
                self._selection_modes[mode])

        if settings.has_key('name'):
            self.set_parameter('name', settings['name'])
        self._plot_name_edit.setText(self.get_parameter('name'))

        # connect after setting initial value
        QObject.connect(self._plot_name_edit,
                        SIGNAL('textChanged(const QString &)'),
                        self._slot_plot_name_changed)        

        if settings.has_key('color_map'):
            self._color_map_name = settings['color_map']
        else:
            self._color_map_name = 'gray'

        self._color_map = colormap.get_color_maps()[self._color_map_name]
        # fix combo box
        for index, color_map_name in enumerate(colormap.get_color_maps()):
            self._color_map_combo_box.insertItem(color_map_name)
            if color_map_name == self._color_map_name:
                self._color_map_combo_box.setCurrentItem(index)

        self.set_color_map(self._color_map_name)

        if settings.has_key('autoupdate'):
            self.set_parameter('autoupdate', settings['autoupdate'])
            self._autoupdate_checkbox.setChecked(unrepr(str(settings['autoupdate'])))
        else:
            self.set_parameter('autoupdate', False)
                           
        self.save_settings()

        self.update_data_names_combo_boxes()

    def _set_autoupdate(self, value):
        self.set_parameter('autoupdate', value)
        self.save_settings()

    def set_color_map(self, color_map_name):
        name = str(color_map_name)
        color_map = colormap.get_color_maps()[name]
        self._plot.set_color_map(color_map)
        self._qwt_plot_widget.replot()
        self.set_parameter('color_map', name)
        self.save_settings()

    def update_data_names_combo_boxes(self):
        self._data_sources = []
        self._input_combo_box.clear()
        for index, port in enumerate(self._node.get_all_input_ports()):
            self._data_sources.append(port) 
            self._input_combo_box.insertItem(port.getName())
            
        if not self._data_source and self._input_combo_box.count():
            self._set_data_source_by_index(0)

        for index, port in enumerate(self._node.get_all_input_ports()):
            if port == self._data_source:
                self._input_combo_box.setCurrentItem(index)

    def update_data(self):
        PlotSettings.update_data(self)
        
        if not self._data_source:
            return

        data = self._data_source.getData()
        
        if data:
            matrix = data.getData()
            self._plot.set_matrix_data(matrix, self._qwt_plot_widget)
            self._qwt_plot_widget.replot()

    def _set_data_source(self, source_port):
        """Specify port containing data to plot."""
        self._data_source = source_port
        self.set_parameter('data_source', source_port.getId())
        self.save_settings()

    def _set_data_source_by_index(self, index):
        self._set_data_source(self._data_sources[index])
        if self.get_parameter('autoupdate'):
            self.update_data()

    def _set_selection_mode(self, mode):
        self._plot.set_selection_mode(str(mode))
        self.set_parameter('selection_mode',
                           str(mode))
        self.save_settings()
