# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.


import os
from qt import QObject, SIGNAL, QPen, Qt
from qwt import QwtPlot
from qtui import QWidgetFactory
from itertools import izip

from GUI.Plots.plot2d.plot2d import Plot2D
from GUI.Plots.plot2d.lineplot import LinePlot
from GUI.Nodes.guiplot2dnodesettings import PlotSettings
from Library.Util.Config import Config
from Library.Util import rangeparser
from Library.Util import numerical
from System.Data.ports import NodePort

class MultiLinePlot(Plot2D):

    _type = "multiline_plot"
    _name = "Multi Line Plot"

    def __init__(self, parent_plot, node, config):
        Plot2D.__init__(self, parent_plot, config, ['lines', 'objects'])
        self._curve_keys = []
        self._colors = Qt.red, Qt.darkYellow, Qt.green, Qt.blue, Qt.cyan, Qt.magenta, Qt.darkRed, Qt.darkGreen, Qt.darkBlue, Qt.darkCyan, Qt.darkMagenta

    def create_settings(self, index, tab_widget, plot_widget, node):
        return MultiLinePlotSettings(node, index, tab_widget, plot_widget, self)

    def set_data(self, x_data, y_data, indices=None):
        self._clear_plot()

        extents = None
        min_x_values = []
        max_x_values = []
        min_y_values = []
        max_y_values = []
        
        make_indices = indices
        for i, (x_values, y_values) in enumerate(izip(x_data, y_data)):
            key = self._parent_plot.insertCurve("curve")

            if not make_indices:
                indices = range(min(len(x_values),len(y_values)))
            (nanx,nany) = numerical.find_all_nan_indices(x_values,y_values)

            if nanx:
                indices = filter(lambda i: not nanx[i], indices)
            if nany:
                indices = filter(lambda i: not nany[i], indices)
            if not indices:
                return

            x_values = map(lambda i : x_values[i], indices)
            y_values = map(lambda i : y_values[i], indices)
                    
            self._parent_plot.setCurveData(key, x_values, y_values)
            self._curve_keys.append(key)

            if self.get_highlighted_indices_by_aspect('lines'):
                pen = QPen(Qt.lightGray)
            else:
                pen = QPen(self._colors[i % len(self._colors)], 1)
            self._parent_plot.setCurvePen(key, pen)
 

        self._parent_plot.replot()

    def remove_from_widget(self):
        self._clear_plot()
        self._parent_plot.remove_plot(self)

    def _clear_plot(self):
        self._parent_plot.clear()
        for key in self._curve_keys:
            self._parent_plot.removeCurve(key)
        self._curve_keys = []

    def deselect_self(self):
        Plot2D.deselect_self(self)
        for i, key in enumerate(self._curve_keys):
            pen = QPen(self._colors[i % len(self._colors)], 1)
            self._parent_plot.setCurvePen(key, pen)

    def select_point(self, point):
        x_map = self._parent_plot.canvasMap(QwtPlot.xBottom)
        y_map = self._parent_plot.canvasMap(QwtPlot.yLeft)
        x_pos = x_map.transform(point.x())
        y_pos = y_map.transform(point.y())
        key, dist, x, y, index = self._parent_plot.closestCurve(x_pos, y_pos)

        if key and dist < 5:
            self.select_values([self._curve_keys.index(key)], "lines")            
        
    def highlight_indices_by_aspect(self, index_list, aspect):
        Plot2D.highlight_indices_by_aspect(self, index_list, aspect)
        if index_list:
            for key in self._curve_keys:
                self._parent_plot.setCurvePen(key, QPen(Qt.lightGray, 1))

    def draw_selection(self, painter, maps, rect):
        # get marked, making sure we dont overstep number of curves
        marked = []
        for index in self.get_highlighted_indices_by_aspect('lines'):
            if index < len(self._curve_keys):
                marked.append(self._curve_keys[index])
                     
        curves = map(self._parent_plot.curve, marked)
        for curve in curves:
            pen = QPen(Qt.red, 2)
            curve.setPen(pen)
            curve.draw(painter, maps[curve.xAxis()], maps[curve.yAxis()])

                

class MultiLinePlotSettings(PlotSettings):
    def __init__(self, node, index, tab_widget, qwt_plot_widget, plot):

        PlotSettings.__init__(self, node, index, tab_widget, qwt_plot_widget,
                              plot)
        
        self._settings_widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                                   'form_multiline_plot_setup.ui'),
                                                      None, tab_widget)
        tab_widget.layout().addWidget(self._settings_widget, 0, 1)
        self._settings = plot.get_settings()
        
        self._x_data_source = 'Indexed'
        self._y_data_source = None
        self._data_sources = []        

        self._init_settings()
        self._init_widget()
        self.update_data_names_combo_boxes()

    def _init_widget(self):
        
        self._name_line_edit = self._settings_widget.child('multiline_edit_plot_name')
        self._x_input_combo_box = self._settings_widget.child('combo_box_x_axis_data')
        self._x_column_combo_box = self._settings_widget.child('combo_box_x_col_row')
        self._x_matrix_range_line_edit = self._settings_widget.child('multiline_edit_x_range')
        self._y_input_combo_box = self._settings_widget.child('combo_box_y_axis_data')
        self._y_column_combo_box = self._settings_widget.child('combo_box_y_col_row')
        self._y_matrix_range_line_edit = self._settings_widget.child('multiline_edit_y_range')
        self._vector_range_line_edit = self._settings_widget.child('multiline_edit_vector_range')
        self._tooltip_button = self._settings_widget.child('push_button_tooltips')
        self._autoupdate_check_box = self._settings_widget.child('autoupdate_checkbox')
        self._update_button = self._settings_widget.child('update_button')

        self._name_line_edit.setText(self._plot_name)
        if self._x_col_row: self._x_column_combo_box.setCurrentItem(0)
        else: self._x_column_combo_box.setCurrentItem(1)
        if self._y_col_row: self._y_column_combo_box.setCurrentItem(0)
        else: self._y_column_combo_box.setCurrentItem(1)
        self._x_matrix_range_line_edit.setText(self._x_matrix_range)
        self._y_matrix_range_line_edit.setText(self._y_matrix_range)
        self._vector_range_line_edit.setText(self._vector_range)
        self._autoupdate_check_box.setChecked(self._autoupdate)
        
        QObject.connect(self._update_button, SIGNAL('clicked()'), self.update_data)
        QObject.connect(self._name_line_edit, SIGNAL('returnPressed()'),
                        self._slot_update_settings_values)
        QObject.connect(self._autoupdate_check_box, SIGNAL('toggled(bool)'),
                        self._slot_update_settings_values)
        QObject.connect(self._x_column_combo_box, SIGNAL("activated(int)"),
                        self._slot_update_settings_values)
        QObject.connect(self._y_column_combo_box, SIGNAL("activated(int)"),
                        self._slot_update_settings_values)
        QObject.connect(self._x_matrix_range_line_edit, SIGNAL('returnPressed()'),
                        self._slot_update_settings_values)
        QObject.connect(self._y_matrix_range_line_edit, SIGNAL('returnPressed()'),
                        self._slot_update_settings_values)
        QObject.connect(self._vector_range_line_edit, SIGNAL('returnPressed()'),
                        self._slot_update_settings_values)

        QObject.connect(self._x_input_combo_box, SIGNAL("activated(int)"),
                        self._set_x_data_source_by_index)
        QObject.connect(self._y_input_combo_box, SIGNAL("activated(int)"),
                        self._set_y_data_source_by_index)
        
        
    def _init_settings(self):
        self._plot_name = 'Multi Plot Line'
        self._x_col_row = False
        self._y_col_row = False
        self._x_matrix_range = '1:end'
        self._y_matrix_range = '1:end'
        self._vector_range = '1:end'
        self._autoupdate = False

        for key, value in self._settings.iteritems():
            if key == 'name':
                self._plot_name = str(value)
            elif key == 'x_col_row':
                if isinstance(value, int):
                    self._x_col_row = not int(value)
                elif value == 'True':
                    self._x_col_row = True
                else:
                    self._x_col_row = False
            elif key == 'y_col_row':
                if isinstance(value, int):
                    self._y_col_row = not int(value)
                elif value == 'True':
                    self._y_col_row = True
                else:
                    self._y_col_row = False
            elif key == 'x_matrix_range':
                self._x_matrix_range = str(value)
            elif key == 'y_matrix_range':
                self._y_matrix_range = str(value)
            elif key == 'vector_range':
                self._vector_range = str(value)
            elif key == 'autoupdate':
                if isinstance(value, bool):
                    self._autoupdate = value
                elif value == 'True':
                    self._autoupdate = True
                elif value == 'False':
                    self._autoupdate = False
            elif key == 'x_data_source':
                for port in self._node.get_all_input_ports():
                    if port.getId() == value:
                        self._set_x_data_source(port)
            elif key == 'y_data_source':                
                for port in self._node.get_all_input_ports():
                    if port.getId() == value:
                        self._set_y_data_source(port)

        if self._settings:
            self.save_settings()

    def update_data(self):
        if not self._x_data_source or not self._y_data_source:
            return

        self._x_matrix_range = str(self._x_matrix_range_line_edit.text())
        self._y_matrix_range = str(self._y_matrix_range_line_edit.text())
        self._vector_range = str(self._vector_range_line_edit.text())
        
        # first test if we have y data:
        y_full_data = self._y_data_source.getData()
        if not y_full_data:
            return
        elif not y_full_data.getData():
            return

        y_data= rangeparser.matlab_parse_matrix_range(self._y_matrix_range,
                                                      y_full_data.getData(),
                                                      self._y_col_row)

        indices = rangeparser.matlab_parse_range(self._vector_range,
                                                 range(len(y_data[0])))
        if indices == range(len(y_data[0])):
            indices = None

        if self._x_data_source == 'Indexed':
            x_data = []
            for i in y_data:
                x_data.append(range(1, len(i) + 1))
        else:
            x_full_data = self._x_data_source.getData()
            if not x_full_data:                
                return
            elif not x_full_data.getData():
                return
            x_data = rangeparser.matlab_parse_matrix_range(self._x_matrix_range,
                                                           x_full_data.getData(),
                                                           self._x_col_row)
            
        self._plot.set_data(x_data, y_data, indices)
        self._save_settings()

    def update_data_names_combo_boxes(self):
        """Updates data combo box

        Updates the data combo box with the ports names and info
        that can be found in the plot node,
        """

        self._data_sources = []
        self._x_input_combo_box.clear()
        self._y_input_combo_box.clear()
        self._x_input_combo_box.insertItem('Indexed')
        self._x_column_combo_box.setEnabled(False)
        self._x_matrix_range_line_edit.setEnabled(False)
        
        for index, port in enumerate(self._node.get_all_input_ports()):
            self._data_sources.append(port)
            dim_string = ''
            if port.datainfo != '':
                dim_string = " (%s)" % port.datainfo.replace('\n', '')
            self._x_input_combo_box.insertItem(str(port.getName() + dim_string))
            self._y_input_combo_box.insertItem(str(port.getName() + dim_string))

        if (not self._x_data_source or not self._y_data_source) and \
               self._x_input_combo_box.count() and self._y_input_combo_box.count():            
            self._set_x_data_source_by_index(0)
            self._set_y_data_source_by_index(0)
            return

        for index, port in enumerate(self._node.get_all_input_ports()):
            if port == self._x_data_source:
                self._x_input_combo_box.setCurrentItem(index + 1)
                self._x_column_combo_box.setEnabled(True)
                self._x_matrix_range_line_edit.setEnabled(True)
            if port == self._y_data_source:
                self._y_input_combo_box.setCurrentItem(index)

    def _set_x_data_source_by_index(self, index):
        if self._data_sources:
            if index == 0:
                self._x_data_source = 'Indexed'
                self._x_column_combo_box.setEnabled(False)
                self._x_matrix_range_line_edit.setEnabled(False)
            else:
                self._x_data_source = self._data_sources[index - 1]
                self._x_column_combo_box.setEnabled(True)
                self._x_matrix_range_line_edit.setEnabled(True)
        else:
            self._x_data_source = None

        self._save_settings()
        if self._autoupdate:
            self.update_data()

    def _set_y_data_source_by_index(self, index):
        if self._data_sources:
            self._y_data_source = self._data_sources[index]
        else:
            self._y_data_source = None

        self._save_settings()
        if self._autoupdate:
            self.update_data()
            
    def _slot_update_settings_values(self, ignore=None):
        self._plot_name = str(self._name_line_edit.text())
        self._x_col_row = not int(self._x_column_combo_box.currentItem())
        self._y_col_row = not int(self._y_column_combo_box.currentItem())
        self._x_matrix_range = str(self._x_matrix_range_line_edit.text())
        self._y_matrix_range = str(self._y_matrix_range_line_edit.text())
        self._vector_range = str(self._vector_range_line_edit.text())
        self._autoupdate = bool(self._autoupdate_check_box.isChecked())
        self._save_settings()

        if self._autoupdate:
            self.update_data()

    def _save_settings(self):
        self.set_parameter('name', self._plot_name)
        self.set_parameter('x_col_row', self._x_col_row)
        self.set_parameter('y_col_row', self._y_col_row)
        self.set_parameter('x_matrix_range', self._x_matrix_range)
        self.set_parameter('y_matrix_range', self._y_matrix_range)
        self.set_parameter('vector_range', self._vector_range)
        self.set_parameter('autoupdate', self._autoupdate)
        if isinstance(self._x_data_source,  NodePort):
            self.set_parameter('x_data_source', self._x_data_source.getId())
        else:
            self.set_parameter('x_data_source', 'Indexed')
        if isinstance(self._y_data_source,  NodePort):
            self.set_parameter('y_data_source', self._y_data_source.getId())

        self.save_settings()
        
    def _set_x_data_source(self, port):
        self._x_data_source = port
        
    def _set_y_data_source(self, port):
        self._y_data_source = port
