# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import sre
from itertools import izip
from sets import Set

class _OntologyTerm:
    """An OntologyTerm holds information on a term (a node in the ontology DAG).

    The parents of an OntologyTerm is extracted from the 'is_a:' tags and the
    'relationship: part_of' tag.
    """

    def __init__(self, id_, name, graphname, definition, comment, alt_ids,
                 subsets):
        self._id = id_
        self._name = name
        self._graphname = graphname
        self._definition = definition
        self._comment = comment
        self._alt_ids = alt_ids
        self._subsets = subsets
        self._parents = []
        self._children = []
        self._genes = Set()
        self._locuslink_ids = Set()
        self._p_value = None

    def get_id(self):
        """Returns the primary identifier of the term"""
        return self._id

    def name(self):
        """Returns the name of the term"""
        return self._name

    def graphname(self):
        """Returns the graphname of the term.

        The graphname is the 'namespace:' tag of the term. The three possible
        graphnames are:
         * biological_process
         * molecular_function
         * cellular_component
        """
        return self._graphname

    def definition(self):
        """Returns the 'def:' tag of the term"""
        return self._definition

    def comment(self):
        """Returns a (optional) comment of the term (may be None)"""
        return self._comment

    def alternate_ids(self):
        """Returns any alternative GO-ids of the term.

        The alternative GO-ids is optional, and if the term does not have any
        alternative GO-ids an empty list is returned.
        """
        return self._alt_ids

    def subsets(self):
        """Returns what this term is a subset of (if defined)

        Subsets defined in GO obo-file
        * goslim_goa "GOA and proteome slim"
        * goslim_yeast "Yeast GO slim"
        * goslim_plant "Plant GO slim"
        * goslim_generic "Generic GO slim"
        * gosubset_prok "Prokaryotic GO subset"
        """
        return self._subsets

    def parents(self):
        """Returns the parent OntologyTerms of the term in a list"""
        return self._parents

    def _add_parent(self, parent):
        """Protected method to set the parents"""
        if parent in self._parents:
            return
        
        self._parents.append(parent)

    def children(self):
        """Returns the children OntologyTerms of the term in a list"""
        return self._children

    def _add_child(self, child):
        """Protected method to set the children"""
        if child in self._children:
            return
        
        self._children.append(child)

    def copy(self):
        """Copies a term without copying references to parents and children"""
        return _OntologyTerm(self.get_id(), self.name(), self.graphname(),
                             self.definition(), self.comment(),
                             self.alternate_ids(), self.subsets())

    def _increase_gene_count(self, locuslink_id):
        oldlength = len(self._locuslink_ids)
        self._locuslink_ids.union_update([locuslink_id])
        if oldlength < len(self._locuslink_ids):
            for parent in self._parents:
                parent._increase_gene_count(locuslink_id)
    
    def add_gene(self, gene):
        """Add gene belonging to this GO term."""
        self._genes.add(gene)
        self._increase_gene_count(gene.locuslink_id())

    def get_genes(self):
        """Get all previously added genes belonging to this GO term."""
        return self._genes

    def get_gene_count(self):
        """Get total number of genes in self and descendents."""
        return len(self._locuslink_ids)

    def set_p_value(self, p_value):
        self._p_value = p_value

    def get_p_value(self):
        return self._p_value


class GeneOntology:
    """Provide an in-memory graph structure of a gene ontology.

    You can extract subgraphs containing only a set of given GOIDs and
    their parents within on of the namespaces molecular_function,
    cellular_component and biological_process.

    An ontology file is parsed and a graph structure (DAG) is built in
    memory.  The graph structure is built from the 'is_a:' tags and
    the 'relationship: part_of' tags in the ontology file. The graph
    structure may be traversed by using the parents()- and
    children()-methods of the OntologyTerm class.
    """
    def __init__(self, filename):
        """Build a new gene ontology from ascii data in filename."""
        self._filename = filename
        self._terms = None

    def get_biological_process_subgraph(self, goid_list):
        """Returns the biological_process subgraph of the GeneOntology.

        The GO-ids in the goid_list are a selection of the graph. All
        leafnodes in the subgraph returned, will be terms with GO-ids
        in goid_list. Some of the terms in goid_list could be internal
        nodes in the subgraph, if a term with goid in goid_list is a
        descendant of such a term.
        """
        return self._get_subgraph('biological_process', goid_list)

    def get_molecular_function_subgraph(self, goid_list):
        """Returns the molecular_function subgraph of the GeneOntology.

        The GO-ids in the goid_list are a selection of the graph. All
        leafnodes in the subgraph returned, will be terms with GO-ids
        in goid_list. Some of the terms in goid_list could be internal
        nodes in the subgraph, if a term with goid in goid_list is a
        descendant of such a term.
        """
        return self._get_subgraph('molecular_function', goid_list)

    def get_cellular_component_subgraph(self, goid_list):
        """Returns the cellular_component subgraph of the GeneOntology.

        The GO-ids in the goid_list are a selection of the graph. All
        leafnodes in the subgraph returned, will be terms with GO-ids
        in goid_list. Some of the terms in goid_list could be internal
        nodes in the subgraph, if a term with goid in goid_list is a
        descendant of such a term.
        """
        return self._get_subgraph('cellular_component', goid_list)

    def _get_subgraph(self, graphname, goid_list):
        if not self._terms:
            parser = _OntologyParser()
            self._terms = parser.build_graph(self._filename)            
        terms = {}
        for term_id in self._terms:
            if self._terms[term_id].graphname() == graphname:
                terms[term_id] = self._terms[term_id]

        return _OntologyGraph(terms, goid_list)
        

class _OntologyGraph:
    """Holds a subgraph of the big ontologygraph.

    All leafnodes of the graph are selected terms, given as the
    selected_goids list in the constructor. All selected terms are not
    necessarily leaf nodes.
    """
    def __init__(self, terms, selected_goids):
        """Constructs a subgraph of the given terms with the selected terms"""
        # holds all terms in the subgraph (a subset of the terms in terms)
        self._terms = {}
        self._build_graph(terms, selected_goids)

    def get_root(self):
        """Returns the root of the subgraph"""
        # only one root in a subgraph
        for term in self._terms.values():
            if not term.parents():
                return term

    def lookup(self, goid):
        """Find term with GO-id goid in the subgraph.

        If it can't be found, None is returned.
        """
        if not self._terms.has_key(goid):
            return None
        
        return self._terms[goid]

    def get_terms(self):
        """Returns a list of all terms in the subgraph."""
        return self._terms.values()

    def add_pvalues(self, p_values, goids):
        """Adds the p_values to the terms with goid in goids."""
        for p_value, goid in izip(p_values, goids):
            term = self.lookup(goid)
            if term:
                term.set_p_value(p_value)
        
    def _build_graph(self, terms, selected_goids):
        # for each term with goid in selected_goids
        selected_terms = []
        for goid in selected_goids:
            if terms.has_key(goid):
                selected_terms.append(terms[goid])
        
        for term in selected_terms:
            # copy gives no children/parents
            if self._terms.has_key(term.get_id()):
                continue
            termcopy = term.copy()
            self._terms[term.get_id()] = termcopy
            for alt_id in term.alternate_ids():
                self._terms[alt_id] = termcopy
            self._recursive_traverse_parents(term)

    def _recursive_traverse_parents(self, original_term):
        if not original_term.parents():
            # root node
            return

        for orig_parent in original_term.parents():
            if self._terms.has_key(orig_parent.get_id()):
                # parent term is already visited
                parent = self._terms[orig_parent.get_id()]
            else:
                # parent term has not been visited, copy it
                parent = orig_parent.copy()
                # put parent into dictionary
                self._terms[parent.get_id()] = parent
                for alt_id in parent.alternate_ids():
                    self._terms[alt_id] = parent
                # traverse parent's parents
                self._recursive_traverse_parents(orig_parent)

            self._terms[original_term.get_id()]._add_parent(parent)
            parent._add_child(self._terms[original_term.get_id()])

    def __str__(self):
        """To allow print()-ing the object nicely in debugging."""
        def string_from_term(term, line_list=[], indent=0):
            line_list.append("\t"*indent + "%s - %s" % (term.get_id(), term.name()))
            for child in term.children():
                string_from_term(child, line_list, indent+1)
            for gene in term.get_genes():
                line_list.append("\t"*(indent+1) + gene.name())
            return line_list

        return "\n".join(string_from_term(self.get_root()))


def _find_first(pattern, string, default=None):
    """Return stripped text of first group in pattern from first match.

    If pattern is not found, default is returned.
    """
    match = sre.search(pattern, string)
    if not match or not match.groups():
        return default
    else:
        return match.group(1)


class _OntologyParser:
    """Parses ontology files.

    The result of a build_graph()-call on a file, is a graph-structure in memory
    and a dictionary to lookup in the graph.
    """
    def __init__(self):
        self._default_graphname = None # default-namespace in GO-file
        self._subsetdefs = {}
        self._parent_ids = {}

    def build_graph(self, filename):
        """Builds the graph structure, based on the returned dict from _parse().

        Adds links between OntologyTerm objects.
        """
        terms = self._parse(filename)
        # build the graph structure
        for term in terms.values():
            # get parent terms of a term
            parents = [terms[parent_id] for parent_id in self._parent_ids[term.get_id()]]
            for parent in parents:
                term._add_parent(parent)
                parent._add_child(term)
        
        return terms

    def _parse(self, filename):
        """Parses the whole file.

        Reads through the whole file on a per-stanza-basis, parsing the first
        stanza as the header of the obo file. All the stanza strings that starts
        with '[Term]' are parsed through the _create_term() method, and all
        stanza strings that starts with '[Typedef]' are ignored. Returned is a
        dictionary with all terms (OntologyTerm) in the file hashed on the
        GO-id. There are no links between the terms at this point.
        """
        f = open(filename)
        header_string = self._read_next_stanza(f)
        self._parse_header(header_string)

        terms = {}

        while True:
            term_string = self._read_next_stanza(f)
            if not term_string:
                break

            if not term_string.startswith("[Term]"):
                continue

            term = self._create_term(term_string)

            if not term:
                continue # don't save obsolete terms

            terms[term.get_id()] = term
            for id_ in term.alternate_ids():
                terms[id_] = term

        return terms

    def _read_next_stanza(self, file_):
        """Reads the next "section" of file_.

        Returns a stanza (either the header of the file, a [Term] stanza or a
        [Typedef] stanza, see http://geneontology.org/GO.format.shtml). Returns
        the stanza with no whitespace at beginnin or end.
        """
        # a stanza is either the header, a [Term] or a [Typedef], separated by a
        # newline character
        stanza = ""
        line = file_.readline()
        while line and line != '\n':
            stanza += line
            line = file_.readline()
            
        return stanza

    def _create_term(self, term_string):
        """Creates an OntologyTerm object from a term_string.

        See http://geneontology.org/GO.format.shtml for details on the file
        format. Tags extracted:
         * id
         * name
         * alt_id
         * namespace
         * def
         * comment
         * is_a
         * relationship: part_of 
         * subset
         * is_obsolete: true
        """
        if sre.search("\nis_obsolete:\s*true", term_string):
            return None

        # comments in the ontology file is handled where necessary, as those
        # regexes that needs to handle them, stops matching at '!' (exclamation
        # mark) which is the start of comment token
        goid = _find_first("\nid:\s*(\w*:?[0-9]+)", term_string)
        name = _find_first("\nname:\s*(.*)", term_string,
                          "Unknown name").split("!")[0].strip()
        graphname = _find_first("\nnamespace:\s*([\w|\s]*(?= \!.*|\n))",
                               term_string, self._default_graphname)
        definition = _find_first("\ndef:\s*(.*)\n", term_string)
        comment = _find_first("\ncomment:\s*(.*)", term_string)

        alt_ids = sre.findall("\nalt_id:\s*(\w*:?[0-9]+)", term_string)
        is_as = sre.findall("\nis_a:\s*(\w*:?[0-9]+)", term_string)
        part_ofs = sre.findall("\nrelationship:\s*part\_of\s*(\w*:?[0-9]+)",
                               term_string)

        subsets = []
        subset_keys = sre.findall("\nsubset:\s*(\w+)", term_string)
        for key in subset_keys:
            if self._subsetdefs.has_key(key):
                subsets.append(self._subsetdefs[key])

        is_as = [is_a.strip() for is_a in is_as]
        part_ofs = [part_of.strip() for part_of in part_ofs]

        self._parent_ids[goid] = is_as + part_ofs
        
        return _OntologyTerm(goid, name, graphname, definition, comment,
                             alt_ids, subsets)

    def _parse_header(self, header_string):
        """Parses the header of the .obo-file.

        See http://geneontology.org/GO.format.shtml for details on the header
        fields. Tags extracted:
         * default-namespace
         * subsetdef
        """
        default_graphname = sre.findall('\ndefault-namespace:\s*(.*)',
                                        header_string)[0]
        self._default_graphname = default_graphname

        subsetdefs = sre.findall('subsetdef:\s*(\w*)\s+\"([\w|\s]*)\"',
                                 header_string)
        for subsetdef in subsetdefs:
            self._subsetdefs[subsetdef[0]] = subsetdef[1]
