# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import Qt, QPen, QColor, SIGNAL, QPixmap, qRed, qGreen, qBlue, qRgb
from qt import QScrollView, QImage, QPixmap, QPoint, QPainter, QObject
from qt import QPopupMenu, QAction, QMessageBox
from qtui import QWidgetFactory
from qtcanvas import QCanvasView, QCanvas, QCanvasSprite, QCanvasPixmapArray
from qtcanvas import QCanvasRectangle, QCanvasLine

import os
from sets import Set

from Library.Util.Config import Config
from Library.Util.lists import flatten
from GUI.NodeConfig import NodeConfig
from GUI.links import create_connection

connect = QObject.connect


class OutlierWidget:

    def __init__(self, plot_widget, node):
        self._plot_widget = plot_widget
        self._plot_node = node
        self._last_filternode = None
        self._canvas_view = None
        
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'filter_widget.ui'),
                                             None, self._plot_widget)
        self._widget.connect(self._widget.child('add_filter_button'),
                            SIGNAL('clicked()'), self._slot_add_filter_clicked)
        
    def show(self, current_diagram_nodes):
        self._create_canvas(current_diagram_nodes)
        self._widget.show()

    def _create_canvas(self, current_diagram_nodes):
        if self._canvas_view:
            self._canvas_view.hide()
        self._canvas_view = OutlierCanvasView(self._widget.child('canvas_box'),
                                              self._plot_node,
                                              current_diagram_nodes)

    def _add_filter_to_selected_node(self):
        node = self._canvas_view.get_selected_node()
        if not node:
            return
        
        self._last_filternode = node

        filters = []
        for plot in self._plot_widget.get_plots():
            n = plot.get_data_length()
            for aspect in plot.get_mappable_aspects():
                indices = plot.get_highlighted_indices_by_aspect(aspect)
                self._last_filternode.add_data_filter(n, indices)
                filters.append((n, indices))
        self._canvas_view.update_nodes()

        return filters
        
    def _slot_add_filter_clicked(self):
        filters = self._add_filter_to_selected_node()
        inactive_filters = False
        for n, indices in filters:
            values = flatten(self._last_filternode.get_last_dimensions())
            if n not in values:
                inactive_filters = True

        if inactive_filters:
            QMessageBox.warning(self._widget, "Possibly inactive filter", "Node <b>%s</b> has not yet processed any data matching the filter you added. Make sure you added the filter to the correct node." % self._last_filternode.getName())

    def get_last_filternode(self):
        return self._last_filternode

        
class OutlierCanvasView(QCanvasView):

    def __init__(self, parent, plot_node, diagram_nodes):
        QCanvasView.__init__(self, parent)

        self._canvas = QCanvas(Config.appWidth * OutlierCanvasNode.RESIZE_SCALE,
                               Config.appHeight * OutlierCanvasNode.RESIZE_SCALE)

        self._canvas.setBackgroundColor(QColor(Config.bgR, Config.bgG, Config.bgB))
        self.setCanvas(self._canvas)

        self._diagram_nodes = diagram_nodes

        self.setHScrollBarMode(QScrollView.AlwaysOn)
        self.setVScrollBarMode(QScrollView.AlwaysOn)

        self._plot_node = plot_node
        self._canvas_nodes = []
        self._selected_node = None
        self._add_canvas_nodes()

    def update_nodes(self):
        self._canvas_nodes = []
        for item in self._canvas.allItems():
            item.setCanvas(None)
        self._selected_node = None
        self._add_canvas_nodes()
        self._canvas.update()

    def _remove_all_filters(self):
        for node in self._diagram_nodes:
            node.remove_filters()
        
    def resizeEvent(self, e):
        QCanvasView.resizeEvent(self, e)
        width = e.size().width()
        height = e.size().height()

        self.resize_canvas_to_fit()

    def resize_canvas_to_fit(self):
        xmin, ymin, xmax, ymax = self.bounding_rect()

        visible_size = (self.visibleWidth(), self.visibleHeight())

        visible_scrollbars_size = (self.visibleWidth() - self.verticalScrollBar().width(),
                                   self.visibleHeight() - self.horizontalScrollBar().height())
        
        canvas_width, canvas_height = self.find_canvas_size((xmax, ymax),
                                                       visible_size,
                                                       visible_scrollbars_size)
        self.canvas().resize(canvas_width, canvas_height)

    def bounding_rect(self):
        """Return bounding rectangle of a set of nodes.

        Returned value is tuple (xmin, ymin, xmax, ymax)
        """
        items = self.canvas().allItems()
        if not items:
            return (0,0,0,0)

        rect = items[0].boundingRect()
        #xmin, ymin, xmax, ymax = items[0].boundingRect()
        xmin, ymin, xmax, ymax = rect.left(), rect.top(), rect.right(), rect.bottom()
        
        for item in items[1:]:
            rect = item.boundingRect()
            #xmin_, ymin_, xmax_, ymax_ = item.boundingRect()
            xmin_, ymin_, xmax_, ymax_ = rect.left(), rect.top(), rect.right(), rect.bottom()
            
            xmin = min(xmin, xmin_)
            ymin = min(ymin, ymin_)
            xmax = max(xmax, xmax_)
            ymax = max(ymax, ymax_)
            
        return (xmin, ymin, xmax, ymax)

    def find_canvas_size(self, items_size, visible_size, visible_scrollbars_size):
        """Calculate size of canvas to show scrollbars only when necessary.

        nodes_size is (max_x, max_y) of all nodes
        visible_size is (width, height) of view without scrollbars
        visible_scrollbars_size is (width, height) of view with scrollbars

        Return (width, height) tuple of best canvas size.
        """
        if items_size[0] <= visible_size[0] and items_size[1] <= visible_size[1]:
            return visible_size
        else:
            canvas_width = max(items_size[0], visible_scrollbars_size[0])
            canvas_height = max(items_size[1], visible_scrollbars_size[1])
            return (canvas_width, canvas_height)
        
    def contentsMouseReleaseEvent(self, event):
        """Method overridden from QCanvasView"""
        point = self.inverseWorldMatrix().map(event.pos())

        items = filter(lambda i : not isinstance(i, QCanvasRectangle),
                       self.canvas().collisions(point))

        if not items:
            item = None
        else:
            item = items[0]
            
        if event.button() == Qt.RightButton:
            menu = QPopupMenu(self)
            if item:
                for text, function in item.get_menu_choices():
                    action = QAction(text, 0, self)
                    action.addTo(menu)
                    connect(action, SIGNAL('activated()'), function)
            action = QAction("Remove all filters from all nodes", 0, self)
            action.addTo(menu)
            connect(action, SIGNAL('activated()'), self._remove_all_filters)
            menu.exec_loop(event.globalPos())
            self.update_nodes()
        elif event.button() == Qt.LeftButton:
            if item and item.is_selectable():
                self._set_selected_node(item)

    def _add_canvas_nodes(self):
        all_nodes = Set(self._diagram_nodes)
        color_nodes = Set([self._plot_node] + self._plot_node.get_all_providing_nodes())
        gray_nodes = all_nodes - color_nodes

        canvas_nodes = {}

        # create all colored nodes (connected to plot)
        for node in color_nodes:
            canvas_node = OutlierCanvasNode(node, self._canvas)
            canvas_nodes[node] = canvas_node

        # create all other nodes
        for node in gray_nodes:
            canvas_node = OutlierCanvasNode(node, self._canvas, grayscale=True)
            canvas_nodes[node] = canvas_node

        # connect nodes with lines and display
        for node, target in canvas_nodes.items():
            for source in [canvas_nodes[provider] for provider in node.get_providing_nodes()]:
                # find start and end points of line
                source_point = (source.x() + source.get_out_point().x(),
                                source.y() + source.get_out_point().y())
                target_point = (target.x() + target.get_in_point().x(),
                                target.y() + target.get_in_point().y())
                
                line_points = create_connection(source_point, target_point)
                
                for (x1, y1), (x2, y2) in zip(line_points[:-1], line_points[1:]):
                    line = OutlierCanvasLineSegment((x1, y1), (x2, y2), self._canvas)
                    line.show()

            target.show()
            self._canvas_nodes.append(target)

    def get_selected_node(self):
        if self._selected_node:
            return self._selected_node.get_node()
        else:
            return None

    def _set_selected_node(self, item):
        for canvas_node in self._canvas_nodes:
            if item == canvas_node:
                item.select(True)
            else:
                canvas_node.select(False)
        
        self._selected_node = item

    def __del__(self):
        # This is needed to keep from segfaulting
        for item in self._canvas.allItems():
            item.setCanvas(None)

        self._canvas.update()


class OutlierCanvasNode(QCanvasSprite):

    RESIZE_SCALE = 0.6
    
    def __init__(self, node, canvas, grayscale=False):

        self._node = node
        self._select_rectangle = None
        
        self._config = NodeConfig(self._node.getNodeRepName())
        fullsize_image = QImage(self._config.getImage())
        width, height = fullsize_image.width(), fullsize_image.height()
        self._image = fullsize_image.smoothScale(width * OutlierCanvasNode.RESIZE_SCALE,
                                                height * OutlierCanvasNode.RESIZE_SCALE)

        if grayscale:
            self._image = self._image.convertDepth(8)
            for i in range(self._image.numColors()):
                rgb = self._image.color(i)
                gray = (qRed(rgb) + qGreen(rgb) + qBlue(rgb)) / 3
                self._image.setColor(i, qRgb(gray, gray, gray))
        
        self._pixmap = QPixmap(self._image)

        if node.get_data_filters():
            filter_image = QImage(os.path.join(Config.iconsdir, "filter.png"))
            painter = QPainter(self._pixmap)
            painter.drawImage(self._image.width() - filter_image.width(), 0,
                          filter_image)
            painter.end()
        
        self._array = QCanvasPixmapArray([self._pixmap])

        QCanvasSprite.__init__(self, self._array, canvas)
        
        x, y = self._node.getPosition()
        self.setX(x * OutlierCanvasNode.RESIZE_SCALE)
        self.setY(y * OutlierCanvasNode.RESIZE_SCALE)

    def get_node(self):
        return self._node

    def select(self, set_selected):
        if set_selected:
            bounding_rect = self.boundingRect()
            self._select_rectangle = QCanvasRectangle(bounding_rect, self.canvas())
            self._select_rectangle.setZ(self.z() + 1)
            self._select_rectangle.setPen(QPen(Qt.red))
            self._select_rectangle.show()
            self.canvas().update()
        else:
            if self._select_rectangle:
                self._select_rectangle.setCanvas(None)
                self._select_rectangle = None
                self.canvas().update()

    def get_in_point(self):
        point = self._config.get_inpoints()[0][0]
        
        return QPoint(point[0] * OutlierCanvasNode.RESIZE_SCALE,
                      point[1] * OutlierCanvasNode.RESIZE_SCALE)
        
    def get_out_point(self):
        point = self._config.get_outpoints()[0][0]

        return QPoint(point[0] * OutlierCanvasNode.RESIZE_SCALE,
                      point[1] * OutlierCanvasNode.RESIZE_SCALE)

    def is_selectable(self):
        """Nodes are selectable as targets for filters."""
        return True

    def get_menu_choices(self):
        """List of pairs of (name, function) for this element."""
        if self._node.get_data_filters().values():
            return ("Remove all filters from %s" % self._node.getName(), self._node.remove_filters),
        else:
            return []
            

class OutlierCanvasLineSegment(QCanvasLine):
    def __init__(self, (x1, y1), (x2, y2), canvas):
        QCanvasLine.__init__(self, canvas)
        self.setPoints(x1, y1, x2, y2)
        # keep lines behind nodes
        self.setZ(-2)

    def is_selectable(self):
        return False

    def get_menu_choices(self):
        """List of pairs of (name, function) for this element."""
        # canvas lines have no menu options available
        return []
