# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qwt import QwtSymbol, QwtPlotMarker
from qt import QPen, QBrush, Qt
from sets import Set
from GUI.Plots.plot2d.plotmap import MappableSelection

class NoSuchAspectError(Exception):
    """Trying to use an aspect which isn't available in a given plot type."""
    pass
    
class Plot2D(MappableSelection):
    """Super class for all types of plots."""

    _type = "unknown_type"
    _name = "Unknown Plot"

    def __init__(self, parent_plot, config, aspects=[]):
        MappableSelection.__init__(self, aspects)
        self._plot_map_list = []
        self._parent_plot = parent_plot
        self._plot_config = config

        parent_plot.add_plot(self)
        # default label size (We really need better handling of settings...)
        self._label_size = 12

    def remove_from_widget(self):
        self._parent_plot.remove_plot(self)

    def set_label_size(self, value):
        self._label_size = value

    def get_label_size(self):
        return self._label_size

    def get_parent_plot(self):
        return self._parent_plot

    def draw(self, painter, rect, x_map, y_map, plot_widget):
        """Optionally draw into plot widget when it redraws.

        painter - QwtPainter
        rect - QRect - bounding rectangle
        x_map, y_map - QwtDiMap
        plot_widget - Plot2DWidget
        """
        pass

    def draw_selection(self, painter, maps, rect):
        """Allow plot to draw selection on top of other markers/curves."""
        pass
        
    def select_area(self, rect):
        pass
    
    def select_point(self, point):
        """Select an object at point.

        Returns whether or not an object was found.
        """
        pass

    def deselect(self, deselect_mapped_points=True):
        if deselect_mapped_points:
            self.deselect_with_mapped()
        self.deselect_self()

    def set_settings(self, settings):
        self._plot_config.set_settings(settings)

    def get_settings(self):
        return self._plot_config.get_settings()

    def get_setting(self, key):
        """Get single plot-setting by key."""
        return self._plot_config.get_setting(key)

    def set_setting(self, key, value):
        """Set single plot-setting by key.

        Any setting set will be stored together with module diagram in
        xml next time diagram is saved.
        """
        self._plot_config.set_setting(key, value)

    def get_marker_setting(self, key):
        """Get single marker setting by key."""
        return self._plot_config.get_marker_setting(key)

    def set_marker_setting(self, key, value):
        """Set single marker setting by key.

        Any setting set will be stored together with module diagram in
        xml next time diagram is saved.
        """
        self._plot_config.set_marker_setting(key, value)
        
    def get_marker_settings(self):
        return self._plot_config.get_marker_settings()

    def set_marker_settings(self, settings):
        self._plot_config.set_marker_settings(settings)
        
    def name(self):
        try:
            return self._plot_config.get_settings()['name']
        except Exception:
            return "Unknown"

    def index(self):
        return self._plot_config.get_index()

    # class methods
    def get_type(cls):
        """Get lowercase underscore-separated name used for xml-tags."""
        return cls._type

    def get_name(cls):
        """Get full name of plot used when presenting plot types to user."""
        return cls._name

    get_type = classmethod(get_type)
    get_name = classmethod(get_name)

    def get_tooltip_at(self, x, y):
        """Override to define tooltips for pixel positions.

        Method should return (tooltip_string, pixel_distance).  When
        tooltip_string is None, pixel_distance is undefined.
        """
        return None, 0

    def get_data_length(self):
        """Get the length of the data set in the plot.

        Needs to be reimplemented in subclasses.
        """
        raise NotImplementedError, "%s does not implement 'get_data_length()'" % self.get_name()


class PlotMarker(QwtPlotMarker):
    
    def __init__(self, parent_plot_widget, data_index, style, colour, size, thickness):
        QwtPlotMarker.__init__(self, parent_plot_widget)
        self._marker_key = None
        self._data_index = data_index
        self._colour = colour
        
        self._current_pen = QPen(colour, thickness)
        self._current_brush = QBrush(colour)
        self._colour_highlight = Qt.red
        self._pen_highlighted = QPen(self._colour_highlight)
        self._brush_highlighted = QBrush(self._colour_highlight)
        self._current_style = style
        self._current_size = size

        symbol = QwtSymbol(self._current_style, self._current_brush, self._current_pen, self._current_size)
        self.setSymbol(symbol)

        self._highlighted = False
        
        
    def set_key(self, key):
        self._marker_key = key
        
     
    def key(self):
        return self._marker_key 


    def set_data_index(self, index):
        """To be used from dendrogram plot, to change the index"""
        self._data_index = index
        
        
    def data_index(self):
        return self._data_index


    def highlight(self, highlight_):
        if highlight_:
            self._highlighted = True
            width = self._current_pen.width()
            self._pen_highlighted.setWidth(width)
            self.symbol().setPen(self._pen_highlighted)
            self.symbol().setBrush(self._brush_highlighted)
            self.setLabelColor(self._colour_highlight)
        else:
            self._highlighted = False
            try:                
                self.setLabelColor(self._colour)
                self.symbol().setPen(self._current_pen)
                self.symbol().setBrush(self._current_brush)
            except RuntimeError, e:
                # Will create an exception if the user has pressed
                # the Update button while something is highlighted
                pass



    def set_style(self, style):
        self._current_style = style
        self.symbol().setStyle(style)


    def set_size(self, size):
        self._current_size = size
        self.symbol().setSize(size)


    def set_thickness(self, thickness):
        self._current_pen.setWidth(thickness)
        self.symbol().pen().setWidth(thickness)


    def set_colour(self, colour):
        pen = self._current_pen
        brush = self._current_brush
        pen.setColor(colour)
        brush.setColor(colour)
        if not self._highlighted:
            self.symbol().setPen(pen)
            self.symbol().setBrush(brush)


    def set_label_colour(self, colour):
        self._colour = colour
        if not self._highlighted:
            self.setLabelColor(colour)
        else:
            self.setLabelColor(self._colour_highlight)
