# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qwt import QwtPlot, QwtPlotPicker, QwtPlotZoomer, QwtPicker
from qwt import QwtEventPattern
from qt import Qt, QPen, QColor, SIGNAL, QPixmap, QCursor, QSizePolicy
from qt import QToolTip, QRect, QMessageBox, QPopupMenu, QPixmap, QAction
from sets import Set

from Library.Util.Config import Config
from GUI.Pixmaps.cursors import Cursors
from GUI.Nodes.guiplot2dnodesettings import _item_to_style
from GUI.Plots.plot2d.outlierselection import OutlierWidget


class Plot2DWidgetTipper(QToolTip):
    """Handle tooltips for a Plot2DWidget."""
    def __init__(self, plot_widget):
        QToolTip.__init__(self, plot_widget.canvas())
        self._plot_widget = plot_widget
        self._limit = 5

    def set_limit(self, limit):
        self._limit = limit
        
    def maybeTip(self, point):
        far = int(1e10)
        closest_distance = far
        closest_tip = None
        
        for plot in self._plot_widget.get_plots():
            tooltip, distance = plot.get_tooltip_at(point.x(), point.y())
            if tooltip != None:
                if distance > self._limit:
                    continue
                if distance < closest_distance:
                    closest_tip = tooltip
                    closest_distance = distance

        if closest_tip:
            self.tip(QRect(point.x()-5, point.y()-5, 10, 10),
                     closest_tip, closest_tip)
        

class Plot2DWidget(QwtPlot):
    """This widget represents one or more plots on top of eachother.

    GUIPlot2DWindow uses a grid of several Plot2DWidgets to show
    several plots side by side.
    """
    def __init__(self, parent, main_window, config, window_id, name=None, title=""):
        QwtPlot.__init__(self, parent, name)
        self.setCanvasLineWidth(0)
        
        self.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored, True)
        self._selected = False
        self.setTitle(title)
        self.setAutoLegend(False)
        self.setCanvasBackground(Qt.white)

        self._gui_mainwindow = main_window
        self._id = window_id # the (x, y) id tuple from the gui window

        self._axis_ratio = 1.0
        self._keep_ratio = False
        self._x_min_set  = 0
        self._x_max_set  = 1000
        self._y_min_set  = 0
        self._y_max_set  = 1000
        self._zoomed    = False
        self._plots = {}
        self._config = config

        self._outlier_menu = QPopupMenu(self)
        self._outlier_widget = OutlierWidget(self, self._gui_mainwindow.node())
        self._show_outlier_widget_action = None
        self._remove_from_last_action = None
        self._delete_all_filters_action = None
        self._run_diagram_action = None
        
        self._init_plot_pickers()
        self._init_zooming()

        self._marker_limit = 30000

        self.connect(self, SIGNAL("plotMousePressed(const QMouseEvent &)"),
                     self._slot_mouse_pressed)

        if config:
            settings = config.get_settings()
        else:
            # to make tests setting config = None work
            settings = {}
        self._load_settings(settings)
        
        self._set_cross_cursor()
        self.canvas().setCacheMode(False)
        self.replot()

        self._tipper = Plot2DWidgetTipper(self)

        self.setMinimumWidth(1)
        self.setMinimumHeight(1)

    def _load_settings(self, settings):
        """Sets settings on self from config."""
        self._load_general_settings(settings)
        self._load_labels_settings(settings)
        self._load_grid_settings(settings)
        self._load_axes_settings(settings)

    def _load_general_settings(self, settings):
        """Sets general settings from dictionary, or defaults."""
        if settings.has_key("background_colour"):
            self.setCanvasBackground(QColor(settings["background_colour"]))

    def _load_labels_settings(self, settings):
        """Sets labels settings from dictionary, or defaults."""
        if settings.has_key("window_title"):
            self.setTitle(settings["window_title"])
            try:
                self._gui_mainwindow.update_title_for_plotwidget(self._id)
            except KeyError:
                # not registered in self._gui_mainwindow
                pass
        if settings.has_key("xaxis_title"):
            self.set_xaxis_title(settings["xaxis_title"])
        if settings.has_key("yaxis_title"):
            self.set_yaxis_title(settings["yaxis_title"])

    def _load_grid_settings(self, settings):
        """Sets grid settings from dictionary, or defaults."""
        self.enableGridX(False) # default
        if settings.has_key("x_grid_maj_enable"):
            self.enableGridX(bool(int(settings["x_grid_maj_enable"])))
            if settings.has_key("x_grid_min_enable"):
                self.enableGridXMin(bool(int(settings["x_grid_min_enable"])))

        self.enableGridY(False) # default
        if settings.has_key("y_grid_maj_enable"):
            self.enableGridY(bool(int(settings["y_grid_maj_enable"])))
            if settings.has_key("y_grid_min_enable"):
                self.enableGridYMin(bool(int(settings["y_grid_min_enable"])))

        majpen = self.gridMajPen()
        minpen = self.gridMinPen()

        # default settings for QPen are ok, except from color of minor grid pen
        minpen.setColor(QColor("#888888"))

        if settings.has_key("maj_grid_colour"):
            majpen.setColor(QColor(settings["maj_grid_colour"]))
        if settings.has_key("min_grid_colour"):
            minpen.setColor(QColor(settings["min_grid_colour"]))
        if settings.has_key("maj_grid_width"):
            majpen.setWidth(int(settings["maj_grid_width"]))
        if settings.has_key("min_grid_width"):
            minpen.setWidth(int(settings["min_grid_width"]))
            
        # Style is saved as index in combobox for
        # WindowSettingsDialog.. hack to load settings...
        if settings.has_key("maj_grid_style"):
            majstyle = _item_to_style(int(settings["maj_grid_style"]))
            majpen.setStyle(majstyle)
        if settings.has_key("min_grid_style"):
            minstyle = _item_to_style(int(settings["min_grid_style"]))
            minpen.setStyle(minstyle)

    def _load_axes_settings(self, settings):
        """Sets axes settings from dictionary, or defaults."""
        set_x_axis = self.set_x_axis # default is to autoscale axis
        
        if settings.has_key("x_axis_auto_scale"):
            if bool(int(settings["x_axis_auto_scale"])):
                self.set_x_axis_autoscale()
            else:
                # set x axis values below (calling set_x_axis_scale
                # will disable autoscaling)
                set_x_axis = self.set_x_axis_scale

        x_min, x_max = 0, 1000
        if settings.has_key("x_axis_min"):
            x_min = float(settings["x_axis_min"])
        if settings.has_key("x_axis_max"):
            x_max = float(settings["x_axis_max"])

        set_x_axis(x_min, x_max)

        # default is to autoscale axis
        set_y_axis = self.set_y_axis
        
        if settings.has_key("y_axis_auto_scale"):
            if bool(int(settings["y_axis_auto_scale"])):
                self.set_y_axis_autoscale()
            else:
                # set y axis values below (calling set_y_axis_scale
                # will disable autoscaling)
                set_y_axis = self.set_y_axis_scale
                
        y_min, y_max = 0, 1000
        if settings.has_key("y_axis_min"):
            y_min = float(settings["y_axis_min"])
        if settings.has_key("y_axis_max"):
            y_max = float(settings["y_axis_max"])

        set_y_axis(y_min, y_max)

        # ratio scaling of axes
        if settings.has_key("ratio"):
            self._keep_ratio = bool(int(settings["ratio"]))
        if settings.has_key("aspect_ratio"):
            self.set_aspect_ratio(float(settings["aspect_ratio"]))

    def reset_axis_settings(self):
        settings = self.get_settings()
        for key in ["x_axis_min", "x_axis_max", "y_axis_min", "y_axis_max",
                    "x_axis_auto_scale", "y_axis_auto_scale", "aspect_ratio",
                    "ratio"]:
            if settings.has_key(key):
                del settings[key]

        # sets default settings for settings deleted
        self._load_axes_settings(settings)

    def get_settings(self):
        """Get all settings associated with this widget."""
        return self._config.get_settings()

    def set_settings(self, settings):
        """Set all settings for this plot widget."""
        # set settings in widget
        self._load_settings(settings)
        # propagate settings to config in node
        self._config.set_settings(settings)

    def _init_zooming(self):
        rubberband_pen = QPen(QColor(Config.zoomRectR, Config.zoomRectG,
                                     Config.zoomRectB), 1)
        self.replot()
        self._zoomer = QwtPlotZoomer(self.canvas())
        self._zoomer.setRubberBandPen(rubberband_pen)
        self._zoomer.setCursorLabelMode(QwtPicker.AlwaysOff)
        self._zoomer.setSelectionFlags(QwtPicker.RectSelection | QwtPicker.DragSelection)
        self._zoomer.setRubberBand(QwtPicker.RectRubberBand)
        # set patterns for zoom, zoom-base and pop-zoom-stack
        mouse_pattern = [QwtEventPattern.MousePattern(Qt.RightButton, Qt.ShiftButton),
                         QwtEventPattern.MousePattern(Qt.MidButton, Qt.NoButton),
                         QwtEventPattern.MousePattern(Qt.LeftButton, Qt.ShiftButton)]

        self._zoomer.setMousePattern(mouse_pattern)
    
    
    def _init_plot_pickers(self):
        rubberband_pen = QPen(QColor(Config.plotPickRectR,
                                     Config.plotPickRectG,
                                     Config.plotPickRectB), 1)
        
        self._picker = QwtPlotPicker(self.canvas())
        self._picker.setRubberBandPen(rubberband_pen)
        self._picker.setCursorLabelMode(QwtPicker.AlwaysOff)
        self._picker.setSelectionFlags(QwtPicker.RectSelection | QwtPicker.DragSelection)
        self._picker.setRubberBand(QwtPicker.RectRubberBand)
        mouse_pattern = [QwtEventPattern.MousePattern(Qt.LeftButton,
                                                      Qt.NoButton)]

        self._picker.setMousePattern(mouse_pattern)
        
        self.connect(self._picker, SIGNAL("selected(const QwtDoubleRect &)"),
                     self._slot_selected)

        self._picker_append = QwtPlotPicker(self.canvas())
        self._picker_append.setRubberBandPen(rubberband_pen)
        self._picker_append.setCursorLabelMode(QwtPicker.AlwaysOff)
        self._picker_append.setSelectionFlags(QwtPicker.RectSelection | QwtPicker.DragSelection)
        self._picker_append.setRubberBand(QwtPicker.RectRubberBand)
        mouse_pattern = [QwtEventPattern.MousePattern(Qt.LeftButton,
                                                      Qt.ShiftButton)]
        self._picker_append.setMousePattern(mouse_pattern)
        
        self.connect(self._picker_append,
                     SIGNAL("selected(const QwtDoubleRect &)"),
                     self._slot_append_selection)

    def add_plot(self, plot):
        """Called from a plot to add itself to this Plot2DWidget."""
        self._plots[plot] = plot

    def limited_insert_markers(self, plot_markers):
        """Insert PlotMarker entries, limiting to current marker limit.

        All markers will have keys set. The limit is to ensure that
        the number of markers does not exceed limitations in QwtPlot.
        The limit is settable with
        Plot2DWidget.set_marker_limit(limit).

        Non-added markers will have key set to 0.
        """

        safe_to_add = self.get_marker_limit() - len(self.markerKeys())
        if safe_to_add < 0:
            safe_to_add = 0

        for marker in plot_markers[0:safe_to_add]:
            marker.set_key(self.insertMarker(marker))
        for marker in plot_markers[safe_to_add:]:
            marker.set_key(0)
        if safe_to_add < len(plot_markers):
            self.marker_warning()

    def get_marker_limit(self):
        return self._marker_limit

    def set_marker_limit(self, limit):
        """Set maximum number of markers addable to plot before warning user.

        Mostly usable for tests of marker limit, where having to
        create and add 30000 markers takes too much time.
        """
        self._marker_limit = limit

    def marker_warning(self):
        """Warn user that number of markers have reached limit."""
        QMessageBox.warning(self, "Marker limit reached.",
"""Plot has tried to add more markers than possible.
There is a limit of 30000 markers in each plot area.

Markers are used in scatterplot for labels, when using
size/colour-mapping and when using text as markers.

Only the first 30000 markers will be shown in this plot area.""")

    def remove(self):
        # Removes the whole widget.
        for plot in self.get_plots():
            self.parent().remove_plot_map(plot)

    def remove_plot(self, plot):
        self.enableYLeftAxis(True)
        self.enableXBottomAxis(True)
        for key, plot_value in self._plots.iteritems():
            if plot == plot_value:
                del self._plots[key]
                self.parent().remove_plot_map(plot)
                break

        self.replot()
        if self.x_axis_autoscales() or self.y_axis_autoscales():
            self._zoomer.setZoomBase()
    
    def set_x_axis_scale(self, x_min, x_max):
        """Sets the min, max values on the x-axis.

        Disables autoscaling.
        """
        self._x_min_set = x_min
        self._x_max_set = x_max
        self.setAxisScale(QwtPlot.xBottom, x_min, x_max)
        self.replot()
        self._zoomer.setZoomBase()

    def get_x_axis_min(self):
        return self.axisScale(QwtPlot.xBottom).lBound()

    def get_x_axis_max(self):
        return self.axisScale(QwtPlot.xBottom).hBound()

    def set_x_axis(self, x_min, x_max):
        """Sets the min, max values on the x-axis.

        Does not affect the autoscaling (if it is enabled).
        """
        self.axisScale(QwtPlot.xBottom).rebuild(
            x_min, x_max,
            self.axisMaxMajor(QwtPlot.xBottom),
            self.axisMaxMinor(QwtPlot.xBottom),
            False) # False -> not logartithmic
        self.updateAxes()

    def set_y_axis_scale(self, y_min, y_max):
        """Sets the min, max values on the y-axis.

        Disables autoscaling.
        """
        self._y_min_set = y_min
        self._y_max_set = y_max
        self.setAxisScale(QwtPlot.yLeft, y_min, y_max)
        self.replot()
        self._zoomer.setZoomBase()

    def get_y_axis_min(self):
        return self.axisScale(QwtPlot.yLeft).lBound()

    def get_y_axis_max(self):
        return self.axisScale(QwtPlot.yLeft).hBound()

    def set_y_axis(self, y_min, y_max):
        """Sets the min, max values on the y-axis.

        Does not affect the autoscaling (if it is enabled).
        """
        self.axisScale(QwtPlot.yLeft).rebuild(
            y_min, y_max,
            self.axisMaxMajor(QwtPlot.yLeft),
            self.axisMaxMinor(QwtPlot.yLeft),
            False) # False -> not logartithmic
        self.updateAxes()

    def set_x_axis_autoscale(self):
        self.setAxisAutoScale(QwtPlot.xBottom)
        self.replot()
        self._zoomer.setZoomBase()
        self._x_min_set = self.get_x_axis_min()
        self._x_max_set = self.get_x_axis_max()

    def set_y_axis_autoscale(self):
        self.setAxisAutoScale(QwtPlot.yLeft)
        self.replot()
        self._zoomer.setZoomBase()
        self._y_min_set = self.get_y_axis_min()
        self._y_max_set = self.get_y_axis_max()
    
    def x_axis_autoscales(self):
        return self.axisAutoScale(QwtPlot.xBottom)

    def y_axis_autoscales(self):
        return self.axisAutoScale(QwtPlot.yLeft)

    def set_major_grid_pen_options(self, qcolor, width, style):
        self._set_grid_pen_options(True, qcolor, width, style)

    def set_minor_grid_pen_options(self, qcolor, width, style):
        self._set_grid_pen_options(False, qcolor, width, style)
    
    def _set_grid_pen_options(self, major, qcolor, width, style):
        if major:
            pen = self.gridMajPen()
        else:
            pen = self.gridMinPen()
        pen.setColor(qcolor)
        pen.setWidth(width)
        pen.setStyle(style)

    def set_data_by_plot(self, plot, xs, ys, indices=None):
        plot.set_data(xs, ys, indices)
        self._gui_mainwindow.auto_plot_map(plot)
        self._zoomer.setZoomBase()

    def get_zoom_base(self):
        return self._zoomer.zoomBase()
        
    def get_plots(self):
        return self._plots.values()
            
    def id(self):
        return self._id
        
    def parent(self):
        return self._gui_mainwindow
        
    def get_plot_by_index(self, index):
        for plot in self._plots.values():
            if plot.index() == index:
                return plot

    # Plot widget settings.

    def get_general_settings(self):
        """Returns the settings for the general tab in WidgetSettingsDialog."""
        return (self.canvasBackground(),)

    def get_labels_settings(self):
        """Returns the settings for the labels tab in WidgetSettingsDialog."""
        return (self.title(), self.axisTitle(QwtPlot.xBottom),
                self.axisTitle(QwtPlot.yLeft))

    def get_grid_settings(self):
        """Returns the settings for the grid tab in WidgetSettingsDialog."""
        majpen = self.gridMajPen()
        minpen = self.gridMinPen()
        return (self.gridXEnabled(), self.gridYEnabled(), majpen.color(),
                majpen.width(), majpen.style(),
                self.gridXMinEnabled(), self.gridYMinEnabled(), minpen.color(),
                minpen.width(), minpen.style())

    def get_axes_settings(self):
        """Returns the settings for the axes tab in WidgetSettingsDialog."""
        return (self.x_axis_autoscales(), self.get_x_axis_min(),
                self.get_x_axis_max(), self.y_axis_autoscales(),
                self.get_y_axis_min(), self.get_y_axis_max(), self._keep_ratio,
                self.aspect_ratio())

    def set_widget_title(self, title):
        self.setTitle(title)
        self.updateLayout()

    def set_background_colour(self, colour):
        self.setCanvasBackground(colour)

    def set_xaxis_title(self, title):
        self.setAxisTitle(QwtPlot.xBottom, title)

    def set_yaxis_title(self, title):
        self.setAxisTitle(QwtPlot.yLeft, title)
    
    def set_axis_scale(self, axis, minvalue, maxvalue, step=0):
        self.setAxisScale(axis, min, max, step)

    # Slots.
    
    def _slot_selected(self, rect):
        # Call select_area(rect) function in all plot instances.
        
        # Select mapped points
        for plot in self.get_plots():
            plot.deselect()

        self._slot_append_selection(rect)

    def _slot_append_selection(self, rect):
        for plot in self.get_plots():
            if rect.width() > 0 and rect.height() > 0:
                plot.select_area(rect)
            elif plot.select_point(rect.center()):
                break

        self._gui_mainwindow.replot_all_widgets()


    def _set_cross_cursor(self):
        cursor_pixmap = QPixmap(Cursors.cross)
        cursor = QCursor(cursor_pixmap) 
        self.canvas().setCursor(cursor)

    def set_keep_aspect_ratio(self, keep_aspect_ratio):
        self._keep_ratio = keep_aspect_ratio

    def set_aspect_ratio(self, axis_ratio):
        self._axis_ratio = axis_ratio

    def aspect_ratio(self):
        return self._axis_ratio

    def _scale_axis(self, axis1, min1, max1, min2, max2, ratio):
        delta_1 = max1 - min1
        delta_2 = max2 - min2
        new_delta_1 = delta_2 * ratio
        add_to_1 = (new_delta_1 - delta_1) / 2.0
        self.setAxisScale(axis1, min1 - add_to_1, max1 + add_to_1)

    def _ratio_scale_axes(self, canvas_width, canvas_height):
        widget_settings  = self.get_settings()
        if self._zoomer.zoomBase() != self._zoomer.zoomRect():
            self._zoomed = True
            return
        elif self._zoomed: # previous state....
            self._zoomed = False
            if widget_settings.has_key('x_axis_auto_scale'):
                if int(widget_settings['x_axis_auto_scale']):
                    self.set_x_axis_autoscale()
            if widget_settings.has_key('y_axis_auto_scale'):
                if int(widget_settings['y_axis_auto_scale']):
                    self.set_y_axis_autoscale()
            return
        if self._keep_ratio:
            x_auto = int(widget_settings['x_axis_auto_scale'])
            y_auto = int(widget_settings['y_axis_auto_scale'])
            x_min  = float(widget_settings['x_axis_min'])
            x_max  = float(widget_settings['x_axis_max'])
            y_min  = float(widget_settings['y_axis_min'])
            y_max  = float(widget_settings['y_axis_max'])
            delta_x = x_max - x_min
            delta_y = y_max - y_min
            if x_auto and y_auto:
                actual_delta_x = self.axisScale(QwtPlot.xBottom).hBound() - self.axisScale(QwtPlot.xBottom).lBound()
                actual_delta_y = self.axisScale(QwtPlot.yLeft).hBound()   - self.axisScale(QwtPlot.yLeft).lBound()
                if (delta_y - actual_delta_y) * self._axis_ratio < (delta_x - actual_delta_x) / self._axis_ratio:
                    # add "air" in y-direction 
                    self.setAxisAutoScale(QwtPlot.xBottom)
                    self._scale_axis(QwtPlot.yLeft, y_min, y_max, x_min, x_max,
                                      float(self._axis_ratio) * (float(canvas_height) / float(canvas_width)))
                else:
                    # add "air" in x-direction
                    self.setAxisAutoScale(QwtPlot.yLeft)
                    self._scale_axis(QwtPlot.xBottom, x_min, x_max, y_min, y_max,
                                      (1.0 /float(self._axis_ratio)) * (float(canvas_width) / float(canvas_height)))
            elif x_auto:
                self._scale_axis(QwtPlot.xBottom, x_min, x_max, y_min, y_max,
                                  (1.0 /float(self._axis_ratio)) * (float(canvas_width) / float(canvas_height)))
            elif y_auto:
                self._scale_axis(QwtPlot.yLeft, y_min, y_max, x_min, x_max,
                                  float(self._axis_ratio) * (float(canvas_height) / float(canvas_width)))
            else:
                return
        self.updateAxes()
        self._zoomer.setZoomBase()


    def drawCanvas(self, painter):
        if self._plots:
            # do not ratio scale if canvas is empty
            if self.markerKeys() or self.curveKeys():
                self._ratio_scale_axes(self.canvas().contentsRect().width(),
                                       self.canvas().contentsRect().height())
        QwtPlot.drawCanvas(self, painter)

    def drawCanvasItems(self, painter, rect, maps, pfilter):
        x_bottom = maps[QwtPlot.xBottom]
        y_left = maps[QwtPlot.yLeft]
        for plot in self._plots.values():
            plot.draw(painter, rect, x_bottom, y_left, self)

        QwtPlot.drawCanvasItems(self, painter, rect, maps, pfilter)

        for plot in self._plots.values():
            plot.draw_selection(painter, maps, rect)

    def _slot_mouse_pressed(self, event):
        self._gui_mainwindow.set_selected(self)

        # Disabled for now..
        if event.button() == Qt.RightButton and event.state() != Qt.ShiftButton:
            self._show_outlier_menu()
            # Moved this here since popup() caused a segfault
            # when running the unittests.
            self._outlier_menu.popup(event.globalPos())

            
    def _show_outlier_menu(self):
        """Method responsible for adding and enable/disable actions to the plot contextmenu."""
        self._outlier_menu.clear()
        
        plots = self._plots.values()
        filter_node = self._outlier_widget.get_last_filternode()
        connected_nodes =  self._gui_mainwindow.node().get_all_providing_nodes()
        connected_nodes.append(self._gui_mainwindow.node())

        something_selected = False
        filter_added = False

        # Find out if any of the markers has been selected
        while plots and not something_selected:
            plot = plots.pop()
            for aspect in plot.get_mappable_aspects():
                if plot.get_highlighted_indices_by_aspect(aspect):
                    something_selected = True
                    break

        # Find out if any of the connected nodes has a filter. 
        for node in connected_nodes:
            if node.get_data_filters():
                filter_added = True
                break

        # Action for openeing the OutlierWidget
        self._show_outlier_widget_action = QAction('Add outlier filter...',
                                                   0, self)
        self._show_outlier_widget_action.addTo(self._outlier_menu)
        self.connect(self._show_outlier_widget_action, SIGNAL('activated()'),
                     self._slot_show_outlier_widget)
        self._show_outlier_widget_action.setEnabled(something_selected)

        # Add action for adding the filter to last "filtered" node
        if filter_node:
            self._remove_from_last_action = QAction('Add outlier filter to %s' \
                                                    % filter_node.getName(),
                                                    0, self)
            self.connect(self._remove_from_last_action,
                         SIGNAL('activated()'),
                         self._slot_remove_from_last_filternode)
            self._remove_from_last_action.addTo(self._outlier_menu)
            self._remove_from_last_action.setEnabled(something_selected)

        else:
            self._remove_from_last_action = None

        # Action for deleting all filters   
        self._delete_all_filters_action = QAction('Remove all filters',
                                                  0, self)
        self.connect(self._delete_all_filters_action, SIGNAL('activated()'),
                     self._slot_delete_all_filters)
        self._delete_all_filters_action.addTo(self._outlier_menu)
        self._delete_all_filters_action.setEnabled(filter_added)

        # Action for running the diagram        
        self._run_diagram_action = QAction('Re-run diagram', 0, self)
        self.connect(self._run_diagram_action, SIGNAL('activated()'),
                     self._slot_run_diagram)
        self._run_diagram_action.addTo(self._outlier_menu)
        self._run_diagram_action.setEnabled(bool(filter_node))

    def _slot_run_diagram(self):
        md = self._gui_mainwindow.node().get_module_diagram()
        md.start()
        
    def _slot_show_outlier_widget(self):
        self._outlier_widget.show(self._gui_mainwindow.node().get_module_diagram().get_nodes())

    def _slot_remove_from_last_filternode(self):
        node = self._outlier_widget.get_last_filternode()

        for plot in self._plots.values():
            n = plot.get_data_length()
            for aspect in plot.get_mappable_aspects():
                indices = plot.get_highlighted_indices_by_aspect(aspect)
                node.add_data_filter(n, indices)

    def _slot_delete_all_filters(self):
        nodes = self._gui_mainwindow.node().get_all_providing_nodes()
        nodes.append(self._gui_mainwindow.node())

        for node in nodes:
            node.remove_filters()
        
    def is_selected(self):
        return self._selected

    def set_selected(self, selected):
        if selected:
            self.canvas().setFocus()
        do_replot = False
        if self._selected != selected:
            do_replot = True
            
        self._selected = selected
        if do_replot:
            self.replot() # Draw or remove focus indicator/rectangle.
