# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.
from sets import Set

# domain exceptions
class NoSuchAspectError(Exception):
    """Trying to use an aspect which isn't available in a given mappable selection."""
    pass

    
class PlotMapConfig:
    """This class holds the settings for a single plot map. It corresponds
    to the PlotAspectMap class."""
    def __init__(self, index):
        self._settings = {}
        self._index = index

    def set_settings(self, settings):
        self._settings = settings.copy()

    def get_settings(self):
        return self._settings
        
    def get_index(self):
        return self._index
        
    def to_xml(self):
        if self._settings:
            xml_list = []
            xml_list.append("\t\t\t<plot_map>")

            settings_keys = self._settings.keys()
            settings_keys.sort()

            for setting in settings_keys:
                xml_list.append("\t\t\t\t<" + str(setting) + ">" +
                                str(self._settings[setting]) +
                                "</" + str(setting) + ">")

            xml_list.append("\t\t\t</plot_map>")
            return xml_list
        else:
            return []



class PlotMapHandler:
    """Class handling mappings between plots."""
    def __init__(self, plotnode):
        self._plot_map_list = []
        self._plot_maps = {}
        self._go_browser_maps = []
        self._plotnode = plotnode

        self._enabled = False
        self._auto_map_enabled = False

    def from_xml_element(self, root):
        if root.find("auto_map_enabled") is not None:
            self.set_auto_map_enabled(bool(int(root.find("auto_map_enabled").text)))
        if root.find("enabled") is not None:
            self.set_enabled(bool(int(root.find("enabled").text)))

    def set_enabled(self, value):
        """Turn on/off plot mapping."""
        self._enabled = value

    def get_enabled(self):
        return self._enabled

    def get_auto_map_enabled(self):
        return self._auto_map_enabled

    def set_auto_map_enabled(self, value):
        """Should automatically map all related plots?"""
        self._auto_map_enabled = value
                
    def get_plot_maps(self):
        """Returns:
            - a list containing current plot mappings."""
        return self._plot_map_list

    def get_go_browser_maps(self):
        """Get all maps between GO browser and plot."""
        return self._go_browser_maps

    def add_plot_map(self, plot1, plot2):
        """Add all possible mappings between two plots."""
        for aspect1 in plot1.get_mappable_aspects():
            for aspect2 in plot2.get_mappable_aspects():
                self.add_aspect_map(plot1, plot2, aspect1, aspect2)

    def add_aspect_map(self, mappable1, mappable2, aspect1, aspect2):
        from GUI.Plots.plot2d.gobrowser import GoBrowser
        # remove duplicates and stop self-mapping
        key = mappable1, mappable2, aspect1, aspect2
        reverse_key = mappable2, mappable1, aspect2, aspect1
        if key in self._plot_maps: return
        if reverse_key in self._plot_maps: return 
        if mappable1 == mappable2: return

        # handle gobrowser-maps differently
        if isinstance(mappable1, GoBrowser) or isinstance(mappable2, GoBrowser):
            # make sure browser is always first in map (which is
            # reversed internally.. ack)
            if isinstance(mappable1, GoBrowser):
                browser_map = _PlotAspectMap(mappable2, mappable1, aspect2, aspect1)
            else:
                browser_map = _PlotAspectMap(mappable1, mappable2, aspect1, aspect2)
            self._go_browser_maps.append(browser_map)
        else:
            plot_map = _PlotAspectMap(mappable1, mappable2, aspect1, aspect2)
            key = mappable1, mappable2, aspect1, aspect2

            self._plot_maps[key] = plot_map
            self._plot_map_list.append(plot_map)

    def remove_plot_aspect_map(self, plot1, plot2, aspect1, aspect2):
        lookup_key = plot1, plot2, aspect1, aspect2
        reverse_key = plot2, plot1, aspect2, aspect1

        for key, plotmap in self._plot_maps.items():
            if key in (lookup_key, reverse_key):
                plot1.remove_plot_map(plotmap)
                plot2.remove_plot_map(plotmap)
                self._plot_map_list.remove(plotmap)
                del self._plot_maps[key]

        for browser_map in self._go_browser_maps[:]:
            browser, plot = browser_map.plots()
            browser_aspect = browser_map.get_aspect_for_plot(browser)
            plot_aspect = browser_map.get_aspect_for_plot(plot)

            if (browser, plot, browser_aspect, plot_aspect) in (lookup_key, reverse_key):
                self._go_browser_maps.remove(browser_map)
        
    def remove_plot_map(self, plot):
        """Remove all plot mapppings to plot."""
        for key, plot_map in self._plot_maps.items():
            if plot_map.contains(plot):
                plot_map.remove()
                self._plot_map_list.remove(plot_map)
                del self._plot_maps[key]

        for browser_map in self._go_browser_maps[:]:
            if plot in browser_map.plots():
                self._go_browser_maps.remove(browser_map)
                
        self.save_settings()
        
    def auto_map_all(self, plot_list):
        """Map all plots to each other."""
        for plot1 in plot_list:
            self.auto_map_plot(plot_list, plot1)
                            
    def auto_map_plot(self, plot_list, plot1):
        """Map plot to all other plots."""
        for plot2 in plot_list:
            self.add_plot_map(plot1, plot2)
        
    def mapping_exists(self, plot1, plot2):
        for plot_map in self._plot_map_list:
            if plot_map.contains(plot1) and plot_map.contains(plot2):
                return True
        return False
                
    def save_settings(self):
        """This method saves plot map handler settings. It also traverse through
        all plot maps and saves the settings of these as well."""
        plot_map_count = len(self._plot_map_list)
        self._plotnode.update_plot_maps(plot_map_count)

        for i, plot_map in enumerate(self._plot_map_list):
            settings = self._create_plot_map_settings(plot_map)
            self._plotnode.set_plot_map_settings(i, settings)

        browser_map_settings = map(self._create_browser_map_settings,
                                   self._go_browser_maps)
        self._plotnode.set_go_browser_map_settings(browser_map_settings)
            
        
    def load_settings(self, plot2dwindow):
        """This method creates a list of plot_maps based on a settings dictionary.
        
        Params:
            - plot2dwindow - needed because it has the list of all plot widgets.
        """
        for settings in self._plotnode.get_plot_map_settings():
            plot1_widget = plot2dwindow.get_plot_widget(
                (int(settings['plot1_widget_x']),
                 int(settings['plot1_widget_y'])))
            plot1 = plot1_widget.get_plot_by_index(int(settings['plot1_index']))
            plot2_widget = plot2dwindow.get_plot_widget(
                (int(settings['plot2_widget_x']),
                 int(settings['plot2_widget_y'])))
            plot2 = plot2_widget.get_plot_by_index(int(settings['plot2_index']))

            try:
                aspect1 = settings['plot1_aspect']
                aspect2 = settings['plot2_aspect']
            except KeyError:
                self.add_plot_map(plot1, plot2)
            else:
                self.add_aspect_map(plot1, plot2, aspect1, aspect2)

        for settings in self._plotnode.get_go_browser_map_settings():
            widget = plot2dwindow.get_plot_widget((int(settings['plot_widget_x']),
                                                   int(settings['plot_widget_y'])))
            plot = widget.get_plot_by_index(int(settings['plot_index']))
            
            aspect = settings['plot_aspect']
            browser = plot2dwindow.get_go_browser()
            browser_aspect, = browser.get_mappable_aspects()
            self.add_aspect_map(plot, browser, aspect, browser_aspect)

    def _create_plot_map_settings(self, plot_map):
        settings = {}

        plot1, plot2 = plot_map.plots()
        
        settings['plot1_index'] = str(plot1.index())
        settings['plot1_widget_x'] = str(plot1.get_parent_plot().id()[0])        
        settings['plot1_widget_y'] = str(plot1.get_parent_plot().id()[1])        
        settings['plot2_index'] = str(plot2.index())
        settings['plot2_widget_x'] = str(plot2.get_parent_plot().id()[0])        
        settings['plot2_widget_y'] = str(plot2.get_parent_plot().id()[1])
        settings['plot1_aspect'] = str(plot_map.get_aspect_for_plot(plot1))
        settings['plot2_aspect'] = str(plot_map.get_aspect_for_plot(plot2))

        return settings

    def _create_browser_map_settings(self, browser_map):
        settings = {}
        
        browser, plot = browser_map.plots()
        settings['plot_index'] = str(plot.index())
        settings['plot_widget_x'] = str(plot.get_parent_plot().id()[0])        
        settings['plot_widget_y'] = str(plot.get_parent_plot().id()[1])        
        settings['plot_aspect'] = str(browser_map.get_aspect_for_plot(plot))

        return settings
                

class _PlotAspectMap:
    """Class representing a mapping between two plots."""
    
    def __init__(self, plot2, plot1, aspect2=None, aspect1=None):
        if not aspect1:
            aspect1 = plot1.get_mappable_aspects()[0]
        if not aspect2:
            aspect2 = plot2.get_mappable_aspects()[0]
            
        self.plot1 = plot1
        self.plot2 = plot2
        
        self.plot1.add_plot_map(self)
        self.plot2.add_plot_map(self)
        
        self._aspects = { plot1 : aspect1, plot2 : aspect2 }
        
    def plots(self):
        return (self.plot1, self.plot2)

    def get_aspect_for_plot(self, plot):
        return self._aspects[plot]
        
    def contains(self, plot):
        return (self.plot1 == plot) or (self.plot2 == plot)
    
    def _connected_plot(self, plot):
        """Returns:
            - the plot connected to parameter plot."""
        if self.plot1 == plot:
            return self.plot2
        else:
            return self.plot1

    def select_points(self, source_plot, index_list):
        """Select points in connected plot based on selected points in the parameter source_plot."""
        other_plot = self._connected_plot(source_plot)
        other_plot.highlight_indices_by_aspect(index_list,
                                               self._aspects[other_plot])
    
    def deselect_points(self, source_plot):
        """Deselect points in plot connected to the parameter source_plot."""
        self._connected_plot(source_plot).deselect_self()

    def remove(self):
        """Remove self from plots, making this instance unusable.

        Should only be called from PlotMapHandler, as it needs to keep
        track of active mappings.
        """
        self.plot1.remove_plot_map(self)
        self.plot2.remove_plot_map(self)
        

class MappableSelection:
    """Mixin for classes to be mappable against each other.
    
    A MappableSelection is a stored set of selections for a set of
    contexts/aspects. For example a table might have its third row
    selected and at the same time it has its first ten columns
    selected. "row" and "column" would be aspects.

    An aspect of a MappableSelection can be mapped against other
    MappableSelections to automatically highlight/select elements in
    other linked MappableSelections.

    To subclass this mixin, specify mappable aspects when calling
    MappableSelection.__init__(self, aspects) in constructor or
    override get_mappable_aspects() to return list of aspects.

    Subclasses would usually also subclass highlight_indices_by_aspect
    if they need direct information about what indices have been
    selected. At other times, just call
    self.get_highlighted_indices_by_aspect(...).
    """
    def __init__(self, aspects=[]):
        self._highlighted = {}
        self._plot_maps = Set()
        self._aspects = aspects

    def get_mappable_aspects(self):
        """Show aspects of this class that can be mapped."""
        return self._aspects

    def select_values(self, values, aspect):
        """Select list of values in this and all mappables mapped to aspect."""
        self.highlight_indices_by_aspect(values, aspect)
        for plot_map in self._plot_maps:
            if plot_map.get_aspect_for_plot(self) == aspect:
                plot_map.select_points(self, values)

    def highlight_indices_by_aspect(self, indices, aspect):
        """Select values only in this mappable.

        Will only highlight indices for a certain context, as defined
        by aspect. An aspect might be 'rows', 'columns' or similar,
        depending on available aspects.
        """
        if aspect not in self.get_mappable_aspects():
            raise NoSuchAspectError, aspect

        if not self._highlighted.has_key(aspect):
            self._highlighted[aspect] = Set()

        self._highlighted[aspect].union_update(indices)

    def get_highlighted_indices_by_aspect(self, aspect):
        if aspect not in self.get_mappable_aspects():
            raise NoSuchAspectError, aspect

        if self._highlighted.has_key(aspect):
            return list(self._highlighted[aspect])
        else:
            return []

    def deselect_self(self):
        """Clear all selections in all aspects."""
        self._highlighted = {}

    def deselect_with_mapped(self):
        """Deselect self and connected mapping."""
        self.deselect_self()
        for plot_map in self._plot_maps:
            plot_map.deselect_points(self)

    def add_plot_map(self, plot_map):
        self._plot_maps.add(plot_map)

    def remove_plot_map(self, plot_map):
        self._plot_maps.remove(plot_map)
