#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Nodes.guiplot2dnodesettings import PlotSettings, Settings
from GUI.Nodes.guiplot2dnodesettings import DatabaseInfoDialog, LabelSettingsDialog
from GUI.Plots.plot2d.dataselector import DataSelector
from Library.Util.rangeparser import InvalidRangeError
from GUI.Plots.plot2d.plot2d import Plot2D, PlotMarker
from Library.Util.Config import Config
from Library.Util import lists, rangeparser, numerical, gui
from qwt import QwtPlotCurve, QwtSymbol, QwtPlot
from qt import QSize, Qt, QColor, QFont, SIGNAL, QPen, QBrush
from qt import QObject, QMessageBox, QPixmap, QPainter, QSize
from qt import QColorDialog, QFontDialog, PYSIGNAL
from qtui import QWidgetFactory

from itertools import izip
from sets import Set

import os
import sys
import math

connect = QObject.connect

class ScatterPlot(Plot2D):
    # constant for mappable aspects of this plot
    _map_aspect = "objects"

    _type = "scatter_plot"
    _name = "Scatter Plot"

    marker_symbols = [QwtSymbol.Ellipse, QwtSymbol.Cross, QwtSymbol.XCross,
                      QwtSymbol.Rect, QwtSymbol.Diamond, QwtSymbol.Triangle]
    
    def __init__(self, parent_plot, node, config):
        """Initializes a new curve containing a Scatter curve."""
        Plot2D.__init__(self, parent_plot, config, [self._map_aspect])

        self._node = node
        self.marker_list = []
        self._curve = QwtPlotCurve(parent_plot)
        self._curve.setStyle(QwtPlotCurve.NoCurve)
        self._curve_key = parent_plot.insertCurve(self._curve)

        self._labels = None
        self._extended_length = 0
        self._marker_map = {}

        # coordinates filtered by indices and nan-removal
        self._x_values = []
        self._y_values = []
        # all coordinates
        self._x_coords = []
        self._y_coords = []
        # indices to be plotted, filtered by nan-removal
        self._indices = []

        default_settings = {}
        default_settings['labels'] = ""
        default_settings['label_size'] = 12
        default_settings['label_index_prefix'] = ""
        default_settings['label_tooltips'] = "None"
        # override defaults with loaded values and store
        default_settings.update(config.get_settings())
        config.set_settings(default_settings)

        default_marker_settings = {}
        default_marker_settings['marker_symbol_index'] = ScatterPlot.marker_symbols.index(QwtSymbol.Ellipse)
        default_marker_settings['marker_size'] = 3
        default_marker_settings['marker_thickness'] = 1
        default_marker_settings['marker_colour'] = (0, 0, 0)
        default_marker_settings['marker_text'] = ""
        # 0 = symbol, 1 = text
        default_marker_settings['toggle_button_id'] = 0
        default_marker_settings['meta_data_input_port'] = ""
        default_marker_settings['min_colour'] = (0, 0, 0)
        default_marker_settings['max_colour'] = (255, 255, 255)
        default_marker_settings['meta_data_dir'] = 0
        default_marker_settings['col_row_index'] = 0
        default_marker_settings['map_to_colour'] = False

        default_marker_settings['map_to_size'] = False

        # override defaults with loaded values and store
        default_marker_settings.update(self.get_marker_settings())
        self.set_marker_settings(default_marker_settings)

    def create_settings(self, index, tab_widget, plot_widget, node):
        """Create GUI for modifying settings of this plot."""
        return ScatterPlotSettings(node, index, tab_widget, plot_widget, self)

    def update(self):
        """Update with correct markers, curve values and attributes.

        Update plot based on current settings and data.
        """
        self._update_markers()
        self._update_labels()

    def get_curve(self):
        return self._curve

    def deselect_with_mapped(self):
        self._deselect_markers()
        Plot2D.deselect_with_mapped(self)

    def deselect_self(self):
        self._deselect_markers()
        Plot2D.deselect_self(self)

    def _deselect_markers(self):
        for i in self.get_highlighted_indices_by_aspect(self._map_aspect):
            if self._marker_map.has_key(i):
                self._marker_map[i].highlight(False)

    def _remove_markers(self):
        """Remove current markers."""
        for m in self.marker_list:
            self._parent_plot.removeMarker(m.key())
        self.marker_list = []
        self._marker_map = {}

    def remove_from_widget(self):
        self._parent_plot.removeCurve(self._curve_key)
        self._remove_markers()

        # remove from widget after other cleanup to make sure plot is
        # replotted correctly
        Plot2D.remove_from_widget(self)

    def _map_data(self, min_size, max_size, data):
        """Scale all values in data between min_size and max_size.

        The smallest value in data will equal min_size, the largest
        will equal max_size, and values between will be scaled
        linearly.
        """
        result = []
        if data:
            size_range = max_size - min_size
            min_data = float(min(data))
            max_data = float(max(data))
            for element in data:
                if (max_data - min_data) != 0:
                    scale_factor = (element - min_data) / (max_data - min_data)
                else:
                    scale_factor = 0
                size = int(min_size + size_range * scale_factor)
                result.append(size)
        return result

    def _get_colour_mapping(self, from_rgb, to_rgb, data):
        return zip(self._map_data(from_rgb[0], to_rgb[0], data),
                   self._map_data(from_rgb[1], to_rgb[1], data),
                   self._map_data(from_rgb[2], to_rgb[2], data))

    def _add_markers(self, x_coords, y_coords, indices=None):
        """Add markers based on the current data set."""
        if indices == None:
            indices = range(len(x_coords))

        self.marker_list = []

        marker_size = QSize(self.get_marker_setting('marker_size'),
                            self.get_marker_setting('marker_size'))
        marker_colour = QColor(*self.get_marker_setting('marker_colour'))
        marker_style = ScatterPlot.marker_symbols[self.get_marker_setting('marker_symbol_index')]
        marker_thickness = self.get_marker_setting('marker_thickness')
        
        self._marker_map = {}
        for i in indices:
            # Get parameters set by MarkerSettingsDialog class at init.
            marker = PlotMarker(self._parent_plot, i, marker_style,
                                marker_colour, marker_size,
                                marker_thickness)
            marker.setXValue(x_coords[i])
            marker.setYValue(y_coords[i])
            marker.set_label_colour(Qt.black) 
            self.marker_list.append(marker)
            self._marker_map[i] = marker

        self._parent_plot.limited_insert_markers(self.marker_list)
        
    def set_data(self, x_coords, y_coords, indices=None):
        """Sets the data for the curve based on the x- and y-koordinates"""
        if not x_coords and not y_coords: return
        self._extended_length = min(len(y_coords), len(x_coords))
        if not indices:
            indices = range(min(len(x_coords), len(y_coords)))
        
        (nanx,nany) = numerical.find_all_nan_indices(x_coords,y_coords)

        if nanx:
            indices = filter(lambda i: not nanx[i], indices)
        if nany:
            indices = filter(lambda i: not nany[i], indices)

        # nothing to plot?
        if not indices:
            return

        x_values = map(lambda i : x_coords[i], indices)
        y_values = map(lambda i : y_coords[i], indices)

        # keep positions with indices
        self._indices = indices
        self._x_values  = x_values
        self._y_values = y_values
        self._x_coords  = x_coords
        self._y_coords = y_coords

        self.update()
        
    def highlight_indices_by_aspect(self, index_list, aspect):
        """Highlight markers given in the index_list, using current range."""
        Plot2D.highlight_indices_by_aspect(self, index_list, aspect)

        for i in index_list:
            if self._marker_map.has_key(i):
                self._marker_map[i].highlight(True)

        # if highlight_indices_by_aspect() was called from another window (e.g.,
        # GoBrowser), we need to redraw the plot for the selection to take
        # effect
        self._parent_plot.replot()
        
    def _get_points_within_area(self, x1, x2, y1, y2):
        index_list = []
        for i, (x, y) in enumerate(izip(self._x_values, self._y_values)):
            if x >= x1 and x <= x2 and y >= y1 and y <= y2:
                index_list.append(i)

        return map(self._indices.__getitem__, index_list)

    def select_area(self, rect):
        """Search for points inside given rectangle."""
        Plot2D.select_area(self, rect)

        index_list = self._get_points_within_area(rect.x1(), rect.x2(),
                                                  rect.y1(), rect.y2())
        self.select_values(index_list, self._map_aspect)

    def select_point(self, point):
        """Highlights a marker if one was hit.
        
        Returns whether a marker/point was hit.
        """
        closest = None
        closest_distance = 0

        # select within +/- 5 pixels
        delta = 5
        # convert delta to plot-distance instead of pixels
        x_map = self._parent_plot.canvasMap(QwtPlot.xBottom)
        y_map = self._parent_plot.canvasMap(QwtPlot.yLeft)
        x_delta = abs(x_map.invTransform(delta) - x_map.invTransform(0))
        y_delta = abs(y_map.invTransform(delta) - y_map.invTransform(0))

        # get only points within size of marker/symbol.  we multiply
        # with 1.2 to add an extra pixel in positive direction to
        # include pixel clicked.
        close_points = self._get_points_within_area(point.x()-x_delta,
                                                    point.x()+(x_delta*1.2),
                                                    point.y()-y_delta,
                                                    point.y()+(y_delta*1.2))

        # for all points within range, get the closest
        for data_index in close_points:
            index = self._indices.index(data_index)
            x = self._x_values[index]
            y = self._y_values[index]
            dx = x_map.transform(point.x()) - x_map.transform(x)
            dy = y_map.transform(point.y()) - y_map.transform(y)
            distance = dx*dx + dy*dy
            if closest == None or distance < closest_distance:
                closest = data_index
                closest_distance = distance

        if closest != None:
            self.select_values([closest], self._map_aspect)

        Plot2D.select_point(self, point)
        return closest != None

    def _get_indices(self, indices):
        valid = Set(indices).intersection(self._indices)

        if "__getitem__" in dir(self._x_coords) and "__getitem__" in dir(self._y_coords):
            # fast way
            x_values = map(self._x_coords.__getitem__, valid)
            y_values = map(self._y_coords.__getitem__, valid)
        else:
            # slow way
            x_values = map(lambda i : self._x_coords[i], valid)
            y_values = map(lambda i : self._y_coords[i], valid)
            
        return x_values, y_values

    def draw_selection(self, painter, maps, rect):
        x_bottom = maps[QwtPlot.xBottom]
        y_left = maps[QwtPlot.yLeft]

        curve = QwtPlotCurve(self._parent_plot)

        indices = self.get_highlighted_indices_by_aspect(self._map_aspect)

        try:
            curve.setData(*self._get_indices(indices))
        except TypeError, e:
            # bad data stuffed into curve, so we won't draw anything
            return
        
        pen = QPen(self.get_curve().pen())
        pen.setColor(QColor(255,0,0))
        brush = QBrush(QColor(255,0,0))
        symbol = QwtSymbol(self.get_curve().symbol())
        symbol.setPen(pen)
        symbol.setBrush(brush)
        curve.setStyle(QwtPlotCurve.NoCurve)
        curve.setSymbol(symbol)
        curve.draw(painter, x_bottom, y_left)

        # draw markers
        for i in indices:
            if self._marker_map.has_key(i):
                m = self._marker_map[i]
                m.draw(painter, x_bottom.transform(m.xValue()),
                       y_left.transform(m.yValue()),
                       rect)

    def _update_markers(self):
        """Update all markers and curves after changing marker settings."""
        if self._x_values and isinstance(self._x_values[0], str):
            return
        if self._y_values and isinstance(self._y_values[0], str):
            return

        self._remove_markers()
        self.get_curve().setData(self._x_values, self._y_values)

        has_text_marker = self.get_marker_setting('toggle_button_id')
        has_labels = self.get_setting('labels') in ('choice_indexed',
                                                    'choice_data')
        has_size_mapping = self.get_marker_setting('map_to_size')
        has_color_mapping = self.get_marker_setting('map_to_colour')

        if has_text_marker or has_labels or has_size_mapping or has_color_mapping:
            self._add_markers(self._x_coords, self._y_coords, self._indices)            
        
        marker_size = QSize(self.get_marker_setting('marker_size'),
                            self.get_marker_setting('marker_size'))
        marker_colour = QColor(*self.get_marker_setting('marker_colour'))
        marker_style = ScatterPlot.marker_symbols[self.get_marker_setting('marker_symbol_index')]
        marker_thickness = self.get_marker_setting('marker_thickness')
        
        self.get_curve().symbol().setStyle(marker_style)
        self.get_curve().symbol().setSize(marker_size)
        self.get_curve().symbol().pen().setWidth(marker_thickness)
        self.get_curve().symbol().pen().setColor(marker_colour)
        self.get_curve().symbol().brush().setColor(marker_colour)

        if has_text_marker:
            self.get_curve().symbol().setStyle(QwtSymbol.None)
            font = QFont("Times", 0, 0, 0)
            colour = QColor(*marker_colour.getRgb())
            for marker in self.marker_list:
                marker.set_style(QwtSymbol.None)
                marker.setLabel(self.get_marker_setting('marker_text'))
                marker.setFont(font)
                marker.set_label_colour(colour)
                marker.setLabelAlignment(Qt.AlignCenter)

        if self.get_marker_setting('map_to_colour'):
            for marker in self.marker_list:
                marker.set_style(marker_style)
            self.get_curve().symbol().setStyle(QwtSymbol.None)

            try:
                data = self._node.get_data(self.get_marker_setting('meta_data_input_port'),
                                           not self.get_marker_setting('meta_data_dir'),
                                           self.get_marker_setting('col_row_index'))
            except:
                data = []
            
            mapping = self._get_colour_mapping(self.get_marker_setting('min_colour'),
                                               self.get_marker_setting('max_colour'),
                                               data)
            for (r,g,b), marker in zip(mapping, self.marker_list):
                marker.set_colour(QColor(r, g, b))

        if self.get_marker_setting('map_to_size'):
            for marker in self.marker_list:
                marker.set_style(marker_style)
            self.get_curve().symbol().setStyle(QwtSymbol.None)

            try:
                data = self._node.get_data(self.get_marker_setting('meta_data_input_port_size'),
                                           not self.get_marker_setting('meta_data_dir_size'),
                                           self.get_marker_setting('col_row_index_size'))
            except:
                data = []

            mapping = self._map_data(self.get_marker_setting('min_size'),
                                     self.get_marker_setting('max_size'), data)
            for size, marker in zip(mapping, self.marker_list):
                marker.set_size(QSize(size, size))

    def _update_labels(self):
        """Update display of labels based on label settings."""
        # no labels when labels are used as markers
        if self.get_marker_setting('toggle_button_id'):
            return
        
        if self.get_setting('labels') == 'choice_indexed':
            for index, marker in self._marker_map.iteritems():
                # set index, making sure it's 1-indexed
                label = self.get_setting('label_index_prefix') + str(index + 1)
                marker.setLabel(label)
        elif self.get_setting('labels') == 'choice_data':
            is_column = self.get_setting('label_data_dir') == 'column'
            index = int(self.get_setting('label_data_index'))
            data = self._node.get_data(self.get_setting('label_data_port'),
                                       is_column, index)
            for index, marker in self._marker_map.iteritems():
                if len(data) > index:
                    text = str(data[index])
                else:
                    text = ""
                marker.setLabel(text)
        else:
            for marker in self.marker_list:
                marker.setLabel("")

        if self.marker_list:
            marker = self.marker_list[0]
            font = QFont(marker.font())
            font.setPointSize(int(self.get_setting('label_size')))
        for marker in self.marker_list:
            marker.setFont(font)
            marker.setLabelAlignment(Qt.AlignRight | Qt.AlignTop)

    def get_tooltip_at(self, x, y):
        """Get (tooltip_string, pixel-distance) near coordinate x, y.

        Coordinates are pixel coordinates in plot area. pixel-distance
        is distance between pixel at plot coordinate x, y and
        marker/curve point representing position showing the tooltip.

        None is returned as tooltip_string when there are no
        markers/curves/tooltips available. Value of pixel-distance is
        then undefined.
        """
        if self.get_setting('label_tooltips') == 'None':
            return (None, 0)

        if not self._x_values:
            return None, 0

        xmap = self._parent_plot.canvasMap(QwtPlot.xBottom)
        ymap = self._parent_plot.canvasMap(QwtPlot.yLeft)
        
        closest = None
        closest_distance = 0

        for index, point_x, point_y in zip(self._indices, self._x_values, self._y_values):
            dx = x - xmap.transform(point_x)
            dy = y - ymap.transform(point_y)
            distance = dx*dx + dy*dy
            if closest == None or distance < closest_distance:
                closest = index
                closest_distance = distance

        distance = math.sqrt(closest_distance)
                
        if self.get_setting('label_tooltips') == 'From data':
            portname = self.get_setting('label_tooltips_data')
            is_column = self.get_setting('label_tooltips_colrow') ==  'column'
            index = self.get_setting('label_tooltips_index')
            label_list = self._node.get_data(portname, is_column, index)
            try:
                return str(label_list[closest]), distance
            except IndexError:
                return "", distance
        else:
            # indexed 
            return str(closest+1), distance

    def get_data_length(self):
        return self._extended_length
    

class ScatterPlotSettings(PlotSettings):
    def __init__(self, node, index, tab_widget, qwt_plot_widget, plot):
        """Constructor.

        Parameters:
        node            -- the node to retrieve data from
        index           -- the index in the listbox of the tab_widget
        tab_widget      -- the tab_widget the associated widget is drawn in
        qwt_plot_widget -- the plot_widget in which the curve we control is drawn
        """

        self._x_data_selector = None
        self._y_data_selector = None
        
        PlotSettings.__init__(self, node, index, tab_widget, qwt_plot_widget,
                              plot)
        self._marker_settings_dialog = MarkerSettingsDialog(tab_widget,
                                                            node,
                                                            qwt_plot_widget, 
                                                            plot)
        self._marker_settings_dialog.init_dialog(False) # hidden line settings.
        self._database_info_dialog = DatabaseInfoDialog(tab_widget,
                                                        node,
                                                        qwt_plot_widget, 
                                                        tab_widget.get_widget_id(),
                                                        index)        
        self._load_settings()
        self._label_settings_dialog = LabelSettingsDialog(self, node,
                                                          qwt_plot_widget,
                                                          plot)

    
        x_data_id = self.get_parameter('x_axis_data')
        x_col_row = not int(self.get_parameter('x_col_row'))
        x_index =  int(self.get_parameter('x_index')) + 1 
        y_data_id = self.get_parameter('y_axis_data') 
        y_col_row = not int(self.get_parameter('y_col_row'))
        y_index =  int(self.get_parameter('y_index')) + 1

        self._x_data_selector = DataSelector(x_data_id, x_col_row, x_index)
        self._y_data_selector = DataSelector(y_data_id, y_col_row, y_index)        

        self._init_widget()
        self._set_widget_values()
        self._qwt_plot_widget = qwt_plot_widget
        self._load_tooltip_settings()
        self.update_data_names_combo_boxes()
        
        self.update_data_names_combo_boxes()

    def _init_widget(self):
        """Initiates the associated widget."""
        self._settings_widget = QWidgetFactory.create(
            os.path.join(Config.uidir, 'form_scatter_plot_setup.ui'),
            self._tab_widget, self._tab_widget)
        
        self._line_edit_plot_name = self._settings_widget.child('line_edit_plot_name')
        self._combo_box_x_axis_data = self._settings_widget.child('combo_box_x_axis_data')
        self._combo_box_y_axis_data = self._settings_widget.child('combo_box_y_axis_data')
        self._combo_box_x_col_row = self._settings_widget.child('combo_box_x_col_row')
        self._combo_box_y_col_row = self._settings_widget.child('combo_box_y_col_row')
        self._spin_box_x_index = self._settings_widget.child('spin_box_x_index')
        self._spin_box_y_index = self._settings_widget.child('spin_box_y_index')
        self._line_edit_range = self._settings_widget.child('line_edit_range')
        self._autoupdate_checkbox = self._settings_widget.child('autoupdate_checkbox')
        self._push_button_label_settings = self._settings_widget.child('push_button_label_settings')
        self._push_button_marker_setup = self._settings_widget.child('push_button_marker_setup')
        self._push_button_database_info = self._settings_widget.child('push_button_database_info')
        update_button = self._settings_widget.child('update_button')
        tooltips_button = self._settings_widget.child('push_button_tooltips')

        connect(self._push_button_marker_setup, SIGNAL('clicked()'),
                self._marker_settings_dialog.display_dialog)
        connect(self._push_button_database_info, SIGNAL('clicked()'),
                self._slot_database_info_dialog_clicked)
        connect(self._push_button_label_settings, SIGNAL('clicked()'),
                self.__slot_push_button_labels_clicked)
        connect(self._line_edit_plot_name,
                SIGNAL('textChanged(const QString &)'),
                self._slot_plot_name_changed)
        connect(self._line_edit_range, SIGNAL('returnPressed()'),
                self._slot_range_return_pressed)
        connect(self._line_edit_range, SIGNAL('textChanged(const QString &)'),
                self._slot_line_edit_range_changed)
        connect(self._autoupdate_checkbox, SIGNAL('toggled(bool)'),
                self._slot_checkbox_changed)        
        connect(update_button, SIGNAL('clicked()'), self.update_data)
        connect(tooltips_button, SIGNAL('clicked()'), self._slot_set_tooltips)

        self._set_inports_in_combo_box(self._combo_box_x_axis_data)
        self._set_inports_in_combo_box(self._combo_box_y_axis_data)
        
        self._tab_widget.layout().addWidget(self._settings_widget, 0, 1)

        self._x_data_selector.set_source_combo(self._combo_box_x_axis_data)
        self._x_data_selector.set_mode_combo(self._combo_box_x_col_row)
        self._x_data_selector.set_index_spinbox(self._spin_box_x_index)

        self._y_data_selector.set_source_combo(self._combo_box_y_axis_data)
        self._y_data_selector.set_mode_combo(self._combo_box_y_col_row)
        self._y_data_selector.set_index_spinbox(self._spin_box_y_index)

        connect(self._x_data_selector, PYSIGNAL("dataChanged()"), self._update_when_autoupdate)
        connect(self._y_data_selector, PYSIGNAL("dataChanged()"), self._update_when_autoupdate)

    def _update_when_autoupdate(self):
        self.save_settings()
        if self.get_parameter('autoupdate'):
            self.update_data()            
    
    def _slot_database_info_dialog_clicked(self):
        self._database_info_dialog.display_dialog(self._plot)
        
    def update_data(self):
        """Updates the plot and sets the data in the plot_widget"""
        PlotSettings.update_data(self)
        x_data = self._x_data_selector.get_list()
        y_data = self._y_data_selector.get_list()

        if not x_data or not y_data:
            # there is no data ports
            self._qwt_plot_widget.set_data_by_plot(self._plot, [], [])
            self._qwt_plot_widget.replot()
            return
  
        if x_data and len(x_data) != len(y_data):
            QMessageBox.information(
                self._tab_widget,
                'Plot - inconsistent data',
                'The number of elements in the X-data is not the same\n' +\
                'as the number of elements in the Y-data. Stripping\n' +\
                'whichever is the longest.')
            x_data = x_data[:len(y_data)]
            y_data = y_data[:len(x_data)]
  
        # make 0-indexed list of indices into dataset using ydata
        # as template as xdata might be empty when 'Indexed'
        try:
            indices = rangeparser.matlab_parse_range(self.get_parameter('range'),
                                                     range(len(y_data)))
        except InvalidRangeError, e:
            QMessageBox.warning(self._tab_widget,
                                'Plot - range notation error:',
                                str(e) + "\nShowing all values.")
            indices = None
  
        self._qwt_plot_widget.set_data_by_plot(self._plot, x_data, y_data,
                                                 indices)
  
        self._load_tooltip_settings()
        self._label_settings_dialog.apply_settings()        
        self._marker_settings_dialog._set_values()
        self._marker_settings_dialog.apply_settings()
    
    def _load_settings(self):
        
        # set defaults
        default_input_port = ''
        input_ports = self._node.get_all_input_ports()
        if input_ports:
            default_input_port = input_ports[0].getName()
        self.set_parameter('x_axis_data', default_input_port)
        self.set_parameter('y_axis_data', default_input_port)
        self.set_parameter('x_col_row', '0')
        self.set_parameter('y_col_row', '0')
        self.set_parameter('x_index', '0')
        self.set_parameter('y_index', '0')
        self.set_parameter('range', '1:end')
        self.set_parameter('labels', '0')
        self.set_parameter('autoupdate', False)

        for key, value in self._plot.get_settings().items():
            self.set_parameter(key, value)

    def save_settings(self):
        if self._x_data_selector and self._y_data_selector:
            self.set_parameter('x_axis_data', self._x_data_selector.get_source_id())
            self.set_parameter('y_axis_data', self._y_data_selector.get_source_id())
            
            if self._x_data_selector.is_column_mode():
                self.set_parameter('x_col_row', 0)
            else:
                self.set_parameter('x_col_row', 1)
            if self._y_data_selector.is_column_mode():
                self.set_parameter('y_col_row', 0)
            else:
                self.set_parameter('y_col_row', 1)

            self.set_parameter('x_index', self._x_data_selector.get_index() - 1)
            self.set_parameter('y_index', self._y_data_selector.get_index() - 1)
        
        PlotSettings.save_settings(self)

    def _set_widget_values(self):
        self._line_edit_range.setText(self.get_parameter('range'))
        self._line_edit_plot_name.setText(self.get_parameter('name'))

        self._autoupdate_checkbox.setChecked(self.get_parameter('autoupdate'))
        if self.get_parameter('autoupdate'):
            self.update_data()
            
    def update_data_names_combo_boxes(self):
        self._x_data_selector.set_active_sources(self._node.get_all_input_ports())
        self._y_data_selector.set_active_sources(self._node.get_all_input_ports())
        
        if not self._node.get_all_input_ports():
            #no input ports -> no label setting
            self._push_button_label_settings.setEnabled(False)
        else:
            self._push_button_label_settings.setEnabled(True)
        self._label_settings_dialog.update_data_names_combo_boxes()

    def __slot_push_button_labels_clicked(self):
        """Slot to run when the push_button_label_settings is pushed."""
        self._label_settings_dialog.display_dialog()


class MarkerSettingsDialog(Settings):

    _marker_symbols = [QwtSymbol.Ellipse, QwtSymbol.Cross,
                       QwtSymbol.XCross, QwtSymbol.Rect, 
                       QwtSymbol.Diamond, QwtSymbol.Triangle]

    def __init__(self, parent, node, qwt_plot_widget, plot):
        Settings.__init__(self)

        self._plot_widget = qwt_plot_widget
        self._node = node
        self._plot = plot
        
        # Setup dialog and tabwidget for both lines and markers.
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'plot_setup.ui'),
                                             None, parent)

        self._show_line_settings = True

        self._line_types = [Qt.SolidLine, Qt.DashLine, Qt.DotLine,
                            Qt.DashDotLine, Qt.DashDotDotLine]

        self._current_font = None

    def init_dialog(self, show_line_settings=True, show_meta_settings=True):
        # Get child widgets
        self._twidget_plot = self._widget.child('tab_plot_setup')
        
        # Buttons
        button_ok = self._widget.child('pbutton_ok')
        button_cancel = self._widget.child('pbutton_cancel')
        button_apply = self._widget.child('pbutton_apply')

        # Connect signals
        connect(button_ok, SIGNAL('clicked()'), self._slot_ok)
        connect(button_cancel, SIGNAL('clicked()'), self._widget.close)
        connect(button_apply, SIGNAL('clicked()'), self._apply_changes)
        
        # Marker tab
        self._marker_tab = self._twidget_plot.child('marker_tab')
        self._cbox_marker_symbol = self._marker_tab.child('cbox_marker_symbol')
        self._bgroup_marker_type = self._marker_tab.child('bgroup_marker_type')
        self._marker_colour_button = self._marker_tab.child('marker_colour_button')
        self._marker_size = self._marker_tab.child('spinbox_marker_size')
        self._marker_thickness = self._marker_tab.child('spinbox_marker_thickness')
        self._label_marker_thickness = self._marker_tab.child('tlabel_marker_thickness')
        self._marker_text = self._marker_tab.child('marker_text')
        self._marker_text_font = self._marker_tab.child('text_marker_font')
        self._gbox_marker_style = self._marker_tab.child('gbox_marker_style')
        self._marker_preview = self._marker_tab.child('marker_preview')
        self._label_colour = self._marker_tab.child('tlabel_colour')
        self._label_msize = self._marker_tab.child('tlabel_msize')

        # Properties
        self._marker_size.setMaxValue(40)
        self._marker_size.setMinValue(1)
        self._marker_thickness.setMaxValue(40)
        self._marker_thickness.setMinValue(1)

        # Connect signals
        connect(self._bgroup_marker_type, SIGNAL('clicked(int)'),
                self._slot_marker_type_changed)
        connect(self._marker_colour_button, SIGNAL('clicked()'),
                self._slot_marker_colour_button_clicked)
        connect(self._marker_text_font, SIGNAL('clicked()'),
                self._slot_font_button_clicked)
        connect(self._marker_size, SIGNAL('valueChanged(int)'),
                self._slot_marker_size_changed)
        connect(self._marker_thickness, SIGNAL('valueChanged(int)'),
                self._slot_marker_thickness_changed)
        connect(self._cbox_marker_symbol, SIGNAL('activated(int)'),
                self._slot_marker_symbol_changed)

        # Line tab
        self._line_tab = self._twidget_plot.child('line_tab')
        self._cbox_line_type = self._line_tab.child('cbox_line_type')
        self._button_line_colour = self._line_tab.child('pbutton_line_colour')
        self._sbox_line_size = self._line_tab.child('sbox_line_size')
        self._line_preview = self._line_tab.child('line_preview')

        # Properties
        self._sbox_line_size.setMaxValue(30)
        self._sbox_line_size.setMinValue(1)

        # Connect signals
        connect(self._button_line_colour,
                                   SIGNAL('clicked()'),
                                   self._slot_line_colour_button_clicked)
        connect(self._sbox_line_size,
                                   SIGNAL('valueChanged(int)'),
                                   self._slot_line_width_changed)
        connect(self._cbox_line_type,
                                   SIGNAL('activated(int)'),
                                   self._slot_line_type_changed)

        # Meta data tab
        self._meta_data_tab = self._twidget_plot.child('meta_data_tab')
        self._cbox_meta_input = self._meta_data_tab.child('cbox_meta_input')
        self._cbox_col_row = self._meta_data_tab.child('cbox_col_row')
        self._sbox_col_row_index = self._meta_data_tab.child('spinbox_index')
        self._groupbox_map_colour = self._meta_data_tab.child('gbox_colourmapping')
        self._min_colour_button = self._meta_data_tab.child('button_min_colour')
        self._max_colour_button = self._meta_data_tab.child('button_max_colour')

        self._cbox_meta_input_size = self._meta_data_tab.child('cbox_meta_input_2')
        self._cbox_col_row_size = self._meta_data_tab.child('cbox_col_row_2')
        self._sbox_col_row_index_size = self._meta_data_tab.child('spinbox_index_2')
        self._groupbox_map_size = self._meta_data_tab.child('gbox_sizemapping')
        self._sbox_min_size = self._meta_data_tab.child('spinbox_min_size')
        self._sbox_max_size = self._meta_data_tab.child('spinbox_max_size')

        # Properties
        self._sbox_min_size.setMinValue(1)
        self._sbox_max_size.setMinValue(1)
        self._sbox_col_row_index.setMaxValue(sys.maxint)
        self._sbox_col_row_index.setMinValue(1)

        # Connect signals
        connect(self._groupbox_map_colour, SIGNAL('toggled(bool)'),
                self._slot_map_to_colour_checked)
        connect(self._groupbox_map_size, SIGNAL('toggled(bool)'),
                self._slot_map_to_size_checked)
        connect(self._min_colour_button, SIGNAL('clicked()'),
                self._slot_min_colour_button_clicked)
        connect(self._max_colour_button, SIGNAL('clicked()'),
                self._slot_max_colour_button_clicked)
        connect(self._sbox_max_size, SIGNAL('valueChanged(int)'),
                self._slot_max_size_button_clicked)
        connect(self._sbox_min_size, SIGNAL('valueChanged(int)'),
                self._slot_min_size_button_clicked)
        connect(self._cbox_meta_input, SIGNAL('activated(int)'),
                self._slot_input_port_changed)
        connect(self._cbox_col_row, SIGNAL('activated(int)'),
                self._slot_meta_data_dir_changed)
        connect(self._cbox_meta_input_size, SIGNAL('activated(int)'),
                self._slot_input_port_size_changed)
        connect(self._cbox_col_row, SIGNAL('activated(int)'),
                self._slot_meta_data_dir_size_changed)

        if not show_line_settings:
            self._show_line_settings = False
            self._twidget_plot.removePage(self._line_tab)
            self._widget.setCaption("Marker Settings")
        if not show_meta_settings:
            self._twidget_plot.removePage(self._meta_data_tab)            
            
        
        marker_pixmaps = self._generate_marker_symbols()
        marker_symbol_desc = ["Ellipse", "Cross", "XCross", "Rect", "Diamond",
                              "Triangle"]
        self._add_combobox_items(self._cbox_marker_symbol, marker_pixmaps,
                                 marker_symbol_desc)

        line_pixmaps = self._generate_lines()
        self._add_combobox_items(self._cbox_line_type, line_pixmaps)

        # Now, try to load saved settings.
        params = {}
        try:
            params = self._load_settings()
        except Exception, e:
            pass

        if params:
            self._parameters = params
        else:
            self._set_default_values()

        self._set_values()

    def display_dialog(self):
        self._set_values()
        # execute it modally
        self._widget.exec_loop()

    def apply_settings(self):
        self._apply_changes()

    def _set_default_values(self):
        marker_symbol_index = 0
        marker_colour = (0, 0, 0)
        marker_size = 3
        marker_thickness = 1
        marker_text = ''
        marker_text_font = ('Times', 12, 0, 0)
        marker_font_list = list(marker_text_font)
        marker_font_list.insert(4, marker_colour)
        marker_font_list.insert(0, marker_text)
        text_marker = tuple(marker_font_list)
        
        self.set_parameter('marker_symbol_index', marker_symbol_index)
        self.set_parameter('marker_text', marker_text)
        self.set_parameter('marker_text_font', marker_text_font)
        self.set_parameter('toggle_button_id', 0)
        self.set_parameter('marker_colour', marker_colour)
        self.set_parameter('marker_size', marker_size)
        self.set_parameter('marker_thickness', marker_thickness)

        if self._show_line_settings:
            self.set_parameter('type', 'Line')
            line_type_index = 0
            line_colour = (0, 0, 255) # blue
            line_width = 2
            
            self.set_parameter('line_type_index', line_type_index)
            self.set_parameter('line_colour', line_colour)
            self.set_parameter('line_width', line_width)


        input_ports = map(lambda port : port.getName(),
                          self._node.get_all_input_ports()) # Get port names.
        default_port = ""
        if input_ports:
            default_port = input_ports[0]

        self.set_parameter('meta_data_input_port', default_port)
        self.set_parameter('meta_data_dir', 0) # column
        self.set_parameter('col_row_index', 0)
        self.set_parameter('meta_data_input_port_size', default_port)
        self.set_parameter('meta_data_dir_size', 0) # column
        self.set_parameter('col_row_index_size', 0)
        self.set_parameter('map_to_colour', 0) # False
        self.set_parameter('map_to_size', 0) # False
        self.set_parameter('min_colour', (0, 0, 0))
        self.set_parameter('max_colour', (0, 0, 0))
        self.set_parameter('min_size', 1)
        self.set_parameter('max_size', 1)

    def _set_values(self):
        symbol_index = self.get_parameter('marker_symbol_index')
        symbol = self._marker_symbols[symbol_index]
        marker_size = self.get_parameter('marker_size')
        marker_thickness = self.get_parameter('marker_thickness')
        r, g, b = self.get_parameter('marker_colour')
        marker_colour = QColor(r, g, b)
        marker_text = self.get_parameter('marker_text')
        font = self.get_parameter('marker_text_font')
        
        self._cbox_marker_symbol.setCurrentItem(symbol_index)
        toggle_button_id = self.get_parameter('toggle_button_id')
        if self._bgroup_marker_type.selectedId() != toggle_button_id:
            self._slot_marker_type_changed(toggle_button_id)
        self._bgroup_marker_type.setButton(toggle_button_id)
        self._marker_colour_button.setPaletteBackgroundColor(marker_colour)
        self._marker_size.setValue(marker_size)
        self._marker_thickness.setValue(marker_thickness)
        self._marker_text.setText(marker_text)
        self._current_font = font

        self._update_marker_preview(self._marker_symbols[symbol_index],
                                    marker_size, marker_thickness, marker_colour)

        if self._show_line_settings:
            line_type_index = self.get_parameter('line_type_index')
            line_style = self._line_types[line_type_index]
            line_width = self.get_parameter('line_width')
            r, g, b = self.get_parameter('line_colour')
            line_colour = QColor(r, g, b)
            
            self._cbox_line_type.setCurrentItem(line_type_index)
            self._button_line_colour.setPaletteBackgroundColor(line_colour)
            self._sbox_line_size.setValue(line_width)
            
            self._update_line_preview(line_width, line_style, line_colour)

        col_row = self.get_parameter('meta_data_dir')
        col_row_index = self.get_parameter('col_row_index')
        col_row_size = self.get_parameter('meta_data_dir_size')
        col_row_index_size = self.get_parameter('col_row_index_size')
        map_to_colour = self.get_parameter('map_to_colour')
        map_to_size = self.get_parameter('map_to_size')
        min_r, min_g, min_b = self.get_parameter('min_colour')
        max_r, max_g, max_b = self.get_parameter('max_colour')
        min_colour = QColor(min_r, min_g, min_b)
        max_colour = QColor(max_r, max_g, max_b)
        min_size = self.get_parameter('min_size')
        max_size = self.get_parameter('max_size')
        
        self._update_meta_data_input_ports(self._cbox_meta_input,
                                           'meta_data_input_port')
        self._update_meta_data_input_ports(self._cbox_meta_input_size,
                                           'meta_data_input_port_size')
        self._cbox_col_row.setCurrentItem(col_row)
        self._cbox_col_row_size.setCurrentItem(col_row_size)
        self._set_max_value(self._sbox_col_row_index, col_row_index,
                            self._cbox_col_row.currentText()=='column',
                            self._cbox_meta_input)
        self._set_max_value(self._sbox_col_row_index_size, col_row_index_size,
                            self._cbox_col_row_size.currentText() == 'column',
                            self._cbox_meta_input_size)
        if map_to_size:
            self._groupbox_map_size.setChecked(True)
        else:
            self._groupbox_map_size.setChecked(False)
        if map_to_colour:
            self._groupbox_map_colour.setChecked(True)
        else:
            self._groupbox_map_colour.setChecked(False)
        self._min_colour_button.setPaletteBackgroundColor(min_colour)
        self._max_colour_button.setPaletteBackgroundColor(max_colour)
        self._sbox_min_size.setValue(min_size)
        self._sbox_max_size.setValue(max_size)
        

    def _apply_changes(self):
        symbol_index = self._cbox_marker_symbol.currentItem()
        marker_colour = self._marker_colour_button.paletteBackgroundColor()
        r = marker_colour.red()
        g = marker_colour.green()
        b = marker_colour.blue()
        marker_size = self._marker_size.value()
        marker_thickness = self._marker_thickness.value()
        marker_text = str(self._marker_text.text())
        marker_font_list = list(self._current_font)
        marker_font_list.insert(4, (r, g, b))
        marker_font_list.insert(0, marker_text)
        text_marker = tuple(marker_font_list)
        
        map_to_colour = 0
        map_to_size = 0
        if self._groupbox_map_colour.isChecked():
            map_to_colour = 1
        if self._groupbox_map_size.isChecked():
            map_to_size = 1

        self.set_parameter('marker_symbol_index', symbol_index)
        self.set_parameter('marker_text', marker_text)
        self.set_parameter('marker_text_font', self._current_font)
        self.set_parameter('toggle_button_id',
                           self._bgroup_marker_type.selectedId())
        self.set_parameter('marker_colour', (r, g, b))
        self.set_parameter('marker_size', marker_size)
        self.set_parameter('marker_thickness', marker_thickness)

        if self._show_line_settings:
            line_type_index = self._cbox_line_type.currentItem()
            line_colour = self._button_line_colour.paletteBackgroundColor()
            r = line_colour.red()
            g = line_colour.green()
            b = line_colour.blue()
            line_width = self._sbox_line_size.value()

            self.set_parameter('line_type_index', line_type_index)
            self.set_parameter('line_colour', (r, g, b))
            self.set_parameter('line_width', line_width)

        port_name = str(self._cbox_meta_input.currentText())
        is_column = str(self._cbox_col_row.currentText()) == 'column'
        index = self._sbox_col_row_index.value() - 1
        port_name_size = str(self._cbox_meta_input_size.currentText())
        is_column_size = str(self._cbox_col_row.currentText()) == 'column'
        index_size = self._sbox_col_row_index_size.value() - 1
        from_colour = self._min_colour_button.paletteBackgroundColor()
        from_colour_r = from_colour.red()
        from_colour_g = from_colour.green()
        from_colour_b = from_colour.blue()
        from_colour = (from_colour_r, from_colour_g, from_colour_b)
        to_colour = self._max_colour_button.paletteBackgroundColor()
        to_colour_r = to_colour.red()
        to_colour_g = to_colour.green()
        to_colour_b = to_colour.blue()
        to_colour = (to_colour_r, to_colour_g, to_colour_b)
        min_size = self._sbox_min_size.value()
        max_size = self._sbox_max_size.value()
        
        self.set_parameter('meta_data_input_port', port_name)
        self.set_parameter('meta_data_dir', self._cbox_col_row.currentItem())
        self.set_parameter('col_row_index', index)
        self.set_parameter('meta_data_input_port_size', port_name_size)
        self.set_parameter('meta_data_dir_size',
                           self._cbox_col_row_size.currentItem())
        self.set_parameter('col_row_index_size', index_size)
        self.set_parameter('map_to_colour', map_to_colour)
        self.set_parameter('map_to_size', map_to_size)
        self.set_parameter('min_colour', from_colour)
        self.set_parameter('max_colour', to_colour)
        self.set_parameter('max_size', max_size)
        self.set_parameter('min_size', min_size)

        self._save_settings()
        self._plot.update()
        self._plot_widget.replot()

    def __merge(self, dict_target, dict_source):
        for key in dict_source.keys():
            if key not in dict_target.keys():
                dict_target[key] = dict_source[key]

    def _load_settings(self):
        # load default settings for all settings not set in file
        self._set_default_values()
        settings = self._plot.get_marker_settings()
        
        # merge default settings with settings loaded from file
        self.__merge(settings, self._parameters)

        return self._convert_dict(settings)

    def _save_settings(self):
        self._plot.set_marker_settings(self._parameters)

    def _generate_marker_symbols(self):
        items = []
            
        for style in self._marker_symbols:
            width = 15
            height = 15
            symbol = QwtSymbol(style, QBrush(Qt.black),
                               QPen(Qt.black), QSize(width, height))
                
            pixmap = QPixmap(width, height+2)
            pixmap.fill(self._cbox_marker_symbol.paletteBackgroundColor())
            painter = QPainter(pixmap)
            symbol.draw(painter, (width/2), (height/2+1))
            
            items.append(pixmap)
                
        return items

    def _generate_lines(self):
        lines = []
        
        for line_type in self._line_types: 
            width = 320
            height = 20
            
            pixmap = QPixmap(width, height)
            pixmap.fill(self._cbox_line_type.paletteBackgroundColor())
            painter = QPainter(pixmap)
            pen = QPen(QColor(0, 0, 0), 1, line_type)
            painter.setPen(pen)
            painter.drawLine(0, (height/2), width, (height/2))
            lines.append(pixmap)

        return lines

    def _add_combobox_items(self, combobox, pixmaps, text_list=[]):
        if len(pixmaps) == len(text_list):
            for i in range(len(pixmaps)):
                combobox.insertItem(pixmaps[i], text_list[i])
        else:
            for pixmap in pixmaps:
                combobox.insertItem(pixmap)

    def _update_meta_data_input_ports(self, cbox, port_param_name):
        """Updates the combo box containing the input ports."""
        cbox.clear()
        for inputport in self._node.get_all_input_ports():
            cbox.insertItem(inputport.getName())

        port = self.get_parameter(port_param_name)
        index = gui.get_index(cbox, port)
        if index != -1:
            cbox.setCurrentItem(index)

    def _update_line_preview(self, width, style, colour):
        pen = QPen(colour, width, style)
        width = 396
        height = self._line_preview.height()
        self.pixmap = QPixmap(width, height)
        self.pixmap.fill(self._line_preview.paletteBackgroundColor())
        painter = QPainter(self.pixmap)
        painter.setPen(pen)
        painter.drawLine(0, height/2, width-1, height/2)
        
        self._line_preview.setPixmap(self.pixmap)

    def _update_marker_preview(self, style, size, thickness, colour):
        pen = QPen(colour, thickness)
        brush = QBrush(colour)
        size_ = QSize(size, size)
        symbol = QwtSymbol(style, brush, pen, size_)
        
        self.pixmap = QPixmap(size+2, size+2)
        self.pixmap.fill(self._marker_preview.paletteBackgroundColor())
        painter = QPainter(self.pixmap)
        symbol.draw(painter, (size/2), (size/2))

        self._marker_preview.setPixmap(self.pixmap)

    def _set_max_value(self, spinbox, default_value, is_column, cbox_inputdata):
        port_name = str(cbox_inputdata.currentText())
        max_size = sys.maxint

        if port_name:
            try:
                max_size = self._node.get_data_size(port_name, is_column)
            except AttributeError:
                pass
        
        spinbox.setValue(default_value+1)
        spinbox.setMaxValue(max_size)

    # Slots
    def _slot_ok(self):
        self._apply_changes()
        self._widget.close() 

    def _slot_marker_type_changed(self, button_id):
        if button_id == 0:
            self._marker_text.setEnabled(False)
            self._marker_text_font.setEnabled(False)
            self._cbox_marker_symbol.setEnabled(True)
            self._marker_size.setEnabled(True)
            self._label_msize.setEnabled(True)
            self._marker_thickness.setEnabled(True)
            self._label_marker_thickness.setEnabled(True)
        elif button_id == 1:
            self._marker_text.setEnabled(True)
            self._marker_text_font.setEnabled(True)
            self._cbox_marker_symbol.setEnabled(False)
            self._marker_size.setEnabled(False)
            self._label_msize.setEnabled(False)
            self._marker_thickness.setEnabled(False)
            self._label_marker_thickness.setEnabled(False)

    def _slot_marker_colour_button_clicked(self):
        colour = QColorDialog.getColor(self._marker_colour_button.paletteBackgroundColor(),
                                       self._widget)
        if colour.isValid():
            self._marker_colour_button.setPaletteBackgroundColor(colour)
            style = self._marker_symbols[self._cbox_marker_symbol.currentItem()]
            thickness = self._marker_thickness.value()
            size = self._marker_size.value()
            
            self._update_marker_preview(style, size, thickness, colour)

    def _slot_marker_size_changed(self, size):
        style = self._marker_symbols[self._cbox_marker_symbol.currentItem()]
        thickness = self._marker_thickness.value()
        colour = self._marker_colour_button.paletteBackgroundColor()
        
        self._update_marker_preview(style, size, thickness, colour)

    def _slot_marker_thickness_changed(self, thickness):
        style = self._marker_symbols[self._cbox_marker_symbol.currentItem()]
        colour = self._marker_colour_button.paletteBackgroundColor()
        size = self._marker_size.value()
        
        self._update_marker_preview(style, size, thickness, colour)

    def _slot_marker_symbol_changed(self, style_index):
        style = self._marker_symbols[style_index]
        colour = self._marker_colour_button.paletteBackgroundColor()
        size = self._marker_size.value()
        thickness = self._marker_thickness.value()

        self._update_marker_preview(style, size, thickness, colour)
        
    def _slot_font_button_clicked(self):
        family, size, weight, italic = self._current_font
        current_font = QFont(family, size, weight, italic)
        font, ok = QFontDialog.getFont(current_font, self._widget)
        if ok:
            family = str(font.family())
            size = font.pointSize()
            weight = font.weight()
            is_italic = font.italic()
            italic = 0
            if is_italic:
                italic = 1
            
            self._current_font = (family, size, weight, italic)

    def _slot_line_colour_button_clicked(self): 
        colour = QColorDialog.getColor(self._button_line_colour.paletteBackgroundColor(),
                                          self._widget)
        if colour.isValid():
            self._button_line_colour.setPaletteBackgroundColor(colour)
            width = self._sbox_line_size.value()
            style = self._line_types[self._cbox_line_type.currentItem()]
            self._update_line_preview(width, style, colour)

    def _slot_line_width_changed(self, width):
        style = self._line_types[self._cbox_line_type.currentItem()]
        colour = self._button_line_colour.paletteBackgroundColor()
        self._update_line_preview(width, style, colour)

    def _slot_line_type_changed(self, index):
        style = self._line_types[index]
        colour = self._button_line_colour.paletteBackgroundColor()
        width = self._sbox_line_size.value()
        self._update_line_preview(width, style, colour)

    def _slot_map_to_colour_checked(self, on):
        if on:
            self._bgroup_marker_type.find(1).setEnabled(False)
            self._bgroup_marker_type.find(0).setEnabled(True)
            self._cbox_marker_symbol.setEnabled(True)
            self._marker_text.setEnabled(False)
            self._marker_text_font.setEnabled(False)
            self._label_colour.setEnabled(False)
            self._marker_colour_button.setEnabled(False)
            self._label_marker_thickness.setEnabled(True)
            self._marker_thickness.setEnabled(True)
            self._bgroup_marker_type.setButton(0)
            if not self._groupbox_map_size.isChecked():
                self._label_msize.setEnabled(True)
                self._marker_size.setEnabled(True)
        else:
            if not self._groupbox_map_size.isChecked():
                self._bgroup_marker_type.find(1).setEnabled(True)
            self._label_colour.setEnabled(True)
            self._marker_colour_button.setEnabled(True)

    def _slot_map_to_size_checked(self, on):
        if on:
            self._bgroup_marker_type.find(1).setEnabled(False)
            self._bgroup_marker_type.find(0).setEnabled(True)
            self._cbox_marker_symbol.setEnabled(True)
            self._marker_text.setEnabled(False)
            self._marker_text_font.setEnabled(False)
            self._marker_size.setEnabled(False)
            self._label_msize.setEnabled(False)
            self._bgroup_marker_type.setButton(0)
            self._marker_thickness.setEnabled(True)
            self._bgroup_marker_type.setButton(0)
            if not self._groupbox_map_colour.isChecked():
                self._label_colour.setEnabled(True)
                self._marker_colour_button.setEnabled(True)
        else:
            if not self._groupbox_map_colour.isChecked():
                self._bgroup_marker_type.find(1).setEnabled(True)
            self._marker_size.setEnabled(True)
            self._label_msize.setEnabled(True)

    def _slot_min_colour_button_clicked(self):
        colour = QColorDialog.getColor(self._min_colour_button.paletteBackgroundColor(),
                                          self._widget)
        if colour.isValid():
            self._min_colour_button.setPaletteBackgroundColor(colour)
            self._groupbox_map_colour.setChecked(True)

    def _slot_max_colour_button_clicked(self):
        colour = QColorDialog.getColor(self._max_colour_button.paletteBackgroundColor(),
                                          self._widget)
        if colour.isValid():
            self._max_colour_button.setPaletteBackgroundColor(colour)
            self._groupbox_map_colour.setChecked(True)

    def _slot_max_size_button_clicked(self):
        self._groupbox_map_size.setChecked(True)

    def _slot_min_size_button_clicked(self):
        self._groupbox_map_size.setChecked(True)
        
    def _slot_input_port_changed(self, index):
        self._set_max_value(self._sbox_col_row_index, 0,
                            self._cbox_col_row.currentText()=='column',
                            self._cbox_meta_input)

    def _slot_input_port_size_changed(self, index):
        self._set_max_value(self._sbox_col_row_index_size, 0,
                            self._cbox_col_row_size.currentText()=='column',
                            self._cbox_meta_input_size)

    def _slot_meta_data_dir_changed(self, index):
        self._set_max_value(self._sbox_col_row_index, 0,
                            self._cbox_col_row.currentText()=='column',
                            self._cbox_meta_input)

    def _slot_meta_data_dir_size_changed(self, index):
        self._set_max_value(self._sbox_col_row_index_size, 0,
                            self._cbox_col_row_size.currentText()=='column',
                            self._cbox_meta_input_size)
