# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QWorkspace, QMessageBox, QTimer, QFileDialog, QDialog, SIGNAL
from GUI.Canvas.mdview import MDView
from System.Data.ModuleDiagram import ModuleDiagram, InvalidXMLException
from System.events import ErrorEvent, InfoEvent
from System.NodeRepository import UnknownNodeError
from Library.Util.Config import Config
from Library.Util.Counter import Counter

import os

def add_to_recent_files(filename):
    filename = os.path.abspath(filename)
    if not Config.recentFiles:
        Config.recentFiles = filename
    else:
        existing_files = Config.recentFiles.split(os.pathsep)
        if filename in existing_files:
            existing_files.remove(filename)
        if len(existing_files) > 4:
            existing_files = existing_files[:4]
        Config.recentFiles = filename + os.pathsep + os.pathsep.join(existing_files)


class Workspace(QWorkspace):
    
    def __init__(self, application, event_server, node_repository, cache):
        self._application = application
        QWorkspace.__init__(self, application._widget)
        
        self._cache = cache
        self._application = application
        self.event_server = event_server
        self.node_repository = node_repository
        self.node_list = [] # nodes that are available
        self.menu_counter = 0
        self.current_diagram = None
        self.diagrams = {}
        self._status_list = self._application.get_node_status_data_list()
        self.file_name = ""
        self.diagram_counter = Counter()
        
        self.setScrollBarsEnabled(True)
        
        self.connect(self, SIGNAL("windowActivated (QWidget *)"), self.set_current_diagram)

    def show_invalid_xml_dialog_box(self):
        QMessageBox.information(None, "Invalid input file",
                                "Can not open invalid SciCraft module diagram file.", QMessageBox.Ok)

    def new_diagram(self):
        """Method for opening a new (empty) diagram"""
        module_diagram = ModuleDiagram(self.node_repository, self.event_server, cache=self._cache)
        caption = "New Diagram - %d"%(self.diagram_counter.get_next_value())

        diagram_id = module_diagram.getId()
        self._application.add_window(diagram_id, caption)
        if self.current_diagram:
            self.current_diagram.set_pointing_tool_as_current_tool()
        self.current_diagram = MDView(self, self._application, module_diagram, self.event_server, None, caption)
        self.current_diagram.showMaximized()
        # Update pointer to MD for the node status dock window.
        self._status_list.set_module_diagram(module_diagram)

        self.diagrams[diagram_id] = self.current_diagram

    def load_diagram(self, filename, at_startup = False):
        """Method for opening a diagram when user has pushed the open button in the GUI"""
        try:
            module_diagram = ModuleDiagram(self.node_repository, self.event_server, filename, cache=self._cache)
            caption = os.path.basename(filename)
            diagram_id = module_diagram.getId()
            self._application.add_window(diagram_id, caption)
            if self.current_diagram:
                self.current_diagram.set_pointing_tool_as_current_tool()            
            self.current_diagram = MDView(self, self._application, module_diagram, self.event_server, filename, caption)
            self.current_diagram.showMaximized()
            # Update pointer to MD for the node status dock window.
            self._status_list.set_module_diagram(module_diagram)

            self.diagrams[diagram_id] = self.current_diagram
            add_to_recent_files(filename)
        except InvalidXMLException, e:
            # make dialogbox, don't open any module diagram
            QTimer.singleShot(0, self.show_invalid_xml_dialog_box)
            if at_startup:
                self.new_diagram()
        except UnknownNodeError, e:
            self.event_server.sendEvent(InfoEvent(None, "INFO", "Could not load diagram %s:\n%s\n\nMaybe you are missing the directory containing the function from your function path?" % (filename, e)))

    
    def save_diagram(self, save_as=0):
        if not self.current_diagram:
            return True
        self.current_diagram.save_diagram(save_as)

    def save_as_diagram(self):
        self.save_diagram(1)

    def save_eps_diagram(self):
        try:
            import pyx
        except ImportError, e:
            QMessageBox.warning(self._application._widget, "Missing Package", "PyX must be installed for this to work.")
            return
        except Exception, e:
            QMessageBox.warning(self._application._widget, "Could not create EPS", "Exception trying to import PyX:\n" + str(e))
            return

        if not self.current_diagram:
            return

        valid_suffices = {".eps": "Encapsulated PostScript (B&W) (*.eps)",
                         ".ps" : "Encapsulated PostScript (Colour) (*.eps)"}

        # The file save dialog
        file_dialog = QFileDialog(Config.last_path_modulediagram, valid_suffices.get(".eps"),
                                 self._application._widget, "fileDialog", True)

        file_dialog.addFilter(valid_suffices[".ps"])

        # Set the selected file to the previous seleced file
        file_dialog.setSelection("")
        file_dialog.setMode(QFileDialog.AnyFile)
        
        # Remember the last file filter used
        file_dialog.setSelectedFilter(0)
        file_dialog.setCaption("Save File")
        
        # User has pushed the save button
        if file_dialog.exec_loop() == QDialog.Accepted:
            file_filter = file_dialog.selectedFilter()
            full_file_name = file_dialog.selectedFile()
  
        # User has cancelled saving
        else:
            return
        
        # Update last saved path
        Config.last_path_modulediagram = os.path.dirname(str(full_file_name))

        # Special for EPS, check if black and white or colour
        if(file_filter.startsWith("Encapsulated PostScript (Colour)")):
            colour_mode = True
        else:
            colour_mode = False

        # Strip the name from the file path
        file_path, file_name = os.path.split(str(full_file_name))
   
        # Strip away the suffix (if any). Will be used for testing
        prefix, suffix = os.path.splitext(file_name)
     
       
        # pyx adds suffix .eps anyway
        suffix = ".eps"
          
        # Check if file already exists. If it does, prompt user
        # for advice
        if os.path.isfile(os.path.join(file_path, prefix + suffix)):
            retval = QMessageBox.warning(self._application._widget,
                                         "SciCraft",
                                         "Overwrite " + prefix + suffix + "?",
                                         QMessageBox.Yes,
                                         QMessageBox.No,
                                         QMessageBox.NoButton)

            # User won't overwrite, recall this method
            if retval == QMessageBox.No:
                self.save_eps_diagram()
                return

        # Ready to write the file        
        eps = self.current_diagram.to_eps(os.path.join(file_path, prefix + suffix),colour_mode)
                        
    def slot_run_diagram(self):
        if not self.current_diagram:
            return
        
        module_diagram = self.current_diagram.get_module_diagram()
        if module_diagram.getIsRunning():
            return
        
        invalid_ports, invalid_parameters, error_message = self.current_diagram.validate_diagram()
        if invalid_ports or invalid_parameters:
            self.event_server.sendEvent(ErrorEvent(self.current_diagram.get_diagram_id(), "VALIDATION", error_message))
            errmsg = "Could not validate the diagram: " + error_message
            #self.event_server.sendEvent(DiagramEvent(int(self.current_diagram.get_diagram_id()), "VALIDATIONERROR", errmsg))
        else:
            self.current_diagram.run_diagram()           

    def slot_stop_diagram(self):
        if not self.current_diagram:
            return

        module_diagram = self.current_diagram.get_module_diagram()
        if not module_diagram.getIsRunning():
            return

        module_diagram.stopMD()

    def set_pointing_tool_as_current_tool(self):
        for view in self.diagrams.values():
            view.set_pointing_tool_as_current_tool()

    def set_linking_tool_as_current_tool(self):
        for view in self.diagrams.values():
            view.set_linking_tool_as_current_tool()

    def set_comment_tool_as_current_tool(self):
        for view in self.diagrams.values():
            view.set_comment_tool_as_current_tool()
            
    def slot_validate_diagram(self):
        if not self.current_diagram:
            return
        
        module_diagram = self.current_diagram.get_module_diagram()
        if module_diagram.getIsRunning():
            return
        
        invalid_ports, invalid_parameters, error_message = self.current_diagram.validate_diagram()
        if invalid_ports or invalid_parameters:
            self.event_server.sendEvent(ErrorEvent(
                self.current_diagram.get_diagram_id(), "VALIDATION", error_message))
        else: 
            self.event_server.sendEvent(InfoEvent(
                self.current_diagram.get_diagram_id(), 'VALIDATION',
                "The module diagram is valid and ready to be executed"))
    
    def slot_add_node(self, number, key = 0):
        if not self.current_diagram:
            return

        if key:
            self.current_diagram.create_node(key)
        else:
            node_key = self.node_list[number]
            self.current_diagram.create_node(node_key)

    def update_canvas(self):
        for diagram in self.diagrams.values():
            diagram.update()
        
    def set_current_diagram(self, active_widget):
        # make sure that only fully existing diagrams are activated
        # (setCurrentDiagram might be called with an object in the
        # middle of its initiation).
        if active_widget in self.diagrams.values():
            self.current_diagram = active_widget
            self.current_diagram.activate()
            self._application.slot_window_clicked(active_widget.get_diagram_id())            
            # Update pointer to MD for the node status dock window.
            self._status_list.set_module_diagram(self.get_current_diagram())
            
    def get_current_diagram(self):
        if not self.current_diagram:
            return None
        
        return self.current_diagram.get_module_diagram()
    
    def remove_canvas(self, id):
        del self.diagrams[id]
        self._application.remove_window(id)
        # Update pointer to MD for the node status dock window.
        self._status_list.set_module_diagram(None)
        if not self.diagrams:
            self.current_diagram = None

    def window_change(self, id):
        self.diagrams[str(id)].setActiveWindow() 
        self.diagrams[str(id)].setFocus()
        # Update pointer to MD for the node status dock window.
        self._status_list.set_module_diagram(self.diagrams[str(id)].get_module_diagram())

    def close_all_windows(self):
        """Close all windows and return False if user aborted the operation."""
        while self.diagrams.values():
            if not self.close_active_window():
                return False

        return True
            
    def maximize_all_windows(self):        
        for diagram in self.diagrams.values():
            diagram.showMaximized()

    def minimize_all_windows(self):
        for diagram in self.diagrams.values():
            diagram.showMinimized()

    def close_active_window(self):
        if not self.current_diagram.close():
            return False

        self.set_current_diagram(self.activeWindow())
        return True
        
    def show_hidden_links(self):
        self.current_diagram.show_hidden_links()
        
    def hide_hidden_links(self):
        self.current_diagram.hide_hidden_links()
