# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.



from qt import QObject, QColor, QPopupMenu, Qt, QAction, QAccel, QFileDialog
from qt import QDialog, QMessageBox, QCloseEvent, SIGNAL, qApp, PYSIGNAL
from qtui import QWidgetFactory

from GUI.rightdockwindow import RightDockWindow

from GUI.Workspace import Workspace
from GUI.EditPrefs import EditPrefs
from GUI.zmlcreator import ZMLCreator
from GUI.viewers import AboutViewer, HelpViewer, ChangelogViewer
from GUI.SciCraftSplash import DummySplash
from GUI.tools.madesigner import MADesigner
from GUI.EventPopup import EventPopup

from rlibinstaller import RLibInstaller

from Library.Util.Config import Config, get_style, write_config
from Library.Util.gui import EventReceiver
from Library.Util.rtester import RTester, RInstallError
from Library.Util.file import CacheDirectory
from Library.Util.rpyclient import rpy_silent_eval, rpy_quit

from System.EventServer import EventServer
from System.NodeRepository import NodeRepository
from System.NodeStatusDataList import NodeStatusDataList
from System.Plugin.PluginRepository import PluginRepository
from System.events import InfoEvent

try:
    from Library.debug.pycute import PyCute
except:
    PyCute = None

import sys
import signal
import os
import qt

connect = QObject.connect

class ApplicationWindow(QObject, EventReceiver):    
    def __init__(self, args, run, debug_mode=False, splash_screen=None):
        """init method"""
        QObject.__init__(self)
        EventReceiver.__init__(self)

        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'mainwindow.ui'))
        self._widget.installEventFilter(self)
        self._widget.setGeometry(Config.appX, Config.appY, Config.appWidth,
                                 Config.appHeight)
        #style used in the qApp
        self._update_style() 
        self._widget.setEraseColor(QColor(Config.bgR, Config.bgG, Config.bgB))
        self.actions_menu = self._widget.child('Actions')
        self.window_menu = self._widget.child('Window')

        self._current_ma_designers = []

        # setup recently opened files
        for child in self._widget.children():
            if isinstance(child, QPopupMenu):
                for i in range(child.count()):
                    if child.text(child.idAt(i)) == "RecentFilesToken":
                        self._recent_files_menu = child
        connect(self._recent_files_menu, SIGNAL("aboutToShow()"), self._update_recent_files)
        connect(self._recent_files_menu, SIGNAL('activated(int)'), self._open_recent_file)

        if not splash_screen:
            splash_screen = DummySplash()

        self._progress_bar_widget = 0
        self.workspace = None
        
        eventServer = EventServer()

        # make sure events are handled by the EventReceiver interface
        eventServer.addEventHandler(self.post_event)

        self._node_modules = NodeRepository.DEFAULT_MODULES[:]
        userfunctionpath = Config.loadpath
        functionpath = Config.systemfunctiondir

        pluginRepository = PluginRepository()
        
        splash_screen.display_message("Loading available nodes")
        self.nodeRepository = NodeRepository(pluginRepository,
                                             self._node_modules,
                                             userfunctionpath, functionpath,
                                             event_server = eventServer)
         
        splash_screen.display_message("Setting up user interface")
        self._widget.statusBar()

        self._cache = CacheDirectory(Config.cachepath)
        self._cache.set_enabled(Config.cacheenabled)
        
        splash_screen.display_message("Initializing workspace")
        self._node_status_data_list = NodeStatusDataList(None)
        self.window_dict = {}
        self.checked_window = None

        splash_screen.display_message("Setting up initial workspace")
        self.workspace = Workspace(self, eventServer, self.nodeRepository, 
                                   self._cache)

        self._init_connections()

        if args:
            for filename in args:
                splash_screen.display_message("Loading module diagram %s" % filename)
                self.workspace.load_diagram(filename, True)
        else:
            splash_screen.display_message("Setting up empty diagram")
            self.workspace.new_diagram()

        dock_window = RightDockWindow(self._widget, self.nodeRepository,
                                      self.workspace,
                                      self.get_node_status_data_list(),
                                      debug_mode)
        self._widget.moveDockWindow(dock_window, Qt.DockRight)
        dock_window.show()
        self._node_tree = dock_window.get_node_tree()
        splash_screen.display_message("Starting SciCraft")

        # remove existing central widget before adding workspace
        self._widget.centralWidget().hide()
        self._widget.setCentralWidget(self.workspace)

        self._update_diagram_state()

        # C-c can kill SciCraft directly
        signal.signal(signal.SIGINT, self._quit_handler)
        # handle hangup
        if os.name == "posix":
            signal.signal(signal.SIGHUP, self._quit_handler)

        self._widget.statusBar().setFixedHeight(27)
        splash_screen.hide()
        if run:
            self.workspace.slot_run_diagram()

    def _quit_handler(self, sig, stack_frame):
        qApp.processEvents()
        qApp.quit()

    def create_new_ma_designer(self, filename=None):
        designer = MADesigner(None, filename)
        designer.show()
        connect(designer, PYSIGNAL("new_editor(filename)"), self.create_new_ma_designer)
        self._current_ma_designers.append(designer)

    def show(self):
        self._widget.show()
        
    def _init_connections(self):
        self.window_menu.setCheckable(1)

        self.pointer_button = self._widget.child('modePointerAction')
        self.link_button = self._widget.child('modeLinkAction')
        self.comment_button = self._widget.child('modeInsertCommentAction')
        self.play_action = self._widget.child('actionsRunAction')
        self.stop_action = self._widget.child('actionsStopAction')
        self.validate_action = self._widget.child("actionsValidateAction")
        self.close_action = self._widget.child("fileCloseDiagramAction")

        connect(self.pointer_button, SIGNAL("activated()"), self.slot_point_diagram)
        connect(self.link_button, SIGNAL("activated()"), self.slot_link_diagram)
        connect(self.comment_button, SIGNAL("activated()"), self.slot_comment_diagram)
        connect(self.play_action, SIGNAL("activated()"), self.workspace.slot_run_diagram)
        connect(self.stop_action, SIGNAL("activated()"), self.workspace.slot_stop_diagram)
        connect(self.validate_action, SIGNAL("activated()"), self.workspace.slot_validate_diagram)
        connect(self._widget.child('fileNewAction'), SIGNAL("activated()"), self.slot_file_new)
        connect(self._widget.child('fileOpenAction'), SIGNAL("activated()"), self.slot_file_open)        
        connect(self._widget.child('fileSaveAction'), SIGNAL("activated()"), self.workspace.save_diagram)
        connect(self._widget.child('fileSaveAsAction'), SIGNAL("activated()"), self.workspace.save_as_diagram)
        connect(self._widget.child('ZMLCreatorAction'), SIGNAL("activated()"), self.slot_creator)
        connect(self._widget.child('fileEPSCreatorAction'), SIGNAL("activated()"), self.workspace.save_eps_diagram)
        connect(self._widget.child('filePreferencesAction'), SIGNAL("activated()"), self.slot_edit_prefs)
        connect(self.close_action, SIGNAL("activated()"), self.workspace.close_active_window)

        connect(self._widget.child('windowMinimizeAllAction'), SIGNAL("activated()"), self._slot_minimize_all)
        connect(self._widget.child('windowMaximizeAllAction'), SIGNAL("activated()"), self.workspace.maximize_all_windows)
        connect(self._widget.child('windowCloseAllAction'), SIGNAL("activated()"), self.workspace.close_all_windows)

        connect(self._widget.child('help_whatsnew'),SIGNAL("activated()")  , self.slot_help_whatsnew)
        connect(self._widget.child('helpHelpAction'), SIGNAL("activated()"), self.slot_help_scicraft)
        connect(self._widget.child('helpAboutQtAction'), SIGNAL("activated()"), self.slot_help_qt)        
        connect(self._widget.child('helpAboutSciCraftAction'), SIGNAL("activated()"), self.slot_help_about)
        connect(self._widget.child('toolsMADesignerAction'), SIGNAL("activated()"), self.create_new_ma_designer)
        connect(self._widget.child('hideHiddenLinksAction'), SIGNAL("activated()"), self.workspace.hide_hidden_links)
        connect(self._widget.child('showHiddenLinksAction'), SIGNAL('activated()'), self.workspace.show_hidden_links)
        connect(self._widget.child('toolsR_Library_InstallerAction'), SIGNAL('activated()'),
                                   self._slot_start_r_library_installer)
        
        if PyCute:
            python_interaction = QAction("Python interaction",
                                         "&Python Interaction",
                                         QAccel.stringToKey("CTRL+I"),
                                         self._widget)
            python_interaction.addTo(self._widget.child('Tools'))
            connect(python_interaction, SIGNAL("activated()"), self._python_interaction)

    def _python_interaction(self):
        locals_ = {"app" : self, "workspace" : self.workspace,
                   "windows" : self.window_dict, "os" : os, "sys" : sys,
                   "qt" : qt}

        pycute = PyCute(locals_)
        pycute.show()

    def get_node_status_data_list(self):
        return self._node_status_data_list

    def get_node_tree(self):
        return self._node_tree

    def handle_event(self, event):
        """Called by EventReceiver interface to handle events in gui thread."""
        class_ = event.get_event_class()
        if class_ == "DiagramEvent":
            self._update_diagram_state()
        elif isinstance(event, InfoEvent):
            EventPopup(self._widget, event, modal=True)
        elif class_ == "ErrorEvent":
            EventPopup(self._widget, event)

    def _update_diagram_state(self):
        """Updates the state of the diagram actions.

        The diagram actions, which consist of "Run"-, "Stop"-, "Validate"-, and
        "Close" diagram, are updated according to the state of the current
        module diagram.
        """
        diagram = self.workspace.get_current_diagram()

        if len(self.window_dict):
            self.validate_action.setEnabled(True)
            self.close_action.setEnabled(True)
        else:
            # if no windows (diagrams) is open
            self.validate_action.setEnabled(False)
            self.close_action.setEnabled(False)
            self.play_action.setEnabled(False)
            self.stop_action.setEnabled(False)
            return
        
        if diagram and diagram.getIsRunning():
            if not diagram.get_has_crashed() and not diagram.getIsInterrupted():
                # running smoothly
                self.play_action.setEnabled(False)
                self.stop_action.setEnabled(True)
            elif diagram.getIsInterrupted():
                # diagram is interrupted but still running
                self.play_action.setEnabled(False)
                self.stop_action.setEnabled(False)
            elif diagram.get_has_crashed():
                self.play_action.setEnabled(True)
                self.stop_action.setEnabled(False)     
        else:
            # default fall-back values
            self.play_action.setEnabled(True)
            self.stop_action.setEnabled(False)

    def set_progress_bar(self, progress):

        if self._progress_bar_widget == progress:
            return

        if self._progress_bar_widget:
            self._progress_bar_widget.hide()
            self._widget.statusBar().removeWidget(self._progress_bar_widget)
            
        self._progress_bar_widget = progress
        self._progress_bar_widget.resize(205, 24)
        self._widget.statusBar().addWidget(self._progress_bar_widget, 205, 1)
        self._progress_bar_widget.show()

    def _update_style(self):
        # get_style() is a function in the Config module
        style = get_style() 
        # qApp is a qApplication "shortcut", derived from Qt        
        qApp.setStyle(style) 

    def slot_file_new(self):
        self.workspace.new_diagram()
        self._update_diagram_state()

    def slot_file_open(self):
        filename = QFileDialog.getOpenFileName(Config.last_path_modulediagram,
                                               "SciCraft diagrams (*.zmd);;All files (*)",
                                               self._widget, "Open Diagram",
                                               "Choose a file").latin1()
        if filename:
            filename = str(filename)
            Config.last_path_modulediagram = os.path.dirname(filename)
            if os.path.isfile(filename):                
                self.workspace.load_diagram(filename)
        
    def _reload_node_repository(self):
        userfunctionpath = Config.loadpath
        functionpath = Config.systemfunctiondir
        self.nodeRepository.updatePaths(self._node_modules, userfunctionpath,
                                        functionpath)
        self._node_tree.update_node_list(dict(self.nodeRepository.getNodeList()))
        
    def slot_edit_prefs(self):
        old_load_path = Config.loadpath
        # show widget
        prefs = EditPrefs(self._widget, self._cache)

        if prefs.exec_loop() == QDialog.Accepted:
            self.workspace.update_canvas()
            self._update_style()
            # has user changed loadpath?
            if Config.loadpath != old_load_path:
                self._reload_node_repository()
                
            self._cache.set_path(Config.cachepath)
            self._cache.set_enabled(Config.cacheenabled)

    def slot_creator(self):
        zmlcreator = ZMLCreator(self._widget)

    def _reset_toggle_tools(self, button_pressed):
        if self.pointer_button.isOn() and button_pressed != self.pointer_button:
            self.pointer_button.toggle()
        if self.link_button.isOn() and button_pressed != self.link_button:
            self.link_button.toggle()
        if self.comment_button.isOn() and button_pressed != self.comment_button:
            self.comment_button.toggle()

    def enable_pointer_tool(self):
        self._reset_toggle_tools(self.pointer_button)
        if not self.pointer_button.isOn():
            self.pointer_button.toggle()
    
    def slot_point_diagram(self):
        self.enable_pointer_tool()
        self.workspace.set_pointing_tool_as_current_tool()
        
    def enable_link_tool(self):
        self._reset_toggle_tools(self.link_button)
        if not self.link_button.isOn():
            self.link_button.toggle()
    
    def slot_link_diagram(self):
        self.enable_link_tool()
        self.workspace.set_linking_tool_as_current_tool()

    def enable_comment_tool(self):
        self._reset_toggle_tools(self.comment_button)
        if not self.comment_button.isOn():
            self.comment_button.toggle()

    def slot_comment_diagram(self):
        self.enable_comment_tool()
        self.workspace.set_comment_tool_as_current_tool()
            
    def slot_help_scicraft(self):
        HelpViewer(self._widget, os.path.join(Config.htmldocpath, "index.html"),
                   Config.htmldocpath)
    
    def slot_help_whatsnew(self):
        viewer = ChangelogViewer(self._widget)

    def slot_help_about(self):
        viewer = AboutViewer(self._widget)
        
    def slot_help_qt(self):
        QMessageBox.aboutQt(self._widget, "About Qt")

    def _slot_minimize_all(self):
        self.workspace.minimize_all_windows()
        for diagram in self.workspace.diagrams.values():
            diagram.showMinimized()

    def save_config(self):
        Config.appWidth = self._widget.width()
        Config.appHeight = self._widget.height()

        if self._widget.x() >= 0:
            Config.appX = self._widget.x()
        if self._widget.y() >= 20:
            Config.appY = self._widget.y()
        else:
            Config.appY = 20
            
        write_config()

    def add_window(self, windowId, name):
        # caller = self.workspace, not inited yet when this function
        # is called for the first time...
        id = self.window_menu.insertItem(name, self.slot_window_clicked) 
        self.window_menu.setItemParameter(id, int(windowId))
        if self.checked_window:
            self.window_menu.setItemChecked(self.checked_window, 0)
        self.window_menu.setItemChecked(id, 1)
        self.checked_window = id
        self.window_dict[windowId] = id
        self._update_diagram_state()
                
    def remove_window(self, windowId):
        self.window_menu.removeItem(self.window_dict[windowId])
        del self.window_dict[windowId]
        self._update_diagram_state()

    def slot_window_clicked(self, window_id):
        id = self.window_dict[str(window_id)]
        self.window_menu.setItemChecked(self.checked_window, 0)
        self.window_menu.setItemChecked(id, 1)
        self.checked_window = id
        self.workspace.window_change(window_id)
        self._update_diagram_state()

    def _update_recent_files(self):
        self._recent_files_menu.clear()
        existing_files = Config.recentFiles.split(os.pathsep)
        for filename in filter(None, existing_files):
            self._recent_files_menu.insertItem(filename)

    def _open_recent_file(self, item_id):
        filename = str(self._recent_files_menu.text(item_id))
        if os.path.isfile(filename):                
            self.workspace.load_diagram(filename)
        
    def eventFilter(self, object_, event):
        """Filter that checks for close events.

        The filter will return True if the event should be
        stopped and False otherwise.
        """
        if isinstance(event, QCloseEvent):
            return not self.workspace.close_all_windows()
        return False

    def _slot_start_r_library_installer(self):

        self._installer = RLibInstaller()
        self._installer.show()
