# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from Library.Util.Config import Config
from GUI.links import LinkControl
from GUI.contextmenus import LinkContext

from qt import QColor, Qt, QPen, QPointArray, QBrush
from qtcanvas import QCanvasLine, QCanvasPolygon

import math

def find_sircle_intersect(start, end, shape):
    start_x, start_y = start
    end_x, end_y = end
    width, height = shape

    sircle_r = math.sqrt(math.pow(width, 2) + math.pow(height,2))/2.
    
    length_x = abs(start_x - end_x)
    length_y = abs(start_y - end_y)
    try:
        r = math.sqrt(math.pow(length_x, 2) + math.pow(length_y, 2))    
        angle = math.asin(length_y/r)
    except ZeroDivisionError:
        return end_x, end_y

    if end_x < start_x:
        start_x *= -1
    if end_y < start_y:
        start_y *= -1
    
    sx = abs(math.cos(angle)*sircle_r + start_x)
    sy = abs(math.sin(angle)*sircle_r + start_y)
    
    return sx, sy

def find_arrow_points(start, end, sx, sy):

    dx = start[0] - end[0]
    dy = start[1] - end[1]
    length = math.sqrt(pow(dx, 2) + pow(dy, 2))
    if not length:
        return 
    
    direction = math.asin(dy/length)
    if dx < 0:
        direction = 3.14 - direction
        
    arrow_length = 15

    point1 = (sx - 3*math.cos(direction),
              sy - 3*math.sin(direction))
    point2 = (sx + arrow_length*math.cos(direction + 0.3),
              sy + arrow_length*math.sin(direction + 0.3))
    point3 = (sx + arrow_length*math.cos(direction - 0.3),
              sy + arrow_length*math.sin(direction - 0.3))
    
    return point1, point2, point3

class Connection:

    def __init__(self, mdview, start_node, end_node, start_point, end_point,
                 width=1, height=1):
        self._mv = mdview

        self._shape = width, height        

        self._start_node = start_node
        self._end_node = end_node

        self._start_point = start_point
        self._end_point = end_point

        self._arrow = None

        self._line = QCanvasLine(self._mv.canvas())

        # return self when asked for the parent
        self._line.getParent = lambda : self

        self._is_selected = False
        self._is_hidden = False

        self._draw()
        self._line.show()
        self._mv.canvas().update()
        self._auto_connect()

    def _draw(self):
        pen = QPen(QColor(Config.lineR, Config.lineG, Config.lineB), Config.lineWidth)
        self._line.setPen(pen)
        
        sx, sy = find_sircle_intersect(self._end_point, self._start_point, self._shape)
        self._line.setPoints(*self._start_point + (sx, sy))
        
        # draw arrow stuff
        if self._arrow:
            self._arrow.setCanvas(None)

        arrow_points = find_arrow_points(self._start_point, self._end_point,
                                         sx, sy)
        if not arrow_points:
            return

        self._arrow = QCanvasPolygon(self._line.canvas())
        point_array = QPointArray(3)
        for i, p in enumerate(arrow_points):
            x, y = p
            point_array.setPoint(i, x, y)
        self._arrow.setPoints(point_array)
        color = QColor(Config.lineR, Config.lineG, Config.lineB)
        self._arrow.setBrush(QBrush(color))
        self._arrow.show()
        
    def move_start_point(self, dx, dy):
        self._start_point = self._start_point[0] + dx, self._start_point[1] + dy
        self._draw()

    def move_end_point(self, dx, dy):
        self._end_point = self._end_point[0] + dx, self._end_point[1] + dy
        self._draw()

    def get_start_point(self):
        return self._start_point

    def get_end_point(self):
        return self._end_point

    def get_start_node(self):
        return self._start_node

    def get_end_node(self):
        return self._end_node

    def get_canvas_line(self):
        return self._line

    def set_highlight(self, value=True):
        if value: color = Qt.red
        else: color = QColor(Config.lineR,
                             Config.lineB,
                             Config.lineG)

        pen = QPen(color, Config.lineWidth)
        self._line.setPen(pen)
        self._arrow.setBrush(QBrush(color))
        self._is_selected = value
        
    def get_nodes(self):
        return self._start_node, self._end_node

    def linkDialog(self):
        link_control = LinkControl(None, self._start_node, self._end_node,
                                   0, 0)
        link_control.show()

    def delete(self):
        if self._line.canvas():
            canvas = self._line.canvas()
            self._line.setCanvas(None)
            if self._arrow:
                self._arrow.setCanvas(None)
            canvas.update()

        self._end_node.removeConnectionsFromNode(self._start_node)
        self._mv.remove_line(self)

    def is_highlighted(self):
        return self._is_selected

    def is_hidden(self):
        return self._is_hidden

    def set_hidden(self, value):
        if value:
            self._line.setCanvas(None)
            self._arrow.setCanvas(None)
        else:
            self._line.setCanvas(self._mv.canvas())
            self._arrow.setCanvas(self._mv.canvas())

        self._mv.canvas().update()        
        
        self._is_hidden = value

    def show_link(self):
        if self._is_hidden:
            self._line.setCanvas(self._mv.canvas())
            self._arrow.setCanvas(self._mv.canvas())
            self._mv.canvas().update()

    def hidden_link(self):
        if self._is_hidden:
            self._line.setCanvas(None)
            self._arrow.setCanvas(None)
            
        self._mv.canvas().update()  
            
    def context_menu(self, point):
        menu = LinkContext(self, self._mv)
        menu.popup(point)
        
    def _auto_connect(self):
        """method that will try to auto-connect ports"""

        output_ports = self._start_node.getOutputPorts()
        input_ports = self._end_node.getInputPorts()
        multi_input_ports = self._end_node.get_multi_input_ports()

        if (len(input_ports) == 1) and (len(output_ports) == 1) and\
               not multi_input_ports:
            input_ports[0].connectToDataSource(output_ports[0])
        elif len(multi_input_ports) == 1 and not input_ports:
            # need to check if one or more port is connected
            multi_port = multi_input_ports[:].pop()            
            emulated_ports = multi_port.get_emulated_input_ports()
            connected = False
            for port in [port.getConnectedInputPorts() for port in output_ports]:                
                if port and filter(lambda x: x in emulated_ports, port):
                    connected = True
                    break
            if not connected:
                for output_port in output_ports:
                    multi_port.add_data_source(output_port)   

    def update(self):
        self._draw()
        
