# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QPopupMenu, QAction, QMessageBox, QPoint, SIGNAL
from qtui import QWidgetFactory
from System.Data.ports import NoDataToCollectError, MissingConnectionError
from Library.Util.Config import Config
from GUI.spreadsheet import SpreadSheetWindow

import os

def calculate_point(point, canvas_view):
    point += QPoint(1, 1)
    point = canvas_view.contentsToViewport(point)
    point = canvas_view.mapToGlobal(point)
    point = canvas_view.inverseWorldMatrix().map(point)

    return point


class LinkContext(QPopupMenu):
    
    def __init__(self, link, canvas_view):

        QPopupMenu.__init__(self, canvas_view)        
        self.link = link
        self._canvas_view = canvas_view

        self.id_data = {} # used to store the data with the popupmenu id as key
        
        self.start_node = link.get_start_node()
        self.end_node = link.get_end_node()
        
        self.insert_data_submenues()

        remove_link_action = QAction("&Remove link", 0, self)
        edit_link_action = QAction("&Edit link", 0, self)
        hide_link_action = QAction("&Hide link", 0, self)
        hide_link_action.setToggleAction(True)
        
        edit_link_action.addTo(self)
        remove_link_action.addTo(self)
        hide_link_action.addTo(self)
        if self.link.is_hidden():
            hide_link_action.toggle()
        
        self.connect(edit_link_action, SIGNAL("activated()"), self.slot_edit_link)
        self.connect(remove_link_action, SIGNAL("activated()"), self.slot_remove_link)
        self.connect(hide_link_action, SIGNAL('toggled(bool)'), self._slot_hide_toggled)
        
        self.sizeHint()

    def popup(self, point):
        point = calculate_point(point, self._canvas_view)
        QPopupMenu.popup(self, point)


    def insert_data_submenues(self):
        top = QPopupMenu(self)
        view_id = self.insertItem("&View data", top)        
        data = self.get_port_data()

        if not data:
            self.setItemEnabled(view_id, False)
            return
        for element in data:
            info = element.get_data_info()
            self.id_data[top.insertItem(info, self.slot_view_data)] = \
                                                   element

    def get_port_data(self):
        data = []
        for port in self.start_node.getOutputPorts(): #output ports
            for input_ in port.getConnectedInputPorts(): #input ports
                if input_ in self.end_node.get_all_input_ports():
                    if port.getData():
                        data.append(port.getData())
        return data

    
    def slot_view_data(self, id_):
        self._sheet_window = SpreadSheetWindow()        
        self._sheet_window.set_data([self.id_data[id_]])
        self._sheet_window.show()

    def slot_edit_link(self):        
        self.link.linkDialog()            
            
    def slot_remove_link(self):
        self.link.delete()
        self._canvas_view.reset_selected_link()

    def get_data_info_string(self):
        outputports = self.start_node.getOutputPorts()
        infostring = "" 
        for output in outputports:
            for port in output.getConnectedInputPorts():
                if port in self.end_node.get_all_input_ports():
                    if output.datainfo:
                        infostring = infostring + output.datainfo

        return infostring
    
    def _slot_hide_toggled(self, bool_):
        self.link.set_hidden(bool_)

class NodeContext(QPopupMenu):

    def __init__(self, guinode, point, canvas_view):
        self.canvas_view = canvas_view
        QPopupMenu.__init__(self, self.canvas_view)
        self.guinode = guinode
        self._port_data_by_id = {} # Both input and output data.
        self.point = point
        create_link_action = QAction("&Create link", 0, self)
        self.connect(create_link_action, SIGNAL("activated()"), self.slot_create_link)
        edit_node_action = QAction("&View node", 0, self)
        self.connect(edit_node_action, SIGNAL("activated()"), self.slot_edit_node)
        remove_node_action = QAction("&Remove node", 0, self)
        self.connect(remove_node_action, SIGNAL("activated()"), self.slot_remove_node)
        rename_action = QAction("Re&name node", 0, self)
        self.connect(rename_action, SIGNAL("activated()"), self._slot_rename_node)
    
        keep_output_action = QAction("&Keep output", 0, canvas_view)
        keep_output_action.setToggleAction(True)
        keep_output_action.setOn(self.guinode.getDiagramNode().get_keep_output())
        self.connect(keep_output_action, SIGNAL("toggled(bool)"), self.slot_toggle_keep_output)
        ports_menu = QPopupMenu(self)
        self._build_ports_menu(ports_menu)
        
        collect_and_run_action = QAction("R&un node", 0, canvas_view)
        self.connect(collect_and_run_action, SIGNAL("activated()"), self.slot_collect_and_run)

        self.show_output_action = None

        create_link_action.addTo(self)
        edit_node_action.addTo(self)
        rename_action.addTo(self)
        remove_node_action.addTo(self)
        keep_output_action.addTo(self)
        collect_and_run_action.addTo(self)
        self.insertSeparator()
        self.insertItem("&Inspect data", ports_menu)
        
        if self.guinode.getDiagramNode().getNodeType() == 'FunctionNode':
            self.show_output_action = QAction("&Show function output", 0, self)
            self.show_output_action.setEnabled(False)
            self.connect(self.show_output_action, SIGNAL("activated()"),
                         self.slot_show_output)
            self.show_output_action.addTo(self)
        
        self.sizeHint()
    
    def popup(self):
        point = calculate_point(self.point, self.canvas_view)
        QPopupMenu.popup(self, point)

    def slot_create_link(self):
        self.canvas_view.set_previous_tool(self.canvas_view.get_current_tool())
        self.canvas_view.set_linking_tool_as_current_tool()
        self.canvas_view.get_current_tool().set_start_of_link(self.guinode, self.point)
        self.canvas_view.get_current_tool().set_link_created_from_context_menu()

    def slot_edit_node(self):
        self.guinode.nodeDialog()

    def slot_remove_node(self):
        self.guinode.delete()
        self.canvas_view.reset_selected_items()

    def slot_toggle_keep_output(self, value):
        self.guinode.getDiagramNode().set_keep_output(value)

    def showEvent(self, event):
        QPopupMenu.showEvent(self, event)
        if self.show_output_action:
            if self.guinode.getDiagramNode().get_formatted_output():
                self.show_output_action.setEnabled(True)
            else:
                self.show_output_action.setEnabled(False)

    def slot_show_output(self):
        outputdialog = QWidgetFactory.create(os.path.join(Config.scicrafthome, "ui",
                                                          "outputdialog.ui"),
                                             self.canvas_view,
                                             self.canvas_view)
        output_textedit = outputdialog.child("output")

        # this slot is never called, unless self.guinode.getDiagramNode() is a
        # FunctionNode
        output_textedit.setText(
            self.guinode.getDiagramNode().get_formatted_output())
        outputdialog.show()

    def slot_collect_and_run(self):
        diagram_node = self.guinode.getDiagramNode()

        try:
            diagram_node.collect_data()
            diagram_node.get_module_diagram().start(diagram_node)
            
        except NoDataToCollectError, e:

            # find which providing nodes do not have keep_output set
            unset_nodes = filter(lambda node : not node.get_keep_output(),
                                 diagram_node.get_providing_nodes())
            if unset_nodes:
                reply = QMessageBox.question(self, "Could not collect data", """
Running a single node requires that all proceding nodes
store output values and then be run atleast once.

Do you wish to enable "Keep output" on all preceding nodes?""",
                                             QMessageBox.Yes, QMessageBox.No |
                                             QMessageBox.Default |
                                             QMessageBox.Escape)
                if reply == QMessageBox.Yes:
                    diagram_node.set_keep_output_in_providing_nodes()
            else:
                QMessageBox.warning(self, "Could not collect data", """
Data was unavailable from one or more of the preceding nodes.

This might be due to setting "Keep output", but not running the
module diagram or node afterwards.

Please try executing the module diagram to make data available.""",
                                    QMessageBox.Ok)
        except MissingConnectionError, e:
            QMessageBox.warning(self, "Could not collect data", str(e))
            
    def _build_ports_menu(self, ports_menu):
        """Builds the Ports popup menu."""
        input_port_menu = QPopupMenu(self)
        output_port_menu = QPopupMenu(self)
        id_in = ports_menu.insertItem("&Input ports", input_port_menu) 
        id_out = ports_menu.insertItem("&Output ports", output_port_menu)
        enable_inspect_data_menu = False
        node = self.guinode.getDiagramNode()
        
        # Fetch data from normal input ports.
        for input_port in node.getInputPorts():
            text = input_port.getName() + " - "
            enable = True
            if input_port.getData() is None:
                text += "no data"
                enable = False
            else:
                text += input_port.getData().get_data_size_str()
            id = input_port_menu.insertItem(text, self.slot_view_data)
            self._port_data_by_id[id] = input_port.getData()
            input_port_menu.setItemEnabled(id, enable)
            enable_inspect_data_menu |= enable
        
        # Fetch data from multi-input ports
        for multi_port in node.get_multi_input_ports():
            for emulated_port in multi_port.get_emulated_input_ports():
                text = emulated_port.getName() + " - "
                enable = True
                if emulated_port.getData() is None:
                    text += "no data"
                    enable = False
                else:
                    text += emulated_port.getData().get_data_size_str()
                id = input_port_menu.insertItem(text, self.slot_view_data)
                self._port_data_by_id[id] = emulated_port.getData()
                input_port_menu.setItemEnabled(id, enable)
                enable_inspect_data_menu |= enable
        ports_menu.setItemEnabled(id_in, input_port_menu.count() != 0)
        # Fetch data from output ports.
        for output_port in node.getOutputPorts():
            enable = True
            if output_port.getData() is None:
                text = output_port.getName() + " - no data"
                enable = False
            else:
                text = output_port.get_data_info()
            id = output_port_menu.insertItem(text, self.slot_view_data)
            self._port_data_by_id[id] = output_port.getData()
            output_port_menu.setItemEnabled(id, enable)
            enable_inspect_data_menu |= enable
        ports_menu.setItemEnabled(id_out, output_port_menu.count() != 0)
        # Disable 'Ports' popup if there are no ports.
        ports_menu.setEnabled(enable_inspect_data_menu)

    def slot_view_data(self, id):
        self._sheet_window = SpreadSheetWindow()
        self._sheet_window.set_data([self._port_data_by_id[id]])
        self._sheet_window.show()

    def _slot_rename_node(self):
        self.guinode.rename()
