# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QPoint, QListViewItem, QObject, SIGNAL, QFileDialog
from qt import QMessageBox, QRect
from qtui import QWidgetFactory
import math
import sre
import types
import os
import copy

from GUI.Canvas.CanvasNode import CanvasNode
from GUI.Canvas.CanvasLabel import CanvasLabel
from GUI.NodeConfig import NodeConfig
from GUI.contextmenus import NodeContext
from GUI.viewers import HelpViewer
from Library.Util.Config import Config
from Library.Util.gui import ValueListViewItem
from System.Data.functionnode import FunctionNode
from System.Data.Parameter import IllegalValueError
connect = QObject.connect
disconnect = QObject.disconnect


class GUINode:

    def __init__(self, canvasview, node):
        
        self.node = node
        
        self._cv = canvasview
        self._canvasnode = None

        self._canvas = self._cv.canvas()
        
        self._position = self.node.getPosition()

        self._selected = False
        self._dialog = None

        self._startLinks = [] # connections that starts in this node
        self._endLinks = [] # connections that ends in this node

        node_config = NodeConfig(node.getNodeRepName())
        
        # if functionnode, use custom icon (if present)
        self.image = None
        if isinstance(node, FunctionNode):
            self.image = node.get_icon()
        if self.image == None:
            self.image = (node_config.getImage())

        self._inpoints = node_config.get_inpoints()
        self._outpoints = node_config.get_outpoints()
        self._canvasnode = CanvasNode(self, self._canvas, self.image, self._position)
        self._canvasnode.show()

        if Config.showLabel:            
            self.label = CanvasLabel(self, self._canvas, self.node.getName())
        else:
            self.label = ""

    def getStartLinks(self):
        return self._startLinks


    def getEndLinks(self):
        return self._endLinks

    def getCanvasNode(self):
        return self._canvasnode

    def getDiagramNode(self):
        return self.node

    def get_closest_entry_line(self, point):
        if not self._inpoints:
            return None        
        return self._get_closest_connection_point_in_points(point, self._inpoints)[0][0]

    def get_closest_exit_line(self, point):
        if not self._outpoints:
            return None
        return self._get_closest_connection_point_in_points(point, self._outpoints)[0][0]    

    def get_closest_entry_line_with_index(self, point):
        """Getting the closest entry point (node input position) to a mouse position."""
        
        # Checking instance type of point, since index of position,
        # is used when loading zmd diagrams.
        if not self._inpoints:
            return (None,0)
        
        if isinstance(point, types.IntType):
            p_len = len(self._inpoints) -1
            if point <= p_len:
                return [self._inpoints[point], point]
            elif point > p_len:
                return [self._inpoints[p_len], p_len]
        else: # should be a QPoint
            return self._get_closest_connection_point_in_points(point, self._inpoints)

    def get_closest_exit_line_with_index(self, point):
        """Getting the closest exit point (node output position) to a mouse position."""
        # Checking instance type of point, since index of position,
        # is used when loading zmd diagrams.
        if not self._outpoints:
            return (None,0)
        
        if isinstance(point, types.IntType):
            p_len = len(self._outpoints)-1
            if point <= p_len:
                return [self._outpoints[point], point]
            elif point > p_len:
                return [self._outpoints[p_len], p_len]
        else: # should be a QPoint
            return self._get_closest_connection_point_in_points(point,self._outpoints)
    
    def _get_closest_connection_point_in_points(self, point, points):
        """Find closest connection point in a set of points."""

        if len(points) == 1:
            return [points[0], 0]

        closestPoint = points[0]
        mousePos = point

        realPoint = closestPoint[0][0] + self.x(), closestPoint[0][1] + self.y()
        xDist = abs(realPoint[0] - mousePos[0])
        yDist = abs(realPoint[1] - mousePos[1])
        bestDist = math.sqrt(pow(xDist, 2) + pow(yDist, 2))

        for point in points[1:]:
            realPoint = point[0][0] + self.x(), point[0][1] + self.y()
            xDist = abs(realPoint[0] - mousePos[0])
            yDist = abs(realPoint[1] - mousePos[1])
            dist = math.sqrt(pow(xDist, 2) + pow(yDist, 2))

            if dist < bestDist:
                closestPoint = point
                bestDist = dist

        return [closestPoint, points.index(closestPoint)]

    def nodeDialog(self):
        if not self._dialog:
            self._dialog = GUINodeDialog(self._cv, self.getDiagramNode())
            self._dialog.show()
        elif not self._dialog.is_shown():
            self._dialog.show()

    def saveGuiSettings(self):
        """Save this node's GUI settings."""
        pass

    def addStartLink(self, start):
        self._startLinks.append(start)

    def addEndLink(self, end):
        self._endLinks.append(end)

    def x(self):
        """
        Returnes the x-coord. for the canvasnode
        """
        
        return self._canvasnode.x()    

    def y(self):
        """
        Returnes the x-coord. for the canvasnode
        """
        return self._canvasnode.y()

    def position(self):
        return QPoint(self.x(), self.y())
    
    def z(self):
        return self._canvasnode.z()

    def setZ(self, z):
        self._canvasnode.setZ(z)

    def height(self):
        return self._canvasnode.height()

    def width(self):
        return self._canvasnode.width()

    def boundingRect(self):
        """
        Returns a bounding rectangle of a node included its label.

        Returned value is tuple (xmin, ymin, xmax, ymax)
        """
        nodeWidth = self.width()
        nodeHeight = self.height()
        nodeX = self.x()
        nodeY = self.y()
        labelWidth = 0
        labelHeight = 0
        labelX = nodeX
        labelY = nodeY
        if self.label:
            labelX = self.label.x()
            labelY = self.label.y()
            labelWidth = self.label.width()
            labelHeight = self.label.height()

        xmin = min(nodeX, labelX)
        ymin = min(nodeY, labelY)
        xmax = max(nodeX+nodeWidth-1, labelX+labelWidth-1)
        ymax = max(nodeY+nodeHeight-1, labelY+labelHeight-1)

        return QRect(QPoint(xmin, ymin), QPoint(xmax, ymax))

    
    def move(self, dx, dy):
        newX = self.x() + dx
        newY = self.y() + dy
            
        self._canvasnode.move(dx, dy)
        self.node.setPosition(newX, newY)
        self._moveLinks(dx, dy)
        
    def setPosition(self, xPos, yPos):
            
        dx = xPos - self.x()
        dy = yPos - self.y()
        self.move(dx, dy)    
    
    def testSnapToGrid(self, x, y):
        granularity = Config.gridGranularity

        testX1 = x - (x%granularity)
        testY1 = y - (y%granularity)
        testX2 = x + (granularity - x%granularity)
        testY2 = y + (granularity - y%granularity)

        if abs(x - testX1) < granularity/4:
            x = testX1
        elif abs(x - testX2) < granularity/4:
            x = testX2

        if abs(y - testY1) < granularity/4:
            y = testY1
        elif abs(y - testY2) < granularity/4:
            y = testY2                
        
        return [x, y]        
    
    def _moveLinks(self, dx, dy):
        for link in self._startLinks:
            link.move_start_point(dx, dy)
        for link in self._endLinks:
            link.move_end_point(dx, dy)

    def delete(self):
        self._canvasnode.delete()
        self._cv.remove_node(self)
        if self.label:
            self.deleteLabel()            

        # Make sure a copy is traversed as each line modifies the list
        # as it is deleted
        for line in list(self._startLinks) + list(self._endLinks):
            line.delete()

        self._canvas.update()

    def remove_link(self, removable_link):
        for link in self._startLinks:
            if link == removable_link:
                self._startLinks.remove(link)

        for link in self._endLinks:
            if link == removable_link:
                self._endLinks.remove(link)        

    def hideLinks(self):
        
        for link in self._startLinks:
            link.hide()
        for link in self._endLinks:
            link.hide()
    
    def redrawLinks(self):
        for link in self._startLinks:
            link.redraw()
        for link in self._endLinks:
            link.redraw()
    
    def showLabel(self):
          self.label = CanvasLabel(self, self._canvas, self.node.getName())

    def moveLabel(self, dx, dy):
        if self.label:
            self.label.move(dx, dy)

    def deleteLabel(self):
        if self.label:
            self.label.delete()
            self.label = ""

    def set_highlight(self, value = True):
        self.label.set_hightlight(value)
        self._selected = value

    def is_highlighted(self):
        return self._selected

    def labelZ(self, z):
        if self.label:
            self.label.newZ(z)
    def newZ(self, z):
        self._canvasnode.newZ(z)

    def context_menu(self, point):
        context = NodeContext(self, point, self._cv)
        context.popup()

    def rename(self):
        """Interactively rename the label of the node."""
        self.label.interactive_rename()


class GUINodeDialog:

    STACK_FILE, STACK_COMBO, STACK_MAX_MIN, STACK_LINE, STACK_EMPTY = range(5)
    

    def __init__(self, parent, node):

        self.node = node
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir, 'nodedialog.ui'),
                                             None, parent)
        
        self._widget.child('description_text').setText(node.getDescription())
        self._widget.setCaption(node.getName())

        connect(self._widget.child('parameters_list'), SIGNAL('selectionChanged(QListViewItem *)'),
                self._slot_parameter_list_changed)                    
        connect(self._widget.child('file_button'), SIGNAL('clicked()'),
                self._slot_file_button_clicked)
        connect(self._widget.child('stack_line_edit'), SIGNAL('textChanged(const QString &)'),
                self._slot_line_edit_changed)
        connect(self._widget.child('stack_combo_box'), SIGNAL('activated(int)'),
                self._slot_combo_box_changed)
        connect(self._widget.child('stack_step_widget'), SIGNAL('valueChanged(int)'),
                self._slot_spin_changed_value)
        connect(self._widget.child('ok_button'), SIGNAL('clicked()'),
                self._slot_ok_pressed)
        connect(self._widget.child('stack_file_line'), SIGNAL('textChanged(const QString &)'),
                self._slot_file_line_changed)
        connect(self._widget.child('help_button'), SIGNAL('clicked()'),
                self._slot_help_pressed)
                                                              
        self._widget.child('stack_widget').raiseWidget(GUINodeDialog.STACK_EMPTY)
        self._copy_parameters = []        
        
        self._update_parameters_list()
        self._update_ports_list()

    def is_shown(self):
        return self._widget.isShown()

    def _update_parameters_list(self):
        param_names = [param.getName() for param in self.node.getParameters()]
        list_view = self._widget.child('parameters_list')

        for parameter in self.node.getParameters():
            param = copy.deepcopy(parameter)
            item = ParameterSortedListViewItem(param_names,
                                               list_view,
                                               param)
            self._copy_parameters.append(param)


    def _update_ports_list(self):

        list_view = self._widget.child('ports_list')
        list_view.clear()

        input_ports = self.node.getInputPorts()
        output_ports = self.node.getOutputPorts()

        if input_ports:
            input_ = QListViewItem(list_view, 'Input Ports')
            input_.setOpen(True)
            
            for port in input_ports:
                item = QListViewItem(input_, port.getName(), port.getDescription())

        if output_ports:
            output = QListViewItem(list_view, 'Output Ports')
            output.setOpen(True)
            
            for port in output_ports:
                item = QListViewItem(output, port.getName(), port.getDescription())

    def _slot_parameter_list_changed(self, item):
        param = item.get_value()
        type_ = param.getType()
        name = param.getName()
        description = param.getDescription()
        restrictions = param.getRestrictions()
        stack = self._widget.child('stack_widget')
        self._widget.child('param_decription').setText(str(description))
        self._widget.child('param_box').setTitle('Parameter: %s' % name)

        if restrictions['values']:
            stack.raiseWidget(GUINodeDialog.STACK_COMBO)
            combo = self._widget.child('stack_combo_box')
            combo.clear()
            
            for value in restrictions['values']:
                combo.insertItem(str(value))
            
            if param.getValue():
                combo.setCurrentText(str(param.getValue()))

        elif type_ != "text" and (restrictions['min'] or restrictions['max']):
            stack.raiseWidget(GUINodeDialog.STACK_MAX_MIN)
            step = self._widget.child('stack_step_widget')
            if restrictions['min']:
                step.setMinValue(restrictions['min'])
            if restrictions['max']:
                step.setMaxValue(restrictions['max'])
            if restrictions['step']:
                step.setLineStep(restrictions['step'])                  

            step.setValue(param.getValue())
                
        elif type_ in ('fileload', 'filesave', 'directory'):
            stack.raiseWidget(GUINodeDialog.STACK_FILE)
            self._widget.child('stack_file_line').setText(param.getValue())
        else:
            stack.raiseWidget(GUINodeDialog.STACK_LINE)
            self._widget.child('stack_line_edit').setText(str(param.getValue()))
            
    def _slot_line_edit_changed(self, qttext):
        text = str(qttext)
        try:
            self._set_parameter_value(text)
        except IllegalValueError, e:
            param = self._widget.child('parameters_list').currentItem().get_value()
            self._widget.child('stack_line_edit').setText(str(param.getValue()))
            if self._widget.parent():
                # Don't want this to be shown when unittesting
                QMessageBox.warning(self. _widget, 'Incorrect value', str(e))


    def _slot_combo_box_changed(self, index):

        value = str(self._widget.child('stack_combo_box').currentText())
        self._set_parameter_value(value)
        
    def _slot_spin_changed_value(self, value):
        self._set_parameter_value(value)

    def _slot_file_button_clicked(self):

        param = self._widget.child('parameters_list').currentItem().get_value()
        if param.getType() == 'filesave':
            name = QFileDialog.getSaveFileName().latin1()
        elif param.getType() == 'fileload':
            name = QFileDialog.getOpenFileName().latin1()
        elif param.getType() == 'directory':
            name = QFileDialog.getExistingDirectory().latin1()

        if name:
            self._widget.child('stack_file_line').setText(name)

    def _set_parameter_value(self, value):        

        item = self._widget.child('parameters_list').currentItem()
        param = item.get_value()        
        param.setValue(value)

        # We want to do this only if setting of the value succeeded 
        item.setText(2, str(value))
            
    def show(self):
        self._widget.show()
        
    def _slot_ok_pressed(self):
        
        for i, param in enumerate(self.node.getParameters()):
            param.setValue(self._copy_parameters[i].getValue())

        self._widget.close()

    def _slot_help_pressed(self):
        HelpViewer(self._widget, os.path.join(Config.htmldocpath,
                                              "node_nodes.html"),
                   Config.htmldocpath)
        
    def _slot_file_line_changed(self, text):
        item = self._widget.child('parameters_list').currentItem()
        item.get_value().setValue(str(text))        
        item.setText(2, str(text))


class ParameterSortedListViewItem(ValueListViewItem):

    def __init__(self, parameters, list_view, param):
        
        ValueListViewItem.__init__(self,
                                   list_view,
                                   param,
                                   param.getName(),
                                   param.getType(),
                                   str(param.getValue()))
        self._parameters = parameters
        self._index = parameters.index(self.get_value().getName())

    def compare(self, i, col, bool):
        i_index = self._parameters.index(i.get_value().getName())

        if i_index > self._index:
            return -1

        return 1
        
