# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qt import QColor, QObject, QPainter, QPen, QPoint, QSizePolicy, QWidget, Qt
from qt import SIGNAL
from qtui import QWidgetFactory

import os

from GUI.viewers import HelpViewer
from GUI.Canvas.CanvasLine import CanvasLine
from GUI.Canvas.CanvasPolygon import CanvasPolygon
from GUI.contextmenus import LinkContext
from System.Data.ports import MultipleInputPort
from Library.Util.Config import Config

class LinkCreationError(Exception):
    pass


def create_connection((x1, y1), (x2, y2)):
    """Create list of points creating connection between x1,y1 and x2,y2."""
    if y1 == y2 or x1 == x2:
        # direct horizontal or vertical line
        return (x1, y1), (x2, y2)
    elif x2 > x1:
        # horizontal line going to right with break
        middle_x = (x1 + x2) / 2
        return (x1, y1), (middle_x, y1), (middle_x, y2), (x2, y2)
    else:
        # vertical line breaking left
        middle_y = (y1 + y2) / 2
        return (x1, y1), (x1, middle_y), (x2, middle_y), (x2, y2)

class LinkControl:
    """Dialog for editing a link in a module diagram."""
    
    def __init__(self, parent, startNode, endNode, startIndex, endIndex):
        self._connections = {}

        self._start_node = startNode
        self._end_node = endNode
        
        self._start_index = startIndex
        self._end_index = endIndex

        self._output_ports = []
        self._input_ports = []

        self._connections = {}
        self._new_connections = {}
        self._deleted_connections = {}
        
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir, "link_control.ui"), None, parent)
        
        self._output_box = self._widget.child('output_box')
        self._input_box = self._widget.child('input_box')

        frame = self._widget.child('line_drawer_frame')
        self._line_draw_widget = LineDrawWidget(frame, self._draw_links, self._widget)

        QObject.connect(self._output_box, SIGNAL("highlighted(int)"),
                     self._slot_clicked_output)
        QObject.connect(self._input_box, SIGNAL("highlighted(int)"),
                     self._slot_clicked_input)

        connect_button = self._widget.child('connect_button')
        disconnect_button = self._widget.child('disconnect_button')
        QObject.connect(connect_button, SIGNAL("clicked()"), self._slot_connect)
        QObject.connect(disconnect_button, SIGNAL("clicked()"),
                     self._slot_disconnect)

        self._name_edit = self._widget.child('name_edit')
        self._type_edit = self._widget.child('type_edit')
        self._description_edit = self._widget.child('description_edit')

        help_button = self._widget.child('help_button')
        ok_button = self._widget.child('ok_button')
        cancel_button = self._widget.child('cancel_button')

        QObject.connect(help_button, SIGNAL("clicked()"), self._slot_help)
        QObject.connect(ok_button, SIGNAL("clicked()"), self._slot_ok)
        QObject.connect(cancel_button, SIGNAL("clicked()"), self._widget.close)

        for port in startNode.getOutputPorts():
            # index in outputPorts equals index in outputBox
            self._output_ports.append(port)
            self._output_box.insertItem(str(port.getName()))

        for port in self._end_node.getInputPorts():
            # index in inputPorts equals index in inputBox
            self._input_ports.append(port)
            self._input_box.insertItem(str(port.getName()))

        for port in self._end_node.get_multi_input_ports():
            # index in inputPorts equals index in inputBox
            self._input_ports.append(port)
            self._input_box.insertItem(port.get_name())

            # for multiinputports we assign connections right away
            for output_port in port.get_data_sources():
                if output_port in self._output_ports:
                    inputIndex = self._input_ports.index(port)
                    outputIndex = self._output_ports.index(output_port)
                    self._connections[(outputIndex,inputIndex)] = (output_port, port)
            
        # generate a dict with connections
        # - using index-ids (from the QListBox-lists) as keys
        # - the element is a list [changed, connected/disconnected, outputID, inputID]
        for connection in self._start_node.get_connections_to_node(self._end_node):
            outputIndex = self._output_ports.index(connection[0])
            inputIndex = self._input_ports.index(connection[1])
            self._connections[(outputIndex,inputIndex)] = connection

    def show(self):
        self._widget.show()

    def _draw_links(self, painter):
        for key in self._connections:
            output, input = self._connections[key]

            outputIndex = self._output_ports.index(output)
            inputIndex = self._input_ports.index(input)

            outputItem = self._output_box.item(outputIndex)
            inputItem = self._input_box.item(inputIndex)

            outputRect = self._output_box.itemRect(outputItem)
            inputRect = self._input_box.itemRect(inputItem)

            # possition of the output port
            if outputRect.isValid(): #returnes FALSE if the item is not visable
                outputPos = QPoint(outputRect.x() + self._output_box.x() + self._output_box.width() + 3,
                                   outputRect.y() + self._output_box.y() + 8)
            else:
                #need to find it if the item is bellow the bottom or over the top
                if (outputIndex < len(self._output_ports)/2):
                    # should the over the top
                    outputPos = QPoint(self._output_box.x() + self._output_box.width() + 3,
                                       self._output_box.y())
                else:
                    # should be bellow the bottom
                    outputPos = QPoint(self._output_box.x() + self._output_box.width() + 3,
                                       self._output_box.y() + self._output_box.height())
                    
            # possition of the input port
            if inputRect.isValid():
                inputPos = QPoint(inputRect.x() + self._input_box.x() - 3,
                                  inputRect.y() + self._input_box.y() + 8)
            else:
                #need to find it if the item is bellow the bottom or over the top
                if (inputIndex < len(self._input_ports)/2):
                    # should the over the top
                    inputPos = QPoint(self._input_box.x(), self._input_box.y())
                else:
                    # should be bellow the bottom
                    inputPos = QPoint(self._input_box.x(),
                                      self._input_box.y() + self._input_box.height())

            painter.drawLine(outputPos, inputPos)
            
    def _slot_clicked_output(self, index):
        port = self._output_ports[index]

        self._name_edit.setText(str(port.getName()))
        self._type_edit.setText(str(port.getDataType()))
        self._description_edit.setText(str(port.getDescription()))

    def _slot_clicked_input(self, index):
        port = self._input_ports[index]

        if isinstance(port, MultipleInputPort):
            self._name_edit.setText(port.get_name())
            self._type_edit.setText("Multi-input port")
            self._description_edit.setText("Several output-ports can be connected to this port")
        else:
            self._name_edit.setText(str(port.getName()))
            self._type_edit.setText(str(port.getDataType()))
            self._description_edit.setText(str(port.getDescription()))

    def _slot_connect(self):
        outputIndex = self._output_box.currentItem()
        inputIndex = self._input_box.currentItem()

        if inputIndex == -1 or outputIndex == -1:
            return
        
        key = (outputIndex,inputIndex)

        if self._deleted_connections.has_key(key):
            del self._deleted_connections[key]
        
        busy_connections = {}
        #Checks if target ports in other connections
        for busy_output_index, busy_input_index in self._connections:
            # ignore multiple_input_nodes (which can have more than
            # one input)
            if isinstance(self._input_ports[busy_input_index],
                          MultipleInputPort):
                continue
            
            if busy_input_index == inputIndex:
                busy_connections[(busy_output_index, busy_input_index)] = 1

        #Deletes conflicting connects found abouve
        for busykey in busy_connections:
            self._disconnect_key(busykey)

        # newConnections is a set, value isn't important
        self._new_connections[key] = 1
        
        if not self._connections.has_key(key):
            self._connections[key] = [self._output_ports[outputIndex],
                                      self._input_ports[inputIndex]]

        # update that sends a paintEvent
        self._line_draw_widget.update()

    def _slot_disconnect(self):
        outputIndex = self._output_box.currentItem()
        inputIndex = self._input_box.currentItem()

        if inputIndex == -1 or outputIndex == -1:
            return
        
        key = (outputIndex,inputIndex)

        self._disconnect_key(key)
        
        self._line_draw_widget.update()

    def _disconnect_key(self, key):
        if self._new_connections.has_key(key):
            del self._new_connections[key]
            
        if self._connections.has_key(key):
            del self._connections[key]
            
            if not self._deleted_connections.has_key(key):
                self._deleted_connections[key] = [self._output_ports[key[0]],
                                                  self._input_ports[key[1]]]
       
    def _slot_ok(self):

        for key in self._deleted_connections:
            output_port, input_port = self._deleted_connections[key]

            if isinstance(input_port, MultipleInputPort):
                input_port.remove_data_source(output_port)
            else:
                input_port.disconnectFromDataSource()

        for key in self._new_connections:
            output_port, input_port = self._connections[key]

            if isinstance(input_port, MultipleInputPort):
                input_port.add_data_source(output_port)
            else:
                input_port.connectToDataSource(output_port)
                output_port.setPosition(int(input_port.getId()), self._start_index,
                                        self._end_index)
        
        self._widget.close()
        del self._widget
        
    def _slot_help(self):
        HelpViewer(self._line_draw_widget, os.path.join(Config.htmldocpath,
                                      "node_User_interface.html"))


class LineDrawWidget(QWidget):
    """Display lines for connections."""

    def __init__(self, parent, draw_function, other_widget):
        QWidget.__init__(self, parent)
        self.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding)
        self._draw_function = draw_function
        self._other_widget = other_widget

    def paintEvent(self, event):
        painter = QPainter(self, True)
        pen = QPen(Qt.black, 2)
        painter.setPen(pen)

        other_origin = self._other_widget.mapToGlobal(QPoint(0,0))
        origin = self.mapToGlobal(QPoint(0,0))
        
        painter.translate(other_origin.x() - origin.x(),
                    other_origin.y() - origin.y())

        self._draw_function(painter)
