# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os

from qt import QDockWindow, QListView, QListViewItem, QObject, QPixmap
from qt import QPopupMenu, QSizePolicy, QSplitter, QTextDrag, QVBoxLayout
from qt import QWidget, SIGNAL

from GUI.viewers import HelpViewer
from GUI.NodeConfig import NodeConfig
from Library.Util.Config import Config
from Library.Util.gui import EventReceiver
from Library.Util.plot.customplotrepository import CustomPlotRepository
from System.NodeRepository import FunctionNodeEntry


class RightDockWindow(QDockWindow):
    """Creates a dock window with the node tree and the node status panel."""

    def __init__(self, parent, node_repository, workspace, node_status_list,
                 debug_mode):
        QDockWindow.__init__(self, QDockWindow.InDock, parent)

        #Dock window properties
        self.setResizeEnabled(1)
        self.setMovingEnabled(0)
        
        # Create a wrapper widget to control layout.
        # Content: a splitter and two list views.
        wrapper_widget = QWidget(self)
        wrapper_widget_layout = QVBoxLayout(wrapper_widget, 1, 2)

        splitter = QSplitter(wrapper_widget)
        splitter.setOrientation(QSplitter.Vertical)

        self.node_tree = NodeTree(splitter, node_repository, workspace,
                                  debug_mode)
        self.node_tree.setSizePolicy(
            QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding,
                        QSizePolicy.Fixed,QSizePolicy.Expanding,
                        self.node_tree.sizePolicy().hasHeightForWidth()))

        node_status_panel = NodeStatusPanel(splitter)
        node_status_panel.setSizePolicy(
            QSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding,
                        QSizePolicy.Fixed, QSizePolicy.MinimumExpanding,
                        node_status_panel.sizePolicy().hasHeightForWidth()))
        
        wrapper_widget_layout.addWidget(splitter)

        self.setWidget(wrapper_widget)

        # Register the status panel with the node status list (model).
        node_status_list.register_displayer(node_status_panel)


    def get_node_tree(self):
        return self.node_tree


class DragListView(QListView):

    def __init__(self, parent):
        QListView.__init__(self, parent, 'Nodes')
        # These must be handles by subclass.
        self.nodeItem = None
        self.dragObject = None

    def contentsMouseMoveEvent(self, event):
        # self._dock_window.nodeItem is reset to None in MDView when item has been released
        if self.nodeItem:
            # initiate drag
            self.dragObject = DragObject(self.nodeItem, self)

            # in case of functionnode, look for custom icon
            node_name = str(self.nodeItem)
            if self.node_repository and isinstance(self.node_repository.get_class_by_node_name(node_name.split(".")[-1]), FunctionNodeEntry):
                node_name = str(self.nodeItem).split(".")[-1]
                icon_path = self.node_repository.get_icon_path_by_node_name(node_name)
                if icon_path:
                    self.dragObject.setImageURL(icon_path)
                    
            self.nodeItem = None
            self.dragObject.drag()

    def contentsMouseReleaseEvent(self, event):
        """Ensure that the dragging stops when the mouse button is released."""
        self.nodeItem = None


class ContextMenu(QPopupMenu):
    """Pops up when a node is rigth-clicked."""

    _help_item_id, _refresh_id = range(2)
    
    def __init__(self, parent_widget, qpoint, item, node_repository):
        QPopupMenu.__init__(self, parent_widget)
        self.node_repository = node_repository
        self.item = item
        self.parent_widget = parent_widget
        self.popup(qpoint)
        QObject.connect(self, SIGNAL("activated(int)"), self.context_action)
        node_name = str(self.item.text(0))

        # Try function nodes first
        self.doc_path = self.node_repository.get_doc_path_by_node_name(node_name)
        menu_item_name = "Node Help Index..."
        if self.doc_path:
            menu_item_name = node_name + " Documentation..."
        # Otherwise, try built-in nodes
        else:
            self.doc_path = self.get_doc_path_by_name(node_name)
            if self.doc_path:
                menu_item_name = "Help on " + node_name + "..."
        # Otherwise, use standard doc page
        if not self.doc_path:
            self.doc_path = os.path.join(Config.htmldocpath, "node_nodes.html")

        self.insertItem(menu_item_name, self._help_item_id)
        self.insertItem("Refresh nodes", self._refresh_id)
        
    def context_action(self, id):
        if id == self._help_item_id:
            HelpViewer(self.parent_widget, self.doc_path)
        elif id == self._refresh_id:
            self.node_repository.reload_functions()
            self.parent_widget.update_node_list(dict(self.node_repository.getNodeList()))
    
    def get_doc_path_by_name(self, name):
        if name == "Edit":
            return os.path.join(Config.htmldocpath, "node_EditNode.html")
        elif name == "Gaussdal":
            return os.path.join(Config.htmldocpath, "node_Gaussdal_Node.html")
        elif name == "Plot2D":
            return os.path.join(Config.htmldocpath, "node_Plotting_in_2D.html")
        elif name == "Plot3D":
            return os.path.join(Config.htmldocpath, "node_Plotting_in_3D.html")
        elif name in ["SpreadSheet", "Input", "Output"]:
            return os.path.join(Config.htmldocpath, "node_Overview.html")
        else:
            return None


class NodeTree(DragListView):
    """Creates a list of available nodes and displays it."""
    
    def __init__(self, parent, node_repository, workspace, debug_mode=False):
        DragListView.__init__(self, parent)
        
        self.workspace = workspace
        self.node_repository = node_repository
        self.debug_mode = debug_mode

        self.addColumn("Nodes")
        self.setRootIsDecorated(True) # open/close sign on root items
        # enable drag and drop
        self.setAcceptDrops(1)
        self.dragObject = None
        
        self.connect(self,
                     SIGNAL("doubleClicked(QListViewItem *, const QPoint &, int)"),
                     self.slot_item_double_clicked)
        self.connect(self,
                     SIGNAL("mouseButtonPressed(int, QListViewItem *, const QPoint &, int)"),
                     self.slotStartDrag)
        self.connect(self,
                     SIGNAL("rightButtonPressed(QListViewItem *, const QPoint &, int)"),
                     self.slot_item_right_clicked)

        self.update_node_list(self.node_repository.getNodeList())            
        self.nodeItem = None

        self.show()


    def update_node_list(self, available_nodes):
        if not self.debug_mode:
            for node in available_nodes.keys():
                if node.startswith("Debug"):
                    del available_nodes[node]
        else:    
            # Hack: only enabling the hack to show Plot3D when using debugmode.
            available_nodes["Plot.Plot3DNode:Plot3D"] = "Plot3D"

        customplots = CustomPlotRepository().get_custom_plots_as_dict()
        if customplots:
            for key in customplots:
                newkey = "Plot.Saved plots:"+key
                available_nodes[newkey] = customplots[key]
        
        self.nodes = {}
        
        parent_elements = {}
        next_dict = parent_elements
        for key in available_nodes:
            keys = str(key).replace(":", ".").split(".")
            next_dict = parent_elements
            for element in keys[:-1]:
                if next_dict.has_key(element):
                    next_dict = next_dict[element]
                elif(element != "Plot3DNode"):
                    next_dict[element] = {}
                    next_dict = next_dict[element]
            next_dict[key] = available_nodes[key]
            
        self._setEntireList(parent_elements)


    def _setEntireList(self, list):
        self.clear()
        self._setListElements(self, list)
        
    def _setListElements(self, parentListItem, list):
        for key in list:
            if isinstance(list[key], dict):
                subItem = QListViewItem(parentListItem)
                subItem.setText(0, key.replace("Node", ""))
                self._setListElements(subItem, list[key])
            elif isinstance(list[key], str):
                itemKey = list.keys()[0]
                type = itemKey.split(".")[0]
                item = QListViewItem(parentListItem)
                item.setText(0, list[key].replace("Node", ""))
                if os.path.exists(os.path.join(Config.pixmapdir, str(type) + "_small.png")):
                    item.setPixmap(0, QPixmap(os.path.join(Config.pixmapdir, str(type) + "_small.png")))
                else:
                    item.setPixmap(0, QPixmap(os.path.join(Config.pixmapdir, "Functions_small.png")))
                self.nodes[item] = key

                
    def slot_item_double_clicked(self, item, qpoint, integer):
        """This is called when a node is double-clicked"""
        if (item in self.nodes) and self.workspace:
            self.workspace.slot_add_node(0, self.nodes[item])
            self.nodeItem = None


    def dragEnterEvent(self, event):
        if event.format().startswith("text/plain"):
            if str(event.encodedData("text/plain")).startswith("SciCraft node"):
                event.accept()
   
        
    def dragMoveEvent(self, event):
        if event.format().startswith("text/plain"):
            if str(event.encodedData("text/plain")).startswith("SciCraft node"):
                if self.dragObject.startPosX is None:
                    self.dragObject.setStartPos(event.pos().x(), event.pos().y())
                elif (abs(event.pos().x()-self.dragObject.startPosX)>5) or \
                    (abs(event.pos().y()-self.dragObject.startPosY)>5 ):
                    self.dragObject.displayPixmap()

                        
    def slotStartDrag(self, button, item, qpoint, integer):
        # only accept drags with left mouse button (1) pressed
        # do not start drag until mouse has moved (see DragListView)
        if button == 1 and (item in self.nodes) and self.workspace:
            self.nodeItem = self.nodes[item]
            
    def slot_item_right_clicked(self, item, qpoint, integer):
        """This slot is called when a node in the list view tree is
        right-clicked
        """
        if item:
            ContextMenu(self, qpoint, item, self.node_repository)

    def contextMenuEvent(self, e):
        """Disables the context menu in this widget"""
        return
        
class DragObject(QTextDrag):
    """This class wraps up node data to be sent to MDView"""
    
    def __init__(self, node_item_key, dragSource):
        QTextDrag.__init__(self, "SciCraft node", dragSource, "")
      
        self.nodeName = node_item_key
        
        if self.nodeName.startswith("Plot.Saved plots"):
            nodeConfig = NodeConfig("Plot2DNode")
        else:
            nodeConfig = NodeConfig(self.nodeName.replace(":", ".").split(".")[-1])
        
        self.imageURL = nodeConfig.getImage()
        
        self.pixmapWidth = None
        self.pixmapHeight = None

        self.startPosX = None
        self.startPosY = None
       
        
    def setStartPos(self, startPosX, startPosY):
        self.startPosX = startPosX
        self.startPosY = startPosY
    
    def setImageURL(self, url):
        self.imageURL = url
        
    def displayPixmap(self):
        pixmap = QPixmap(self.imageURL)
        QTextDrag.setPixmap(self, pixmap)
        self.pixmapWidth = self.pixmap().width()
        self.pixmapHeight = self.pixmap().height()


class NodeStatusPanel(QListView, EventReceiver):
    """Displays the status of the nodes in a module diagram.
        
    Relies on a model of the node status, and simply displays
    the data of the model. Updating the model is not the responsibility
    of this class.
    """
    
    def __init__(self, parent):
        QListView.__init__(self, parent)
        EventReceiver.__init__(self)

        self.parent = parent
        self._list_items = {}
        
        self.addColumn("Status")
        self.addColumn("Node")

    
    def model_update(self, instance, model):
        """Updates the view of the model."""
        # Iterate through data in model and find changes and additions.
        # We update only the elements that changes.
        for key, data in model.items():
            if not self._list_items.has_key(key):
                self._list_items[key] = QListViewItem(self, None)
            # Update columns that have changed, if any.
            item = self._list_items[key]
            for index in range(0, 2):
                if item.text(index) != data[index]:
                    item.setText(index, data[index])
        
        # Remove nodes that have been deleted
        tmp_list = []
        for key in self._list_items:
            if not key in model:
                self.takeItem(self._list_items[key])
                tmp_list.append(key)
        for entry in tmp_list:
            self._list_items.pop(entry)


    def handle_event(self, event):
        """Handles update in the GUI thread."""
        controller, model = event
        self.model_update(controller, model)
