# -*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os

from qt import SIGNAL, QApplication, QMessageBox, QFileDialog, QString
from qt import Qt, QCursor, QPoint, QObject, QClipboard, QFontMetrics

# must be imported to be able to call methods on a QTable retrieved from ui
from qttable import QTable

from qtui import QWidgetFactory
from GUI.viewers import HelpViewer
from GUI.Plots.plot2d.dataselector import DataSelector
from Library.Nodes.File.OutputNode import OutputNode
from Library.Util.Config import Config
from Library.Util.lists import get_repr_dim
from Numeric import array
from System.Data.Data import Data
from System.FileHandlers.Output.OctaveFileWriter import OctaveFileWriter

connect = QObject.connect

class UnknownExtensionError(Exception):
    """Signalises that a filename extension is unknown."""
    pass


def auto_adjust_cells(table):
    """Shrink/expand cells in table to fit data."""

    #since QTable cannot update the correct size of it's own QHeader we have
    #to do some magic... Below QFontMetrics is used to find the bounding
    #rectangle on all sections in the header. Then the width of the
    #rectangle is measured up against the previous width set. initially, we
    #set the comparison width to zero. wenn all sections is compared up
    #against the width variable, we can set the left margin of the QTable.

    font = table.verticalHeader().font()
    fm = QFontMetrics(font)

    width = 0

    for index in range(table.verticalHeader().count()):
        label = table.verticalHeader().label(index)
        width = max(fm.boundingRect(label).width(), width)

    # setting a border and padding since it currently isn't possible to
    # retrieve this from Qt.
    padding_and_border = 12

    table.setLeftMargin(width + padding_and_border)

    for y in range(table.numRows()):
        table.adjustRow(y)
    for x in range(table.numCols()):
        table.adjustColumn(x)


class SpreadSheetWindow:
    """This window contains the spreadsheet.

    This Window is a basic spreadsheet.
    """
    def __init__(self):
        """Create window displaying spreadsheet.

        Functionality of menus in user interface is connected to
        handler. table_widget is inserted into user interface.
        """
        self._window = QWidgetFactory.create(os.path.join(Config.uidir,
                                                    "spreadsheetwindow.ui"),
                                             None);

        self._table = QTable(self._window.child('box'))
        self._handler = SpreadSheetWidget(self._table)

        self._window.connect(self._window.child("save_action"),
                             SIGNAL("activated()"),
                             self._handler._slot_save_data)
        self._window.connect(self._window.child("windowCloseAction"),
                             SIGNAL("activated()"),
                             self._handler._slot_close)
        self._window.connect(self._window.child("editCopyAction"),
                             SIGNAL("activated()"),
                             self._handler._slot_copy_to_clipboard)
        self._window.connect(self._window.child("editSelect_AllAction"),
                             SIGNAL("activated()"),
                             self._handler._slot_select_all)
        self._window.connect(self._window.child("viewColumn_wiseAction"),
                             SIGNAL("toggled(bool)"), 
                             self._handler._slot_view_orientation_col)
        self._window.connect(self._window.child("viewRow_wiseAction"),
                             SIGNAL("toggled(bool)"), 
                             self._handler._slot_view_orientation_row)
        self._window.connect(self._window.child("helpHelpAction"),
                             SIGNAL("activated()"),
                             self._handler._slot_help)

    def show(self):
        self._window.show()
        self._window.raiseW()

    def set_data(self, data_list):
        self._handler.set_data(data_list)


class DataComposer:
    """Spreadsheet allowing user to fill it with data from ports.
    
    Data added to spreadsheet will always be filtered by a list of
    selected indices as provided on creation.
    """
    def __init__(self, indices=[]):
        self._window = QWidgetFactory.create(os.path.join(Config.uidir,
                                                    "datacomposerwindow.ui"),
                                             None);
        self._data_selector = DataSelector()
        self._data_selector.set_source_combo(self._window.child('source_combo'))
        self._data_selector.set_mode_combo(self._window.child('mode_combo'))
        self._data_selector.set_index_spinbox(self._window.child('index_spinbox'))

        self._table = self._window.child('table')

        self._datagrid = DataGrid(indices)
        self._datasets_in_columns = True

        connect(self._window.child('add_button'), SIGNAL("clicked()"),
                self._add_selected_data_to_spreadsheet)
        connect(self._window.child('orient_columns_action'), SIGNAL("activated()"),
                self._slot_display_datasets_as_columns)
        connect(self._window.child('orient_rows_action'), SIGNAL("activated()"),
                self._slot_display_datasets_as_rows)
        connect(self._window.child('delete_action'), SIGNAL("activated()"),
                self._slot_remove_selected_datasets)
        connect(self._window.child('copy_action'), SIGNAL("activated()"),
                self._slot_copy_selected)

    def update_ports(self, port_list):
        """Update list of available data to add."""
        self._data_selector.set_active_sources(port_list)

    def show(self):
        self._window.show()

    def _add_selected_data_to_spreadsheet(self):
        if self._data_selector.get_source_id() == None:
            return

        vector, description = self._data_selector.get_described_list()
        # we have to explicitly check for None as some vectors might
        # be empty, but we'd still like to add them
        if vector == None:
            return
        
        self._datagrid.add_data(description, vector)
        self._update_table()

    def _update_table(self):
        if self._datasets_in_columns:
            row_titles = self._datagrid.get_index_names()
            col_titles = self._datagrid.get_entry_names()
            data = self._datagrid.get_string_data()
        else:
            col_titles = self._datagrid.get_index_names()
            row_titles = self._datagrid.get_entry_names()
            data = zip(*self._datagrid.get_string_data())

        self._table.setNumRows(len(row_titles))
        self._table.setNumCols(len(col_titles))
        for col, rowset in enumerate(data):
            for row, entry in enumerate(rowset):
                self._table.setText(row, col, entry)

        for index, name in enumerate(row_titles):
            self._table.verticalHeader().setLabel(index, name)
        for index, name in enumerate(col_titles):
            self._table.horizontalHeader().setLabel(index, name)

        auto_adjust_cells(self._table)
        
    def _slot_display_datasets_as_rows(self):
        self._datasets_in_columns = False
        self._update_table()

    def _slot_display_datasets_as_columns(self):
        self._datasets_in_columns = True
        self._update_table()

    def _slot_remove_selected_datasets(self):
        # find selected indices
        if self._datasets_in_columns:
            indices = filter(lambda i : self._table.isColumnSelected(i, True),
                             range(self._table.numCols()))
        else:
            indices = filter(lambda i : self._table.isRowSelected(i, True),
                             range(self._table.numRows()))

        # find names
        all_names = self._datagrid.get_entry_names()[:]
        names = [all_names[i] for i in indices]

        for name in names:
            self._datagrid.remove_data(name)

        self._update_table()

    def _slot_copy_selected(self):
        """Copy selected cells to clipboard."""
        # only one selection
        if self._table.numSelections() < 1:
            return
        
        selection = self._table.selection(0)

        row_indices = range(selection.topRow(), selection.bottomRow() + 1)
        col_indices = range(selection.leftCol(), selection.rightCol() + 1)

        rows = []

        for row in row_indices:
            rows.append("\t".join(map(lambda col : unicode(self._table.text(row, col)),
                                      col_indices)))
        text = "\n".join(rows)

        cb = QApplication.clipboard()
        cb.setText(text, QClipboard.Clipboard)
        if cb.supportsSelection():
            cb.setText(text, QClipboard.Selection)
    

class SpreadSheetWidget:
    def __init__(self, table):
        self._table = table

        self._columnwise_orientation = True
        self._data = None
        self._file_writer_node = OutputNode()
        self._indices = None

    def set_view_orientation_columnwise(self, flag):
        """Sets the view orientation of the table.

        If a vector is shown columnwise, the whole vector is in one column,
        spanning multiple rows. If a vector is shown rowwise, the whole vector
        is in one row, spanning multiple columns. The flag is True if the data
        is to be displayed columnwise, else False.
        """
        self._columnwise_orientation = flag;

    def show(self):
        """Shows the window if data has been set."""
        if self._data:
            self._auto_adjust_cells()
        else:
            QMessageBox.information(self._window, "No data in spreadsheet",
                                    "There is no data in the spreadsheet.\n" +\
                                    "Most likely this means you have not\n" +\
                                    "run the diagram yet.", QMessageBox.Ok)

    def set_data(self, data, indices=None):
        """Sets the data that the spreadsheet shall show.

        The data provided is a list of System.Data.Data.Data-objects.
        """
        if not data:
            return

        self._indices = indices

        self._data = data
        self._table.setNumRows(0)
        self._table.setNumCols(0)
        for data_object in data:
            raw_data = data_object.getData()
            if type(raw_data) == type(array(0)):
                raw_data = raw_data.tolist()

            if indices:
                raw_data = [raw_data[i] for i in indices]

            name = data_object.getName()
            if type(raw_data) == type(list()):
                if get_repr_dim(raw_data) == 1:
                    self._add_vector_to_table(name, raw_data)
                elif get_repr_dim(raw_data) == 0:
                    # we have a list with dimension 0, which means it is
                    # empty: we assert to rule out an "impossible situation"
                    assert len(raw_data) == 0
                    self._add_scalar_to_table(name, "")
                else:
                    self._add_matrix_to_table(name, raw_data)
            else:
                self._add_scalar_to_table(name, raw_data)

        for i, index in enumerate(indices or []):
            if self._columnwise_orientation:
                self._table.verticalHeader().setLabel(i, str(index+1))
            else:
                self._table.horizontalHeader().setLabel(i, str(index+1))
                    
    def _add_matrix_to_table(self, name, matrix):
        if self._columnwise_orientation:
            offset = self._table.numCols()
            if len(matrix) > self._table.numRows():
                self._table.setNumRows(len(matrix))

            longest_col = self._get_longest_row(matrix)
            
            self._table.setNumCols(offset + longest_col)
            for col in range(longest_col):
                self._table.horizontalHeader().setLabel(
                    offset + col, "%s %d" % (name, col + 1))

            for row_index in range(len(matrix)):
                for col_index, value in enumerate(matrix[row_index]):
                    self._table.setText(row_index, offset + col_index,
                                        str(value))

        else:
            offset = self._table.numRows()
            if len(matrix) > self._table.numCols():
                self._table.setNumCols(len(matrix))

            longest_col = self._get_longest_row(matrix)

            self._table.setNumRows(offset + longest_col)
            for row in range(longest_col):
                self._table.verticalHeader().setLabel(offset + row,
                                                      "%s %d" % (name, row + 1))

            for col_index in range(len(matrix)):
                for row_index, value in enumerate(matrix[col_index]):
                    self._table.setText(offset + row_index, col_index,
                                        str(value))
            

    def _add_vector_to_table(self, name, vector):
        if self._columnwise_orientation:
            offset = self._table.numCols()
            if len(vector) > self._table.numRows():
                self._table.setNumRows(len(vector))

            self._table.setNumCols(offset + 1)
            self._table.horizontalHeader().setLabel(offset, name)

            for row_index, value in enumerate(vector):
                if type(value) == type(list()):
                    if value:
                        value = value[0]
                    else:
                        # empty list - skip this column
                        continue
                self._table.setText(row_index, offset, str(value))
        else:
            offset = self._table.numRows()
            if len(vector) > self._table.numCols():
                self._table.setNumCols(len(vector))
                
            self._table.setNumRows(offset + 1)
            self._table.verticalHeader().setLabel(offset, name)

            for col_index, value in enumerate(vector):
                if type(value) == type(list()):
                    value = value[0]
                self._table.setText(offset, col_index, str(value))

    def _add_scalar_to_table(self, name, scalar):
        if self._columnwise_orientation:
            offset = self._table.numCols()
            if self._table.numRows() < 1:
                self._table.setNumRows(1)
            self._table.setNumCols(offset + 1)
            self._table.horizontalHeader().setLabel(offset, name)
            self._table.setText(0, offset, str(scalar))
        else:
            offset = self._table.numRows()
            if self._table.numCols() < 1:
                self._table.setNumCols(1)
            self._table.setNumRows(offset + 1)
            self._table.verticalHeader().setLabel(offset, name)
            self._table.setText(offset, 0, str(scalar))

    def _get_longest_row(self, vectors):
        longest_row = 0
        for row in vectors:
            if longest_row < len(row):
                longest_row = len(row)
        return longest_row

    def _adjust_window(self):
        """Make sure window is centered and sized reasonably."""
        self._window.adjustSize()
        new_x = (QApplication.desktop().width() - self._window.width()) / 2
        new_y = (QApplication.desktop().height() - self._window.height()) / 2
        self._window.move(new_x, new_y)

    def _auto_adjust_cells(self):
        """Shrink/expand cells to fit data."""
        auto_adjust_cells(self._table)

    def _save_data(self, filepath):
        """Saves the data to the given filepath."""
        writers = self._file_writer_node.get_writers_by_extension()
        path, ext = os.path.splitext(filepath)
        # find the writer
        if ext[1:] not in writers.keys():
            raise UnknownExtensionError, ext 

        writer = writers[ext[1:]] # not the dot
        writer.write_data(self._data, filepath) 

    def _copy_to_clipboard(self):
        """Copy the contents of the table to the clipboard."""
        COL_SEP = "\t"
        ROW_SEP = os.linesep
        contents = QString("")

        if self._table.numSelections() < 1:
            return

        # only one selection
        selection = self._table.selection(0)
        
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        for row in range(selection.topRow(), selection.bottomRow() + 1):
            for col in range(selection.leftCol(), selection.rightCol() + 1):
                if col > selection.leftCol():
                    contents.append(COL_SEP)
                contents.append(self._table.text(row, col))

            contents.append(ROW_SEP)

        cb = QApplication.clipboard()

        cb.setText(contents, cb.Clipboard)
        if cb.supportsSelection():
            cb.setText(contents, cb.Selection)

        QApplication.restoreOverrideCursor()

    def _reload_data_to_table(self):
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        self._table.setUpdatesEnabled(False)
        self.set_data(self._data, self._indices)
        self._auto_adjust_cells()
        self._table.setUpdatesEnabled(True)
        QApplication.restoreOverrideCursor()
        # update to remove "ghosts" in cells not updated
        self._table.updateContents()

    # slots
    def _slot_view_orientation_col(self, on):
        self.set_view_orientation_columnwise(on)
        self._reload_data_to_table()

    def _slot_view_orientation_row(self, on):
        self.set_view_orientation_columnwise(not on)
        self._reload_data_to_table()

    def _slot_save_data(self):
        """Queries the user for a file name to save data to."""
        filters = self._file_writer_node.get_filters()
        filepath = QFileDialog.getSaveFileName("",
                                               filters,
                                               self._window,
                                               "save file dialog",
                                               "Choose a filename to save " +\
                                               "the data under")
        if filepath:
            try:
                self._save_data(str(filepath))
            except UnknownExtensionError, e:
                QMessageBox.information(self._window, "Unknown file type",
                                        "Could not recognize the file type " +\
                                        "from the extension %s." % e,
                                        QMessageBox.Ok)
            except Exception, e:
                QMessageBox.information(self._window, "Error writing file",
                                        str(e), QMessageBox.Ok)

    def _slot_close(self):
        self._window.close()

    def _slot_help(self):
        HelpViewer(self._window, os.path.join(Config.htmldocpath, 
                                              "node_SpreadSheet_Node.html"),
                   Config.htmldocpath)

    def _slot_copy_to_clipboard(self):
        self._copy_to_clipboard()

    def _slot_select_all(self):
        self._table.removeSelection(self._table.currentSelection())
        self._table.selectCells(0, 0, self._table.numRows(),
                                self._table.numCols())
        

class DataGrid:
    """Represents a grid of data-entries, model of spreadsheet.

    A grid isn't inherently horizontal or vertical, it's just a set of
    equal-length vectors with names of each vector. Viewing the
    vectors as either row-wise or column-wise is up to a view-layer.

    The grid is two-dimensional, of size number_of_vectors *
    entries_per_vector.

    A grid will only contain those elements of a vector at indices
    provided on creation.
    """
    def __init__(self, indices=[]):
        """No indices specified indicates full datasets should be used."""
        self._indices = indices
        self._data = []
        self._names = []
    
    def length(self):
        """Get length of grid, equal to length of all its vectors."""
        if not self._indices:
            return max(map(len, self._data))
        return len(self._indices)

    def num_entries(self):
        """Get number of vectors comprising this grid."""
        return len(self._data)

    def get_entry_names(self):
        """Get list of names for all datasets."""
        return self._names

    def get_index_names(self):
        """Get list of names of indices shared by datasets."""
        if not self._indices:
            return map(str, range(1, 1 + self.length()))
            
        return map(lambda i : str(i+1), self._indices)

    def get_string_data(self):
        """Get list of list of strings from datasets."""
        # pad entries to match length
        rows = []
        for entry in self._data:
            pad_list = [""]*(self.length() - len(entry))
            rows.append(map(str, entry) + pad_list)

        return rows
            
    def add_data(self, name, vector):
        """Add new named dataset."""
        if self._indices:
            indices = [i for i in self._indices if i < len(vector)]
            self._data.append([vector[i] for i in indices])
        else:
            self._data.append(vector[:])
            
        self._names.append(name)
    
    def remove_data(self, name):
        """Remove named vector."""
        index = self._names.index(name)
        del self._data[index]
        del self._names[index]
