# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from Library.Util.experiment import Experiment, BioReplica, MicroArray, Scan
from Library.Util.Config import Config
from Library.Util.gui import ValueListViewItem

from qt import QBrush, Qt, SIGNAL, QObject, QColor, QPen, PYSIGNAL, QInputDialog
from qt import QLineEdit, QMessageBox, QFileDialog, QPoint, QPointArray
from qt import QScrollView
from qtcanvas import QCanvasView, QCanvas, QCanvasRectangle, QCanvasLine
from qtui import QWidgetFactory

import os
import math
import cPickle

connect = QObject.connect
disconnect = QObject.disconnect

class MACanvasView(QCanvasView):
    def __init__(self, *args):
        QCanvasView.__init__(self, *args)
        self._item = None
        self._offset_x = 0
        self._offset_y = 0
        self._scale = 1.0
        self._rotation = 0.0
        self._canvas = QCanvas(640, 480)
        self._canvas.setBackgroundColor(QColor(255, 225,196))
        self.setCanvas(self._canvas)
        self._active_item = None
        self.setHScrollBarMode(QScrollView.AlwaysOn)
        self.setVScrollBarMode(QScrollView.AlwaysOn)
        
    def resizeEvent(self, e):
        QCanvasView.resizeEvent(self, e)
        width = e.size().width()
        height = e.size().height()

        self.resize_canvas_to_fit()

    def resize_canvas_to_fit(self):
        xmin, ymin, xmax, ymax = self.bounding_rect()

        visible_size = (self.visibleWidth(), self.visibleHeight())

        visible_scrollbars_size = (self.visibleWidth() - self.verticalScrollBar().width(),
                                   self.visibleHeight() - self.horizontalScrollBar().height())
        
        canvas_width, canvas_height = self.find_canvas_size((xmax, ymax),
                                                       visible_size,
                                                       visible_scrollbars_size)
        self.canvas().resize(canvas_width, canvas_height)

    def bounding_rect(self):
        """Return bounding rectangle of a set of nodes.

        Returned value is tuple (xmin, ymin, xmax, ymax)
        """
        items = self.canvas().allItems()
        if not items:
            return (0,0,0,0)

        rect = items[0].boundingRect()
        #xmin, ymin, xmax, ymax = items[0].boundingRect()
        xmin, ymin, xmax, ymax = rect.left(), rect.top(), rect.right(), rect.bottom()
        
        for item in items[1:]:
            rect = item.boundingRect()
            #xmin_, ymin_, xmax_, ymax_ = item.boundingRect()
            xmin_, ymin_, xmax_, ymax_ = rect.left(), rect.top(), rect.right(), rect.bottom()
            
            xmin = min(xmin, xmin_)
            ymin = min(ymin, ymin_)
            xmax = max(xmax, xmax_)
            ymax = max(ymax, ymax_)
            
        return (xmin, ymin, xmax, ymax)

    def find_canvas_size(self, items_size, visible_size, visible_scrollbars_size):
        """Calculate size of canvas to show scrollbars only when necessary.

        nodes_size is (max_x, max_y) of all nodes
        visible_size is (width, height) of view without scrollbars
        visible_scrollbars_size is (width, height) of view with scrollbars

        Return (width, height) tuple of best canvas size.
        """
        if items_size[0] <= visible_size[0] and items_size[1] <= visible_size[1]:
            return visible_size
        else:
            canvas_width = max(items_size[0], visible_scrollbars_size[0])
            canvas_height = max(items_size[1], visible_scrollbars_size[1])
            return (canvas_width, canvas_height)

    def get_active_item(self):
        return self._active_item

    def set_active_item(self, item):
        if self._active_item:
            self._active_item.dehighlight()
        
        self._active_item = item

        if isinstance(item, FollowingLine):
            self.emit(PYSIGNAL('line_clicked(comparison)'),
                      (item.get_comparison(),))
        else:
            self.emit(PYSIGNAL('line_clicked(comparison)'), (None,))
        
        if item:
            item.highlight()

        self.canvas().update()

    def get_canvas_rect_at_position(self, x, y):
        """Get canvas item for bio material at position."""
        items = filter(lambda item : isinstance(item, BioMaterialCanvasItem),
                       self.canvas().collisions(QPoint(x, y)))
        
        if items:
            return items[0]
        
    def contentsMousePressEvent(self, e):
        point = self.inverseWorldMatrix().map(e.pos())
        items = self.canvas().collisions(point)
        if items:
            self._item = items[0]
            self._offset_x = point.x() - self._item.x()
            self._offset_y = point.y() - self._item.y()

            self.set_active_item(self._item)
        else:
            self.set_active_item(None)

        self.setFocus()

    def contentsMouseMoveEvent(self, e):
        point = self.inverseWorldMatrix().map(e.pos())
        if isinstance(self._item, BioMaterialCanvasItem):
            new_x = point.x() - self._offset_x
            new_y = point.y() - self._offset_y
            x, y = self._validate_sprite_position(new_x, new_y)
            
            self.emit(PYSIGNAL('moving_bio_canvas_item(x, y, item)'),
                      (x, y, self._item))            

    def contentsMouseReleaseEvent(self, e):
        self._item = None
        
        point = self.inverseWorldMatrix().map(e.pos())
        items = self.canvas().collisions(point)
        if items:
            item = items[0]
            x, y = item.x(), item.y()
            if isinstance(item, BioMaterialCanvasItem):
                self.emit(PYSIGNAL('canvas_mouse_released(x, y)'), (x, y))
            elif isinstance(item, FollowingLine):
                self.emit(PYSIGNAL('line_clicked(comparison)'),
                          (item.get_comparison(),))
        else:
            self.emit(PYSIGNAL('canvas_mouse_released(x, y)'), (point.x(), point.y()))
            
    def contentsMouseDoubleClickEvent(self, e):
        point = self.inverseWorldMatrix().map(e.pos())
        items = self.canvas().collisions(point)
        if items and isinstance(items[0], BioMaterialCanvasItem):
            try:
                item = items[0]
                ok = False
                text = str(QInputDialog.getText(item.get_text(), 'Enter name', QLineEdit.Normal,
                                                item.get_text(), self)[0])
                
                if text:
                    item.set_text(str(text))
                    self.emit(PYSIGNAL('canvas_name_changed(text, bio)'),
                              (text, item.get_bio_material()))
                    
            except e:
                print e

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Delete:
            if isinstance(self._active_item, BioMaterialCanvasItem):
                self.emit(PYSIGNAL("delete_material(material)"),
                          (self._active_item.get_bio_material(),))
            elif isinstance(self._active_item, FollowingLine):
                self.emit(PYSIGNAL("delete_comparison(comparison)"),
                          (self._active_item.get_comparison(),))

    def _validate_sprite_position(self, x, y):
        new_x = x
        new_y = y
        
        canvas_width = self.canvas().width()
        canvas_height = self.canvas().height()
        
        if x < 0:
            new_x = 0
        elif x > canvas_width or (x+self._item.width()) > canvas_width:
            new_x = canvas_width - self._item.width()

        if y < 0:
            new_y = 0
        elif y > canvas_height or (y+self._item.height()) > canvas_height:
            new_y = canvas_height - self._item.height()

        self._item.move(new_x, new_y)

        return new_x, new_y        
                
    def delete_item_for_bio_material(self, bio):
        for item in self.canvas().allItems():
            if isinstance(item, BioMaterialCanvasItem):
                if bio == item.get_bio_material():
                    item.delete()
                    self.emit(PYSIGNAL('line_clicked(comparison)'), (None,))
                    break

    def delete_item_for_comparison(self, comparison):
        for item in self.canvas().allItems():
            if isinstance(item, FollowingLine):
                if comparison == item.get_comparison():
                    item.delete()
                    self.emit(PYSIGNAL('line_clicked(comparison)'), (None,))
                    break

class BioMaterialCanvasItem(QCanvasRectangle):
    def __init__(self, canvas, bio_material):
        x, y = bio_material.get_position()
        QCanvasRectangle.__init__(self, x, y, 120, 40, canvas)
        self._text = bio_material.get_name()
        pen = QPen()
        pen.setWidth(4)
        self.setPen(pen)
        self.setBrush(QBrush(Qt.white))
        self._listeners = []
        self._canvas = canvas
        self._bio = bio_material
        self._highlighted = False
        self.show()
        
    def get_text(self):
        return self._text

    def get_bio_material(self):
        return self._bio

    def set_text(self, text):
        self._text = text
        self.update()
        self._canvas.update()
        self.move(self.x(), self.y()) # force a redraw

    def add_move_listener(self, listener):
        self._listeners.append(listener)

    def move(self, xpos, ypos):
        QCanvasRectangle.move(self, xpos, ypos)
        for listener in self._listeners:
            listener(xpos + self.width()/2, ypos + self.height()/2)
        self._canvas.update()

    def delete(self):
        self.setCanvas(None)
        self._canvas.update()
        del self

    def highlight(self):
        self._highlighted = True
        self.update()

    def dehighlight(self):
        self._highlighted = False
        self.update()

    def drawShape(self, painter):
        fm = painter.fontMetrics()
        rect = fm.boundingRect(self._text)
        padding = 10
        border = 2

        self.setSize(rect.width() + 2 * (padding + border),
                     rect.height() + 2 * (padding + border))

        if self._highlighted:
            border_brush = QBrush(Qt.red)
        else:
            border_brush = QBrush(Qt.black)
        
        painter.fillRect(self.x(), self.y(), self.width(), self.height(),
                         border_brush)
        painter.fillRect(self.x() + border, self.y() + border,
                         self.width() - 2*border, self.height() - 2*border,
                         QBrush(Qt.white))
       
        painter.drawText(padding + border + self.x() - 1,
                         self.y() + self.height() - padding - border, self._text)        

class FollowingLine(QCanvasLine):
    def __init__(self, canvas, comparison):
        QCanvasLine.__init__(self, canvas)
        self._pen = QPen()
        self._pen.setWidth(2)
        self._pen.setColor(Qt.black)
        self.setPen(self._pen)
        self._comparison = comparison

    def delete(self):
        canvas = self.canvas()
        self.setCanvas(None)
        canvas.update()

    def get_comparison(self):
        return self._comparison

    def _get_direction(self):
        dx = self.startPoint().x() - self.endPoint().x()
        dy = self.startPoint().y() - self.endPoint().y()

        length = math.sqrt(pow(dx, 2) + pow(dy, 2))
        if not length:
            return 0

        direction = math.asin(dy / length)
        if dx < 0:
            direction = 3.14 - direction

        return direction

    def drawShape(self, painter):
        QCanvasLine.drawShape(self, painter)
        xpos = (self.startPoint().x() + self.endPoint().x()) / 2
        ypos = (self.startPoint().y() + self.endPoint().y()) / 2

        direction = self._get_direction()
        
        arrow_length = 10

        painter.drawLine(xpos, ypos, xpos - arrow_length*math.cos(direction + 0.5),
                         ypos - arrow_length*math.sin(direction + 0.5))
        painter.drawLine(xpos, ypos, xpos - arrow_length*math.cos(direction - 0.5),
                         ypos - arrow_length*math.sin(direction - 0.5))

    def areaPoints(self):
        direction = self._get_direction()

        boundary = 10
        
        points = QPointArray(4)
        points.setPoint(0,
                        self.startPoint().x() + boundary*math.cos(direction - math.pi/2),
                        self.startPoint().y() + boundary*math.sin(direction - math.pi/2))
        points.setPoint(1,
                        self.startPoint().x() + boundary*math.cos(direction + math.pi/2),
                        self.startPoint().y() + boundary*math.sin(direction + math.pi/2))
        points.setPoint(2,
                        self.endPoint().x() + boundary*math.cos(direction + math.pi/2),
                        self.endPoint().y() + boundary*math.sin(direction + math.pi/2))

        points.setPoint(3,
                        self.endPoint().x() + boundary*math.cos(direction - math.pi/2),
                        self.endPoint().y() + int(boundary*math.sin(direction - math.pi/2)))
                        
        return points
    
    def moveEnd(self, xpos, ypos):
        self.setPoints(self.startPoint().x(), self.startPoint().y(),
                       xpos, ypos)
        
    def moveStart(self, xpos, ypos):
        self.setPoints(xpos, ypos, 
                       self.endPoint().x(), self.endPoint().y())

    def highlight(self):
        self._pen.setColor(Qt.red)
        self.setPen(self._pen)
        
    def dehighlight(self):
        self._pen.setColor(Qt.black)
        self.setPen(self._pen)
    

class ComparisonEditor:
    """Shows currently selected comparison in designer."""
    def __init__(self, main_widget):
        self._main_widget = main_widget
        
        self._view = main_widget.child('listview')
        self._from_dye_edit = main_widget.child('from_dye')
        self._to_dye_edit = main_widget.child('to_dye')
        self._editing_stack = main_widget.child('stack')
        self._stack_frame = main_widget.child('stack_frame')
        self._ma_name_edit = main_widget.child('ma_name')
        self._bio_name_edit = main_widget.child('bio_name')
        self._scan_name_edit = main_widget.child('scan_name')
        self._scan_file_name_edit = main_widget.child('scan_filename')
        
        self._comparison = None
        self._view.clear()
        self._view.setDisabled(True)
        self._view.header().hide()

        self.set_comparison(None)

        # buttons connections
        connect(self._main_widget.child('delete_button'), SIGNAL("clicked()"),
                self.delete_current)
        connect(self._main_widget.child('bio_button'), SIGNAL("clicked()"),
                self.add_bio_replica)
        connect(self._main_widget.child('ma_button'), SIGNAL("clicked()"),
                self.add_hybridisation)
        connect(self._main_widget.child('scan_button'), SIGNAL("clicked()"),
                self.add_scan)

        # widget signals
        connect(self._from_dye_edit, SIGNAL('textChanged(const QString &)'),
                self._slot_from_dye_changed)
        connect(self._to_dye_edit, SIGNAL('textChanged(const QString &)'),
                self._slot_to_dye_changed)
        connect(self._view, SIGNAL('currentChanged(QListViewItem *)'),
                self._slot_update_stack)
        connect(main_widget.child('scan_pmt_combo'),
                SIGNAL('activated(const QString &)'),
                self._slot_scan_pmt_changed)
        connect(self._main_widget.child('scan_laser_combo'),
                SIGNAL('activated(const QString &)'),
                self._slot_scan_laser_changed)
        connect(self._main_widget.child('ma_dye_swap'),
                SIGNAL('activated(int)'), self._slot_ma_dye_swap_changed)
        connect(self._bio_name_edit,
                SIGNAL('textChanged(const QString &)'),
                self._slot_bio_name_changed)
        connect(self._ma_name_edit, SIGNAL('textChanged(const QString &)'),
                self._slot_ma_name_changed)
        connect(self._scan_name_edit, SIGNAL('textChanged(const QString &)'),
                self._slot_scan_name_changed)
        connect(self._scan_file_name_edit, SIGNAL('textChanged(const QString &)'),
                self._slot_scan_file_name_changed)

    def set_comparison(self, comparison):
        self._comparison = comparison
        if not comparison:
            self._view.setDisabled(True)
            self._from_dye_edit.setDisabled(True)
            self._to_dye_edit.setDisabled(True)
        else:
            self._view.setEnabled(True)
            self._from_dye_edit.setEnabled(True)
            self._to_dye_edit.setEnabled(True)

        self.update_bio_material_names()
        self._update_dye_colors()
        self.update_view()
        self._editing_stack.raiseWidget(self._main_widget.child('empty_stack_page'))
        self._stack_frame.setTitle("")

    def get_comparison(self):
        return self._comparison

    def _get_parents(self, item):
        parents = 0
        current = item
        while current.parent():
            parents += 1
            current = current.parent()
        return parents

    def update_view(self):
        self._view.clear()
        if not self._comparison:
            return
        for replica in self._comparison.get_bio_replicas():
            replica_item = ValueListViewItem(self._view, replica, replica.get_name())
            for ma in replica.get_micro_arrays():
                ma_item = ValueListViewItem(replica_item, ma, ma.get_name())
                for scan in ma.get_scans():
                    ValueListViewItem(ma_item, scan, scan.get_name())
                self._view.setOpen(ma_item, True)
            self._view.setOpen(replica_item, True)
        self._view.triggerUpdate()    
        
    def add_bio_replica(self):
        if not self._comparison:
            return

        replica = self._comparison.add_bio_replica()
        self.update_view()
        self._jump_to(replica.get_name())
        return replica

    def add_hybridisation(self):
        current = self._view.currentItem()
        if not current:
            return

        # get bio replica
        while current.parent():
            current = current.parent()
            
        bio_replica = current.get_value()
        ma = bio_replica.add_micro_array()
        self.update_view()
        self._jump_to(bio_replica.get_name(), ma.get_name())
        return ma

    def _jump_to(self, bio, ma=None, scan=None):
        item = self._view.firstChild()
        if not item:
            return 
        
        while str(item.text(0)) != bio:
            item = item.nextSibling()
            if not item:
                return

        if ma:
            item = item.firstChild()
            if not item:
                return 
        
            while str(item.text(0)) != ma:
                item = item.nextSibling()
                if not item:
                    return

        if scan:
            item = item.firstChild()
            if not item:
                return 
        
            while str(item.text(0)) != scan:
                item = item.nextSibling()
                if not item:
                    return

        self._view.setCurrentItem(item)
        self._view.ensureItemVisible(item)

    def add_scan(self):
        current = self._view.currentItem()
        if not current:
            return

        # make sure we are beneath a bio
        if not current.parent():
            return

        # jump up to ma if on scan
        if current.parent().parent():
            current = current.parent()

        ma = current.get_value()
        bio = current.parent().get_value()

        scan = ma.add_scan()
        self.update_view()
        self._jump_to(bio.get_name(), ma.get_name(), scan.get_name())
        return scan

    def delete_current(self):
        current = self._view.currentItem()
        if not current:
            return

        parents = self._get_parents(current)
        current_item = current.get_value()
        ma_name = None
        bio_name = None
        if parents == 0: # current is a bio replica
            self._comparison.remove_bio_replica(current_item)
        elif parents == 1: # current is a hybridisation
            bio = current.parent().get_value()
            bio.remove_micro_array(current_item)
            bio_name = bio.get_name()
        elif parents == 2: # current is a raw data file
            ma = current.parent().get_value()
            ma.remove_scan(current_item)
            ma_name = ma.get_name()
            bio_name = current.parent().parent().get_value().get_name()

        self._editing_stack.raiseWidget(self._main_widget.child('empty_stack_page'))
        self._stack_frame.setTitle("")

        self.update_view()
        self._jump_to(bio_name, ma_name)

    def get_list_view(self):
        return self._view

    def _update_dye_colors(self):
        if self._comparison:
            from_dye, to_dye = self._comparison.get_colors()
        else:
            from_dye, to_dye = "", ""
        
        self._from_dye_edit.setText(from_dye)
        self._to_dye_edit.setText(to_dye)

    def _slot_from_dye_changed(self, q_string):
        if not self._comparison:
            return
        
        old_from, to_dye = self._comparison.get_colors()
        self._comparison.set_colors(str(q_string), to_dye)

    def _slot_to_dye_changed(self, q_string):
        if not self._comparison:
            return

        from_dye, old_to = self._comparison.get_colors()
        self._comparison.set_colors(from_dye, str(q_string))

    def _slot_update_stack(self, item):
        edit_item = item.get_value()
        if isinstance(edit_item, BioReplica):
            widget = self._main_widget.child('bio_stack_page')
            self._current_bio = edit_item
            self._update_bio_edit_box()
        elif isinstance(edit_item, MicroArray):
            widget = self._main_widget.child('ma_stack_page')
            self._current_ma = edit_item
            self._update_ma_edit_box()
        elif isinstance(edit_item, Scan):
            widget = self._main_widget.child('scan_stack_page')
            self._update_scan_edit_box(edit_item)
        self._editing_stack.raiseWidget(widget)

        self._main_widget.child('stack_frame').setTitle(edit_item.get_name())

    def _update_ma_edit_box(self):
        disconnect(self._ma_name_edit, SIGNAL('textChanged(const QString &)'),
                   self._slot_ma_name_changed)
        self._main_widget.child('ma_name').setText(self._current_ma.get_name())
        if not self._current_ma.is_dye_swapped():
            self._main_widget.child('ma_dye_swap').setCurrentItem(0)
        else:
            self._main_widget.child('ma_dye_swap').setCurrentItem(1)
        connect(self._ma_name_edit, SIGNAL('textChanged(const QString &)'),
                self._slot_ma_name_changed)
        self._main_widget.child('ma_micro_array_name').setText(self._current_ma.get_array_name())

    def _update_bio_edit_box(self):
        disconnect(self._bio_name_edit,
                   SIGNAL('textChanged(const QString &)'),
                   self._slot_bio_name_changed)
        self._bio_name_edit.setText(self._current_bio.get_name())
        connect(self._bio_name_edit,
                SIGNAL('textChanged(const QString &)'),
                self._slot_bio_name_changed)

    def _update_scan_edit_box(self, scan):
        disconnect(self._scan_name_edit,
                   SIGNAL('textChanged(const QString &)'),
                   self._slot_scan_name_changed)
        disconnect(self._scan_file_name_edit,
                   SIGNAL('textChanged(const QString &)'),
                   self._slot_scan_file_name_changed)
        self._current_scan = scan
        self._scan_name_edit.setText(scan.get_name())
        self._main_widget.child('scan_pmt_combo').setCurrentItem(scan.get_pmt())
        self._main_widget.child('scan_laser_combo').setCurrentItem(scan.get_laser())
        
        filename = scan.get_filename()
        if filename:
            self._scan_file_name_edit.setText(filename)
        else:
            self._scan_file_name_edit.setText('Not generated')
        connect(self._scan_name_edit,
                SIGNAL('textChanged(const QString &)'),
                self._slot_scan_name_changed)
        connect(self._scan_file_name_edit,
                SIGNAL('textChanged(const QString &)'),
                self._slot_scan_file_name_changed)

    def _slot_scan_pmt_changed(self, text):
        if str(text)== 'Not set':
            self._current_scan.set_pmt(Scan.NOT_SET)
        elif str(text) == 'Low':
            self._current_scan.set_pmt(Scan.LOW)
        elif str(text) == 'Medium':
            self._current_scan.set_pmt(Scan.MEDIUM)
        elif str(text) == 'High':
            self._current_scan.set_pmt(Scan.HIGH)            
        
    def _slot_scan_laser_changed(self, text):
        if str(text)== 'Not set':
            self._current_scan.set_laser(Scan.NOT_SET)
        elif str(text) == 'Low':
            self._current_scan.set_laser(Scan.LOW)
        elif str(text) == 'Medium':
            self._current_scan.set_laser(Scan.MEDIUM)
        elif str(text) == 'High':
            self._current_scan.set_laser(Scan.HIGH)        

    def _slot_ma_dye_swap_changed(self, value):
        if not value:
            self._current_ma.set_dye_swapped(False)
        else:
            self._current_ma.set_dye_swapped(True)

    def _slot_ma_name_changed(self, string):
        self._current_ma.set_name(str(string))
        self.update_view()

    def _slot_scan_name_changed(self, string):
        self._current_scan.set_name(str(string))
        self.update_view()
        
    def _slot_bio_name_changed(self, string):
        self._current_bio.set_name(str(string))
        self.update_view()

    def _slot_scan_file_name_changed(self, string):
        self._current_scan.set_filename(str(string))
        self.update_view() 
    

    def update_bio_material_names(self):
        """Notify editor that names have changed."""
        box = self._main_widget.child('comparison_box')

        if self._comparison:
            bio1, bio2 = self._comparison.get_bio_materials()
            box.setTitle('Comparing %s to %s' % (bio1.get_name(), bio2.get_name()))
        else:
            box.setTitle('No comparison selected')


class MADesigner(QObject):
    TOOL_MOVE, TOOL_CREATE_BIO, TOOL_CREATE_COMPARISON = 0, 1, 2
    def __init__(self, parent, experiment_path=None):
        QObject.__init__(self)
        self._experiment = Experiment()
        self._from_bio = None
        self._tool = MADesigner.TOOL_MOVE        
        self._active_material = None
        self._moving_item = None
        self._current_edit_item = None
        self._parent = parent

        self._filename = None

        self._main_window = QWidgetFactory.create(os.path.join(Config.uidir,
                                                               "madesigner.ui"),
                                                  None, parent)

        self._stack = self._main_window.child('stack')
        self._bio_page = self._main_window.child('bio_stack_page')
        self._ma_page = self._main_window.child('ma_stack_page')
        self._scan_page = self._main_window.child('scan_stack_page')

        self._tool_bio_material_action = self._main_window.child('toolsBioMaterialAction')
        self._tool_pointer_action = self._main_window.child('toolsPointerAction')
        self._tool_comparison_action = self._main_window.child('toolsComparisonAction')
        micro_array_action = self._main_window.child('insert_micro_array_names')
        generate_filename_action = self._main_window.child('generate_filenames_action')

        connect(self._tool_pointer_action, SIGNAL('toggled(bool)'),
                self._slot_toggled_pointer_action)
        connect(self._tool_bio_material_action, SIGNAL('toggled(bool)'),
                self._slot_toggled_bio_material_action)
        connect(self._tool_comparison_action, SIGNAL('toggled(bool)'),
                self._slot_toggled_comparison_action)
        
        connect(micro_array_action, SIGNAL('activated()'),
                self._slot_micro_array_names_action)
        connect(generate_filename_action, SIGNAL('activated()'),
                self._slot_generate_filenames)

        connect(self._main_window.child('experimentNewAction'),
                SIGNAL('activated()'), self._slot_new_experiment)
        connect(self._main_window.child('experimentOpenAction'),
                SIGNAL('activated()'), self._slot_open_experiment)
        connect(self._main_window.child('experimentSaveAsAction'),
                SIGNAL('activated()'), self._slot_save_as_experiment)
        connect(self._main_window.child('experimentSaveAction'),
                SIGNAL('activated()'), self._slot_save_experiment)

        self._comparison_editor = ComparisonEditor(self._main_window)
        
        self._bio_name_edit = self._main_window.child('bio_name')
        self._ma_name_edit = self._main_window.child('ma_name')
        self._ma_dye_combo = self._main_window.child('ma_dye_swap')
        self._scan_pmt_combo = self._main_window.child('scan_pmt_combo')
        self._scan_laser_combo = self._main_window.child('scan_laser_combo')
        
        canvas_box = self._main_window.child('box')
        self._cv = MACanvasView(canvas_box)

        connect(self._cv, PYSIGNAL('canvas_mouse_released(x, y)'),
                self._handle_canvas_mouse_click)
        connect(self._cv, PYSIGNAL('moving_bio_canvas_item(x, y, item)'),
                self._slot_moving_bio_item)
        connect(self._cv, PYSIGNAL('line_clicked(comparison)'),
                self._slot_clicked_comparison)
        connect(self._cv, PYSIGNAL('delete_comparison(comparison)'),
                self._slot_delete_comparison)
        connect(self._cv, PYSIGNAL('delete_material(material)'),
                self._slot_delete_material)
        connect(self._cv, PYSIGNAL('canvas_name_changed(text, bio)'),
                self._slot_bio_name_changed)

        if experiment_path:
            self._load_experiment(experiment_path)
        
    def show(self):
        self._main_window.show()

    def _slot_delete_comparison(self, comparison):
        self._experiment.remove_comparison(*comparison.get_bio_materials())
        self._cv.delete_item_for_comparison(comparison)

    def _slot_delete_material(self, bio_material):
        for comparison in self._experiment.get_comparisons()[:]:
            if bio_material in comparison.get_bio_materials():
                self._slot_delete_comparison(comparison)

        self._experiment.remove_bio_material(bio_material)
        self._cv.delete_item_for_bio_material(bio_material)

    def _slot_open_experiment(self):
        filename = str(QFileDialog.getOpenFileName("", "Experiment (*.exp)",
                                                   self._main_window,
                                                   "open experiment dialog",
                                                   "Choose an experiment to open"))
        if filename:
                self._open_experiment(filename)

    def _slot_save_as_experiment(self):
        filename = str(QFileDialog.getSaveFileName("", "Experiment (*.exp)",
                                                   self._main_window,
                                                   "save experiment as dialog",
                                                   "Save experiment as..."))

        if filename:
            filename = _add_extension(filename)

            if os.path.exists(filename):
                result = QMessageBox.warning(self._main_window, "File Already Exists",
                                             "The file already exists. Do you wish to overwrite it?",
                                             QMessageBox.Yes, QMessageBox.No | QMessageBox.Default | QMessageBox.Escape)
                if result == QMessageBox.No:
                    self._slot_save_as_experiment()
                    return
            
            self._save_experiment(filename)

    def _slot_save_experiment(self):
        if not self._filename:
            self._slot_save_as_experiment()
        else:
            self._save_experiment(self._filename)

    def _slot_update_stack(self, item):
        self._current_edit_item = item.get_value()
        if isinstance(self._current_edit_item, BioReplica):
            self._stack.raiseWidget(self._bio_page)
            self._update_bio_edit_box()
        elif isinstance(self._current_edit_item, MicroArray):
            self._stack.raiseWidget(self._ma_page)
            self._update_ma_edit_box()
        elif isinstance(self._current_edit_item, Scan):
            self._stack.raiseWidget(self._scan_page)
            self._update_scan_edit_box()

        self._main_window.child('stack_frame').setTitle(self._current_edit_item.get_name())

    def get_experiment(self):
        return self._experiment

    def set_tool(self, tool):
        self._tool = tool
        if tool == MADesigner.TOOL_MOVE:
            self._tool_pointer_action.setOn(True)
            self._tool_bio_material_action.setOn(False)
            self._tool_comparison_action.setOn(False)
        elif tool == MADesigner.TOOL_CREATE_BIO:
            self._tool_pointer_action.setOn(False)
            self._tool_bio_material_action.setOn(True)
            self._tool_comparison_action.setOn(False)
        elif tool == MADesigner.TOOL_CREATE_COMPARISON:
            self._tool_pointer_action.setOn(False)
            self._tool_bio_material_action.setOn(False)
            self._tool_comparison_action.setOn(True)            

    def _handle_canvas_mouse_click(self, x, y):
        """Respond to mouse clicks in canvas coordinates."""
        if self._tool == MADesigner.TOOL_CREATE_BIO:
            self._add_bio_material(x,y)
            self.set_tool(MADesigner.TOOL_MOVE)
            self._cv.canvas().update()
        elif self._tool == MADesigner.TOOL_CREATE_COMPARISON:
            if not self._from_bio:
                self._from_bio = self._get_bio_material_at(x, y)
                if not self._from_bio:
                    self.set_tool(MADesigner.TOOL_MOVE)
            else:
                to_bio = self._get_bio_material_at(x, y)
                if to_bio:
                    self._add_comparison(self._from_bio, to_bio)
                self._from_bio = None
                self.set_tool(MADesigner.TOOL_MOVE)
        else:
            self._active_material = self._get_bio_material_at(x, y)

    def _get_bio_material_at(self, x, y):
        """Get bio material located at canvas coordinates."""
        for bio in self._experiment.get_bio_materials():
            if bio.get_position() == (x, y):
                return bio        
        
    def get_tool(self):
        return self._tool
        
    def get_active_material(self):
        return self._active_material 

    def _delete_active_material(self):
        self._experiment.remove_bio_material(self._active_material)
        if self._experiment.get_bio_materials():
            self._active_material = self._experiment.get_bio_materials()[-1]
        else:
            self._active_material = None

    def get_moving_item(self):
        """Get current dragged item."""
        return self._moving_item 

    def _set_moving_item(self, item):
        self._moving_item = item
        
    def _slot_toggled_pointer_action(self, bool_):
        if bool_:
            self.set_tool(MADesigner.TOOL_MOVE)
        if not self._tool_comparison_action.isOn() and \
               not self._tool_bio_material_action.isOn():
            self.set_tool(MADesigner.TOOL_MOVE)

        
    def _slot_toggled_bio_material_action(self, bool_):
        if bool_:
            self.set_tool(MADesigner.TOOL_CREATE_BIO)
        else:
            self.set_tool(MADesigner.TOOL_MOVE)

    def _slot_toggled_comparison_action(self, bool_):
        if bool_:
            self.set_tool(MADesigner.TOOL_CREATE_COMPARISON)
        else:
            self.set_tool(MADesigner.TOOL_MOVE)

    def _add_bio_material_to_canvas(self, bio):
        rect = BioMaterialCanvasItem(self._cv.canvas(), bio)
            
    def _add_bio_material(self, x, y):
        bio = self._experiment.add_bio_material(x, y)
        self._add_bio_material_to_canvas(bio)

    def _add_comparison_to_canvas(self, comparison):
        from_bio, to_bio = comparison.get_bio_materials()
        start_pos = from_bio.get_position()
        end_pos = to_bio.get_position()

        line = FollowingLine(self._cv.canvas(), comparison)
        line.setZ(-1)
        line.show()
        
        start_rect = self._cv.get_canvas_rect_at_position(*start_pos)
        end_rect = self._cv.get_canvas_rect_at_position(*end_pos)
        start_rect.add_move_listener(line.moveEnd)
        end_rect.add_move_listener(line.moveStart)

        start_rect.move(*start_pos)
        end_rect.move(*end_pos)        

    def _add_comparison(self, from_bio, to_bio):

        # First check if this connectiong already exists
        sets = [comparison.get_bio_materials() for comparison in \
                self._experiment.get_comparisons()]
        if (self._from_bio, to_bio) in sets or \
               (to_bio, self._from_bio) in sets:
            return
        
        comparison = self._experiment.add_comparison(from_bio, to_bio)
        self._add_comparison_to_canvas(comparison)
        
    def _slot_moving_bio_item(self, x, y, item):
        item.get_bio_material().set_position(x, y)

    def _slot_clicked_comparison(self, comparison):
        self._comparison_editor.set_comparison(comparison)

    def _slot_bio_name_changed(self, text, bio):
        bio.set_name(text)
        self._comparison_editor.update_bio_material_names()
        
    def _slot_micro_array_names_action(self):
        array_assigner = ArrayAssigner(self._main_window, self._experiment)
        array_assigner.show()

    def _slot_generate_filenames(self):
        self._experiment.assign_filenames()
        self._comparison_editor.update_view()

    def _load_experiment(self, filename):
        """Load experiment from named file.

        File format is expected to be a pickled experiment.
        """
        experiment = cPickle.load(open(filename))
        self._experiment = experiment
        self._filename = filename
        basename = os.path.basename(filename)
        self._main_window.setCaption("Micro Array Experiment Designer (%s)" % basename)

        for bio in experiment.get_bio_materials():
            self._add_bio_material_to_canvas(bio)

        self._cv.resize_canvas_to_fit()

        for comparison in experiment.get_comparisons():
            self._add_comparison_to_canvas(comparison)
        self._cv.canvas().update()            

    def _save_experiment(self, filename):
        """Save experiment to filename."""
        cPickle.dump(self.get_experiment(), open(filename, "w"))
        self._filename = filename
        basename = os.path.basename(filename)
        self._main_window.setCaption("Micro Array Experiment Designer (%s)" % basename)

    def _open_experiment(self, filename):
        """Handle user requests to open experiment."""
        if self.get_experiment().is_empty():
            self._load_experiment(filename)
        else:
            self.emit(PYSIGNAL("new_editor(filename)"), (filename,))

    def get_current_filename(self):
        """Get path of current experiment."""
        return self._filename

    def _slot_new_experiment(self):
        """User interactively starts a new experiment."""
        self.emit(PYSIGNAL("new_editor(filename)"), (None,))


        
class ArrayAssigner:

    def __init__(self, parent, experiment):
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir,
                                                          'madesigner-arrayassigner.ui'),
                                             None, parent)
        
        self._experiment = experiment
        self._edit_box = self._widget.child('names_edit')
        self._edit_box.setText("\n".join(self._experiment.get_array_names()))
        close_button = self._widget.child('close_button')
        assign_button = self._widget.child('assign_button')
        connect(close_button, SIGNAL('clicked()'), self._slot_close_dialog_clicked)
        connect(assign_button, SIGNAL('clicked()'), self._slot_assign_button_clicked)
                
    def show(self):
        self._widget.exec_loop()

    def _slot_close_dialog_clicked(self):
        self._set_values()
        self._widget.close()
        
    def _slot_assign_button_clicked(self):
        self._set_values()
        self._experiment.assign_names_to_micro_arrays()
        
        
    def _set_values(self):
        text_array = str(self._edit_box.text()).strip().splitlines()
        if text_array:
            self._experiment.set_array_names(text_array)

def _add_extension(filename):
    """Add eperiment extension to filename, if no extension exist."""
    basename, ext = os.path.splitext(filename)
    if ext:
        return filename
    else:
        return filename + ".exp"
