# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os

from cPickle import UnpicklingError

from Library.Util.rtester import RTester
from System.Data.Node import Node
from System.FileHandlers.Input.BASEFileReader import BASEFileReader
from System.FileHandlers.Input.GaussDalFileReader import GaussDalFileReader
from System.FileHandlers.Input.MoleculeFileReader import MoleculeFileReader
from System.FileHandlers.Input.OctaveFileReader import OctaveFileReader
from System.FileHandlers.Input.RFileReader import RFileReader
from System.FileHandlers.Input.experimentfilereader import ExperimentFileReader
from System.FileHandlers.Input.matfilereader import MatFileReader
from System.FileHandlers.Input.textfilereader import TextFileReader, TextFileSettings


class FileNotSetError(Exception):
    """Raised when performing operations on a file with no valid filename"""
    pass

class NoOutputError(Exception):
    """Raised when executing node and no output is set on the output ports"""
    pass

class InputNode(Node):
    """Node that reads diverse data files from disk."""
    
    READERS = [MatFileReader, BASEFileReader, OctaveFileReader, RFileReader,
               MoleculeFileReader, TextFileReader, GaussDalFileReader,
               ExperimentFileReader]
    unknown_format_string = "Unknown format"
    
    def __init__(self, name="InputNode"):
        Node.__init__(self, name)
        self._filename_parameter = self.create_parameter("filename", "text", "")
        self._format_parameter = self.create_parameter("fileformat", "text", "")

        self._textfilesettings = None # the Separator object
        
        # cache variables...
        self._cached_data = []
        self._cached_data_filename = None
        self._cached_data_fileformat = None
        self._cached_data_timestamp = None
        self._cached_data_textfilesettings = None

    def set_filename(self, path, force_format=""):
        """Set the file path of this node.

        If force_format is set, the path is believed to be a valid one
        as long as it exists. Otherwise, the following conditions must
        be met by path, to make it a valid one:

         * the whole path must exist (either as absolute path or relative to
           module diagram path)
         * the extension (case insensitive) must be known
         
        If all of the above conditions are met, True is returned.
        """
        name, ext = os.path.splitext(path)
        abspath = self._make_path_absolute(path)
        if os.path.exists(abspath) and force_format:
            self._filename_parameter.setValue(path)
            self._format_parameter.setValue(force_format)
            return True
            
        if os.path.exists(abspath) and \
               ext[1:].lower() in map(str.lower, self._known_extensions()):
            self._filename_parameter.setValue(path)
            reader_class = self._get_reader_class_from_extension(ext[1:])
            self._format_parameter.setValue(reader_class.fileDescription)
            return True
        else:
            self._filename_parameter.setValue("")
            self._format_parameter.setValue("")

    def get_filename(self):
        """Get the file path of this node.

        If no valid path is set, None is returned. For conditions for a path to
        be valid, see set_filename(). filename might be a relative path, and if
        so, it is relative to module diagram's path.
        """
        return self._filename_parameter.getValue()

    def get_fileformat(self):
        """Get the fileformat of the set file."""
        formatstring = self._format_parameter.getValue()
        if formatstring:
            return formatstring
        else:
            return InputNode.unknown_format_string

    def get_known_formats(self):
        """Get the fileformats known to the file reader node.

        First item in list is a string signalizing an unknown format.
        """
        formats = [InputNode.unknown_format_string]
        real = [reader.fileDescription for reader in InputNode.READERS]
        formats.extend(real)
        return formats

    def get_fileformat_filters(self):
        """Get a string appropriate for filters in QFileDialog."""
        filters = "All files (*)"
        for reader in InputNode.READERS:
            extensions = ""
            for ext in reader.supportedFileExtensions:
                if extensions:
                    extensions += " "
                    
                extensions += "*.%s" % ext
                
            filters += ";;%s (%s)" % (reader.fileDescription, extensions)

        return filters

    def set_fileformat(self, formatstring):
        """Set the fileformat of the set file."""
        for reader in InputNode.READERS:
            if formatstring == reader.fileDescription:
                self._format_parameter.setValue(formatstring)
                return
        
        self._format_parameter.setValue("")

    def _known_extensions(self):
        """Get a list of all extensions known to the file reader node."""
        extensions = []
        for reader in InputNode.READERS:
            extensions.extend(reader.supportedFileExtensions)
        return extensions

    def _make_path_absolute(self, path):
        """Makes the given path absolute.

        path could be a path relative to module diagram path (if module diagram
        is set and has a filename). If module diagram does not have a path, path
        is assumed to be relative to working directory.
        """
        if os.path.isabs(path):
            return path
        if self.moduleDiagram and self.moduleDiagram.getFilename():
            return os.path.join(os.path.dirname(self.moduleDiagram.getFilename()), path)
        else:
            return os.path.abspath(path)

    def _get_reader_class_from_extension(self, ext):
        """Get a reader based on extension."""
        ext = ext.lower()
        for reader in InputNode.READERS:
            if ext in map(str.lower, reader.supportedFileExtensions):
                return reader

    def _get_reader_class_from_formatstring(self, formatstring):
        """Get a reader based on extension."""
        for reader in InputNode.READERS:
            if formatstring == reader.fileDescription:
                return reader

    def get_file_reader_class_name(self):
        """Get the name of the file reader currently set."""
        format = self._format_parameter.getValue()
        reader_class = self._get_reader_class_from_formatstring(format)
        if reader_class:
            return reader_class.name
        else:
            return "Unknown"

    def _has_valid_cache(self):
        """Check if self._cached_data() is a valid mirror of file on disk."""

        # using self._textfilesettings toXML to determine wether it has changed
        if self._textfilesettings:
            textfilesettings = self._textfilesettings.toXML()
        else:
            textfilesettings = ""

        return self._cached_data \
               and self._cached_data_filename == self._make_path_absolute(self.get_filename()) \
               and self._cached_data_fileformat == self._format_parameter.getValue() \
               and os.path.exists(self._cached_data_filename) \
               and self._cached_data_timestamp == os.path.getmtime(self._cached_data_filename) \
               and self._cached_data_textfilesettings == textfilesettings

    def _cache_data(self, data):
        """Put data in cache."""
        self._cached_data = data
        self._cached_data_filename = self._make_path_absolute(self.get_filename())
        self._cached_data_fileformat = self._format_parameter.getValue()
        self._cached_data_timestamp = os.path.getmtime(self._cached_data_filename)

        if self._textfilesettings:
            self._cached_data_textfilesettings =  self._textfilesettings.toXML()
        else:
            self._cached_data_textfilesettings = ""
    
            
    def _get_cached_data(self):
        return self._cached_data

    def _read_file(self):
        """Read the file set with set_filename and return list of data."""
        if not self.get_filename():
            raise FileNotSetError, "File has not been set"

        if not self._has_valid_cache():
            fname = self._make_path_absolute(self.get_filename())
            format = self._format_parameter.getValue()
            reader_class = self._get_reader_class_from_formatstring(format)

            if not reader_class: # unknown format
                return []

            # special treatment of TextFileReader: If textfilesettings have been set, and manual read is requested call special method read_data_manual in TextFileReader.
            if reader_class.name == "TextFileReader" and self._textfilesettings and not self._textfilesettings.automatic:
                self._cache_data(reader_class().read_data_with_settings(fname,self._textfilesettings))
            else:
                self._cache_data(reader_class().read_data(fname))
            
        return self._get_cached_data()

    def get_available_output(self):
        """Get the available data objects from the file as list of Data.

        If an error occurs, an empty list is returned.
        """
        try:
            RTester.r_install_test()
            from rpy import RException
        except Exception, e:
            # Failed to import rpy, RException will never be thrown in that case and
            # we need to just put something there that we will never see            
            class DummyRException(Exception):
                pass
            RException = DummyRException
        
        datalist = []
        try:
            datalist = self._read_file()
        except FileNotSetError:
            pass
        except RException: # exception from r file reader
            pass 
        except IOError: # exception from matlab-reader
            pass 
        except TypeError: # exception from BASE file reader
            pass 
        except UnpicklingError: # exception from experiment file reader
            pass
        except IndexError: # exception from Octave file reader
            pass
        
        return datalist

    def set_available_output_as_ports(self):
        """Set the data available in set file as output ports."""
        datanames = [data.getName() for data in self.get_available_output()]
        for port in self.getOutputPorts()[:]:
            if port.getName() not in datanames:
                self.remove_port(port)
        
        for dataname in datanames:
            self.create_output_port(dataname)

    def execute(self):
        # if we have outputports but no available data, we have an error
        datalist = self.get_available_output()
        if not datalist and self.getOutputPorts():
            raise NoOutputError, "No output was found in file '%s'" % self.get_filename()
        
        for data in datalist:
            port = self.getOutputPortByName(data.getName())
            if port:
                port.setData(data)
        
        Node.execute(self)


    def toXML(self):
        """Inserts filereadersettings into NodeXML
           Assumes last object returned from Node.toXML() is '</node>'"""

        if not self._textfilesettings:
            return Node.toXML(self)

        xml_list = Node.toXML(self)
        xml_file = self._textfilesettings.toXML()

        return xml_list[:-1]+xml_file+xml_list[-1:]

    def from_xml_element(self, root):
        Node.from_xml_element(self, root)

        textfilexml = root.find("TextFileSettings")
        if textfilexml:
            self._textfilesettings = TextFileSettings()
            self._textfilesettings.from_xml_element(textfilexml)

    def get_textfilesettings(self):
        if not self._textfilesettings:
            self._textfilesettings = TextFileSettings()
        return self._textfilesettings
