# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from System.Data.Node import Node
from System.Data.Data import Data, RData
from Library.Util.unrepr import unrepr
from System.FileHandlers.Output.OctaveFileWriter import OctaveFileWriter
from System.FileHandlers.Output.RFileWriter import RFileWriter
from System.FileHandlers.Output.MoleculeFileWriter import MoleculeFileWriter
import os
import sets

# domain exceptions
class DuplicateNameError(Exception):
    """Same storage name for two variables in FileWriter."""

class UnknownWriterError(Exception):
    """FileWriter unset or unknown."""

class WriteError(Exception):
    """Error when writing data to file."""


class OutputNode(Node):
    """Node writing all inputs to file using available file writers.

    This node uses a multiinputport to receive data from all nodes
    connected to it and store everything in one file using one of the
    readers.
    """
    _file_writers = [OctaveFileWriter(), RFileWriter(), MoleculeFileWriter()]
    
    def __init__(self, name="OutputNode"):
        Node.__init__(self, name)

        self._multi_port = self.create_multi_input_port("input")
        self.setVariable("filename", "")
        self.setVariable("writer", "")
        self.setVariable("names", {})
        self._storage_names = {}
        
    def get_filename(self):
        """Get name of file to store all dataobjects on execute."""
        return self.getVariable("filename")

    def set_filename(self, filename):
        """Sets filename to use when storing data."""
        self.setVariable("filename", filename)

    def get_file_writers(self):
        """Gets all available file writers which do actual data storage."""
        return self._file_writers

    def set_stored_name(self, port, name):
        """Set a storage name for data from a given output port."""
        names = unrepr(self.getVariable("names"))
        names[port.getId()] = name
        self.setVariable("names", str(names))

    def get_storage_names(self):
        """Get list of stored variable name, emulated input port pairs."""
        storage_names = []
        names = unrepr(self.getVariable("names"))
        for port in self._multi_port.get_emulated_input_ports():
            connected_port = port.getDataSource()
            if connected_port.getId() in names:
                storage_names.append((names[connected_port.getId()], port))
            else:
                storage_names.append((connected_port.getName(), port))
        return storage_names

    def set_file_writer(self, file_writer):
        """Sets active file writer to use at next execution.

        file_writer is a unique name/identifier of one of the
        available file writers. OutputNode.get_file_writers()
        returns all file writer objects. FileWriter.get_filetype() on
        a file writer object will give the identifying string.
        """
        self.setVariable("writer", file_writer)

    def get_file_writer(self):
        """Get name of active file writer.

        The name used is the same as get_filetype() returns on a
        FileWriter instance.
        """
        return self.getVariable("writer")

    def get_file_writer_class(self):
        """Get name of file writer class, mostly to set icon in GUI."""
        for writer in self._file_writers:
            if writer.get_filetype() == self.getVariable("writer"):
                return writer.get_name()
        return None

    def _get_renamed_data(self):
        """Collect data and rename to user specified variable names."""
        data_objects = []
        names = unrepr(self.getVariable("names"))
        for port in self._multi_port.get_emulated_input_ports():
            connected_port = port.getDataSource()
            if connected_port.getId() in names:
                data_name = names[connected_port.getId()]
            else:
                data_name = connected_port.getName()
            try:
                data = port.getData()
                if isinstance(data, RData):
                    renamed_data = RData(data_name)
                    renamed_data.copy_data(data)
                else:
                    renamed_data = Data(data_name)
                    renamed_data.copy_data(data)
            except Exception, e:
                print "am i ready?", self.isReady()
                print "port:", port
                print "e:", e
            data_objects.append(renamed_data)
        return data_objects

    def execute(self):
        """Store data from ports to set filename using set filewriter."""
        # nothing connected?
        if not self._multi_port.get_emulated_input_ports():
            return
        
        file_writer = self.getVariable("writer")
        if not file_writer:
            raise UnknownWriterError, "No file type set in OutputNode."

        writers = dict([(writer.get_filetype(),
                         writer) for writer in self._file_writers])

        if file_writer not in writers:
            raise UnknownWriterError, "Invalid file type set in OutputNode."

        active_writer = writers[file_writer]
        path = self.getVariable("filename")
        if not path:
            raise WriteError, "No filename specified in OutputNode."
        
        # convert relative path to be relative to module diagram
        # before sending as absolute path to writer        
        if not os.path.isabs(path) and self.moduleDiagram:
            mdpath = self.moduleDiagram.getFilename()
            if mdpath:
                path = os.path.join(os.path.dirname(mdpath), path)

        renamed_data = self._get_renamed_data()
        # make sure data has unique names
        names = map(Data.getName, renamed_data)
        dupes = filter(lambda name : names.count(name) > 1, sets.Set(names))
        if dupes:
            raise DuplicateNameError, ", ".join(dupes) + " has duplicate entries in FileWriter."
        
        active_writer.write_data(self._get_renamed_data(), path)

    def get_filters(self):
        """Get QFileDialog style filters for filetypes supported."""
        filters = "All files (*)"
        for writer in self._file_writers:
            extensions = " ".join(["*." + ext for ext in writer.supportedFileExtensions])
            filters += ";;%s (%s)" % (writer.description, extensions)
        return filters

    def get_writers_by_extension(self):
        """Return table to lookup writer based on extension."""
        table = {}
        for writer in self._file_writers:
            for ext in writer.supportedFileExtensions:
                table[ext] = writer
        return table

