# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.


from System.Data.Node import Node
from System.Data.Data import Data
from itertools import izip

class IllegalValueError(Exception):
    """Used when input type is incorrect"""
    pass
    
class NotEnoughValuesError(Exception):
    pass



class UserQueryNode(Node):

    def __init__(self, name):
        Node.__init__(self, name)
        self._input_port = self.create_multi_input_port('query input')
        self._output_port = self.create_output_port('query output')

        self._values = []
        self._value_labels = []
        self._input_types = []
        self._number_of_values = 0

    def from_xml_element(self, root):
        Node.from_xml_element(self, root)
        
        query_root = root.find("query_setup")
        number_of_values = int(query_root.get("number_of_values"))
        types = query_root.get("types")
        labels = query_root.get("labels")

        self.set_number_of_values(number_of_values)
        if types:
            self.set_value_types(types.split(":"))
        if labels:
            self.set_value_labels(labels.split(":"))

        if number_of_values == 0:
            return

        types = []
        values = {}

        for query_value in query_root.findall("query_value"):
            value_id = query_value.get("id")
            values.setdefault(value_id, []).append(query_value.text)

        keys = values.keys()
        keys.sort()
        for key in keys:
            self.add_values(values[key])

    def set_value_types(self, types):
        self._input_types = []
        if not isinstance(types, list): types = [types]            
        for type_ in types:
            if type_ == 'Integer':
                self._input_types.append([int, type_])
            elif type_ == 'Text':
                self._input_types.append([str, type_])
            elif type_ == 'Float':
                self._input_types.append([float, type_])
            else:
                raise TypeError, '%s is not a valid type for this node' % type_

    def set_number_of_values(self, number):
        if number < self._number_of_values:
            self._strip_values(number)
            self._strip_labels(number)
        self._number_of_values = int(number)

    def add_values(self, values):
        converted_values = []
        for value, types in izip(values, self._input_types):
            type_, name = types
            try:
                converted = type_(value)
            except ValueError:
                raise IllegalValueError, 'Recived input of type %s, expected %s' \
                      % (type(value), name)
            converted_values.append(converted)

        if isinstance(values, list) and len(values) == 1:
            converted_values = converted_values.pop()

        self._values.append(converted_values)
        
        
    def get_number_of_values(self):
        return self._number_of_values

    def get_input_types(self):
        return self._input_types

    def get_values(self):
        return self._values

    def execute(self):
        data = Data('output', self._values)
        self._output_port.setData(data)

        
    def toXML(self):
        base = Node.toXML(self)
        query_xml = []
        labels = ":".join(self._value_labels)
        types = ":".join(map(lambda list_: list_[1], self._input_types))
        
        query_xml.append('\t<query_setup number_of_values="%s" types="%s" labels="%s">' \
                         % (self._number_of_values, types, labels))

        if not self._values:
            for type_ in self._input_types:
                query_xml.append('\t\t<query_value id="%s"></query_value>' % 0)
        else:
            for index, value in enumerate(self._values):
                if self._number_of_values > 1:
                    for data in value:
                        query_xml.append('\t\t<query_value id="%s">%s</query_value>' % \
                                         (index, data))
                else:
                    query_xml.append('\t\t<query_value id="%s">%s</query_value>' % \
                                     (index, value))
        query_xml.append('\t</query_setup>')
        for x in query_xml:
            base.insert(-1, x)
        return base
        

    def _strip_values(self, number):
        """Removes values if the number of inputs are lowere.

        For example, if the user changes the number of inputs for
        each row to 2 instead of 3, the last element in each row
        will be removed.
        """
        
        if not self._values:
            return
        if not isinstance(self._values[0], list):
            return

        diff = number - self._number_of_values
        new_list = []
        for list_ in self._values:
            new_values = list_[:diff]
            if len(new_values) == 1:
                new_list.append(new_values[0])
            else:
                new_list.append(list_[:diff])
        self._values = new_list

    def _strip_labels(self, number):
        diff =  number - self._number_of_values
        new_list = self._value_labels[:diff]
        self._value_labels = new_list


    def _remove_old_values(self):
        self._values = []


    def set_value_labels(self, labels):
        self._value_labels = labels

    def get_value_labels(self):
        return self._value_labels
