# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import Numeric
import copy

from Library.Util import numerical
from GUI.Plots.plot2d.plotmap import PlotMapHandler, PlotMapConfig
from System.Data.Node import Node
from System.Data.Data import Data

# helper method
def _parse_as_settings(root):
    settings = {}
    for tag in root.getchildren():
        settings[tag.tag] = tag.text or ""
    return settings

class Plot2DNode(Node):
    def __init__(self, name):
        Node.__init__(self, name)
        self._plot_map_handler = PlotMapHandler(self)
        self._multi_input = self.create_multi_input_port("plot input")

        # plot node settings
        self._widgets = {}
        self._horisontal_widgets = 0
        self._vertical_widgets = 0
        
        self._plot_maps = {}
        self._go_browser_maps = []

        # set one by one widgets as default
        self.update_widgets(1, 1)

    def from_xml_element(self, root):
        Node.from_xml_element(self, root)
        plotnode = root.find("plotnode_config")
        self.update_widgets(int(plotnode.find("horisontal_widgets").text),
                            int(plotnode.find("vertical_widgets").text))

        plot_map_config = plotnode.find("plot_map_config")
        if plot_map_config:
            handler = self.get_plot_map_handler()
            handler.from_xml_element(plot_map_config)

            plot_maps = plot_map_config.findall("plot_map")
            self.update_plot_maps(len(plot_maps))
            for i, plot_map in enumerate(plot_maps):
                self.set_plot_map_settings(i, _parse_as_settings(plot_map))
            

        for widget_element in plotnode.findall("widget_config"):
            x_coord = int(widget_element.find("x_coord").text)
            y_coord = int(widget_element.find("y_coord").text)
            
            widget_config = self.get_widget_config(x_coord, y_coord)
            widget_config.from_xml_element(widget_element)

        # handle port settings for custom plots
        input_ports = plotnode.find("input_ports_used")
        if input_ports:
            for port in input_ports.findall("ip"):
                self.create_input_port(port.find("name").text,
                                       port.find("type").text or "",
                                       port.find("description").text or "")

        plot_map = plotnode.find("plot_map_config")
        if plot_map:
            go_maps = map(_parse_as_settings, plot_map.findall("go_browser_map"))
            self.set_go_browser_map_settings(go_maps)

    def execute(self):
        """Executes this node"""
        Node.execute(self)
        # a plot node currently does no processing of data

    def get_data_size(self, port_name, is_column):
        """Get size of row/column at named port."""
        port = self.getInputPortByName(port_name)
        if not port:
            raise Exception, '[Plot2DNode]: no port called %s' % port_name

        data = port.getData()
        return data.get_data_size(is_column)
            
    def get_data(self, port_name, is_column, index):
        """Get row or column from matrix data at port."""
        port = self.getInputPortByName(port_name)
        if not port:                       
            raise Exception, '[Plot2DNode]: no port called %s' % port_name
        
        data = port.getData()
        if data:
            if data.get_col_indices() or data.get_row_indices():
                return data.get_sub_data_with_nan(is_column, index)
            else:                
                return data.get_sub_data(is_column, index)
        else:
            return []

    def get_all_data_by_name(self, port_name):
        port = self.getInputPortByName(port_name)
        if not port:
            return []

        data = port.getData()

        if data:
            if data.get_col_indices() or data.get_row_indices():
                return data.get_all_data_with_nan()
            else:
                return port.getData().getData()            
        else:
            return []

    def toXML(self):
        self.nodeRepName = "PlotNodes.Plot2DNode"
        xml_list = Node.toXML(self)
        xml_swap = xml_list[:-1]
        xml_swap.extend(self.to_xml())
        xml_swap.append(xml_list[-1])
        return xml_swap

    def to_xml(self):
        """Convert plot specific settings to xml as list of strings."""
        xml_list = []
        xml_list.append("\t<plotnode_config>")
        xml_list.append("\t\t<horisontal_widgets>%s</horisontal_widgets>" % self._horisontal_widgets)
        xml_list.append("\t\t<vertical_widgets>%s</vertical_widgets>" % self._vertical_widgets)

        for widget in self._widgets.values():
            xml_list.extend(widget.to_xml())

        handler = self.get_plot_map_handler()
        xml_list.append("\t\t<plot_map_config>")
        xml_list.append("\t\t\t<auto_map_enabled>%d</auto_map_enabled>" % handler.get_auto_map_enabled())
        xml_list.append("\t\t\t<enabled>%d</enabled>" % handler.get_enabled())

        for index in self._plot_maps:
            xml_list.extend(self.get_plot_map_config(index).to_xml())

        for settings in self.get_go_browser_map_settings():
            xml_list.append("\t\t\t<go_browser_map>")
            for key, value in settings.items():
                xml_list.append("\t\t\t\t<%s>%s</%s>" % (key, value, key))
            xml_list.append("\t\t\t</go_browser_map>")

        xml_list.append("\t\t</plot_map_config>")    
        xml_list.append("\t</plotnode_config>")
        return xml_list

    def custom_plot_to_xml(self):
        xml_list = self.to_xml()
        xml_swap = xml_list[:-1]
        extend_list = []
        
        ip_dict = self.get_input_port_descriptions()
        if ip_dict:
            extend_list.append("\t\t<input_ports_used>")
            for key in ip_dict:
                extend_list.append("\t\t\t<ip>")
                extend_list.append("\t\t\t\t<name>" + key + "</name>")
                extend_list.append("\t\t\t\t<type>" + ip_dict[key][0] + "</type>")
                extend_list.append("\t\t\t\t<description>" + ip_dict[key][1] + "</description>")                
                extend_list.append("\t\t\t</ip>")
            extend_list.append("\t\t</input_ports_used>")
            
        xml_swap.extend(extend_list)
        xml_swap.append(xml_list[-1])
        
        return xml_swap
        
    def remove_input_port(self, port_name):
        self.remove_port(self.getInputPortByName(port_name))

    def get_input_port_descriptions(self):
        """Get input ports datatype and description as dictionary by name.

        Example:
        ports_in_dictionary = {'portname1':('matrix', 'matrix data'),
                               'portname2':('scalar', 'color of marks')} 
        """
        ports_in_dictionary = {}

        for port in self.getInputPorts():
            ports_in_dictionary[port.getVariableName()] = (port.getDataType(), port.getDescription())

        return ports_in_dictionary

    def get_all_emulated_input_ports(self):
        """Gather all emulated ports from multiinputports."""
        return self._multi_input.get_emulated_input_ports()

    def get_plot_map_handler(self):
        return self._plot_map_handler

    def set_widget_settings(self, x_coord, y_coord, widget_settings):
        """Save settings to widget at coordinate."""
        self.get_widget_config(x_coord, y_coord).set_settings(widget_settings)

    def get_widget_settings(self, x_coord, y_coord):
        """Get settings for widget at coordinate."""
        return self.get_widget_config(x_coord, y_coord).get_settings()
    
    def update_plots(self, x_coord, y_coord, number_of_plots):
        """Create plot configs for number of plots at widget at coordinate.

        This method must be called every time the gui wishes to
        add/remove a plot to/from a specific widget so that the class
        can ready the data structures needed to save future settings.
        """
        self.get_widget_config(x_coord, y_coord).update_plots(number_of_plots)

    def get_number_of_plots(self, x_coord, y_coord):
        """Get number of plots at widget at coordinate."""
        return self.get_widget_config(x_coord, y_coord).get_number_of_plots()

    def update_widgets(self, horisontal_widgets, vertical_widgets):
        """Create widget configs for given layout.

        This method must be called every time the gui wishes to
        add/remove one or more widgets so that the class can ready the
        data structures needed to save future settings.
        """
        self._horisontal_widgets = horisontal_widgets
        self._vertical_widgets = vertical_widgets
        
        temp_dict = {}
        for x in range(self._horisontal_widgets):
            for y in range(self._vertical_widgets):
                if not self._widgets.has_key((x, y)):
                    temp_dict[(x, y)] = _WidgetConfig(x, y)
                else:
                    temp_dict[(x, y)] = self._widgets[(x, y)]
        self._widgets = temp_dict.copy()

    def set_marker_settings(self, x_coord, y_coord, index, marker_settings):
        """Save marker settings to plot at index in widget at coordinates."""
        plot_config = self.get_plot_config(x_coord, y_coord, index)
        plot_config.set_marker_settings(marker_settings)

    def get_marker_settings(self, widget_x_coord, widget_y_coord, index):
        plot_config = self.get_plot_config(widget_x_coord, widget_y_coord,
                                            index)
        return plot_config.get_marker_settings()

    def get_horisontal_widgets(self):
        return self._horisontal_widgets    

    def get_vertical_widgets(self):
        return self._vertical_widgets

    def update_plot_maps(self, number_of_plot_maps):
        """Create plotmaps before use."""
        temp_dict = {}
        for index in range(number_of_plot_maps):
            if not self._plot_maps.has_key(index):
                temp_dict[index] = PlotMapConfig(index)
            else:
                temp_dict[index] = self._plot_maps[index]
        self._plot_maps = temp_dict.copy()
        
    def set_plot_map_settings(self, index, settings):
        """Saves settings for plot map of index."""
        plot_map_config = self.get_plot_map_config(index)
        plot_map_config.set_settings(settings)

    def get_plot_map_settings(self):
        """Retrieves list of dictionaries with settings for all plot maps."""
        return map(PlotMapConfig.get_settings, self._plot_maps.values())

    def get_widget_config(self, *key):
        """Get config for widget in zero-indexed position (x, y)."""
        return self._widgets[key]

    def get_plot_config(self, widget_x_coord, widget_y_coord, index):
        _widget_config = self.get_widget_config(widget_x_coord, widget_y_coord)
        return _widget_config.get_plot_config(index)

    def get_plot_map_config(self, index):
        return self._plot_maps[index]

    def get_go_browser_map_settings(self):
        """Get list with settings/dict for each mapping to a GO browser."""
        return self._go_browser_maps

    def set_go_browser_map_settings(self, maps):
        """Set list with settings/dict for each mapping to a GO browser."""
        self._go_browser_maps = maps
    

class _WidgetConfig:
    def __init__(self, x_coord, y_coord):
        self._x_coord = x_coord
        self._y_coord = y_coord

        self._settings = {}
        self._plots = {}
        self._number_of_plots = 0

    def from_xml_element(self, root):
        settings = {}
        for tag in root.getchildren():
            if tag.tag not in ("plot_config", "x_coord", "y_coord"):
                settings[tag.tag] = tag.text or ""

        self.update_plots(len(root.findall("plot_config")))

        for plot_config in root.findall("plot_config"):
            index = int(plot_config.find("index").text)
            config = self.get_plot_config(index)
            config.from_xml_element(plot_config)

        self.set_settings(settings)

    def set_settings(self, widget_settings):
        self._settings = widget_settings.copy()

    def get_settings(self):
        return self._settings

    def update_plots(self, number_of_plots):
        """Updates self._plots dictionary and self._number_of_plots"""
        temp_dict = {}
        for index in range(number_of_plots):
            if not self._plots.has_key(index):
                temp_dict[index] = PlotConfig(index)
            else:
                temp_dict[index] = self._plots[index]
        self._plots = temp_dict.copy()

        self._number_of_plots = len(self._plots)

    def get_plot_config(self, index):
        if self._plots.has_key(index):
            return self._plots[index]
        return None

    def to_xml(self):
        self.__update_standard_parameters()

        xml_list = []
        xml_list.append("\t\t<widget_config>")

        for setting, value in self._settings.items():
            xml_list.append("\t\t\t<%s>%s</%s>" % (str(setting), str(value),
                                                   str(setting)))

        for plot in self._plots.values():
            xml_list.extend(plot.to_xml())

        xml_list.append("\t\t</widget_config>")
        return xml_list

    def __update_standard_parameters(self):
        self._settings["x_coord"] = str(self._x_coord)
        self._settings["y_coord"] = str(self._y_coord)

    def get_number_of_plots(self):
        return self._number_of_plots
    
    def get_widget_x_coord(self):
        return self._x_coord

    def get_widget_y_coord(self):
        return self._y_coord


class PlotConfig:
    def __init__(self, index):
        self._settings = {}
        self._marker_settings = {}
        self._index = index

    def from_xml_element(self, root):
        marker_settings = root.find("plot_marker_settings")
        if marker_settings:
            self.set_marker_settings(_parse_as_settings(marker_settings))
        
        for tag in root.getchildren():
            if tag.tag not in ("plot_marker_settings", "index"):
                self.set_setting(tag.tag, tag.text or "")

    def set_settings(self, plot_settings):
        self._settings = plot_settings.copy()

    def get_settings(self):
        return self._settings

    def set_setting(self, key, value):
        self._settings[key] = value

    def get_setting(self, key):
        return self._settings[key]

    def set_marker_setting(self, key, value):
        self._marker_settings[key] = value

    def get_marker_setting(self, key):
        return self._marker_settings[key]

    def set_marker_settings(self, marker_settings):
        self._marker_settings = marker_settings.copy()

    def get_marker_settings(self):
        return self._marker_settings

    def to_xml(self):
        if self._settings:
            self.__update_standard_parameters()
        
            xml_list = []
            xml_list.append("\t\t\t<plot_config>")

            settings_keys = self._settings.keys()
            settings_keys.sort()

            for setting in settings_keys:
                xml_list.append("\t\t\t\t<" + str(setting) + ">" +
                                str(self._settings[setting]) +
                                "</" + str(setting) + ">")

            if self._marker_settings:
                xml_list.append("\t\t\t\t<plot_marker_settings>")
                marker_setting_keys = self._marker_settings.keys()
                marker_setting_keys.sort()
                for setting in marker_setting_keys:
                    xml_list.append("\t\t\t\t\t<" + str(setting) + ">" +
                                    str(self._marker_settings[setting]) +
                                    "</" + str(setting) + ">")
                xml_list.append("\t\t\t\t</plot_marker_settings>")
            xml_list.append("\t\t\t</plot_config>")

            return xml_list
        else:
            return []

    def __update_standard_parameters(self):
        self._settings["index"] = str(self._index)


    def get_index(self):
        return self._index


