# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from thirdparty import ElementTree

import math

from qt import QHBox, QSpinBox, QLabel, QObject, SIGNAL, QComboBox, QCheckBox

connect = QObject.connect

class Choice:
    def __init__(self, ident, label):
        self._ident = ident
        self._label = label
        self._listeners = []

    def get_id(self):
        return self._ident

    def get_label_text(self):
        return self._label

    def create_xml(self, extra={}):
        """Used from subclasses to generate xml."""
        attributes = {"ident" : self._ident}
        attributes.update(extra)
        return ElementTree.Element("choice", attributes)

    def add_listener(self, listener):
        self._listeners.append(listener)

    def remove_listener(self, listener):
        self._listeners.remove(listener)

    def notify_listeners(self):
        """Called from subclasses when value or choices change."""
        for listener in self._listeners:
            listener(self.get_id())

    def update_ports(self, ports):
        """Overridable by subclasses.

        Called by plotsystem to notify port changes.
        """
        pass


class IntegerChoiceController:
    def __init__(self, choice, spinbox):
        self._choice = choice
        self._spinbox = spinbox
        connect(self._spinbox, SIGNAL("valueChanged(int)"), choice.select)
        choice.add_listener(self.update_values)
        self.update_values(None)

    def disconnect(self):
        self._choice.remove_listener(self.update_values)

    def update_values(self, ident):
        self._spinbox.setMinValue(self._choice.get_min())
        self._spinbox.setMaxValue(self._choice.get_max())
        self._spinbox.setValue(self._choice.current_value())
        

class IntegerChoice(Choice):
    def __init__(self, ident=None, label=None, minimum=0, maximum=0, default=None):
        Choice.__init__(self, ident, label)
        self._min = minimum
        self._max = maximum

        if default:
            self._value = default
        else:
            self._value = minimum

        self._initial_value = None

    def get_min(self):
        return self._min

    def get_max(self):
        return self._max

    def select(self, value):
        self._initial_value = None
        old_value = self._value
        self._value = value

        if value != old_value:
            self.notify_listeners()
            
    def current_value(self):
        return self._value

    def set_range(self, minimum, maximum):
        old_values = self._min, self._max, self._value
        
        self._min = minimum
        self._max = maximum

        if self._initial_value != None:
            self._value = self._initial_value

        self._clip_value()
        
        if old_values != (self._min, self._max, self._value):
            self.notify_listeners()

    def _clip_value(self):
        if self._value < self._min:
            self._value = self._min
        if self._value > self._max:
            self._value = self._max

    def to_xml_element(self):
        if self._initial_value == None:
            return self.create_xml({"value" : str(self._value)})
        else:
            return self.create_xml({"value" : str(self._initial_value)})

    def from_xml_element(self, root):
        self._initial_value = int(root.get("value"))
        self._value = self._initial_value
        self._clip_value()

    def create_controller_widget(self, parent, target=None):
        if target:
            return IntegerChoiceController(self, target)
        else:
            box = QHBox(parent)
            QLabel(self.get_label_text(), box)
            spinbox = QSpinBox(box)
            return IntegerChoiceController(self, spinbox)


class ListChoiceController:
    def __init__(self, choice, combo):
        self._choice = choice
        self._combo = combo
        connect(self._combo, SIGNAL("activated(const QString&)"), self._changed)
        choice.add_listener(self.update_values)
        self.update_values(None)

    def disconnect(self):
        self._choice.remove_listener(self.update_values)

    def _changed(self, qstring):
        text = str(qstring)
        if text != self._choice.current_text():
            self._choice.select(text)

    def update_values(self, ident):
        texts = self._choice.get_texts()
        self._combo.clear()
        self._combo.insertStrList(texts)
        if texts:
            self._combo.setCurrentItem(texts.index(self._choice.current_text()))


class ListChoice(Choice):
    def __init__(self, ident, label, texts, values):
        Choice.__init__(self, ident, label)
        self._texts = texts or []
        self._values = values or []
        self._selected = None
        self._initial_value = None
        
        if texts:
            self._selected = texts[0]

    def get_texts(self):
        return self._texts

    def get_values(self):
        return self._values

    def current_text(self):
        return self._selected

    def current_value(self):
        try:
            return self._values[self._texts.index(self._selected)]
        except ValueError, e:
            return None

    def select(self, text):
        old = self._selected
        self._selected = text
        self._initial_value = None

        if text != old:
            self.notify_listeners()

    def set_choices(self, texts, values):
        texts = texts or []
        values = values or []

        if self._initial_value != None:
            self._selected = self._initial_value
            
        old_value = self.current_value()
        
        self._texts = texts
        self._values = values

        if not texts:
            self._selected = None
        elif self._selected not in texts:
            self._selected = texts[0]

        self.notify_listeners()

    def from_xml_element(self, root):
        value = root.get("value")
        self._initial_value = value
        if value in self._texts:
            self._selected = value
        elif self._texts:
            self._selected = self._texts[0]
        else:
            self._selected = None

    def to_xml_element(self):
        if self._initial_value != None:
            return self.create_xml({"value" : self._initial_value})
        elif self._selected != None:
            return self.create_xml({"value" : self._selected})
        else:
            return self.create_xml()

    def create_controller_widget(self, parent, combo=None):
        if combo:
            return ListChoiceController(self, combo)
        else:
            box = QHBox(parent)
            QLabel(self.get_label_text(), box)
            combo = QComboBox(box)
            return ListChoiceController(self, combo)


class PortChoiceController:
    def __init__(self, choice, combo):
        self._choice = choice
        self._combo = combo
        self._ports = []
        connect(self._combo, SIGNAL("activated(int)"), self._changed)
        choice.add_listener(self.update_values)
        self.update_values(None)

    def disconnect(self):
        self._choice.remove_listener(self.update_values)

    def _changed(self, int):
        port = self._ports[int]
        if port != self._choice.current_port():
            self._choice.select(port)

    def update_values(self, ident):
        self._ports = self._choice.get_values()
        self._combo.clear()
        self._combo.insertStrList(self._choice.get_texts())
        self._combo.setCurrentItem(self._ports.index(self._choice.current_port()))


class PortChoice(Choice):
    def __init__(self, ident=None, label=None):
        Choice.__init__(self, ident, label)
        self._current = None
        self._ports = []
        self._initial_port_id = None

    def update_ports(self, ports):
        self._ports = ports

        if self._initial_port_id != None:
            for port in ports:
                if port.getId() == self._initial_port_id:
                    self._current = port
                    break
            self._initial_port_id = None
            
        if self._current not in ports:
            self._current = None

        self.notify_listeners()

    def get_texts(self):
        names = ["No source"]
        for port in self._ports:
            names.append(port.getName())

        return names

    def get_values(self):
        return [None] + self._ports

    def select(self, port):
        old_port = self._current
        self._current = port
        if old_port != port:
            self.notify_listeners()

    def current_port(self):
        return self._current

    def current_value(self):
        if self._current and self._current.getData():
            return self._current.getData().getData()

    def data_instance(self):
        if self._current:
            return self._current.getData()

    def to_xml_element(self):
        if self._initial_port_id:
            return self.create_xml({"port_id" : self._initial_port_id})
        elif self._current:
            return self.create_xml({"port_id" : self._current.getId()})
        else:
            return self.create_xml()
    
    def from_xml_element(self, root):
        if root.get("port_id"):
            if self._ports:
                for port in self._ports:
                    if port.getId() == root.get("port_id"):
                        self._current = port
            else:
                self._initial_port_id = root.get("port_id")

        if self._current:
            self.notify_listeners()

    def create_controller_widget(self, parent, combo=None):
        if combo:
            return PortChoiceController(self, combo)
        else:
            box = QHBox(parent)
            QLabel(self.get_label_text(), box)
            combo = QComboBox(box)
            return PortChoiceController(self, combo)


class BooleanChoiceController:
    def __init__(self, choice, checkbox):
        self._choice = choice
        self._checkbox = checkbox
        connect(self._checkbox, SIGNAL("toggled(bool)"), self._changed)
        choice.add_listener(self._update_values)
        self._checkbox.setChecked(choice.current_value())

    def disconnect(self):
        self._choice.remove_listener(self._update_values)

    def _changed(self, new_value):
        if self._choice.current_value() != new_value:
            self._choice.select(new_value)

    def _update_values(self, ident):
        self._checkbox.setChecked(self._choice.current_value())


class BooleanChoice(Choice):
    def __init__(self, ident, label, initial_value=False):
        Choice.__init__(self, ident, label)
        self._value = initial_value

    def current_value(self):
        return self._value

    def select(self, value):
        value = bool(value)
        old_value = self._value
        self._value = value

        if old_value != value:
            self.notify_listeners()

    def from_xml_element(self, root):
        self._value = root.get("value") == "True"

    def to_xml_element(self):
        return self.create_xml({"value" : str(self._value)})

    def create_controller_widget(self, parent, widget=None):
        if widget:
            return BooleanChoiceController(self, widget)
        else:
            box = QHBox(parent)
            QLabel(self.get_label_text(), box)
            checkbox = QCheckBox(box)
            return BooleanChoiceController(self, checkbox)
                

