# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from GUI.Plots.plot2d.imageplot import matrix_to_qimage
from Library.Nodes.Plot.develplotnode import IndexSet

from Numeric import ravel, transpose
import Numeric

from qt import QPen, QBrush, QColor, Qt
from qwt import QwtPlotCurve, QwtSymbol, QwtCurve
from itertools import izip
from sets import Set

class DummyPainter:
    """Do-nothing-painter."""
    def __init__(self):
        pass
    
    def disconnect(self):
        """Remove from any listeners."""
        pass

    def update_painter(self):
        """Paint into Qwt plot."""
        pass
    

class PlotComponent:
    """Paintable plot element to be added by UserPlots."""
    
    def __init__(self, name=None):
        self._name = name
        self._selection = []
        self._listeners = []
        self._interactive_selection = True
        self._disabled_modes = Set()

    def get_name(self):
        """Get the name of this component.

        PlotComponents with the same name imply representing the same
        dataset. UserPlot uses the name of components to transfer
        selection state between old components to new components with
        same name when they are replaced during replotting.
        """
        return self._name

    def set_selection(self, selection):
        """Set current selection state.

        selection should be same structure as returned by
        get_selection().
        """
        self._selection = selection

    def get_selection(self):
        """Get current selection state."""
        return self._selection

    # used by painters

    def add_repaint_listener(self, listener):
        """Add listener to notify when component has changed."""
        self._listeners.append(listener)

    def remove_repaint_listener(self, listener):
        """Remove existing listener."""
        self._listeners.remove(listener)

    # called by users / plot-designers

    def disable_interactive_selection(self):
        """Make component not respond to select_area()."""
        self._interactive_selection = False

    def disable_selection_mode(self, mode_key):
        """Disable a named selection mode for this component.

        The selection modes available, and the type of key they use is
        up to each component. It's implied that when a selection mode
        is disabled, the component no longer highlights elements in
        that mode, neither on interactive selection nor mapping.
        """
        self._disabled_modes.add(mode_key)

    def enable_selection_mode(self, mode_key):
        """Reenable a previously disabled selection mode."""
        self._disabled_modes.remove(mode_key)

    # usable by subclasses

    def _has_interactive_selection(self):
        """Check if selections are caused interactively or through mapping."""
        return self._interactive_selection

    def notify_painters(self):
        """Signal all registered painters that they should repaint."""
        for listener in self._listeners:
            listener()

    def is_selection_mode_enabled(self, mode_key):
        """Query a given selection mode.

        All selection modes are initially enabled, and can be disabled
        through PlotComponent.disable_selection_mode(...).
        """
        return not mode_key in self._disabled_modes

    # should be overridden by subclasses

    def get_mappable_index_sets_within(self, x1, x2, y1, y2):
        """Return list of index sets of indices within area."""
        return []

    def create_painter(self, qwt_plot):
        """Create painter responsible for drawing into Qwt plot on updates."""
        return DummyPainter()

    def select_area(self, x1, y1, x2, y2):
        """Respond to selection being made.

        A usual response is to change internal selection state and
        notify painter(s).
        """
        pass

    def mapped_select(self, indices, length):
        """Respond to indices being selected in other plots.

        To enable mapping/data brushing, any component of same length
        should select the same indices as other components of that
        length.
        """
        pass


class MultiLinePainter:
    """Painter for MultiLineComponent."""
    def __init__(self, component, qwt_plot):
        self._component = component
        self._plot = qwt_plot
        self._curves = []
        self._component.add_repaint_listener(self.update_painter)
        self.update_painter()

    def disconnect(self):
        """Disconnect listening when no longer used."""
        self._component.remove_repaint_listener(self.update_painter)

    def update_painter(self):
        """Update plot with new line data."""
        colors = [Qt.blue, Qt.cyan, Qt.magenta, Qt.darkRed , Qt.darkGreen,
                  Qt.darkBlue, Qt.darkCyan, Qt.darkMagenta, Qt.red,
                  Qt.darkYellow, Qt.green]
        for key in self._curves:
            self._plot.removeCurve(key)
        self._curves = []

        x_points, y_points = self._component.get_line_points()

        selected_lines = self._component.get_selected_lines()

        if not selected_lines:
            for i, (x, y) in enumerate(izip(x_points, y_points)):
                curve = QwtPlotCurve(self._plot)
                curve.setData(x, y)
                if not selected_lines:
                    pen = QPen(colors[i % len(colors)])
                elif i in selected_lines:
                    pen = QPen(Qt.red)
                else:
                    pen = QPen(Qt.lightGray)
                curve.setPen(pen)
                self._curves.append(self._plot.insertCurve(curve))
        else:
            # first do gray lines...
            for i, (x, y) in enumerate(izip(x_points, y_points)):
                if i in selected_lines:
                    continue
                curve = QwtPlotCurve(self._plot)
                curve.setData(x, y)
                pen = QPen(Qt.lightGray)
                curve.setPen(pen)
                self._curves.append(self._plot.insertCurve(curve))

            # ...then do selected lines
            for i in selected_lines:
                curve = QwtPlotCurve(self._plot)
                curve.setData(x_points[i], y_points[i])
                pen = QPen(Qt.red)
                curve.setPen(pen)
                self._curves.append(self._plot.insertCurve(curve))

        # then one curve for selected values
        curve = QwtPlotCurve(self._plot)
        x_points, y_points = self._component.get_selected_points()
        curve.setData(x_points, y_points)
        curve.setStyle(QwtCurve.NoCurve)
        brush = QBrush(Qt.red)
        symbol = QwtSymbol()
        symbol.setStyle(QwtSymbol.Ellipse)
        symbol.setSize(4)
        symbol.setBrush(brush)
        
        curve.setSymbol(symbol)
        self._curves.append(self._plot.insertCurve(curve))

        self._plot.set_dirty()


class PointsPainter:
    """Scatter plot painter for PointsComponent."""
    def __init__(self, component, qwt_plot):
        self._component = component
        self._plot = qwt_plot
        self._curves = []
        self._component.add_repaint_listener(self.update_painter)
        self._normal_curve = None
        self.update_painter()

    def disconnect(self):
        """Disconnect listening when no longer used."""
        self._component.remove_repaint_listener(self.update_painter)

    def get_tooltip_text(self, curve_key, index):
        """Get tooltip for point on curve."""
        if not self._component.get_selection():
            return self._component.get_tooltip_text(index)

        # the index has to be translated to get the correct index, if
        # the curve had been selected in.

        if curve_key == self._normal_curve:
            indices = Set(range(len(self._component.get_points()[0])))
            indices.difference_update(self._component.get_selection())
            indices = list(indices)
            indices.sort()
            return self._component.get_tooltip_text(indices[index])
        else:
            return self._component.get_tooltip_text(self._component.get_selection()[index])

    def update_painter(self):
        """Add points to plot according to coordinates in component."""
        for key in self._curves:
            self._plot.remove_tooltip_curve(key)
        self._curves = []

        # curve for not selected values
        curve = QwtPlotCurve(self._plot)
        x_points, y_points = self._component.get_non_selected_points()
        curve.setData(x_points, y_points)
        curve.setStyle(QwtCurve.NoCurve)
        brush = QBrush(QColor(*self._component.get_color()))
        symbol = QwtSymbol()
        if self._component.get_draw_only_selected():
            # for some plots, points should only be displayed when
            # they are selected (density plot), but we still want them
            # inside the plot so we can use tooltips.
            symbol.setStyle(QwtSymbol.None)
        else:
            symbol.setStyle(QwtSymbol.Ellipse)
        symbol.setSize(self._component.get_size())
        symbol.setBrush(brush)
        
        curve.setSymbol(symbol)
        key = self._plot.insert_tooltip_curve(curve, self)
        self._curves.append(key)
        self._normal_curve = key

        # then one curve for selected values
        curve = QwtPlotCurve(self._plot)
        x_points, y_points = self._component.get_selected_points()
        curve.setData(x_points, y_points)
        curve.setStyle(QwtCurve.NoCurve)
        brush = QBrush(Qt.red)
        pen = QPen(Qt.darkRed)
        symbol = QwtSymbol()
        symbol.setStyle(QwtSymbol.Ellipse)
        symbol.setSize(self._component.get_size())
        symbol.setBrush(brush)
        symbol.setPen(pen)
        
        curve.setSymbol(symbol)
        self._curves.append(self._plot.insert_tooltip_curve(curve, self))
        self._plot.set_dirty()


class PointsComponent(PlotComponent):
    """Plot component for a set of x/y-coordinates.

    Usable for scatter plot like plots.
    """
    def __init__(self, x_coords, y_coords, name=None):
        PlotComponent.__init__(self, name)
        self._x_coords = []
        self._y_coords = []

        self._draw_only_selected = False
        
        if x_coords and y_coords:
            self._x_coords = Numeric.array(x_coords)
            self._y_coords = Numeric.array(y_coords)

        # gray
        self._color = 160, 160, 164
        self._size = 4
        self._tooltips = []

    def set_selection(self, selection):
        PlotComponent.set_selection(self, selection)
        self.notify_painters()

    def get_points(self):
        return self._x_coords, self._y_coords

    def get_non_selected_points(self):
        """Get x and y-coordinates for non-selected points."""
        if not self.get_selection():
            return self._x_coords, self._y_coords

        selected_x = []
        selected_y = []

        length = len(self._x_coords)

        bad_array = Numeric.array(self.get_selection())
        filtered_bad = Numeric.take(bad_array,
                                    Numeric.nonzero(bad_array < length))
        good = Numeric.ones(length)
        Numeric.put(good, filtered_bad, Numeric.zeros(length))
        good_indices = Numeric.nonzero(good)

        selected_x = Numeric.take(self._x_coords, good_indices)
        selected_y = Numeric.take(self._y_coords, good_indices)

        return selected_x, selected_y

    def set_tooltip_texts(self, text_list):
        """Set list of strings to use for tooltips in plot.

        If a static list is not enough, get_tooltip_text() can
        be overridden for a more dynamic approach.
        """
        self._tooltips = text_list

    def get_tooltip_text(self, index):
        """Get tooltip string describing 0-indexed value index.

        Override this method if you need a more dynamic approach than
        static strings.
        """
        if index < len(self._tooltips):
            return self._tooltips[index]

    def get_selected_points(self):
        """Get all x/y-coordinates for selected points."""
        selected_x = []
        selected_y = []

        max_index = len(self._x_coords) - 1
        a = Numeric.array(self.get_selection())
        indices = Numeric.take(a, Numeric.nonzero(a <= max_index))

        selected_x = Numeric.take(self._x_coords, indices)
        selected_y = Numeric.take(self._y_coords, indices)

        return selected_x, selected_y

    def create_painter(self, qwt_plot):
        """Create painter to be responsible for drawing into qwt_plot."""
        return PointsPainter(self, qwt_plot)

    def get_mappable_index_sets_within(self, x1, x2, y1, y2):
        """Get indices of points within area as list of IndexSets."""
        return [IndexSet(self._get_indices_within(x1, x2, y1, y2),
                         len(self._x_coords))]

    def _get_indices_within(self, x1, x2, y1, y2):
        """Get indices of points within area."""
        x = self._x_coords
        y = self._y_coords
        return Numeric.nonzero((x>=x1) & (x<=x2) & (y>=y1) & (y<=y2))

    def select_area(self, x1, x2, y1, y2):
        """Highlight points within area to respond to mouse selection."""
        if self._has_interactive_selection():
            self.set_selection(self._get_indices_within(x1, x2, y1, y2))
        else:
            self.set_selection([])

    def get_color(self):
        """Get color set for points."""
        return self._color

    def set_color(self, red, green, blue):
        """Set color to use for points."""
        self._color = red, green, blue

    def get_size(self):
        """Get size set for each point."""
        return self._size

    def set_size(self, size):
        """Set size to use when painting each point.

        Size should ideally be represented in a device independent
        way, but is currently represented by pixel size.
        """
        self._size = size

    def mapped_select(self, indices, length):
        """Highlight points when length matches data set."""
        if length == len(self._x_coords):
            self.set_selection(indices)

    def set_draw_only_selected(self, value):
        """Make this component only be visible when points are selected.

        This is a useful feature when combining it with an ImageComponent
        for e.g. a density plot, allowing points to be used to show
        highlighted areas on the density plot.
        """
        self._draw_only_selected = value

    def get_draw_only_selected(self):
        """Get state as set by PointsComponent.set_draw_only_selected()."""
        return self._draw_only_selected


class MultiLineComponent(PlotComponent):
    """Component for drawing multiple lines, optionally with provided x."""

    # selection modes
    OBJECT_MODE, LINES_MODE = range(2)
    
    def __init__(self, dim_1, dim_2=None, name=None):
        PlotComponent.__init__(self, name)
        if dim_2:
            self._y_coords = dim_2
            self._x_coords = dim_1
        else:
            self._y_coords = dim_1
            self._x_coords = [range(1, 1+len(line)) for line in dim_1]
        self._selections = []
        self._selected_lines = []

    def get_line_points(self):
        """Get all x and y-coordinates for all lines.

        Gets the tuple ([line_1_x, line_2_x, ...], [line_1_y, line_2_y, ...]).
        """
        return self._x_coords, self._y_coords

    def select_area(self, x1, x2, y1, y2):
        """Select lines and indices on each line within area."""
        if self._has_interactive_selection():
            self._selections = self._get_indices_within(x1, x2, y1, y2)
        else:
            self._selections = []
            
        index_set = IndexSet([], len(self._y_coords[0]))
        for objects in self._selections:
            if objects:
                index_set.add_set(IndexSet(objects, len(self._y_coords[0])))

        self._selected_lines = []
        for i, line in enumerate(self._selections):
            if line:
                self._selected_lines.append(i)

        self.notify_painters()

    def _get_indices_within(self, x1, x2, y1, y2):
        """Get a list of indices within the area for each line."""
        selections = []
        for x_line, y_line in izip(self._x_coords, self._y_coords):
            x = Numeric.array(x_line)
            y = Numeric.array(y_line)
            selections.append(Numeric.nonzero((x>=x1) & (x<=x2) & (y>=y1) &
                                              (y<=y2)))
        return selections

    def get_selected_lines(self):
        """Get lines indices selected by mapped_select or select_area()."""
        return self._selected_lines

    def get_selected_points(self):
        """Get (x_coords, y_coords) for all selected points."""
        x_selected, y_selected = [], []
        x_points, y_points = self.get_line_points()

        for i, line in enumerate(self._selections):
            x_selected.extend(map(lambda x : x_points[i][x], line))
            y_selected.extend(map(lambda x : y_points[i][x], line))
        return x_selected, y_selected

    def create_painter(self, qwt_plot):
        """Create painter drawing lines into qwt_plot."""
        return MultiLinePainter(self, qwt_plot)

    def get_mappable_index_sets_within(self, x1, x2, y1, y2):
        object_indices = Set()
        line_indices = []
        for i, indices in enumerate(self._get_indices_within(x1, x2, y1, y2)):
            if indices:
                object_indices.update(indices)
                line_indices.append(i)

        index_sets = []
        if self.is_selection_mode_enabled(self.OBJECT_MODE):
            index_sets.append(IndexSet(list(object_indices),
                                       len(self._y_coords[0])))
        if self.is_selection_mode_enabled(self.LINES_MODE):
            index_sets.append(IndexSet(line_indices, len(self._y_coords)))

        return index_sets

    def mapped_select(self, indices, length):
        if length == len(self._y_coords):
            if self.is_selection_mode_enabled(self.LINES_MODE):
                self._selected_lines = indices
                self.notify_painters()
        if length == len(self._y_coords[0]):
            if self.is_selection_mode_enabled(self.OBJECT_MODE):
                self._selections = [indices for i in range(len(self._y_coords))]
                self.notify_painters() 


class SingleLineComponent(MultiLineComponent):
    """Special case of MultiLine.

    Added for saner syntax and better behaviour concerning LINES_MODE
    selection.
    """
    def __init__(self, dim_1, dim_2=None, name=None):
        """Create single line based on one or two lists of values."""

        # and-expression to keep dim_2 as None, or place it in a list,
        # as MultiLineComponent expects a matrix, not a list
        MultiLineComponent.__init__(self, [dim_1], dim_2 and [dim_2], name)

        # we disable LINES_MODE as we only have one line
        self.disable_selection_mode(self.LINES_MODE)
        
        
class ImageComponent(PlotComponent):

    def __init__(self, matrix, image_area = None, color_map=None, name=None):
        PlotComponent.__init__(self, name)
        self._matrix = matrix
        self._color_map = color_map

        self._image = matrix_to_qimage(self._matrix, self._color_map)

        self._selected_pixels_by_row = []
        self._selected_pixels_by_column = []
        self._selected_rows = []
        self._selected_columns = []    
        self._image_area = image_area
        self._selection_disabled = False

        if not self._image_area:
            height, width = Numeric.shape(self._matrix)
            self._image_area = (0.5, width+0.5, 0.5, height + 0.5)

    def get_image(self):
        return self._image

    def create_painter(self, qwt_plot):
        return ImagePainter(self, qwt_plot)

    def get_mappable_index_sets_within(self, x1, x2, y1, y2):
        if self._selection_disabled:
            return []
        
        row_length, column_length = Numeric.shape(self._matrix)        
        total_size = row_length * column_length

        if x1 > column_length:
            x1, x2 = 0, 0
        else:
            x1 = int(round(x1)) - 1
            x2 = min(int(round(x2)), column_length)

        if y1 > row_length:
            y1, y2 = 0, 0
        else:            
            y1 = int(round(y1)) - 1
            y2 = min(int(round(y2)), row_length)

        rows = range(y1, y2)
        columns = range(x1, x2)

        # remove negative values
        row_indices = filter(lambda value : value >= 0, rows)
        column_indices = filter(lambda value : value >= 0, columns)
        
        row_matrix = Numeric.reshape(range(total_size), (row_length,
                                                         column_length))
        
        column_matrix = []
        for row in range(row_length):
            col = []
            for column in range(column_length):
                col.append(column*row_length + row)
            column_matrix.append(col)

        pixels_by_row_indices = ravel(transpose(row_matrix[y1:y2])[x1:x2])
        pixels_by_column_indices = ravel(transpose(column_matrix[y1:y2])[x1:x2])

        return (IndexSet(pixels_by_row_indices, total_size),
                IndexSet(pixels_by_column_indices, total_size),
                IndexSet(row_indices, row_length),
                IndexSet(column_indices, column_length))

    def select_area(self, x1, x2, y1, y2):
        if self._selection_disabled:
            self.notify_painters()
            return
        
        index_sets = self.get_mappable_index_sets_within(x1, x2, y1, y2)
        
        self._selected_pixels_by_row = index_sets[0].get_indices()
        self._selected_pixels_by_column = index_sets[1].get_indices()
        self._selected_rows = index_sets[2].get_indices()
        self._selected_columns = index_sets[3].get_indices()

        self.notify_painters()

    def get_selected_pixels_by_row(self):
        return self._selected_pixels_by_row

    def get_selected_pixels_by_column(self):
        return self._selected_pixels_by_column

    def get_selected_rows(self):
        return self._selected_rows

    def get_selected_columns(self):
        return self._selected_columns

    def mapped_select(self, indices, length):
        if self._selection_disabled:
            self.notify_painters()
            return
        
        rows, columns = Numeric.shape(self._matrix)
        total_length = rows*columns

        selected = []
        if total_length == length:
            self._selected_pixels_by_row = indices
            self._selected_pixels_by_column = indices
            selected = indices
        else:
            self._selected_pixels_by_row = []
            self._selected_pixels_by_column = []                     

        if rows == length:
            self._selected_rows = indices
            selected = indices
        else:
            self._selected_rows = []
            
        if columns == length:
            self._selected_columns = indices
            selected = indices
        else:
            self._selected_columns = []

        self.notify_painters()

    def get_image_area(self):
        return self._image_area

    def get_selection(self):
        return (self._selected_pixels_by_row,
                self._selected_pixels_by_column,
                self._selected_rows,
                self._selected_columns)

    def set_selection(self, selections):
        self._selected_pixels_by_row = selections[0]
        self._selected_pixels_by_column = selections[1]
        self._selected_rows = selections[2]
        self._selected_columns = selections[3]

    def set_selection_disabled(self, value):
        self._selection_disabled = value

    def get_selection_disabled(self):
        return self._selection_disabled


class ImagePainter:

    def __init__(self, component, qwt_plot):
        self._component = component
        self._plot = qwt_plot
        self._image = None
        self._component.add_repaint_listener(self.update_painter)
        self.update_painter()

    def disconnect(self):
        if self._image:
            self._plot.remove_image(self._image)
        self._component.remove_repaint_listener(self.update_painter)

    def update_painter(self):
        self._image = self._component.get_image()
        area = self._component.get_image_area()
        selection = self._component.get_selection()
        self._plot.add_image(self._image, area, selection)
        self._plot.set_dirty()
