# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from System.Data.Node import Node
from Library.Util.Config import Config
from thirdparty import ElementTree
from Library.Nodes.Plot.choice import IntegerChoice, ListChoice, PortChoice

import glob
import os
import traceback
import sys

from qtui import QWidgetFactory
from qwt import QwtPlotCurve
from inspect import isclass
from sets import Set

# domain exceptions

class NoSuchPortChoiceError(Exception):
    pass


class DevelPlotNode(Node):
    def __init__(self, name):
        Node.__init__(self, name)
        self._plot_classes = {}
        self._panes = {}
        self._active_plots = {}
        self._listener = None
        self._types = {}
        self._mapping_active = True
        self._active_pane_listeners = []
        self._failed_imports = {}

        self.set_layout(1, 1)

        plots = {}
        for path in Config.userplotdir, Config.plotdir:
            loader = UserPlotLoader(path)
            plots.update(loader.get_plots())
            self._failed_imports.update(loader.get_failed_imports())
        self.set_available_plots(plots)
        self.create_multi_input_port("plotinput")

    def set_available_plots(self, classes):
        """Use plot_name:plot_class dict as available plots."""
        self._plot_classes = classes

    def get_available_plot_types(self):
        """Get list of names of available plot types."""
        types = []
        for plot_type, plot_class in self._plot_classes.items():
            types.append((plot_type, plot_class.get_type_name()))
        return types

    def add_plot_by_type_at(self, plot_type, xpos, ypos):
        """Create instance of plot_type at layout position."""
        pane = self._panes[xpos, ypos]
        if plot_type in self._plot_classes:
            plot = self._plot_classes[plot_type]()
        else:
            plot = MissingPlot(plot_type)
        self._types[plot] = plot_type
        plot.update_ports(self.get_all_input_ports())
        pane.add_plot(plot)
        if self._listener:
            self._listener(xpos, ypos)
        return plot

    def mapped_selection_changed(self, index_sets, source):
        """Callback listening to mapped_selections in PlotPane."""
        for pane in self._panes.values():
            if pane == source:
                continue
            for plot in pane.get_plots():
                for index_set in index_sets:
                    plot.mapped_select(index_set.get_indices(),
                                       index_set.get_dataset_length())
                    for component in plot.get_cached_components():
                        component.mapped_select(index_set.get_indices(),
                                                index_set.get_dataset_length())

    def set_layout(self, width, height):
        """Set layout of plot squares to use."""
        self.setVariable("layoutwidth", width)
        self.setVariable("layoutheight", height)
        for x in range(width):
            for y in range(height):
                if (x, y) not in self._panes:
                    pane = PlotPane(self)
                    self._panes[x, y] = pane
        self.set_active_pane(self.get_pane_at(0, 0))
    
    def get_layout(self):
        """Get current layout, (width, height)."""
        return (int(self.getVariable("layoutwidth")),
                int(self.getVariable("layoutheight")))

    def get_plots_at(self, xpos, ypos):
        """Get all plot instances added to layout position."""
        if (xpos, ypos) not in self._panes:
            return []

        return self._panes[xpos, ypos].get_plots()

    def from_xml_element(self, root):
        Node.from_xml_element(self, root)
        self.set_layout(*self.get_layout())
        for plot_tag in root.findall("plot"):
            x = int(plot_tag.get('x'))
            y = int(plot_tag.get('y'))
            plot = self.add_plot_by_type_at(plot_tag.get('plot_type'), x, y)
            plot.from_xml_element(plot_tag)

    def toXML(self):
        xml_list = Node.toXML(self)
        # we insert our own xml within the tags
        for pos, pane in self._panes.items():
            for plot in pane.get_plots():
                plot_type = self._types[plot]
                x, y = pos
                plot_element = plot.to_xml_element(x, y, plot_type)
                xml_list = xml_list[:-1] + ElementTree.tostring(plot_element).split("\n") + xml_list[-1:]
        return xml_list

    def execute(self):
        Node.execute(self)
        for pane in self._panes.values():
            for plot in pane.get_plots():
                plot.update_ports(self.get_all_input_ports())

    def remove_plot(self, plot):
        del self._types[plot]
        for pane in self._panes.values():
            if plot in pane.get_plots():
                pane.remove_plot(plot)
                return

    def get_pane_at(self, x, y):
        return self._panes[x, y]

    def get_active_pane(self):
        return self._active_pane

    def set_active_pane(self, pane):
        self._active_pane = pane
        for listener in self._active_pane_listeners:
            listener()

    def add_active_pane_listener(self, listener):
        """Add a listener to be signaled when active pane changes."""
        self._active_pane_listeners.append(listener)

    def get_missing_plot_types(self):
        missing = []
        for pane in self._panes.values():
            for plot in pane.get_plots():
                if isinstance(plot, MissingPlot):
                    missing.append(plot.get_real_type())
        return missing

    def get_failed_plot_imports(self):
        """Get filename:report_string dictionary over failed imports."""
        return self._failed_imports


class PlotPane:
    """Bundle of plots in same layout square/plot window pane."""
    def __init__(self, plot_node):
        self._plots = []
        self._active_plot = None
        self._component_change_listeners = []
        self._plot_node = plot_node
        self._active_plot_listeners = []
        
    def add_plot(self, plot):
        self.set_active_plot(plot)
        self._plots.append(plot)
        plot.set_replot_listener(self._notify_component_change_listeners)
        self._notify_component_change_listeners()

    def get_plots(self):
        return self._plots

    def set_active_plot(self, plot):
        self._active_plot = plot
        self._notify_active_plot_listeners()

    def remove_plot(self, plot):
        self._plots.remove(plot)
        if plot == self._active_plot:
            if self._plots:
                self.set_active_plot(self._plots[-1])
            else:
                self.set_active_plot(None)
        self._notify_component_change_listeners()

    def get_active_plot(self):
        """One plot in each pane is always the active plot.

        The only exception ofcourse is when there are no plots in the
        pane.
        """
        return self._active_plot

    def add_component_change_listener(self, listener):
        self._component_change_listeners.append(listener)

    def disconnect_component_change_listener(self, listener):
        self._component_change_listeners.remove(listener)
        
    def _notify_component_change_listeners(self):
        for listener in self._component_change_listeners:
            listener()

    def add_active_plot_listener(self, listener):
        self._active_plot_listeners.append(listener)

    def disconnect_active_plot_listener(self, listener):
        self._active_plot_listeners.remove(listener)

    def _notify_active_plot_listeners(self):
        for listener in self._active_plot_listeners:
            listener()

    def select_area(self, x1, x2, y1, y2):
        for plot in self.get_plots():
            for component in plot.get_cached_components():
                component.select_area(x1, x2, y1, y2)

        sets = {}
        for plot in self.get_plots():
            for index_set in plot.get_mappable_index_sets_within(x1, x2, y1, y2):
                length = index_set.get_dataset_length()
                if length not in sets:
                    sets[length] = index_set
                else:
                    sets[length].add_set(index_set)
        
        self._plot_node.mapped_selection_changed(sets.values(), self)


class UserPlotLoader:
    def __init__(self, path):
        """Load all userplots from path."""
        self._plots = {}
        self._failed = {}
        files = glob.glob(os.path.join(path, "*.py"))
        filenames = {}
        for f in files:
            candidates = {}
            try:
                execfile(f, candidates)

                for key, value in candidates.items():
                    if not isclass(value) or value == UserPlot:
                        continue
            
                    if issubclass(value, UserPlot):
                        value.set_ui_path(path)
                        self._plots[key] = value
                        filenames[key] = f
            except:
                etype, value, tb = sys.exc_info()
                self._failed[f] = "".join(traceback.format_exception(etype, value, tb, None))

        # we try instantiating them, to check for some problems early
        for name, plot_class in self._plots.items()[:]:
            try:
                plot = plot_class()
            except:
                del self._plots[name]
                etype, value, tb = sys.exc_info()
                self._failed[filenames[name]] = "".join(traceback.format_exception(etype, value, tb, None))

    def get_plots(self):
        """Get all loaded plots as dictionary of name:plot_class."""
        return self._plots

    def get_failed_imports(self):
        """Get source:report_string dictionary over failed imports.

        source can be filename of failed file, or name of a bad class.
        """
        return self._failed


class UserPlot:
    """All user-created plots subclass UserPlot."""

    # should be overriden by user to set new name
    name = None
    
    _ui_path = ""

    # user overridable methods - should call parent-method
    def __init__(self):
        self._choice_order = []
        self._choices = {}
        self._replot_listener = None
        self._components = {}
        self._cached_components = []
        self._updated_initially = False
        self._indexed_components = False
        self._title = None
        self._x_label = None
        self._y_label = None
        self._ui_file = None
        self._reverse_y = False
        self._use_floating_scale = False

    def get_components(self):
        return []

    def input_changed(self, ident):
        pass

    def mapped_select(self, indices, length):
        pass

    # utility methods for user

    def add_input(self, choice):
        ident = choice.get_id()
        self._choices[ident] = choice
        self._choice_order.append(ident)
        choice.add_listener(self.input_changed)

    def get_input(self, ident):
        return self._choices[ident]

    def add_port_input(self, ident, label_name):
        self.add_input(PortChoice(ident, label_name))

    def add_list_input(self, ident, label, texts, values):
        self.add_input(ListChoice(ident, label, texts, values))

    def update_list_input(self, ident, texts, values):
        """Change choices for earlier created list input."""
        self._choices[ident].set_choices(texts, values)

    def select_list_index(self, ident, index):
        """Select one of the choices using 0-indexed number."""
        self._choices[ident].select(self._choices[ident].get_texts()[index])

    def add_integer_input(self, ident, label, minimum, maximum, default=None):
        self.add_input(IntegerChoice(ident, label, minimum, maximum, default))

    def update_integer_input(self, ident, minimum, maximum):
        """Change range of earlier created input."""
        self._choices[ident].set_range(minimum, maximum)

    def get_value(self, ident):
        """Get data from named choice."""
        return self._choices[ident].current_value()

    def get_data_instance(self, port_ident):
        """Get Data-instance from named port choice."""
        if port_ident not in self._choices:
            raise NoSuchPortChoiceError, port_ident
        choice = self._choices[port_ident]
        return choice.data_instance()

    def replot(self):
        """Callable from input_changed().

        Tells GUI classes that this plot has changes and they should
        call get_persistent_components(). It also updates the cached
        components returned to GUI calling get_cached_components().
        """
        named_components = {}
        self._cached_components = []

        for component in self.get_components():
            name = component.get_name()
            
            if name in self._components:
                component.set_selection(self._components[name].get_selection())
            if name:
                named_components[name] = component
            self._cached_components.append(component)
            
        self._components = named_components
        if self._replot_listener:
            self._replot_listener()

    def set_title(self, title):
        self._title = title

    def set_x_label(self, label):
        self._x_label = label

    def set_y_label(self, label):
        self._y_label = label

    def set_user_interface(self, filename):
        """Use GUI defined in filename for plot settings.

        The filename should be an absolute filename to a Qt Designer
        made interface of a widget containing child widgets with names
        matching choices. If the plot includes an integer input with
        ident"number", the interface should contain a spinbox named "number".
        """
        
        self._ui_file = filename

    def set_should_reverse_y(self, value=True):
        """Request flipped y-axis in plot.

        It's only possible to turn on/off requesting a flipped axis,
        calling this function with a False value will let the plot be
        displayed in the plot panes current mode.
        """
        self._reverse_y = value

    def set_use_floating_scale(self):
        self._use_floating_scale = True

    # to be called by plot system
    def get_should_reverse_y(self):
        return self._reverse_y

    def get_use_floating_scale(self):
        return self._use_floating_scale
    
    def get_title(self):
        return self._title

    def get_x_label(self):
        return self._x_label

    def get_y_label(self):
        return self._y_label

    def get_cached_components(self):
        if not self._updated_initially:
            self._updated_initially = True
            self.replot()
        return self._cached_components

    def update_ports(self, ports):
        for choice in self._choices.values():
            choice.update_ports(ports)

    def get_choices(self):
        return [self._choices[ident] for ident in self._choice_order]

    def get_type_name(cls):
        if cls.name:
            return cls.name
        else:
            return cls.__name__

    get_type_name = classmethod(get_type_name)

    def set_ui_path(cls, path):
        cls._ui_path = path

    set_ui_path = classmethod(set_ui_path)

    def from_xml_element(self, root):
        for choice_tag in root.findall("choice"):
            ident = choice_tag.get("ident")
            if ident in self._choices:
                self._choices[ident].from_xml_element(choice_tag)

    def to_xml_element(self, x, y, plot_type):
        root = ElementTree.Element("plot", {"x" : str(x), "y" : str(y),
                                            "plot_type" : plot_type})
        for choice in self._choices.values():
            root.append(choice.to_xml_element())

        return root

    def set_replot_listener(self, listener):
        self._replot_listener = listener

    def enable_indexed_components(self):
        self._indexed_components = True

    def disable_indexed_components(self):
        self._indexed_components = False

    def get_mappable_index_sets_within(self, x1, x2, y1, y2):
        sets = []
        components = []
        for i, component in enumerate(self.get_cached_components()):
            for index_set in component.get_mappable_index_sets_within(x1, x2, y1, y2):
                if index_set.get_indices() and i not in components:
                    components.append(i)
                sets.append(index_set)
        if self._indexed_components:
            sets.append(IndexSet(components, len(self.get_cached_components())))
        return sets

    def get_widget(self, parent):
        ui_file = self._get_user_interface()
        if ui_file:
            widget = QWidgetFactory.create(ui_file, None, parent)
            widget.show()
            return widget

    def _get_user_interface(self):
        if self._ui_file:
            return os.path.join(self.__class__._ui_path, self._ui_file)
        

class MissingPlot(UserPlot):
    def __init__(self, real_type):
        UserPlot.__init__(self)
        self._xml_object = None
        self._real_type = real_type

    def from_xml_element(self, element):
        self._xml_object = element

    def to_xml_element(self, x, y, plot_type):
        return self._xml_object

    def get_type_name(self):
        return "%s (missing)" % self._real_type

    def get_real_type(self):
        return self._real_type


class IndexSet:
    def __init__(self, indices, length):
        self._indices = indices
        self._length = length

    def get_indices(self):
        return self._indices

    def get_dataset_length(self):
        return self._length

    def add_set(self, other_set):
        try:
            self._indices = list(Set(self._indices).union(Set(other_set.get_indices())))
        except:
            raise
        self._indices.sort()

    def __eq__(self, other_set):
        return self._indices == other_set.get_indices() and self._length == other_set.get_dataset_length()

    def __ne__(self, other_set):
        return not self == other_set

    def __repr__(self):
        return "<IndexSet length:%d indices:%s>" % (self.get_dataset_length(), ",".join(map(str, self.get_indices()))[:20])
        
