# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from System.Data.Node import Node
from System.Data.Data import Data, RData, is_list, has_dimnames

class DissectionNode(Node):

    def __init__(self, name):
        Node.__init__(self, name)
        self._outputs = {}
        self._index_outputs = {}
        self._input_port = self.create_multi_input_port('struct input')

    def get_all_struct_data(self):
        structs = []
        for port in self.get_all_input_ports():
            if port.getData():
                data = port.getData()
                if isinstance(data, RData):
                    if is_list(data.get_pointer()) or has_dimnames(data.get_pointer()):                    
                        structs.append(data)
                elif isinstance(data.getData(), dict):
                    structs.append(data)

        return structs

    def set_output(self, struct_name, sub_elements = []):
        if sub_elements:
            port_name = "%s-%s" % (struct_name, '-'.join(sub_elements))
        else:
            port_name = struct_name

        port = self.create_output_port(port_name)
        self._outputs[port_name] = (struct_name, sub_elements)

    def execute(self):
        self._convert_index_outputs()    

        for port_name, value in self._outputs.items():
            struct_name, elements = value
            port = self.getOutputPortByName(port_name)
            if port:
                self.set_output_data(port, struct_name, elements[:])

    def _convert_index_outputs(self):
        input_structs = self.get_all_struct_data()
        structs = {}
        for data in input_structs:
            if isinstance(data, RData):
                structs[data.getName()] = data.get_data_with_dimnames()
            else:
                structs[data.getName()] = data.getData()        
        
        for port, index in self._index_outputs.items():
            try:
                struct_index = index[0]
                skeys = structs.keys()
                skeys.reverse()
                skey = skeys[struct_index]
                struct = structs[skey]
                
                port_name = skey
                for i in index[1:]:
                    keys = struct.keys()
                    keys.sort()
                    keys.reverse()
                    key = keys[i]
                    
                    struct = struct[key]
                    port_name += "-" + key
                    
                target_port = self.getOutputPortByName(port)
                target_port.setName(port_name)
                sub_items = port_name.split("-")[1:]
                
                self._outputs[port_name] = (skey, sub_items)
            except IndexError:
                # Can't find this output, igoring it.
                pass

        self._index_outputs = {}
        

    def set_output_data(self, port, struct_name, sub_elements):
        structs = self.get_all_struct_data()
        return_data = None
        
        for input_struct in structs:
            if input_struct.getName() == struct_name:
                if isinstance(input_struct, RData):
                    return_data = input_struct.get_data_with_dimnames()
                else:
                    return_data = input_struct.getData()
                break

        if not return_data:
            return

        data_name = struct_name
        sub_elements.reverse()        
        while sub_elements:
            element = sub_elements.pop()
            return_data = return_data[element]
            data_name = element

        port.setData(Data(data_name, return_data))
        
    def clear_output(self):
        for port in self.getOutputPorts():
            self.remove_port(port)

        self._outputs = {}

    def remove_output(self, port_name):
        port = self.getOutputPortByName(port_name)
        self.remove_port(port)
        

    def toXML(self):
        xml = Node.toXML(self)

        dissect_xml = []
        dissect_xml.append("\t\t<dissect_output>")

        for port, values in self._outputs.items():
            struct_name, sub_elements = values
            dissect_xml.append("\t\t\t<struct_item port_name='%s' struct='%s' sub_elements='%s' />" \
                               % (port, struct_name, ",".join(sub_elements)))

        dissect_xml.append("\t\t</dissect_output>")

        for xml_string in dissect_xml:
            xml.insert(-1, xml_string)

        return xml

    def from_xml_element(self, root):
        Node.from_xml_element(self, root)

        dissect_element = root.find('dissect_output')
        if dissect_element:
            for struct_item in dissect_element.findall('struct_item'):
                port_name = struct_item.get('port_name')
                struct = struct_item.get('struct')
                elements = struct_item.get('sub_elements')

                if elements:
                    elements = elements.split(",")
                else:
                    elements = []

                self._outputs[port_name] = (struct, elements)
            for index_item in dissect_element.findall('index_item'):

                port_name = index_item.get('port_name')
                index = [int(i) for i in index_item.get('index').split(',')]

                self._index_outputs[port_name] = index
                
        
            

            

                                
            
        


        
        
        
        
    
