# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import array as arrayMod
import copy
import os
import re
import string
import types
from sys import maxint

import Numeric

from System.Data import Data, Node

class EditNode( Node.Node ):
    """This class provides a number of commands to manipulate datasets.
       The functions are invoked by calling the method execCmd with the
       relevant command name.

       NB! Each command is implemented as a method in this class. However,
       if the methods are invoked directly, important parts of the error
       checking is bypassed, and an exception may be raised.

    """
    ltag = "    "
    ctag = ">>> EditNode"

    errPrefix = ">>> EDITNODE WARNING: "
    errNewline = "\n\t"

    # Some "constants". Makes the code more understandable.
    # Used for accessing dataObject list.
    dtoNAME = 0
    dtoDATA = 1
    dtoRANK = 2
    dtoTYPE = 3
    dtoDIMS = 4

    # Used for accessing chain list.
    chnNAME = 0
    chnINPUTNAMES = 1
    chnDIMENSIONS = 2
    chnCOMMANDS = 3
    chnOUTPUTPORT = 4
    
    # Format of ARGS_command variable:
    # [(argumentName, tuple with accepted types, description, 0 or 1, where 1 = required)]
    # To allow any type, use an empty tuple -> ().
    # Arguments prefixed with '__#' are internal arguments, and should not be set by the user.
    # Their value is actually not used, only key existance is checked.
    ARGS_transpose = [
        ("data", (list, Numeric.ArrayType), "The matrix/array or cube to be transposed.", 1),
        ("axis", (tuple,), "New axis-order. Syntax: (axisA, axisB,...,axisZ)", 0)
        ]

    ARGS_subset = [
        ("data", (list, Numeric.ArrayType), "The dataset to be reduced.", 1),
        ("upperLeft", (tuple,), "Upper left position of subset.", 1),
        ("lowerRight", (tuple,), "Lower right position of subset.", 1),
        ("dimension", (tuple,), "The dimension of the dataset to be reduced, specified as a tuple.", 0)
        ]
    
    ARGS_removeColumns = [
        ("data", (list, Numeric.ArrayType), "The dataset to operate on.", 1),
        ("columns", (tuple, list), "The columns to be removed.", 1),
        ("dimension", (int,), "Number of columns in the datastructure to edit.", 0),
        ("minIndex", (int,), "The index of the first column. Not used by user.", 0)
        ]

    ARGS_removeRows = [
        ("data", (list, Numeric.ArrayType), "The dataset to operate on.", 1),
        ("rows", (tuple, list), "The rows to be removed.", 1),
        ("dimension", (int,), "Number of rows in the datastructure to edit.", 0),
        ("minIndex", (int,), "The index of the first column. Not used by user.", 0)
        ]

    ARGS_vectorsToMatrix = [
        ("__#multipleOperands", None, "Specifies that the command can take multiple operands", 0),
        ("data", (list,), "The vectors to operate on. Must be a list with names of input ports.", 1),
        ("axis", (int,), "Specifies what axis to add in. 0 makes the vectors rows, 1 makes them columns.", 1)
        ]

    ARGS_joinMatrices = [
        ("__#multipleOperands", None, "Specifies that the command can take multiple operands", 0),
        ("data", (list, Numeric.ArrayType), "The matrices to join. Join-dimension must match.", 1),
        ("axis", (int,), "Specifies what axis to join in. 0 join vertically, 1 joins horizontally.", 1)
        ]

    ARGS_passThrough = [
        ("data", (), "The data to be passed on.", 1)
        ]

    ARGS_noop = [
        ("data", (), "", 1)
        ]


    # These are shown in the GUI, and used for matching on parameters in '__parseParams'.
    paramCommand = "Command"
    paramTranspose = "Transpose"
    labelTranspose = "Transpose"
    paramRemoveRows = "Remove rows"
    labelRemoveRows = "Remove rows"
    paramRemoveCols = "Remove columns"
    labelRemoveCols = "Remove columns"
    paramJoinMatrices = "Join matrices"
    labelJoinMatrices = "Join matrices"
    paramVectorsToMatrix = "Vectors to matrix"
    labelVectorsToMatrix = "Vectors to matrix"
    paramSubsetUL = "Subset, upper left"
    paramSubsetLR = "Subset, lower right"
    labelSubset = "Subset"
    labelPassThrough = "Pass through"
    labelTranspose = "Transpose"
    
    lookupInternalName = {labelTranspose: "transpose", labelRemoveRows: "removeRows", labelRemoveCols: "removeColumns", labelJoinMatrices: "joinMatrices", labelVectorsToMatrix: "vectorsToMatrix", labelSubset: "subset", labelPassThrough: "passThrough", "noop": "noop"}
    




    def __init__( self, name ):
        Node.Node.__init__(self, name)
        self.description = "The edit node makes it possible to manipulate data before passing them on. Currently, an edit node can only perform one operation specified by the user and additionally a set of operations specified by another node, for example a plot node."
        # Format for command chain tuple:
        # (name:str, [portName:str]*:list, portDataDim:tuple, [command:str, args:dict]*:list, nameOfOutputPort:str)
        # [portName] or portDataDim can be None, but not both of them!
        self.commandChains = []
        self.previousExtCmds = []

        # Add default input port ('operand1')
        desc = "Incoming data. You can delete or rename this port if you want. Note that you for some commands "
        desc += "must add additional input ports."
        self.create_input_port("operand1", "unknown", desc) 

        # Add 'availableAliases' output port.
        desc = "A list with the names of all data that is available to this node.\n"
        desc += "The data sent on this port is used by another node for specifying what "
        desc += "data the editing commands should operate on. If not needed, the port can be deleted."
        self.create_output_port("availableAliases", "list", desc)

        # [WARNING] This is specific for version 0.4x
        # Add EditNode output port 'editnode_output'.
        desc = "The result of the command executed on the EditNode."
        self.create_output_port("editnode_output", "unknown", desc)

        # Adding the command selection tab and parameters for the commands.
        self.create_parameter(self.paramCommand, "Text",
                          ("The command the edit node should execute.\n" +
                           "Please choose the apropriate command and then set the parameters they need.  Note that " +
                           "in this version all available input ports are passed to the command. All commands, " +
                           "except 'vectorsToMatrix and 'joinMatrices', takes only one input."),
                          None, None, None, [self.labelPassThrough, self.labelRemoveRows,self.labelRemoveCols, self.labelTranspose,
                                             self.labelVectorsToMatrix, self.labelJoinMatrices, self.labelSubset])
        self.create_parameter(self.paramRemoveRows, "Text",
                          ("Please enter the indices of the rows you want to remove, separated by commas. " +
                           "The first row has index 1. Note that you can specify ranges like 'startIndex-stopIndex'. " +
                           "Below are two simple examples.\n'1,6,3,12'\n'1,3-15,17,39'"))
        self.create_parameter(self.paramRemoveCols, "Text",
                          ("Please enter the indices of the columns you want to remove, separated by commas. " +
                           "The first column has index 1. Note that you can specify ranges like 'startIndex-stopIndex'. " +
                           "Below are two simple examples.\n'1,6,3,12'\n'1,3-15,17,39'"))
        self.create_parameter(self.paramVectorsToMatrix, "Integer",
                          ("This specifies along what axis to join the vectors.\n0 makes the vectors rows.\n" +
                           "1 makes the vectors columns.")
                          , 0, 1, 1)
        self.create_parameter(self.paramJoinMatrices, "Integer",
                          ("This specifies along what axis to join the matrices in.\n" +
                           "0 joins the matrices along the horizontal axis.\n" +
                           "1 joins the matrices along the vertical axis.\n\n" +
                           "Note that the matrices must have the same dimensions, except along the " +
                           "join axis."),
                          0, 1, 1)
        self.create_parameter(self.paramSubsetUL, "Text",
                          ("This specifies the upper left coordinate of the subset. The first row and column have index 1." +
                           "If the dataset is a list, specify one index. If it is a matrix, specify two indices, separated " +
                           "by a comma.\nNote that both this parameter and the parameter 'subset2' must be specified."))
        self.create_parameter(self.paramSubsetLR, "Text",
                          ("This specifies the lower right coordinate of the subset. The first row and column have index 1." +
                           "If the dataset is a list, specify one index. If it is a matrix, specify two indices, separated " +
                           "by a comma.\nNote that both this parameter and the parameter 'subset1' must be specified."))
        self.create_parameter("ignoreThisParameter4ever", "Hidden",
                          ("DO NOT EDIT THIS PARAMETER!\nIt is used by other nodes for sending commands to this edit node."))

        # Add the one and only command chain.
        self.addCommandChain("default", [], None, "editnode_output")





    def addCommandChain( self, name, dataVariables, dataDimensions, outPortName = "\"\"\"\"\"\"\"\""):
        """Adds a new command chain to the editnode. The chain is initialized with zero commands.

           Params:
             - name           : string  A name for the command chain.
             - dataVariables  : list    A list with names of the input data for the command chain.
                                        This must be names of input ports for the node.
             - dataDimensions : tuple   A tuple specifying the dimensions of the input data.
                                        It is used for dimension matching if no names of input data
                                        is specified, possibly dimension checking otherwise.
             - outPortName    : string  A name for the output port. If not specified, it is set to
                                        'chName_output'.

           Returns:
             A tuple with two elements.
             (0, errMsg) if an error occurs.
             (1, chainIndex) if succcess."""
        # Check inputs.
        if not isinstance(dataDimensions, tuple):
            if not dataDimensions == None:
                return (0, "Data dimensions must be specified as a tuple.")
        if dataVariables == None and dataDimensions == None:
            return (0, "Either the names or the dimension of the operand data must be specified.")
        if not isinstance(name, str) or len(name) == 0:
            return (0, "The name of the command chain must be specified as a string.")
        if outPortName == "\"\"\"\"\"\"\"\"":
            outPortName = name + "_output"
        else:
            if not isinstance(outPortName, str) or len(outPortName) < 1:
                return (0, "The name of the output port must be specified as a string.")
        chIndex = len(self.commandChains)
        self.commandChains.append([name, dataVariables, dataDimensions, [], outPortName])
        self.create_output_port(outPortName, "unknown", "Output port for the result of the command chain '" + name + "'.")

        return (1, chIndex)





    def addCommandToChain( self, cmdChainName, cmdName, args ):
        """Adds a command to the specified command chain.

           Params:
             - cmdChainName : string      Name of the command chain to add to.
             - cmdName      : string      Name of the command.
             - args         : dictionary  A dictionary with arguments to be passed to the command.

           Returns:
             A tuple with two elements.
             (0, errMsg) if error occurs.
             (1, commandIndex) if success."""
        mtag = self.ctag + ".addCommandToChain <<<" + os.linesep + self.ltag
        # Find the correct chain. Two chains can not have the same name!
        # Insted of a name, an index could be used.
        commandAdded = 0
        for chainIndex in range(0, len(self.commandChains)):
            chain = self.commandChains[chainIndex]
            if chain[0] == cmdChainName:
                cmdList = self.commandChains[chainIndex][self.chnCOMMANDS]
                cmdIndex = len(cmdList)
                cmdList.append([cmdName, args])
                return (1, cmdIndex)
        return (0, "Unable to add command. Specified command chain not found!")





    def getCommandChains( self ):
        """Returns a reference to the list containing the command chains.

           Returns:
             A list with command chains.

           Note:
             The index value returned when invoking method addCommandChain(),
             can be used in the list returned."""
        return self.commandChains





    def listCommands( self, category = None ):
        """Returns a list with the available commands. This method shall eventually be used
           to populate a list in the GUI. The current format may not be suitable."""
        commands = []
        cmds = dir(self)
        for posCmd in cmds:
            if re.compile("_EditNode__cmd_").match(posCmd):
                # Build command description list:
                # ["commandName", argumentList]
                # argumentList = list with tuples (argName, argType, description, required)
                # The argumentlist must be defined as ARGS_commandname
                command = []
                command.append(posCmd[15:])

                argList = getattr(self, "ARGS_" + command[0], None)
                add = 1
                if argList != None:
                    if category != None:
                        # See if the method can take the specified type of data input.
                        # If arg "data" is not found, assume command can be applied to any type.
                        add = 0
                        found = 0
                        for arg in argList:
                            if arg[0] == "data":
                                found = 1
                                for type in arg[1]:
                                    if type == category:
                                        add = 1
                        if not found:
                            add = 1

                if add:
                    command.append(argList)
                    commands.append(command)
                
        return commands
            
    



    def execCmd( self, cmdName, args ):
        """This is a wrapper method for all the editing commands in this class."""
        mtag = self.ctag + ".execCmd(3) <<<" + os.linesep + self.ltag
        try:
            call = getattr(self, "_EditNode__cmd_" + str(cmdName))
        except Exception, e:
            return (0, "Non-existing command '" + str(cmdName) + "'.")
        validation = self._EditNode__automaticValidation(cmdName, args)
        if validation[0] != 1:
            missing = ""
            invalid = ""
            if len(validation[2]) > 0:
                missing = "Missing arguments: " + str(validation[2]) + ". "
            if len(validation[3]) > 0:
                invalid = "Invalid arguments: " + str(validation[3]) + ". "
            msg = "Validation failed. " + missing + invalid
            return (0, msg)
        res = call(args)

        return res     


    def _has_textual_first_element(self, arrayData):
        if hasattr(arrayData, "__iter__"):
            iter = arrayData.__iter__()
            return self._has_textual_first_element(iter.next())
        # We have a atomic datatype.
        if type(arrayData) is types.StringType:
            return True
        return False


    def _create_numeric_array(self, data):
        """
        """
        mtag = self.ctag + "._create_numeric_array(2) <<<" + os.linesep
        ret_data = None
        # Check whether there is a string in the first element of the array.
        # Must make this recursve, we do not know the number of dimensions.
        # If there ia a non-string in the first element, we are good to go
        # with the Numeric.array() method directly. If a string-array is
        # passed without proper arguments, Numeric converts each char to a
        # float...
        if not self._has_textual_first_element(data):
            try:
                ret_data = Numeric.array(data, Numeric.Float)
            except:
                pass
        if ret_data is None:
            try:
                ret_data = Numeric.array(data, Numeric.PyObject)
            except:
                pass
        return ret_data 
           
    

    def __automaticValidation( self, cmdName, args ):
        """This method validates data for commands on the following ways:
              1) Checks if all required arguments are given.
              2) Checks the types of all incoming arguments.

            Params:
              - cmdName   : string       The name of the command to validate input for.
                                         Example: 'transpose' for __cmd_transpose().
              - args      : dictionary   The set of arguments passed to the command.

            Returns:
              A tuple of one int and three lists: (status, validArgs, missingArgs, invalidArgs)
              Status is 1 if there are no missing or invalid arguments."""
        mtag = self.ctag + ".__automaticValidation(3) <<<" + os.linesep + self.ltag
        argList = getattr(self, "ARGS_" + str(cmdName), None)
        if argList == None:
            return (0, [],[],[])
        valid = []
        invalid = []
        missing = []
        status = 1

        for arg in argList:
            # Ignore internal arguments.
            if arg[0][0:2] == "__#":
                continue
            # Arg format: (name, (types), desc, required)
            # The name of the argument is used as key in the dictionary.
            exists = args.has_key(arg[0])
            if (not exists) and arg[3]:
                missing.append(arg[0])
            elif exists:
                # Check type. "Shallow" check, i.e. elements in lists and arrays are not checked.
                typeOK = 0
                typeReceived = type(args.get(arg[0]))
                if arg[1] == ():
                    typeOK = True
                else:
                    for typeDef in arg[1]:
                        if typeReceived is typeDef:
                            valid.append(arg[0])
                            typeOK = True
                            break
                if not typeOK:
                    invalid.append(arg[0])

        if len(invalid) > 0 or len(missing) > 0:
            status = 0

        return (status, valid, missing, invalid)




            
    def __cmd_transpose( self, args ):
        mtag = self.ctag + ".cmd_transpose(2) >>>" + os.linesep + self.ltag
        # Set default values for non-mandatory arguments
        axis = None
        
        data = self._create_numeric_array(args.get("data"))
        if data is None:
            return (0, "Invalid dataset provided for transpose command.")

        # If we receive a "normal" list with non-list elements,
        # make it into a one row matrix. If not done, transpose will not do anything. 
        if Numeric.rank(data) == 1:
            data = self._create_numeric_array([data])

        # Validate other arguments.
        if args.has_key("axis"):
            tmp_axis = args.get("axis")
            if len(tmp_axis) != Numeric.rank(data):
                msg = "\"axis\" - matrix and axis tuple has different rank ("
                msg += str(Numeric.rank(data)) + "/" + str(len(tmp_axis)) + ")"
                return (0, msg)
            else:
                axis = tmp_axis
                
        transposed = Numeric.transpose(data, axis)

        # Convert the Numeric.ArrayType back to lists and return result.
        return (1, transposed.tolist())





    def __cmd_subset( self, args ):
        """This method returns a subset of a dataset.
           It currently supports operation on lists and on matrices. All
           parameters decribed below must be passed through the dictionary
           'args'.
           NB! Coordinates starts at 1, not at 0!

           Params:
             - args         : dictionary
               'data'       : list, Numeric.ArrayType
               'upperLeft'  : tuple
               'lowerRight' : tuple"""
        mtag = self.ctag + ".__cmd_subset(2) <<<" + os.linesep + self.ltag
        upperLeft = args.get("upperLeft")
        lowerRight = args.get("lowerRight")
        
        # Convert to Numeric.ArrayType, find rank and decide what to do.
        array = self._create_numeric_array(args.get('data'))
        if array is None:
            return (0, "Invalid dataset provided for subset command.")
        rank = Numeric.rank(array)
        # Make sure the coordinates are of the valid type, e.g. tuple length equals rank.
        if len(upperLeft) != rank or len(lowerRight) != rank:
            return (0, "Coordinates for subset does not match dataset dimensions.")

        # We have a list.
        if rank == 1:
            upperLeft = upperLeft[0] -1
            lowerRight = lowerRight[0] -1
            # Validate actual value of coordinates. Do not tolerate negative index.
            if (not type(upperLeft) is int) or upperLeft < 0 or upperLeft >= len(array):
                msg = "Upper left coordinate is not valid! Must be an integer in range [1, " + str(len(array)) + "]."
                return (0, msg)
            if (not type(lowerRight) is int) or lowerRight < 0 or lowerRight >= len(array):
                msg = "Lower right coordinate is not valid! Must be an integer in range [1, " + str(len(array)) + "]."
                return (0, msg)            
            # Should we wrap around or perform a "normal" slice?
            if upperLeft < lowerRight:
                try:
                    array = array[upperLeft: lowerRight +1]
                    array = array.tolist()
                except Exception, e:
                    return (0, "Unexpected error: " + str(e))
                return (1, array)
            elif upperLeft > lowerRight:
                try:
                    tmpA = array[0: lowerRight +1].tolist()
                    tmpB = array[upperLeft:].tolist()
                    array = tmpA + tmpB
                except Exception, e:
                    return (0, "Unexpected error: " + str(e))
                return (1, array)
            else:
                # Must be equal, and meaning must be fetch single element.
                return (1, array[upperLeft: upperLeft +1].tolist())


        # We have a matrix.
        # [NOTE] We can handle "edge-wrapping" subsets here, but this kind of overlaps with
        #        the functionality of removeCols and removeRows. Is this of any practical use?
        #        Order of columns and/or rows will be changed.

        if rank == 2:
            # Validate coordinates.
            uLr = upperLeft[0] -1
            uLc = upperLeft[1] -1
            lRr = lowerRight[0] -1
            lRc = lowerRight[1] -1
            rows = len(array)
            cols = len(array[0])
            if (uLr < 0 or uLr >=  rows) or (uLc < 0 or uLc >= cols):
                msg = "Upper left coordinate is not valid! Must be in range [1, 1] to ["
                msg += str(rows) + ", " + str(cols) + "]."
                return (0, msg)
            if (lRr < 0 or lRr >=  rows) or (lRc < 0 or lRc >= cols):
                msg = "Lower right coordinate is not valid! Must be in range [1, 1] to ["
                msg += str(rows) + ", " + str(cols) + "]."
                return (0, msg)
            if uLr > lRr:
                return (0, "Row number for upper coordinate cannot be higher than lower coordinate.")
            # Decide what actions to take. A normal slice or something special...
            if (uLr <= lRr) and (uLc < lRc):
                # Do a normal slice.
                array = array[uLr: lRr +1, uLc: lRc +1]
                return (1, array.tolist())
            else:
                # Not implemented yet.
                return (0, "Not able to handle wrapping subsets of matrices.")

        if rank == 3:
            # Not implemented yet.
            return (0, "Subset operation on cubes has not yet been implemented.")
        return (0, "Subset operation on the provided datatype is not supported.")





    def _preprocessIndices( self, indices, maxIndex, minIndex = 0 ):
        """Internal method. Generates a list with column or row indices to keep for
           use with the method Numeric.take().

           Params:
             - indices   : list    A list with the indices to *remove* from the dataset.
             - maxIndex  : int     The highest valid index.

           Returns:
             (1, tuple) if everything is OK.
             (0, tuple) if some indices was invalid. These are ignored."""
        mtag = self.ctag + "._preprocessIndices(3) <<<" + os.linesep + self.ltag
        raw_indList = []
        if type(indices) is tuple:
            # Convert tuple to list.
            for index in indices:
                raw_indList.append(index)
        else:
            raw_indList = indices
        raw_indList.sort()
        previousIndex = -1;
        processedIndices = []
        noInvalidIndices = 1
        invalid = []
        for index in raw_indList:
            if not type(index) is int:
                noInvalidIndices = 0
            else: 
                if index < minIndex or index > maxIndex:
                    noInvalidIndices = 0
                    invalid.append(index)
                else:
                    # Index is valid. Add it to list if it does not already exist.
                    if index != previousIndex:
                        processedIndices.append(index)
                    previousIndex = index
        # Transform indices from [minIndex, maxIndex] to [0, maxIndex - minIndex]
        processedIndices = map(lambda x : x - minIndex, processedIndices)
        # Transform list with indices to remove to list with indices to keep.
        tmpList = []
        if len(processedIndices) == 0:
            processedIndices.append(-999) # Add index that will never satisfy if-test below.
        searchIndex = 0
        for index in range(0, maxIndex - minIndex + 1):
            if index == processedIndices[searchIndex]:
                if searchIndex < (len(processedIndices) -1):
                    searchIndex += 1
            else:
                tmpList.append(index)

        processedIndices = tuple(tmpList)
        return (noInvalidIndices, processedIndices, invalid)


            
    def __cmd_removeColumns( self, args, dimension = 1 ):
        """This method removes the specified columns from a dataset.
           If the argument 'dimension' is set, the command will only be
           executed if the 'dimension' argument and the number of columns
           in the dataset is the same.
           (Below; ! means mandatory, ? means optional.)

           Params:
             - args          : dictionary   : A dictionary with the arguments for the command.
               ! 'data'      : list/array   : The dataset to operate on.
               ! 'columns'   : tuple/list   : A list with the columns to remove.
               ? 'dimension' : int          : The number of columns in dataset to edit.
             - dimension     : int          : Used internally. If value is 1, columns are removed.
                                              If it is 0, rows are removed."""
        mtag = self.ctag + ".__cmd_removeColumns(2/3) <<<" + os.linesep + self.ltag

        data = self._create_numeric_array(args.get("data"))
        if data is None:
            return (0, "Invalid dataset provided for 'removeColumns' command.")
        rank = Numeric.rank(data)
        userDimension = args.get("dimension")
        if isinstance(userDimension, int):
            # Check number of columns in dataset.
            dataColumns = len(data)
            if rank == 2:
                dataColumns = len(data[0])
                
            if userDimension != dataColumns:
                return (0, "Column mismatch between dataset and user specification. (" + str(dataColumns) + "/" + str(userDimension) + ")")

        if rank == 1:
            data = self._create_numeric_array([data])
            # "Fall-through" to next section - matrices.
            rank = 2
            
        if rank == 2:
            if args.get("minIndex") != None:
                min = args.get("minIndex")
                max = len(data[0]) -1 + min
            else:
                max = len(data[0])
                min = 1
            userIndicesList = args.get("columns")
            if dimension == 0:
                # We are removing rows instead of columns.
                max = len(data)-1 + min
                userIndicesList = args.get("rows")
            (status, keepTuple, invalid) = self._preprocessIndices(userIndicesList, max, min)
            if keepTuple != ():
                res = Numeric.take(data, keepTuple, dimension)
                res = res.tolist()
            else:
                res = []
            return (1, res)

        msg = "Dataset of unsupported rank (" + str(rank) + ") received."
        return (0, msg)


    
    def __cmd_removeRows( self, args ):
        """Internal method. Removes the specified columns from a dataset. The
           parameters are passed through the dictionary 'args'.

           Params:
             - """
        mtag = self.ctag + ".__cmd_removeRows(2) <<<" + os.linesep + self.ltag
        data = self._create_numeric_array(args.get('data'))
        if data is None:
            return (0, "Invalid dataset provided for 'removeRows' command.")
        # Must add a dummy "columns" argument to pass automatic validation.
        args["columns"] = [-1]

        return self._EditNode__cmd_removeColumns(args, 0)


    def __cmd_vectorsToMatrix( self, args ):
        """Internal method. Take a number of vectors and transforms them into a matrix.

           Params:
             - args : dictionary   : A dictionary with arguments. See allowed keys below.
                      'data'       : list   A list of vectors.
                      'axis'       : int    0 -> vectors treated as rows
                                            1 -> vectors treated as columns

           Returns:
             (0, errMsg) if an error occurs.
             (1, list) if operations succeds."""
        mtag = self.ctag + ".__cmd_vectorsToMatrix <<<" + os.linesep + self.ltag
        vectors = None
        number = -1
        length = -1
        axis = args.get("axis")
        try:
            axis = int(axis)
            if axis < 0 or axis > 1:
                return (0, "Axis must be either 0 or 1.")
            vectors = args.get("data")
            number = len(vectors)
            # The code below is a fix, because if only one vector is found, it is not wrapped by a list.
            length = number
            if isinstance(vectors[0], list):
                length = len(vectors[0])
            else:
                number = 1
                vectors = [vectors]
        except Exception , e:
            return (0, "Unable to process input information. 'axis' or 'data' argument is invalid.")
        if number < 1:
            return (0, "No vectors for processing received.")

        if axis == 0:
            matrix = []
            for vector in vectors:
                if not isinstance(vector, list) or len(vector) != length:
                    return (0, "At least one of the vectors received is invalid.")
                else:
                    matrix.append(vector)

            
        if axis == 1:
            matrix = []
            # Must check that every vector has the same length.
            for vector in vectors:
                if not isinstance(vector, list) or length != len(vector):
                    return (0, "At least one of the vectors received is invalid.")
            for rowIndex in range(0, length):
                row = []
                for vector in vectors:
                    row.append(vector[rowIndex])
                matrix.append(row)

        # Test matrix with Numeric.array()
        if self._create_numeric_array(matrix) is None:
            return (0, "Unable to validate matrix! Possible causes: strings in matrix.")

        return (1, matrix)

   
    def __cmd_joinMatrices( self, args ):
        """Takes a number of matrices and joins them along the specified axis.

            Params:
              - args : dictionary   : A dictionary with the arguments for the command.
                       'data'       : list  A list with matrices. All matrices must have the same
                                            dimension, except along the join axis.
                       'axis'       : int   0 -> vertical join
                                            1 -> horizontal join

            Returns:
              (1, matrix) if operation succeded.
              (0, errMsg) otherwise."""
        mtag = self.ctag + ".__cmd_joinMatrices(2) <<<" + os.linesep + self.ltag
        data = args.get("data")
        axis = args.get("axis")
        if axis < 0 or axis > 1:
            return (0, "Invalid join axis specified. Must be either 0 (vertical join) or 1 (horizontal join).")
        if len(data) < 2:
            return (0, "Too few matrices received. Must be 2 or more.")
        newMatrix = []
        try:
            newMatrix = Numeric.concatenate(data, axis)
            newMatrix = newMatrix.tolist()
        except Exception, e:
            return (0, "Join operation failed. Error reported: " + str(e))

        return (1, newMatrix)




    
    def __dimensionMatching( self, dimensions1, dimensions2 ):
        """
           Params:
             - dimensions1 : tuple
             - dimensions2 : tuple

           Returns:
             (-1, msg) on error, (0, None) on no match, (1, dimIndex) on match."""
        mtag = self.ctag + ".__dimensionMatching(3) <<<" + os.linesep + self.ltag
        if not isinstance(dimensions1, tuple) or not isinstance(dimensions2, tuple):
            return (-1, "Invalid input.")
        rank = len(dimensions1)
        if rank != len(dimensions2):
            msg = "The dimensions given does not have the same rank. Unable to continue."
            return (-1, msg)
        matches = []
        for dimIndex in range(0, rank):
            if dimensions1[dimIndex] == dimensions2[dimIndex]:
                matches.append(dimIndex)
        if len(matches) == 0:
            return (0, matches)
        return (1, matches)




        
    def __cmd_passThrough( self, args ):
        data = args.get("data")
        if data == None:
            return (0, "No data received!")
        return (1, data)


    
    def __cmd_noop( self, args ):
        return (1, None)


    
    def __findDataPort( self, datalist, nameList, dimension, forceDimensionMatch = 0 ):
        """Searches for inputport to get data from. If no port with a matching name
           is found, the search is continued by matching dimension.

           Params:
             - datalist            : list    See execute() for format of the list elements.
             - name                : list    A list with the names to search for.
             - dimension           : tuple   The dimension of the data to edit. -1 for
                                             'zero-dimensional' data.
             - forceDimensionMatch : boolean Should dimension matching be performed when
                                             name matches?

           Returns:
             Returns a list with two lists: the the name of and data on the matching inputports. Returns a empty list if
             no matching ports are found."""
        mtag = self.ctag + ".__findDataPort(4/5) <<<" + os.linesep + self.ltag
        # Search for input port with name as specified.
        names = []
        data = []
        # Must check this. If nameList is not a list, the second for-loop below fails.
        if not isinstance(nameList, list):
            nameList = []
        for port in datalist:
            for name in nameList:
                if port[0] == name:
                    if forceDimensionMatch:
                        if isinstance(dimension, tuple):
                            if self._EditNode__dimensionMatching(dimension, port[4])[0] == 1:
                                names.append(port[0])
                                data.append(port[1])
                            else:
                                # Data may have been transposed. Only "total" transpose is supported.
                                dimension = list(dimension)
                                dimension.reverse()
                                dimension = tuple(dimension)
                                if self._EditNode__dimensionMatching(dimension, port[4])[0] == 1:
                                    names.append(port[0])
                                    data.append(port[1])
                    else:
                        names.append(port[0])
                        data.append(port[1])
        # [WARNING] Is this wanted behaviour? This means that dimension matching search is not performed if
        #           we have a match on names.
        if len(data) > 0:
            return [names, data]
        # No matching name found. Search for data using dimension matching.
        if isinstance(dimension, tuple) and len(dimension) > 0:
            rank = len(dimension)
            for port in datalist:
                if rank == port[2]:
                    if self._EditNode__dimensionMatching(dimension, port[4])[0] == 1:
                        names.append(port[0])
                        data.append(port[1])
        return [names, data]





    def __textualTypeDescription( self, object ):
        """Internal method. Returns a textual description of the data received.

           Params:
             - object : object  The object to determine type for.

           Returns:
             A string describing the object type."""
        dtype = type(object)
        if dtype is None:
            return "unknown"
        if dtype is int:
            return "integer"
        if dtype is float:
            return "decimal"
        if dtype is arrayMod.ArrayType or dtype is Numeric.ArrayType:
            return "matrix"
        if dtype is list:
            try:
                tmp = self._create_numeric_array(object)
                if tmp != None:
                    if Numeric.rank(tmp) >= 2:
                        return "matrix"
            except:
                pass
            return "list"
        if dtype is tuple:
            return "tuple"
        if dtype is string:
            return "string"
        return "unknown"



    def __validateRange( self, rangeDef, min, max ):
        """Validates and writes out a range of integer indices.

           Params:
             - rangeDef  : str  : A range definition.
                                  Syntax: '1,5,7,10-25'
             - min       : int  : The minimum index allowed.
             - max       : int

           Returns:
             (1, list) if success. List is a list with all integer indices written out.
             (0, errMsg) otherwise."""
        mtag = self.ctag + ".__validateRange <<<" + os.linesep + self.ltag
        invSynErrMsg = "Invalid range definition syntax!"
        intRange = []
        elements = rangeDef.split(",")
        index = -1
        for element in elements:
            notNumber = 0
            createRange = 0
            validate = []
            element = element.strip()
            try:
                validate.append(int(element))
            except:
                notNumber = 1
            if notNumber:
                if len(element) == 0:
                    return (0, invSynErrMsg)
                rangeElements = element.split("-")
                if len(rangeElements) != 2:
                    return (0, invSynErrMsg)
                start = rangeElements[0].strip()
                stop = rangeElements[1].strip()
                try:
                    validate.append(int(start))
                    validate.append(int(stop))
                    createRange = 1
                except:
                    msg = "Invalid sub-range definition. All indices must be integers."
                    return (0, msg)

            # Validate index/indices.
            for index in validate:
                if index < min or index > max:
                    

                    msg = "Index out of bounds: " + str(index) + ". Must be in range [" + str(min) + ", " + str(max) + "]."
                    return (0, msg)
            if len(validate) == 1:
                intRange.append(validate[0])
            elif len(validate) == 2:
                for index in range(validate[0], validate[1] +1):
                    intRange.append(index)

        intRange.sort()
        return (1, intRange)



                
    
    def __validateIntTuple( self, tupleDef ):
        """Validates a text defining a integer tuple.

           Params:
             - tupleDef  : str  : A text defining a integer tuple.
                                  '1,2,3' -> (1,2,3)
                                  '1'     -> (1,)
                                  '(1)'   -> (1,)

           Returns:
             (1, tuple) if success.
             (0, errMsg) otherwise."""
        mtag = self.ctag + ".__validateTuple(2) <<<" + os.linesep + self.ltag
        if tupleDef == None:
            return (0, "Received 'None'.")
        if type(tupleDef) is not types.StringType:
            return (0, "Received illegal data type: " + str(type(tupleDef)))
        tupleDef = tupleDef.strip()
        if len(tupleDef) == 0:
            return (0, "Received no tuple data (empty string).")
        if tupleDef[0]  == '(': tupleDef = tupleDef[1:]
        if tupleDef[-1] == ')': tupleDef = tupleDef[:-1]
        elements  = tupleDef.split(",")
        validated = []
        for element in elements:
            try:
                validated.append(int(element))
            except:
                return (0, "Unable to coerce data to tuple of integers.")
        return (1, tuple(validated))


    def __cleanAndOptimizeExtCmdInput( self, extCmdInput ):
        mtag = self.ctag + ".__cleanAndOptimizeExtCmdInput <<<" + os.linesep + self.ltag
        # Sort on command input. This assumes chains processing is independent on input,
        # i.e. order of chains can be swapped.
        optimized = []
        data = {}
        for cmd in extCmdInput:
            input = None
            try:
                input = cmd[1].get("data")
            except:
                optimized.append(cmd)
                continue
                
            if len(input) == 1:
                input = str(input[0])
                if data.get(input) == None:
                    data[input] = [cmd]
                else:
                    data.get(input).append(cmd)
            else:
                # Don't know what to do...
                # Ignore, put directly in result list.
                optimized.append(cmd)

        lastCommand = ""
        for commands in data.values():
            for cmd in commands:
                name = cmd[0]
                args = cmd[1]
                if name == "removeRows":
                    data = args.get("data")
                    if len(data) == 1:
                        data = str(data[0])
                        if name != lastCommand:
                            optimized.append(cmd)
                        else:
                            curRows = args.get("rows")
                            curDims = args.get("dimension")
                            prevDims = optimized[-1][1].get("dimension")
                            if isinstance(curRows, list) and curDims == prevDims:
                                # Get last command in this chain.
                                for row in curRows:
                                    optimized[-1][1]["rows"].append(row)
                            else:
                                optimized.append(cmd)
                                lastCommand = "invalid"
                    if lastCommand != "invalid":
                        lastCommand = "removeRows"
                elif name == "removeColumns":
                    data = args.get("data")
                    if len(data) == 1:
                        data = str(data[0])
                        if name != lastCommand:
                            optimized.append(cmd)
                        else:
                            curCols = args.get("columns")
                            curDims = args.get("dimension")
                            prevDims = optimized[-1][1].get("dimension")
                            if isinstance(curCols, list) and curDims == prevDims:
                                # Get last command in this chain.
                                for col in curCols:
                                    optimized[-1][1]["cols"].append(row)
                            else:
                                optimized.append(cmd)
                                lastCommand = "invalid"
                    if lastCommand != "invalid":
                        lastCommand = "removeColumns"
                elif name == "transpose":
                    lastCommand = ""
                elif name == "subset":
                    lastCommand = ""
                elif name == "vectorsToMatrix":
                    lastCommand = ""
                elif name == "joinMatrices":
                    lastCommand = ""
                    
        # Must add argument "minIndex" = 0 to ensure correct handling of indices.
        # All indices received from other nodes are assumed to start at 0 if "minIndex" is not present.
        for cmdIndex in range(0, len(optimized)):
            try:
                if optimized[cmdIndex][1].get("minIndex") == None:
                    optimized[cmdIndex][1]["minIndex"] = 0
            except:
                pass
        return optimized

                   
            

        
    def __parseParams( self, params ):
        """This method parses input received from the GUI (user input), and
           represents them in the correct format if possible.

           Params:
             - commandName : str        : The name of the command to validate input for.
             - params      : dictionary : A dictionary with user input.

           Returns:
             A list with 2 elements:
             [cmdName, dictionary with correctly formatted arguments] if success.
             'None' otherwise.
        """
        mtag = self.ctag + ".__parseParams(2) <<<" + os.linesep + self.ltag
        args = {}
        # Find command that we are parsing input for.
        cmdName = None
        for param in params:
            if param.getName() == "Command": # and param.isValueSet():
                cmdName = str(param.getValue())
        # The code below is very specific, and very good...
        # The names of the params must be the same as specified in build().
        if cmdName == self.labelRemoveRows:
            for param in params:
                name = param.getName()
                hasValue = param.isValueSet()
                if name == self.paramRemoveRows:# and hasValue:
                    (status, result) = self._EditNode__validateRange(param.getValue(), 1, maxint)
                    if status == 0:
                        return ["noop", {}]
                    args["rows"] = result

        elif cmdName == self.labelRemoveCols:
            for param in params:
                name = param.getName()
                hasValue = param.isValueSet()
                if name == self.paramRemoveCols:# and hasValue:
                    (status, result) = self._EditNode__validateRange(param.getValue(), 1, maxint)
                    if status == 0:
                        return ["noop", {}]
                    args["columns"] = result

        elif cmdName == self.labelTranspose:
            # Nothing needed here at the moment (no needed parameters).
            pass

        elif cmdName == self.labelSubset:
            for param in params:
                name = param.getName()
                hasValue = param.isValueSet()
                if name == self.paramSubsetUL:# and hasValue:
                    (status, result) = self._EditNode__validateIntTuple(param.getValue())
                    if status == 0:
                        return ["noop", {}]
                    args["upperLeft"] = result
                if name == self.paramSubsetLR:# and hasValue:
                    (status, result) = self._EditNode__validateIntTuple(param.getValue())
                    if status == 0:
                        return ["noop", {}]
                    args["lowerRight"] = result
                    
        elif cmdName == self.labelVectorsToMatrix:
            for param in params:
                name = param.getName()
                hasValue = param.isValueSet()
                if name == self.paramVectorsToMatrix:# and hasValue:
                    try:
                        args["axis"] = int(str(param.getValue()))
                    except:
                        pass
        elif cmdName == self.labelJoinMatrices:
            for param in params:
                name = param.getName()
                hasValue = param.isValueSet()
                if name == self.paramJoinMatrices:# and hasValue:
                    try:
                        args["axis"] = int(str(param.getValue()))
                    except:
                        pass
        # [NOTE] Last part of elif is due to default command not being returned from GUI.
        #        The selection reads "passThrough", but still only empty string is returned.
        #        This happens if user does not manipulate the selection widget.
        elif cmdName == self.labelPassThrough or cmdName == "":
            return ["passThrough", args]
        else:
            cmdName = "noop"
            args = {}

        return [cmdName, args]

        
    def execute( self ):
        mtag = self.ctag + ".execute(1) <<<" + os.linesep + self.ltag
        # Used for accessing commandDesc list.
        COMMAND = 0
        ARGS = 1

        # Keep track of which output ports have their data value set.
        opHasValueSet = {}

        # Recreate command chain if it does not exist. Needed when loading from file.
        if len(self.commandChains) == 0:
            self.addCommandChain("default", [], None, "editnode_output")
        if not isinstance(self.previousExtCmds, list):
            self.previousExtCmds = []
        errorFreeExec = 1

        # Reset chain data.
        self.commandChains[0][self.chnINPUTNAMES] = []
        self.commandChains[0][self.chnCOMMANDS] = []
        self.commandChains[0][self.chnDIMENSIONS] = None

        #### Fetch all input data and generate some info about it.
        inData = []
        inDataByNames = {}
        for inPort in self.inPorts:
            dataObject = []
            # [WARNING] According to documentation in NodePort, getVariableName()
            #           shall be renamed to getName().
            dataObject.append(inPort.getVariableName())
            data = inPort.getData()
            if isinstance(data, Data.Data):
                dataObject.append(data.getData())
            else:
                dataObject.append(data)
            del(data)
            from Library.Util.lists import get_repr_dim
            dataObject.append(get_repr_dim(dataObject[self.dtoDATA]))
            # Not sure what the method getDataType() of NodePort returns.
            #dataObject.append(inPort.getDataType())
            dataObject.append(type(dataObject[self.dtoDATA]))
            # Create dimension tuple.
            if dataObject[self.dtoRANK] == 0:
                dataObject.append((-1,))
            else:
                dims = []
                dimData = dataObject[self.dtoDATA]
                for dim in range(0, dataObject[self.dtoRANK]):
                    dims.append(len(dimData))
                    try:
                        dimData = dimData[0]
                    except:
                        pass
                dataObject.append(tuple(dims))
            inData.append(dataObject)
            inDataByNames[dataObject[self.dtoNAME]] = inData[-1]
            
        # This time we only get one args-dictionary because of limit on one user command in EditNode.
        commandDef = self._EditNode__parseParams(self.getParameters())
        self.commandChains[0][self.chnCOMMANDS].append(commandDef)
        
        # Add all input ports.
        # [HACK] Another hack, fix this later... Again, this only works/is needed in v. 0.4x
        #        We maps input ports to chains.
        inputToChainName = {}
        del self.commandChains[0][self.chnINPUTNAMES][0:]
        for inPort in inData:
            self.commandChains[0][self.chnINPUTNAMES].append(inPort[1])
            inputToChainName[inPort[0]] = "editnode_output"

        if len(self.commandChains[0][self.chnINPUTNAMES]) < 1:
            cmds = []


        #### Add commands received in parameter 'ignoreThisParameter4ever', if any.
        incomingCommandsData = None
        params = self.getParameters()
        for param in params:
            if param.getName() == "ignoreThisParameter4ever":
                incomingCommandsData = param.getValue()
                
        try:
            if str(incomingCommandsData) == "" or str(incomingCommandsData) == "[]":
                incomingCommandsData = []
            else:
                incomingCommandsData = eval(str(incomingCommandsData))
        except Exception, e:
                incomingCommandsData = []

        tmp_added = 0        
        if isinstance(incomingCommandsData, list):
            tmp_count = str(len(incomingCommandsData))
            inCmds = self._EditNode__cleanAndOptimizeExtCmdInput(incomingCommandsData)
            tmp_prevExtCmds = copy.deepcopy(inCmds)
            inCmds = copy.deepcopy(self.previousExtCmds) + inCmds
            if isinstance(inCmds, list) and len(inCmds) > 0:
                lastChainIndex = -1
                for cmd in inCmds:
                    if isinstance(cmd, list) and len(cmd) == 2:
                        name = cmd[0]
                        args = cmd[1]
                        data = args.get("data")[0]
                        # [HACK]
                        data = inputToChainName[data]
                        # Decide what chain we want to add the command to.
                        # Inspect argument 'data' to do this. Match on input port.
                        added = 0
                        for chainIndex in range(0, len(self.commandChains)):
                            if self.commandChains[chainIndex][self.chnOUTPUTPORT] == data:
                                # We have a matching command chain. Add command.
                                self.commandChains[chainIndex][self.chnCOMMANDS].append(cmd)
                                tmp_added += 1
                                lastChainIndex = chainIndex
                                added = 1
                        if lastChainIndex != -1 and added == 0:
                            self.commandChains[lastChainIndex][self.chnCOMMANDS].append(cmd)
                        elif lastChainIndex == -1 and added == 0:
                            # See if there is a output port that matches, and if so, create a new chain.
                            chainAdded = 0
                            for outPort in self.outPorts:
                                if outPort.getName() == data:
                                    # Found output port.
                                    (status, msg) = self.addCommandChain(data, [data],inDataByNames.get(data)[self.dtoDIMS], data)
                                    if status == 1:
                                        self.addCommandToChain(data, name, args)
                                        tmp_added += 1
                                        chainAdded = 1
        #### Execute command-chains.
        # The code below is written specifically for version 0.4x.
        # The following limitations apply:
        #  * only one command chain (created by default in build())


        # Excute commands. There will be only one command if no other node sends commands.
        cmds = self.commandChains[0][self.chnCOMMANDS]

        # Check if the command must have multiple operands or not.
        multipleOperands = 0
        cmds[0][0] = str(self.lookupInternalName.get(str(cmds[0][0]), str(cmds[0][0])))
        name = cmds[0][0]
        args = getattr(self, "ARGS_" + name, None)
        if isinstance(args, list):
            # Look for internal argument '__#multipleOperands'.
            for arg in args:
                if arg[0] == "__#multipleOperands":
                    multipleOperands = 1
                    break
        # Fetch available operands.
        operand = self.commandChains[0][self.chnINPUTNAMES]

        if len(operand) == 1:
            if multipleOperands == 0:
                operand = operand[0]
            else:
                cmds = []
                operand = None
        elif len(operand) > 1:
                if multipleOperands == 0:
                    cmds = []
                    operand = None
                else:
                    # Do nothing. 'operand' is valid.
                    pass
            
        for cmd in cmds:
            name = cmd[0]
            args = cmd[1]
            args["data"] = operand
            (status, result) = self.execCmd(name, args)
            if status == 0:
                errorFreeExec = 0
                # Do not pass any data on.
                result = operand = None
                break
            operand = result
        result = operand

        if errorFreeExec:
            self.previousExtCmds = self.previousExtCmds + tmp_prevExtCmds
        # Find outport.
        # [NOTE] If user updates output port, the chain tuple must be updated as well!
        if isinstance(result, list):
            result = Numeric.array(result)
        result = Data.Data("editnode_output", result)
        outPort = None
        for port in self.outPorts:
            if port.getName() == self.commandChains[0][self.chnOUTPUTPORT]:
                outPort = port
                break
        if outPort != None:
            outPort.setData(result)
            outPort.setDataType(self._EditNode__textualTypeDescription(result.getData()))
            opHasValueSet[outPort] = 1

        #### Send list of available data to the port 'availableAliases' if it exists.
        inPortNames = []
        availableAliases = None
        for port in self.outPorts:
            if port.getName() == "availableAliases":
                availableAliases = port
        for port in self.inPorts:
            if port.getName() != "availableAliases":
                inPortNames.append(port.getName())
        if availableAliases != None:
            availableAliases.setData(Data.Data("availableAliases", inPortNames))
            opHasValueSet[availableAliases] = 1

        # Try to match ports that have no value with inport with same name.
        # This is essentially a pass-through function.
        for outPort in self.outPorts:
            if not opHasValueSet.has_key(outPort):
                for inPort in self.inPorts:
                    if inPort.getName() == outPort.getName():
                        opHasValueSet[outPort] = 1
                        outPort.setData(inPort.getData())

        Node.Node.execute(self)

    def toXML( self ):
        # Overload the Node.toXML() method to prevent 'ignoreThisParameter4ever' to saved to disk.
        for param in self.getParameters():
            if param.getName() == "ignoreThisParameter4ever":
                param.setValue("")
        return Node.Node.toXML( self )

    
    def isParametersReady(self):
        """This is a hack to pass MD validation"""
        return True
