#!/usr/bin/python
#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from System.Data import Data
from System.Data.Node import Node
from copy import copy

class MissingVtkError(Exception):
    pass

class GaussdalNode(Node):
    def __init__(self, name="Gaussdal_node"):
        Node.__init__(self, name)
        self.description = "GaussDal molecule mapping editor"
        self.outputPortName = "molecule"
        self.inputPortName= "molecule"
        self.create_input_port(self.inputPortName, "", "")
        self.create_output_port(self.outputPortName,"" , "Modified molecule data.")
        self.inputPortName2 = "molecule_mapping"
        self.outputPortName2 = "molecule_mapping"
        self.create_input_port(self.inputPortName2, "", "", "", True)
        self.create_output_port(self.outputPortName2, "", "Molecule mapping")
        self.inputPortName3 = "col_names"
        self.create_input_port(self.inputPortName3, "", "")
        self.create_output_port("col_names", "", "");
        self.is_executed = 0

        self.manual_mapping = False
        self.molecule_mapping = []
        self.inputport_mapping = False
        self.col_names = None

        # Specify per default which columns contain which data
        self.col_mol_id = 0
        self.col_atom_no = 1
        self.col_x = 2
        self.col_y = 3
        self.col_z = 4
        self.local_data_cols = [5]
        self.global_data_cols = []

        self.molecule_list = None
        self.molecule_browser = None

        self.old_table = []
    
    def get_output_col_names(self):
        """Returns the column names of the output table after mapping"""
        col_names = []
        for i in self.global_data_cols:
            col_names.append(self.col_names[i])
        n_mapped_atoms = len(self.find_smallest_molecule().atoms)
        if self.manual_mapping:
            n_mapped_atoms = len(self.molecule_list[0].mapping)
        for atom in range(n_mapped_atoms):
            for i in self.local_data_cols:
                col_names.append(self.col_names[i] + str(atom + 1))
        return col_names

    def make_table(self):
        """Return the resulting output table after the mapping of atoms.
        """
        n_molecules = len(self.molecule_list)

        # Find the most #atoms in a molecule
        n_atoms = 0
        for molecule in self.molecule_list:
            if len(molecule.atoms) > n_atoms:
                n_atoms = len(molecule.atoms)

        # Create the resulting table to put on the port
        output = []
        for m in self.molecule_list:
            row = []
            mapping = {}
            #if mapping method is set to manual
            if self.manual_mapping:
                mapping = m.mapping
            elif self.inputport_mapping and not self.manual_mapping:
                mapping = m.port_mapping
            else:
                for k in range(len(self.find_smallest_molecule().atoms)):
                    mapping[k+1] = k
            # Global data as first columns
            for i in self.global_data_cols:
                row.append(m.atoms[0].data[i])
            for i in mapping:
                out = []
                for j in self.local_data_cols:
                    if j == self.col_x:
                        out.append(m.atoms[mapping[i]].coords[0])
                    elif j == self.col_y:
                        out.append(m.atoms[mapping[i]].coords[1])
                    elif j == self.col_z:
                        out.append(m.atoms[mapping[i]].coords[2])
                    else:
                        out.append(m.atoms[mapping[i]].data[j])
                row += out
            output.append(row)
        return output
        
    def get_molecules(self):
        return self.molecule_browser

    def update_molecules(self):
        self.molecule_list = self.make_molecule_list(self.data)
        self.ref_mol_browser = Molecule_browser(self.molecule_list)  # left molecule
        self.test_mol_browser = Molecule_browser(self.molecule_list) # right molecule
        self.molecule_browser = Molecule_browser(self.molecule_list)
        return self.molecule_browser
        
    def execute(self):        
        """Is executed when the user presses play"""
        try:
            import vtk
        except ImportError:
            raise MissingVtkError("Running GaussdalNode requires VTK with python bindings to be installed.")
            
        if not self.is_executed:
            # Get input port data
            inport = self.getInputPortByName(self.inputPortName)
            self.data = inport.getData().getData()
            self.old_table = copy(self.data)
        
            #old_table = inport.getData().getData()
            #print 'done collecting molecule data from port'

            self.update_molecules()

            col_inport = self.getInputPortByName(self.inputPortName3)
            self.col_names = col_inport.getData().getData()
        else:
            inport = self.getInputPortByName(self.inputPortName)
            new_table = inport.getData().getData()
            if not new_table == self.old_table:
                self.data = new_table
                self.old_table = new_table
        # get mapping data from input port if any
        inport2 = self.getInputPortByName(self.inputPortName2)
        if inport2.getData():
            self.molecule_mapping = inport2.getData().getData()
            self.make_mapping_from_port(self.molecule_mapping)

        output_data = self.make_table()

        # send modified molecules to all outputs (should be only one)
        #for outPort in self.outPorts:
        #    outPort.setData(Data.Data("variable", output_data))

        molecule_outport = self.getOutputPortByName(self.outputPortName)
        if molecule_outport:
            molecule_outport.setData(Data.Data("variable", output_data))

        mapping_outport = self.getOutputPortByName(self.outputPortName2)
        if mapping_outport:
            mapping_outport.setData(Data.Data("variable", self.molecule_mapping))

        self.getOutputPortByName("col_names").setData(Data.Data("variable", self.get_output_col_names()))

        self.is_executed = 1
        Node.execute(self)
        
    def make_molecule_list(self, table):
        """Traverses the table of input molecules (moleculeid, atomic_number, x, y, z)
        and returns a list of molecules.
        """
        coord_cols = (self.col_x, self.col_y, self.col_z)
        molecules = []
        molecule_data = {}
        for row in table:
            molecule_data[row[self.col_mol_id]] = []
        for row in table:
            molecule_data[row[self.col_mol_id]].append(row)
        for moleculeid in molecule_data:
            molecules.append(Molecule(
                moleculeid,
                molecule_data[moleculeid],
                self.col_atom_no,
                coord_cols))

        return molecules

    def find_smallest_molecule(self):
        """Finds the molecule with least number of atoms and
        returns this molecule. This will be reference molecule
        for the automatic molecule mapping
        """
        if self.molecule_list:
            smallest_molecule = self.molecule_list[0]
            for molecule in self.molecule_list:
                
                if len(molecule.atoms) < len(smallest_molecule.atoms):
                    smallest_molecule = molecule
        else:
            smallest_molecule = None

        return smallest_molecule

    def make_arbitrary_mapping(self):
        """Maps the first n atoms of each molecule,
        n being the number of atoms in the smallest molecule.
        """
        reference_molecule = self.find_smallest_molecule()
        atoms_to_be_mapped = len(reference_molecule.atoms)
        mapped = {}
        for molecule in self.molecule_list:
            mapped = {}
            for i in range(atoms_to_be_mapped):
                mapped[i+1] = i
                #molecule.select_atom(molecule.atoms[i][1])
            molecule.mapping = mapped
                
    def make_mapping_from_port(self, mapping_matrix):
        """Maps atoms according to provided mapping matrix,
        where each row contains a vector representing a
        molecule. molecule_list must have same order as
        mapping_matrix..
        """
        format_ok = len(mapping_matrix) is len(self.molecule_list)
        if format_ok:
            row = 0
            for mol in self.molecule_list:
                col = 0
                mapping = {}
                for i in range(len(mapping_matrix[col])):
                    if not mapping_matrix[row][col] == 0:
                        mapping[col+1] = int(mapping_matrix[row][col]) #col+1?
                    col += 1
                row += 1
                mol.port_mapping = mapping
            #self.set_port_mapping()
            
        

    def procrustes(self, reference_molecule_id):
        from Library.Util.molecular import Procrustes
        reference_molecule = self.molecule_list[reference_molecule_id]
        reference_coords = reference_molecule.get_coords()
        reference_mapping = reference_molecule.mapping
        procrustes = Procrustes(reference_coords, reference_mapping)
        for molecule in self.molecule_list:
            coords = molecule.get_coords()
            mapping = molecule.mapping
            procrus_coords = procrustes.procrustes(coords, mapping)
            molecule.set_coords(procrus_coords)
            molecule.update_model()
            molecule.update_actors()

    def set_default_mapping(self):
        self.manual_mapping = False
        self.inputport_mapping = False

    def set_manual_mapping(self):
        self.manual_mapping = True

    def set_port_mapping(self):
        self.inputport_mapping = True
        #self.manual_mapping = False
                
class Atom:
    def __init__(self, atomic_n, id, coords, actor, data):
        self.textActor = None
        self.id = id
        self.selected = 0
        self.actor = actor
        self.coords = coords
        self.data = data
        self.atomic_n = atomic_n

        # imported here to make vtk optional
        from vtk import vtkVectorText
        # mapping number for rendering
        self.text = vtkVectorText()
    
    def select(self, as_index, camera):
        """Invert color and annotate the atom with an index.

        camera - the camera in the QVTKRenderWindowInteractor which the index
        actor is always to face."""

        # imported here to make vtk optional
        from vtk import vtkFollower, vtkPolyDataMapper
        
        self.selected = not self.selected

        # negate color
        color = self.actor.GetProperty().GetColor()
        newColor = []
        newColor.append(1 - color[0])
        newColor.append(1 - color[1])
        newColor.append(1 - color[2])
        self.actor.GetProperty().SetColor(newColor)

        if self.selected:
            self.text.SetText(str(as_index))
            textMapper = vtkPolyDataMapper()
            textMapper.SetInput(self.text.GetOutput())
            self.textActor = vtkFollower()
            self.textActor.SetMapper(textMapper)
            self.textActor.SetScale(0.4, 0.4, 0.4)
            self.textActor.AddPosition(self.coords)
            self.textActor.AddPosition([0.0, -0.6, 0.5])
            self.textActor.GetProperty().SetColor([1.0, 0.0, 0.0])
            self.textActor.SetCamera(camera)

    def deselect(self):
        self.selected = not self.selected

        # Negate color
        self.actor.GetProperty().SetColor(map(lambda a:1-a, self.actor.GetProperty().GetColor()))

        # Delete index actor
        del self.textActor
        self.textActor = None
        #print 'self.textActor: ', self.textActor

def make_model(atoms, coords):
    """Return a chempy model

    atoms - a list of atomic numbers
    coords - a list of atom coordinates
    """
    from Library.Util import molecular
    from chempy.models import Indexed
    from chempy import Atom,Bond

    model = Indexed()

    # append the atoms onto it
    for i in range(len(atoms)):
        new_atom = molecular.createAtom(molecular.atomic_no2symbol(atoms[i]))

        # This id must be consistent across instances of the same molecule.
        # An atom id may also be included in the input data table.
        # Could possibly be based on coordinates.
        new_atom.id = i

        model.atom.append(new_atom)

    for i in range(len(atoms)):
        model.atom[i].coord = coords[i]

    # Bonds
    for i in range(len(atoms)):
        for j in range(len(atoms)):
            if i != j:
                x = coords[i][0] - coords[j][0]
                y = coords[i][1] - coords[j][1]
                z = coords[i][2] - coords[j][2]
                d = molecular.alen((x, y, z))
                bond_order = molecular.bond_order(atoms[i], atoms[j], d)
                if bond_order > 0:
                    new_bond = molecular.createBond(i, j, bond_order)
                    model.bond.append(new_bond)
    return model
    
class Molecule:
    def __init__(self, id, atom_table, col_atom_no, coords):
        self.id = id
        self.atom_table = atom_table
        self.col_atom_no = col_atom_no
        self.coord_cols = coords
        self.atoms = []
        self.mapping = {}
        self.port_mapping = {}
        self.i = 0

        # VTK actors
        self.actors = []
        self.model = self.make_model()
        self.update_actors()

        for i in range(len(atom_table)):

            # Find the atom's actor
            atom_actor = None
            for actor in self.actors:
                if hasattr(actor, "id"):
                    if actor.id == i:
                        atom_actor = actor

            self.atoms.append(
                Atom(
                    atom_table[i][col_atom_no],
                    i,
                    (atom_table[i][coords[0]], atom_table[i][coords[1]], atom_table[i][coords[2]]),
                    atom_actor,
                    atom_table[i]))

    def update_model(self):
        atoms = []
        coords = []
        for atom in self.atoms:
            atoms.append(atom.atomic_n)
            coords.append(atom.coords)
        self.model = make_model(atoms, coords)

    def update_actors(self):
        from GUI.Plots.Plots3D.MolPlot3D.VtkGenerator import VtkGenerator
        hidden_atoms = []
        (self.actors, hidden_actors, bond_actors) = VtkGenerator(self.model).getBallAndStickActors(hidden_atoms)
        
    def get_actors(self):
        # Molecule balls and sticks
        actors = []
        actors += self.actors

        # atom selection index actors
        for atom in self.atoms:
            actor = atom.textActor
            if actor:
                actors.append(atom.textActor)
        return actors

    def select_atom(self, atom_id, camera):
        """camera - the camera in the QVTKRenderWindowInteractor which the index
        actor is always to face.
        """
        if not self.atoms[atom_id].selected:
            self.i += 1
            self.mapping[self.i] = atom_id
            self.atoms[atom_id].select(self.i, camera)
        else:
            del self.mapping[self.i]
            self.atoms[atom_id].select(self.i, camera)

    def deselect_latest_selected_atom(self):
        # Delete last mapping entry
        if self.mapping:
            last_mapping_idx = self.mapping.keys()[-1]
            last_selected_atom = self.mapping[last_mapping_idx]
            del self.mapping[last_mapping_idx]
            # Update atom actors
            self.atoms[last_selected_atom].deselect()

            self.i -= 1

    def clear_atom_selection(self):
        # Delete all mapping entries for the current molecule
        if self.mapping:
            mapping_indexes = self.mapping.keys()
            atom_indexes = []
            for index in mapping_indexes:
                atom_indexes.append(self.mapping[index])
            for atom_index in atom_indexes:
                self.atoms[atom_index].deselect()
            self.i -= len(self.mapping)
            self.mapping = {}

    def get_coords(self):
        """Returns a list of the coordinates of the molecule's atoms"""
        coords = []
        for a in self.atoms:
            coords.append(a.coords)
        return coords

    def set_coords(self, coords):
        for i in range(len(self.atoms)):
            self.atoms[i].coords = coords[i]

    def make_model(self):
        """Makes a chempy molecule model from the data that is received from an InputNode which is reading a GaussDal data table."""
        atoms = []
        coords = []
        for row in self.atom_table:
            atoms.append(row[self.col_atom_no])
            coords.append([row[self.coord_cols[0]], row[self.coord_cols[1]], row[self.coord_cols[2]]])
        
        return make_model(atoms, coords)
        
        

class Molecule_browser:
    """Convenience class for browsing through lists"""
    def __init__(self, molecules, init=0):
        self._test_molecule_index = -1
        self._reference_index = 0
        self.molecules = molecules
        self.i = init
        self.moleculeids = []
        for id in self.molecules:
            self.moleculeids.append(id)

    def _loop(self, n):
        return n % len(self.molecules)

    def molecule_gen(self):
        """Yields molecule instances"""
        for m in self.molecules:
            yield m

    def current(self):
        """Returns the current molecule object."""
        return self.molecules[self._loop(self.i)]

    def current_index(self):
        return self.i

    def next(self):
        self.i += 1
        return self.molecules[self._loop(self.i)]
        
    def previous(self):
        self.i -= 1
        return self.molecules[self._loop(self.i)]

    def specific_by_id(self, id):
        """2DO: continue in sequence from this moleculeid when next()"""
        for m in self.molecules:
            if m.id == id:
                return m

    def specific_by_index(self, n):
        """2DO: continue in sequence from this moleculeid when next()"""
        self.i = n
        return self.molecules[self._loop(self.i)]

    def specific(self, n):
        """2DO: continue in sequence from this moleculeid when next()"""
        return self.molecules[n]
        
    def get_reference_molecule(self):
        return self.molecules[self._reference_index]
    
    def set_reference_molecule_by_index(self, n):
        self._reference_index = n
        
    def get_next_test_molecule(self):
        """Returns the first molecule which is not a reference molecule
        and sets it as the current molecule
        """
        self._test_molecule_index += 1
        self._test_molecule_index %= len(self.molecules)
        while self._reference_index == self._test_molecule_index and self._test_molecule_index < len(self.molecules):
            self._test_molecule_index += 1
        self._test_molecule_index %= len(self.molecules)
        self.i = self._test_molecule_index
        return self.molecules[self._test_molecule_index]

