# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import copy
import re
import traceback

from System.Data.Data import Data
from System.Data.Node import Node

class MoleculeEditorNode(Node):
    """This class provides a number of commands to manipulate molecules."""

    def __init__(self, name):
        Node.__init__(self, name)
        self.description = "This molecule editor node makes it possible to manipulate molecules before passing them on. The node can perform a set of operations specified interactively."

        self.history = []
        self.cachedData = []
        self.originalData = []

        self.inputPortName = "molecule"
        self.outputPortName = "molecule"
        
        self.create_input_port(self.inputPortName, "", "Source molecule data.")
        self.create_output_port(self.outputPortName,"" , "Modified molecule data.")
        

    def applyCommand(self, functionObject):
        """Apply function on molecule-models. Function is expected to
        take a single parameter, the models to be modified.
        """

        try:
            functionObject(self.cachedData)
            self.history.append(functionObject)
        except StandardError,e:
            print "Error:",str(e),dir(e)
            traceback.print_exc()


    # Should this method be used through rpc/rmi, a patch-based scheme
    # should be employed to keep from sending the entire model each
    # time.
    def getMoleculeModels(self):
        return self.cachedData


    def execute(self):
        """Running the node fills it with a new molecule, and
        simultaneously passes it on. If it already received the same
        molecule before, and alterations have been performed on it,
        pass on the alterations instead."""
        
        # copy all molecules from all inputports (should be only one input port though)
        data = []
        models = []

        inPort = self.getInputPortByName(self.inputPortName)
        models = inPort.getData().getData()

        try:
#         if data has changed:
            if self.shouldUpdateData(models):
                
#             copy data
#             reset commands/history / ignore new commands
                self.originalData = copy.deepcopy(models)
                self.cachedData = copy.deepcopy(models)
        except:
            traceback.print_exc()

        # send modified molecules to all outputs (should be only one)
        for outPort in self.outPorts:
            #Make sure to send a copy to outputport, so next node won't
            #see modifications done in this node before running diagram.
            outPort.setData(Data("variable", copy.deepcopy(self.cachedData)))
            
        Node.execute(self)
        

    def shouldUpdateData(self, newData):
        if not self.originalData: return True
        if len(newData) != len(self.originalData): return True

        for modelIndex in range(len(newData)):
            if len(newData[modelIndex].atom) != len(self.originalData[modelIndex].atom):
                print "Number of atoms were different."
                return True
            for atomIndex in range(len(newData[modelIndex].atom)):
                newSignature = newData[modelIndex].atom[atomIndex].get_signature()
                originalSignature = self.originalData[modelIndex].atom[atomIndex].get_signature()
                if newSignature != originalSignature:
                    print "Two atoms were different:",originalSignature,"and", newSignature
                    return True
            if len(newData[modelIndex].bond) != len(self.originalData[modelIndex].bond):
                print "Number of bonds were different."
                return True
            for bondIndex in range(len(newData[modelIndex].bond)):
                newBond = newData[modelIndex].bond[bondIndex]
                originalBond = self.originalData[modelIndex].bond[bondIndex]
                if newBond.index != originalBond.index:
                    print "Two bonds were different:", str(oldBond.index),"and",str(originalBond.index)
                    return True
                if newBond.order != originalBond.order:
                    print "Two bonds were different in order:", str(oldBond.order),"and",str(originalBond.order)
                    return True
                    
                
        return False

            
            
