# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from System.FileHandlers.Input.matfilereader import MatFileReader
from System.FileHandlers.Output.OctaveFileWriter import OctaveFileWriter
from System.Data.Data import Data
from System.Exceptions.ZPluginError import ZPluginError
from System.Plugin import Plugin, PluginInterrupted
from Library.Util.Config import Config
from Library.Util.Counter import Counter
from Library.Util.InterruptibleCommand import InterruptibleCommand, NoSuchExecutableError

import os
import re
import tempfile
import time

class OctavePlugin(Plugin):
    """Plugin responsible for running octavescripts and returning data."""

    def __init__(self):
        Plugin.__init__(self)
        self.fileWriter = OctaveFileWriter()
        self.fileReader = MatFileReader()
        self._counter = Counter()
        
        self.scriptDescriptor = 0
        self.scriptPath       = 0
        self.outputDescriptor = 0
        self.outputPath       = 0
        self.inputDescriptor  = 0
        self.inputPath        = 0

        # holds stdout/err from last run
        self._last_stdout = "Not run"
        self._last_stderr = "Not run"
        
    def execute(self, functionName, functionFileName, inputPorts,
                outputPorts, parameters, interruptCallback):
        """Execute octave script and return result.

        Return:
            - List[Data] - a List with Data-objects which are the results from the octave-process
                 
        This method is responsible with executing an octave script.
        It will first write he scriptfile which can be used by octave,
        then it will parse the data returned by octave and return this
        as Data-objects.
        """

        variables = self._collect_variables(inputPorts, parameters)
        output_port_names = map(lambda port : port.getName(), outputPorts)

        path, script_name = os.path.split(functionFileName)
        # split "jalla.balla.m" into ("jalla.balla", ".m")
        function_name = os.path.splitext(script_name)[0]

        return self.execute_function(function_name, variables,
                                     output_port_names, interruptCallback, path)

    def execute_function(self, function_name, variables, output_names,
                        interruptCallback=lambda:False, file_path=None):
        """Execute named octave function and return result as Data objects.

        function_name - name of function to call
        variables - list of name/value tuples
        output_names - names of output data objects
        interruptCallback - predicate function to see if execution should stop

        Results are returned as a list of Data objects, each object is
        named from output_names.

        example:
        >>> op = OctavePlugin()
        >>> output = op.execute_function("rand", (('N', 2), ('M', 3)), ("out",))
        >>> output[0].getData()
        array([[ 0.99374181,  0.42468256,  0.31429112],
               [ 0.92701024,  0.42134142,  0.45148054]])
        
        """
        (self.scriptDescriptor, self.scriptPath) = tempfile.mkstemp()
        (self.outputDescriptor, self.outputPath) = tempfile.mkstemp()
        (self.inputDescriptor,  self.inputPath)  = tempfile.mkstemp()
        
        self._write_script_file(function_name, variables, output_names, file_path)        
        # running the command
        try:
            start_time = time.time()
            try:
                cmd = InterruptibleCommand("octave", "-q", self.scriptPath)
            except NoSuchExecutableError:
                raise NoSuchExecutableError, """Unable to find the Octave executables.
                It seems as if Octave is not installed on your system."""
            cmd.run_interruptible()

            error_string = cmd.read_stderr()
            while error_string == 0:
                ## check if we are going to interrupt
                if interruptCallback():
                    cmd.interrupt()
                    raise PluginInterrupted("The octave process was interrupted") 
                time.sleep(0.1)
                # check if stdout is read and readable, this is needed
                # with intensive octave processing or else the QProcess
                # will never finish
                cmd.check_stdout()
                error_string = cmd.read_stderr()

            self._last_stdout = cmd.read_stdout()
            self._last_stderr = error_string
            # command has finished running
            # now, have to check if there was a warning in the output from Octave

            has_error = (re.search("(.*?)error(.*?)", error_string) != None)
            has_missing_file = error_string.find("No such file or directory") != -1

            if has_error and not has_missing_file:
                self.setUnlocked()
                raise ZPluginError, "Octave returned with an error when running the script: %s.\n\nError reported:\n%s" % (function_name, error_string)

            outData = []
            if len(output_names) > 0:
                outData = self.fileReader.read_data(self.outputPath)
        finally:
            # now, delete the temp-files generated
            self.clean()

        return outData

    def _write_script_file(self, function_name, variables, output_names, file_path):
        try:
            scriptFile = os.fdopen(self.scriptDescriptor, "w")
        except:
            raise ZPluginError.ZPluginError, str(e)

        loadPath = ""

        # start writing the scriptfile
        # appending Config.loadpath (which is optional) and
        # systemfunctiondir to the octave loadpath
        
        paths = [Config.systemfunctiondir]
        if file_path:
            paths.append(file_path)
        
        if Config.loadpath:
            for p in Config.loadpath.split(";"):
                paths.append(p)
                
        for element in paths:
            if os.name == "nt":
                element = os.path.abspath(element)
                element = element.replace(":", "")
                loadPath += ":/cygdrive/"
                loadPath += '/'.join(str(element).split(os.sep)) + "//"
            else:
                loadPath += ":" + '/'.join(str(element).split(os.sep)) + "//"
            
        scriptFile.write('path("'+ loadPath + '", LOADPATH);\n')

        if variables:
            variable_dict = {}
            for name, value in variables:
                variable_dict[name] = value

            self.fileWriter.write_variable_dict(variable_dict, self.inputPath)

            #writing input-line
            scriptFile.write("load " + self.inputPath + "\n")

        outVariable = []
        for name in output_names:
            #fetching names on the output variables
            outVariable.append(str(name))

        tempOut = re.sub("\[", "", str(outVariable))
        tempOut = re.sub("\]", "", str(tempOut))

        #writing function-line
        if tempOut:
            scriptFile.write("[" + re.sub("'", "", str(tempOut)) + "] = " +
                             function_name + "(")
        else:
            scriptFile.write(function_name + "(")
            
        if variables:
            names = map(lambda var : var[0], variables)
            tempIn = re.sub("\[", "", str(names))
            tempIn = re.sub("\]", "", str(tempIn))
            scriptFile.write(re.sub("'", "", str(tempIn)))
        scriptFile.write(");\n")

        #writing output-line
        scriptFile.write("save -mat-binary " + self.outputPath + " " +
                         re.sub(",", "", re.sub("'", "", str(tempOut))) + "\n")
        scriptFile.close()

        #done writing the scriptfile

    def _collect_variables(self, inputPorts, parameters):
        """Gather variable names and values from ports and parameters.

        Create a list with the values from parameters and input ports
        we want to use.  Parameters with a value set gets precedence
        over port with same name.

        Returns list of name/value pairs of variables.
        """
        variables = []
        
        for port in inputPorts:
            data = port.getData()
            variables.append((self._get_unique_name(data.getName()), data.getData()))
        
        for param in parameters:
            if param.isValueSet():
                variables.append((self._get_unique_name(param.getName()), param.getValue()))
        return variables

    def _get_unique_name(self, name):
        return str(name) + '_' + str(self._counter.get_next_value())
    

    def clean(self):
        self.setUnlocked()
        try:
            if os.path.exists(self.scriptPath):
                os.remove(self.scriptPath)
            if os.path.exists(self.outputPath): 
                os.fdopen(self.outputDescriptor).close()
                os.remove(self.outputPath)            
            if os.path.exists(self.inputPath):
                os.fdopen(self.inputDescriptor).close()
                os.remove(self.inputPath)
        except IOError, e:
            print 'Could not delete file: ', e
            
             
    def get_stdout_from_last_run(self):
        return self._last_stdout

    def get_stderr_from_last_run(self):
        return self._last_stderr
