# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import re
import time
import sys

from System.Plugin import Plugin, PluginInterrupted
from System.Data import Data
from System.Exceptions import ZPluginError

def _run_function(script, function_name, variables):
    old_globals = dict(globals())

    try:
        execfile(script, globals())
        retval = globals()[function_name](*variables)
    finally:
        globals().clear()
        globals().update(old_globals)
    return retval

class PythonPlugin(Plugin):

    def __init__(self):
        """Constructor which creates a PythonPlugin-object"""
        Plugin.__init__(self)
        self._start_time = 0
        
        # holds what was printed to stdout/err during last run
        self._last_stdout = "Not run"
        self._last_stderr = "Not run"

    def _clean_dprint_and_raise(self, current_action):
        self.clean()
        raise PluginInterrupted, "PythonPlugin interrupted - " + current_action

    def get_stdout_from_last_run(self):
        """Returns what the python plugin wrote to stdout during last run."""
        return self._last_stdout

    def get_stderr_from_last_run(self):
        """Returns what the python plugin wrote to stderr during last run."""
        return self._last_stderr

    def execute(self, functionName, functionFileName, nodeInputs,
                nodeOutputs, nodeParams, interruptCallback):
        """Execute python script and return result.

        Return:
            - List[Data] - a List with Data-objects which are the results from the python-script.
                  
        General usage: When making a python-script for use as a
        function within SciCraft, you can do pretty much anything you
        want. However, if your function returns more than one
        variable, you have to return them as elements in a
        dictionary. If your function returns one variable,
        PythonPlugin will assume it is supposed to have the same name
        as the output-port. In case there are more than one
        output-port, and the result from the function is not a
        dictionary, the plugin will raise an error, claiming the given
        function and xml-declaration are inconsistent.
        
        Also note that objects outside the function-shell cannot be
        referenced from whithin the function.  This means that
        e.g. imports need to be done within the function.
        """

        variables = []
        for inp in nodeInputs:
            variables.append(inp.getData().getData())

        for par in nodeParams:
            variables.append(par.getValue())

        self._start_time = time.time()

        if interruptCallback():
            self._clean_dprint_and_raise("before retrieving data")

        func = functionName

        if interruptCallback():
            self._clean_dprint_and_raise("before executing")

        try:
            res = _run_function(functionFileName, functionName, variables)

        except PluginInterrupted, e:
            raise e
        except Exception, e:
            import sys, traceback
            trace_string = ""
            for level in traceback.format_tb(sys.exc_info()[2]):
                trace_string += level
            trace_string += str(e)
            print "exception in pythonplugin\n", trace_string, e
            raise ZPluginError.ZPluginError, "Error when running python-script:\n%s" % str(e)

        data = []
        if not isinstance(res, dict):
            numOfOutputs = len(nodeOutputs)            
            if numOfOutputs > 1:
                print "Exception in pythonplugin. Method declaration inconsistent with actual method behavior. No dictionary found!"
                raise ZPluginError, 'Inconsistency between returned data from function and xml-declaration.'
            elif numOfOutputs == 1:
                data.append(Data.Data(nodeOutputs[0].getName(), res))
                return data
        else:
            for key in res.keys():
                data.append(Data.Data(key, res[key]))
        return data

            
