# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os
import sys
import time
import tempfile
from threading import Lock

from Library.Util.file import StreamHolder
from Library.Util.rtester import RTester
from Library.Util.Config import Config
from Library.Util.rpyclient import rpy_eval, rpy_silent_eval
from System.Plugin import Plugin, PluginInterrupted, get_python_command
from System.Data.Data import Data, RData
from System.Exceptions.ZPluginError import ZPluginError
from System.FileHandlers.Input.RFileReader import RFileReader

r_mutex = Lock()
    
class RPlugin(Plugin):

    def __init__(self):
        """Constructor which creates an RPlugin-object"""
        Plugin.__init__(self)

        RTester.r_install_test()
        global rpy
        import rpy

        # Placeholders for output from the function.
        self._last_stdout = "Not run"
        self._last_stderr = "Not run"

    def execute(self, function_name, function_filename, input_ports,
                output_ports, parameters, interrupt_callback):

        r_mutex.acquire()
        try:
            try:
                function_filename = function_filename.replace('\\', '/')                

                # Non-R data needs to be "loaded" into R. We keep a
                # list of references to the RData to keep it from
                # being garbage-collected through the RData
                # destructor.
                wrapped_data = []
                input_names = []
                for port in input_ports:
                    data = port.getData()
                    if not isinstance(data, RData):
                        data = RData(data.getName(), data.getData())
                        wrapped_data.append(data)

                    input_names.append(data.get_pointer())

                for param in parameters:
                    wrapped_data.append(RData(param.getName(), param.getValue()))
                    input_names.append(wrapped_data[-1].get_pointer())

                output_name = self._get_unique_output_name()

                self._test_loading_of_function(function_name, function_filename)

                if interrupt_callback():
                    raise PluginInterrupted, "RPlugin interrupted during execution"
        
                old_path = os.getcwd()
                try:
                    rpy_silent_eval("setwd('%s')" % os.path.dirname(str(function_filename)))
                    rpy_silent_eval("source('%s')" % os.path.basename(function_filename))
                    rpy_silent_eval("%s <- %s(%s)" %(output_name, function_name, ",".join(input_names)))
                    names = rpy_eval("names(%s)" % output_name) or []
                    if not isinstance(names, list):
                        names = [names]

                    data_list = []
                    for name in names:
                        data = RData(name)
                        rpy_silent_eval('%s<-%s$%s' %(data.get_pointer(), output_name, name))
                        data_list.append(data)
                    rpy_silent_eval("rm(%s)" % output_name)
                    rpy_silent_eval("gc()")

                finally:
                    try:
                        rpy_silent_eval("setwd('%s')" % old_path.replace("\\", "/"))
                    except Exception, e:
                        print e
                        rpy_silent_eval("setwd('%s')" % Config.scicrafthome.replace("\\", "/"))

            except PluginInterrupted:
                raise            
            except Exception, e:
                raise ZPluginError, 'Got the following error when ' + \
                      'executing %s:\n%s' % (function_name, str(e))
        finally:
            self.clean()
            r_mutex.release()
            
        return data_list

    def _get_unique_output_name(self):

        unique = False

        while not unique:
            test_name = os.path.basename(tempfile.mktemp())
            test_name = test_name.replace('-','')
            if test_name not in rpy_eval("ls()"):
                rpy_silent_eval('%s <- "reserved"' % test_name)
                unique = True
                

        return test_name

    def _test_loading_of_function(self, function_name, function_filename):

        try:
            rpy_silent_eval("source('%s')" % function_filename)            
            rpy_silent_eval("rm(%s)" % function_name)
            rpy_silent_eval("gc()")
        except rpy.RException, e:
            raise ZPluginError, \
                  "Error when loading function:\n" + \
                  "Function file: " + function_filename + "\n" + \
                  "Error reported: " + str(e)
        
    def clean(self):
        pass
    
    def get_stdout_from_last_run(self):
        return self._last_stdout

    def get_stderr_from_last_run(self):
        return self._last_stderr

        
