# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os, sys, types, re
from qt import QStyleFactory, qApp

class ConfigMeta(type):
    def __getattr__(self, key):
        """Control where to look for attributes, and in what order."""
        # Not very efficient access, very cacheable if necessary,
        # currently readability is more important though.
        
        attributes = {}
        for k in (self.readonly,
                  self.defaults,
                  self.modified):
            attributes.update(k)
        if key in attributes:
            return attributes[key]

        raise AttributeError, key

    def __setattr__(self, key, value):
        if key in self.readonly:
            pass # value is silently ignored
        elif key in self.defaults:
            self.modified[key] = value
        else:
            raise AttributeError, "'Config' object has no settable attribute '%s'"%key
        

class Config:
    """Class used to take care of configuration, mostly related to the GUI.
    Config is used as a singleton through the class interface and never
    instantiated.
    
    Usage:

    Config.resetDefaults() - resets all settings to default
    Config.key = value     - set value
    Config.key => value    - get value
    """

    __metaclass__ = ConfigMeta


    # Default values. These dicts should not have overlapping values.
    # Keys not found in one of these dicts should not be set by user,
    # this is not currently enforced though.
    
    # normal defaults
    defaults = {}
    # values not settable by user
    readonly = {}

    # Values modified by user or configuration file, only these values
    # will be saved back to file.
    modified = {}

    def setDefaults(cls):
        """Set all default values. Should not be called outside
        module. To reset to defaults, use Config.resetToDefaults()."""

        cls.readonly['appname'] = "SciCraft"
        cls.readonly['version'] = "0.18.0"
        # Hack to find main scicraft directory
        import Library.Util.Config
        path = os.path.dirname(Library.__file__)[:-len("Library")]
        # Remove 'library.zip' from the path which will be there
        # after building scicraft with py2exe on Windows.
        start_zip = path.find('library.zip')
        if start_zip > -1:
            path = path[:start_zip]
        cls.readonly['scicrafthome'] = path
    
        
        cls.readonly['globalposixconf'] = '/etc/scicraft.conf'
        if os.name == "posix":
            cls.readonly['homedir'] = os.environ["HOME"] 
            cls.readonly['scicraftuserdir'] = ".scicraft"
            cls.readonly['rcfile'] = "scicraftrc"
        else:
            cls.readonly['homedir'] = os.environ["USERPROFILE"] 
            cls.readonly['scicraftuserdir'] = "SciCraft"         
            cls.readonly['rcfile'] = "scicraftrc.txt"

        cls.readonly['userconfigdir'] = os.path.join(cls.readonly['homedir'], cls.readonly['scicraftuserdir'])

        cls.readonly['nodeconfdir'] = os.path.join(cls.readonly['scicrafthome'], "nodeconf")
        cls.readonly['pixmapdir'] = os.path.join(cls.readonly['scicrafthome'], "icons", "nodes")
        cls.readonly['iconsdir'] = os.path.join(cls.readonly['scicrafthome'], "icons", "pixmaps")
        cls.readonly['cursorsdir'] = os.path.join(cls.readonly['scicrafthome'], "icons", "cursors")
        cls.readonly['datadir'] = os.path.join(cls.readonly['scicrafthome'], "data")
        cls.readonly['singlefragmentdir'] = os.path.join(cls.readonly['datadir'], "fragments", "single")
        cls.readonly['doublefragmentdir'] = os.path.join(cls.readonly['datadir'], "fragments", "double")
        cls.readonly['htmldocpath'] = os.path.join(cls.readonly['scicrafthome'], "doc", "manual", "html")
        cls.readonly['systemfunctiondir'] = os.path.join(cls.readonly['scicrafthome'], "toolboxes")
        cls.readonly['customplotdir'] = os.path.join(cls.readonly['scicrafthome'], "customplots")
        cls.readonly['customplot2ddir'] = os.path.join(cls.readonly['userconfigdir'], "customplots")
        cls.readonly['userplotdir'] = os.path.join(cls.readonly['userconfigdir'], "userplots")
        cls.readonly['plotdir'] = os.path.join(cls.readonly['scicrafthome'], "Library", "plots")
        cls.readonly['uidir'] = os.path.join(cls.readonly['scicrafthome'], "ui")
        
        cls.defaults['nodeStyle'] = ""
        cls.defaults['loadpath'] = ""

        cls.defaults['last_path_dataset'] = ""
        cls.defaults['last_path_modulediagram']  = ""       
 
        cls.defaults['cachepath'] = cls.readonly['userconfigdir']
        cls.defaults['cacheenabled'] = False
        cls.defaults['cachewarnlimit'] = 100 # megabytes

        cls.defaults['enable_keep_output'] = True

        cls.defaults['style']="QPlatinumStyle"

        cls.defaults['showLabel']=1

        cls.defaults['bgR']=208
        cls.defaults['bgB']=208
        cls.defaults['bgG']=208

        cls.defaults['appX'] = 30
        cls.defaults['appY'] = 30
        cls.defaults['appWidth'] = 800
        cls.defaults['appHeight'] = 640


        cls.defaults['lineB'] = 255
        cls.defaults['lineG'] = 255
        cls.defaults['lineR'] = 255

        cls.defaults['rectB'] = 0
        cls.defaults['rectG'] = 0
        cls.defaults['rectR'] = 0

        cls.defaults['plotPickRectB'] = 0
        cls.defaults['plotPickRectG'] = 0
        cls.defaults['plotPickRectR'] = 255

        cls.defaults['zoomRectB'] = 0
        cls.defaults['zoomRectG'] = 0
        cls.defaults['zoomRectR'] = 0

        cls.defaults['lineWidth'] = 2

        cls.defaults['showSplash'] = 0

        cls.defaults['use_composite'] = True
        cls.defaults['delete_composite_script'] = True
        cls.defaults['composite_script_path'] = ''

        # colon-separated list of warning-types
        cls.defaults['ignored_warnings'] = ''

        cls.defaults['gridGranularity'] = 10

        cls.defaults['logging'] = 0
        cls.defaults['printStderr'] = 0

        cls.defaults['recentFiles'] = ""

    def writeToFile(cls, filename):
        """Save all settings to file."""
        outputFile = open(filename, "w")
        outputFile.write(cls.writeToString())
        outputFile.close()

    def readFromFile(cls, filename):
        """Read settings from file."""
        inputFile = open(filename)
        cls.readFromString(inputFile.read())
        
    def writeToString(cls):
        """Save all settings to string, as would when writing to
        file."""
        output = "# User-configuration for %s - version %s\n"%(Config.appname, Config.version)
        output += "# This file was automatically saved. Users can modify\n"
        output += "# this file, but any comments will be deleted.\n"
        for key,value in cls.modified.items():
            if value == None or value == 'None':
                value = ""
            output += key + " = " + str(value) + "\n"
        return output

    def readFromString(cls, inputString):
        lines = inputString.splitlines()
        # remove comments and empty lines
        lines = filter(lambda l : not l.lstrip().startswith('#'), lines)
        lines = filter(lambda l : l.strip(), lines)
        for l in lines:
            try:
                key, value = l.split('=')
                key = key.strip()
                value = value.strip()
                try:
                    # try to convert to integer, unless decimal
                    if not value.count('.'):
                        value = int(value)
                except ValueError:
                    pass

                # convert to boolean
                if value == 'True':
                    value = True
                elif value == 'False':
                    value = False

                setattr(cls, key, value)
                
            except ValueError:
                pass
            except AttributeError:
                pass
            except IOError, e:
                pass

    def resetToDefaults(cls):
        """Reset modifiable values, leaving only defaults."""
        cls.modified.clear()


    resetToDefaults = classmethod(resetToDefaults)
    setDefaults = classmethod(setDefaults)
    writeToString = classmethod(writeToString)
    readFromString = classmethod(readFromString)
    writeToFile = classmethod(writeToFile)
    readFromFile = classmethod(readFromFile)
    

def write_config():
    try:
        # test directory
        if not os.path.exists(Config.userconfigdir):
            os.mkdir(Config.userconfigdir)
                       
        configFileName = os.path.join(Config.userconfigdir, Config.rcfile)
        Config.writeToFile(configFileName)
    except IOError, e:
        pass
                    
def read_config(fileName=None):
    try:
        if fileName:
            Config.rcfile = os.path.abspath(fileName)

        configFileName = os.path.join(Config.userconfigdir, Config.rcfile)
        Config.readFromFile(configFileName)
        
        write_config()
    except Exception, e:
        pass

def get_style():    
    if not hasattr(Config, "style"):
        Config.style = __extractStyle(qApp.style())

    if ((Config.style[0] != 'Q') or
        (Config.style[-5:] != "Style") or
        (Config.style.find(" ") > 0)):
        Config.style = "QPlatinumStyle" #default style

    try:
        return QStyleFactory.create(Config.style[1:-5].lower())
        #Strip the leading "Q" and trailing "Style".
    except NameError:
        return QStyleFactory.create("platinum")


Config.setDefaults()
