# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import qt
import os
import time
import sys

class InterruptibleCommand:
    """Class to run shell commands that is interruptible

    This class uses the qt.QProcess class for running external shell
    commands. This class is the only one (that we found...) that is totally
    supported in both Linux and Windows.

    The class' intended use is solely for running external shell
    commands/processes which you *might* want to interrupt at some time. There
    is also functionality for reading from stdout and stderr, and writing to
    stdin. 

    Example:
    cmd = InterruptibleCommand("sleep", "2")
    cmd.runInterruptible()
    
    This is equivalent to running
    $ sleep 2
    in a shell.
    """
    

    def __init__(self, command, *args):
        """Constructs a command, but does not run it.

        Arguments:
        command -- the command to run (without the path)
        *args   -- the arguments to the command
        
        """
        self._proc = qt.QProcess()
        self._proc.addArgument(command)
        for arg in args:
            self._proc.addArgument(arg)
        self._pid = 0
        self._interrupted = True
        self._is_posix = os.name == 'posix'
        self._command = command
                   
        if not self._executable_in_path(command):
            if not os.path.exists(command):
                raise NoSuchExecutableError, "Command '%s' not found" % command
            

    def _executable_in_path(self, base_command):
        """Return true if base_command is somewhere in the PATH environment variable.

        Arguments:
        base_command -- the command to search the path for
        """
        pathlist = os.getenv("PATH").split(os.pathsep)
        executable_found = False
        for path in pathlist:
            if os.path.exists(os.path.join(path, base_command)):
                executable_found = True
            if not self._is_posix:
                if os.path.exists(os.path.join(path, base_command) + ".exe"):
                    executable_found = True
                    self._command += ".exe"
                elif os.path.exists(os.path.join(path, base_command) + ".com"):
                    executable_found = True
                    self._command += ".com"
                elif os.path.exists(os.path.join(path, base_command) + ".bat"):
                    executable_found = True
                    self._command += ".bat"
                elif 'library.zip' in sys.path:                    
                    executable_found = True
                    self._command += ".exe"
        return executable_found
        
        
    def run_interruptible(self):
        """Runs the command by starting the QProcess-instance."""
        self._interrupted = False
        self._proc.start()
        
        if self._is_posix:
            self._pid = self._proc.processIdentifier()


    def interrupt(self, wait=True):
        """Interrupts the running QProcess-instance and returns the exit-status.
        
        If the process is running, it tries to terminate and then sleeps for
        half a second. Then, if the process is still running, it kills it. It
        also tries to handle defunct processes.

        Parameters:
        wait=True -- wait for process to finish after kill (default is True)
        
        """
        self._interrupted = True
        if self._proc.isRunning():
            self._proc.tryTerminate()
            max_wait_time = 5 # 2.5 sec
            while max_wait_time > 0 and self._proc.isRunning():
                time.sleep(0.5)
                max_wait_time -= 1
            if self._proc.isRunning():
                self._proc.kill()
                self._proc.isRunning() # -> removes defunct processes... (coincidence?)
                while wait and not self._proc.exitStatus() and self._proc.isRunning():
                    time.sleep(0.1)
                
        return self._proc.exitStatus()


    def read_stdout(self):
        """Returning stdout as a string from the process.

        If the process is still running or is interrupted it returns 0.
        All line separators are converted to \n, regardless of what os is
        running.
        """
        # don't read until it is finished
        if self.is_running():
            return 0
        else:
            output = str(self._proc.readStdout())
            if os.linesep != '\n':
                out = output.replace(os.linesep, "\n")
                return out
            else:
                return output


    def read_stdout_blocking(self):
        """Returning stdout as a string from the process, waiting for stdout to
        finish before it returns.
        """
        std_out = self.read_stdout()
        while std_out == 0:
            time.sleep(0.1)
            std_out = self.read_stdout()
        return std_out
    

    def readline_stdout(self):
        """Returns the next unread line from the process' stdout as a string.

        Returns the next unread line from the process' stdout as a string, if it
        is possible to read a whole line from stdout. If it is not and the
        process is running, it will raise an OutputStreamNotReadyError
        exception. If the process has finished running it returns an empty
        string ('').
        """
        if self._proc.canReadLineStdout():
            return str(self._proc.readLineStdout())
        elif self._proc.isRunning():
            raise OutputStreamNotReadyError, "stdout not ready yet"
        else:
            return ''

    def read_stderr(self):
        """Returning stderr as a string from the process.

        If the process is still running or is interrupted it returns 0.
        All line separators are converted to \n, regardless of what os is
        running.
        """
        # don't read until it is finished
        if self.is_running():
            return 0
        else:
            output = str(self._proc.readStderr())
            if os.linesep != '\n':
                out = output.replace(os.linesep, "\n")
                return out
            else:
                return output


    def read_stderr_blocking(self):
        """Returning stderr as a string from the process, waiting for stderr to
        finish before it returns.
        """
        std_err = self.read_stderr()
        while std_err == 0:
            time.sleep(0.1)
            std_err = self.read_stderr()
        return std_err


    def readline_stderr(self):
        """Returns the next unread line from the process' stderr as a string.

        Returns the next unread line from the process' stderr as a string, if it
        is possible to read a whole line from stderr. If it is not and the
        process is running, it will raise an OutputStreamNotReadyError
        exception. If the process has finished running it returns an empty
        string ('').
        """
        if self._proc.canReadLineStderr():
            return str(self._proc.readLineStderr())
        elif self._proc.isRunning():
            raise OutputStreamNotReadyError, "stderr not ready yet"
        else:
            return ''
    

    def close_stdin(self):
        """Closes the stdin file for the running process.

        Closes the stdin file (equivalent to ^D in an interactive shell).
        """
        self._proc.closeStdin()


    def write_stdin(self, string):
        """Writes the given string to stdin and flushes.

        After string is written to the stdin file, the file is flushed

        Arguments:
        string -- the string to be written to stdin
        """
        self._proc.writeToStdin(string)
        self._proc.flushStdin()

    def check_stdout(self):
        self._proc.canReadLineStdout()


### get/set/is methods...

    def _get_pid(self):
        """Returns pid if process is started.

        If the running system is not linux (POSIX) then it raises a
        NotImplementedError exception.
        """
        if not self._is_posix:
            raise NotImplementedError, "method not implemented under windows"
        return self._pid


    def is_running(self):
        """Returns true if the process is running, and not interrupted."""
        return self._proc.isRunning() and not self._interrupted
        

    def is_interrupted(self):
        """Returns true if the process is interrupted."""
        return self._interrupted

class OutputStreamNotReadyError(Exception):
    pass


class NoSuchExecutableError(Exception):
    pass
