# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import gzip 
import xdrlib

class RVariableNameIdentifier:
    """
        This class was written because RPy (<= 0.3.5) was unable to execute a
        function we needed to give the user a list of variables present in a
        R data file. Hopefulle, RPy will be fixed, this class can then be
        deleted.
        
        [NOTE] If we want, we are able to return types of the variables.
    """

    def __init__(self):
        self.DEBUG   = 0    # Flag for printing debugging information.
        self.USERERR = 1    # Flag for printing error msgs the user should see.
        
        self.bufferSize = 8192 # Byte

    def findRVars(self, fileName):
        """
            Parses a R data file and returns the names of the variables stored
            in the file.
            
            Params:
                fileName - string : The name of the file to parse.
                
            Returns:
                A list with the variable names.
                If no variables were found, a empty list is returned.
        """
        # Determine type of file.
        fType  = "UNDETERMINED"
        isZipped = 0
        try:
            fp = file(fileName, "rb")
            fType = fp.read(4)            
            # Are there more partial headers for compressed file used by R?
            if fType == '\x1F\x8B\x08\x00':
                isZipped = 1
                fp.seek(0)
                zfp = gzip.GzipFile(fileobj = fp)
                fType = zfp.read(4)
                zfp.close()
        finally:
            fp.close()
        if fType == "RDA2":
            res = self.__findRVarsASCII(fileName, zipped = isZipped)
            return res
        # Not sure if RDB2 is used, or if it works with xdr routines...
        elif fType == "RDX2" or fType == "RDB2":
            return self.__findRVarsBinary(fileName, zipped = isZipped)
    
    def __findRVarsBinary(self, fileName, zipped=0):
        """
            Parses a binary R data file and returns the names of the variables
            in the file.
            
            Params:
                fileName - string : The name of the file to parse.
                zipped   - int    : Indicates wether the data file is
                                    zipped or not.
                
            Returns:
                A list with the variable names.
                If no variables were found, a empty list is returned.            
        """
        target    = '\x00\x00\x04\x02\x00\x00\x00\x01\x00\x00\x14\t'
        targetLen = len(target)
        names = []
        # Sanitycheck.
        if ( self.bufferSize <= targetLen ):
            message = "[__findRVarsBinary] Buffer too small!"
            message += " Should be at least the size of the target expression."
            # A general exception? What's the point of that?
            raise Exception, message
        if zipped:
            fpRaw = file(fileName, "rb", self.bufferSize)
            fp = gzip.GzipFile(fileobj = fpRaw)
        else:
            fp = file(fileName, "rb", self.bufferSize)
        
        # Skip file type definition, magic number, version and some more.
        fp.seek(19)
                       
        # While more data, read in.
        binData = "#BOGUS#SCICRAFT#SPECIFIC#VALUE#"
        up = None
        isComplex     = 0
        complexMode   = 0
        lastByte      = ''
        while binData != '':
            # Size of binData won't be static because of prefixing 
            # and unscheduled datafetching. Is this very bad?
            prefix = ""
            if binData != "#BOGUS#SCICRAFT#SPECIFIC#VALUE#":
                # Make sure we don't miss a match on the buffer border.
                prefix = binData[-targetLen:]
            # Read data from file.
            binData = fp.read(self.bufferSize)
            if binData != '':
                binData = prefix + binData
            index = 0
            # Search for varaible definitions.
            delimIndex = binData.find(target, index)
            while delimIndex != -1:
                isComplex = 0
                lastByte = binData[delimIndex -4:delimIndex]
                # Is this the end of a complex variable structure?
                if lastByte == '\x00\x00\x00\xfe':
                    complexMode -= 1
                index = delimIndex + targetLen
                if len(binData) - index < 4:
                    # This should not happen often.
                    # Just read in the data needed.
                    binData = binData[index:] + fp.read(len(binData) - index)
                    index = 0
                # Create unpacker to unpack binary data.
                up = xdrlib.Unpacker(binData[index:])
                varNameLength = up.unpack_uint()
                if len(binData) - index < varNameLength:
                    binData = binData[index:] + fp.read(len(binData) - index)
                    index = 0
                    up = xdrlib.Unpacker(binData[index:])
                # Must reposition after call to unpack_fstring().
                # Don't know why.
                flagPos = up.get_position() + varNameLength
                name = up.unpack_fstring(varNameLength)
                up.set_position(flagPos)
                # Fetch flags; definition of type and statusflags.
                flags = up.unpack_uint()
                if self.__hasAttrs(flags):
                    isComplex = 1
                if isComplex:
                    if complexMode == 0:
                        # Add name.
                        names.append(name)
                    complexMode += 1

                else:
                    if complexMode == 0:
                        # We are not inside a complex object. Add the name.
                        names.append(name)
                delimIndex = binData.find(target,  index)
        
        fp.close()
        if zipped:
            fpRaw.close()
        return names
            

    def __findRVarsASCII(self, fileName):
        """
            Parses a R ASCII data file and returns the names of the variables
            in the file.
            
            Params:
                fileName - string : The name of the file to parse.
                
            Returns:
                A list with the variable names.
                If no variables were found, a empty list is returned.            
        """
        # Deprecated, but okay to maintain anyway. See Rinternals.h for defs.
        #allTypes = [0,1,2,3,4,5,6,7,8,9,10,13,14,
        #   15,16,17,18,19,20,21,22,23,99]          # R.1,8,1
        #simple = [0,9,13,14,15,16,19]
        #complexO = [3,6]
        complexMode = 0
        isComplex   = 0
        if zipped:
            fpRaw = file(fileName, "r", self.bufferSize)
            fp = gzip.GzipFile(fileobj = fpRaw)
        else:
            fp = file(fileName, "r", self.bufferSize)
        names = []
        line = "BOGUS START! IGNORE THIS!"
        prevLine = ""

        def a_line():
            return fp.readline().strip()
       
        while line != '':
            prevLine = line
            line = a_line()
            if line == "1026":
                isComplex = 0
                lines = []
                lines.append(a_line()) # 1 (unknown meaning)
                lines.append(a_line()) # 5129 (unknown meaning)
                lines.append(a_line()) # Length
                lines.append(a_line()) # Name
                try:
                    lines.append(int(a_line())) # Typedef
                except ValueError:
                    lines.append(-1) # Invalid typedef.                
                if lines[0] == "1" and lines[1] == "5129" and lines[4] > 0:
                    if prevLine == "254":
                        complexMode -= 1                    
                    # Valid definition.
                    hasAttrs = self.__hasAttrs(lines[4])

                    # Might be more checks here later, for instance special 
                    # handling of some types.
                    # The determination of isComplex MUST work!
                    if hasAttrs:
                        isComplex = 1

                    if not complexMode:
                        # Add name
                        names.append(lines[3])

                    if isComplex:
                        complexMode += 1

        fp.close()
        if zipped:
            fpRaw.close()
        return names


    def __getType(self, flags):
        """
            Helper method to determine variable type from a integer value
            stored in the data file.
            
            Params:
                flags - int : The integer value to extract the type from.
            
            Returns:
                An integer representing a R type. See 'Rinternals.h' (R source 
                code) for details.
        """
        return (flags & 255)


    def __hasAttrs(self, flags):
        """
            Helper method to determine if a variable has attributes, based on a
            integer value stored in the data file.
            
            Params:
                flags - int : The integer value to extract the information from..
            
            Returns:
                '1' if variable has attributes.
                '0' if variable does not have attributes.
        """
        if (flags & (1 << 9)) > 0:
            return 1
        return 0
       
       
    def __hasTag(self, flags):
        """
            Helper method to determine if a variable has tag, based on a
            integer value stored in the data file.
            
            Params:
                flags - int : The integer value to extract the information from..
            
            Returns:
                '1' if variable has tag.
                '0' if variable does not have tag.
        """
        if (flags & (1 << 10)) > 0:
            return 1
        return 0
        
        
    def __isObject(self, flags):
        """
            Helper method to determine if a variable is classified as an
            object, based on a integer value stored in the data file.
            
            Params:
                flags - int : The integer value to extract the information from..
            
            Returns:
                '1' if variable is classified as an object.
                '0' if variable is not classified as an object.
        """
        if (flags & (1 << 8)) > 0:
            return 1
        return 0
