# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.



from qt import qRgb
from Numeric import arange


_colormap = {}

def get_color_maps():
    """ return colormap as dictionary. """
    return _colormap

def look_up_available_colormaps():
    """ Look up and return available colormaps,that exists in _colormap.

    If _colormap is empty, this will be initialized here.
    
    """
    return _colormap.keys()

def get_colormap_by_key(cmap_key):
    """ get colormap by key.
    
    Return value:
    On success a valid colormap is returned,
    on failure a False is returned.

    """

    if _colormap.has_key(cmap_key):
        return _colormap.get(cmap_key)
    else:
        return False
    
def _init_colormap():
    """ Initializing colormap. Should be done ONCE"""
    _colormap['gray'] = _create_gray_cmap()
    _colormap['jet'] = _create_jet_cmap()
    _colormap['autumn'] = _create_autumn_cmap()
    _colormap['bone'] = _create_bone_cmap()
    _colormap['hot'] = _create_hot_cmap()
    _colormap['cool']=_create_cool_cmap()
    _colormap['rainbow'] = _create_rainbow_cmap()
    _colormap['red'] = _create_red_cmap()
    _colormap['green'] = _create_green_cmap()
    _colormap['blue'] = _create_blue_cmap()
    _colormap['summer']= _create_summer_cmap()
    _colormap['winter'] = _create_winter_cmap()
    _colormap['spring'] = _create_spring_cmap()
    
def _create_gray_cmap():
    """ Create a gray scaled colormap."""
    rgb_list = []
    n = 256
    for i in range(n):
        rgb_list.append(qRgb(i,i,i))
    return rgb_list

def _create_jet_cmap():
    """
    Create a jet scaled color map.

    Colors are ranging from dark blue and to dark red.
    Algorithm is taken from jet.m http://octave.sf.net

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(0, 0, 0.5*255 ) )
    else:
        number = n
        x = arange(0, 1+1.0/number, 1.0/(number-1))
        r = (( (x >= 3.0/8.0) & (x < 5.0/8.0) ) * (4.0 * x - 3.0/2.0) + \
             ( (x >= 5.0/8.0) & ( x < 7.0/8.0 ))) + \
             (x >= 7./8) * (-4.0 * x + 9./2.)
        
        g =  ((x >= 1./8) & (x < 3./8)) * (4.0 * x - 1./2)\
            + ( (x >= 3./8) & (x < 5./8)) + \
            ( (x >= 5./8) & (x < 7./8)) * (-4. * x + 7./2.)
        b = (x < 1./8) * (4. * x + 1./2) +\
            ( (x >= 1./8) &( x < 3./8))\
            + ((x >= 3./8) & (x < 5./8)) * (-4 * x + 5./2)
        for i in range(n):
            rgb_list.append( qRgb( r[i]*255,g[i]*255,b[i]*255 ) )
    return rgb_list

def _create_autumn_cmap():
    """
    Create a autumn color map.

    Colors are ranging from red to orange.
    Algorithm is taken from autumn.m http://octave.sf.net

    """

    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(0, 0, 0.5*255 ) )
    else:
        number = n
        for i in range(number):
            rgb_list.append(qRgb(255, i, 0))
    return rgb_list

def _create_bone_cmap():
    """
    Create a differentgreyscale color map.

    Colors ranging from black to white, but with  a higher value for the blue component.
    Gives an x-ray image appearance. 
    
    Algorithm is taken from bone.m http://octave.sf.net

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(0, 0, 0 ) )
    else:
        number = n
        x = arange(0, 1+1.0/number, 1.0/(number-1))
        r = ((x < 3./4.)) * (7./8. * x) + ((x >= 3./4.)) * (11./8. * x - 3./8.)
        g = ((x < 3./8.)) * (7./8. * x)\
            + ((x >= 3./8.) & (x < 3./4.)) * (29./24. * x - 1./8.)\
            + ((x >= 3./4.)) * (7./8. * x + 1./8.)
        b = ((x < 3./8.)) * (29./24. * x) + ((x >= 3./8.)) * (7./8. * x + 1./8.)
        
        for i in range(n):
            rgb_list.append( qRgb( r[i]*255,g[i]*255,b[i]*255 ) )
    return rgb_list


def _create_hot_cmap():
    """
    Create a hot color map.

    Colors varies smoothly from black through shades of red, orange, and yellow, to white..
    Algorithm is taken from hot.m http://octave.sf.net

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(0, 0, 0 ) )
    else:
        number = n
        x = arange(0, 1+1.0/number, 1.0/(number-1))
        r = ((x < 2./5.)) * (5./2. * x) + ((x >= 2./5.))
        g = ((x >= 2./5.) & (x < 4./5)) * (5./2. * x - 1) + ((x >= 4./5.))
        b = ((x >= 4./5.)) * (5*x - 4)
        for i in range(n):
            rgb_list.append( qRgb( r[i]*255,g[i]*255,b[i]*255 ) )
    return rgb_list

def _create_cool_cmap():
    """
    Create a cool color map.

    Colors are shades of cyan and magenta. It varies smoothly from cyan to magenta.
    Algorithm is taken from cool.m http://octave.sf.net

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(0, 255, 255 ) )
    else:
        number = n
        for i in range(n):
            rgb_list.append( qRgb( i,255-i,255 ) )
    return rgb_list

def _create_rainbow_cmap():
    """
    Create a rainbow color map.

    Algorithm is taken from rainbow.m http://octave.sf.net

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(255, 0, 0 ) )
    else:
        number = n
        x = arange(0, 1+1.0/number, 1.0/(number-1))
        r = (x < 2./5) + ( (x >= 2./5) & (x < 3./5)) * (-5. * x + 3)\
            + ( (x >= 4./5)) * (10./3 * x - 8./3)
        g = (x < 2./5) * (5./2 * x) + ( (x >= 2./5) & (x < 3./5))\
            + ( (x >= 3./5) & (x < 4./5)) * (-5. * x + 4)
        b = ( (x >= 3./5) & (x < 4./5)) * (5 * x - 3) + ((x >= 4./5))
        for i in range(n):
            rgb_list.append( qRgb( r[i]*255,g[i]*255,b[i]*255 ) )
    return rgb_list

def _create_red_cmap():
    """
    Create a red color map.

    Colors varies from black to red. 

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(0, 0, 0 ) )
    else:
        number = n
        for i in range(n):
            rgb_list.append( qRgb( i,0,0 ) )
    return rgb_list

def _create_green_cmap():
    """
    Create a green color map.

    Colors varies from black to green. 

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(0, 0, 0 ) )
    else:
        number = n
        for i in range(n):
            rgb_list.append( qRgb( 0,i,0 ) )
    return rgb_list

def _create_blue_cmap():
    """
    Create a blue color map.

    Colors varies from black to blue. 

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(0, 0, 0 ) )
    else:
        number = n
        for i in range(n):
            rgb_list.append( qRgb( 0,0,i ) )
    return rgb_list

def _create_summer_cmap():
    """
    Create a summer color map.

    Colors are shades of green and yellow. 
    Algorithm is taken from summer.m in octave.sf.net

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(0, 255*0.5, 255*0.4 ) )
    else:
        number = n
        for i in range(n):
            rgb_list.append( qRgb(i, 0.5*255+i/2., 0.4*255 ) )
    return rgb_list

def _create_winter_cmap():
    """
    Create a winter color map.

    Colors are shades of blue and green. 
    Algorithm is taken from winter.m in octave.sf.net

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(0, 0, 255 ) )
    else:
        number = n
        for i in range(n):
            rgb_list.append( qRgb(0, i, 255 - i/2.0 ) )
    return rgb_list

def _create_spring_cmap():
    """
    Create a spring color map.

    Colors are shades of magenta and yellow.
    Algorithm is taken from spring.m in octave.sf.net

    """
    n = 256
    rgb_list = []
    if n == 1:
        rgb_list.append(qRgb(255, 0,255 ) )
    else:
        number = n
        for i in range(n):
            rgb_list.append( qRgb(255, i, 255-i ) )
    return rgb_list


_init_colormap()
