# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.



from System.Data.functionnode import FunctionNode
from Library.Util import lists

def find_node_threads(nodes):
    """Find seperated branches/threads of nodes """
    # first find start nodes without any input ports
    node_thread_list = []
    for node in nodes:
        if node.isStartNode():
            # find the thread of nodes that starts in this node
            node_thread_list.append(search_next_nodes(node, []))


    if len(node_thread_list) > 1:
        # search the threads to see if some of them are joined
        for i in range(len(node_thread_list)):
            match = False
            try:
                match = check_matching_elements(node_thread_list[i],
                                                node_thread_list[i+1])
            except IndexError:
                break
            if match:
                lists.merge_lists(node_thread_list[i], (node_thread_list[i+1]))
                node_thread_list.remove(node_thread_list[i+1])    
                
    return node_thread_list
            
def search_next_nodes(node, node_list):
    """Finds nodes conneted to 'node' and adds these 
    to 'node_list'.
    """
    
    node_list.append(node)
    successors = node.get_successing_nodes()

    if successors:
        for successor in successors:
            search_next_nodes(successor, node_list)

    lists.remove_duplicates(node_list)
    return node_list



def check_matching_elements(search_list, target_list):
    """Checks if any of the elements in search_list can be found in
    target_list
    """
    for node in search_list:
        if node in target_list:
            return True

    return False


def find_start_function_nodes(nodes, functions):
    """Method that finds star function nodes, that is nodes that are not connected
    to any other function nodes.
    """

    start_nodes = []
    for node in functions[:]:
        # lets remove all function nodes that we know cant be part of a cluster
        parents = node.get_providing_nodes()
        childs = node.get_successing_nodes()
        alone = True
        for related in parents + childs:

            if related in functions:
                alone = False
                break
        if alone:
            functions.remove(node)
            continue 

        # find start (function nodes without function nodes as parents)
        start_node = True
        for parent in parents:
            if parent in functions:
                start_node = False
                break
        if start_node:
            start_nodes.append(node)

    return start_nodes

def find_combinable_nodes(start_nodes, functions):
    """Mehtod that scans through clusters to see if they are
    really part of the same cluster.
    """
    combinable = []
    # now we find the clusters that starts with the start_nodes
    for start in start_nodes:
        combinable.append(look_for_clusters(start, [], functions))

    # check if some of the clusters are really the same cluster
    for cluster in combinable:
        lists.remove_duplicates(cluster) #remove duplicates

    for x in range(len(combinable)):
        for y in range(x+1, len(combinable)):
            match = False
            try:
                match = check_matching_elements(combinable[x], combinable[y])
            except IndexError:                
                break
            if match:
                # we have a cluster that shares one or more objects
                lists.merge_lists(combinable[x], combinable[y])
                combinable.remove(combinable[y])
                                                  
    # now we have a dictionary with clusters, now we need to check if
    # these clusters are safe
    return combinable


def test_cluster(cluster):
    """Method that checks if a clusters is safe. A cluster is
    unsafe if two connected cuntion nodes are also connected through
    a node that is not a function node or have a different plugin.
    """
    unsafe_nodes = []
    for node in cluster[:]:
        parents = node.get_providing_nodes()
        function_parents = []
        other_parents = []
        for parent in parents:
            if parent in cluster:
                function_parents.append(parent)
            else:
                other_parents.append(parent)
                
        for other in other_parents:
            for function in function_parents:
                if look_for_node(other, function):
                    cluster.remove(node)
                    unsafe_nodes.append(node)
                    break
                

    # now we have a list with safe and unsafe nodes. We need to find all function nodes
    # 'behind' the unsafe nodes and remove these from the safe list

    for node in cluster:
        for unsafe in unsafe_nodes[:]:
            if look_for_node(node, unsafe):
                cluster.remove(node)
                unsafe_nodes.append(node)            
                
    return (cluster, unsafe_nodes)
    
def look_for_node(source, target):
    """Method that search through all of source's preceding nodes
    to check see if it can find nodes in the targets list
    """
    
    parents = source.get_providing_nodes()

    if target in parents:
        return True
    elif not parents:
        return False
    else:
        check = False
        for node in parents:
            check = look_for_node(node, target)
            if check:
                break
        return check

def find_octave_function_nodes(thread):
    """Mehtod that finds all octave function nodes."""

    function_nodes = []
    for node in thread:
        if isinstance(node, FunctionNode) and node.get_plugin_type() == 'OctavePlugin':
            function_nodes.append(node)
        
    return function_nodes
        

def look_for_clusters(main_node, cluster_list, functions):
    """Method that finds clusters, i.e. function nodes that
    are direcly connected. Note that these clusters might not be safe.
    """
    
    if main_node in cluster_list:
        # we have already been through this one
        return cluster_list
    
    if main_node in functions:
        cluster_list.append(main_node)
        
    childs = main_node.get_successing_nodes()
    for child in childs:
        look_for_clusters(child, cluster_list, functions)
        
    return cluster_list



def find_octave_groups(nodes):
    """Main method on this module that is used to find safe
    clusters of octave function nodes.

    These clusers are safe to execute within one instance of a plugin.
    """
    
    diagram_clusters = []
    threads = find_node_threads(nodes)
    
    for thread in threads:
        thread_clusters = []
        unsafe_nodes = []    
        functions = find_octave_function_nodes(thread) 

        start_nodes = find_start_function_nodes(thread, functions) # find start nodes
        unsafe_nodes.append(start_nodes)

        while unsafe_nodes:
            combinable_nodes = find_combinable_nodes(unsafe_nodes.pop(), functions)
            for nodes in combinable_nodes:
                (safe, unsafe) = test_cluster(nodes)
                if safe:
                    thread_clusters.append(safe)
                if unsafe and unsafe != nodes:
                    unsafe_nodes.append(unsafe)
                        
        diagram_clusters.extend(thread_clusters)

    return diagram_clusters



def find_input_nodes(cluster):
    """Finds all nodes that delivers data to the cluster"""
    input_nodes = []
    for node in cluster:
        inputs = node.get_providing_nodes()
        for input_ in inputs:
            if input_ not in cluster:
                input_nodes.append(input_)

    return input_nodes
        
def find_output_nodes(cluster):
    """Finds all nodes that recives data from the cluster"""
    output_nodes = []
    for node in cluster:
        outputs = node.get_successing_nodes()
        for output in outputs:
            if output not in cluster:
                output_nodes.append(output)

    return output_nodes

def find_safe_group_order(cluster):

    order = []
    nodes = cluster[:]
    while nodes:
        node = nodes[0]
        nodes.remove(node)
        inputs = node.getInputPorts()
        safe = True
        for port in inputs:
            if port.getDataSource().getOwnerNode() in nodes:
                safe = False
        if safe:
            order.append(node)
        else:
            nodes.append(node)

    return order
    
