# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import re
import random
import cPickle
import os

class _Named:
    """A named item, subclassable for naming functionality."""
    def __init__(self, name=""):
        self._name = name
        
    def get_name(self):
        """Get name associated with this microarray."""
        return self._name

    def set_name(self, name):
        self._name = name

class Scan(_Named):
    """Represents a datafile resulting from a laserscanned microarray."""
    # possible last and pmt-settings (laser frequency)
    NOT_SET, LOW, MEDIUM, HIGH = range(4)

    def __init__(self, name):
        _Named.__init__(self, name)
        self._pmt = Scan.NOT_SET
        self._laser = Scan.NOT_SET
        self._filename = ""

    def get_pmt(self):
        """Get PMT setting used for this scan.

        Setting can be None, Scan.HIGH, Scan.MEDIUM or Scan.LOW.
        """
        return self._pmt

    def set_pmt(self, value):
        self._pmt = value

    def get_laser(self):
        """Get laser setting used for this scan.

        Setting can be None, Scan.HIGH, Scan.MEDIUM or Scan.LOW.
        """
        return self._laser

    def set_laser(self, value):
        self._laser = value

    def set_filename(self, filename):
        self._filename = filename

    def get_filename(self):
        return self._filename
    

class MicroArray(_Named):
    """A physical micro array, with smears of biomaterial."""
    def __init__(self):
        _Named.__init__(self)
        self._dye_swapped = False
        self._scans = []
        self._array_name = ""
    
    def is_dye_swapped(self):
        """Is dye reversed in samples compared to main comparison direction?"""
        return self._dye_swapped

    def set_dye_swapped(self, value):
        """Indicate that dye is swapped compared to main colors."""
        self._dye_swapped = value

    def add_scan(self):
        value = 1
        for scan in self._scans:
            m = re.match(r"Raw Data File (\d+)", scan.get_name())
            if m:
                value = int(m.group(1)) + 1
        scan = Scan("Raw Data File %d" % value)
        self._scans.append(scan)
        return scan

    def get_scans(self):
        return self._scans

    def set_array_name(self, name):
        self._array_name = name

    def get_array_name(self):
        return self._array_name

    def remove_scan(self, scan):
        """Remove an existing scan."""
        self._scans.remove(scan)

    def clear_array_name(self):
        self._array_name = ""

        
class BioReplica(_Named):
    """Biological material from two individuals to be compared."""
    def __init__(self):
        _Named.__init__(self)
        self._arrays = []
        
    def add_micro_array(self):
        """Create and return a new micro array for this bio material."""
        ma = MicroArray()
        value = 1
        for array in self._arrays:
            m = re.match(r"Hybridisation (\d+)", array.get_name())
            if m:
                value = int(m.group(1)) + 1
        ma.set_name("Hybridisation %d" % value)
        self._arrays.append(ma)
        return ma

    def get_micro_arrays(self):
        """Get all micro arrays defined for this bio material."""
        return self._arrays

    def remove_micro_array(self, ma):
        self._arrays.remove(ma)


class BioMaterial(_Named):
    """A bio material is a potentially treated population."""
    def __init__(self, x, y):
        """Create a new bio material at position x, y."""
        _Named.__init__(self)
        self._position = x, y

    def get_position(self):
        return self._position

    def set_position(self, x, y):
        self._position = x, y

class Comparison:
    """Holds all bio materials/micro arrays/scans between two bio_materials."""
    def __init__(self, from_bio_material, to_bio_material):
        self._bio_replicas = []
        self._from = from_bio_material
        self._to = to_bio_material
        # dyes used in from/to biological matter used in micro arrays
        self._colors = "", ""
    
    def add_bio_replica(self):
        bio = BioReplica()
        value = 1
        for material in self._bio_replicas:
            m = re.match(r"BioReplica (\d+)", material.get_name())
            if m:
                value = int(m.group(1)) + 1
        bio.set_name("BioReplica %d" % value)
        self._bio_replicas.append(bio)
        return bio

    def get_bio_replicas(self):
        return self._bio_replicas

    def remove_bio_replica(self, bio_replica):
        self._bio_replicas.remove(bio_replica)

    def get_bio_materials(self):
        """Get from/to tuple of bio materials for this comparison."""
        return self._from, self._to

    def get_colors(self):
        return self._colors

    def set_colors(self, from_color, to_color):
        self._colors = from_color, to_color


class Experiment(_Named):
    """An experiment with bio materials and comparisons."""
    def __init__(self):
        _Named.__init__(self)
        self.set_name("New Experiment")
        self._bio_materials = []
        self._comparisons = []
        self._array_names = []

    def add_bio_material(self, x, y):
        """Create a new bio material at position x, y."""
        t = BioMaterial(x, y)
        value = 1
        for bio_material in self._bio_materials:
            m = re.match(r"BioMaterial (\d+)", bio_material.get_name())
            if m:
                value = int(m.group(1)) + 1
        t.set_name("BioMaterial %d" % value)
        self._bio_materials.append(t)
        return t

    def get_bio_materials(self):
        return self._bio_materials

    def remove_bio_material(self, bio_material):
        self._bio_materials.remove(bio_material)

    def add_comparison(self, from_bio_material, to_bio_material):
        comparison = Comparison(from_bio_material, to_bio_material)
        self._comparisons.append(comparison)
        return comparison

    def get_comparisons(self):
        return self._comparisons

    def remove_comparison(self, from_bio_material, to_bio_material):
        for comparison in self._comparisons[:]:
            if (from_bio_material, to_bio_material) == comparison.get_bio_materials():
                self._comparisons.remove(comparison)
                return

    def set_array_names(self, array):
        """Set a list of possible MicroArray names.

        These names can later be randomly assigned to all micro arrays in the
        experiment.
        """
        self._array_names = array

    def get_array_names(self):
        return self._array_names
                
    def _get_all_micro_arrays(self):
        arrays = []
        for comparison in self._comparisons:
            for bio in comparison.get_bio_replicas():
                arrays.extend(bio.get_micro_arrays())
        return arrays

    def assign_names_to_micro_arrays(self):
        self._clear_array_names()
        i = 0
        all_arrays = self._get_all_micro_arrays()
        shuffled_names = self._array_names[:]
        random.shuffle(shuffled_names)
        for i, (array, name) in enumerate(zip(all_arrays, shuffled_names)):
            array.set_array_name(name)
            i += 1
            
        
        return i

    def _clear_array_names(self):
        for array in self._get_all_micro_arrays():
            array.clear_array_name()

    def assign_filenames(self):
        replace = lambda string: string.replace(" ", "_")
        for comparison in self._comparisons:
            from_bio, to_bio = [material.get_name() for material in comparison.get_bio_materials()]
            from_dye, to_dye = comparison.get_colors()
            for bio in comparison.get_bio_replicas():
                for ma in bio.get_micro_arrays():
                    array_name = ma.get_array_name()
                    if not array_name:
                        array_name = ma.get_name()

                    from_, to = from_bio, to_bio
                    if ma.is_dye_swapped():
                        to, from_ = from_, to
                    
                    from_ += from_dye
                    to += to_dye
                        
                    for scan in ma.get_scans():
                        scan.set_filename('%s-%s-%s-%s-%s' %(replace(from_),
                                                             replace(to),
                                                             replace(bio.get_name()),
                                                             replace(array_name),
                                                             replace(scan.get_name())))

    def is_empty(self):
        """See if this experiment contains any information at all."""
        return not self._array_names and not self._bio_materials and not self._comparisons
                        

class ExperimentStorage:
    """Storage for all experiments of current user."""

    def __init__(self, storage_path):
        self._path = storage_path
        self._experiment_file = os.path.join(self._path, "experiments")
        self._experiments = {}

        if os.path.exists(self._experiment_file):
            self._experiments = cPickle.load(open(self._experiment_file))

    def get_experiment_names(self):
        """Get all experiments stored by user."""
        return self._experiments.keys()

    def store_experiment(self, experiment):
        """Store experiment somewhat hidden in users home directory."""
        self._experiments[experiment.get_name()] = experiment
        if not os.path.exists(self._path):
            os.mkdir(self._path)
        cPickle.dump(self._experiments, open(self._experiment_file, "w"))

    def load_by_name(self, name):
        return self._experiments[name]
    
