#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

"""Utility functionality concerned with files."""

from Library.Util.Config import Config

import cPickle
import os
import stat
import sys

class PathList:
    
    def __init__(self):
    
        self._path_list = Config.loadpath
        if self._path_list:
            self._path_list = self._path_list.split(";")
        else:
            self._path_list = []
    
    def add_path(self, path):
        self._path_list.append(path)

    def get_paths(self):
        return self._path_list[:]

    def get_invalid(self):
        invalid = []
        for f in self._path_list:
            if not os.path.isdir(f):
                invalid.append(f)
        return invalid

    def remove_paths(self, paths):
        for f in paths:
            index = self._path_list.index(f)
            self._path_list.pop(index)
        
    def write_paths(self):
        if not self.get_invalid():
            Config.loadpath = ";".join(self._path_list)
        else:
            raise InvalidPathException(1337)
        
class InvalidPathException(Exception):

    def __init__(self, value=0):
        self.value = value

    def __str__(self):
        return "Tried to write invalid path's to config", self.value
        
    

def loader(file_path):
    """Function that will search through the file_path directory
    adds all *.py files and directories to the all variable
    """

    # Need to remove this when scicraft has be built with py2exe
    
    file_path = file_path.replace('library.zip','')
    all = []

    valid_ext = "py" 
    
    # This is needed because scicraft.exe running under Windows doesn't always use
    # the scicrafthome as current working directory. In these cases loading of the
    # Library classes will fail
    if 'library.zip' in sys.path: # Running exe-file under Windows
        dirname = os.path.dirname(file_path)
        if dirname.startswith(os.sep):
            # Need to strip the leading \ incase Config.scicrafthome is not set yet
            dirname = dirname[1:]
        path = os.path.join(os.path.abspath(Config.scicrafthome), dirname)
    else:
        path = os.path.dirname(file_path)

    file_list = os.listdir(path) 
    for file in file_list: 
        file_path = os.path.join(path,file)         
        if (file.split(".")[-1] == valid_ext or os.path.isdir(file_path)) and file != "__init__.py" and file != ".svn": 
            all.append(file.split(".")[0])  

    return all


class CacheDirectory:
    """A filesystem based directory of cached ports."""
    def __init__(self, path):
        self._root = path
        self._path = os.path.join(path, "cache")
        self._enabled = True

    def get_used_space(self):
        """Get space used by cache-files in bytes."""
        if not os.path.exists(self._path):
            return 0
        
        used_space = 0
        for entry in os.listdir(self._path):
            path = os.path.join(self._path, entry)
            if os.path.isfile(path):
                used_space += os.stat(path)[stat.ST_SIZE]
        return used_space

    def clear_files(self):
        """Remove all cached files in directory."""
        if not os.path.exists(self._path):
            return
        
        for f in os.listdir(self._path):
            path = os.path.join(self._path, f)
            os.unlink(path)

        try:
            os.rmdir(self._path)
        except:
            pass

    def store_entry(self, key, data):
        """Store data entry to disk.

        Key is unique id for this data, and also lookup key.
        Data can be any pickleable data.
        """
        if not self._enabled: return
        
        if not os.path.exists(self._path):
            os.mkdir(self._path)

        path = os.path.join(self._path, key)
        cPickle.dump(data, open(path, "w"))

    def get_entry(self, key):
        """Retrieve previously stored entry from directory."""
        if not self._enabled: return {}
        
        path = os.path.join(self._path, key)
        try:
            return cPickle.load(open(path))
        except:
            return {}

    def set_path(self, new_path):
        """Set root path to be used when storing cache entries.

        Entries will not be placed directly in the path, but gathered
        in a created subdirectory named"cache".
        """
        self._root = new_path
        self._path = os.path.join(new_path, "cache")

    def get_path(self):
        return self._root

    def set_enabled(self, value):
        """Control if cache is active or not.

        A disabled cache will not store or gather data on requests.
        """
        self._enabled = value

    def get_enabled(self):
        return self._enabled

class StreamHolder:
    """Simple class to capture and hold a stream.

    May, e.g., be used to capture stdout (or stderr) on an occasion where some
    function writes to stdout (or stderr) during its execution, and you don't
    want the output to be visible. It is also possible to retrieve the data the
    function wrote to stdout (or stderr) by calling the read-function.

    EXAMPLE:
    holder = StreamHolder()
    sys.stdout = holder
    print "asdfasdf" # asdfasdf will not be printed in output
    # reset stdout
    sys.stdout = sys.__stdout__
    print "asdfasdfasdf" # asdfasdfasdf will be printed in output
    print holder.read() # will print asdfasdf in output
    """
    def __init__(self):
        self.data = ""

    def write(self, text):
        """Writes text to the stream."""
        self.data = self.data + text

    def read(self):
        """Retrieves all data that was written to the stream."""
        return self.data

    def flush(self):
        # Needs to be implemented since it will be called at times. 
        pass
        
