#-*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from qtui import QWidgetFactory
from qt import SIGNAL as get_signal
from qt import QTimer, QObject, SIGNAL, QListView, QListViewItem
from os.path import join as joinpath
from Library.Util.Config import Config
from Library.Util.lists import get_rows, get_cols
from Queue import Queue, Empty

import thread

DEBUG = False

def size(widget):
    """Takes the size of a widget, and returns it as a complex number,
        which is very handy for size and position calculations.
        It may look like misuse of the complex type, but it's quite
        widespread to use it in a similar manner.
    """
    return complex(widget.width(), widget.height())

def unsize(cnum):
    """Returns a complex number as a tuple."""
    return (cnum.real, cnum.imag)

def unqstring(qstring):
    """Converts a QString to a Latin-1 (iso-8859-1)"""
    return str(unicode(qstring).encode("l1"))

def get_index(combobox, text):
    """Returns the index of item with the specified text.
        Returns -1 if combobox is empty and 0 if no items match.
    """
    if combobox.count() == 0:
        return -1

    for index in range(combobox.count()):
        if combobox.text(index) == text:
            return index

    return 0


class UiWrapper:
    """Hides a lot of complexity when working with .ui files."""

    def __init__(self, parent, uiname):
        self.filename = joinpath(Config.uidir, uiname + ".ui")
        self.widget = QWidgetFactory.create(self.filename, None, parent)
        self._addchildren(self.widget)

    def _addchildren(self, widget):
        for child in widget.children():
            self.__dict__[child.name()] = child
            if child.children():
                self._addchildren(child)

    def get_widget(self):
        """Returns the main widget."""
        return self.widget

    def _new(self, function, special_name=None):
        """Connects a function and a special name together.
            A special name, in this connection, is a string
            that gives the type and name of a widget, and
            which signal to use, if different from the default
            signal.

            Some examples of "special names" are:
                btn_Ok
                btn_Ok_clicked
                lsv_Input_doubleClicked
                cmb_Selection

            The convensions used here are quite similar to
            the much loved and hated naming conventions
            that are widely used in Windows programming.

            Currently supported types and signals:

            type    widget      signal                qt calls the function with
            --------------------------------------------------------------------
            btn     QButton     clicked (¤)           void
            lsv     QListView   doubleClicked (¤)     QListViewItem, QPoint, int
            lsv     QListView   clicked               QListViewItem
            lsv     QListView   selectionChanged      void
            cmb     QComboBox   activated (¤)         QString
            txt     QLineEdit   returnPressed (¤)     void
            txt     QLineEdit   changed               QString

            (¤ means default, void means nothing)
        """
        
        # Check the function
        if not callable(function):
            return False
        function_name = function.func_name
        
        # Get the right names
        if special_name is None:
            special_name = function_name
        signal_name = ""
        # number of underscores
        nus = special_name.count("_")
        if nus == 2:
            type, name, signal_name = special_name.split("_")
            widget_name = type + "_" + name
        elif nus == 1:
            type, name = special_name.split("_")
            widget_name = special_name
        else:
            return False

        # Get the right widget
        widget = None
        if widget_name in self.__dict__:
            widget = self.__dict__[widget_name]
        if widget is None:
            return False

        # Get the right signal
        # (Just paste in a signal from the qt documentation to add more,
        # but remove the void-returntype. Some lines of code is commented
        # out for easy later inclusion)
        # This could be moved out to a single method, if wished for
        signal = None
        if type == "btn":
            signal = get_signal("clicked()")
        elif type == "lsv":
            if signal_name == "clicked":
                signal = get_signal("clicked(QListViewItem *item)")
            elif signal_name == "selectionChanged":
                signal = get_signal("selectionChanged()")
            else: #doubleClicked
                signal = get_signal("doubleClicked(QListViewItem *, const QPoint &, int)")
        elif type == "txt":
            if signal_name == "changed":
                signal = get_signal("textChanged(const QString &)")
            else: #returnPressed
                signal = get_signal("returnPressed()")
        elif type == "cmb":
            #activated
            signal = get_signal("activated(const QString &)")
        elif signal is None:
            return False

        # Connect and return
        self.widget.connect(widget, signal, function)
        return True

    def autoconnect(self, object):
        """Tries to connect the right functions in an object
            with the correct widgets and signals.
        """
        for methodname in dir(object):
            if "_" in methodname[3:]:
                self._new(getattr(object, methodname))


class InputOutputView:
    """Combine two QListViews and a bunch of TreeNodes in a mindboggingly easy way."""

    def __init__(self, inputQListView, outputQListView, forest, selection=[]):
        # a TreeView is a QListViewWrapper with a tree of TreeNodes
        self._input_treeview = TreeView(QListViewWrapper(inputQListView), forest)
        self._output_treeview = TreeView(QListViewWrapper(outputQListView))

        # apply the given selection
        if selection:
            self.set_selection(selection)

        # keep track of data changes
        # (must be set after set_selection)
        self._dirty = False

        self.widgets = {}

    def set_widgets(self, btn_Clear, btn_Remove, btn_Add):
        """Widgets to sync with the lists
            in terms of beeing enabled or disabled
        """
        self.widgets["clear"] = btn_Clear
        self.widgets["remove"] = btn_Remove
        self.widgets["add"] = btn_Add
    
    def set_dirty(self, dirty=True):
        """Tag the data as changed or not."""
        self._dirty = dirty

    def sync_enabled(self):
        """Synchronizes the buttons with the data,
            in terms of beeing enabled or disabled
        """
        # Sync the widgets
        if self.widgets:
            state = self.has_data()
            for key in ["clear", "remove"]:
                self.widgets[key].setEnabled(state)
            state = self._output_treeview.has_selected() or (len(self._output_treeview) == 1)
            self.widgets["remove"].setEnabled(state)
            state = self._input_treeview.has_selected() or (len(self._input_treeview) == 1)
            self.widgets["add"].setEnabled(state)

    def is_dirty(self):
        """Has the data changed? Am I dirty?"""
        return self._dirty

    def add_to_output(self, item):
        """Adds an item from the input view to the output view."""
        if item:
            treenode = self._input_treeview.to_treenode(item)
            try:
                added = self._output_treeview.add_flat_treenode(treenode)
            except KeyError:
                added = []
            self.set_dirty(bool(added))

    def remove_from_output(self, item):
        """Removes an item from the output view."""
        if item:
            treenode = self._output_treeview.to_treenode(item)
            try:
                removed = self._output_treeview.remove_flat_treenode(treenode)
            except KeyError:
                removed = []
            self.set_dirty(bool(removed))

    def add_selected(self):
        """Add the items selected in the input view to the output view."""
        treenodes = self._input_treeview.get_selected_treenodes()
        if treenodes:
            try:
                added = self._output_treeview.add_flat_treenodes(treenodes)
            except KeyError:
                added = []
            self.set_dirty(bool(added))

    def remove_selected(self):
        """Remove the items selected in the output view."""
        treenodes = self._output_treeview.get_selected_treenodes()
        if treenodes:
            try:
                removed = self._output_treeview.remove_flat_treenodes(treenodes)
            except KeyError:
                removed = []
            self.set_dirty(bool(removed))

    def set_selection(self, selections):
        """Uses a selection on the input view to add items to the output view."""
        items = self._input_treeview.items_by_selections(selections)
        treenodes = self._input_treeview.to_treenode(items)
        if treenodes:
            try:
                added = self._output_treeview.add_flat_treenodes(treenodes)
            except KeyError:
                added = []
            self.set_dirty(bool(added))

    def get_active_selection(self):
        """Returns the active (name, selection)."""
        return self._output_treeview.get_active_selection()

    def clear(self):
        """Removes all output ports."""
        if self.has_data():
            self._output_treeview.clear()
            self.set_dirty(True)

    def __len__(self):
        """The length of the list of outputs."""
        return len(self._output_treeview)

    def has_data(self):
        """Do we have any outputs?"""
        return bool(len(self))

    def has_selected(self):
        """Is one of the inputs selected?"""
        return self._input_treeview.has_selected()


class TreeView:
    """A powerful binding between trees and widgets
    
        Some variable names you'll encounter here:
            treenode                a TreeNode object
            treenodes               a list with TreeNode objects
            listviewwrapper         a QListViewWrapper object
            item                    a QListViewItem object
            key                     a tuple of (treenode.index2d, treenode)
            anything                a key, a treenode or an item
            self.mylist             a QListViewWrapper object
    """

    def __init__(self, listviewwrapper, treenodes=None):
        """Save the given listviewwrapper and treenode(s) internally."""
        self.mylist = listviewwrapper
        if treenodes:
            if isinstance(treenodes, list):
                self.add_treenodes(treenodes)
            elif isinstance(treenodes, TreeNode):
                self.add_treenode(treenodes)
            else:
                raise TypeError("treenodes must be a treenode or a list of treenodes")

    def is_key(self, anything):
        if isinstance(anything, tuple) and len(anything) == 2:
            return anything[0] == anything[1].index2d
        else:
            return False

    def to_key(self, anything):
        """Conversion function.
            Note that QListViewItem can only be converted to a key if it has been added.
        """
        if isinstance(anything, TreeNode):
            return (anything.index2d, anything)
        elif isinstance(anything, QListViewItem):
            return self.mylist.item2key(anything)
        elif isinstance(anything, list):
            return map(self.get_key, anything)
        elif self.is_key(anything):
            return anything
        else:
            raise TypeError("Parameter must be a TreeNode, a QListViewItem, a key or a list of one of them.")

    def to_item(self, anything):
        """Conversion function.
            Note that TreeNode can only be converted to QListViewItem if it has been added.
        """
        if self.is_key(anything):
            return self.mylist.key2item(anything)
        elif isinstance(anything, TreeNode):
            return self.mylist.key2item(self.get_key(anything))
        elif isinstance(anything, list):
            return map(self.get_item, anything)
        elif isinstance(anything, QListViewItem):
            return anything
        else:
            raise TypeError("Parameter must be a TreeNode, a QListViewItem, a key or a list of one of them.")

    def to_treenode(self, anything):
        """Conversion function.
            Note that QListViewItem can only be converted to TreeNode if it has been added.
        """
        if self.is_key(anything):
            return anything[1]
        elif isinstance(anything, QListViewItem):
            return self.mylist.item2key(anything)[1]
        elif isinstance(anything, list):
            if len(anything):
                if self.is_key(anything[0]):
                    return dict(anything).itervalues()
                elif isinstance(anything[0], QListViewItem):
                    fun = self.mylist.item2key
                    return dict(map(fun, anything)).itervalues()
                elif isinstance(anything[0], TreeNode):
                    return anything
                else:
                    return map(self.to_treenode, anything)
            else:
                return []
        elif isinstance(anything, TreeNode):
            return anything
        else:
            raise TypeError("Parameter must be a TreeNode, a QListViewItem, a key or a list of one of them.")

    def add_treenode(self, treenode, recursively=True, parent=None):
        """Add a treenode recursively.
            Returns the generated toplevel item.
            Raises a KeyError on duplicate entries.
        """
        if DEBUG:
            if not isinstance(treenode, TreeNode):
                raise TypeError("treenode must be a TreeNode")
            if not (parent is None or isinstance(parent, QListViewItem)):
                raise TypeError("parent must be a QListViewItem")
        if recursively:
            self.current = self.add_treenode(treenode=treenode, recursively=False, parent=parent)
            for child in treenode:
                self.add_treenode(treenode=child, recursively=True, parent=self.current)
            self.current = self.current.parent()
        else:
            return self.mylist.append(parent=parent, key=self.to_key(treenode), textlist=[treenode.name, treenode.info])

    def add_treenodes(self, treenodes):
        """Add several treenodes recursively.
            Returns the generated items.
            Raises a KeyError on duplicate entries.
        """
        items = []
        keyerror = None
        for treenode in treenodes:
            try:
                items.append(self.add_treenode(treenode))
            except KeyError, the_error:
                keyerror = the_error
        if keyerror:
            raise keyerror
        return items
    
    def add_flat_treenode(self, treenode):
        """Add a treenode, and none if its children.
            Return the genereted item.
            Raises a KeyError on duplicate entries.
        """
        return self.add_treenode(treenode, False)
    
    def add_flat_treenodes(self, treenodes):
        """Add a list of treenodes non-recursively.
            Return the generated items.
            Raises a KeyError on duplicate entries.
        """
        items = []
        keyerror = None
        for treenode in treenodes:
            try:
                items.append(self.add_flat_treenode(treenode))
            except KeyError, the_error:
                keyerror = the_error
        if keyerror:
            raise keyerror
        return items

    def remove_treenode(self, treenode, recursively=True):
         """Removes a treenode and all its children."""
         if DEBUG:
            if not isinstance(treenode, TreeNode):
                raise TypeError("treenode must be a TreeNode")
         if recursively:
            for child in treenode:
                self.mylist.pop(self.to_key(child))
                self.remove_treenode(child, True)
            self.mylist.pop(self.to_key(treenode))
         else:
            return self.mylist.pop(self.to_key(treenode))

    def remove_flat_treenode(self, treenode):
        """Removes a treenode."""
        return self.remove_treenode(treenode, False)

    def remove_flat_treenodes(self, treenodes):
        """Removes several treenodes."""
        rem = self.remove_treenode
        return [rem(treenode, False) for treenode in treenodes]
        
    def get_selected_items(self):
        return self.mylist.get_selected()

    def get_selected_keys(self):
        return self.to_key(self.get_selected_items())

    def get_selected_treenodes(self):
        return self.to_treenode(self.get_selected_items())
    
    def get_active_selection(self):
        """Get the treenodes as a list of index2d.
            This result is one of the main reasons for this class to live.
        """
        return dict(self.mylist.keys()).keys()

    def item_by_selection(self, selection):
        """Takes a selection, returns an item."""
        for key in self.mylist.keys():
            if key[0] == selection:
                return self.to_item(key)

    def items_by_selections(self, selections):
        """Takes a list of selections, returns items."""
        return [self.mylist[key] for key in self.mylist.keys() if key[0] in selections]

    def __len__(self):
        return len(self.mylist)

    def clear(self):
        """Removes all the items and treenodes."""
        self.mylist.clear()

    def has_selected(self):
        """Is something selected?"""
        return bool(self.get_selected_items())


class QListViewWrapper(dict):
    """Wraps a QListView and makes it usable.

        Deals only with QListViews, QListViewItems
        and keys to, QListViewItems. Keys must be unique.
        QListViewWRapper knows nothing of what the keys
        really are, as long as they works as keys.
    """

    def __init__(self, qlistview):
        """Takes a qlistview."""
        self.listview = qlistview

    def item2key(self, item):
        """Takes an item, returns a key."""
        return self.keys()[self.values().index(item)]

    def key2item(self, key):
        """Takes a key, returns an item."""
        return self[key]

    def append(self, parent, key, textlist):
        """Appends an item by a parent, key and the given text.
            The parent is a QListView or QListViewItem.
            If the parent is Null, the internal QListView is used.
            Raises a KeyError if it's a duplicate key.
            Returns the produced QListViewItem.
        """
        if key in self:
            raise KeyError("Duplicate key")
        if parent is None:
            self[key] = QListViewItem(self.listview, *textlist)
        elif isinstance(parent, (QListView, QListViewItem)):
            self[key] = QListViewItem(parent, *textlist)
        else:
            msg = "parent must be None, a QListView or a QListViewItem."
            raise TypeError(msg + "Right now it's: %s" % str(type(parent)))
        return self[key]

    def pop(self, key):
        """Remove an item by a given key, and return the item."""
        item = self[key]
        self.listview.takeItem(item)
        del self[key]
        return item

    def get_selected(self):
        """Return the last item left or a list of selected items."""
        if len(self) == 1:
            # as a list, yes
            return self.values()
        else:
            return [item for item in self.itervalues() if item.isSelected()]

    def clear(self):
        """Remove all items."""
        dict.clear(self)
        self.listview.clear()


class TreeNode:
    """A treenode that you can build oh so pretty trees from."""
    
    def __init__(self, name, data):
        self.name = name
        self.data = data
        self.children = []
        self.info = ""
        self.index2d = []
        
    def __len__(self):
        return len(self.children)

    def __getitem__(self, index):
        return self.children[index]

    def __repr__(self):
        return self.name

    def __str__(self, depth=0):
        s = "\t" * depth + "* " + self.name
        if self.info:
            s += " " + self.info 
        lself = len(self)
        if lself:
            plural_ending = "ren" * (lself > 1)
            s += " with %i child" % lself + plural_ending + ":\t" + str(self.index2d) + "\n"
        else:
            s += "\t" + str(self.index2d) + "\n"
        depth += 1
        for child in self.children:
            s += child.__str__(depth) + "\n"
        if s.endswith("\n"):
            s = s[:-1]
        return s

    def add_child(self, node):
        self.children.append(node)

    def get_children(self):
        return self.children

    def set_info(self):
        """Set the info field automagically.
            Really handy if the data in the node is RObject derived.
        """
        rows = get_rows(self.data)
        cols = get_cols(self.data)
        if rows + cols < 0:
            try:
                self.data + 1
                self.info = "[scalar]"
            except TypeError:
                pass 
        elif cols == -1:
            pass
            #self.info = "[%i children]" % rows
        else:
            if rows + cols > 0:
                self.info = "[%ix%i matrix]" % (rows, cols)

    def has_children(self):
        return bool(self.children)

    def get_by_index(self, index2d):
        children = self.children
        for index in index2d:
            try:
                treenode = children[index]
            except IndexError:
                raise IndexError("The 2d-index tries to reach children that doesn't exist!")
            if treenode.has_children():
                children = treenode.get_children()
            else:
                return children[index]
        raise IndexError("Unable to lookup any treenode with the given index.")


class ComboWrapper(list):

    def __init__(self, combobox, synclist=[]):
        """- combobox is a QComboBox.
            - synclist is a list of widgets to enable and disable
              synchronously with the combobox.
        """
        list.__init__(self)
        self.combobox = combobox
        self.combobox.clear()
        self._synclist = synclist

    def set_selected(self, string):
        """Sets the selected item, takes a string."""
        self.combobox.setCurrentItem(self.index(string))

    def get_selected(self):
        """Return the selected item."""
        return self[self.combobox.currentItem()]

    def select_last(self):
        """Selects the last item."""
        if len(self):
            index = self.index(self[-1])
            self.combobox.setCurrentItem(index)
        else:
            self.combobox.setCurrentItem(-1)

    def set_enabled(self, enabled=True):
        """Set the combobox to enabled or disabled."""
        self.combobox.setEnabled(enabled)
        # Synchronize enable/disable on the the widgets
        [widget.setEnabled(enabled) for widget in self._synclist]

    def clear(self):
        """Clear out the list."""
        list.__init__(self, [])
        self.combobox.clear()

    def remove(self, string):
        """Removes an item, takes a string."""
        index = self.index(string)
        string = self[index]
        del self[index]
        return string

    def fill(self, given_list, sort=True):
        """Fills the listbox with items,
            selects the last item and refreshes.
        """
        stringlist = given_list
        if sort:
            stringlist.sort()
        list.__init__(self, filter(bool, stringlist))
        self.refresh()
        self.select_last()

    def remove_selected(self):
        """Removes the selected item."""
        index = self.combobox.currentItem()
        string = self[index]
        del self[index]
        return string

    def append(self, string):
        """Appends an item."""
        list.append(self, string)
        self.refresh()
        self.select_last()

    def __delitem__(self, index):
        """Removes an item."""
        list.__delitem__(self, index)
        self.refresh()
        if index < len(self):
            self.combobox.setCurrentItem(index)
        else:
            self.select_last()
        if len(self) == 0:
            self.combobox.clear()

    def refresh(self):
        """Refresh the combobox with the internal list.
            Disables the combobox if it's empty.
        """
        if len(self):
            self.combobox.clear()
            for string in self:
                self.combobox.insertItem(string)
            if not self.combobox.isEnabled():
                self.set_enabled()
        else:
            if self.combobox.isEnabled():
                self.set_enabled(False)

    def has_data(self):
        """Do we have any data?"""
        return bool(len(self))


class EventReceiver:
    """Receive data in any thread, process it in event thread.

    Subclasses must implement handle_event(self, event). The passed object
    can be any object the application needs to pass between threads.

    Any thread can safely call post_event on an instance of a subclass with
    some data, and is guaranteed that handle_event is called with the same
    argument in Qt's main gui thread.

    A threadsafe Queue implementation is used to pass data between threads.
    """
    def __init__(self):
        self._unhandled_events = Queue()
        self._timer = QTimer(self)
        self._timer.start(100) # ms
        QObject.connect(self._timer, SIGNAL("timeout()"), self._handle_events)

    def post_event(self, event):
        """Callable by any thread wanting data processed by event thread."""
        self._unhandled_events.put(event)

    def _handle_events(self):
        try:
            while True:
                event = self._unhandled_events.get(0)
                # call method in subclass
                self.handle_event(event)
        except Empty:
            pass


class ValueListViewItem(QListViewItem):
    """QListView entry with connected data of any type."""
    def __init__(self, parent, value, *text):
        QListViewItem.__init__(self, parent, *text)
        self._value = value

    def get_value(self):
        return self._value


def list_view_to_dict(list_view):
    result = {}
    child = list_view.firstChild()
    current = child
    while current:
        result[str(current.text(0))] = list_view_to_dict(current)
        current = current.nextSibling()
    return result

def list_view_column_text_to_dict(list_view):

    
    if hasattr(list_view, "listView"):
        num_cols = list_view.listView().columns()
    else:
        num_cols = list_view.columns()
    
    result = {}
    child = list_view.firstChild()
    current = child
    while current:
        key = tuple([str(current.text(x)) for x in range(num_cols)])        
        result[key] = list_view_column_text_to_dict(current)
        current = current.nextSibling()
    return result
