# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from vtk import vtkMath, vtkTransform
from math import acos, sqrt

class Vector3D(list):
    def __init__(self, endPoint=(0,0,0), startPoint=(0,0,0)):
        """Create a new Vector3D from an existing vector or a 3D coordinate tuple."""
        list.__init__(self)
        self.extend((endPoint[0] - startPoint[0],
                     endPoint[1] - startPoint[1],
                     endPoint[2] - startPoint[2],0))

    def getPerpendicular(self, angle = 0):
        a = [0,0,0]
        b = [0,0,0]
        vtkMath.Perpendiculars(self[0:3], a, b, 0)
        return Vector3D(b)

    def dot(self, otherVector):
        """Compute the dot product of this vector and other."""
        # ignore homogenous coordinate
        return self[0]*otherVector[0] + self[1]*otherVector[1] + self[2]*otherVector[2]
    
    def angle(self, otherVector):
        """Compute angle (in degrees) between this vector and other."""

        dotProduct = self.dot(otherVector)
        lengthProduct = self.length() * otherVector.length()
        divresult = dotProduct / lengthProduct
        if divresult >= 1.0:
            divresult = 0.999999999999999
        elif divresult <= -1.0:
            divresult = -0.999999999999999
        result = acos(divresult)
        return result

    def rotateAboutPerpendicular(self, angle):
        """Rotate angle radians about an arbitrary perpendicular axis."""
        perpendicular = self.getPerpendicular(0)
        self.rotateAboutAxis(perpendicular, angle)

    def rotateAboutAxis(self, axis, angle):
        """Rotate angle radians about a given axis."""
        transform = vtkTransform()
        transform.RotateWXYZ(angle*vtkMath.RadiansToDegrees(), axis[0:3])

        result = transform.TransformPoint(self[0:3])
        self[0:3] = result

    def setLength(self, length):
        """Convenience method for setting the length of the vector, keeping
        the original direction."""
        self.normalize()
        self.multiply(length)

    def cross(self, vector):
        """Return the cross-product of this vector and another."""
        result = [0.0, 0.0, 0.0]
        vtkMath.Cross(self[0:3], vector[0:3], result)
        
        return Vector3D(result)

    def copy(self):
        return Vector3D(self)

    def length(self):
        return sqrt(self[0]*self[0]
                    + self[1]*self[1]
                    + self[2]*self[2])

    def normalize(self):
        """Make this vector a unit vector (length = 1.0) in same direction."""
        if self.length() < 0.001:
            print "Warning: normalizing zerovector."
        vec = self[0:3]
        vtkMath.Normalize(vec)
        self[0:3] = vec

    def multiply(self, value):
        for i in range(3):
            self[i] *= value

    def add(self, vector):
        for i in range(3):
            self[i] += vector[i]

    def negate(self):
        """Flip vector in opposite direction."""
        self.multiply(-1.0)
