# -*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from chempy import feedback, Bond
from Library.Util.molecular import getAttached, InvalidReplacePointError, getRelatedBonds
from Library.Util.math3d.Point3D import Point3D
from Library.Util.math3d.Vector3D import Vector3D

import copy, math

class Fragment:
    """Represents a molecule-part to be connected to other molecules."""

    def __init__(self, model, dummyList):
        """Create a new fragment by copying the model, with dummyList
        as a list of indices of attachementpoints. These will be
        replaced when attaching fragment to a model.
        Currently only a single attachment point is implemented.
        """

        self.model = copy.deepcopy(model)
        self.dummyList = dummyList

    def getModel(self):
        return self.model

    def getNumAttachmentPoints(self):
        return len(self.dummyList)

    def attachToMolecule(self, otherModel, replaceList):

        """Attach this fragment to molecule, using replaceList as list
        of atoms in otherModel to be replaced. All atoms in dummyList
        and all atoms in replaceList will be removed.

        Fragment:   H-O-H       Molecule: H-C=H
                    ^                       |
                    |                       H <- in replaceList
                    - In self.dummyList

        Result:
                    H-C=H
                      |
                      O
                      |
                      H

        When attaching fragments having two attachment points/dummies,
        the fragment is positioned to the best position for attaching
        at the first attachment point, then rotated along that bond to
        fit the second as best as it can.
        """

        # make sure dummy-atoms exist
        if not self.dummyList:
            return
        # make sure fragment and number to attach is similar
        elif len(self.dummyList) != len(replaceList):
            return
        # make sure dummy and "replace-atoms" are only connected to a single atom each
        for i in range(len(replaceList)):
            attachedToDummy = getAttached(self.model, self.dummyList[i])
            attachedToReplace = getAttached(otherModel, replaceList[i])
            if len(attachedToReplace) != 1:
                raise InvalidReplacePointError("Only one atom should be attached to replaced atom %d, instead was: "%i + str(attachedToReplace))
            elif len(attachedToDummy) != 1:
                raise InvalidReplacePointError("Only one atom should be attached to dummy atom %d, instead was: "%i + str(attachedToDummy))
            # Make sure bond-orders are the same
            dummyBond = getRelatedBonds(self.model, self.dummyList[i])[0]
            replaceBond = getRelatedBonds(otherModel, replaceList[i])[0]
        
        feedback['actions'] = 0
        
        modelConnectAtom = otherModel.atom[attachedToReplace[0]]
        modelReplaceAtom = otherModel.atom[replaceList[0]]
        fragmentConnectAtom = self.model.atom[attachedToDummy[0]]
        fragmentDummyAtom = self.model.atom[self.dummyList[0]]
        
        fragmentVector = Vector3D(startPoint = fragmentConnectAtom.coord,
                                  endPoint = fragmentDummyAtom.coord)
        modelVector = Vector3D (startPoint = modelReplaceAtom.coord,
                                endPoint = modelConnectAtom.coord)
        
        # Align fragment with model so dummy and replacer align with the real
        # atoms that connect to eachother.
        rotationAxis = fragmentVector.cross(modelVector)
        angle = fragmentVector.angle(modelVector)

        offsetVector = Vector3D(startPoint = fragmentConnectAtom.coord,
                                endPoint = modelReplaceAtom.coord)

        for atom in self.model.atom:
            pos = Point3D(atom.coord)
            # move all points in fragment so fragmentConnectAtom reaches
            # modelReplaceAtom
            pos.add(offsetVector)
            
            # rotate points in fragment so binding aligns
            axisStart = Point3D(modelReplaceAtom.coord)
            pos.rotateAboutFreeAxis(rotationAxis, axisStart, angle)
            atom.coord = tuple(pos[0:3])
            
        if len(self.dummyList) > 1:
            attachedToSecondDummy = getAttached(self.model, self.dummyList[1])
            attachedToSecondReplace = getAttached(otherModel, replaceList[1])

            # check for connections where both dummies and replaced atoms are connected to same atom, this involves increasing bonds
            if attachedToSecondDummy == attachedToDummy and attachedToReplace == attachedToSecondReplace:
                bondOrderTotalFragment = getRelatedBonds(self.model, self.dummyList[0])[0].order + getRelatedBonds(self.model, self.dummyList[1])[0].order
                bondOrderTotalModel = getRelatedBonds(otherModel, replaceList[0])[0].order + getRelatedBonds(otherModel, replaceList[1])[0].order

                if bondOrderTotalFragment != bondOrderTotalModel:
                    raise InvalidReplacePointError("Can't connect connect fragment with total valence %d to atom with valence %d."%(bondOrderTotalFragment, bondOrderTotalModel))
                
                otherModel.delete_list(replaceList)
                self.model.delete_list(self.dummyList)
                otherModel.merge(self.model)
                b = Bond()
                b.index = [otherModel.index_atom(modelConnectAtom), otherModel.index_atom(fragmentConnectAtom)]
                b.order = bondOrderTotalModel
                otherModel.add_bond(b)
            else:       
                # rotate fragment if necessary, to shorten the path between the second attachment points
                rotationVector = Vector3D(modelVector)
                rotationPosition = Point3D(modelConnectAtom.coord)
                secondFragmentConnectAtom = self.model.atom[attachedToSecondDummy[0]]
                secondModelConnectAtom = otherModel.atom[attachedToSecondReplace[0]]

                fragmentPosition = Point3D(secondFragmentConnectAtom.coord)
                modelPosition = Point3D(secondModelConnectAtom.coord)

                # find angle giving shortest distance
                bestAngle = 0.0
                shortestDistance = fragmentPosition.distanceTo(modelPosition)

                steps = 20
                for step in range(steps):
                    angle = 2*math.pi*step/steps
                    fragmentPosition.rotateAboutFreeAxis(rotationVector, rotationPosition, angle)
                    newDistance = fragmentPosition.distanceTo(modelPosition)
                    if  newDistance < shortestDistance:
                        shortestDistance = newDistance
                        bestAngle = angle

                # rotate atoms
                for a in self.model.atom:
                   atomPosition = Point3D(a.coord)
                   atomPosition.rotateAboutFreeAxis(rotationVector, rotationPosition, bestAngle)
                   a.coord = tuple(atomPosition[0:3])
                
                connections = []
                for i in range(len(replaceList)):
                    modelAtom = otherModel.atom[getAttached(otherModel, replaceList[i])[0]]
                    fragmentAtom = self.model.atom[getAttached(self.model, self.dummyList[i])[0]]
                    bondOrder = getRelatedBonds(self.model, self.dummyList[i])[0].order
                    connections.append((modelAtom, fragmentAtom, bondOrder))

                otherModel.delete_list(replaceList)
                self.model.delete_list(self.dummyList)
                otherModel.merge(self.model)
                for c in connections:
                    b = Bond()
                    b.index = [otherModel.index_atom(c[0]), otherModel.index_atom(c[1])]
                    b.order = c[2]
                    otherModel.add_bond(b)
        else:
            # clean up
            otherModel.delete_atom(replaceList[0])
            self.model.delete_atom(self.dummyList[0])
            otherModel.merge(self.model)
            b = Bond()
            b.index = [otherModel.index_atom(modelConnectAtom), otherModel.index_atom(fragmentConnectAtom)]
            otherModel.add_bond(b)

