# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import copy

from Numeric import Float, array, identity, matrixmultiply
from chempy import Atom
from chempy import Bond
from chempy.pdb import PDB
from math import cos, pi, sin, sqrt

from Library.Plugins.OctavePlugin import OctavePlugin
from Library.Util.math3d.Point3D import Point3D
from Library.Util.math3d.Vector3D import Vector3D

def center_molecule(molecule):
    """Returns a centered molecule.

    The format of the molecule is [(x1, y1, z1), (x2, y2, z2), ...] where each
    tuple specifies the coordinates of an atom in the molecule."""

    mean = (0.0, 0.0, 0.0)
    for v in molecule:
        mean = (mean[0] + v[0], mean[1] + v[1], mean[2] + v[2])
    l = len(molecule)
    mean = (mean[0] / l, mean[1] / l, mean[2] / l)
    for i in range(len(molecule)):
        molecule[i] = (molecule[i][0] - mean[0],
                       molecule[i][1] - mean[1],
                       molecule[i][2] - mean[2])
    return molecule

class Procrustes:
    """A wrapper class for the Octave Procrustes routine.

    The format of a molecule is [(x1, y1, z1), (x2, y2, z2), ...] where each
    tuple specifies the coordinates of an atom in the molecule.
    """
    def __init__(self, reference_molecule, reference_mapping):
        """reference_molecule - coordinates of the reference molecule.
        reference_mapping - a vector which specifies that atom reference_mapping[i] is mapped to
        index i.
        """
        self.reference_molecule = reference_molecule
        self.n_mappings = len(reference_mapping)
        self.reference_mapping = self.dict2list(reference_mapping)
        self.octave_plugin = OctavePlugin()

    def procrustes(self, test_molecule, mapping):
        """Performs procrustes on the test_molecule, comparing it to the
        instance reference molecule and returns the resulting molecule.
        
        test_molecule - coordinates of the test molecule.
        test_mapping - a vector which specifies that atom test_mapping[i] is mapped to
        index i.
        """
        output = self.octave_plugin.execute_function("procrus_wrapper", (('Uj', test_molecule),
                                                                         ('vj', self.dict2list(mapping)),
                                                                         ('Ur', self.reference_molecule),
                                                                         ('vr', self.reference_mapping)),
                                                                         ("out",))
        return output[0].getData()

    def dict2list(self, dict):
        """Translates a dictionary mapping into a list (vector) mapping"""
        list = range(self.n_mappings + 1)
        for i in range(len(list)):
            list[i] = 0
            if dict.has_key(i):
                list[i] = dict[i] + 1
        #print "list mapping:", list
        return list[1:]

def atomic_no2symbol(n):
    """Returns the atomic symbol of a limited set of atoms with of the atomic
    number n. If the atomic symbol is not listed, 'Unknown' is returned"""
    symbols = {
        1 : 'H',
        6 : 'C',
        7 : 'N',
        8 : 'O',
        11: 'Na',
        15: 'P',
        16: 'S',
        17: 'Cl',
        20: 'Ca',
        25: 'Mn',
        30: 'Zn',
        35: 'Br'
        }
    if symbols.has_key(n):
        return symbols[n]
    else:
        return 'Unknown'

# Br    C     Cl    F     H     I     N     O     P     S     Si
_single_bond_length = [\
[2.28, 1.94, 2.14, 1.76, 1.41, None, 2.14, None, 2.20, 2.27, 2.16],\
[1.94, 1.54, 1.77, 1.38, 1.08, 2.14, 1.47, 1.43, 1.84, 1.82, 1.85],\
[2.14, 1.77, 1.99, 1.63, 1.28, 2.32, 1.97, 1.70, 2.03, 1.99, 2.02],\
[1.76, 1.38, 1.63, 1.42, 0.92, 2.57, 1.36, 1.42, 1.54, 1.58, 1.56],\
[1.41, 1.08, 1.28, 0.92, 0.74, 1.60, 1.01, 0.97, 1.42, 1.34, 1.48],\
[None, 2.14, 2.32, 2.57, 1.60, 2.67, None, None, 2.47, None, 2.43],\
[2.14, 1.47, 1.97, 1.36, 1.01, None, 1.46, 1.36, None, None, 1.74],\
[None, 1.43, 1.70, 1.42, 0.97, None, 1.36, 1.48, 1.54, None, 1.63],\
[2.20, 1.84, 2.03, 1.54, 1.42, 2.47, None, 1.54, 2.21, None, None],\
[2.27, 1.82, 1.99, 1.58, 1.34, None, None, None, None, 2.05, 2.15],\
[2.16, 1.85, 2.02, 1.56, 1.48, 2.43, 1.74, 1.63, None, 2.15, 2.32]]

# C     N     O     S
_double_bond_length = [\
[1.34, 1.30, 1.22, None],\
[1.30, 1.25, None, None],\
[1.22, None, 1.21, None],\
[None, None, None, 1.89]]\

# C     N
_triple_bond_length = [\
[1.20, 1.16],\
[1.16, 1.10]]\

def bond_order(atom1, atom2, d, slack=0.1):
    """Return the order of bonding between atom1 and atom2 which are d
    angstroms apart.
    
    A slack, 10 per cent default, is allowed.
    """

    # Try 1st order bonds first

    atom_index = {35:0, 6:1, 17:2, 9:3, 1:4, 53:5,\
    7:6, 8:7, 15:8, 16:9, 14:10} # atomic number:column
    if not atom_index.has_key(atom1)\
    or not atom_index.has_key(atom2):
        return 0

    row = atom_index[atom1]
    col = atom_index[atom2]
    index_d = _single_bond_length[row][col]
    if index_d and d < index_d * (1 + slack) and d > index_d * (1 - slack):
        return 1
    
    # Then try 2nd order bonds

    atom_index = {6:0, 7:1, 8:2, 16:3}
    if not atom_index.has_key(atom1)\
    or not atom_index.has_key(atom2):
        return 0

    row = atom_index[atom1]
    col = atom_index[atom2]
    index_d = _double_bond_length[row][col]
    if index_d and d < index_d * (1 + slack) and d > index_d * (1 - slack):
        return 2

    # Then try 3rd order bonds

    atom_index = {6:0, 7:1}
    if not atom_index.has_key(atom1)\
    or not atom_index.has_key(atom2):
        return 0

    row = atom_index[atom1]
    col = atom_index[atom2]
    index_d = _triple_bond_length[row][col]
    if index_d and d < index_d * (1 + slack) and d > index_d * (1 - slack):
        return 3

def alen((x,y,z)):
    """Find the length of a vector."""
    return sqrt(x*x+y*y+z*z)

def rotation_matrix(angle, vec):
    """Generate rotational matrix for rotation of point about
    arbitrary vector [x,y,z].
    """

    x = vec[0]
    y = vec[1]
    z = vec[2]
    
    s = sin(angle)
    c = cos(angle)
    mag = alen((x,y,z))

    x /= mag
    y /= mag
    z /= mag

    xx = x * x
    yy = y * y
    zz = z * z
    xy = x * y
    yz = y * z
    zx = z * x
    xs = x * s
    ys = y * s
    zs = z * s
    one_c = 1 - c

    M = array(((0,0,0,0),
               (0,0,0,0),
               (0,0,0,0),
               (0,0,0,0)), Float)

    M[0][0] = (one_c * xx) + c
    M[0][1] = (one_c * xy) - zs
    M[0][2] = (one_c * zx) + ys
    M[0][3] = 0.0
    M[1][0] = (one_c * xy) + zs
    M[1][1] = (one_c * yy) + c
    M[1][2] = (one_c * yz) - xs
    M[1][3] = 0.0
    M[2][0] = (one_c * zx) - ys
    M[2][1] = (one_c * yz) + xs
    M[2][2] = (one_c * zz) + c
    M[2][3] = 0.0
    M[3][0] = 0.0
    M[3][1] = 0.0
    M[3][2] = 0.0
    M[3][3] = 1.0
    return M

def translation_matrix((vx,vy,vz)):
    M = identity(4,Float)
    M[0][3] = vx
    M[1][3] = vy
    M[2][3] = vz
    return M

def genCompleteMatrix(avec, bvec, PointT, PointB):
    """Create a transformation matrix that rotates a point around
    PointT, then offsets it with PointB-PointT. The rotation is the
    same as if avec is rotated to match bvec."""
    
    # translate -T
    negT = Point3D(PointT)
    for i in range(3):
        negT[i] *= -1
    
    T = translation_matrix(negT[0:3])
    cvec = avec.cross(bvec)
    angle = avec.angle(bvec)
    if  abs(angle) < 0.0001:
        R = identity(4)
    else:
        if cvec.length() < 0.0001:
            xvec = Vector3D((1,0,0))
            # cant quite remember what this does
            if avec.angle(xvec) < 1.0 or abs(avec.angle(xvec) - pi) < 1.0:
                xvec = Vector3D((0,1,0))
            cvec = avec.cross(xvec)
        R = rotation_matrix(angle, cvec)
    T2 = translation_matrix(PointB[0:3])
    temp = matrixmultiply(R,T)
    return matrixmultiply(T2, temp)

def getRelatedBonds(model, atomIndex):
    """Get all bonds to/from atom."""
    bonds = []
    for b in model.bond:
        if atomIndex in b.index:
            bonds.append(b)
    return bonds

def createAtom(symbol):
    a = Atom()
    a.symbol = symbol
    return a

def createBond(indexA, indexB, order):
    """Create a bond object for a given model, using atom indices into that model."""
    b = Bond()
    b.index = [indexA, indexB]
    b.order = order
    return b

def getAttached(model, index):
    """Get indices of all atoms attached through bonds to a certain atom."""
    attached = []
    for b in model.bond:
        if b.index[0] == index:
            if not attached.count(b.index[1]):
                attached.append(b.index[1])
        elif b.index[1] == index:
            if not attached.count(b.index[0]):
                attached.append(b.index[0])
    return attached

class InvalidReplacePointError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


def getAngleBetweenPlanes(a, b, c, d):
    """Find the current angle between the planes ABC and BCD."""
    
    # point e is the endpoint of be, a line perpendicular to bc, from b
    # point f is the endpoint of cf, a line perpendicular to bc, from c

    ba = Vector3D(startPoint=b,endPoint=a)
    bc = Vector3D(startPoint=b,endPoint=c)
    cb = Vector3D(startPoint=c,endPoint=b)
    cd = Vector3D(startPoint=c,endPoint=d)

    be = bc.cross(ba.cross(bc))
    cd = cb.cross(cd).cross(cb)

    return be.angle(cd)

def deleteAtoms(molecules, atoms):
    emptyMolecules = []
    for m in molecules:
        for a in atoms:
            index = m.index_atom(a)
            if index != -1:
                m.delete_atom(index)
        if not m.atom:
            emptyMolecules.append(m)
    for m in emptyMolecules:
        molecules.remove(m)

def cleanModel(model):
    """Remove all references to possible vtk-structures from bonds in
    model.
    """
    for b in model.bond:
        if hasattr(b, "actor"):
            del b.actor

def connectToFragment(molecules, replacedAtoms, fragment):
    """Connect single-attachment-point fragment at each of atoms to be
    replaced."""
    for m in molecules:
        for a in replacedAtoms:
            index = m.index_atom(a)
            if index != -1:
                fragmentCopy = copy.deepcopy(fragment)
                try:
                    fragmentCopy.attachToMolecule(m, (index,))
                except InvalidReplacePointError, e:
                    pass

def connectToDualPointFragment(molecules, replacedAtoms, fragment):
    """Connect double-attachment-point fragment at the two atoms to be
    replaced."""
    molA = molB = None
    for m in molecules:
        index = m.index_atom(replacedAtoms[0])
        if index != -1:
            molA = m
        index = m.index_atom(replacedAtoms[1])
        if index != -1:
            molB = m

    if molA and molB:
        if molA != molB:
            molA.merge(molB)
            molecules.remove(molB)
        fragmentCopy = copy.deepcopy(fragment)
        # ignore more than two first atoms to be replaced
        indices = map(lambda atom : molA.index_atom(atom), replacedAtoms[:2])
        if not -1 in indices:
            try:
                fragmentCopy.attachToMolecule(molA, indices)
            except InvalidReplacePointError, e:
                pass
        

class ConnPDB(PDB):
   def fromList(self,list):

      model = PDB.fromList(self, list)
      
# read atoms
      cnt = 0
      indexToCount = {}
      at = None
      for rec in list:
         if (rec[0:4]=='ATOM') or (rec[0:6]=='HETATM'):
            index = int(rec[6:11])
            indexToCount[index] = int(cnt)
            cnt = cnt + 1            
         elif (rec[0:6]=='CONECT') or (rec[0:6]=='HETATM'):
            currentAtom = indexToCount[int(rec[7:11])]
            connectedAtoms = rec[12:31].split()
            for b in connectedAtoms:
               bond = Bond()
               bond.index = [currentAtom, indexToCount[int(b)]]
               bond.order = 1 #FIXME
               model.bond.append(bond)
      return(model)



from Library.Util.molecular import Fragment
Fragment = Fragment.Fragment
