# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import Numeric

from Library.Util import lists

class ClusterGenerator:

    def __init__(self):
        self.cutoff = None
        self._cluster_index = 1

    def cluster_builder(self, matrix, order):
        
        if isinstance(matrix, Numeric.ArrayType):
            self._matrix = matrix.tolist()
        else:
            self._matrix = matrix
        if isinstance(order, Numeric.ArrayType):
            self._order = lists.flatten(order.tolist())
        else:
            self._order = order

        self.clusters = []
        while self._matrix.count([]) != len(self._matrix):
            maximum_row = self._get_maximum_row()
            self.clusters.append(self._cluster_traverser(maximum_row))

        return self.clusters

    def set_cutoff_level(self, level):
        self.cutoff = level
        for cluster in self.clusters:
            cluster_list = [cluster]
            while cluster_list:
                cluster = cluster_list[0]
                if cluster.get_height()<=level:
                    cluster.set_cluster_mode(1)
                else:
                    cluster.set_cluster_mode(0)
                    if cluster.get_left_cluster():
                        cluster_list.append(cluster.get_left_cluster())
                    if cluster.get_right_cluster():
                        cluster_list.append(cluster.get_right_cluster())
                cluster_list.remove(cluster)

    def reset_cutoff(self):
        for cluster in self.clusters:
            cluster_list = [cluster]
            while cluster_list:
                cluster = cluster_list[0]
                cluster.set_cluster_mode(0)
                if cluster.get_left_cluster():
                    cluster_list.append(cluster.get_left_cluster())
                if cluster.get_right_cluster():
                    cluster_list.append(cluster.get_right_cluster())
                cluster_list.remove(cluster)

    def get_cutoff(self):
        return self.cutoff      
                
    def _cluster_traverser(self, row):

        index = self._matrix.index(row)
        self._matrix[index] = []
        index += 1 #index is used a the id for the cluster

        left_cluster = None
        right_cluster = None
        left_object = None
        right_object = None
        left_index = None
        right_index = None

        if row[0] > 0:
            left_cluster = self._cluster_traverser(self._matrix[int(row[0])-1])            
        else:
            left_object = row[0]
            left_index = self._order.index(abs(row[0])) + 1
        if row[1] > 0:
            right_cluster = self._cluster_traverser(self._matrix[int(row[1])-1])            
        else:
            right_object = row[1]
            right_index = self._order.index(abs(row[1])) + 1

        return Cluster(index, row[-1], left_cluster, right_cluster,
                       left_object, right_object, left_index, right_index)

    def _get_maximum_row(self):
        maximum_row = [0]
        for row in self._matrix:
            if row and row[-1] > maximum_row[-1]:
                maximum_row = row
        return maximum_row

    def _check_matrix(self):
        for row in self.matrix:
            if row: return True

    
class Cluster:

    #cluster modes
    OPEN = 0 #no cutoff
    BAR = 1 #cutoff mode #1
    TRIANGLE = 2 #cutoff mode #2

    def __init__(self, number, height, left_cluster, right_cluster,
                 left_object, right_object, left_index, right_index):
        self.width = 10
        self.cluster_number = number
        self.left_cluster = left_cluster
        self.right_cluster = right_cluster
        self.height = height
        self.selected = False
        self.left_object = left_object
        self.right_object = right_object
        self.left_index = left_index
        self.right_index = right_index
        self.cluster_mode = self.OPEN
        
    def get_cluster_number(self):
        return self.cluster_number

    def get_left_cluster(self):
        return self.left_cluster

    def get_right_cluster(self):
        return self.right_cluster

    def get_left_object(self):
        """object at bottom of cluster-tree.. only leaf clusters have left_object != None"""
        return self.left_object

    def get_right_object(self):
        """object at bottom of cluster-tree.. only leaf clusters have right_object != None"""
        return self.right_object

    def get_left_index(self):
        """Returns the left index for the cluster (where the left 'foot' is drawn from). May change when cluster changes mode."""
        if not self.left_index:
            self.left_index = self.left_cluster.get_left_index() + \
                              (self.left_cluster.get_right_index() - self.left_cluster.get_left_index())/2.
            
        return self.left_index

    def get_right_index(self):
        """Returns the left index for the cluster (where the left 'foot' is drawn from). May change when cluster changes mode."""
        if not self.right_index:
            self.right_index = self.right_cluster.get_left_index() + \
                               (self.right_cluster.get_right_index() - self.right_cluster.get_left_index())/2.

        return self.right_index
            
    def get_height(self):
        return self.height

    def get_cluster_mode(self):
        return self.cluster_mode
    
    def _reset_left_index(self):
        """Method that reset the left index of a cluster.

        In practice this mean that we set left_index = None.
        Important that we dont reset the index of leaf clusters"""

        if not self.get_left_object():
            self.left_index = None
            
    def _reset_right_index(self):
        """Method that reset the right index of a cluster.

        In practice this mean that we set right_index = None.
        Important that we dont reset the index of leaf clusters"""

        if not self.get_right_object():
            self.right_index = None

            
    def set_left_index(self, index):
        """This method is called when we go from cluster_mode:BAR  to cluster_mode:TRIANGLE"""
        self.left_index = index

    def set_right_index(self, index):
        self.right_index = index

    def set_selected(self):
        self.selected = True


    def set_deselected(self):
        self.selected = False
        
    def change_cluster_mode(self, parent_plot):
        self.cluster_mode = (self.cluster_mode + 1)%3

        if self.cluster_mode == self.OPEN:
            self._reset_left_index()
            self._reset_right_index()
            if parent_plot and parent_plot.get_left_cluster() == self:
                parent_plot._reset_left_index()
            if parent_plot and parent_plot.get_right_cluster() == self:
                parent_plot._reset_right_index()
        if self.cluster_mode == self.TRIANGLE:
            #find left index
            cluster_list = [self]
            while cluster_list:
                left_cluster = cluster_list[0]
                if left_cluster.get_left_cluster():
                    cluster_list.append(left_cluster.get_left_cluster())
                else:
                    left_index = left_cluster.get_left_index()
                cluster_list.remove(left_cluster)

            #find right index
            cluster_list = [self]
            while cluster_list:
                right_cluster = cluster_list[0]
                if right_cluster.get_right_cluster():
                    cluster_list.append(right_cluster.get_right_cluster())
                else:
                    right_index = right_cluster.get_right_index()
                cluster_list.remove(right_cluster)

            #index for left or right index for parent cluster
            parent_index = left_index +((right_index -left_index)/2.0)

            #update index
            self.set_left_index(left_index)
            self.set_right_index(right_index)
            if parent_plot and parent_plot.get_left_cluster() == self:
                parent_plot.set_left_index(parent_index)
            if parent_plot and parent_plot.get_right_cluster() == self:
                parent_plot.set_right_index(parent_index)
                
    def is_open(self):
        if self.cluster_mode == self.OPEN:
            return True
        else:
            return False

    def set_cluster_mode(self, mode):
        self.cluster_mode = mode



