# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

# domain exceptions

from Library.Util import lists
import Numeric

class InvalidRangeError(Exception):
    """Invalid range."""
    pass
class InvalidMatrix(Exception):
    """Used when a matrix cant be converted to Numeric.array"""
    pass

def parse_range(range_statement, max_length):
    """This method takes two parameters:
    range_statement is a string formatted as a range. The format is as follows:
    range_statement = "*" | "" | range
    range = span | span, range
    span = int | int - int

    Examples of use are: "1", "1-3", "1-3, 7, 20, 30-41" and "*"

    max_length is used when the range is "*" or "" meaning a full range
    without "holes". It denotes what "full" actually is.

    The method returns a list of indices reflecting the chosen range.
    For example, the range "1-3, 6, 10-12" will produce the list:
    [0,1,2,5,9,10,11] and the range "*" will produce the list:
    [0,1,2,3, ... , max_length - 1].

    Note that values are decremented by 1
    to be compatible with the "python way" of indexing lists.
    """

    result_list = []

    if range_statement.strip() == "*" or range_statement.strip() == "":

        return range(max_length)

    # Split on commas: "1,3,4-6" -> ["1","3","4-6"]
    comma_splitted = range_statement.split(",")

    for subrange in comma_splitted:
        # Split on dashes: ["1","3","4-6"] -> ["1", "3", ["4","6"]]
        single_elements = subrange.split("-")

        # It was a single value, not a range.
        if len(single_elements) == 1 and int(single_elements[0]) <= max_length:
            result_list.append(int(single_elements[0]) - 1)

        # It was a range
        elif len(single_elements) == 2:

            if int(single_elements[0]) > max_length:
                break
            
            start_at = int(single_elements[0]) - 1
            stop_before = int(single_elements[1])
            for elem in range(start_at, stop_before):
                if elem < max_length:
                    result_list.append(elem)

    return result_list


def matlab_parse_range(range, data):
    """Tool for parsing ranges given in matlab ++ notation.
    
    Examples of matlab notations:
    -----------------------------
    
        m:n      - from element m to element n
        m:s:n    - from element m to element n with step s
        m:end    - from elmenet m to last element
        m:end-5  - from elmenet m to last element - 5
        
        m, n     - element m and element n
        m:n, o   - from element m to element n, and element o
        
    """

    range_statements = range.lower().split(',')
    result = []
    
    for statement in range_statements:
        result.extend(_parse_statement(statement, data))
        
    return result
    
def matlab_parse_matrix_range(range_, data, is_column):

    if is_column:
        try:
            data = Numeric.transpose(data).tolist()
        except TypeError, e:
            if str(e) == 'an integer is required':                
                raise InvalidMatrix, 'This data can not be parse ' + \
                      'since it is an invalid link'
            else:
                raise
    return matlab_parse_range(range_, data)


def _parse_statement(range_statement, data):
    """Handles parsing of one statement, eg. m:s:n 
    """

    max_index = len(data)
    range_elems = range_statement.lower().split(':')
    
    start_index = 1
    step_index = 1
    stop_index = max_index
    
    # parse statement:
    
    if len(range_elems) > 3:
        raise InvalidRangeError, "Wrong number of arguments in range expression."
    
    elif len(range_elems) == 1: # notation: 'start'
        start_index = _parse_token(range_elems[0], max_index)
        stop_index = start_index
        
    else: # notation: 'start:stop' or 'start:step:stop'
        start_index = _parse_token(range_elems[0], max_index)
        stop_index = _parse_token(range_elems[-1], max_index)
    
        # notation: 'start:step:stop'
        if len(range_elems) == 3:
            step_index = int(range_elems[1])

    # build the new data set:
    
    # convert from matlab indexes to python list indexes:
    start_index -= 1
    stop_index -= 1

    should_reverse = False

    if step_index < 0:
        step_index = -step_index
        start_index, stop_index = stop_index, start_index
        should_reverse = True

    if len(data) == 0:
        result = []
    elif stop_index < start_index:
        if should_reverse:
            raise InvalidRangeError, "Stop index in range cannot exceed start index in %s when using negative step." % range_statement
        else:
            raise InvalidRangeError, "Start index in range cannot exceed stop index in %s." % range_statement
    elif start_index == stop_index:
        result = [data[start_index]]
    else:
        if step_index == 1:
            result = data[start_index:stop_index+1]
        else:
            result = []
            curr_index = start_index
            while (curr_index < stop_index):
                result.append(data[curr_index])
                curr_index += step_index

    if should_reverse:
        result.reverse()
    return result


def _parse_token(token, max_index):
    """Handles parsing of one token, eg. m 
    """
    
    if token.find('end') != -1:
        end_statement = token.split('-')
        
        if len(end_statement) > 1:
            end_index = int(end_statement[-1])
            
            if end_index > max_index:
                raise InvalidRangeError, "The result from \'end-%i\' is out of bounds." % end_index
            else:
                return max_index - end_index
        else:
            return max_index
    else:
        try:
            value = int(token) 
        except ValueError:
            raise InvalidRangeError, "Part of statement (\'" + token + "\') not a number."
        return value    
    
