# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import os
import sys
import time
import signal
import thread
import random
import string
import re

from Library.Util.Config import Config
from threading import RLock

# Hack to add Pyro to our python path
thirdparty = os.path.join(Config.scicrafthome, "thirdparty")
if thirdparty not in sys.path:
    sys.path.append(thirdparty)

import Pyro.core
import Pyro.constants
import Pyro

# Handle sending nan/inf to R until Python 2.5
Pyro.config.PYRO_PICKLE_FORMAT = 0

class _Handler:
    """Singleton for accessing rpyserver."""
    def __init__(self):
        self._obj = None
        self._rpy_mutex = RLock()
        self._pid = None
     
    def mutex_wrap(self, func_name):
        def do_with_mutex(name, args):
            self._rpy_mutex.acquire()
            try:
                if not self._obj:
                    # Make sure connection to Pyro/rpyserver has been
                    # started before calling function
                    self._start_connection()
                return getattr(self._obj, name)(*args)
            finally:
                self._rpy_mutex.release()
        
        return lambda *args : do_with_mutex(func_name, args)

    def _start_connection(self):
        def update_printing(f):
            while True:
                line = f.readline()
                if line:
                    sys.stdout.write(line)
                else:
                    break

        server_path = os.path.join(Config.scicrafthome, "rpyserver.py")

        if re.search("scicraft.*\.exe", sys.argv[0]):
            self._stdin, self._stdout = os.popen2(os.path.join(Config.scicrafthome, 'rpyserver.exe'))
        else:
            self._stdin, self._stdout = os.popen2('python "%s"'% server_path)

        # password uses normal random module to get values, so is not
        # cryptologically secure yet. should be replaced with urandom
        # module when we support python 2.4
        password = ''.join([random.choice(string.letters + string.digits) for i in range(32)])
        self._stdin.write("%s\n" % password)
        self._stdin.flush()
        
        line = ""
        while not line.startswith("pid: "):
            line = self._stdout.readline()
        self._pid = int(line[5:].strip())
        while not line.startswith("PYRO://"):
            line = self._stdout.readline()
        Pyro.core.initServer(0)
        uri = Pyro.core.PyroURI(line)
        self._obj = uri.getProxy()
        self._obj._setIdentification(password)

        # print data from other process to this process' stdout
        thread.start_new_thread(update_printing, (self._stdout,))

        # initialize handler and error handling
        self._obj.init_handler()

    def __del__(self):
        try:
            self._obj.rpy_silent_eval("quit(save='no')")
        except:
            pass

    def quit(self, timeout):
        if not self._pid:
            # process never started, no point in stopping it
            return

        if self._rpy_mutex.acquire(blocking=False):
            try:
                try:
                    self._obj.rpy_silent_eval("quit(save='no')")
                except Pyro.errors.ConnectionClosedError, e:
                    pass                    
            finally:
                # release it just in case
                self._rpy_mutex.release()
        else:
            # we have to stop a running evaluation
            if sys.platform == 'win32':
                 import win32api, win32con
                 handle = win32api.OpenProcess(win32con.PROCESS_TERMINATE, 0, self._pid)
                 win32api.TerminateProcess(handle, 0)
            else:
                os.kill(self._pid, signal.SIGHUP)
                start_time = time.time()
                while (time.time() - start_time) < timeout:
                    if os.waitpid(self._pid, os.WNOHANG) != (0, 0):
                        return
                    time.sleep(0.1)
                os.kill(self._pid, signal.SIGKILL)

# instanciate singleton
_handler = _Handler()

# Set up interface for accessing rpy/Pyro threadsafely, including some
# convenciene functions
rpy_eval = _handler.mutex_wrap("rpy_eval")
rpy_converted_eval = _handler.mutex_wrap("rpy_converted_eval")
rpy_silent_eval = _handler.mutex_wrap("rpy_silent_eval")
rpy_gc = lambda : rpy_silent_eval("gc()")
rpy_assign = _handler.mutex_wrap("rpy_assign")
rpy_clear = lambda : rpy_silent_eval("rm(list=ls())\ngc()")
rpy_quit = _handler.quit
