# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from Library.Plugins.OctavePlugin import OctavePlugin
from Library.Util.Counter import Counter
from Library.Util import lists
from Library.Util import diagramanalyser
from Library.Util.Config import Config
from System.Plugin import PluginInterrupted
from System.Data.Parameter import Parameter
from System.Data.Data import Data
import time
import tempfile
import os
import Queue


class OctaveRunner:
    """This class will be responsible for running all function nodes that
    are part of a cluster inside a single octave instance
    """
    
    def __init__(self, number):

        self._number_of_clusters = number
        self._queue = Queue.Queue(number)
        self._function_nodes = []
        self._counter = Counter()
        self._plugin = OctavePlugin()
        self._input_mapping = {}
        self._output_mapping = {}
        self._cluster_results = []
        self._silly_mutex = Queue.Queue()
        self._silly_mutex.put("something")
        self._interrupted = False
        self._exception = None

    def execute(self, node):
        self._queue.put(node)
        self._function_nodes.append(node)
        while not self._queue.full():
            time.sleep(0.1)

        try:
            dummy = self._silly_mutex.get_nowait()
            start_time = time.time()
            self._run_cluster()
        except Queue.Empty, e:
            pass

        while not self._cluster_results:            
            time.sleep(0.1)
            if self._interrupted:
                raise PluginInterrupted, 'The Octave plugin was interrupted'
            elif self._exception:
                self._interrupted = True
                raise self._exception
            
        results = self._map_to_old_names(node, self._cluster_results)
        return results

    def _run_cluster(self):
        if Config.composite_script_path:
            (desc, script_file) = tempfile.mkstemp(suffix='.m',
                                                   dir=Config.composite_script_path)
        else:
            (desc, script_file) = tempfile.mkstemp(suffix='.m')

        os.fdopen(desc).close() # dont want to use the file descriptor
            
        # mkstemp might return a name with '-', which cant be used in octave
        script_file = script_file.replace('-','_')
        script_name = os.path.basename(script_file).replace('.m','')
        script_path = os.path.dirname(script_file)

        self._map_variables()
        input_variables = self._generate_input_variables()
        output_variables = self._generate_output_variables()

        self._generate_script(script_name, script_file)
        try:
            try:
                self._cluster_results = self._plugin.execute_function(script_name,
                                                                      input_variables,
                                                                      output_variables,
                                                                      self.isInterrupted,
                                                                      script_path)
            except PluginInterrupted, e:
                self._interrupted = True
            except Exception, e:
                self._exception = e
                
        finally:
            self.clean(script_file) 

    def _get_next_unique_id(self):
        return 'sc' + str(self._counter.get_next_value())


    def _generate_unique_names(self):
        """Generated unique names on variables

        Generates uniqie names on input ports, parameters and output ports.
        These names are mapped to the original names and returned as a dict.
        """
        input_ = {}
        output = {}
        for node in self._function_nodes:
            result_in = {}
            result_out = {}
            
            for var in node.getInputPorts() + node.getParameters():
                result_in[var.getName()] = self._get_next_unique_id()
            for var in node.getOutputPorts():
                result_out[var.getName()] = self._get_next_unique_id()
                
            input_[node.getName()] = result_in
            output[node.getName()] = result_out

        
        return (input_, output)
            
            
    def _map_variables(self):
        """Maps name on input ports to dependent output ports.

        This method maps the name on the input ports to the 
        unique name on the connected output ports on noes in
        the same cluster.
        """

        self._input_mapping, self._output_mapping = \
                             self._generate_unique_names()
        for node in self._function_nodes:
            ports = node.getInputPorts()
            for port in ports:
                owner_port = port.getDataSource()
                owner_node = owner_port.getOwnerNode()
                if owner_node in self._function_nodes:
                    # we have a port which is connected to
                    # a function node in the cluster
                    self._input_mapping[node.getName()][port.getName()] = \
                                                                        self._output_mapping[owner_node.getName()] \
                                                                        [owner_port.getName()]
        
        return self._input_mapping, self._output_mapping


    def _find_safe_input_variables(self):

        safe_variables = {}
        
        for node in self._function_nodes:
            result = []
            for port in node.getInputPorts():
                if port.getDataSource().getOwnerNode() not in self._function_nodes:
                    result.append(port.getName())
            for param in node.getParameters():
                result.append(param.getName())
            safe_variables[node.getName()] = result

        return safe_variables 



    def _generate_input_variables(self):

        safe_vars = self._find_safe_input_variables()
        variables = []
        
        for node in safe_vars:
            safe_inputs = []
            for function in self._function_nodes:
                if function.getName() == node:
                    for var in function.getInputPorts() + \
                        function.getParameters():
                        if var.getName() in safe_vars[node]:
                            safe_inputs.append(var)
                    break
            for var in safe_inputs:
                safe_name = self._input_mapping[node][var.getName()]
                if isinstance(var, Parameter):
                    value = var.getValue()
                else:
                    value = var.getData().getData()
                variables.append((safe_name, value))
                                
        return tuple(variables)


    def _generate_output_variables(self):

        variables = []
        for node in self._function_nodes:
            for var in node.getOutputPorts():
                unique_name = self._output_mapping[node.getName()][var.getName()]
                variables.append(unique_name)
                
        return tuple(variables)


    def _generate_script(self, script_name, file_path = None):

        if not file_path:
            (descriptor, file_path) = tempfile.mkstemp()
            os.fdopen(descriptor).close()
            
        script = open(file_path, 'w')
        safe_inputs = self._find_safe_input_variables()
        all_outputs = self._generate_output_variables()
        all_inputs = []
        function_strings = []

        for (node_name, vars_) in safe_inputs.items():
            for var in vars_:
                all_inputs.append(self._input_mapping[node_name][var])
            
        safe_order = diagramanalyser.find_safe_group_order(self._function_nodes)        

        for node in safe_order:
            function = os.path.splitext(os.path.basename(node.getFunctionFileName()))[0] 
            inputs = node.getInputPorts()
            params = node.getParameters()
            outputs = node.getOutputPorts()
            input_vars = []
            output_vars = []
            for var in inputs + params:
                input_vars.append(self._input_mapping[node.getName()][var.getName()])
            for var in outputs:
                output_vars.append(self._output_mapping[node.getName()][var.getName()])

            function_strings.append(
                lists.to_string(output_vars) + ' = ' +
                function + '(' +
                lists.to_string(input_vars).replace('[','').replace(']','') +
                ');\n')
            

        script.write(
            'function ' +
            lists.to_string(list(all_outputs)) +
            ' = ' + script_name  + '(' +
            lists.to_string(list(all_inputs)).replace('[','').replace(']','') +
            ')\n')

        for string in function_strings:
            script.write(string)

        script.close()

    def _map_to_old_names(self, node, cluster_results):
        results = []
        ports = node.getOutputPorts()
        for port in ports:
            unique = self._output_mapping[node.getName()][port.getName()]            
            for res in cluster_results[:]:
                if res.getName() == unique:
                    res.setName(port.getName())
                    results.append(res)
                    cluster_results.remove(res)

        return results

    def clean(self, path=None):
        self._plugin.clean()
        if path:
            if os.path.exists(path):
                try:
                    os.unlink(path)
                except OSError:
                    pass
                
    def get_plugin(self):
        return self._plugin
    
    def isInterrupted(self):
        for node in self._function_nodes:
            if node.isInterrupted():
                return True

        return False
