# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

__all__ = ["update"]

from thirdparty import ElementTree

import sre
import os.path

class DocumentVersion:
    def __init__(self, major=0, minor=0, patch=0):
        self._version = major, minor, patch

    def get_tuple(self):
        """Get version as tuple (major, minor, patch)."""
        return self._version

    def __cmp__(self, other):
        other_major, other_minor, other_patch = other.get_tuple()
        major, minor, patch = self.get_tuple()
        if other_major > major:
            return -1
        elif other_major < major:
            return 1
        elif other_minor > minor:
            return -1
        elif other_minor < minor:
            return 1
        elif other_patch > patch:
            return -1
        elif other_patch < patch:
            return 1
        else:
            return 0

    def from_document(root):
        """Factory for creating DocumentVersion from version tag.

        Version tag is read from root element of ElementTree-parsed
        xml.  A document with xml <root_tag version='0.10'></root_tag>
        will be parsed as version 0.10.0.
        """
        version_string = root.get("version") or ""
        # remove text-comments at end
        version_string = sre.sub(r"(.*\.\d+).*$", r"\1", version_string)
        
        version = map(int, filter(None, version_string.split(".")))
        version = version + [0]*(3-len(version))
        return DocumentVersion(*version)

    from_document = staticmethod(from_document)

    def __hash__(self):
        major, minor, patch = self.get_tuple()
        return major < 10 | minor < 5 | patch

    def to_string(self):
        return ".".join(map(str, self.get_tuple()))
            

class DocumentUpdater:
    def __init__(self):
        self._conversions = {}
        
    def add_conversion(self, (major, minor, patch), converters):
        """Add converter for version major.minor.patch."""
        self._conversions[DocumentVersion(major, minor, patch)] = converters

    def update(self, document_root):
        versions = self._conversions.keys()
        versions.sort()
        for version in versions:
            if DocumentVersion.from_document(document_root) < version:
                for converter in self._conversions[version]:
                    converter(document_root)
                document_root.set("version", version.to_string())

# converters

def combine_filewriter_and_outputnode(tree):
    """Rename pre-0.16 FileWriterNode to OutputNode."""
    for node in tree.findall("node"):
        class_element = node.find("class")
        if class_element.text.endswith("OutputNode"):
            # change name
            class_element.text = sre.sub(r"\bOutputNode$", "FileWriterNode",
                                         class_element.text)
            # replace parameters with  variable settings
            for parameter in node.find("input").findall("parameter"):
                if parameter.find("name").text == "Filetype":
                    filetype = parameter.find("value").text
                elif parameter.find("name").text == "Filename":
                    filename = parameter.find("value").text
                node.find("input").remove(parameter)
                
            variables = node.find("variables")
            writer_element = ElementTree.SubElement(variables, "writer")
            writer_element.text = filetype
            filename_element = ElementTree.SubElement(variables, "filename")
            filename_element.text = filename

            # create multiport
            multiport = ElementTree.SubElement(node.find("input"), "multiport")
            ElementTree.SubElement(multiport, "name").text = "input"

            names = {}
            # for each existing port, create connection to multiport
            # instead
            for port in node.find("input").findall("port"):
                connected = ElementTree.SubElement(multiport, "connected_port")
                connected.text = port.findtext("outputport")
                names[port.findtext("outputport")] = port.findtext("name")
                node.find("input").remove(port)

            names_element = ElementTree.SubElement(variables, "names")
            names_element.text = repr(names)

def rename_filewriter_to_outputnode(tree):
    for node in tree.findall("node"):
        node.find("class").text = sre.sub(r"\FileWriterNode$", "OutputNode",
                                          node.findtext("class"))

def convert_old_input_node_to_file_writer_node(tree):
    for node in tree.findall("node"):
        if not sre.search("InputNode", node.findtext("class")):
            continue
        
        params = node.find("input").findall("parameter")
        for param in params:
            if param.find("name").text == "Filename":
                param.find("name").text = "filename"
                param.find("type").text = "text"
            elif param.find("name").text == "FileReader":
                param.find("name").text = "fileformat"

def convert_multiple_readers_to_one(tree):
    """Convert pre-0.11 input nodes to one input node."""
    for node in tree.findall("node"):
        class_element = node.find("class")
        if class_element.text.startswith("FileHandlers.Input."):
            reader_part = class_element.text.split(".")[-1]
            reader = sre.match("(.*)FileReader", reader_part).group(1)

            class_element.text = "FileHandlers.InputNode"
            
            parameter = ElementTree.SubElement(node.find("input"), "parameter")
            name_element = ElementTree.SubElement(parameter, "name")
            name_element.text = "FileReader"
            value_element = ElementTree.SubElement(parameter, "value")
            value_element.text = reader
            type_element = ElementTree.SubElement(parameter, "type")
            type_element.text = "text"
            
def convert_multiple_writers_to_one(tree):
    """Convert pre-0.11 output nodes to one output node."""
    for node in tree.findall("node"):
        class_element = node.find("class")
        if class_element.text.startswith("FileHandlers.Output."):
            reader_part = class_element.text.split(".")[-1]
            reader = sre.match("(.*)FileWriter", reader_part).group(1)

            class_element.text = "FileHandlers.OutputNode"

            for param_name in ("Filetype", "FileWriter"):
                parameter = ElementTree.SubElement(node.find("input"), "parameter")
                name_element = ElementTree.SubElement(parameter, "name")
                name_element.text = param_name
                value_element = ElementTree.SubElement(parameter, "value")
                value_element.text = reader
                type_element = ElementTree.SubElement(parameter, "type")
                type_element.text = "text"

def update_struct_dissect_to_new_xml(tree):
    """Convert pre 0.16 struct dissect node to use proper xml instead of portname magic."""
    for node in tree.findall("node"):
        class_element = node.find("class")
        if class_element.text.endswith("StructDissectionNode"):
            output = node.find('output')
            if output:
                ports = output.findall('port')
                if ports:
                    dissect_output = ElementTree.SubElement(node, 'dissect_output')

                    for port in ports:
                        name = port.find('name').text
                        struct_sub_items_parts = name.split('-')

                        if len(struct_sub_items_parts) == 1:
                            
                            ElementTree.SubElement(dissect_output,
                                                   'struct_item',
                                                   port_name = name,
                                                   struct = struct_sub_items_parts,
                                                   sub_elements = '')
                        else:
                            ElementTree.SubElement(dissect_output,
                                                   'struct_item',
                                                   port_name = name,
                                                   struct = struct_sub_items_parts[0],
                                                   sub_elements = ','.join(struct_sub_items_parts[1:]))

def convert_rdissect_to_struct_dissect(tree):
    for node in tree.findall("node"):
        class_element = node.find("class")
        if class_element.text.endswith("RDissectionNode"):
            class_element.text = "Tools.StructDissectionNode"
            name = node.find('name')
            unique_id = sre.match("RDissectionNode(.*)", name.text).group(1)
            name.text = "StructDissectionNode" + unique_id
            
            dissect_output = ElementTree.SubElement(node, 'dissect_output')

            rdissection = node.find('rdissection')
            if rdissection:
                items = rdissection.findall('item')
                if items:
                    for item in items:
                        name = item.get('name')
                        item_index = item.get('index').replace('(', '').replace(')','')
                        item_index = item_index.replace(' ', '')
                        ElementTree.SubElement(dissect_output,
                                               'index_item',
                                               port_name = name,
                                               index = item_index)

                node.remove(rdissection)

def rename_structdissect_to_dissect(tree):
    for node in tree.findall('node'):
        class_element = node.find('class')
        if class_element.text.endswith("StructDissectionNode"):
            class_element.text = "Tools.DissectionNode"
            name = node.find('name')
            unique_id = sre.match("StructDissectionNode(.*)", name.text).group(1)
            name.text = "DissectionNode" + unique_id

def convert_plot_data_source(tree):
    output_ports = {}
    nodes = tree.findall('node')

    for node in nodes or []:
        name = node.find('name').text
        inputs = node.find('output')
        output_ports[name] = {}
        if inputs:
            for port in inputs.findall('port') or []:
                output_ports[name][port.find('name').text] = port.find('id').text

    for node in nodes or []:
        if node.find('class').text.endswith('Plot2DNode'):
            input_ports = {}
            input_ = node.find('input')
            if input_:
                multiport = input_.find('multiport')
                if multiport:
                    for connected in multiport.findall('connected_port') or []:                        
                        if 'id' in connected.keys():
                            input_ports[connected.attrib['id']] = connected.text
                for input_port in input_.findall('port') or []:
                    name = input_port.find('name')
                    id_ = input_port.find('id')
                    if name != None and id_ != None:
                        input_ports[name.text] = id_.text
                    

            plotnode_config = node.find('plotnode_config')
            if plotnode_config == None:
                return

            for widget_config in plotnode_config.findall('widget_config'):
                config = widget_config.find('plot_config')

                if config == None:
                    continue

                if config.find('plot_type').text in ['scatter_plot',
                                                     'normal_qq_plot',
                                                     'line_plot']:
                    x_data_source = config.find('x_axis_data')
                    y_data_source = config.find('y_axis_data')
                    if len(x_data_source.text.split(':')) == 2:
                        node, dataname = [x.strip() for x in x_data_source.text.split(':')]
                        if output_ports.has_key(node):
                            if output_ports[node].has_key(dataname):
                                id_ = output_ports[node][dataname]
                                for port_id, connected_id in input_ports.items():
                                    if connected_id == id_:
                                        x_data_source.text = port_id
                                        break
                    else:
                        # assume that it is NOT using a multi input port
                        for port_name, connected_id in input_ports.items():
                            if port_name == x_data_source.text:
                                x_data_source.text = connected_id


                    if len(y_data_source.text.split(':')) == 2:
                        node, dataname = [y.strip() for y in y_data_source.text.split(':')]
                        if output_ports.has_key(node):
                            if output_ports[node].has_key(dataname):
                                id_ = output_ports[node][dataname]
                                for port_id, connected_id in input_ports.items():
                                    if connected_id == id_:
                                        y_data_source.text = port_id
                                        break
                    else:
                        # assume that it is NOT using a multi input port
                        for port_name, connected_id in input_ports.items():
                            if port_name == y_data_source.text:
                                y_data_source.text = connected_id


def rename_plotnode_to_plot3dnode(tree):
    for node in tree.findall("node") or []:
        class_element = node.find('class')
        if class_element.text.endswith('.PlotNode:Plot3D'):
            class_element.text = "PlotNodes.Plot3DNode:Plot3D"
                                
def rename_ports_from_textfilereader(tree):
    # finding inputnodes
    inputnodes = [node for node in tree.findall("node") if "File.InputNode" in node.find("class").text]

    for inputnode in inputnodes:
        # finding parameters
        params = inputnode.find('input').findall('parameter')
        fileformat_params = [param for param in params if param.find('name').text == 'fileformat' ]
        filename_params   = [param for param in params if param.find('name').text == 'filename' ]

        # if textfilereader, update portnames to reflect textfilereader naming scheme and update settings
        if fileformat_params and filename_params and fileformat_params[0].find("value").text=='Textfile':
 
            # new variable name! 
            new_var_name = os.path.basename(filename_params[0].find("value").text)
            inputnode.find('output').find("port").find("name").text = new_var_name

            # emulate old settings
            filesettings = ElementTree.fromstring("""        <TextFileSettings automatic="False">
            <remove lines="0"/>
            <headers column="False" row="False"/>
            <separators>
                <row><i>Space</i><i>Tab</i></row>
                <column><i>Newline</i></column>
            </separators>   </TextFileSettings>""")
            inputnode.append(filesettings)    

                                
_zmd_updater = DocumentUpdater()
_zmd_updater.add_conversion((0, 11, 0), [convert_multiple_readers_to_one,
                                         convert_multiple_writers_to_one])
_zmd_updater.add_conversion((0, 16, 0), [combine_filewriter_and_outputnode,
                                         rename_filewriter_to_outputnode,
                                         update_struct_dissect_to_new_xml,
                                         convert_rdissect_to_struct_dissect,
                                         rename_structdissect_to_dissect,
                                         convert_old_input_node_to_file_writer_node])
_zmd_updater.add_conversion((0, 17, 0), [convert_plot_data_source])
_zmd_updater.add_conversion((0, 18, 0), [rename_plotnode_to_plot3dnode, rename_ports_from_textfilereader])

def update(document_root):
    """Update ZMD document to current version."""
    _zmd_updater.update(document_root)
