# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import re 
import Numeric
import os
import tempfile
import types
from sets import Set
from itertools import izip
from cPickle import PicklingError

from Library.Util.rtester import RTester
from Library.Util import lists
from Library.Util.Counter import Counter
from Library.Util import numerical
from Library.Util.rpyclient import rpy_assign, rpy_gc, rpy_silent_eval, rpy_eval, rpy_converted_eval

import copy
pointer_counter = Counter()

class Data:
    """This class is a representation of a data object.  It can be
    passed around in SciCraft, and usually contains metadata.
    """
   
    def __init__(self, name, data=None):
        self.nameFilter = re.compile("^[^\ \n\r]+.*$")
        self.setName(name)

        self.data = data
        self._info = None
        self._col_indices = Set()
        self._row_indices = Set()
        self._col_filter_length = None
        self._row_filter_length = None

    def setData(self, data):
        self.data = data
        self._info = data_info(data)
        self._col_indices = Set()
        self._row_indices = Set()
        self._col_filter_length = None
        self._row_filter_length = None

    def setName(self, name):
        if not self.nameFilter.match(name):
            raise AttributeError, "Invalid name. Must match the following regex: " + str(self.nameFilter.pattern) \
                + "\nBad name: " + str(name)
        self.name = str(name)
        
    def getName(self):
        return self.name
        
    def getData(self):
        return self.data

    def get_data_info(self):
        """Returns the name and type/size of the data object.
        
        Examples: 'X - 6x4', 'MyX - molecule', 
                  'D - Dictionary with keys: a, b'
        """
        return "%s - %s" % (self.name, self.get_data_size_str())

    def get_data_size_str(self):
        """Return type/size of data object, e.g. '6x4'."""
        if not self._info:
            self._info = data_info(self.getData())
        return self._info

    def get_sub_data(self, is_column, index, data = None):
        """Extract vector row or column from matrix."""
        if not data:
            data = self.getData()
        if isinstance(data, dict):
            return data
        
        dim  = lists.get_repr_dim(data)
        if is_column:
            if dim == 2:
                return [row[index] for row in data]
            else:
                return [data[index]]
        else: #is row
            if dim == 2:
                return data[index]
            else:
                if index > 0:
                    raise IndexError, 'Index %d out of range' % index
                return data

    def get_sub_data_with_nan(self, is_column, index):
        data = insert_nan_values(self.getData(),
                                 self._col_indices,
                                 self._row_indices)
        return self.get_sub_data(is_column, index, data)        

    def get_all_data_with_nan(self):
        return insert_nan_values(self.getData(),
                                 self._col_indices,
                                 self._row_indices)        

    def get_data_size(self, is_column):
        """Get number of rows or columns of data."""
        data= self.getData()
        dim = lists.get_repr_dim(data)
        if is_column:
            # retrieve number of cols
            if dim == 2:
                return len(data[0])
            elif dim == 1:
                return len(data)
            else:
                return 1            
        else:
            # retrieve number of rows
            if dim == 2:
                return len(data)
            else:
                return 1

    def get_col_indices(self):
        """Returns list with indices of removed columns."""
        return list(self._col_indices)

    def get_row_indices(self):
        """Returns list with indices of removed rows."""
        return list(self._row_indices)

    def set_col_indices(self, indices):
        self._col_indices.update(indices)

    def set_row_indices(self, indices):
        self._row_indices.update(indices)

    def get_dim(self):
        """Returns (# of rows, # of cols)"""

        # Numeric.shape returnes (rows, cols), but will
        # return (cols,) if there is only one row
        try:
            dim = Numeric.shape(self.getData())
        except TypeError:
            # Numeric can't handle arrays with both numbers and strings
            return self.get_data_size(False), self.get_data_size(True)
        if not dim:
            return 0, 0
        elif len(dim) == 1:
            # The data is now a vector
            return 0, dim[0]
        else:
            return dim
        

    def get_original_col_length(self):
        cols = self.get_dim()[1]
        return cols + len(self._col_indices)

    def get_original_row_length(self):
        rows  = self.get_dim()[0]
        return rows + len(self._row_indices)

    def _create_index_map(self, length, removed_indices):
        index_map = {}
        skipped = 0
        for i in xrange(length):
            if i not in removed_indices:
                index_map[i] = i - skipped
            else:
                index_map[i] = i
                skipped += 1

        return index_map

    def remove_indices(self, n, indices):
        """Removes rows/cols if data set has len n in one or both directions"""
        rows = self.get_original_row_length()
        columns = self.get_original_col_length()

        # Numeric.array needs to be converted to list for processing and back
        # to an array when finished.
        is_numeric = False
        data = self.getData()
        
        if isinstance(data, Numeric.ArrayType):
            data = data.tolist()
            is_numeric = True

        if columns == n:
            index_map = self._create_index_map(columns, self._col_indices)
            valid_indices = filter(lambda index : index < columns,
                                   Set(indices) - self._col_indices)
            valid_indices.sort()
            
            if rows >= 1:
                for row in data:
                    for i, index in enumerate(valid_indices):
                        del row[index_map[index] - i]
            else:
                for i, index in enumerate(valid_indices):
                    del data[index_map[index] - i]

            self.set_col_indices(valid_indices)
            self._col_filter_length = n
            
        if rows == n:
            index_map = self._create_index_map(rows, self._row_indices)
            valid_indices = filter(lambda index: index < rows,
                                   Set(indices) - Set(self._row_indices))
            valid_indices.sort()

            for i, index in enumerate(valid_indices):
                del data[index_map[index] - i]
                
            self.set_row_indices(valid_indices)
            self._row_filter_length = n

        if is_numeric:
            data = Numeric.array(data)
        
        col_indices = self.get_col_indices()
        row_indices = self.get_row_indices()
        col_filter_length = self._col_filter_length
        row_filter_length = self._row_filter_length
        self.setData(data)
        self._col_indices = Set(col_indices)
        self._row_indices = Set(row_indices)
        self._col_filter_length = col_filter_length
        self._row_filter_length = row_filter_length

    def copy_data(self, data):        

        self.setData(copy.deepcopy(data.getData()))
        self._col_indices = Set(data.get_col_indices())
        self._row_indices = Set(data.get_row_indices())
        self._col_filter_length = data.get_col_filter_length()
        self._row_filter_length = data.get_row_filter_length()

    def has_dimension(self, dim):
        if dim == 0:
            return False
        
        if not isinstance(self.getData(), dict):
            return dim in self.get_dim()

        data = self.getData()
        return self._check_struct_dimension(dim, data)    

    def has_filter_dimension(self, dim):
        """Helper method that 'remembers' removed indices."""
        if not dim:
            return False

        if len(self._col_indices) == len(self._row_indices):
            return self.has_dimension(dim - len(self._col_indices))

        if self._col_indices:
            col_check = self.has_dimension(dim - len(self._col_indices))
        elif not self._col_indices and self._row_indices:
            col_check = self.has_dimension(dim - len(self._row_indices))
        
        if self._row_indices:
            row_check = self.has_dimension(dim - len(self._row_indices))
        elif self._col_indices and not self._row_indices:
            row_check = self.has_dimension(dim - len(self._col_indices))
            
        return col_check or row_check

    def _check_struct_dimension(self, dim, data):
        for value in data.values():
            if isinstance(value, dict):
                if self._check_struct_dimension(dim, value):
                    return True
            else:
                try:
                    if dim in Numeric.shape(value):
                        return True
                except:
                    pass
        return False

    def set_col_filter_length(self, n):
        self._col_filter_length = n

    def get_col_filter_length(self):
        return self._col_filter_length

    def set_row_filter_length(self, n):
        self._row_filter_length = n

    def get_row_filter_length(self):
        return self._row_filter_length
    

class RData(Data):

    def __init__(self, name, data = None):
        Data.__init__(self, name)
        
        self._pointer = "p" + str(pointer_counter.get_next_value())

        self._cached_data = None
        self._force_cache_update = False

        RTester.r_install_test()
        global rpy
        import rpy

        self.setData(data)

    def __del__(self):
        if self._pointer:
            try:
                rpy_silent_eval("if (exists('%s')) { rm(%s) }" % (self._pointer, self._pointer))
            except:
                # we ignore connection-problems etc. at shutdown
                pass

    def __getstate__(self):
        filename = tempfile.mktemp()
        filename = filename.replace("\\", "/")
        rpy_silent_eval('save(file="%s", list="%s")' % (filename, self._pointer))
        data = open(filename).read()
        os.unlink(filename)
        return {'name' : self.getName(), 'data' : data}

    def __setstate__(self, state):
        self.__init__(state['name'])

        filename = tempfile.mktemp()
        filename = filename.replace("\\", "/")
        open(filename,"w").write(state['data'])
        rpy_silent_eval('tempenv = new.env()')
        rpy_silent_eval('load(file="%s", env=tempenv)' % filename)
        key = rpy_eval('ls(tempenv)')
        rpy_silent_eval('assign("%s", tempenv$%s)' % (self._pointer, key))
        rpy_silent_eval("rm(tempenv)")
        rpy_gc()
        
    def get_pointer(self):
        return self._pointer
  
    def getData(self,bypass_cache = False):
        # use cached data if available, not instucted to ignore, and data not marked as changed
        if self._cached_data and not bypass_cache and not self._force_cache_update:
            return self._cached_data

        if self._force_cache_update:
            # forced update no longer needed
            self._force_cache_update = False

        try:
            self._cached_data = rpy_eval(self._pointer)
            return self._cached_data
        except PicklingError, e:
            self._cached_data =  rpy_converted_eval(self._pointer)
            return self._cached_data
        except TypeError, e:
            self._cached_data = rpy_converted_eval(self._pointer)
            return self._cached_data
  
    def setData(self, data):
        # cache obviously no longer valid
        self._force_cache_update = True

        if isinstance(data, list) and data and not type(data[0]) in types.StringTypes:
            data = Numeric.array(data)
        rpy_assign(self._pointer, data)
        self._col_indices = Set()
        self._row_indices = Set()
        self._col_filter_length = None
        self._row_filter_length = None

    def get_data_size_str(self):
        val = rpy_eval("dim(%s)"%self._pointer)
        if val:
            return "R(%s)" % "x".join(map(str, val))
        else:
            return "RObject"

    def get_data_with_dimnames(self, pointer = None):
        """ Returns both the data and the dim names if the data is of type 'numeric'.

        The returned result will be a dict where the key 'data' contains the data and
        'dimnames'.
        """

        if not pointer:
            pointer = self._pointer

        if is_character(pointer):
            return rpy_eval(pointer) 
        elif is_vector(pointer) or is_matrix(pointer):
            dim_names = rpy_eval('dimnames(%s)' % pointer)
            if dim_names:
                if dim_names[0] == None:
                    dim_names[0] = [None]
                if dim_names[1] == None:
                    dim_names[1] = [None]
                return {'data': rpy_eval(pointer), 'dimnames' : dim_names}
            else:
                return rpy_eval(pointer)
        elif is_list(pointer):
            dict_ = {}
            names = rpy_eval('names(%s)' % pointer)
            if names:
                if isinstance(names, list):
                    for name in names:
                        new_pointer = '%s$%s' % (pointer, name)
                        dict_[name] = self.get_data_with_dimnames(new_pointer)
                else:
                    return rpy_eval(pointer)

            return dict_

    def copy_data(self, data):
        # cache no longer valid
        self._force_cache_update = True
            
        data.get_pointer()
        rpy_silent_eval("%s <- %s" %(self._pointer, data.get_pointer()))        
        self.set_col_indices(data.get_col_indices())
        self.set_row_indices(data.get_row_indices())
        self._col_filter_length = data.get_col_filter_length()
        self._row_filter_length = data.get_row_filter_length()

    def _load_library(self):
        """ Tries to load the library required for using this R-data.

        Will work of the data is an R list and it contains the package
        attribute."""

        if not is_list(self._pointer):
            # assume that we can handle this
            return True        

        lib = None
        try:
            lib = rpy_eval('attributes(attributes(%s)$class)$package' % self._pointer)
            if lib:
                rpy_silent_eval("library(%s)" % lib)
                return True
        except rpy.RException:
            # Don't really care why we get this exception
            pass
        return False

    def get_original_row_length(self):

        dim = self.get_dim()

        if dim:
            return dim[0] + len(self._row_indices)
        else:
            return len(self._row_indices)

    def get_original_col_length(self):
        dim = self.get_dim()
        if dim:
            return dim[1] + len(self._col_indices)
        else:
            return len(self._col_indices)

    def remove_indices(self, n, indices):
        # cache no longer valid!
        self._force_cache_update = True

        if not is_list(self._pointer):
            Data.remove_indices(self, n, indices)
            return

        if not self._load_library():
            return
        
        rows = self.get_original_row_length()
        columns = self.get_original_col_length()
        
        if columns == n:
            index_map = self._create_index_map(columns, self._col_indices)
            valid_indices = filter(lambda index : index < columns,
                                   Set(indices) - self._col_indices)
            valid_indices.sort()

            remove_indices = []
            for index in valid_indices:
                remove_indices.append(index_map[index] + 1)

            rpy_assign('remove_indices', remove_indices)
            rpy_silent_eval('%s <- %s[,-remove_indices]' % ((self._pointer,)*2))

            self.set_col_indices(valid_indices)
            self._col_filter_length = n
                
        if rows == n:
            index_map = self._create_index_map(rows, self._row_indices)
            valid_indices = filter(lambda index : index < rows,
                                   Set(indices) - self._row_indices)
            valid_indices.sort()
            
            remove_indices = []
            for index in valid_indices:
                remove_indices.append(index_map[index] + 1)

            rpy_assign('remove_indices', remove_indices)
            rpy_silent_eval('%s <- %s[-remove_indices,]' % ((self._pointer,)*2))
                  
            self.set_row_indices(valid_indices)
            self._row_filter_length = n

        

    def get_dim(self):
        if not is_list(self._pointer):
            return Data.get_dim(self)

        if not self._load_library():
            return (0,0)

        dim = rpy_eval('dim(%s)' % self._pointer)
        if dim:
            return dim
        else:
            return (0,0)

    def has_dimension(self, dim):
        """Checks if this RData object contains something with the dimension 'dim'."""
        if dim in self.get_dim():
            # Should cover matrix
            return True
        
        if is_vector(self._pointer) or is_character(self._pointer) \
           or is_matrix(self._pointer):
            return self._check_vector_and_matrix_dimension(dim, self._pointer)

        if is_list(self._pointer):
            return self._check_list_dimension(dim, self._pointer)
            
        return False

    def _r_operation(self, op, pointer = None):
        # cache not necessarily valid anymore
        self._force_cache_update = True    
        
        if not pointer:
            pointer = self._pointer
            
        return rpy_eval('%s(%s)' % (op, pointer))

    def _check_list_dimension(self, dimension, pointer):
        """ Recursive method for dim checking inside a R-list."""
        names = self._r_operation('names', pointer)
        
        if not names:
            if is_vector(pointer) or is_matrix(pointer):
                return self._check_vector_and_matrix_dimension(dimension, pointer)
            else:
                return False
        
        if isinstance(names, str):
            names = [names]

        for name in names:
            p = '%s$%s' % (pointer, name)
            if is_vector(p) or is_character(p) or is_matrix(p):
                if self._check_vector_and_matrix_dimension(dimension, p):
                    return True
            elif is_list(p):
                return self._check_list_dimension(dimension, p)

        return False

    def _check_vector_and_matrix_dimension(self, dimension, pointer):
        if is_vector(pointer)or is_character(pointer):
            dim = self._r_operation('length', pointer)
            return dimension == dim
        elif is_matrix(pointer):
            dim = self._r_operation('dim', pointer)
            return dimension in dim
                


# is_* methods used to check RData
def is_character(pointer):
    return bool(rpy_eval('is.character(%s)' % pointer))


def is_vector(pointer):
    if is_character(pointer):
        return False
    elif is_list(pointer):
        return False

    return bool(rpy_eval('is.vector(%s)' % pointer)) 

    
def is_matrix(pointer):
    return bool(rpy_eval('is.matrix(%s)' % pointer))


def is_list(pointer):
    return bool(rpy_eval('is.list(%s)' % pointer))

def has_dimnames(pointer):
    return bool(rpy_eval('dimnames(%s)' % pointer))

def data_info(data):
    if isinstance(data, str): # Single string
        return 'String'
    elif isinstance(data, int) or isinstance(data, float):
        return 'Scalar'
    elif isinstance(data, dict):
        return 'Struct'
    elif isinstance(data, list) or isinstance(data, tuple):
        try:
            return data_info(Numeric.array(data))
        except TypeError:
            # Using lists.describe_variable() as backup in
            # case the row length differs through the data set
            return lists.describe_variable(data)
    elif isinstance(data, Numeric.ArrayType):
        if data.typecode() == Numeric.typecodes['Character']:
            shape = Numeric.shape(data)
            return '%dxString' % shape[0]
        elif data.typecode() in Numeric.typecodes['Integer'] or \
                 data.typecode() in Numeric.typecodes['Float']:

            dims = Numeric.shape(data)
            if dims == (0,):
                return '0x0'
            elif len(dims) == 1:
                return '1x%d' % dims[0]
            else:
                return 'x'.join(map(str, dims))
    return 'Unknown'

    
def insert_nan_values(raw_data, col_indices, row_indices):
    # need to copy the data since we dont want to alter the input data
    data = copy.deepcopy(raw_data)
    
    shape = Numeric.shape(data)
    if isinstance(data, Numeric.ArrayType):
        data = data.tolist()

    if len(shape) == 1:
        # column vector
        col_length = shape[0] + len(col_indices)
        row_length = 1
    else:
        col_length = len(data[0]) + len(col_indices)
        row_length = len(data) + len(row_indices)

    is_row_vector = False
    if row_length > 1:
        is_row_vector = isinstance(data[0], list)

    nan = numerical.define_nan()

    for x in range(row_length):
        if x in row_indices:
            if is_row_vector:
                data.insert(x, [nan for y in range(col_length)])
            else:
                data.insert(x, nan)
            continue

        for y in range(col_length):
            if y in col_indices:
                if len(shape) == 2:
                    data[x].insert(y, nan)
                else:
                    data.insert(y, nan)

    return data
