# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.


from Library.Util.runners import OctaveRunner
from Library.Util.NullEventServer import NullEventServer
from Library.Util import diagramanalyser, zmdupdater
from Library.Util.Config import Config
from System.Data.ModuleObject import ModuleObject
from GUI.Canvas.canvascomment import CanvasCommentData
from System.Data.functionnode import FunctionNode
from System.Data.Node import Node
from System.events import StartDiagramEvent, DiagramEvent, NodeEvent
from System.NodeRepository import NodeRepository
from System.events import InfoEvent
from thirdparty import ElementTree

from xml.parsers.expat import ExpatError
from threading import Thread

import os
import string
import time
import sys

# domain exception
class InvalidXMLException(Exception):
    """Error parsing XML."""


# helper function
def convert_relative_path(relative_path, old_super_path, new_super_path):
    """Change path to be relative to new_super_path.

    The converted path, relative to new_super_path should reflect the
    same absolute path as the original relative_path relative to
    old_super_path.
    """
    # we only convert relative paths... weed out any absolutes
    if os.path.isabs(relative_path):
        return relative_path

    old_drive = os.path.splitdrive(old_super_path)[0]
    new_drive = os.path.splitdrive(new_super_path)[0]
    if old_drive != new_drive:
        return os.path.join(old_super_path, relative_path)
    
    shared = []

    old_elements = old_super_path.split(os.path.sep)
    new_elements = new_super_path.split(os.path.sep)
    
    for old, new in zip(old_elements, new_elements):
        if old == new:
            shared.append(old)
        else:
            break

    result = relative_path.split(os.path.sep)

    up = [os.path.pardir] * (len(new_elements) - len(shared))
    down = old_elements[len(shared):]

    return os.path.join(*up + down + result)


class ModuleDiagram(ModuleObject):
    """Holds all information needed about nodes and itself.
        Including methods for parsing itself to xml, building itself from xml,
        add / remove nodes, executing itself.
    """
   

    def __init__(self, nodeRepository=NodeRepository(),
                 eventServer=NullEventServer(), filename=None, cache=None):
        ModuleObject.__init__(self)

        # ASSORTED VARIABLES
        self._canvas_comments = []
        self._node_list = []
        self._repository = nodeRepository
        self._event_server = eventServer
        self._has_crashed = False
        self._filename = None
        self._cache = cache
        
        # Seconds to wait for nodes to finish before exiting MD
        # thread.
        self._max_time_to_wait_for_nodes = 15.0
        
        # INTERNAL VARIABLES
        self._finished_nodes = []
        self._waiting_nodes  = []
        self._running_nodes = []
        self._interrupted  = False

        # keep threads in list to keep from being garbage collected
        self._threads = []
        
        # register errorhandler
        self._event_server.addEventHandler(self._handle_critical_events)

        # Builds diagram from XML in file, if filename is given
        if filename:
            self._buildFromFile(filename)
            self._filename = filename

        # Make sure all variables are set and the diagram itself is "clean"
        self.clean() 

    def from_xml_element(self, root):
        ModuleObject.from_xml_element(self, root)

        # replace id with new, runtime-unique id
        self.set_id(ModuleObject.counter.get_next_value())

        # get canvascoments tag, and handle the tags inside it
        canvas_comments = root.find("canvascomments")
        if canvas_comments:
            for comment_element in canvas_comments.findall("canvascomment"):
                comment_data = CanvasCommentData()
                comment_data.from_xml_element(comment_element)
                self.add_canvas_comment(comment_data)

        # remember connections while parsing so ports can be connected
        # afterwards
        connect_list = {}
        output_list = {}
        input_list = {}
        multi_input_connections = {}

        # get Nodes, iterate through them, handle them
        for node_element in root.findall("node"):
            # get regular Node properties
            node = self.add_node_by_name(node_element.findtext("class", ""),
                                         node_element.findtext("name"))
            node.from_xml_element(node_element)

            outputs = node_element.find("output")
            if outputs:
                for port_element in outputs.findall("port"):
                    port = node.create_output_port(port_element.findtext("name"))
                    port.from_xml_element(port_element)

                    port_id = int(port_element.findtext("id"))
                    output_list[port_id] = port

            inputs = node_element.find("input")
            if inputs:
                for port_element in inputs.findall("port"):
                    port = node.create_input_port(port_element.findtext("name"))
                    port.from_xml_element(port_element)

                    # remember connection
                    output = port_element.findtext("outputport")
                    if output:
                        connect_list[port] = int(output)

                for multiport_element in inputs.findall("multiport"):
                    port = node.create_multi_input_port(multiport_element.findtext("name"))

                    # remember connections
                    connections = []
                    for connected in multiport_element.findall("connected_port"):
                        connections.append((connected.get("id"), int(connected.text)))

                    multi_input_connections[port] = connections

        # connect ports
        for inputport, outputport_id in connect_list.items():
            inputport.connectToDataSource(output_list[outputport_id])

        for port, connections in multi_input_connections.items():
            for port_id, output_port_id in connections:
                emulated_port = port.add_data_source(output_list[output_port_id])
                if port_id:
                    emulated_port.set_id(port_id)

    def start(self, single_node=None):
        """Wrapper method for starting this is object in a new
            thread. Made because Threads does not support restarting.
        """
        self._interrupted = False
        self._has_crashed = False
        runner = Thread(target=self._run, args=(single_node,))
        runner.start()
        # keep list of running threads
        self._threads.append(runner)

    def start_blocking(self, single_node=None):
        """Start module diagram and return only when its finished.

        If single_node is specified, only run that node.
        """
        self.start(single_node)
        while self.getIsRunning():
            time.sleep(0.1)
        return 


    def _handle_critical_events(self, event):
        """Eventhandler, stopping the diagram when critical events occur."""
        criticalEvents = ["EXCEPTION", "PATHNOTSPECIFIED"]
        if event.get_module_diagram_id() != self.getId():
            return
        if event.get_event_class() == "NodeEvent" and \
           event.getType() in criticalEvents:
            self._has_crashed = True
            # Need to remove the crashed node from the running list.
            self.setFinished(event.getNode())    


    def stopMD(self):
        """Method to stop the running Module Diagram"""
        self._interrupted = True
        self._event_server.sendEvent(DiagramEvent(self.id, "HALT", "The Diagram was stopped."))
        for node in self._node_list:
            if node.is_alive():
                node.interrupt()

    
    def validate(self):
        """Return:
                a list consisting of:
                    - List of NodeInputPort which doesn't have their data set
                    - List of nodes which have not defined all parameters
                    - String containing human-readable errormessage
        """

        inputports = []
        paramnodes = []

        for node in self._node_list:
            ports = node.getNonvalidInputPorts()
            for port in ports:
                inputports.append(port)

            if not node.isParametersReady():
                paramnodes.append(node)

        errorMessage = self._parse_validation(inputports, paramnodes)
        return (inputports, paramnodes, errorMessage)
        
    
    def _parse_validation(self, invalidPorts, invalidParameters):
        """Method for parsing validation results into text"""

        errorMessage = "Validation Report"
        if invalidPorts or invalidParameters:
            if invalidPorts:
                errorMessage = errorMessage +\
                    "\nThese inputports have no connections:"
                for port in invalidPorts:
                    errorMessage = errorMessage +\
                        "\n    The port: \'" + port.getVariableName() +\
                        "\' on \'" + port.getOwnerNode().name + "\'"
            
            if invalidParameters:
                errorMessage = errorMessage +\
                    "\nThese nodes have parameters without values:"
                for node in invalidParameters:
                    errorMessage = errorMessage + "\n    The node: \'" +\
                        node.name + "\'"
        else:        
            return "No errors found."
        return errorMessage

    def _run_node(self, node):
        """
        Params:
            - node : Node - the node to be started

        Simple method for setting up and starting a node
        """
        self._running_nodes.append(node)
        runner = Thread(target=node.start_blocking)
        runner.start()
        self._threads.append(runner)

        # wait until node has started
        while not node.is_started():
            time.sleep(0.05)

        event = NodeEvent(self, node, 
            "START", "The node '" + node.getName() + "' started")
        self.sendEvent(event)


    def _run(self, single_node=None):
        """Executes a modulediagram along with all its nodes.

        Requires that run-variables have been setup, so should be executed from
        start or start_blocking.

        If single_node is specified, only run that node.
        """
        start_time = time.time()

        if single_node:
            num_nodes_to_run = 1
        else:
            num_nodes_to_run = len(self.get_nodes())
        
        self._event_server.sendEvent(StartDiagramEvent(self.id,
                                                       "The diagram started",
                                                       num_nodes_to_run))
        
        # Clean all nodes from previous runs (including plot nodes)
        if not single_node:
            for node in self._node_list:
                node.reset()

        self._finished_nodes = []
        self._waiting_nodes = []
        self._running_nodes = []
    
        # finds starting nodes - nodes with no input ports
        if single_node:
            starting_nodes = [single_node]
        else:
            starting_nodes = filter(lambda node : node.isStartNode(),
                                    self._node_list)

        if Config.use_composite and not single_node:
            self._set_runners()
        else:
            self._clean_runners()

        # Starts the nodes that have no inputports
        for node in starting_nodes:
            if not self._interrupted:
                self._run_node(node)

        # go through waiting list and launch all nodes until none are left waiting
        while (self._waiting_nodes or self._has_alive_nodes()) and \
                not self._has_crashed and not self._interrupted:
            for node in self._waiting_nodes[:]:
                if not node.is_alive() and node.isReady() and \
                       not self._interrupted:
                    self._waiting_nodes.remove(node)
                    self._run_node(node)
                    
            time.sleep(0.05)


        # post-run cleanup

        runTime = time.time() - start_time

        event = None

        if self._has_crashed:
            event = DiagramEvent(self.id, "EXCEPTION", "The diagram crashed")
            for node in self._node_list:
                if node.is_alive():
                    if isinstance(node, FunctionNode):
                        node.interrupt()
            self.clean()
        else:
            if not self._interrupted:
                event = DiagramEvent(self.id, "FINISH", "The diagram finished")
                self._event_server.sendEvent(DiagramEvent(self.id, "RUNTIME", "Total runtime: " + str(runTime) + "second(s)\n"))
            else:
                # Keep ModuleDiagram thread alive as long as nodes are running
                while self._has_alive_nodes():
                    time.sleep(0.5)
                event = DiagramEvent(self.id, "FINISH", "The Diagram finished after interruption.")
                
        
        # Make an effort to let running nodes finish.
        timeCounter = self._max_time_to_wait_for_nodes
        period = 0.1
        while self._running_nodes and timeCounter > 0.0:
            time.sleep(period)
            timeCounter -= period
            
        self._event_server.sendEvent(event)

    def wait(self):
        """Return after all running threads have completed."""
        return
        
    def _has_alive_nodes(self):
        """Determine if nodes are still alive and running."""
        for node in self._node_list:
            if node.is_alive():
                return True
        return False

 
    def addNextNodes(self, list):
        """
        Params:
            - list: List - a List of Node'objects which are to be added

        Adds a list of nodes to the waiting to be started queue
        """
        for node in list:
            if self._waiting_nodes.count(node) == 0:
                self._waiting_nodes.append(node)

    def remove_node(self, node):
        """
        Params:
            - node : Node - a Node-object which is about to be removed


        Removes the 'node' from the modulediagram"""

        # node has link to ports and ports to owner; deletion of nodes must be resolved.
            
        self._node_list.remove(node)
                
        inports = node.getInputPorts()
        outports = node.getOutputPorts()

        for port in inports:
            port.disconnectFromDataSource()

        for outport in outports:
            for inport in outport.getConnectedInputPorts():
                inport.disconnectFromDataSource() 

        node.reset()
        return (1, None)


    def add_node_by_name(self, nodestring, name):
        """Creates a node, adds it to the modulediagram and returns it to the
            caller.
            
            Params:
                * nodestring - String :
                    A string identifying  the node to be added

        """
        node = self._repository.createNodeByName(nodestring)
        node.set_module_diagram(self)

        if not name:
            name = node.getName()
            if name.endswith('Node'):
                name = name[:-4]

            count = 1
            for n in self._node_list:
                if n.getName().startswith(name):
                    count += 1
            name = '%s %d' %(name, count)
                    
        node.setName(name)
        node.set_keep_output(Config.enable_keep_output)
        self._node_list.append(node)
                
        return node

    def add_node(self, node):
        """Add a previously created node to the diagram."""

        node.set_module_diagram(self)
        self._node_list.append(node)


    def get_nodes(self):
        """
        Return:
            - List - A list of the nodes contained in this module diagram.
        """

        return self._node_list

    def get_node_by_name(self, name):
        """Returns the nodes with this class"""
        for node in self._node_list:
            if node.getName() == name:
                return node
        

    def getRunningNodes(self):
        return self._running_nodes


    def getWaitingNodes(self):
        return self._waiting_nodes


    def get_finished_nodes(self):
        return self._finished_nodes

    def add_canvas_comment(self, comment):
        """Add a complete CanvasCommentData instance to module diagram."""
        self._canvas_comments.append(comment)

    def create_canvas_comment(self, text, xpos, ypos, font=None,
                              font_colour = None, background_colour=None):
        canvas_comment_data = CanvasCommentData(text, xpos, ypos, font,
                                                font_colour, background_colour)
        self._canvas_comments.append(canvas_comment_data)
        return canvas_comment_data

    def get_canvas_comments(self):
        return self._canvas_comments

    def del_canvas_comment(self, canvas_comment):
        self._canvas_comments.remove(canvas_comment)

    def toXML(self):
        """
        Return:
            - String - an xml representation of the modulediagram and its nodes.

        creates an xml representation and returns it.
        """
        self.update_link_connection_status()
        returnlist = []
        returnlist.append('<modulediagram version="%s">'%Config.version)

        for line in ModuleObject.toXML(self):
            returnlist.append("\t" + line)

        for node in self.get_nodes():
            for line in node.toXML():
                returnlist.append("\t" + line)

        canvas_comments = self.get_canvas_comments()
        if canvas_comments:
            returnlist.append("\t<canvascomments>")        
            for canvas_comment in canvas_comments:
                for line in canvas_comment.to_xml():
                    returnlist.append("\t\t" + line)
            returnlist.append("\t</canvascomments>")

        returnlist.append("</modulediagram>")
        return string.join(returnlist,"\n")

    def save_to_file(self, filename):
        """Save module diagram xml to filename."""
        # modify relative paths in inputnodes to reflect new location
        # of module diagram

        old_names = {}
        old_filename = self.getFilename()
        self.setFilename(filename)
        if old_filename and old_filename != filename:
            old_path, old_file = os.path.split(old_filename)
            new_path, new_file = os.path.split(filename)
            nodes = [node for node in self.get_nodes() if node.getNodeType() == "InputNode"]
            for node in nodes:
                old_names[node] = node.get_filename()
                node.set_filename(convert_relative_path(old_names[node],
                                                        os.path.abspath(old_path),
                                                        os.path.abspath(new_path)),
                                  force_format=node.get_fileformat())
        try:
            open(filename, "w").write(self.toXML())
        except:
            # roll back changes since we couldn't write file
            self.setFilename(old_filename)
            for node, old_name in old_names.items():
                node.set_filename(old_name, force_format=node.get_fileformat())
            raise

    def setFinished(self, node):
        """
        Params:
            - node : Node - the finished node

        Updates the list of finished and running nodes.
        """
        if node in self._running_nodes:
            self._running_nodes.remove(node)
            self._finished_nodes.append(node)

    def from_xml_string(self, xml):
        """Populate diagram based on xml string."""
        # parse diagram
        try:
            root = ElementTree.fromstring(xml)
        except ExpatError, e:
            raise InvalidXMLException, "Error parsing xml: %s" % e
        except SyntaxError, e:
            raise InvalidXMLException, "Error parsing xml: %s" % e

        # update document to structure in latest version of SciCraft
        zmdupdater.update(root)

        self.from_xml_element(root)

    def _buildFromFile(self, filename):
        """
        Params:
            - xmlstring : String - the xml which will build this ModuleDiagram

        Builds the ModuleDiagram and nodes in the xml.
        """
        if not os.path.exists(filename):
            return
        
        try:
            xmlstring = open(filename).read()
        except Exception, e:
            raise IOError, str("Failed to open: " + os.path.abspath(filename))

        self.from_xml_string(xmlstring)

        if self._cache:
            base_name = os.path.splitext(os.path.split(filename)[-1])[0]
            node_data = self._cache.get_entry(base_name)
            cached_nodes = filter(Node.get_keep_output, self.get_nodes())

            for node in cached_nodes:
                for port in node.getOutputPorts():
                    try:
                        data = node_data[node.getName()][port.getName()]
                        port.setData(data)
                    except KeyError:
                        pass


    def getConnections(self):
        """Get all OutputPort->InputPort connections for a ModuleDiagram.

        The connections returned include connections to emulated input
        ports.

        returns:
            list of tuples (outputPort, inputPort)
        """
        connections = []
        for node in self.get_nodes():
            inputPorts = node.get_all_input_ports()
            for inputPort in inputPorts:
                source = inputPort.getDataSource()
                if source:
                    connections.append((source, inputPort))

        return connections


    def add_node_link(self, start_node_id, end_node_id):
        """Add GUI link between two nodes.
        
        A GUI link is a link drawn between two nodes where there are no
        ports connected to eachother, also called a connectionless link.
        Information about these must be stored to be able to redraw them
        when generating a md from XML. The concept is purely graphical!
        """
        link_str = "@%s-%s" % (start_node_id, end_node_id)
        links = self.getVariable("connectionless_links")
        if links:
            if links.find(link_str) == -1: # do not add more than once
                self.setVariable("connectionless_links", links + link_str)
        else:
            self.setVariable("connectionless_links", link_str)


    def remove_node_link(self, start_node_id, end_node_id):
        """Remove GUI link between two nodes."""
        link_str = "@%s-%s" % (start_node_id, end_node_id)
        links = self.getVariable("connectionless_links")
        if links:
            self.setVariable("connectionless_links",
                links.replace(link_str, ""))


    def update_link_connection_status(self):
        """Remove links with port connections from the connectionless links list."""
        node_connections = {}
        for (sport, eport) in self.getConnections():
            start_node_id = sport.getOwnerNode().getId() 
            end_node_id = eport.getOwnerNode().getId()
            node_connections[start_node_id, end_node_id] = "ignore"
        links = self._parse_connectionless_links_variable()
        for (start_id, end_id) in links:
            if node_connections.has_key((start_id, end_id)):
                self.remove_node_link(start_id, end_id)
    
    
    def get_connectionless_links(self):
        """Get all Node->Node links without port connections."""
        self.update_link_connection_status()
        return self._parse_connectionless_links_variable();


    def on_close(self):
        """Allow module diagram to do some cleanup."""
        if not self._cache: return
        if not self.getFilename(): return
        
        node_data = {}
        
        for node in self.get_nodes():
            if node.get_keep_output():
                port_data = {}
                for port in node.getOutputPorts():
                    data = port.getData()
                    if data:
                        port_data[port.getName()] = data

                if port_data:
                    node_data[node.getName()] = port_data

        if node_data:
            sys.stdout.flush()
            base_name = os.path.splitext(os.path.split(self.getFilename())[-1])[0]
            try:
                self._cache.store_entry(base_name, node_data)
                mb_size = self._cache.get_used_space() / (1024*1024)
                if Config.cachewarnlimit and mb_size >= Config.cachewarnlimit:
                    self.sendEvent(InfoEvent(self.getId(), "INFO",
                                             """Warning: Stored cache size for diagram %s is %s MB.\n\nThe warning limit can be changed in the preferences menu."""% (base_name, mb_size)))
            except IOError, e:
                self.sendEvent(InfoEvent(self.getId(), "INFO",
                                         """Could not save cache for diagram %s:\n\n%s"""% (base_name, e)))

            
    def clean(self):
        self._finished_nodes = []
        self._waiting_nodes  = []
        self._running_nodes = []

    def sendEvent(self, event):
        """Send an event to the eventserver.
            This method is to be used by the nodes/objects in a module 
            diagram, instead of them accessing the event server reference
            directly.
        """
        self._event_server.sendEvent(event)


    def getIsInterrupted(self):
        return self._interrupted

    def _get_running_threads(self):
        """Get all active threads spawned by this object."""
        self._threads = filter(Thread.isAlive, self._threads)
        return self._threads

    def getIsRunning(self):
        return True and self._get_running_threads()

    def get_has_crashed(self):
        return self._has_crashed

    def setFilename(self, filename):
        """Set filename this module diagram is based on."""
        self._filename = filename


    def getFilename(self):
        """Get filename this module diagram is based on."""
        return self._filename

    def _set_runners(self):
        octave_groups = diagramanalyser.find_octave_groups(self.get_nodes())
        for group in octave_groups:
            runner = OctaveRunner(len(group))
            for node in group:
                node.set_runner(runner, group)
                self._waiting_nodes.append(node)

    def _clean_runners(self):
        """Used to remove runners on function nodes"""
        octave_groups = diagramanalyser.find_octave_groups(self.get_nodes())
        for group in octave_groups:
            for node in group:
                node.remove_runner()


    def _valid_node_id(self, id):
        """Check if module diagram has a node with the given id."""
        for node in self._node_list:
            if node.getId() == id:
                return True
        return False
        
    
    def _parse_connectionless_links_variable(self):
        """Parse variable "connectionless_links" into a list.
        
        This method also removes invalid entries in the variable string.
        Each connectionless link is represented with a tuple:
        (start_node_id, end_node_id). These are returned in a list.
        Excepted string format of the variable: @sid-eid
        Example: "@1-6@6-2@6-8" (3 links)
        """
        # defensive programming here! (contents come from XML-file)
        separator = "@"
        links = self.getVariable("connectionless_links")
        if not links or type(links) != type(""):
            return []
        validated_string = ""
        node_id_pairs = []
        # parse variable: @start-end
        links = links.split(separator)[1:] # remove first ""
        for link in links:
            ids = link.split("-")
            if len(ids) != 2: continue
            if not self._valid_node_id(ids[0]) or\
               not self._valid_node_id(ids[1]): continue
            node_id_pairs.append((ids[0], ids[1]))
            validated_string += separator + link
        self.setVariable("connectionless_links", validated_string)
        return node_id_pairs
        

