# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from xml.dom import minidom
from StringIO import StringIO

from System import Tools
from exceptions import Exception
from Library.Util.Counter import Counter
from thirdparty.ElementTree import ElementTree

class ModuleObject:
    """Named object with unique ID within a ModuleDiagram."""

    counter = Counter()

    def __init__(self):
        self.name = ""
        self.description = ""
        self.id = str(ModuleObject.counter.get_next_value())

        # list of extra text variables which should be stored in the
        # xml
        self.variables = {} 
        # xml which will be added
        self.customXML = ""
        # object tree representing the xml. can be traversed using its
        # __dict__. Provided for convenience
        self.customLambda = None

    def from_xml_element(self, root):
        """Set information in ModuleObject based on ElementTree xml element."""
        if root.find("name") is not None:
            self.setName(root.find("name").text)
        if root.find("id") is not None:
            self.set_id(root.find("id").text)
        if root.find("description") is not None:
            self.setDescription(root.find("description").text or "")

        if root.find("variables"):
            for variable in root.find("variables"):
                self.setVariable(variable.tag, variable.text)

        if root.find("CustomXML"):
            text = StringIO()
            child, = root.find("CustomXML").getchildren()
            tree = ElementTree(child)
            tree.write(text)
            self.setCustomXML(text.getvalue())

    def clean(self):
        pass

    def toXML(self):
        returnlist = []
        returnlist.append("<name>" + str(self.getName()) + "</name>")
        returnlist.append("<id>" + str(self.getId()) + "</id>")
        returnlist.append("<description>" + str(self.getDescription()) +"</description>")
        if self.variables:
            returnlist.append("<variables>")
            for key in self.variables:
                returnlist.append("\t<%s>%s</%s>" % (key,self.variables[key], key))
            returnlist.append("</variables>")
        returnlist.append("<CustomXML>")
        returnlist.append(self.customXML)
        returnlist.append("</CustomXML>")
        
        return returnlist

    # ------------------ SET / GET Methods -------------------------
    
    def setName(self, name):
        self.name = str(name)
    
    def getName(self):
        return str(self.name)
    
    def setDescription(self, description):
        self.description = str(description)
    
    def getDescription(self):
        return self.description
    
    def getId(self):
        return self.id

    def set_id(self, new_id):
        ModuleObject.counter.ensure_higher_than(int(new_id))
        self.id = str(new_id)
    
    def setVariable(self,name,value):
        self.variables[str(name)] = str(value)

    def getVariables(self):
        return self.variables

    def getVariable(self,name):
        if self.variables.has_key(str(name)):
            return self.variables[str(name)]
        return False    
    
    def setCustomXML(self,text):
        try:
            dom = minidom.parseString(str(text))
        except Exception:
            return False    
        self.customXML = str(text) 
        self.customLambda = Tools.xml2lambda(dom)
        return True
    
    def getCustomXMLasLambda(self):
        return self.customLambda
