# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.


from System.Data.ModuleObject import ModuleObject

class IllegalValueError(Exception):
    pass



class Parameter(ModuleObject):
    """Class for a parameter of a node. These are set by the user before executing a module diagram,
    and represents essential information needed by the node in order to execute"""

        
   
 
    def __init__(self):
        """
        Init function that creates Parmeter-object"""
        ModuleObject.__init__(self)

        self.type = "Text"   # supported types are Integer / Float / Text / Fileload / Filesave . default is "Text"
        self.validTypes = ["integer", "float", "text", "directory", "fileload", "filesave", "hidden"]

        self.valueSet = 0

        self.restrictions = {
            "max"       : None,
            "min"       : None,
            "values"    : None,
            "default"   : ""  ,
            "step"      : None }

            # max       Maximum allowed value. Maximum allowed length strength for Text. Ignored for File-operations.
            # min       Same as max except the obvious difference
            # values    If given, restrict the allowed values of this parameter to these choices. This option overrides all others. 
            # default   The default value for this parameter
            # step      If given, for Integer/Floats the value must be a multiplum of this value. Used for step-boxes.

        self._value = None

        self.isNamed = 0

        self.PLUGINVariables = None
        # NOT PROPERLY USED
        self.typeVariables = {}

    def from_xml_element(self, root):
        """Set information in Parameter based on ElementTree xml element."""
        ModuleObject.from_xml_element(self, root)
        if root.find("isNamed") is not None:
            self.setIsNamed(bool(int(root.find("isNamed").text)))
        self.setType(root.findtext("type", ""))

        if root.find("restrictions"):
            element = root.find("restrictions")
            if element.find("values"):
                values = map(lambda value_element : value_element.text,
                             element.find("values").findall("value"))
            else:
                values = None
            
            self.setRestrictions(element.findtext("min"),
                                 element.findtext("max"),
                                 values,
                                 element.findtext("step"),
                                 element.findtext("default"))
        if root.find("value") != None and root.find("value").text:
            self.setValue(root.find("value").text)

    def setRestrictions(self, min = None, max = None, values = None, step = None, default = ""):
        if max:
            try:
                self.restrictions['max'] = int(max)
            except ValueError:
                pass
    
        if min:
            try: 
                self.restrictions['min'] = int(min)
            except ValueError:
                pass
        
        if step:
            try:
                self.restrictions['step'] = int(step)
            except ValueError:
                pass
        if default:
            try:
                self.restrictions['default'] = self.convertParamValue(default)
                self.valueSet = True
            except:
                pass

        # Need this since values will be a list with strings when read from xml.
        if values:
            convert = str
            if self.type == "integer":
                convert = int
            elif self.type == 'float':
                convert = float
            values = map(convert, values)
        self.restrictions['values'] = values


    def getRestrictions(self):
        return self.restrictions 

    def convertParamValue(self, value):
        """Method that will try to convert a value to the correct data type.
            Will not do anything on values for file handlers.
        """
        if self.type in ('integer', 'float') and not value:
            value = 0

        if self.type == "integer":
            return int(value)
        elif self.type == "float":
            return float(value)
        elif self.type == "text":
            return str(value)
        else:
            return value
        
        
    def checkRestrictions(self, value):
        """Method that will check if the value is inside the given 
            restrictions. Will return a tuple of two: (valid, errormsg)
        """
        if self.restrictions["values"] and value not in self.restrictions["values"]:
            return (False, "Value not in list of allowed values.")
        
        if self.type.lower() in ("integer", "float"):
            # check min/max values
            if self.restrictions['max']:
                if value > self.restrictions['max']:
                    return (False, "Value exceeds maximum allowed value " +\
                        "(%s > %s)." % (value, self.restrictions['max']))
            if self.restrictions['min']:
                if value < self.restrictions['min']:
                    return (False, "Value is smaller than minimum allowed " +\
                    "value (%s < %s)." % (value, self.restrictions['min']))
        elif self.type.lower() == "text":
            if self.restrictions['max']:
                if len(value) > self.restrictions['max']:
                    return (False, "Text string is too long (%s > %s)." %\
                        (value, self.restrictions['max']))
            if self.restrictions['min']:
                if len(value) < self.restrictions['min']:
                    return (False, "Text string is too short (%s < %s)." %\
                        (value, self.restrictions['min']))                
        
        return (True, "")


    def isValueSet(self):
        """Returns 'True' if parameter value has been set"""
        return self.valueSet


    def setValue(self, value):
        """
        Params:
               - value : Object - The value the parameter will get as its value
        """
        try:
            self._value = self.convertParamValue(value)
        except:
            raise IllegalValueError, \
                  "Illegal value for %s, type should be %s" % (self.name, self.type)
        
        (valid, errormsg) = self.checkRestrictions(self._value)
        if valid:
            self.valueSet = True
        else:
            raise IllegalValueError, "Value for parameter '" + str(self.name) +\
                  "' did not pass the restriction test: " + errormsg

    
    def getValue(self):
        """
        Get the value of this parameter.

        If this parameter has a value set, it is returned. If it has
        not a value set, but a default value, the default value is
        returned. If it does not have a default value either, but has
        restrictions on what values may be set, the first of these is
        returned. If that does not apply either, in case of an integer
        or a float, the minimum value or 0 is returned, and in case of
        a string parameter, an empty string is returned.
        
        Return:
               - Object - The value of the parameter
        """
        if self.isValueSet() and self._value is not None:
            return self._value
        if self.restrictions["default"]:
            return self.restrictions["default"]
        if self.restrictions["values"]:
            return self.restrictions["values"][0]
        if self.type in ["text", "directory", "fileload", "filesave"]:
            return ""
        if self.restrictions["min"]:
            return self.restrictions["min"]
        if self.type == "integer":
            return 0
        if self.type == "float":
            return 0.0
        return ""

    def toXML(self):
        """
        Return:
               - List[String] - List of lines containg XML instance of this instance
        Makes XML-code of the Parameter Object       """
        def toString(x):
            if x or x==0:
                return str(x)
            else:
                return ""       
 
        returnlist = []
        returnlist.append("<parameter>")
        for line in ModuleObject.toXML(self):
            returnlist.append("\t"+line)
        returnlist.append("\t<isNamed>" + toString(self.getIsNamed()) + "</isNamed>")
        returnlist.append("\t<value>"+ toString(self.getValue()) + "</value>")
        returnlist.append("\t<type>" + toString(self.getType()) + "</type>")
        returnlist.append("\t<restrictions>")
        returnlist.append("\t\t<max>" + toString(self.restrictions["max"]) + "</max>")
        returnlist.append("\t\t<min>" + toString(self.restrictions["min"]) + "</min>")
        if self.restrictions['values']:
            returnlist.append("\t\t<values>")
            for value in self.restrictions['values']:
                returnlist.append("\t\t\t<value>" + toString(value) + "</value>")
            returnlist.append("\t\t</values>")        
        returnlist.append("\t\t<step>" + toString(self.restrictions["step"]) + "</step>")
        returnlist.append("\t\t<default>" + toString(self.restrictions["default"]) + "</default>")
        returnlist.append("\t</restrictions>")
        returnlist.append("</parameter>")
        return returnlist


    def setType(self, type):
        if type.lower() in self.validTypes:
            self.type = type.lower()
        else:
            self.type = "Unknown"

    
    def getType(self):
        return self.type

    
    def setIsNamed(self, isNamed):
        if isNamed == '0' or not isNamed:
            self.isNamed = 0
        else:
            self.isNamed = 1

    
    def getIsNamed(self):
        return self.isNamed

# DEPRECATED!!!
    def getVariableName(self):
        return self.name


# DEPRECATED!!!!
    def setLabel(self,label):
        self.description = label

        
