# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from System.Data.ModuleObject import ModuleObject
from System.Data.Data import Data

# domain exceptions
class NoDataToCollectError(Exception):
    """No data was available when reading from output port."""

class MissingConnectionError(Exception):
    """Tried doing an operation on the non-existing connection to this port."""


class NodePort(ModuleObject):
    """
    The NodePort class contain all methods and variables common to input-
    and output ports. A Node object has a number of NodePorts which work as
    the gateways for in-coming and out-going data. A NodePort is responible for one
    Data object."""

    def __init__(self, ownerNode, name="", datatype="", description=""):
        ModuleObject.__init__(self)

        self.name = str(name)
        self.data = None
        self.datatype = str(datatype)
        self.description = description
        self.ownerNode = ownerNode
        self.datainfo = "" #Information about dimensions.

    def from_xml_element(self, root):
        ModuleObject.from_xml_element(self, root)
        self.setDataType(root.findtext("type", ""))

    def getOwnerNode(self):
        return self.ownerNode

    def setOwnerNode(self, ownerNode):
        self.ownerNode = ownerNode

    def getVariableName(self): # deprecated
        """
        Return:
            - String - the variable name of this port"""

        return self.name

    def getData(self):
        """
        Return:
            - Data - the data object on this port""" 

        return self.data

    def setData(self, data):
        """
        Params:
            - data : Data - the Data object that should be put on this port"""
        from System.Data import Data
        import types
        self.data = data
        if self.data:
            self.datainfo = self.data.get_data_info()


    def get_data_info(self):
        return self.datainfo

        
    def getDataType(self):
        """
        Return:
            - String - the data type of this port"""

        return self.datatype

    def setDataType(self, type):
        self.datatype = type

    def clean(self):
        self.data = None

    def toXML(self):
        return ModuleObject.toXML(self)

    def remove(self):
        self.getOwnerNode().remove_port(self)


class NodeInputPort(NodePort):
    """
    The NodeInputPort handles a Data object going in to the Node containing this
    NodePort. The data stays on this port until the node wants it --- and gets it."""
    
    def __init__(self, owner=None, name="", datatype="", datadescription="",
                 optional=False):
        """Create a now port owned by node owner."""
        NodePort.__init__(self,owner,name,datatype,datadescription)

        self.nodePort = None
        self.isnamed = 0
        self._optional = optional

    def getData(self):
        if self.isOptional():
            if not self.getDataSource():
                return Data(self.getName()+"_data")
        return NodePort.getData(self)


    def from_xml_element(self, root):
        NodePort.from_xml_element(self, root)
        if root.find("isOptional") is not None:
            self.setOptional(root.find("isOptional").text == "True")
        if root.find("isNamed") is not None:
            self.setIsNamed(root.find("isNamed").text == "1")

    def collect_data(self):
        """Get data from connected output port."""
        source = self.getDataSource()

        if not source and self.isOptional():
            return
        elif not source:
            raise MissingConnectionError, "Missing connection to port %s." % self.getName()

        data = self.getDataSource().getData()
        if data:
            self.setData(data)
        else:
            raise NoDataToCollectError, "Data not available in port " + \
                  self.getDataSource().getName()

    def connectToDataSource(self, outputPort):
        """
        Params:
               - outputPort : NodeOutputPort - the NodeOutputPort that this
               input port gets its data from

        Exception:
               - TypeError - raised if the ports to be connected do not
               have compatible types

        Creates a reference to the NodeOuputPort that sends its data to this port,
        and at the same time a reference from the NodeOutputPort to this port"""

        if not outputPort:
            raise TypeError, 'NodeOuputPort received was \'None\'!' 

        if self.nodePort:
                self.nodePort.removeConnectionToInputPort(self)

        self.nodePort = outputPort

        outputPort.addConnectionToInputPort(self)

    def disconnectFromDataSource(self):
        """Removes the connection to connected NodeOutputPort."""
        if self.nodePort:
            self.nodePort.removeConnectionToInputPort(self)
       
        self.nodePort = None        

    def getDataSource(self):
        """Get connected NodeOutputPort."""
        return self.nodePort
    
    def toXML(self):
        returnlist = []
        returnlist.append("<port>")
        for line in NodePort.toXML(self):
            returnlist.append("\t" + line)            

        refnodeport = self.getDataSource()
        
        refnodeportid = ""
        if refnodeport:
            refnodeportid = str(refnodeport.getId())

        returnlist.append("\t<isNamed>" + str(self.getIsNamed()) + "</isNamed>")
        returnlist.append("\t<isOptional>" + str(self.isOptional()) + "</isOptional>")
        returnlist.append("\t<type>" + str(self.getDataType()) + "</type>")    
        returnlist.append("\t<outputport>" + refnodeportid + "</outputport>")
        returnlist.append("</port>")
        return returnlist

    def setIsNamed(self,boolean):
        # Developers note: isNamed refers to how a plugin(r and python
        # uses this by now) should invoke the call to external
        # program. If any more of these types of variables appear,
        # they should be encapsulated into another datastructure
        # here
        
        if boolean:
            self.isnamed = 1
        else:
            self.isnamed = 0
        
    def getIsNamed(self):
        return self.isnamed

    def isOptional(self):
        """Check if this port needs to be connected for attached Node to run."""
        return self._optional

    def setOptional(self, value=True):
        """Set if this port needs to be connected for attached Node to run."""
        self._optional = value


class NodeOutputPort(NodePort):
    """
    Represents the NodePorts handling a Data object departing from a Node.
    When a node is finished executing, it places all its result data on the different
    NodeOutputPorts"""
    def __init__(self,owner, name="", datatype="", datadescription=""):
        NodePort.__init__(self,owner,name,datatype, datadescription)
        
        self._connected_ports = []

    def getConnectedInputPorts(self):
        """Return:
               - List[NodeInputPort] - all the InputNodePorts that this port points to
        """
        return self._connected_ports

    def get_connected_nodes(self):
        """Returns a list with nodes connected to this port"""
        ports = []
        for port in self._connected_ports:
            ports.append(port.getOwnerNode())
        return ports

    def addConnectionToInputPort(self, inputPort):
        """Params:
               - inputPort : NodeInputPort - the port to be added to this port
        """
        self._connected_ports.append(inputPort)

    def ship_data(self):
        """Send set data to all connected ports."""
        for port in self.getConnectedInputPorts():
            port.setData(self.getData())

    def getPosition(self):
        """Get position of port on node."""
        # The reason this is so ugly right now is the use of the
        # "position" variable in the xml
        var = self.getVariable("position")
        if not var:
            self.setPosition(-1, 0, 0)
            var = self.getVariable("position")

        # variable is text in format ('nodeInputPortID', [pointIndexOnOutput, pointIndexOnInput])
        # example: "('18', [1, 0])"
        # nodeInputPortID is ignored and outdated

        # convert it to [outputIndex, InputIndex] list
        varlist = var[1:-1].split(",")
        position = map(int,(varlist[1].strip()+varlist[2])[1:-1].split(" "))

        return position
        
    def setPosition(self, inputPortID, startIndex, endIndex):
        """Set position of port, target, and index of point at start
        and end of link."""
        var = "(%d, [%d, %d])"%(inputPortID, startIndex, endIndex)
        self.setVariable("position", var)

    def removeConnectionToInputPort(self, inputPort):
        """Params:
               - inputPort : NodeInputPort - the port to be removed from this port
        """
        if inputPort in self._connected_ports:
            self._connected_ports.remove(inputPort)
    
    def toXML(self):
        returnlist = []
        returnlist.append("<port>")
        returnlist.append("\t<type>" + str(self.getDataType()) + "</type>")
        for line in NodePort.toXML(self):
            returnlist.append("\t"+line)
        returnlist.append("</port>")
        return returnlist


class EmulatedInputPort(NodeInputPort):
    """Acts as inputport but dynamically created/deleted by MultipleInputPort."""
    def __init__(self, multi_input, *args):
        NodeInputPort.__init__(self, *args)
        self._multi_input = multi_input
        
    def disconnectFromDataSource(self):
        self._multi_input.remove_data_source(self.getDataSource())


class MultipleInputPort:
    """Port able to receive data from multiple (including zero) sources."""
    def __init__(self, name, owner_node=None):
        self._name = name
        self._source = []
        self._input_ports = {}
        self._owner = owner_node

    def get_name(self):
        return self._name
        
    def add_data_source(self, source):
        """Add a new connection from a NodeOutputPort."""
        if source.getOwnerNode():
            node_name = source.getOwnerNode().getName()
            # only keep part after last period
            if node_name.count('.'):
                node_name = node_name[node_name.rindex('.')+1:]
            name = "%s: %s" % (node_name, source.getName())
        else:
            name = "Unknown: %s" % source.getName()

        if source in self._input_ports.keys():
            return # we dont want to add the same port twice
        input_port = EmulatedInputPort(self, self._owner, name, "", "")
        input_port.connectToDataSource(source)
        self._source.append(source)
        self._input_ports[source] = input_port
        return input_port

    def get_data_sources(self):
        """Return all connected NodeOutputPorts."""
        return self._source

    def remove_data_source(self, source):
        """Remove connection from NodeOutputPort."""
        source.removeConnectionToInputPort(self._input_ports[source])
        self._source.remove(source)
        del self._input_ports[source]

    def get_emulated_input_ports(self):
        """Return a list of generated ports, one for each connection."""
        return self._input_ports.values()

    def to_xml(self):
        """Generate recreatable XML from internal settings.

        Output is returned as list of strings without newline.

        """
        emulated_ports = self.get_emulated_input_ports()
        output = []
        output.append("<multiport>")
        output.append("\t<name>%s</name>"%self.get_name())
        for emulated_port in emulated_ports:
            output_port = emulated_port.getDataSource()
            output.append('\t<connected_port id="%s">%s</connected_port>' % (emulated_port.getId(),
                                                                             output_port.getId()))
        output.append("</multiport>")
        return output

    def collect_data(self):
        """Collect data from all data sources."""
        for port in self.get_emulated_input_ports():
            port.collect_data()
