# -*- coding: utf-8 -*-
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from thirdparty import ElementTree

class _ZMLReader:
    """ Class that reads and parses ZML files.
    """
    def __init__(self, filename, data=None):
        self._filename = filename
        self._data = None
        self._read(data)
        
    def get_data(self):
        return self._data
    
    def _read(self, data=None):
        try:
            fh = open(self._filename, "r")
        except Exception, e:
            raise Exception, "Failed opening %s for reading" % self._filename

        zmlstring = fh.read()
            
        # generate DOM object
        try:
            root = ElementTree.fromstring(zmlstring)
        except Exception, e:
            raise Exception, "There is an ZML code error in the ZML string:\n%s\n\n%s " % (str(zmlstring), str(e))

        if data:
            self._data = data
        else:
            self._data = FunctionData()
        subobjects = {}

        # build general object structure

        # find output ports
        for port in root.find("output") or []:
            item = OutputItemData(self._data)
            for parameter in port:
                item.set_parameter(parameter.tag, parameter.text)
            self._data.add_outputport(item)

        if root.find("input"):
            # find parameters
            for param in root.find("input").findall("parameter"):
                item = ParameterItemData(self._data)

                restrictions = param.find("restrictions")
                if restrictions:
                    value_list = []
                    if restrictions.find("values"):
                        for value in restrictions.find("values").findall("value"):
                            value_list.append(value.text)
                    item.set_values(value_list)

                    for restriction_element in restrictions:
                        if restriction_element.tag != "values":
                            item.set_restriction(restriction_element.tag,
                                                 restriction_element.text)

                for param_element in param:
                    if param_element.tag != "restrictions":
                        item.set_parameter(param_element.tag, param_element.text)

                self._data.add_parameter(item)

            # find inputports
            for port in root.find("input").findall("port"):
                item = ItemData(self._data)

                for input_port_element in port:
                    item.set_parameter(input_port_element.tag,
                                       input_port_element.text)
                    
                self._data.add_inputport(item)

        # find subobjects
        if root.find("subobjects"):
            for object_element in root.find("subobjects").findall("object"):
                for key, value in object_element.items():
                    subobjects[value] = object_element.text

        for key in subobjects:
            self._data.add_subobject(key, subobjects[key])

        # find general parameters
        for element in root:
            if element.tag not in ("subobjects", "input", "output"):
                self._data.set_parameter(element.tag, element.text)


        
class ZMLWriter:
    """ Class that writes ZML files.
    """
    def __init__(self, filename, data):
        self._filename = filename
        self._data = data
        
    def write(self):
        try:
            fh = open(self._filename, "w")
            fh.write(self._data.to_zml())
            fh.close()
        except Exception, e:
            raise Exception, "Failed during write operation. Filename: %s.\n\nError Message: %s" % (self._filename, str(e))
        
class FunctionData:
    """ This class is the root in the data structure containing information about the functionnode.
    
        ZML data structure:

        class name                    zml tag             comment
        -----------------------------------------------------------------------------------------------------------
        FunctionData                  <functionnode>      Main class.
           |
           |- ItemData                <input>             Not subclassed, nescessary functionality in ItemData
           |- OutputItemData          <output>            Subclassed from ItemData (extra functionailty: subobject)
           |- ParameterItemData       <parameter>         Subclassed from ItemData (extra functionailty: restrictions, values)
    """
    def __init__(self, filename=None):
        self._inputport_list = []
        self._outputport_list = []
        self._parameter_list = []
        self._dict = {}

        if filename:
            reader = _ZMLReader(filename, self)
            
    def add_inputport(self, item = None):
        if item == None:
            item = ItemData(self, "input")
        self._inputport_list.append(item)
        return item

    def add_outputport(self, item = None):
        if item == None:
            item = OutputItemData(self, "output")    
        self._outputport_list.append(item)
        return item
        
    def add_parameter(self, item = None):
        if item == None:
            item = ParameterItemData(self, "parameter")    
        self._parameter_list.append(item)
        return item
        
    def add_subobject(self, portname, value):
        for port in self._outputport_list:
            if port.get_parameter("name") == portname:
                port.set_subobject(value)
        
    def delete_item(self, data):
        if self._inputport_list.count(data) > 0:
            self._inputport_list.remove(data)
        elif self._outputport_list.count(data) > 0:
            self._outputport_list.remove(data)
        elif self._parameter_list.count(data) > 0:
            self._parameter_list.remove(data)
        
    def get_inputport_list(self):
        return self._inputport_list

    def get_outputport_list(self):
        return self._outputport_list
        
    def get_parameter_list(self):
        return self._parameter_list

    def set_parameter(self, name, value):
        self._dict[name] = value
        
    def get_parameter(self, name):
        if self._dict.has_key(name):
            return self._dict[name]
        else:
            return ""
            
    def move_item_up(self, item):
        if self._inputport_list.count(item) > 0:
            orig_index = self._inputport_list.index(item)
            orig_list = self._inputport_list
            
        elif self._outputport_list.count(item) > 0:
            orig_index = self._outputport_list.index(item)
            orig_list = self._outputport_list
            
        elif self._parameter_list.count(item) > 0:
            orig_index = self._parameter_list.index(item)
            orig_list = self._parameter_list
            
        else:
            return
        
        orig_list.remove(item)
        orig_list.insert(orig_index - 1, item)
        
    def move_item_down(self, item):
        if self._inputport_list.count(item) > 0:
            orig_index = self._inputport_list.index(item)
            orig_list = self._inputport_list
            
        elif self._outputport_list.count(item) > 0:
            orig_index = self._outputport_list.index(item)
            orig_list = self._outputport_list
            
        elif self._parameter_list.count(item) > 0:
            orig_index = self._parameter_list.index(item)
            orig_list = self._parameter_list
            
        else:
            return
        
        orig_list.remove(item)               
        orig_list.insert(orig_index + 1, item)        
        
    def to_zml(self):
        zml = "<function>\n"
        for key in self._dict:
            zml = zml + "\t<" + key + ">" + self._dict[key] + "</" + key + ">\n"

        zml = zml + "\t<subobjects>\n"
        for port in self._outputport_list:
            if port.get_subobject() != "":
                zml = zml + "\t\t<object port='" + port.get_parameter("name") + "'>"
                zml = zml + port.get_subobject() + "</object>\n"
        zml = zml + "\t</subobjects>\n"                

        zml = zml + "\t<input>\n"
        for item in self._inputport_list:
            zml = zml + "\t\t<port>\n" + item.to_zml() + "\t\t</port>\n"
        for item in self._parameter_list:
            zml = zml + "\t\t<parameter>\n" + item.to_zml() + "\t\t</parameter>\n"
        zml = zml + "\t</input>\n"

        zml = zml + "\t<output>\n"
        for item in self._outputport_list:
            zml = zml + "\t\t<port>\n" + item.to_zml() + "\t\t</port>\n"
        zml = zml + "\t</output>\n"

        zml = zml + "</function>\n"
        
        return zml

    def save_file(self, path):
        f = open(path, "w")
        f.write(self.to_zml())
               
                
class ItemData:
    """ Part of the data structure containing information about the functionnode.
        This class contains detailed information about eg. inputports, outputports and parameters.
    """
    def __init__(self, parent, title = None):
        self._parent = parent
        self._dict = {}
        if title != None:
            self.set_parameter("name", title)

    def set_parameter(self, name, value):
        self._dict[name] = value
        
    def get_parameter(self, name):
        if self._dict.has_key(name):
            return self._dict[name]
        else:
            return ""

    def to_zml(self):
        zml = ""
        for key in self._dict:
            zml = zml + "\t\t\t<" + key + ">" + self._dict[key] + "</" + key + ">\n"
        return zml            

    def delete(self):
        self._parent.delete_item(self)
        
    def move_up(self):
        self._parent.move_item_up(self)
        
    def move_down(self):
        self._parent.move_item_down(self)
        
        
class OutputItemData(ItemData):
    def __init__(self, parent, title = None):
        ItemData.__init__(self, parent, title)
        self._subobject = ""
        
    def set_subobject(self, value):
        self._subobject = value
        
    def get_subobject(self):
        return self._subobject
        
        
class ParameterItemData(ItemData):
    """ Handles data for parameter objects.
    """
    def __init__(self, parent, title = None):
        ItemData.__init__(self, parent, title)
        self._restrictions = {}
        self._values = []

    def set_restriction(self, name, value):
        self._restrictions[name] = value
        
    def get_restriction(self, name):
        if self._restrictions.has_key(name):
            return self._restrictions[name]
        else:
            return ""
        
    def set_values(self, values):
        self._values = values
               
    def get_values(self):
        return self._values
        
    def to_zml(self):
        zml = ItemData.to_zml(self)
        
        zml = zml + "\t\t\t<restrictions>\n"
        for key, value in self._restrictions.items():
            zml = zml + "\t\t\t\t<" + key + ">" + (value or "") + "</" + key + ">\n"
        
        zml = zml + "\t\t\t\t<values>\n"
        for value in self._values:
            zml = zml + "\t\t\t\t\t<value>" + value + "</value>\n"
        zml = zml + "\t\t\t\t</values>\n"
        
        zml = zml + "\t\t\t</restrictions>\n"
    
        return zml
