# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.
# [TODO]
#   * Optimize data return in read_data(). Start converting to int, if fails, restart.

import os
import re
import string

from System.Data import Data
from System.FileHandlers.Input.FileReader import FileReader

class BASEFileReader(FileReader):

    # Must allow "#" in start of headers and other "meta-data".
    _IDENTIFIER = re.compile("#?BASEfile")
    _ASSAYS_SECTION = re.compile("#?section( |\t)assays")
    _DATA_INDICATOR = re.compile("#?%\r?\n")
    _BLANK_LINE = re.compile("\r?\n")
    _SPOT_AREA = re.compile("#?section( |\t)spots")
 
    # Used for debugging.
    ctag = ">>> BASEFileReader"
    ltag = "    "

    supportedFileExtensions = ["base"]
    variable_name_support = True
    name = "BASEFileReader"
    fileDescription = "BASE"
    description = "BASE data file format. " +\
            "Only specified columns are passed on in the module " +\
            "diagram. The reader supports the standard and custom BASE " +\
            "files in variants serial, merged on reporter and per " +\
            "position. For information about BASE, visit the homepage at " +\
            "http://base.thep.lu.se."

    def __init__(self):
        """Creates a BASEFileReader-object."""
        FileReader.__init__(self)
        self.current_filename = ""
        self.variable_info = {}
        self._cachedVarnames = None

        # Tells wether the BASE file is serial or non-serial/merged.
        self._isSerial = False
        self._fileInfo  = {"filename": "", "datafileSize": "",
            "datafileMTime": ""}
        self._cacheInfo = {"varsInspected": False, "parsedFilename": "",
            "parsedDatafileSize": "", "parsedDatafileMTime": ""}

        # The byte positions for the data sections in the data file;
        # the position for the line containing the data marker.
        self._spot_sections_positions = []
        self._data = {}
    

    def validate_file(self, sourcePath):
        """Make sure the file exists, that we can read it and that is follows
            the BASE format.

            Returns:
                Filehandle to specified data file on success.
                Raises an exception otherwise ('IOError', 'TypeError' 
                or 'Exception').
        """
        mtag = self.ctag + ".validate_file() <<<" + os.linesep + self.ltag
        # Make sure the file exists, and that we can read from it.
        mtag = self.ctag + ".validate_file() <<<" + os.linesep + self.ltag
        # Make sure the file exists, and that we can read from it.
        if not os.access(sourcePath, os.F_OK):
            raise IOError, "The file '" + sourcePath + "' does not exist."
        if not os.access(sourcePath, os.R_OK):
            raise IOError, "Read access to file '" + sourcePath + "' denied."
        try:        
            fd = open(sourcePath, "r")
        except Exception, e:
            raise Exception, "Unexpected error while opening data " +\
                "file:\n'" + sourcePath + "': " + str(e) 
        
        # Make sure we are dealing with a BASE data file.
        # A proper one should have 'BASEfile' on the first line.
        line = fd.readline()
        if not BASEFileReader._IDENTIFIER.match(line):
            fd.close()
            raise TypeError, "The specified file does not follow the BASE " +\
                "format. The start string 'BASEfile' (or '#BASEfile') is " +\
                "not found."

        if sourcePath != self._fileInfo["filename"]:
            self._cacheInfo["varsInspected"] = False
        self._fileInfo["filename"]      = sourcePath
        self._fileInfo["datafileSize"]  = os.stat(sourcePath)[6]
        self._fileInfo["datafileMTime"] = os.stat(sourcePath)[8]

        return fd


    def _is_cached(self, sourcePath):
        """Tells whether the file specified has been processed already or not.
            The method uses internal variable _fileInfo, and must be called
            after _validate_file and/or _get_variable_names.
        """
        if self._cacheInfo["varsInspected"]:
            if self._cacheInfo["parsedFilename"] == sourcePath:
                if self._cacheInfo["parsedDatafileSize"] == \
                    self._fileInfo["datafileSize"]:
                    if self._cacheInfo["parsedDatafileMTime"] == \
                        self._fileInfo["datafileMTime"]:
                        return True
        return False


    def _seek_to(self, filedesc, regexObj):
        """Seeks forward in the file until a line matching the
            regular expression object is found, or end of file is reached.

            Params:
                * filedesc  - FileObject
                    File descriptor for file to search in.
                * regexObj  - SRE_Pattern
                    Regular expression pattern object. Can be created with
                    're.compile()'.

            Return:
                'None' if no match, filedescriptor position 
                *before* matching line otherwise.
        """
        matchPos = filedesc.tell()
        while matchPos < self._fileInfo["datafileSize"]:
            matchPos = filedesc.tell()
            line = filedesc.readline()
            if regexObj.match(line):
                return matchPos
        return None


    def _parse_headers(self, filedesc, stopRegexObj):
        """Read lines from data file and extracts headers.
            Note that re.match() is used with the regex.

            Returns:
                A tuple of two: (stopTopkenPos, dictionary)
                stopTokenPos specifies the line *before* where the 
                stoptoken  was found.
                A dictionary with header names and corresponding values.
                All values are stored in a list. The length of the list
                can be zero (header present, but no value).
        """
        headers = {}
        line = filedesc.readline()
        stopTokenPos = -1 # The pos for the line containing the stop token.
        while not stopRegexObj.match(line):
            columns = line.split("\t")
            columns[-1] = columns[-1].rstrip()
            headers[columns[0]] = columns[1:]
            stopTokenPos = filedesc.tell()
            line = filedesc.readline()
            if filedesc.tell() >= self._fileInfo["datafileSize"]:
                del(headers)
                raise AssertionError, "Stop token (" + stopRegexObj.pattern +\
                    ") does not exist in data file. Unable to separate " +\
                    "header section from data section. Processing aborted."
        return (stopTokenPos, headers)


    def _get_variable_names(self, sourcePath):
        variableNames = []
        filedesc = self.validate_file(sourcePath)
        if not isinstance(filedesc, file):
            return None

        if self._is_cached(sourcePath):
            # File already processed. Use existing results.
            return self._cachedVarnames

        # Read section 'assays'.
        line = self._seek_to(filedesc, BASEFileReader._ASSAYS_SECTION)
        if line is None:
            raise AssertionError, "Section 'assays' not found in BASE " +\
                "datafile '" + sourcePath + "'. Unable to process file content."

        (dontCare, assayHeaders) = self._parse_headers(filedesc, BASEFileReader._DATA_INDICATOR)
        for key in assayHeaders:
            variableNames.append("assays." + key)
            self._data["assays." + key] = assayHeaders[key]
        # Read the datasection in section 'assays'.
        assayInfo = []
        line = filedesc.readline()
        while not BASEFileReader._BLANK_LINE.match(line):
            # Do NOT strip before splitting!
            line = line.split("\t")
            line[-1] = line[-1].rstrip()
            assayInfo.append(line)
            line = filedesc.readline()
        variableNames.append("assays.info")
        self._data["assays.info"] = assayInfo
        # Locate all section 'spots', read headers and store position.
        while True:
            line = ""
            while (not BASEFileReader._SPOT_AREA.match(line) and
                filedesc.tell() < self._fileInfo["datafileSize"]):
                line = filedesc.readline()
            if filedesc.tell() >= self._fileInfo["datafileSize"]:
                break
            (stopPos, spotsHeaders) = self._parse_headers(filedesc, BASEFileReader._DATA_INDICATOR)
            self._spot_sections_positions.append(stopPos)
            # Generate variable names
            # If we have data for a single assay in the section.
            if len(spotsHeaders["assays"]) == 1:
                self._isSerial = True
                columnList = []
                prefix = spotsHeaders["assays"][0] + "."
                # Handle header 'column".
                for column in spotsHeaders["columns"][0:-1]: # Remove 'assayData'.
                    entry = prefix + column
                    variableNames.append(entry)
                    columnList.append(entry)
                # Handle header 'assayFields'.
                for field in spotsHeaders["assayFields"]:
                    entry = prefix + field
                    variableNames.append(entry)
                    columnList.append(entry)
                self._data[prefix + "columns"] = columnList
            # If we have a merged BASEfile (multiple assays in one section).
            else:
                self._isSerial = False
                columnList = []
                for column in spotsHeaders["columns"][0:-1]: # Remove 'assayData'.
                    variableNames.append(column)
                    columnList.append(column)
                for assay in spotsHeaders["assays"]:
                    for field in spotsHeaders["assayFields"]:
                        entry = assay + "." + field
                        variableNames.append(entry)
                        columnList.append(entry)
                self._data["columns"] = columnList

        self._cachedVarnames = variableNames
        self._cacheInfo["varsInspected"] = True
        self._cacheInfo["parsedFilename"] = self._fileInfo["filename"]
        self._cacheInfo["parsedDatafileSize"] = self._fileInfo["datafileSize"]
        self._cacheInfo["parsedDatafileMTime"] = self._fileInfo["datafileMTime"]

        return variableNames


    def read_data(self, sourcePath):
        """Parses a BASE data file and returns the data found in it as
           a list of System.Data.Data objects.

           Params:
             - sourcePath   : string   Path to the data file.
        """
        mtag = ">>> BASEFileReader.read_data() <<<" + os.linesep + self.ltag
        fd = self.validate_file(sourcePath)
        # _get_variable_names() must be run before reading data.
        if not self._cacheInfo["varsInspected"]:
            self._get_variable_names(sourcePath)
        if len(self._spot_sections_positions) < 1:
            # Unable to find any 'section spots' in data file.
            raise Exception, "The BASE data file does not contain any data!"

        # Read assay data.
        # This may be a large amout of data. The current approach might cause an out-of-memory error.
        # [WARNING] The format does not tell what type the data is. We must take a informed guess.
        #           The type must be consistent throughout the column. Problem: how to represent missing values
        #           for non-string types? Currently, None is used for both missing data and invalid data!
        for spot in range(0, len(self._spot_sections_positions)):
            # Seek to the line before 'section spots' and skip data until data indicator is found.
            fd.seek(self._spot_sections_positions[spot])
            while fd.tell() < self._fileInfo["datafileSize"]:
                line = fd.readline()
                if BASEFileReader._DATA_INDICATOR.match(line):
                    break

            if not self._isSerial:
                columns = self._data["columns"]
            else:
                # Fetch colums for current spots section.
                key = self._data["assays.info"][spot][0] + ".columns"
                columns = self._data[key]

            (status, result) = self._fetchDataByColumns(fd, columns)
            if status == 0:
                raise Exception, "Error occured during data read: " + result
        fd.close()

        # Add the info from the 'assays' section.
        for key in self._cachedVarnames:
            if key.find("assays.") > -1:
                result = [Data.Data(key, self._data[key])] + result
        
        return result


    def _fetchDataByColumns(self, fileDescriptor, columnNames):
        """Helper method for read_data().Extracts data from the data file,
           and wraps it into Data objects, one for each column.

           Params:
              - fileDescriptor : file object    Must be at the correct position, e.g.
                                                the next readline must result in a data line.
              - colmnNames     : List           Contains the names of the columns.

           Returns:
             A tuple of length two: (status, data)
             Status is 0 on error, 1 if everything is OK!
             Data is a list of Data-objects if OK, a textual error message otherwise.
        """
        mtag = self.ctag + "._fetchDataByColumns(3) <<<" + os.linesep + self.ltag
        numberOfColumns = len(columnNames)
        result = []
        data = []
        # Initialize list to hold the columns.
        for index in range(0, numberOfColumns):
            data.append([])
        line = fileDescriptor.readline()
        # Read data until end-of-file or an empty line is found.
        while line and line != "\r\n" and line != "\n":
            cells = string.split(line, "\t")
            if len(cells) != numberOfColumns:
                return (0, "Missing data cell!")
            cells[len(cells) -1] = string.rstrip(cells[len(cells) -1])
            for index in range(0, numberOfColumns):
                data[index].append(cells[index])
            line = fileDescriptor.readline()

        # Create Data objects to hold the data. One Data object per column.
        for column in range(0, numberOfColumns):
            # Try to convert to float. Check first, then convert.
            # This is double work amount, but that is better than double datavolume in memory.
            toFloat = 1
            for valueIndex in xrange(0, len(data[column])):
                try:
                    float(data[column][valueIndex])
                except:
                    toFloat = 0
                    break
            if toFloat:
                for valueIndex in xrange(0, len(data[column])):
                    data[column][valueIndex] = float(data[column][valueIndex])
            
            result.append(
            Data.Data(columnNames[column],
                      data[column]))

        return (1, result)


    def get_variable_info(self, filename, variable_name = None, node = None):

        variable_info = {}

        if node:
            filename = self.convert_to_absolute_filepath(filename, node.get_module_diagram().getFilename())
        
        if not os.access(filename, os.R_OK) or os.path.isdir(filename):
            return variable_info
        try:
            data_array = self.read_data(filename)
        except Exception, e:
            return variable_info

        for data in data_array:
            variable_info[data.getName()] = "1x" + str(len(data.getData()))

        if variable_name:
            if variable_info.has_key(variable_name):
                return variable_info[variable_name]
            else:
                return ""
        else: 
            return variable_info

                
