# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import re
import Numeric

from System import Tools
from System.Data.Data import Data
from System.FileHandlers.Input.FileReader import FileReader

class OctaveFileReader(FileReader):
    """This class is able to read data in Octave-format from a file"""
    
    supportedFileExtensions = ["oct"]
    variable_name_support = True
    description = "GNU Octave file format."    
    fileDescription = "Octave"
    name = "OctaveFileReader"

    def __init__(self):
        """Constructor for creating an OctaveFileReader-object"""
        FileReader.__init__(self)

        self._start_reg = re.compile(".*\#.*")    
        self._name_reg = re.compile("# name: (\w*)\n")
        self._type_reg = re.compile("# type: (\w*.\w*)\n")
        self._row_reg = re.compile("# rows: (\w*)")
        self._col_reg = re.compile("# columns: (\w*)")
        self._elem_reg = re.compile("# elements: (\w*)")
        self._length_reg = re.compile("# length: (\w*)")
        self._range_reg = re.compile("# base, limit, increment")
        self._gobal_reg = re.compile("global (\w*)")
        self._ndims_reg = re.compile("# ndims: (\w*)")


    def read_data(self, sourcePath):
        """
           Params:
                  - dataPath : string - path to the file that will be read
           Return:
                  - List[Data] - Returnes a List with Data-objects

           Function which will read a file written with octave ASCII format
        """
 
        try: 
            file_ = open(sourcePath, "r")
            file_text = file_.readlines() 
            file_.close() 
        except IOError, e:
            raise IOError, "File %s not found" % sourcePath

        file_text.reverse()
        data_sets = []

        # loop for handling new variables.
        while file_text:
            name = ""
            type_ = ""
            line = ''

            while file_text and (not self._name_reg.match(line)):
                line = file_text.pop()

            if not file_text:
                break

            name = self._name_reg.match(line).group(1).strip()
            type_ = self._type_reg.match(file_text.pop()).group(1).strip()
            data = self._parse_data_objects(type_, file_text)
            
            data_sets.append(Data(name, data))

        return data_sets

    def _parse_data_objects(self, type_, strings):

        if type_ == "scalar":
            return self._parse_scalar(strings)
        elif type_ == "complex scalar":
            return self._parse_complex_scalar(string)
        elif type_ == "string" or type_ == "string array":
            return self._parse_string(strings)
        elif type_ == "range":
            return self._parse_range(strings)
        elif type_ == "matrix":
            return self._parse_matrix(strings)
        elif type_ == "struct":
            return self._parse_struct(strings)
        else:
            print 'Unknown type:', type_
            return ""
                
    def _parse_complex_scalar(self, input_):
        data = input_.pop()
        real, imag = eval(data)
        return complex(real, imag)

    def _parse_scalar(self, input_):
                
        return Tools.stringToNumber(input_.pop().strip())    

    def _parse_string(self, input_):
        elements = int(self._elem_reg.match(input_.pop()).group(1))
        if elements > 1:
            # We have an array of strings
            string_array = []
            for x in range(elements):
                length = self._length_reg.match(input_.pop())
                string_array.append(input_.pop().strip())
            return Numeric.array(string_array)
        else:
            input_.pop()
            return input_.pop().strip()
        
    def _parse_range(self, input_):
        useless_text = input_.pop()
        base, limit, increment = map(int, input_.pop().split(' '))
        return Numeric.array(range(base, limit+1, increment))

    def _parse_matrix(self, input_):

        dimensions = 0
        dimensions_length = 0

        # create support variables
        if self._ndims_reg.match(input_[-1]):
            # Matrix is multidimensional
            dimensions  = self._ndims_reg.match(input_.pop()).group(1)
            dimensions_length = map(int, input_.pop().strip().split(' '))
        else:
            # Standard type
            dimensions  = 2
            dimensions_length = (int(self._row_reg.match(input_.pop()).group(1)),
                            int(self._col_reg.match(input_.pop()).group(1)))
        items = Numeric.product(dimensions_length)
        
        matrix = []

        # read and parse data (storing as strings)
        finished = False
        while((not finished) and input_):
            matrix += re.split('\s', input_.pop().strip())
            if len(matrix)>=items:

                # probably finished, removing extra empty cells that may have appeared        
                while('' in matrix):
                    matrix.remove('')
                
                # checking if it still looks like we're finished:
                if len(matrix)==items:
                    finished = True
                    
     


        # Need to find out what kind of number this is, default is float
        test_number = matrix[0]
        type_ = float

        try:
            # An int?
            a = int(test_number)
            type_ = int

        except OverflowError:
            # A long?
            a = long(test_number)
            type_ = long

        except ValueError:
             # We have a float
             pass


        # Sanity check
        assert len(matrix) == Numeric.product(dimensions_length)

        # Due to NaN, each object must be converted seperatly
        for x in xrange(len(matrix)):
            try:
                matrix[x] = (type_(matrix[x]))
            except: #Might be NaN or some other number that we can't convert.
                matrix[x] = None

        if dimensions==2: # standard type
            matrix = Numeric.reshape(matrix,dimensions_length)
        else:
            matrix = Numeric.transpose(Numeric.reshape(matrix , list(dimensions_length)[::-1] ))
        
        return matrix

    def _parse_struct(self, input_):

        length = int(self._length_reg.match(input_.pop()).group(1))
        
        result = {}
        for x in xrange(length):
            name = self._name_reg.match(input_.pop()).group(1).strip()
            type_ = self._type_reg.match(input_.pop()).group(1).strip()
            data = self._parse_data_objects(type_, input_)
            result[name] = data

        return result
        

