# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import time
import Numeric

from System.FileHandlers.Output.FileWriter import FileWriter
from System.Exceptions.ZPluginError import ZPluginError
from Library.Util.Config import Config


class OctaveFileWriter(FileWriter):
    """This is class for writing Octave-files to disk.
    It inherits from FileWriter, but writes over all the inherited methods.
    """
    supportedFileExtensions = ["oct"]
    fileDescription = "Octave"
    description = "GNU Octave file format."
    name = "OctaveFileWriter"

    def write_data(self, data_objects, filename):
        """Write data objects to a file in Octave ASCII format.
        
        Params:
            - data : List[Data] - A List with Data objects
            - destinationPath : String - A String with the path to write
                                             the data to

        Exception:
            - IOError - File cannot be written
        """


        vars_ = {}
        for data in data_objects:
            vars_[data.getName()] = data.getData()
            
        self.write_variable_dict(vars_, filename)

        
        
    def _write_data_object(self, name, data, file_):

        self._write_name(name, file_)

        if isinstance(data, list):
            self._write_numeric(Numeric.reshape(data, Numeric.shape(data)),
                                file_)
        elif isinstance(data, str):
            self._write_string(data, file_)
        elif isinstance(data, int) or isinstance(data, float):
            self._write_scalar(data, file_)
        elif isinstance(data, Numeric.ArrayType):
            self._write_numeric(data, file_)
        elif isinstance(data, dict):
            self._write_struct(data, file_)
        else:
            print 'Unrecognicable object. Could not write %s' % file_.name


    def _write_name(self, name, file_):
        file_.write('# name: %s\n' % name)        

    def _write_numeric(self, data, file_):
        if data.typecode() in Numeric.typecodes['Character']:
            # string array
            self._write_string(data, file_)
            return
          
        shape = Numeric.shape(data)

        if len(shape) == 1: # one-dimensional ; faking it into a two-dim one
            shape = (1,shape[0])
            data = [data]

        # start generating file
        file_.write('# type: matrix\n')

        if len(shape) < 3: # case regular two-dimensional
            rows, cols = shape[0], shape[1]
            file_.write('# rows: %s\n# columns: %s\n' % (rows, cols))

        else: # case; multidimensional array. Not supported in early octave versions
            file_.write('# ndims: ' + str(len(shape))+"\n" )
            file_.write(' '.join(map(str,shape)) +"\n")

            rows = shape[0]

            # data is transposed because that's how it octave reads them
            # (reshaping is just for ease of iterating over it
            data = Numeric.reshape(Numeric.transpose(data),(rows,-1))
                       
        for x in xrange(rows):
            file_.write(' %s\n' % ' '.join(map(str, data[x])))            

    def _write_string(self, data, file_):
        shape = Numeric.shape(data)
        if len(shape) == 1:
            data = [data]

        file_.write('# type: string\n')
        file_.write('# elements: %d\n' % len(data))
        for string in data:
            file_.write('# length: %d\n' % len(string))
            file_.write('%s\n' % ''.join(map(str, string)))

    def _write_scalar(self, data, file_):
        file_.write('# type: scalar\n')
        file_.write('%s\n' % data)

    def _write_struct(self, data, file_):
        file_.write('# type: struct\n')
        file_.write('# length: %d\n' % len(data))
        for name, data in data.items():
            self._write_data_object(name, data, file_)
            
    


    def write_variable_dict(self, vars_, filename):
        
        try:
            file_ = open(filename, 'wb')
        except IOError, e:
            raise ZPluginError, "Could not open file for " +\
                  "writing. Check path, permissions and free space."
        
        file_.write('# Created by %s %s %s\n' %(Config.appname, Config.version,
                                                time.asctime(time.localtime(time.time()))))

        for name, value in vars_.items():
            self._write_data_object(name, value, file_)
        
        try:
            file_.close()
        except IOError, e:
            print 'Could not close %s' % filename

            
