# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from Library.Nodes import Tools, Plot, File, Debug, Input, View

from Library.Util.Config import Config
from System.Plugin.PluginRepository import PluginRepository
from System.Data.functionnode import FunctionNode
from System.Data.Plot3DNode import Plot3DNode
from System.events import InfoEvent
from Library.Nodes.Plot.Plot2DNode import Plot2DNode
from xml.parsers.expat import ExpatError
from System.Data.Node import Node
# used when checking if a class is a subclass of Node
from System.Data.Node import Node as super_class
from thirdparty import ElementTree

import copy
import os
import re
import types
from xml.parsers import expat

# domain exceptions
class UnknownNodeError(Exception):
    """Tried to create node not known to repository."""
    pass

class MissingTagsError(Exception):
    """Essential tags missing when parsing xml."""
    pass

class NodeEntry:
    def __init__(self):
        pass

    def get_full_name(self):
        return ".".join(self.get_categories() + [self.get_class_name()])


def _get_compulsory_text(root, tag_name):
    if root.find(tag_name) is not None:
        return root.findtext(tag_name)
    else:
        raise MissingTagsError, "Missing tag %s from xml." % repr(tag_name)


def _get_document_root(filename):
    try:
        document = ElementTree.parse(filename)
    except expat.error, e:
        if expat.ErrorString(e.code) == expat.errors.XML_ERROR_NO_ELEMENTS:
            raise MissingTagsError, "No tags found in %s." % filename
        else:
            raise
    except SyntaxError, e:
        if str(e).startswith("no element found: line 1,"):
            raise MissingTagsError, "No tags found in %s." % filename
        else:
            raise
    try:
        root = document.getroot()
    except IndexError:
        raise MissingTagsError, "Missing function tag in file %s." % filename
    return root


class FunctionNodeEntry(NodeEntry):
    """Factory for FunctionNodes.

    Information is loaded from XML and instances of FunctionNode can
    then be created from FuncionNodeEntry.create_node().
    """
    def __init__(self, filename, categories):
        NodeEntry.__init__(self)

        path, zmlfile = os.path.split(filename)
        
        self._categories = categories

        self._parameters = []
        self._inputs = []
        self._outputs = []       
        self._sub_objects = {}

        # parse document, possibly raising exception
        root = _get_document_root(filename)

        # function to call
        self._name = _get_compulsory_text(root, "title")
        self._class_name, ext = os.path.splitext(zmlfile)
        self._description = root.findtext("description", "")
        self._plugin = _get_compulsory_text(root, "plugin")
        self._scriptfile = os.path.join(path, _get_compulsory_text(root, "filename"))
        self._htmlLink = root.findtext("htmllink")
        self._menuAddress = root.findtext("menuaddress", None)
        self._doc_path = root.findtext("doc_path")
        self._icon = root.findtext("icon")

        self._node_element = root

        if root.find("subobjects"):
            for object_element in root.find('subobjects').findall("object"):
                self._sub_objects[object_element.get("port")] = object_element.text.split("$")

        for input_element in root.findall("input"):
            for parameter in input_element.findall("parameter"):
                self._parameters.append(parameter)

            for port in input_element.findall("port"):
                if port.findtext("isNamed") == "1":
                    named = "isNamed=1"
                else:
                    named = 0
                self._inputs.append((port.findtext("name"),
                                     port.findtext("type", ""),
                                     port.findtext("description"),
                                     named))

        for output_element in root.findall("output"):
            for port in output_element.findall("port"):
                self._outputs.append((port.findtext("name"),
                                      port.findtext("type", ""),
                                      port.findtext("description")))

    def create_node(self):
        """Create the FunctionNode instance represented by this entry."""
        newNode = FunctionNode(self._name, PluginRepository())
        newNode.from_xml_element(self._node_element)

        # setting the name used in the node repository
        newNode.setNodeRepName(self.get_full_name())

        newNode.setFileName(self._scriptfile)
        newNode.setMenuAddress(self._menuAddress)

        newNode.set_plugin_type(self._plugin)

        for values  in self._inputs:
            isNamed = values[3]
            values = values[:3]
            port = apply(newNode.create_input_port,values)
            port.setIsNamed(isNamed)

        for values in self._outputs:
            apply(newNode.create_output_port,values)

        # setting sub object
        newNode.set_sub_objects(self._sub_objects)
        
        # setting icon
        if self._icon != None:
            newNode.set_icon(self._icon)

        return newNode
    
    def get_doc_path(self):
        return self._doc_path

    def get_icon_path(self):
        return self._icon

    def get_class_name(self):
        return self._class_name

    def get_categories(self):
        if self._menuAddress:
            return self._menuAddress.split(".")
        else:
            return self._categories

    def get_name(self):
        return self._name


class PythonNodeEntry(NodeEntry):
    """Represents an available python based node."""
    def __init__(self, class_name, categories, class_):
        NodeEntry.__init__(self)
        self._class_name = class_name
        self._categories = categories
        self._class = class_        

    def get_class_name(self):
        return self._class_name

    def get_categories(self):
        return self._categories

    def create_node(self):
        node = self._class(self.get_class_name())
        node.setNodeRepName(self.get_full_name())
        return node

    def get_name(self):
        return self.get_class_name()    
    

class PythonNodeSource:
    """Gives repository access to all nodes in a set of modules."""
    def __init__(self, modules):
        """Generate a list of entries from all nodes in modules.

        The modules parameter is a list of modules containing class
        modules within. Each class module should contain a class of
        the same name.

        It is a requirement that all modules contain the __all__
        attribute showing the class modules within, or that all class
        modules are already available in the modules' namespace.
        """
        self._modules = modules

    def get_entries(self):
        entries = []
        for module in self._modules:
            # does the following line have any function at all?
            #attributes = map(lambda a : getattr(module, a), dir(module))

            class_modules = {}
            classes = {}

            category = module.__name__.split('.')[-1]

            attributes = {}
            for a in dir(module):
                attributes[a] = None
            
            if hasattr(module, "__all__"):
                for a in module.__all__:
                    if not attributes.has_key(a):
                        __import__(module.__name__ + "." + a)
                        
            for name in dir(module):
                sub_module = getattr(module, name)

                if not isinstance(sub_module, types.ModuleType):
                    continue

                for class_name in dir(sub_module):
                    node_class = getattr(sub_module, class_name)

                    if not isinstance(node_class, types.ClassType):
                        continue

                    if not issubclass(node_class, Node):
                        continue

                    if node_class == Node:
                        continue

                    entries.append(PythonNodeEntry(class_name, [category], node_class))

        return entries

class FunctionNodeSource:
    """Gives repository access to function nodes in a directory."""
    def __init__(self, path, event_server):
        self._entries = []

        for root, dirs, files in os.walk(path):
            zml_files = filter(lambda entry : entry.endswith('.zml'), files)
            names = map(lambda entry : os.path.splitext(entry)[0], zml_files)

            # all function nodes shall have Toolboxes as first
            # category (unless they have an address, which is
            # FunctionNodeEntry's responsibility).
            categories = ["Toolboxes"]
            # split path into categories and remove empty entries
            categories.extend(filter(None, root[len(path):].split(os.sep)[1:]))
            
            for name in names:
                zml_file = os.path.join(root, name + '.zml')
                if not os.access(zml_file, os.R_OK):
                    continue
                try:
                    self._entries.append(FunctionNodeEntry(zml_file, categories))
                except ExpatError, e:
                    message = "Invalid zml-file for function %s:\n%s" % (name, e)
                    event_server.sendEvent(InfoEvent(None, "INFO", message))
                except MissingTagsError, e:
                    message = "Invalid zml-file for function %s:\n%s" % (name, e)
                    event_server.sendEvent(InfoEvent(None, "INFO", message))
                except SyntaxError, e:
                    message = "Invalid zml-file for function %s:\n%s" % (name, e)
                    event_server.sendEvent(InfoEvent(None, "INFO", message))

            if '.svn' in dirs:
                dirs.remove('.svn') # don't visit SubVersion directories
    
    def get_entries(self):
        """Return all entries accessible through this source."""
        return self._entries

class NodeRepository:
    """Repository of available nodes in SciCraft.

    All available nodes, both subclasses of System.Data.Node.Node and function
    nodes are made available through the NodeRepository interface.
    """

    DEFAULT_MODULES = Debug, Tools, File, Input, Plot, View
                      
    
    def __init__(self, pluginRepository=PluginRepository(),
                 node_modules=None, loadPath=None, functionPath=None,
                 event_server=None):
        """Constructor for NodeRepository. Makes a NodeRepository.

        pluginRepository - store of all plugins
        node_modules - modules containing class modules (modules with
                       same-named classes)
        loadPath - path to user-defined .zml-files (functions)
        functionPath - path to included .zml-files (functions)
        """
        self._pluginRepository = pluginRepository
        self._event_server = event_server

        # remember paths for reloading
        self._node_modules = None
        self._load_path = None
        self._function_path = None
        self.updatePaths(node_modules, loadPath, functionPath)
        self._report_duplicates()

    def reload_functions(self):
        """Reload all functions from paths loaded earlier."""
        self.updatePaths(self._node_modules, self._load_path,
                         self._function_path)

    def updatePaths(self, node_modules=None, loadPath=None, functionPath=None):
        """Update this repository with nodes/functions in new paths."""

        self._node_modules = node_modules
        self._load_path = loadPath
        self._function_path = functionPath

        # clear settings
        self._sources = []
        self._entries = {}

        if loadPath or functionPath:
            for path in self.__create_load_path_list(loadPath, functionPath):
                self._sources.append(FunctionNodeSource(path, self._event_server))

        if node_modules:
            self._sources.append(PythonNodeSource(node_modules))

        self._duplicates = {}

        for source in self._sources:
            for entry in source.get_entries():
                if not self._entries.has_key(entry.get_class_name()):
                    self._entries[entry.get_class_name()] = entry
                else:
                    self._duplicates.setdefault(entry.get_class_name(), []).append(entry)

        # to ensure backwards compatibilty in ZML-files
        if self._entries.has_key('VectorSpreadsheetNode'):
            self._entries['VectorSpreadsheetNode'] = self._entries['SpreadSheetNode']

    def _report_duplicates(self):
        duplicate_names = []
        keept_names = []
        for name, duplicates in self._duplicates.items():
            keept_names.append(self._entries[name])
            [duplicate_names.append(x) for x in map(NodeEntry.get_full_name, [self._entries[name]] + duplicates)]
        
        if self._duplicates:
            message = "Duplicate function-entries\n\n"
            message += "\n".join(duplicate_names)
            message += "\n\nOnly one node will use the same name:\n"
            message += '\n'.join([x.get_full_name() for x in keept_names])
            message += "\nOther nodes with same name will be ignored."
            if self._event_server:
                self._event_server.sendEvent(InfoEvent(None, "INFO", message))
            
    def createNodeByName(self, name):
        """Returns a single node as requested with the name:string.

        The name is the fully qualified node name as returned by
        getNodeList(). This name is also the same as read from
        XML.
        """
        plotvalue = False
        isCustomPlot = False
        oldname = name

        base_name = name.split(":")[0].split(".")[-1]

        if name.count(":"):
            first, last = name.split(':', 1)

            if base_name == "Plot3DNode":
                name = "Library.Nodes." + first
                plotvalue = last.split(".", 1)[-1]

                path = name.split(".")
                plotnode = Plot3DNode(name, plotvalue)
                plotnode.nodeRepName = oldname
                plotnode.setName(oldname.split(":", 1)[0])
                return plotnode
                    
            # checking if the node to be created is a custom plot
            if base_name == "Customplots" or base_name == "Saved plots":
                #in the case of a custom Plot2D node
                customPlotNode = Plot2DNode(name)
                customPlotNode.nodeRepName = name
                customPlotNode.setName(name.split(":")[1])

                #setting the custom settings
                path = os.path.join(Config.customplot2ddir, name.split(":")[1] + ".zpl")
                stream = open(path, 'r')
                xml_string = stream.read()
                root = ElementTree.Element("node")
                root.insert(0, ElementTree.fromstring(xml_string))
                customPlotNode.from_xml_element(root)

                return customPlotNode
        else:
            if self._entries.has_key(base_name):
                return self._entries[base_name].create_node()
            raise UnknownNodeError, "Can not find the node " + name + " in the repository."

    def get_doc_path_by_node_name(self, name):
        if self._entries.has_key(name):
            doc_path = self._entries[name].get_doc_path()
            if doc_path:
                filename = self._entries[name]._scriptfile
                full_path = os.path.join(os.path.split(filename)[0], doc_path)
                if os.path.exists(full_path) and os.path.isfile(full_path):
                    return full_path

        return None
        
    def get_icon_path_by_node_name(self, name):
        if self._entries.has_key(name):
            icon_path = self._entries[name].get_icon_path()
            if icon_path:
                filename = self._entries[name]._filename
                # path given in zml file is absolute
                if os.path.isabs(icon_path):
                    full_path = icon_path
                # path given in zml file is relative, merge with path to zml-file
                else:
                    full_path = os.path.normpath(os.path.join(os.path.split(filename)[0], icon_path))

                if os.path.isfile(full_path):
                    return full_path

        return None
        
    def get_class_by_node_name(self, name):
        if self._entries.has_key(name):
            return self._entries[name]
        return None

    def getNodeList(self):
        """Return list of (fully-qualified-class-name, name) pairs.

        The fully qualified class name can be used with
        createNodeByName() to retrieve an instance of the node.
        """
        nodes = {}
        for class_name, entry in self._entries.items():
            nodes[entry.get_full_name()] = entry.get_name()
            
        return nodes

    def __create_load_path_list(self, loadpath, functionpath):
        """Return list of current loadpaths."""

        loadPathList = []

        if loadpath:
            # correct possible wrong directory separators
            if   os.name == "nt":
                loadpath = loadpath.replace("/",  os.sep)
            elif os.name == "posix":
                loadpath = loadpath.replace("\\", os.sep)

            userPathList = loadpath.split(";")

            for path in userPathList:
                path = path.strip()

                if not path:
                    continue
        
                if functionpath and os.path.abspath(path) == os.path.abspath(functionpath):
                    continue

                if not os.path.exists(path):
                    print "Warning. Loadpath", path, "does not exist."
                    continue

                loadPathList.append(path)

        if functionpath:
            loadPathList.insert(0, functionpath)

        return loadPathList



    def _search_nodes(self):
        """Method that will search and find nodes"""
        nodes = {}        
        dict_ = copy.copy(globals())
        
        for key in dict_:    
            for attr in dir(dict_[key]):
                a = getattr(dict_[key], attr)
                if type(a) == types.ClassType and a.__name__ != "Node":
                    if issubclass(a, super_class):
                        # join e.g. ['EditNodes', 'MoleculeEditorNode'] to EditNodes.MoleculeEditorNode
                        dict_key = '.'.join(a.__module__.split(".")[2:])
                        nodes[dict_key] = a.__name__

        return nodes

