# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

import time
from threading import RLock


class NodeStatusDataList:
    """ 
        Maintains node status data for the nodes in a module diagram
        designated as active. The status data is propagated to registered
        displayers/listeners on each status update. Displayers/listeners must
        implement the method 'model_update(self, instance, model)', where 
        instance is the NodeStatusDataList initiating the method call. 'model'
        is a dictionary with node status data. Node id is used as keys, the
        entry is a list with status data.
    """
    
    
    def __init__(self, module_diagram=None):
        self._data_list = {}
        self._module_diagram = module_diagram
        self._displayers = []
        self._mutex = RLock()    
        
    
    def set_module_diagram(self, md):
        """ Sets a new active module diagram and update the status data. """
        self._module_diagram = md
        self.update_list()            
        
    
    def register_displayer(self, displayer):
        """ 
            Register the specified displayer and propagates the current
            node status data to it.
        """
        if displayer is not None and displayer not in self._displayers:
            self._displayers.append(displayer)
            # Propagate current model to new displayer at once.
            self._mutex.acquire()
            try:
                try:
                    displayer.model_update(self, self._data_list)
                except Exception, e:
                    pass
            finally:
                self._mutex.release()

    
    def remove_displayer(self, displayer):
        try:
            self._displayers.remove(displayer)
        except ValueError, te:
            pass
   
   
    def update_list(self):
        """ 
            Causes the status data to be fetched from the nodes in the
            active module diagram and the new status data is then propagated 
            to the registered displayers.
        """
        self._mutex.acquire()
        try:
            self._data_list = {}
            if self._module_diagram is not None:
                nodes = self._module_diagram.get_nodes()
                for node in nodes:
                    self._data_list[node] = self._extract_node_data(node)
            # Notify displayers
            id_ = -1
            if self._module_diagram is not None:
                id_ = self._module_diagram.getId()
            for displayer in self._displayers:
                displayer.post_event((self, self._data_list))
        finally:
            self._mutex.release()

   
    def _extract_node_data(self, node):
        """ Extracts the data we want from a node """
        data = []
        data.append(node.get_status_string())
        # We do not want node names to have 'Node' at the end.
        data.append(node.getName().replace("Node ", " "))
        data.append(node.is_alive())
        return data

