# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.
  
from Library.Util.Config import Config
from System import Tools
from System.Exceptions import ZPluginError

import os 
import re 
import time 
import types

class PluginRepository:
    
    def __init__(self):
        self.totalProcLimit = 10 # Limit of active processes pr plugin.
        try:
            self.pluginProcLimits =  self.__getPluginProcLimits()
        except ValueError, e:
            print e
        
        self.availablePlugins = self.__getAvailablePlugins() # List of plugins found in the plugin-directory.
        self.activePlugins = {} # Dictionary with lists of plugin-processes. Plugin-names are used as keys.
        self.TIMEOUT = 0.4 # Timeout. Used to wait for busy plugin-processes to become available.
        self.numOfIdleProcs = 1 # Max number of idle processes pr plugin.

        

    def getPlugin(self, plugin):
        """
           Params:
                  - plugin : String - a String object with the name of the plugin requested
           Return:
                  - Plugin - a plugin object depending on the request
           Exception:
                  - ZPluginError - Unknown Plugin

           This method will return a Plugin corresponding to the String parameter given. If
           the Plugin requested is not found, the method will return None"""

        if type(plugin) is not types.StringType:
            raise TypeError, "Illegal parameter. Expected string."
        
        # find out if plugin exists in directory
        if plugin not in self.availablePlugins:
            raise ZPluginError.ZPluginError, "Plugin %s not found" % plugin

        UNLOCKED = 0 # Plugin instance is not locked.

        keys = self.activePlugins.keys()
        retPlugin = None # Reference to the plugin returned by this method.
        if plugin not in keys: # No instances of plugin 'plugin' is currently running.
            retPlugin = self.__createPluginInstance(plugin) # Creating new instance of 'plugin'.
            self.activePlugins[plugin] = [retPlugin] # Creating new instance of 'plugin'.
            retPlugin.setLocked()
            return retPlugin
        else: 
            while 1: 
                for process in self.activePlugins[plugin]:
                    if process.getLockStatus() == UNLOCKED: 
                        retPlugin = process
                        retPlugin.setLocked()
                        return retPlugin # Suitable unlocked plugin-instance found.
                if plugin not in self.pluginProcLimits.keys(): self.pluginProcLimits[plugin] = 10
                if len(self.activePlugins[plugin]) < self.pluginProcLimits[plugin]: # Process limit not reached.
                    retPlugin = self.__createPluginInstance(plugin)
                    self.activePlugins[plugin] += [retPlugin]
                    retPlugin.setLocked()
                    return retPlugin
                time.sleep(self.TIMEOUT) # No more instances allowed. Waiting, and retrying...


    def __createPluginInstance(self, plugin):
        process = None
        # Creating dynamic code for creating an instance of plugin 'plugin', and executing it.
        text = "from Library.Plugins.%s import %s \n" % (plugin, plugin)
        text += "process = %s()" % plugin
        exec(text)
        return process



    def getProcessLimit(self):
        return self.processLimit



    def setProcessLimit(self, limit):
        if type(limit) is not IntType:
            raise TypeError, "Illegal parameter. Expected integer."
        if limit < 1:
            raise ValueError, "Process-limit cannot be lower than 1."
        self.processLimit = limit


    def __getAvailablePlugins(self):
        # get list of files in directory
        availablePlugins = []
        list = os.listdir(os.path.join(Config.scicrafthome, "Library", "Plugins"))

        # strip of .py extensions: and make a list of files consisting only of those files.
        for name in list:
            parts = name.split('.')
            if len(parts)>1:    
                if parts[1] == "py":            
                    availablePlugins.append(parts[0])

        # remove the __init__ files             
        availablePlugins.remove("__init__")
        return availablePlugins



    def __getPluginProcLimits(self):
        file = []
        usrdir = Config.userconfigdir
        usrfile = os.path.join(Config.userconfigdir, 'pluginProcs.cfg')
        if not os.path.exists(usrfile):
            try:
                plugProcFile = open(usrfile, 'w')
                plugProcFile.write('OctavePlugin 3\nRPlugin 5\nPythonPlugin 10\n')
            except IOError, e:
                print e
        try:
            file = open(usrfile,'r').readlines()
        except IOError, e:
            print e

        lineformat = re.compile('^(\w*) (\d*)$')

        pluginProcLimits = {}

        for line in file:
            mat = lineformat.match(line)
            if mat == None:
                raise ValueError, 'Invalid format in file pluginProcs.cfg'
            else:
                pluginProcLimits[mat.groups()[0]] = Tools.stringToNumber(mat.groups()[1])
        return pluginProcLimits
