# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

class Event:
    """Event superclass.

    The intended use is to subclass this class for each category of events.
    """
    def __init__(self, event_class, event_type, description,
                 module_diagram_id=None):
        """
            Params:
                * event_class - string :
                    The event class. This is a string describing what kind of
                    event this is, for instance 'ErrorEvent' or 'DiagramEvent'.
                    Subclasses should provide an appropriate value.
                * event_type - string :
                    The type of the event. This is a string describing the
                    actual event, for instance 'PATHNOTFOUND' or 'START'.
                    Legal values should be specified in the subclasses
                    (e.g. for each class of events).
                * description - string :
                    A string describing the event. Detailed information and
                    information for the user will be contained here.
                * module_diagram_id - integer : 
                    The ID on the module diagram this event belongs to. If the
                    event occurs outside a module diagram, this should be 
                    'None' (which is the default).
        """
        self.module_diagram_id = module_diagram_id
        self.event_class = event_class
        self.event_type = event_type
        self.description = description

    def get_event_class(self):
        return self.event_class

    def getType(self):
        return self.event_type

    def getDescription(self):
        return self.description

    def get_module_diagram_id(self):
        return self.module_diagram_id

 
class NodeEvent(Event):    
    def __init__(self, module_diagram_id, node, event_type, description):
        """
        Event types
        - START               : Node has started execution.
        - HALT                : Node has (been) halted.
        - FINISH              : Node has finished execution.
        - RUNTIME             : The time it took to execute the node.
        - EXCEPTION           : An exception occured inside the node.
        - PATHNOTSPECIFIED    : The path to datasource was not
                                specified when node started.
        - INFO                : Info that should be shown to user
        """
        Event.__init__(self, "NodeEvent", event_type, description,
                       module_diagram_id = module_diagram_id.getId())
        
        self.node = node
        event_types = ["START", "HALT", "FINISH", "RUNTIME", "EXCEPTION",
                      "PATHNOTSPECIFIED", "INFO"]
        
        if not event_type in event_types:
            raise ValueError, "Illegal type for node event: " + event_type

    def getNode(self):
        return self.node


class ErrorEvent(Event):    
    def __init__(self, module_diagram_id, event_type, description):
        """
        Valid event types:
        Event descriptions;
        LOAD          :
        FILEFORMAT    :
        FILENOTFOUND  :
        XMLPARSE      :
        VALIDATION    : 
        NODECREATION  : Creation of node failed in NodeRepository.
        """
        Event.__init__(self, "ErrorEvent", event_type, description,
                       module_diagram_id = module_diagram_id)
        
        # legal error events
        self.event_types = ["LOAD", "FILEFORMAT", "FILENOTFOUND", "XMLPARSE",
                           "VALIDATION", "NODECREATION"]
        
        if event_type not in self.event_types:
            raise ValueError, "Illegal type for error event: " + event_type


class DiagramEvent(Event):    
    def __init__(self, module_diagram_id, event_type, description):
        Event.__init__(self,  "DiagramEvent", event_type, description,
                       module_diagram_id = module_diagram_id)
        
        # Legal event types, add new diagram events here.
        self.event_types = ["START", "HALT", "FINISH", "RUNTIME", "EXCEPTION",
                           "VALIDATIONERROR"]

        if event_type not in self.event_types:
            raise ValueError, "Illegal type for diagram event: " + event_type


class StartDiagramEvent(DiagramEvent):
    def __init__(self, module_diagram_id, description, num_nodes):
        DiagramEvent.__init__(self, module_diagram_id, "START", description)
        
        self._num_nodes = num_nodes

    def get_num_nodes(self):
        return self._num_nodes


class InfoEvent(Event):
    def __init__(self, module_diagram_id, event_type, description):
        self.event_types = ["VALIDATION", "INFO"]
        Event.__init__(self, "InfoEvent", event_type, description,
                       module_diagram_id = module_diagram_id)
        
        if event_type not in self.event_types:
            raise ValueError, "Illegal type for info event: " + event_type


class IgnorableWarningEvent(Event):
    def __init__(self, module_diagram_id, warning_id, message):
        """Create warning event that user can choose to ignore.
        
        warning_id identifies type of message, so user can choose to
        ignore messages of same type later.
        """
        Event.__init__(self, "Warning", "Ignorable", message, module_diagram_id)
        
        self._warning_id = warning_id
        self._message = message

    def get_warning_id(self):
        return self._warning_id
