#!/usr/bin/python
#-*- coding:utf-8 -*-
#
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from Library.Util.Config import Config
from Library.Util.rtester import RTester, RInstallError

import os
import re
import sys
from qt import QApplication, QObject, SIGNAL, QObject, qApp
from qtui import QWidgetFactory

class RLibInstaller:

    # File containing the libraries that needs to updated
    library_file = os.path.join(Config.scicrafthome, 'rlibs.txt')

    INTRO_PAGE = 0
    INSPECT_PAGE = 1
    INSTALL_LIST_PAGE = 2
    INSTALLING_PAGE = 3
    FINISH_PAGE = 4
    
    
    def __init__(self):        
        self._widget = QWidgetFactory.create(os.path.join(Config.uidir, 'rlibwizard.ui'),
                                             None, None)
        # We do not want to show the help button
        self._widget.helpButton().hide()

        if not self._load_rpy():
            print 'The python library rpy could not be imported. See http://rpy.sf.net on how to install this library'
            sys.exit(0)

        # bool used for keeping track  of whether we already have inspected/installed
        self._inspection_done = False
        self._install_done = False        

        self._libraries = {}
        self._missing_libraries = {}
        self._failed_libraries = []
        
        file_ = open(RLibInstaller.library_file)
        for line in file_.readlines():
            if line.startswith('#'):
                continue
            line = line.strip()
            if line.find('-') != -1:
                library, package = line.split('-')
                if self._libraries.has_key(package):
                    self._libraries[package].append(library)
                else:
                    self._libraries[package] = [library]
            else:
                self._libraries[line] = [line]

        QObject.connect(self._widget, SIGNAL('selected(const QString &)'),
                        self._slot_page_change)

    def _slot_page_change(self, page_title):
        
        current_page = self._widget.currentPage()
        if current_page == self._widget.page(RLibInstaller.INSPECT_PAGE):
            self._run_inspection()
        elif current_page == self._widget.page(RLibInstaller.INSTALLING_PAGE):
            self._install_libraries()
        elif current_page == self._widget.page(RLibInstaller.FINISH_PAGE):
            self._widget.cancelButton().hide()
            self._widget.setFinishEnabled(current_page, True)
            
    def show(self):
        self._widget.show()

    def _load_rpy(self):
        try:
            global rpy
            import rpy
        except:
            return False
        return True

    def _check_libraries_installed(self):
        text = str(self._widget.child('check_r_text_edit').text())        
        for package, library in self._libraries.items():
            for lib in library:
                qApp.processEvents()
                text += '%s... ' % lib
                self._widget.child('check_r_text_edit').setText(text)                 
                try:
                    rpy.r('library(%s)' % lib)
                    message = 'OK'
                except rpy.RException:
                    if self._missing_libraries.has_key(package):
                        self._missing_libraries[package].append(lib)
                    else:
                        self._missing_libraries[package] = [lib]
                    message = 'FAILED'
        
                text += '%s\n' % message
        text += '\nDone'
        self._widget.child('check_r_text_edit').setText(text)
        self._clean_up_r()

    def _clean_up_r(self):
        # Used to unload all libraries that has been loaded
        for package, library in self._libraries.items():
            for lib in library:
                try:
                    rpy.r('detach(package:%s)' % lib)
                except:
                    pass

    def _check_r_is_installed(self):
        text = str(self._widget.child('check_r_text_edit').text())
        text += 'R... '
        self._widget.child('check_r_text_edit').setText(text)
        
        try:
            installed = RTester.r_install_test()
        except RInstallError:
            installed = False

        text = str(self._widget.child('check_r_text_edit').text())
        if not installed:
            text += """FAILED

Installation will be aborted. Either R is not installed on this system or not properly configured.
Please visit http://scicraft.org/?cat=1&page=7&doc=node_R_problems.html for a description on how to fix this problem if R is installed."""
            self._widget.setFinishEnabled(self._widget.currentPage(), True)
        else:
            text += "OK\n\n"
            
        self._widget.child('check_r_text_edit').setText(text)

    def _run_inspection(self):
        if self._inspection_done:
            return

        # disable next and back button
        self._configure_back_next(False)
        self._check_r_is_installed()
        self._check_libraries_installed()
        self._configure_back_next(True)
            
        self._inspection_done = True
        
        if not self._missing_libraries:
            self._widget.cancelButton().hide()
            self._widget.setFinishEnabled(self._widget.currentPage(), True)
            self._widget.setNextEnabled(self._widget.currentPage(), False)
        else:
            self._widget.cancelButton().show()
            self._widget.setFinishEnabled(self._widget.currentPage(), False)
            self._widget.setNextEnabled(self._widget.currentPage(), True)

        text = str(self._widget.child('missing_libs_text').text())
        for package, library in self._missing_libraries.items(): 
            for lib in library:
                text += lib + '\n'

        self._widget.child('missing_libs_text').setText(text)

    def _install_libraries(self):
        if self._install_done:
            return

        self._configure_back_next(False)
        
        failed = False
        text = str(self._widget.child('installing_text').text())
        text += 'Installing...\n\n'
        rpy.r('source("http://www.bioconductor.org/getBioC.R")')
        
        for package, libraries in self._missing_libraries.items():
            qApp.processEvents()
            text += '%s... ' % package
            self._widget.child('installing_text').setText(text)
            try:
                if rpy.os.environ.has_key('R_LIBS'):
                    rpy.r('getBioC(c("%s"), lib="%s")' % (package, rpy.os.environ['R_LIBS']))
                else:
                    rpy.r('getBioC(c("%s"))' % package)
                for lib in libraries:
                    if not check_installed(lib):
                        raise
                text += 'Success\n'
            except:
                failed = True
                text += 'FAILED\n'
                self._failed_libraries.append(package)
            self._widget.child('installing_text').setText(text)

        if failed:
            finish_text = "The following libraries were not installed successfully:\n"
            for library in self._failed_libraries:
                finish_text += library + '\n'
            finish_text += "\nEither the library was unavailable or something is wrong with your installation of R. Please contact us at scicraft@scicraft.org if you are unable to rectify this problem. """
        else:
            finish_text = "All libraries were successfully installed."

        self._widget.child('finish_text').setText(finish_text)
        
        self._install_done = True
        self._configure_back_next(True)

    def _configure_back_next(self, allowed):
        self._widget.setNextEnabled(self._widget.currentPage(), allowed)
        self._widget.setBackEnabled(self._widget.currentPage(), allowed)

def check_installed(library):
    libraries = rpy.r('installed.packages()')
    for list_ in libraries:
        if library in list_:
            return True
    return False

def main():
    app = QApplication([])
    wizard = RLibInstaller()
    QObject.connect(app, SIGNAL("lastWindowClosed()"), app.quit)
    wizard.show()
    app.exec_loop()
    
    
if __name__ == "__main__":
    main()
