# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

from Library.Util.Config import Config
from System.Data.rtoken import RToken
import os
import sys

# Hack to add Pyro to our python path
thirdparty = os.path.join(Config.scicrafthome, "thirdparty")
if thirdparty not in sys.path:
    sys.path.append(thirdparty)

import Pyro.core
import Pyro.constants
import Pyro
import rpy

# to enable transfer of inf/nan-values until Python2.5 arrives
Pyro.config.PYRO_PICKLE_FORMAT = 0

def convert_it(value):
    """Replace non-convertible R-objects in struct/nested list with token."""
    if isinstance(value, list):
        for i in range(len(value)):
            value[i] = convert_it(value[i])
        return value
    elif isinstance(value, dict):
        for key in value:
            value[key] = convert_it(value[key])
        return value
    elif hasattr(value, "autoconvert") and hasattr(value, "as_py") \
             and hasattr(value, "local_mode"):
        return RToken()
    else:
        return value


class RPyHandler:
    def init_handler(self):
        # Make sure exception handling behaves on rpy <0.4.6 running
        # non-interactively
        rpy.r("options(error = expression(NULL))")
        
    def rpy_eval(self, text):
        """Evaluate R-statement and return result.

        Can throw PicklingError if trying to return value with
        non-convertible, in that case, use rpy_converted_eval instead.
        """
        value = rpy.r(text)
        sys.stdout.flush()
        return value

    def rpy_converted_eval(self, text):
        """Evaluate R-statement and return result.

        This method will return RToken instances instead of
        non-convertible R-objects. Because of the conversion through
        traversal of the entire structure, this method is slower than
        rpy_eval.
        """
        value = rpy.r(text)
        sys.stdout.flush()
        return convert_it(value)

    def rpy_silent_eval(self, text):
        """Evaluate R-statement and ignore possible return value.

        This method should be used for all R-evaluation where return
        values are not kept as it keeps incidental results from being
        converted to python and then transported through Pyro. Should
        always be faster than or equal to rpy_eval.
        """
        mode = rpy.get_default_mode()
        rpy.set_default_mode(rpy.NO_CONVERSION)
        try:
            rpy.r(text)
            sys.stdout.flush()
        finally:
            rpy.set_default_mode(mode)

    def rpy_assign(self, name, value):
        """Assign python object as value of variable in R."""
        mode = rpy.get_default_mode()
        rpy.set_default_mode(rpy.NO_CONVERSION)
        try:
            rpy.r.assign(name, value)
        finally:
            rpy.set_default_mode(mode)


def main():
    print "pid:", os.getpid()
    Pyro.core.initServer(0)
    daemon = Pyro.core.Daemon()
    daemon.setAllowedIdentifications([sys.stdin.readline().strip()])
    # We only allow one connection at a time access to rpy, as rpy
    # isn't threadsafe
    obj = Pyro.core.SynchronizedObjBase()
    handler = RPyHandler()
    obj.delegateTo(handler)
    uri = str(daemon.connect(obj, "rpyhandler"))
    slash_index = uri.find(":",5)
    uri = "PYRO://127.0.0.1" + uri[slash_index:]
    print uri
    sys.stdout.flush()
    daemon.requestLoop()

if __name__=="__main__":
    main()
