#!/usr/bin/python
#-*- coding:utf-8 -*-
#
# SciCraft - An interactive graphics front end to advanced data mining
# Copyright (C) 2002-2005 SciCraft Development Team
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# As a special exception, SciCraft Development Team gives permission
# to link this program with  Qt non-commercial edition, and distribute
# the resulting executable, without including the source code for the
# Qt non-commercial edition in the source distribution.

"""This is the main run-file for SciCraft.

Usage:
> python scicraft.py <xml-file>

The xml-file is optional
"""
import os
import sys
import qt
import getopt
import re
from os import linesep as nl
from qt import SIGNAL, QObject

connect = QObject.connect

from Library.Util.Config import Config, read_config
from Library.Util.Logging import Logging
from Library.Util.rpyclient import rpy_quit
from GUI.SciCraftSplash import SciCraftSplash


def _r_path_loader():
    """This function will search the Windows registry to see if R
    is installed and in that case, add the install path to the path environment
    """

    import _winreg as wreg
    try:
        key = wreg.OpenKey(wreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\R-core\\R")
        path = wreg.QueryValueEx(key, "InstallPath")[0]
    except Exception, e: # R is not installed
        return False

    # Add path to the environment (will only affect the current python session)
    os.environ["PATH"] += ";" + str(path) + "\\bin"

    #Final check that everything is working:
    (stdin, stdout, stderr) = os.popen3("RTerm --version")
    if not re.match("R [0-9]+\.[0-9]+(\.[0-9]*)?", stderr.read()):
        return False
    return True


def _octave_path_loader():
    """Add the path to octave to the path envrironment.

    This functino is only executed on Windows.
    """

    import _winreg as wreg
    location_50 = "SOFTWARE\\GNU Octave\\Cygwin\\mounts v2\\/usr/bin"
    location_60 = "SOFTWARE\\GNUOctave 2.1.60\\Cygwin\mounts v2\\/usr/bin"
    location_64 = "SOFTWARE\\GNUOctave 2.1.64\\Cygwin\mounts v2\\/bin"
    path = None
    
    try: # look for 2.1.64
        key = wreg.OpenKey(wreg.HKEY_LOCAL_MACHINE, location_64)
        path = wreg.QueryValueEx(key, "native")[0]
    except:
        try: # look for 2.1.60
            key = wreg.OpenKey(wreg.HKEY_LOCAL_MACHINE, location_60)
            path = wreg.QueryValueEx(key, "native")[0]
        except:
            try: # look for 2.1.50
                key = wreg.OpenKey(wreg.HKEY_LOCAL_MACHINE, location_50)
                path = wreg.QueryValueEx(key, "native")[0]
            except:
                pass
            
    if not path:        
        return False
        
    os.environ["PATH"] += ";" + str(path)

    #Final check    
    string = os.popen('octave -v').read()
    if not string.find("GNU Octave") != -1:
        return False

    return True    

def sanity_check():
    """This function will try to import all libraries needed. 
    
    In the case of an error, display the result to the user.
    """

    rpy_check = True
    octave_check = True

    if os.name == "nt":
        # need to use popen3 here since R on Windows will only write to stderr
        (stdin, stdout, stderr) = os.popen3("RTerm --version")
        if not re.match("R [0-9]+\.[0-9]+(\.[0-9]*)?", stderr.read()):
            rpy_check = _r_path_loader()
    else:
        ustring = os.popen("R --version").read()
        if ustring.find("R") != 0:
            rpy_check = False

    string = os.popen('octave -v').read()
    if not string.find("GNU Octave") != -1:
        if os.name == "nt":
            octave_check = _octave_path_loader()
        else:
            octave_check = False
    
    if not rpy_check or not octave_check:
        error_app = qt.QApplication(sys.argv)

        main_window = qt.QMainWindow(None)
        main_window.setFixedWidth(400)
        main_window.setFixedHeight(295)
        main_window.setCaption("SciCraft - Warning")
        main_window.show()
        
        text_window = qt.QTextEdit(main_window)
        text_window.setGeometry(5, 5, 390, 250)
        text_window.show()
        
        error_app.setMainWidget(main_window)

        error_msg = "SciCraft startup test failed:" + nl*2
        if not rpy_check:
            error_msg += "Failed to load the rpy library. R might not "
            error_msg += "be installed and/or not in your path environment." + nl

        if not octave_check:
            error_msg += "Octave test failed. Octave might not be installed "
            error_msg += "and/or not in your path environment." + nl

        if not octave_check or not rpy_check and os.name == "nt":
            error_msg += nl + "See the following website for more information "
            error_msg += "about setting the environment path: " + nl
            error_msg += "http://www.scicraft.org/index.php"
            error_msg += "?cat=1&page=4&doc=node_Windows.html" + nl

        error_msg += nl*2 + "Press Accept to try starting SciCraft "
        error_msg += "anyway or Quit to exit." + nl
            
        text_window.setText(error_msg)
        text_window.setReadOnly(True)
                
        accept_button = qt.QPushButton("Accept", main_window)
        accept_button.setGeometry(156, 260, 45, 30)
        accept_button.show()
        
        quit_button = qt.QPushButton("Quit", main_window)
        quit_button.setGeometry(204, 260, 45, 30)
        quit_button.show()
                
        def quit():
            error_app.exit(0)

        def accept():
            error_app.exit(1)

        main_window.connect(accept_button, qt.SIGNAL("clicked()"), accept)
        main_window.connect(quit_button, qt.SIGNAL("clicked()"), quit)
        main_window.connect(error_app, qt.SIGNAL("lastWindowClosed()"), quit)
        main_window.setActiveWindow()
        return error_app.exec_loop()
    return True


# parse command line arguments before proceeding.
def __process_args(argv):
    optlist, args = getopt.getopt(argv, 'hvc:dD', ['help', 'version', 'config=',
                                                   'magic','designer', 'debug'])
    start_designer = False
    debug_mode = False
    for (option, arg) in optlist:
        if option in ('-h', '--help'):
            print 'Version: %s' % Config.version
            print 'Usage: scicraft [options] [file]...'
            print '-h, --help'
            print '\tDisplay this help text and exit.'
            print '-v, --version'
            print '\tShow version.'
            print '-r, --run'
            print '\tRun specified module diagram immediately.'
            print '-c, --config=FILE'
            print '\tUse FILE as user configuration file.'
            print '-d, --designer'
            print '\tStart MADesigner directly.'
            print '-D, --debug'
            print '\tRun SciCraft in debug mode. May also be enabled by setting'
            print '\tthe environment variable SCDEBUG.'
            sys.exit()
        elif option in ('-v', '--version'):
            print 'Version: %s' % Config.version
            sys.exit()
        elif option in ('-c', '--config'):
            read_config(arg)
        elif option in ('-d', '--designer'):
            start_designer = True
        elif option in ('-D', '--debug'):
            debug_mode = True
            
    return args, start_designer, debug_mode


def start_gui(run = False):
    args, start_designer, debug_mode = __process_args(sys.argv[1:])
    # also allow environment to set debug mode
    if os.getenv("SCDEBUG"):
        debug_mode = True

    read_config()

    app = qt.QApplication(sys.argv)

    # make sure stuff using Config isn't instantiated indirectly by
    # import before configuration has been read
    
    from GUI.application import ApplicationWindow
    application = ApplicationWindow(args, run, debug_mode, SciCraftSplash())
    # pp.slotFileQuit = scicraft.slotFileQuit
    app.setMainWidget(application._widget)
    application.show()
    if start_designer:
        application.create_new_ma_designer()

    # process hanging events
    connect(app, SIGNAL("lastWindowClosed()"), app.processEvents)

    app.exec_loop()
    application.save_config()
    rpy_quit(5.0)

def main():
    run = False
    Logging() # start logging to file if enabled in Config
    import sys
    if '--magic' in sys.argv:
        code = sys.stdin.read()
        eval(compile(code, 'interactive', 'exec'))
        sys.exit()
    
    if '--run' in sys.argv:
        run = True
        sys.argv.remove('--run')
    
    if '-r' in sys.argv:
        run = True
        sys.argv.remove('-r')
        
    
    if sanity_check():
        # insert 'import loadfile'-statement here when creating windows
        # must be after sanity_check()
        #import loadfile
        start_gui(run)


if __name__ == "__main__":
    main()

