function [outX,outY,cov,iter]=modified_simplex(X,extFunc,tol,maxIter)

% Modified simplex algorithm
%
% [outX,outY,cov,iter]=modified_simplex(X,extFunc,tol,maxIter)
%
% Input variables
% X:       the initial simplex with the vertices arranged in rows, 
%          i.e. in k-dimensional space, k+1 vertices are needed,
%          and X must have dimensions (k+1,k)
% extFunc: an external function from which the responses are
%          retrieved. The name should be placed in apostrophes 
% tol:     Stop criterion; standard deviation of the responses as a
%          percentage of the mean response.
% maxIter: maximum number of iterations
%
% outX:    the output simplex, best point is the last element
% outY:    the corresponding responses
% cov:     standard deviation of the responses as a percentage of
%          the mean response. 
% iter:    number of iterations in the run
%
% The function does not account for limitations in the input
% variable space.

% Reference:
% Morgan, e.M., Burton, K.W., Nickless, F.: Optimization using the
% modified simplex methond, Chem. Intell. Lab., Syst, 7 (1990), 209-222

% April 8, 2003
% Bjrn Fjeldstad

[rows, cols]=size(X);
if~(rows==cols+1),
   error('Number of columns must be equal to number of rows + 1'),
end
if~(rank(X)==cols),
   error('The vectors does not span the variable space'),
end

cov=100;
iter=0;

outY=feval(extFunc,X); %Calculates the responses in the initial simplex
outX=X;

while (iter<maxIter && cov>tol && rank(outX)==cols)
  % Flow diagram from the reference.
  % There are no check whether any vertex is retained k+1 times,
  % which is recommended therein. 
   
  [rankResp,rankIdx]=sort(outY); %Rank from below
  
  outX=outX(rankIdx,:);          %Sorterer InnSimpl
  outY=outY(rankIdx,:);
  
  W=outX(1,:);
  w=outY(1);
  N=outX(rows-1,:);
  n=outY(rows-1);
  B=outX(rows,:);
  b=outY(rows);
  Nw=outX(2,:); % My own addition. Equal to N in the 2D-case
  nw=outY(2);
  

  % Find reflection point
  P=sum(outX(2:rows,:))/cols;  % Eq (1)
  R=P+(P-W);                   % Eq (2)   
  r=feval(extFunc,R);

  if r>b
    % Make E
    E=P+2*(P-W);        % Eq (3)
    e=feval(extFunc,E);
    
    if e>r
      %new simplex BNE
      outX(1,:)=E;
      outY(1)=e;
      % disp("1. BNE");
    else
      %new simplex BNR
      outX(1,:)=R;
      outY(1)=r;
      % disp("2. BNR");
    end %e>r
    
  elseif r>n
    %new simplex BNR
    outX(1,:)=R;
    outY(1)=r;  
    % disp("3. BNR");
  elseif r>w
    % make Cr
    Cr=P+.5*(P-W);        % Eq (4)
    cr=feval(extFunc,Cr);
    if cr>r
      %new simplex BNCr
      outX(1,:)=Cr;
      outY(1)=cr;
      % disp("4. BNCr");
    else 
      %make Rmark from Nw
      Pmark=sum(outX([1 3:rows],:))/cols;    % Eq (1)
      Rmark=Pmark+(Pmark-Nw);                % Eq (2)
      %new Simplex B-Cr-Rmark
      outX([1 2],:)=[Cr; Rmark];
      outY([1 2])=[cr; feval(extFunc,Rmark)];
      % disp("5. BCrRmark");
    end % if cr>r

  else
    %make Cw
    Cw=P-.5*(P-W);
    cw=feval(extFunc,Cw);
    if cw>w
      %new simplex BNCw
      outX(1,:)=Cw;
      outY(1)=cw;
      % disp("6. BNCw");
    else 
      %make Rmark from Nw
      Pmark=sum(outX([1 3:rows],:))/cols;  % Eq (1)
      Rmark=Pmark+(Pmark-Nw);                 % Eq (2)   
      % new simplex BCwRmark
      outX([1 2],:)=[Cw; Rmark];
      outY([1 2])=[cw; feval(extFunc,Rmark)];
      % disp("7. BCwRmark");
    end % if cw>w

  end % if r>b

  SD=sqrt((rows*sumsq(outY)-sum(outY)^2)/(rows*(rows-1))); %Eq (8)
  MR=abs(sum(outY))/rows; %Mean response
  cov=100*SD/MR;
  iter=iter+1;
end % while-loop

outY=feval(extFunc,outX);
[rankResp,rankIdx]=sort(outY);
outX=outX(rankIdx,:); 
outY=outY(rankIdx,:);



