function [outX,outY,cov,iter]=supermodified_simplex(X,extFunc,tol,maxIter)

% Super-modified simplex algorithm 
%
% [outX,outY,cov,iter]=supermodified_simplex(X,extFunc,tol,maxIter) 
%
% Input variables
% X:       the initial simplex with the vertices arranged in rows, 
%          i.e. in k-dimensional space, k+1 vertices are needed,
%          and X therefore has dimensions (k+1,k)
% extFunc: an external function from which the responses are
%          retrieved. The name should be placed in apostrophes 
% tol:     Stop criterion; standard deviation of the responses as a
%          percentage of the mean response.
% maxIter: maximum number of iterations
%
% outX:    the output simplex, best point is the last element
% outY:    the corresponding responses
% cov:     standard deviation of the responses as a percentage of
%          the mean response. 
% iter:    number of iterations in the run
%
% The function does not account for boundaries in the input space

% Reference:
% Morgan, E.M., Burton, K.W., Nickless, F.: Optimization using the
% super-modified simplex methond, Chem. Intell. Lab., Syst, 8
% (1990), 97-107

% April 8, 2003
% Bjrn Fjeldstad


margin=.2; % Set to constant, may be set as an input variable

[rows, cols]=size(X);
if~(rows==cols+1),
   error('Number of columns must be equal to number of rows + 1'),
end
if~(rank(X)==cols),
   error('The vectors does not span the variable space'),
end

cov=100;
iter=0;

outY=feval(extFunc,X); %Calculates the responses in the initial simplex
outX=X;

while (iter<maxIter && cov>tol && rank(outX)==cols)
  
  % Flow diagram in the reference.
  % There are no check whether any vertex is retained k+1 times,
  % which is recommended there. 
  
  [rankResp,rankIdx]=sort(outY); % Rank from below
  
  outX=outX(rankIdx,:);          % Sorting
  outY=outY(rankIdx,:);
  
  W=outX(1,:); %Picking worst vertex
  w=outY(1);
  
  % Find reflection point
  P=sum(outX(2:rows,:))/cols;  % Eq (1), generalized
  p=feval(extFunc,P);
  R=P+(P-W);                   % Eq (2)   
  r=feval(extFunc,R);
  
  Yopt=(w-p)/(w-2*p+r)+.5; % eq (5)

  if (Yopt<-1)
    Yopt=-1;
  elseif (Yopt> -1*margin && Yopt<0)
    Yopt=-1*margin;
  elseif (Yopt<margin && Yopt>0)
    Yopt=margin;
  elseif (Yopt<1 && Yopt>1-margin)
    Yopt=1-margin;
  elseif (Yopt<1+margin && Yopt>1)
    Yopt=1+margin;
  elseif (Yopt>3)
    Yopt=3;
  end
  
  %Make O
  O=Yopt*P+(1-Yopt)*W;
  o=feval(extFunc,O);
  
      
  if o>r
    outX(1,:)=O;
    outY(1)=o;
  else
    outX(1,:)=R;
    outY(1)=r;
  end
  
  SD=sqrt((rows*sumsq(outY)-sum(outY)^2)/(rows*(rows-1))); %Eq (8)
  MR=abs(sum(outY))/rows; %Mean response
  cov=100*SD/MR;
  iter=iter+1;
end % while-loop

outY=feval(extFunc,outX);
[rankResp,rankIdx]=sort(outY);
outX=outX(rankIdx,:); 
outY=outY(rankIdx,:);



