
function [emscData,emscParam] = emsc(data)
%,offset,slope,slopeSquared,xValues,xValuesSquared,extraModelParam)

offset = 1;
slope = 1;
slopeSquared = 0;
xValues = 1;
xValuesSquared = 1;
extraModelParam = data(2,:) - data(47,:);

%Performs Extended Multiplicative Signal Correction (EMSC) on data
%
% [emscData,emscParam] = emsc(data, slope, offset, slopeSqared, ...
%				 xValues, xValuesSquared, extraModelParam)
%
% Input Variables:
%  data:        Data matrix to be transformed
%  offset:      1 for including additive model term. 0 otherwise.
%  slope:       1 for including multiplicative model term. 0
%               otherwise. The multiplicative term is chosen as the
%               mean spectrum. 
%  slopeSquared: 1 for including squared an additive model term
%                (the square of the mean spectrum). 0 otherwise.
%  xValues:     1 for including the x-variable values
%               (ex. Wavelengths in spectra) in the model. 0
%               otherwise. Additive.
%  xValuesSquared: 1 for including the square of the x-variable
%               values. 0 otherwise. Additive
%  extraModelParam: Extra model Parameter matrix to include in the
%               model. The matrix must have the same number of
%               columns as x-variables. [] or 0 if not used.
%
% Output Variables:
%   emscData:    EMSC transformed data matrix
%   emscParam:   Correction parameters for the different model
%                terms: [offsetPar, xValuesPar,xValuesSqPar,
%                slopeSqPar, extraModPar, slopePar]

%author M.B. Rye, Oct 10, 2002


X = data;

CondNumber = 1e19;

[N,M] = size(X);
MSCWeights = ones(1,M);

[N2,M2] = size(extraModelParam);
Xmod = [];
Ref = mean(X)';


%Offset   
if offset == 1
  Xmod = [ones(M,1)];
end % if

%xValues
if xValues == 1
  lambda = (1:M);
  Xmod = [Xmod lambda'];
end % if

%xValuesSquared
if xValuesSquared == 1
  lambda = (1:M);
  meanLambda = mean(lambda');
  lambdaSq = (lambda - meanLambda).^2;
  Xmod = [Xmod lambdaSq'];
end % if

%Slope Squared
if slopeSquared == 1
  mRef = mean(Ref);
  RefSq = (Ref - mRef).^2;
  Xmod = [Xmod RefSq];
end % if

% Extra Model Parameters
if extraModelParam == 0
  extraModelParam = [];
end % if 

flagVal = !isempty(extraModelParam);

if flagVal == 1
  Xmod = [Xmod extraModelParam']; 
end % if

%Slope 
if slope == 1
  Xmod = [Xmod Ref];
end % if
  
diagW = diag(MSCWeights);
XX = Xmod'*diagW*Xmod;

% Ridging if CondNumber is specified. Calculetion of inverse XX.

if CondNumber>0
  [u,S,v]=svd(XX);
  s=diag(S)';
  sCorrected=max(s, (s(1)/CondNumber)); 
  XXCorrected=u*(diag(sCorrected))*v';
  InvXX = inv(XXCorrected);
else
  InvXX=inv(XX);
end % if

MSCParam = InvXX*Xmod'*diagW*X';
MSCParam = MSCParam';

%estimating Xnew

Xnew = X;
j = 1;
    
%Offset
if offset == 1
  ai = MSCParam(:,j);
  Xnew = Xnew-(ai*ones(1,M));
  j = j+1;
end % if

% xValues
if xValues == 1
  ei = MSCParam(:,j);
  Xnew = Xnew-(ei*lambda);
  j = j+1;
end % if

% xValues Squared
if xValuesSquared == 1
  fi = MSCParam(:,j);
  Xnew = Xnew-(fi*lambdaSq);
  j = j+1;
end % if
    
% Slope Squared
if slopeSquared == 1
  b2i = MSCParam(:,j);
  Xnew = Xnew - (b2i*RefSq'); 
  j = j+1;
end % if
    
%Extra Model Parameters
if flagVal == 1
  H = MSCParam(:,j:j+(N2-1));
  j = j+N2;
end % if

%Slope
if slope == 1
  bi = MSCParam(:,j);
  Xnew = Xnew./(bi*ones(1,M));  
  j = j+1;
end % if
    
emscData = Xnew;
emscParam = MSCParam;






        
    

       
   







