function [oscData, noise, B, T, P] = osc(xData, yData)
aMax = 3

% Performs orthogonal signal correction on data based on NIPALS algorithm
% Can be extended to be used in a cross-validation procedure.
% Data should be scaled centered before using the algorithm
%
%[oscData, noise, B, T, P] = osc(xData, yData, aMax)
%
%Input Variables:
%  xData:      Data matrix (nSamp, nVar) to be transformed
%  yData:      Related yData vector (nSamp). The function only
%              supports one y-variable.
%  aMax:       Number of principal components.
% 
%Output Variables:
%  oscData:    Transformed data matrix (nSamp, nVar)
%  noise:      Data removed (filtered) from matrix (nSamp, nVar)
%  B:          Regression Coefficient matrix (nVar, aMax)
%  T:          Scores matrix (nSamp, aMax)
%  P:          Loadings matrix (nVar, aMax)
%
%Related Functions:
%  plsrnipals.m, osccorrect.m
%References:
%  Wold et. al. Chemometrics and Intelligent laboratory Systems 44
%  (1998) 175-185
%  Sjblom et.al. Cemometrics and intelligent loboratory systems 44
%  (1998) 229-244

%author M.B. Rye, Oct 22, 2002

[nSamp, nXVar] = size(xData);
E = xData;
F = yData;
itLim = 1e-12;

regCoeff = zeros(nXVar, aMax);
scores = zeros(nSamp, aMax);
loadings = zeros(nXVar,aMax);

for j = 1:aMax
  [U,S,V] = svd(E);
  scores = U*S;
  tStart = scores(:,1);
  tLast = tStart;
  %tMean = mean(tStart);
  %tStart = tStart - tMean;
  
  itVal = 1;
  while itVal > itLim
    %tiOrt = (1-F*inv(F'*F)*F')*tiStart;
    tNew = tStart - (F*inv(F'*F)*F'*tStart); 
    %tOrt = tOrt + tMean;
    itVal = (norm(tNew-tLast))/(norm(tNew));
    tLast = tNew;
    w = E'*tNew/(tNew'*tNew); 
    w = w/(w'*w);
    tStart = E*w;
  end % while
  
  [W,T2,U,P2,Q,b] = plsrnipals(E,tNew,nXVar,0);
  
  tOrt = E*b;
  p = (E'*tOrt)/(tOrt'*tOrt);
    
  E = E - tOrt*p';
  
  T(:,j) = tOrt;
  P(:,j) = p;
  B(:,j) = b;
  
end % for

  
oscData = E;
noise = T*P';
