function [out, outDer,varRat,varDerRat]=LSPsmooth(data,n,m)

% Least squares polynomial smoothing
% 
% [out, outDer,varRat,varDerRat]=LSPsmooth(data,n,m)
%
% Input variables:
% data: raw data sequence, column vector with length k
% n:    the width of the window to fit each point  
% m:    order of the polynom, m<n-1
%
% out:       smoothed data
% outDer:    first derivative of the smoothed data
% varRat:    variance ratios for the smoothed data
% varDerRat: variance ratios for the derivative of the smoothed curve
%

% Reference
% Bialkowski, S.E.: Generalized digital smoothing filters made easy
% by matrix calculations, Anal. Chem, 61(1989), 1308-1309
% 

% April 4., 2003
% Bjrn Fjeldstad

k=size(data,1);
out=zeros(k,1);
outDer=zeros(k,1);

nl=floor((n-1)/2);
nr=ceil((n-1)/2);

%Uses the Vandermonde matrix. Columns are flipped to follow the
%notation used in the reference

fullV=fliplr(vander([1:n]));
V=fullV(:,1:m+1);

Vplus=inv(V'*V)*V';  %Eq (3)
VVplus=V*Vplus;

%Another possibilty, suggested in the reference
%[Q,R]=V    
%out=Q*Q'*x  (5d)

D=diag([1:m],1);     %Eq (7)
DerVVplus=V*D*Vplus;  

%Estimate of the first nl+1 points
x=data(1:n);
out(1:nl+1)=VVplus(1:nl+1,:)*x;        %Eq (4)
outDer(1:nl+1)=DerVVplus(1:nl+1,:)*x;  %Eq (8)

%Estimate of the last nr+1 points
x=data(k-n+1:k);
out(k-nr-1:k)=VVplus(n-nr-1:n,:)*x;        %Eq (4)
outDer(k-nr-1:k)=DerVVplus(n-nr-1:n,:)*x;  %Eq (8)

%Moving window, from (nl+2) to (k-nr-1)
for i=(nl+2):(k-nr-1)
  x=data(i-nl:i+nr);
  out(i)=VVplus(nl+1,:)*x;        %Eq (4)
  outDer(i)=DerVVplus(nl+1,:)*x;  %Eq (8)
end

%Calculating the variance ratios
%Equal for all points, except for the resp. nl/nr end points.

varRat=zeros(k,1);
varDerRat=zeros(k,1);
varRatData=diag(V*Vplus);                % Eq (10a)
varDerRatData=diag(V*D*inv(V'*V)*D'*V'); % Eq (10b)

%Insert data
varRat(1:nl)=varRatData(1:nl);
varRat(nl+1:k-nr-1)=varRatData(nl+1);
varRat(k-nr:k)=varRatData(n-nr:n);

varDerRat(1:nl)=varDerRatData(1:nl);
varDerRat(nl+1:k-nr-1)=varDerRatData(nl+1);
varDerRat(k-nr:k)=varDerRatData(n-nr:n);
