function [T,order] = HCA(X,method,metric,opt)
% [T,order] = HCA(X,method,metric,[opt])
% Uses the agnes() function in R to compute a hierarchical
% cluster analysis
%
%
% INPUT VARIABLES
%
% $$$ ##X                  = data matrix (objects x variables)
% $$$ ##method             = string with clustering type. String can be
% $$$ ##                    "average" (group average method), "single"
% $$$ ##                    (single linkage), "complete" (complete
% $$$ ##                    linkage), "ward" (Ward's method), and 
% $$$ ##                    "weighted" (weighted average linkage).
% $$$ ##metric             = string with what metric to use. Only
% $$$ ##                     possible ones are"euclidean" and "manhattan"
%
% $$$ ##opt                = option vector. If opt(1) = 0 (default) X is
% $$$ ##                     asymmetric, if opt = 1 we
% $$$ ##                     assume X is a symmetric dissimilarity matrix
% $$$ ##OUTPUT VARIABLE
%
% $$$ ##T                  = matrix containing the results from the
% $$$ ##                    clustering. Each row is a cluster. col no.1
% $$$ ##                    is cluster no., col no.2 cluster in left
% $$$ ##                    branch, col. no. 3 cluster in right branch, 
% $$$ ##                    col. no. 4 is height of cluster, col. no. 5
% $$$ ##                    is x coordinate of cluster
%
% $$$ ##order              = order of objects from left to right in
% $$$ ##                     dendrogram. Necessary for plotting dendrograms
%
% $$$ ##Copyright (C), B.K. Alsberg, bka@phys.chem.ntnu.no
% $$$ ##Last update: 8/5-2004
%
% Copyright(C), The SciCraft Team, www.scicraft.org
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.




if nargin < 4,
  opt = zeros(1,2);
  disp('********** no plotting of dendrogram now ***********');
  disp(' ***** use R_HCA_dendrogram.m later  if needed ********');
end;

if length(opt) < 2,
  opt2 = [0 opt];
  opt = opt2;
end;


%%%%%%%%%%%%%%% setting of system file names %%%%%%%%%%%%%%%%%
% The following is to create unique temp file names
% for each run to minimize the chances of clashes:
rnm = round(10000*rand(1));
comfile = ['run' int2str(rnm) 'R.R'];
datafile = 'data.inp';
dendrogram = ['merge' int2str(rnm) '.dat'];
ordername = ['order' int2str(rnm) '.dat'];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%% R-calling process %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%dostr = ['save ' datafile ' X -ascii '];
%eval(dostr);

loc_ascii_dump(datafile,X);

% Write R command file
loc_write_to_R(comfile,datafile,method,metric,opt,dendrogram,ordername);

if strcmp(computer,'i386-pc-linux-gnu')
  callstr = ['R BATCH <' comfile];
else
  callstr = ['R <' comfile ' --no-save '];
end;

system(callstr);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%% Reading HCA results from R %%%%%%%%%%%%%%%%
order = loc_read_order(ordername);

% Now deleting order files:
if strcmp(computer,'i386-pc-linux-gnu')
  str = ['rm ' ordername ];
  system(str);
  str2 = ['rm ' comfile ];
%  system(str2)
else
  str = ['del ' ordername];
  system(str);
  str2 = ['del ' comfile ];
  system(str2);
end;

T = loc_read_merge_file(dendrogram);
% Now deleting dendrogram files:
if strcmp(computer,'i386-pc-linux-gnu')
  str = ['rm ' dendrogram];
  system(str);
else
  str = ['del ' dendrogram];
  system(str);
end;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%% dendrogram procedure %%%%%%%%%%%%%%%%%%%%%%%
if opt(1) == 1,
  [nn,mm] = size(T);
  clf
  for j = 1:nn
    [L(j),R(j)] = loc_T_trav3(T,order,j);
  end;
  maxy = max(T(:,4))*1.3;
  miny = max(T(:,4))*0.1;
  axax = [0 (nn+2) -miny maxy];
  axis(axax)
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%% local files %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [] = loc_ascii_dump(filename,X)
% [] = loc_ascii_dump(file,X)
% Since Octave is stupid enough not to follow the matlab syntax
% for save, I have to spend extra effort to make a function that
% produce an output file with a similar type of output. The output
% from the Octave save -ascii command adds a lot of extra trash on
% top which we would have to remove later. Arrgggh!!!!

pr = 10;

fid = fopen(filename,'w');

str0 = '8';

[n,m]=size(X);

str = ['%' str0 '.' int2str(pr) 'f'];
for i = 1:n
 for j = 1:m-1
  fprintf(fid,[str ' ' ],X(i,j));
 end;
 ux = [str '  \n'];
 fprintf(fid,ux,X(i,m));
end;
  
fclose(fid);


function [] = loc_write_to_R(comfile,datafile,method,metric,opt,dendrogram,ordername)
% [] = loc_write_to_R(comfile,datafile,method,metric,opt,dendrogram,ordername)
% Writes a command file (comfile) which reads 
% from the datafile (datafile)
% if opt(2) == 1, we want to run agnes with a DISSIMILARITY matrix

fid = fopen(comfile,'w');

str1 = ['metric = "' metric '"'];
str2 = ['method = "' method '"'];
str2_1 = ['diss = inherits(as.dist(X), "dist")'];

if opt(2) == 0,
  str3 = ['den <- agnes(X,', str1, ',', str2, ')\n'];
else
  str3 = ['den <- agnes(as.dist(X),', str2_1, ',', str1, ',', str2, ')\n'];  
end;

str4 = ['newmerge <- data.frame(den$merge,sort(den$height))\n'];
% In R we need to extract information from the dendrogram object
% into separate files:

%extr_str1 = 'merge.dat';
extr_str1 = dendrogram;
extr_str2 = 'height.dat';
%extr_str3 = 'order.dat';
extr_str3 = ordername;
extr_str4 = 'ac.dat';
extr_str5 = 'diss.dat';

%fprintf(fid,['X <- read.table("',datafile,'", header = TRUE)\n']);
fprintf(fid,['X <- read.table("',datafile,'")\n']);

fprintf(fid,['library(cluster)\n']);

fprintf(fid,str3);
fprintf(fid,str4);
%fprintf(fid,['plot(den,which.plots=2)\n']);

fprintf(fid,['write.table(newmerge, file = "',extr_str1, '")\n']);
%fprintf(fid,['write.table(den$height, file = "',extr_str2, '")\n']);
fprintf(fid,['write.table(den$order, file = "',extr_str3, '")\n']);
fprintf(fid,['write.table(den$ac, file = "',extr_str4, '")\n']);
%fprintf(fid,['write.table(den$diss, file = "',extr_str5, '")\n']);

fclose(fid)

%%%%%%%%%% reader functions %%%%%%%%%%%%%%

function T = loc_read_merge_file(filename)
% T = loc_read_merge_file(filename)

T = zeros(1,4);
fid = fopen(filename,'r');
i = 1;
% First we read a line that is not supposed to be kept
linein = fgetl(fid);
linein = fgetl(fid);
if strcmp(computer,'i386-pc-linux-gnu')
  cmpos = 1;
else
  cmpos = 2;
end;

while isstr(linein),
  w=wordfind(linein);
  T(i,1) = i;
  if cmpos == 1,
    T(i,1+1) = str2num(w{2});
    T(i,2+1) = str2num(w{3});
    T(i,3+1) = str2num(w{4});
  else
    T(i,1+1) = str2num(w{2}');
    T(i,2+1) = str2num(w{3}');
    T(i,3+1) = str2num(w{4}');
  end;
  linein = fgetl(fid);
  i=i+1;
end;
fclose(fid);

%% now that we have the T matrix we need to extend it with one
%extra column which contains the x coordinate of every cluster

function order = loc_read_order(filename)
% order = loc_read_order(filename)
% Reads the agnes order information. This is the
% left to right order in which the different objects
% are plotted in a dendrogram

order = zeros(1,2);
fid = fopen(filename,'r');
i = 1;
% First we read a line that is not supposed to be kept
linein = fgetl(fid);
linein = fgetl(fid);

if strcmp(computer,'i386-pc-linux-gnu')
  cmpos = 1;
else
  cmpos = 2;
end;


while isstr(linein),
  w=wordfind(linein);
  order(i,1) = i;
  if cmpos == 1
    order(i,1+1) = str2num(w{2});
  else
    order(i,1+1) = str2num(w{2}');
  end;
  linein = fgetl(fid);
  i=i+1;
end;
fclose(fid)


function [Li,Ri] = loc_T_trav3(T,order,i)
%  [Li,Ri] = loc_T_trav3(T,order,i)
% The function traverses the T matrix containing info
% on the different clusters and returns the x-axis position
% for a cluster
%
% INPUT VARIABLES
%
% T                = matrix containing clusters. Each row a cluster
%                    col.1=idx, col.2=left branch, col.3= right
%                    branch, col.4=height
% order            = object no. appearing from left to right in
%                    dendrogram
% i                = index of highest cluster 
%
% OUTPUT VARIABLES
%
% Li,Ri           = left and right branch x-positions of cluster i
%
% Copyright (C), B.K. Alsberg, bka@phys.chem.ntnu.no
% Laast update: 8/5-2004
%


offset = (max(T(:,4)))*0.03;

left =  T(i,2);
right = T(i,3);

if left < 0 & right < 0,
  % we have a leaf cluster where left and right branches 
  % only contain one object each. Here we terminate the 
  % recursion
  pleft = abs(left);
  pright = abs(right);
  Li = find(order(:,2)==pleft);
  Ri = find(order(:,2)==pright);
  x = Li + (Ri-Li)/2;
  % Now we do some plotting:
  % first the horizontal line:
  line([Li Ri],[T(i,4) T(i,4)]); hold on
  % Now we have two vertical lines to the bottom:
  % First the left branch:
  line([Li Li],[T(i,4) 0])
  text(Li,-offset,int2str(pleft))
  % The the right branch
  line([Ri Ri],[T(i,4) 0])
  text(Ri,-offset,int2str(pright))
  
else
  if left > 0,%i.e. we have a cluster in left branch
    [leftx,rightx] = loc_T_trav3(T,order,left);
    Li = leftx + (rightx-leftx)/2;
    line([Li Li],[T(i,4) T(left,4)]); hold on    
  else
    % we have an object
    pleft = abs(left);
    Li = find(order(:,2)==pleft);
    line([Li Li],[T(i,4) 0]); hold on
    text(Li,-offset,int2str(pleft))    
  end;
  if right > 0,%i.e. we have a cluster in right branch
    [leftx,rightx] = loc_T_trav3(T,order,right);
    Ri = leftx + (rightx-leftx)/2;
    line([Ri Ri],[T(i,4) T(right,4)]); hold on
  else
    % we have an object
    pright = abs(right);
    Ri = find(order(:,2)==pright);
    line([Ri Ri],[T(i,4) 0]); hold on
    text(Ri,-offset,int2str(pright))    
  end;
  % Now we do some plotting:
  % first the horizontal line:
  line([Li Ri],[T(i,4) T(i,4)]); hold on
end;


