function [Xh,Wh] = WaveletDenoiseX(X,method,wavelet,par,L)
% [Xh,Wh] = WaveletDenoiseX(X,method,wavelet,par,L)
% Perform wavelet denoising on raw data and returns filteret raw data
% (this is in contrast to WaveletDenoiseW which performs the
% operation on the wavelet coefficient matrix directly)
%
%
% X             = data matrix with N rows and M columns
%                 M = 2^J. This will also work fine when N=1
% method        = string indicating type of denoise method. Choices are:
%                 'Visu'
%                 'SURE'
%                 'Hybrid'
%                 'MinMax'
%                 'MAD'
% wavelet       = string indicating type of wavelet family to be used.
%                 Choices are:
%                 'Haar', 
%                 'Beylkin'
%                 'Coiflet'
%                 'Daubechies'
%                 'Symmlet'
%                 'Vaidyanathan'
%                 'Battle'
% par           = indicates support and vanishing moments of the
%                 wavelets. The allowed value for each wavelet
%                 family are:
%                 Haar           : discontinuous
%                 Coiflet        : 1,2,3,4 or 5.
%                 Daubechies     : 4,6,8,10,12,14,16,18 or 20
%                 Symmlet        : 4 to 10
%                 Vaidyanathan   : no moment condition
%                 Battle         : Par gives degree of spline.
%                                  Vanishing moments=Par+1
% L             = Coarsest Level of V_0;  L << J (Default=0)
%
% OUTPUT VARIABLES
%
% Xh            = raw data matrix after denoising
% Wh            = the corresponding wavelet coefficient matrixc
%                 (one for every row) after the wavelet denoising
%
%
% Copyright(C), B.K. Alsberg, The SciCraft Team, www.scicraft.org
% Last update: 8/3-06
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

if nargin < 5,
  L =0;
end;

[N,M] = size(X);

qmf = MakeONFilter(wavelet,par);

for i = 1:N
  [y,coef] = NormNoise(X(i,:),qmf);
  [Xh(i,:), Wh(i,:)]  = WaveShrink(y,method,L,qmf);
end;


