function [W,qmf] = WaveletTransform(X,wavelet,par,L)
% [W,qmf] = WaveletTransform(X,wavelet,par,L)
% Performs the Fast Wavelet Transform (FWT) on data
% with 2^J elements. The FWT is performed on each ROW in X
%
% INPUT VARIABLES
%
% X             = data matrix with N rows and M columns
%                 M = 2^J. This will also work fine when N=1
% wavelet       = string indicating type of wavelet family to be used.
%                 Choices are:
%                 'Haar', 
%                 'Beylkin'
%                 'Coiflet'
%                 'Daubechies'
%                 'Symmlet'
%                 'Vaidyanathan'
%                 'Battle'
% par           = indicates support and vanishing moments of the
%                 wavelets. The allowed value for each wavelet
%                 family are:
%                 Haar           : discontinuous
%                 Coiflet        : 1,2,3,4 or 5.
%                 Daubechies     : 4,6,8,10,12,14,16,18 or 20
%                 Symmlet        : 4 to 10
%                 Vaidyanathan   : no moment condition
%                 Battle         : Par gives degree of spline.
%                                  Vanishing moments=Par+1
% L             = Coarsest Level of V_0;  L << J (Default=0)
%
% OUTPUT VARIABLES
%
% W             = Matrix with wavelet coefficients, N rows and M
%                 columns
% qmf           = quadrature mirror filter vector which identifies
%                 the wavelet type used. Is required for the
%                 InverseWaveletTransform.m routine
%
%
% Copyright(C), B.K. Alsberg, The SciCraft Team, www.scicraft.org
% Last update: 7/3-06
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 2
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


[N,M] = size(X);

if round(loge(M,2))-loge(M,2) ~= 0,
  error('You have not M=2^J number of variables');
end

if nargin < 4,
  L=0;
end


qmf = MakeONFilter(wavelet,par);


% Calculating the wavelet coefficients for each spectrum
for i = 1:N
 [y,coef] = NormNoise(X(i,:),qmf);
 W(i,:) = FWT_PO(y,L,qmf);
end;



