function [full_model,sub_models,errors] =
crossval2(data,segment_index,aMax,dfl,signif)

%CROSSVAL.M
%crossval performs crossvalidation with plsr-decomp. Significance-testing and prediction errors are calculated from the sub-models 
%
%input: 		data - struct(X,Y)
%			segment_index - cross validation index of the partial segments (cell{nSegments})
%			aMax - max. number of components to calculate
%			dfl - flag for deflating (0/1)
%			signif - flag for significance tests (0=none,1=testing on Ws, 2=testing on Bs, 3 = both W&B)
%
%output:		full_model - struct(T,P,W,Q,B,Yhat,Wpval,Bpval)
%			sub_models - struct(T,P,W,Q,B,Yhat)
%			errors - struct(MSEP,MSEC,SEP)
%
%
%[full_model,sub_models,errors] = crossval2(data,segment_index,aMax,dfl)
%
%
% $$$ ## This file is part of SciCraft.
% $$$ ##
% $$$ ## SciCraft is free software; you can redistribute it and/or modify it
% $$$ ## under the terms of the GNU General Public License as published by
% $$$ ## the Free Software Foundation; either version 2, or (at your option)
% $$$ ## any later version.
% $$$ ##
% $$$ ## SciCraft is distributed in the hope that it will be useful, but
% $$$ ## WITHOUT ANY WARRANTY; without even the implied warranty of
% $$$ ## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
% $$$ ## General Public License for more details.
% $$$ ##
% $$$ ## You should have received a copy of the GNU General Public License
% $$$ ## along with SciCraft; see the file COPYING.  If not, write to the Free
% $$$ ## Software Foundation, 59 Temple Place - Suite 330, Boston, MA
% $$$ ## 02111-1307, USA.
% $$$ 
% $$$ ## -*- texinfo -*-

%arg check
if nargin >5
  help crossval
  error('try')
elseif nargin == 4
  signif=0;
  dfl = 1;
elseif nargin == 3
  dfl = 1;
  signif = 0;
elseif nargin ==2
  aMax = min(size(data.X))
  dfl = 1;
  signif = 0;
      
elseif nargin == 1
  dfl = 1;
  aMax = min(size(data.X));
  segment_index = makeXval('full')
elseif nargin==5
disp('all args passed')
else
error('Usage: [full_model,sub_models,errors] = crossval2(data,segment_index,aMax,dfl)')
  
end

s = segment_index;
nSegments = size(s,2);

%deflating data?
if dfl==1
  disp('Deflating data...')
  [X,Vdef] = deflate(data.X);
  Y = data.Y;
  disp('Deflating data complete.');
else
  X = data.X;
  Y = data.Y;
end


[nSamples,nVariablesX] = size(X);
[nSamples,nVariablesY] = size(Y);
%warning(num2str(nVariablesY))


% pls1 or 2 ? (just for testing puroses...)
if nVariablesY > 1.54
  pls2 = 1;
else
  pls2 = 0;
end


%% NB nSegments > aOpt (required in jackknife-W)

%initialize variables
EP = zeros(nSamples,nVariablesY,aMax+1); %prediction error
Yhat = zeros(nSamples,nVariablesY,aMax); %Y-estimates
W = zeros(nVariablesX,aMax,nSegments);
P = zeros(nVariablesX,aMax,nSegments);
T = zeros(nSamples,aMax,nSegments);
Q = zeros(nVariablesY,aMax,nSegments);
B = zeros(nVariablesX,nVariablesY,aMax,nSegments);


%crossvalidation loop
for i=1:nSegments %for each x-validation segment
  keep = logical(s{i}); %samples used in model
  nSampSub = length(keep)- sum(keep);
  Xcal = X(keep,:);
  Ycal = Y(keep,:);
  Xtrue = X(~keep,:);
  Ytrue = Y(~keep,:);
  %center
  [Xcal_m,Ycal_m,mx,my] = centerXY(Xcal,Ycal);
  Xtrue = Xtrue - ones(nSampSub,1)*mx;
  Ytrue = Ytrue - ones(nSampSub,1)*my;
  
  EP(~keep,:,1) = Ytrue; %pc 0
  
  % model component loop 
  for a = 1:aMax
    [W(:,1:a,i),T(keep,1:a,i),P(:,1:a,i),Q(:,1:a,i),B(:,:,a,i)] = plsr(Xcal_m,Ycal_m,a,0);
    that = Xtrue*W(:,1:a,i); %expand t s for comparison with full model
    T(~keep,1:a,i) = that;
    Yhat(~keep,:,a) = Xtrue*B(:,:,a,i);
    EP(~keep,:,a+1) = Yhat(~keep,:,a)-Ytrue;
  end % for a
  
end % for segment i

SEP = EP.^2; %squared error of prediction
MSEP = mean(SEP,1); %mean squared error of prediction
SEP = reshape(SEP,nSamples*nVariablesY,aMax+1);

%store
errors.MSEP = reshape(MSEP,nVariablesY,aMax+1);
errors.SEP = SEP;
sub_models.Yhat = Yhat;  %Warning! Yhat after aMax PCs, not aOpt
sub_models.B = B;
disp('size of sub_models b');
size(sub_models.B);

sub_models.W = W;
sub_models.P = P;
sub_models.Q = Q;
sub_models.T = T;


%%%%%%%%%%%%%%%% CALIBRATION MODEL %%%%%%%%%%%%%%%%%%%%%%

%initialize variables
W = zeros(nVariablesX,aMax);
P = zeros(nVariablesX,aMax);
T = zeros(nSamples,aMax);
Q = zeros(nVariablesY,aMax);
Bc = zeros(nVariablesX,nVariablesY,aMax);
YHATc = zeros(nSamples,nVariablesY,aMax);
MSEC = zeros(aMax+1,nVariablesY); %mean square error of calibration
MSEC(1,:) = var(Y); %PC 0

%center
[X_m,Y_m,mx,my] = centerXY(X,Y);

for a = 1:aMax
  [W(:,1:a),T(:,1:a),P(:,1:a),Q(:,1:a),Bc(:,:,a)] = plsr(X_m,Y_m,a,0);
  YHATc(:,:,a) = X_m*Bc(:,:,a);
  EC = YHATc(:,:,a)-Y_m;
  MSEC(a+1,:) = sum(EC.^2)/(nSamples-1-a);
end %for

%store 
full_model.W = W;
full_model.T = T;
full_model.P = P;
full_model.Q = Q;
full_model.B = Bc;
% adding (bka)

full_model.mx = mx;
full_model.my = my;

full_model.Yhat = YHATc; 
errors.MSEC = MSEC';



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% SIGNIFICANCE TESTING %%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%% INFLATE AND ROTATE SUBMODELS %%%%%%%
if aMax >1 % only need to rotate if more than one component
  disp('Rotating...');
  [sub_models] = rotateSub(sub_models,full_model);
  disp('Rotation complete');
end

if dfl
disp('Inflating...');
[sub_models,full_model] = inflate(Vdef,full_model,sub_models,aMax);	
disp('Inflating complete.');
end

% jackknife on W's
%h=figure, plot(errors.MSEP','-o')
%hold on, plot(errors.MSEC')
%set(gca,'YLim',[errors.MSEP(1,1)])
%aMax = input('Number of components to use? : ')
%close(h)

%%% Optimal number of components ... %%%%%
if size(errors.MSEP,1)>1
  RMSEP = sqrt(mean(errors.MSEP));
else
  RMSEP = sqrt(errors.MSEP)
end

[minVal,aOpt]=min(RMSEP);
aOpt = aOpt-1;
%keyboard
if aOpt<1
disp('Aopt is at zeros comp.!. Setting aOpt = 1 ...');
aOpt = 1;
end
full_model.aOpt = aOpt;
if (signif == 1 | signif ==3)
disp('Significance Ws');
[full_model.Wpval] = jackknifeW(data,sub_models,full_model,aOpt,dfl);
end

% jacknife on B's
if (signif == 2 | signif ==3)
  disp('Significance Bs');
  %warning(num2str(pls2))
  [full_model.Bpval] = jackknifeB(full_model,sub_models,aOpt,pls2);
end
full_model.aOpt = aOpt; 
disp('Done.')


