function [pval, Tsq] = hotelling_test_jk_matlab (x, m)
% $$$ ## Copyright (C) 1996, 1997  Kurt Hornik
% $$$ ##Modified by endre anderssen to use correct degrees of freedom to
% $$$ ##when used in a jack knife context
% $$$ ##
% $$$ ##
% $$$ ## This file is part of Octave.
% $$$ ##
% $$$ ## Octave is free software; you can redistribute it and/or modify it
% $$$ ## under the terms of the GNU General Public License as published by
% $$$ ## the Free Software Foundation; either version 2, or (at your option)
% $$$ ## any later version.
% $$$ ##
% $$$ ## Octave is distributed in the hope that it will be useful, but
% $$$ ## WITHOUT ANY WARRANTY; without even the implied warranty of
% $$$ ## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
% $$$ ## General Public License  more details.
% $$$ ##
% $$$ ## You should have received a copy of the GNU General Public License
% $$$ ## along with Octave; see the file COPYING.  If not, write to the Free
% $$$ ## Software Foundation, 59 Temple Place - Suite 330, Boston, MA
% $$$ ## 02111-1307, USA.
% $$$ 
% $$$ ## -*- texinfo -*-
% $$$ ## @deftypefn {Function File} {[@var{pval}, @var{tsq}] =} hotelling_test (@var{x}, @var{m})
% $$$ ## For a sample @var{x} from a multivariate normal distribution with unknown
% $$$ ## mean and covariance matrix, test the null hypothesis that @code{mean
% $$$ ## (@var{x}) == @var{m}}.
% $$$ ##
% $$$ ## Hotelling's T^2 is returned in @var{tsq}.  Under the null,
% $$$ ## @math{(n-p) T^2 / (p(n-1))} has an F distribution with @math{p} and
% $$$ ## @math{n-p} degrees of freedom, where @math{n} and @math{p} are the
% $$$ ## numbers of samples and variables, respectively.
% $$$ ##
% $$$ ## The p-value of the test is returned in @var{pval}.
% $$$ ##
% $$$ ## If no output argument is given, the p-value of the test is displayed.
% $$$ ## @end deftypefn
% $$$ 
% $$$ ## Author: KH <Kurt.Hornik@ci.tuwien.ac.at>
% $$$ ## Description: Test  mean of a multivariate normal

if (nargin ~= 2)
  disp('usage:  hotelling_test (x, m)');
end %if
[n,p] = size(x);
[k,l] = size(m);
if ((n==1) | (p==1))
  if (~(k==1 & l==1))
    error('hotelling_test: if x is a vector, m must be a scalar');
  end % 2.if
  n = length (x);
  p = 1;
elseif (n>1&p>1)
  if (n <= p)
    error('hotelling_test: x must have more rows than columns');
  end %2.if
  if (length (m) == p)
    m = reshape (m, 1, p);
  else
    error('hotelling_test: if x is a matrix, m must be a vector of length columns (x)');
  end %2.if
else
  error('hotelling_test: x must be a matrix or vector');
end %1.if

d    = mean (x) - m;
Tsq  = (1/(n-1)) * d * (cov (x) \ d');
pval = 1 - f_cdf ((n-p) * Tsq / (p * (n-1)), p, n-p);

if (nargout == 0)
  printf('  pval: %g\n', pval);
end

